import random
import numpy as np
import statistics as stats
import os
import sys
from scipy.sparse import coo_matrix


CHR = ['X','Xd','Y']

nGen = 100000
init_pop = 50
q = 0.1 #initial drive frequency
c = 1
d = 1 #drive strength
b = 2.4

alpha = 10**(-2.4)

h = 0.1
sf = 1.0
sm = 0

fm = 2 #female and male max matings
mm = 20

nSims = 5

wXX = 1
wXdX = 1 - h*sf
wXdXd = 1 - sf
wXY = 1
wXdY = 1 - sm

mean_pops = []
extinctGens = []

allExtGens = []
drivePresent = []

for sim in range(nSims):
    XXvals = []
    XdXvals  =[]
    XdXdvals = []
    XYvals = []
    XdYvals = []
    SRvals=[]
    extinctGen = 0
    pops = []


    nXX = int(((1-q)**2)*init_pop/2)
    nXdX = int(2*q*(1-q)*init_pop/2)
    nXdXd = int((q**2)*init_pop/2)
    nXY = int(init_pop*(1-q)/2)
    nXdY = int(init_pop*q/2)


    for i in range(nGen):
        newXX = 0
        newXdX = 0
        newXdXd = 0
        newXY = 0
        newXdY = 0

        ftot = nXX + nXdX + nXdXd
        mtot = nXY + nXdY

        f_id_list = [i for i in range(ftot)]
        m_id_list = [i for i in range(mtot)]

        num_f_matings = [fm for i in range(ftot)]
        num_m_matings = [mm for i in range(mtot)]

        f_weights = [1/ftot for i in range(ftot)] # equal probability of being selected
        m_weights = [1/ftot for i in range(mtot)]

        s = int(min(np.sum(num_f_matings),np.sum(num_m_matings)))

        f_matings = []
        m_matings = []


        for j in range(s): # sample individuals at random until they use their allotted matings. then we set their probability of being chosen to 0.
            f = random.choices(f_id_list,f_weights)[0]
            m = random.choices(m_id_list,m_weights)[0]
            f_matings.append(f)
            m_matings.append(m)

            num_f_matings[f] -= 1
            if num_f_matings[f] == 0:
                f_weights[f] = 0

            num_m_matings[m] -= 1
            if num_m_matings[m] == 0:
                m_weights[m] = 0

        v = [1 for i in range(s)]

        M = coo_matrix((v,(f_matings,m_matings)),shape=(ftot,mtot)) #create a matrix to track the matings each female has received
        M.sum_duplicates()

        S = []

        for j in range(nXY): #create a matrix where row i is the sperm male i gives in a mating
            S.append([0.5,0,0.5])
        for j in range(nXdY):
            S.append([0,c*(1+d)/2,c*(1-d)/2])

        S = np.array(S)
        sperm_mat = M@S # this then gives the sperm carried by each female

        for j in range(ftot):
            births = int(np.floor(b) + np.random.binomial(1, b-np.floor(b), 1)[0])
            for k in range(births):
                try:
                    pat_chr = random.choices(CHR,weights=sperm_mat[j,:])[0] #sample paternal chromosome according to how much of each the female is carrying
                    if j < nXX: # wild type female
                        mat_chr = 'X'
                    elif j > nXX - 1 and j < nXX + nXdX: #heterozygous female
                        mat_chr = random.choices(CHR,weights=[0.5,0.5,0])[0]
                    else: #homozygous
                        mat_chr = 'Xd'

                    if mat_chr == 'X':
                        if pat_chr == 'X':
                            newXX += 1
                        elif pat_chr == 'Xd':
                            newXdX += 1
                        elif pat_chr == 'Y':
                            newXY += 1

                    elif mat_chr == 'Xd':
                        if pat_chr == 'X':
                            newXdX += 1
                        elif pat_chr == 'Xd':
                            newXdXd += 1
                        elif pat_chr == 'Y':
                            newXdY += 1
                except:
                    print('')

        tot= ftot+mtot

        previousGen = [nXX,nXdX,nXdXd,nXY,nXdY]

        survivalProb = max(1-(tot*alpha),0)

        nXX = np.random.binomial(newXX,survivalProb*wXX)
        nXdX = np.random.binomial(newXdX,survivalProb*wXdX)
        nXdXd = np.random.binomial(newXdXd,survivalProb*wXdXd)
        nXY = np.random.binomial(newXY,survivalProb*wXY)
        nXdY = np.random.binomial(newXdY,survivalProb*wXdY)

        distn = [nXX,nXdX,nXdXd,nXY,nXdY]
        print(distn)


        if i > 100: #allow a burnin period of 100 generations
            pops.append(np.sum(distn))
            XXvals.append(nXX)
            XdXvals.append(nXdX)
            XdXdvals.append(nXdXd)
            XYvals.append(nXY)
            XdYvals.append(nXdY)
            try:
                SRvals.append((nXX+nXdX+nXdXd)/(nXY+nXdY))
            except:
                print('')



        if distn[0] + distn[1] + distn[2] == 0 or distn[3] + distn[4] == 0: # at extinction, check whether or not drive was present in the preceding generation
            if (previousGen[1] + previousGen[2] + previousGen[4])/np.sum(previousGen) > 0.2:
                drivePresent.append(True)
            else:
                drivePresent.append(False)
            print('extinction generation ', i+1)
            extinctGen = i+1
            extinctGens.append(extinctGen)
            break

        if i == nGen - 1: # in this case the population persisted
            extinctGen = i+1
            extinctGens.append(extinctGen)
            if (previousGen[1] + previousGen[2] + previousGen[4])/np.sum(previousGen) > 0.2:
                drivePresent.append(True)
            else:
                drivePresent.append(False)

        if pops:
            mean_pops.append(np.mean(pops))



        newXX = 0
        newXdX = 0
        newXdXd = 0
        newXY = 0
        newXdY = 0
