(* Content-type: application/vnd.wolfram.mathematica *)

(*** Wolfram Notebook File ***)
(* http://www.wolfram.com/nb *)

(* CreatedBy='Mathematica 12.0' *)

(*CacheID: 234*)
(* Internal cache information:
NotebookFileLineBreakTest
NotebookFileLineBreakTest
NotebookDataPosition[       158,          7]
NotebookDataLength[   1221691,      26708]
NotebookOptionsPosition[   1188479,      26205]
NotebookOutlinePosition[   1188840,      26221]
CellTagsIndexPosition[   1188797,      26218]
WindowFrame->Normal*)

(* Beginning of Notebook Content *)
Notebook[{

Cell[CellGroupData[{
Cell["\<\
X-linked meiotic drive can boost population size and persistence\
\>", "Title",
 CellChangeTimes->{{3.813651645117669*^9, 
  3.813651659263967*^9}},ExpressionUUID->"f963561f-d28a-4a18-ba87-\
157215e0ae56"],

Cell["Carl Mackintosh, Andrew Pomiankowski, and Michael F Scott", "Subtitle",
 CellChangeTimes->{{3.813651663235971*^9, 
  3.8136516773311377`*^9}},ExpressionUUID->"f0b9a4d6-bcd4-48f7-9312-\
c0514f50288c"],

Cell[CellGroupData[{

Cell["Overview", "Section",
 CellChangeTimes->{{3.814018860757616*^9, 
  3.8140188621077747`*^9}},ExpressionUUID->"4310ba37-7551-4d2a-a43b-\
1de62e5d9dda"],

Cell[TextData[{
 "This notebook analyses the eco-evolutionary dynamics of X-linked meiotic \
drive. \n\nIn this file, we assume that the birth rate declines linearly with \
population size. This assumption means that increased birth rates will \
increase the population size, which we expect to be generally true. Different \
assumptions about density-dependence may mean that increases in birth rate \
can actually decrease population size. We explore one such possibility in \
another Supplementary notebook. Further supplementary material contains \
python code used to conduct stochastic simulations of population size and \
persistence when there is X-linked meiotic drive. \n\nSince the spread of \
meiotic drive alleles may be affected the degree of polyandry via sperm \
competition, we investigate cases where females mate once (",
 Cell[BoxData[
  FormBox[
   RowBox[{
    SubscriptBox["\[Lambda]", "f"], "=", "1"}], TraditionalForm]],
  FormatType->"TraditionalForm",ExpressionUUID->
  "ea4cc969-b2d2-4458-82e2-93ec0ae56a0c"],
 "), twice (",
 Cell[BoxData[
  FormBox[
   RowBox[{
    SubscriptBox["\[Lambda]", "f"], "=", "2"}], TraditionalForm]],
  FormatType->"TraditionalForm",ExpressionUUID->
  "a9e0c51e-5318-4fa5-8bd7-f6f424049cc2"],
 "), or many times (",
 Cell[BoxData[
  FormBox[
   RowBox[{
    SubscriptBox["\[Lambda]", "f"], "\[Rule]", "\[Infinity]"}], 
   TraditionalForm]],
  FormatType->"TraditionalForm",ExpressionUUID->
  "f56877d5-7300-4f39-8dc2-bf574cd22701"],
 "), each of which is presented in a separate section. The first section \
(Common Notation and Substitutions), should be entered before evaluating any \
of these sections. "
}], "Text",
 CellChangeTimes->{{3.813655240250414*^9, 3.813655477749702*^9}, {
  3.813655517394024*^9, 3.813655929474389*^9}, {3.813928023019766*^9, 
  3.813928026347568*^9}, {3.814018820664456*^9, 3.814018843545726*^9}, {
  3.814018936514036*^9, 3.814018960689776*^9}, {3.8140920410257273`*^9, 
  3.814092050969368*^9}},ExpressionUUID->"49957b2d-c3fb-449d-9a2a-\
f23d490ae09c"]
}, Closed]],

Cell[CellGroupData[{

Cell["Common Notation and Substitutions [ENTER FIRST] ", "Section",
 CellChangeTimes->{{3.814018923386217*^9, 3.8140189329701023`*^9}, {
  3.814019466209839*^9, 
  3.8140194721274967`*^9}},ExpressionUUID->"ecb2965d-0ae6-4644-846b-\
a4438bf10635"],

Cell[TextData[{
 "We track adult (diploid) genotypes at an X linked locus that experiences \
meiotic drive in males. There are therefore 5 genotypes: 3 female genotypes \
and 2 male genotypes. In this notebook we use upper case D for the wildtype \
allele (which could also be called ",
 Cell[BoxData[
  FormBox[
   SubscriptBox["S", "t"], TraditionalForm]],
  FormatType->"TraditionalForm",ExpressionUUID->
  "f53f08b8-c8c0-435c-8de5-230bfe759d83"],
 " or ",
 Cell[BoxData[
  FormBox["X", TraditionalForm]],
  FormatType->"TraditionalForm",ExpressionUUID->
  "c19ad89e-3e7a-4f12-9b87-df63c4e9a53d"],
 ") and lower case d to give the meiotic drive allele (which could also be \
called ",
 Cell[BoxData[
  FormBox[
   SubscriptBox["S", "r"], TraditionalForm]],
  FormatType->"TraditionalForm",ExpressionUUID->
  "eb1bf42b-8c4e-4210-8af6-2517b35758cc"],
 " or ",
 Cell[BoxData[
  FormBox[
   SubscriptBox["X", "d"], TraditionalForm]],
  FormatType->"TraditionalForm",ExpressionUUID->
  "3b0239e2-6199-4506-95ac-3489558a9fce"],
 "). The 5 adult genotypes are therefore\nfemales:\n",
 Cell[BoxData[
  FormBox[
   SubscriptBox["S", "t"], TraditionalForm]],
  FormatType->"TraditionalForm",ExpressionUUID->
  "f8bffd34-98f6-4957-aea1-ab47a51cb6aa"],
 Cell[BoxData[
  FormBox[
   SubscriptBox["S", "t"], TraditionalForm]],
  FormatType->"TraditionalForm",ExpressionUUID->
  "5b409ad5-ce4b-4341-baff-ce64c7bf91c9"],
 " = ",
 Cell[BoxData[
  FormBox["X", TraditionalForm]],ExpressionUUID->
  "de1bdf2a-1743-4a84-aaf4-c68f30aaec73"],
 Cell[BoxData[
  FormBox["X", TraditionalForm]],ExpressionUUID->
  "8fb9ddf6-101f-42dd-921e-f1a7f460be25"],
 "= DDf\n",
 Cell[BoxData[
  FormBox[
   SubscriptBox["S", "r"], TraditionalForm]],
  FormatType->"TraditionalForm",ExpressionUUID->
  "6c914f89-983e-46ae-8f78-6c14642c0573"],
 Cell[BoxData[
  FormBox[
   SubscriptBox["S", "t"], TraditionalForm]],
  FormatType->"TraditionalForm",ExpressionUUID->
  "84638b1b-f1d1-49dc-bbe8-76a34fa923a1"],
 " = ",
 Cell[BoxData[
  FormBox[
   SubscriptBox["X", "d"], TraditionalForm]],ExpressionUUID->
  "ca1ad55f-bc21-4ba8-b09d-b820773d326f"],
 Cell[BoxData[
  FormBox["X", TraditionalForm]],ExpressionUUID->
  "ebbbe5e3-af83-428c-a314-0e882084065a"],
 " = Ddf\n",
 Cell[BoxData[
  FormBox[
   SubscriptBox["S", "r"], TraditionalForm]],
  FormatType->"TraditionalForm",ExpressionUUID->
  "f44b26a6-6782-4333-9f0d-14ed8f434e42"],
 Cell[BoxData[
  FormBox[
   SubscriptBox["S", "r"], TraditionalForm]],
  FormatType->"TraditionalForm",ExpressionUUID->
  "8419305f-35ec-4983-aa7f-a9e294302bd7"],
 " = ",
 Cell[BoxData[
  FormBox[
   SubscriptBox["X", "d"], TraditionalForm]],ExpressionUUID->
  "43e8627d-683d-4265-8abc-e93abbb5ad4d"],
 Cell[BoxData[
  FormBox[
   SubscriptBox["X", "d"], TraditionalForm]],ExpressionUUID->
  "081e7bb1-cdcf-4c8f-b563-2766b0cdef2b"],
 " = ddf\nmales:\n",
 Cell[BoxData[
  FormBox[
   SubscriptBox["S", "t"], TraditionalForm]],ExpressionUUID->
  "8be3796c-1662-49ca-a4ed-ec96184be421"],
 "Y = ",
 Cell[BoxData[
  FormBox["X", TraditionalForm]],ExpressionUUID->
  "618ad4d4-c302-43d4-9e6f-f2cd95858eae"],
 Cell[BoxData[
  FormBox["Y", TraditionalForm]],ExpressionUUID->
  "36c8f206-87e7-4bb2-9c09-f93c8db1cf01"],
 " = Dm\n",
 Cell[BoxData[
  FormBox[
   SubscriptBox["S", "r"], TraditionalForm]],ExpressionUUID->
  "a13b9d35-bcd3-4906-81b3-662b6cf04851"],
 "Y = ",
 Cell[BoxData[
  FormBox[
   SubscriptBox["X", "d"], TraditionalForm]],ExpressionUUID->
  "8d7bf35b-8b46-411d-91b3-824cc4a87a9b"],
 Cell[BoxData[
  FormBox["Y", TraditionalForm]],ExpressionUUID->
  "2568c88d-6f0e-4b62-af76-bd13551fe287"],
 " = dm\nsuch that\nijf is the density of females with genotype ij \nim is \
the density of males with genotype i"
}], "Text",
 CellChangeTimes->{{3.8140190605131702`*^9, 3.814019392288312*^9}, {
  3.8140757952060537`*^9, 3.8140758541839437`*^9}, {3.814077180589332*^9, 
  3.81407720443826*^9}, {3.81407730059503*^9, 
  3.814077315912006*^9}},ExpressionUUID->"bbf45c70-2c7b-4d53-8e72-\
65085059f88a"],

Cell[TextData[{
 "We assume that the X-linked meiotic drive allele biases transmission in \
males, such that a ",
 Cell[BoxData[
  RowBox[{"(", 
   FractionBox[
    RowBox[{"(", 
     RowBox[{"1", "+", "\[Delta]"}], ")"}], "2"], ")"}]],
  CellChangeTimes->{{3.7814428720012417`*^9, 3.781442894240843*^9}, {
   3.814008329923299*^9, 3.8140083738240767`*^9}, {3.814019442825118*^9, 
   3.814019443216568*^9}, {3.81401966596344*^9, 3.814019666468874*^9}, {
   3.81407591820485*^9, 3.814075935309269*^9}, {3.814075978592771*^9, 
   3.814075985638269*^9}, {3.814079326819665*^9, 3.8140793341532917`*^9}, {
   3.8140794393433743`*^9, 3.814079615488195*^9}, {3.8140796473810863`*^9, 
   3.8140798375136547`*^9}},ExpressionUUID->
  "cc635356-31e0-4484-b8e6-f706222a3908"],
 "fraction of gametes produced by a meiotic drive male (dm) carry the drive \
allele (d). Furthermore, we assume that the ejaculate size of meiotic drive \
males, relative to wildtype males, is c. "
}], "Text",
 CellChangeTimes->{{3.814080029254443*^9, 
  3.8140801640639563`*^9}},ExpressionUUID->"159b7ed4-0e7f-402a-a842-\
a76148633e12"],

Cell["\<\
We  use the following substitutions to convert between these genotype \
densities and the total density of males (maleNum) and females (femaleNum), \
and the frequency of the drive allele among males (pm) and females (pf), and \
the inbreeding coefficient (F). These quantities are defined by \
substitutionEquations, subeq can be used to convert from the genotype \
densities into male/female densities and allele frequencies. \
\>", "Text",
 CellChangeTimes->{{3.814077206472991*^9, 3.814077212502678*^9}, {
  3.814077289440152*^9, 3.814077399069392*^9}, {3.81407752412192*^9, 
  3.814077623571116*^9}},ExpressionUUID->"eecf2aaf-f731-49f3-bfda-\
0e917861cdd7"],

Cell[CellGroupData[{

Cell[BoxData[{
 RowBox[{
  RowBox[{"substitutionEquations", "=", 
   RowBox[{"{", "\[IndentingNewLine]", 
    RowBox[{
     RowBox[{"femaleNum", "==", 
      RowBox[{"DDf", "+", "Ddf", "+", "ddf"}]}], ",", "\[IndentingNewLine]", 
     RowBox[{"maleNum", "\[Equal]", 
      RowBox[{"(", 
       RowBox[{"dm", "+", "Dm"}], ")"}]}], ",", "\[IndentingNewLine]", 
     RowBox[{"pm", "==", 
      FractionBox["dm", 
       RowBox[{"dm", "+", "Dm"}]]}], ",", "\[IndentingNewLine]", 
     RowBox[{
      RowBox[{"(", 
       RowBox[{
        RowBox[{
         RowBox[{
          RowBox[{"(", 
           RowBox[{"1", "-", "pf"}], ")"}], "^", "2"}], 
         RowBox[{"(", 
          RowBox[{"1", "-", "F"}], ")"}]}], "+", 
        RowBox[{"F", 
         RowBox[{"(", 
          RowBox[{"1", "-", "pf"}], ")"}]}]}], ")"}], "==", 
      FractionBox["DDf", 
       RowBox[{"DDf", "+", "Ddf", "+", "ddf"}]]}], ",", "\[IndentingNewLine]", 
     RowBox[{
      RowBox[{"2", "pf", 
       RowBox[{"(", 
        RowBox[{"1", "-", "pf"}], ")"}], 
       RowBox[{"(", 
        RowBox[{"1", "-", "F"}], ")"}]}], "==", 
      FractionBox["Ddf", 
       RowBox[{"DDf", "+", "Ddf", "+", "ddf"}]]}], ",", "\[IndentingNewLine]", 
     RowBox[{
      RowBox[{
       RowBox[{
        RowBox[{
         RowBox[{"(", "pf", ")"}], "^", "2"}], 
        RowBox[{"(", 
         RowBox[{"1", "-", "F"}], ")"}]}], "+", 
       RowBox[{"F", " ", "pf"}]}], "==", 
      FractionBox["ddf", 
       RowBox[{"DDf", "+", "Ddf", "+", "ddf"}]]}]}], "\[IndentingNewLine]", 
    "}"}]}], ";"}], "\[IndentingNewLine]", 
 RowBox[{"subeq", "=", 
  RowBox[{"Simplify", "[", 
   RowBox[{"Solve", "[", 
    RowBox[{"substitutionEquations", ",", 
     RowBox[{"{", 
      RowBox[{"dm", ",", "Dm", ",", "DDf", ",", "Ddf", ",", "ddf"}], "}"}]}], 
    "]"}], "]"}]}]}], "Input",
 CellChangeTimes->{{3.814077223260071*^9, 3.8140772805761766`*^9}, {
  3.814077501177114*^9, 3.814077520478444*^9}},
 CellLabel->
  "In[1269]:=",ExpressionUUID->"933bbd05-fc28-431a-9273-1f317681a387"],

Cell[BoxData[
 RowBox[{"{", 
  RowBox[{"{", 
   RowBox[{
    RowBox[{"dm", "\[Rule]", 
     RowBox[{"maleNum", " ", "pm"}]}], ",", 
    RowBox[{"Dm", "\[Rule]", 
     RowBox[{"maleNum", "-", 
      RowBox[{"maleNum", " ", "pm"}]}]}], ",", 
    RowBox[{"DDf", "\[Rule]", 
     RowBox[{
      RowBox[{"-", "femaleNum"}], " ", 
      RowBox[{"(", 
       RowBox[{
        RowBox[{"-", "1"}], "+", "pf"}], ")"}], " ", 
      RowBox[{"(", 
       RowBox[{"1", "+", 
        RowBox[{
         RowBox[{"(", 
          RowBox[{
           RowBox[{"-", "1"}], "+", "F"}], ")"}], " ", "pf"}]}], ")"}]}]}], 
    ",", 
    RowBox[{"Ddf", "\[Rule]", 
     RowBox[{"2", " ", 
      RowBox[{"(", 
       RowBox[{
        RowBox[{"-", "1"}], "+", "F"}], ")"}], " ", "femaleNum", " ", 
      RowBox[{"(", 
       RowBox[{
        RowBox[{"-", "1"}], "+", "pf"}], ")"}], " ", "pf"}]}], ",", 
    RowBox[{"ddf", "\[Rule]", 
     RowBox[{"femaleNum", " ", "pf", " ", 
      RowBox[{"(", 
       RowBox[{"F", "+", "pf", "-", 
        RowBox[{"F", " ", "pf"}]}], ")"}]}]}]}], "}"}], "}"}]], "Output",
 CellChangeTimes->{{3.814077219052555*^9, 3.814077280927372*^9}, 
   3.814077589788096*^9, 3.814080197254603*^9, 3.814189384027315*^9, 
   3.814195950473256*^9, 3.8141965029755163`*^9},
 CellLabel->
  "Out[1270]=",ExpressionUUID->"16d45062-7510-4809-b453-50a35732a81e"]
}, Open  ]],

Cell["\<\
In the absence of any competition for resources, each female will survive to \
produce b surviving juveniles. Density dependence acts to reduce the survival \
of females to reproduction, survival of juveniles, and/or female fecundity \
(which are all modelled equivalently) by a factor of (1-\[Alpha] N) where N \
is the total density of males and females (maleNum+femaleNum). That is \
\[Alpha] is the per adult competitive effect on reproductive rate. totAdults \
gives the total population density of males and females. \
\>", "Text",
 CellChangeTimes->{{3.814078082733198*^9, 3.814078132242939*^9}, {
  3.814078167961803*^9, 3.814078323095675*^9}, {3.814078358488497*^9, 
  3.814078360367148*^9}, {3.814078400246647*^9, 3.81407846691296*^9}, {
  3.814078506703103*^9, 3.814078506939747*^9}, {3.814078603846031*^9, 
  3.8140786211249657`*^9}},ExpressionUUID->"6495c331-3979-4c05-ad07-\
e0a3df45f3ae"],

Cell[BoxData[
 RowBox[{
  RowBox[{"totAdults", " ", "=", " ", 
   RowBox[{
   "DDf", " ", "+", " ", "Ddf", " ", "+", " ", "ddf", " ", "+", " ", "Dm", 
    " ", "+", " ", "dm"}]}], ";"}]], "Input",
 CellLabel->
  "In[1271]:=",ExpressionUUID->"b8cdac10-be99-4143-9f4a-cbcfca14e837"],

Cell[TextData[{
 "Before juveniles reach adulthood, they experience selection according to \
their genotype. The relative fitness of different genotypes is given by\n",
 Cell[BoxData[
  FormBox[
   RowBox[{
    SubscriptBox["W", "XX"], "=", "1"}], TraditionalForm]],ExpressionUUID->
  "a77f4836-0e4d-4928-b7f2-d7e6ec7f207d"],
 "\n",
 Cell[BoxData[
  FormBox[
   RowBox[{
    SubscriptBox["W", 
     RowBox[{
      SubscriptBox["X", "d"], "X"}]], "=", 
    RowBox[{"1", "-", 
     RowBox[{"h", " ", "sf"}]}]}], TraditionalForm]],ExpressionUUID->
  "4561688b-357e-42bb-a620-375e399f8e16"],
 "\n",
 Cell[BoxData[
  FormBox[
   RowBox[{
    SubscriptBox["W", 
     RowBox[{
      SubscriptBox["X", "d"], 
      SubscriptBox["X", "d"]}]], "=", 
    RowBox[{"1", "-", "sf"}]}], TraditionalForm]],ExpressionUUID->
  "86946e7d-c0ba-433b-91d3-ff462383969c"],
 "\n",
 Cell[BoxData[
  FormBox[
   RowBox[{
    SubscriptBox["W", "XY"], "=", "1"}], TraditionalForm]],ExpressionUUID->
  "55c484f0-9132-4953-82b1-56540e9766f6"],
 "\n",
 Cell[BoxData[
  FormBox[
   RowBox[{
    SubscriptBox["W", 
     RowBox[{
      SubscriptBox["X", "d"], "Y"}]], "=", 
    RowBox[{"1", "-", "sm"}]}], TraditionalForm]],ExpressionUUID->
  "648095e2-ef28-4ed0-9797-f79c3b79c3a7"],
 "\nWe use ijfprime and imprime to specify the densities of each genotype \
after one generation (recursion/difference equations for change in genotype \
densities). For example, DDfprime is the density of females with genotype DD \
after one generation. "
}], "Text",
 CellChangeTimes->{{3.814078662208776*^9, 3.8140787408128223`*^9}, {
  3.81407914095951*^9, 3.814079300391842*^9}, {3.814080008982128*^9, 
  3.8140800183201637`*^9}},ExpressionUUID->"440e961e-4cff-444f-9479-\
667039b4eae6"]
}, Closed]],

Cell[CellGroupData[{

Cell[TextData[{
 "Allele Frequency and Population Size Dynamics - females mate with one male \
(",
 Cell[BoxData[
  FormBox[
   RowBox[{
    SubscriptBox["\[Lambda]", "f"], "=", "1"}], TraditionalForm]],
  FormatType->"TraditionalForm",ExpressionUUID->
  "31aa7438-73b4-4d17-92da-86a18422caea"],
 ")"
}], "Section",
 CellChangeTimes->{{3.781270219531413*^9, 3.78127022211313*^9}, {
  3.814005687940864*^9, 3.814005718641315*^9}, {3.814006090928898*^9, 
  3.814006091279448*^9}, {3.814018602075158*^9, 
  3.8140186161748466`*^9}},ExpressionUUID->"64106deb-c0a7-4c27-9d28-\
350dff73efe7"],

Cell[CellGroupData[{

Cell["Recursion equations", "Subsection",
 CellChangeTimes->{{3.781442815965227*^9, 
  3.7814428221797113`*^9}},ExpressionUUID->"39c0b246-04bf-4ac8-9a23-\
0d4f89147337"],

Cell[TextData[{
 "We consider each pairwise combination of male and female genotypes and the \
probability that each mating type produces each offspring genotype. We assume \
random mating so that each mating combination happens in proportion to the \
frequency of males with each genotype (e.g., ",
 Cell[BoxData[
  FractionBox["dm", 
   RowBox[{"(", 
    RowBox[{"Dm", "+", "dm"}], ")"}]]],
  CellChangeTimes->{{3.7814428720012417`*^9, 3.781442894240843*^9}, {
   3.814008329923299*^9, 3.8140083738240767`*^9}, {3.814019442825118*^9, 
   3.814019443216568*^9}, {3.81401966596344*^9, 3.814019666468874*^9}, {
   3.81407591820485*^9, 3.814075935309269*^9}, {3.814075978592771*^9, 
   3.814075985638269*^9}, {3.814079326819665*^9, 3.8140793341532917`*^9}, {
   3.8140794393433743`*^9, 3.814079615488195*^9}, {3.8140796473810863`*^9, 
   3.8140798375136547`*^9}},ExpressionUUID->
  "7e9fcc47-b6c8-474a-8d8d-364dde5ebf1a"],
 "is the frequency of drive males among males). These equations track all \
mating combinations that produce each juvenile genotype. The rate of birth \
and density-dependence competition (b*(1-\[Alpha] totAdults)) is not directly \
dependent on genotype. "
}], "Text",
 CellChangeTimes->{{3.781442917577629*^9, 3.7814429407924557`*^9}, {
  3.8140194969339123`*^9, 3.814019651741583*^9}, {3.814079370297378*^9, 
  3.814079448867035*^9}, {3.814079848942235*^9, 
  3.814079997046238*^9}},ExpressionUUID->"952fafe7-9bdc-4222-85bb-\
f6cec541a6e0"],

Cell[BoxData[{
 RowBox[{
  RowBox[{
   RowBox[{"DDfjuveniles", "=", 
    RowBox[{"b", "*", 
     RowBox[{"(", 
      RowBox[{"1", "-", 
       RowBox[{"\[Alpha]", " ", "totAdults"}]}], ")"}], 
     RowBox[{"(", 
      RowBox[{
       RowBox[{"DDf", "*", 
        FractionBox["Dm", 
         RowBox[{"(", 
          RowBox[{"Dm", "+", "dm"}], ")"}]], "*", 
        FractionBox["1", "2"]}], " ", "+", "\[IndentingNewLine]", 
       RowBox[{
        FractionBox["Ddf", "2"], "*", 
        FractionBox["Dm", 
         RowBox[{"(", 
          RowBox[{"Dm", "+", "dm"}], ")"}]], "*", 
        FractionBox["1", "2"]}]}], ")"}]}]}], ";"}], 
  "\[IndentingNewLine]"}], "\[IndentingNewLine]", 
 RowBox[{
  RowBox[{"Ddfjuveniles", "=", 
   RowBox[{"b", "*", 
    RowBox[{"(", 
     RowBox[{"1", "-", 
      RowBox[{"\[Alpha]", " ", "totAdults"}]}], ")"}], 
    RowBox[{"(", 
     RowBox[{
      RowBox[{"DDf", "*", 
       FractionBox["dm", 
        RowBox[{"(", 
         RowBox[{"Dm", "+", "dm"}], ")"}]], "*", 
       RowBox[{"(", 
        FractionBox[
         RowBox[{"(", 
          RowBox[{"1", "+", "\[Delta]"}], ")"}], "2"], ")"}]}], "+", 
      "\[IndentingNewLine]", 
      RowBox[{
       FractionBox["Ddf", "2"], 
       RowBox[{"(", 
        RowBox[{
         RowBox[{
          FractionBox["Dm", 
           RowBox[{"(", 
            RowBox[{"Dm", "+", "dm"}], ")"}]], "*", 
          FractionBox["1", "2"]}], "+", 
         RowBox[{
          FractionBox["dm", 
           RowBox[{"(", 
            RowBox[{"Dm", "+", "dm"}], ")"}]], "*", 
          RowBox[{"(", 
           FractionBox[
            RowBox[{"(", 
             RowBox[{"1", "+", "\[Delta]"}], ")"}], "2"], ")"}]}]}], ")"}]}], 
      "+", "\[IndentingNewLine]", 
      RowBox[{"ddf", "*", 
       FractionBox["Dm", 
        RowBox[{"(", 
         RowBox[{"Dm", "+", "dm"}], ")"}]], "*", 
       FractionBox["1", "2"]}]}], ")"}]}]}], ";"}], "\[IndentingNewLine]", 
 RowBox[{
  RowBox[{"ddfjuveniles", "=", 
   RowBox[{"b", "*", 
    RowBox[{"(", 
     RowBox[{"1", "-", 
      RowBox[{"\[Alpha]", " ", "totAdults"}]}], ")"}], 
    RowBox[{"(", 
     RowBox[{
      RowBox[{
       FractionBox["Ddf", "2"], "*", 
       FractionBox["dm", 
        RowBox[{"(", 
         RowBox[{"Dm", "+", "dm"}], ")"}]], "*", 
       RowBox[{"(", 
        FractionBox[
         RowBox[{"(", 
          RowBox[{"1", "+", "\[Delta]"}], ")"}], "2"], ")"}]}], "+", 
      "\[IndentingNewLine]", 
      RowBox[{"ddf", "*", 
       FractionBox["dm", 
        RowBox[{"(", 
         RowBox[{"Dm", "+", "dm"}], ")"}]], "*", 
       RowBox[{"(", 
        FractionBox[
         RowBox[{"(", 
          RowBox[{"1", "+", "\[Delta]"}], ")"}], "2"], ")"}]}]}], ")"}]}]}], 
  ";"}], "\[IndentingNewLine]", 
 RowBox[{
  RowBox[{
   RowBox[{"Dmjuveniles", "=", 
    RowBox[{"b", "*", 
     RowBox[{"(", 
      RowBox[{"1", "-", 
       RowBox[{"\[Alpha]", " ", "totAdults"}]}], ")"}], 
     RowBox[{"(", 
      RowBox[{
       RowBox[{"DDf", "*", 
        FractionBox["Dm", 
         RowBox[{"(", 
          RowBox[{"Dm", "+", "dm"}], ")"}]], "*", 
        FractionBox["1", "2"]}], "+", "\[IndentingNewLine]", 
       RowBox[{
        FractionBox["Ddf", "2"], "*", 
        RowBox[{"(", 
         RowBox[{
          RowBox[{
           FractionBox["Dm", 
            RowBox[{"(", 
             RowBox[{"Dm", "+", "dm"}], ")"}]], "*", 
           FractionBox["1", "2"]}], "+", 
          RowBox[{
           FractionBox["dm", 
            RowBox[{"(", 
             RowBox[{"Dm", "+", "dm"}], ")"}]], "*", 
           RowBox[{"(", 
            FractionBox[
             RowBox[{"(", 
              RowBox[{"1", "-", "\[Delta]"}], ")"}], "2"], ")"}]}]}], ")"}]}],
        "+", "\[IndentingNewLine]", 
       RowBox[{"DDf", "*", 
        FractionBox["dm", 
         RowBox[{"(", 
          RowBox[{"Dm", "+", "dm"}], ")"}]], "*", 
        RowBox[{"(", 
         FractionBox[
          RowBox[{"(", 
           RowBox[{"1", "-", "\[Delta]"}], ")"}], "2"], ")"}]}]}], ")"}]}]}], 
   ";"}], "\[IndentingNewLine]"}], "\[IndentingNewLine]", 
 RowBox[{
  RowBox[{"dmjuveniles", "=", 
   RowBox[{"b", "*", 
    RowBox[{"(", 
     RowBox[{"1", "-", 
      RowBox[{"\[Alpha]", " ", "totAdults"}]}], ")"}], 
    RowBox[{"(", 
     RowBox[{
      RowBox[{
       FractionBox["Ddf", "2"], "*", 
       RowBox[{"(", 
        RowBox[{
         RowBox[{
          FractionBox["dm", 
           RowBox[{"(", 
            RowBox[{"Dm", "+", "dm"}], ")"}]], "*", 
          RowBox[{"(", 
           FractionBox[
            RowBox[{"(", 
             RowBox[{"1", "-", "\[Delta]"}], ")"}], "2"], ")"}]}], "+", 
         RowBox[{
          FractionBox["Dm", 
           RowBox[{"(", 
            RowBox[{"Dm", "+", "dm"}], ")"}]], "*", 
          FractionBox["1", "2"]}]}], ")"}]}], "+", "\[IndentingNewLine]", 
      RowBox[{"ddf", "*", 
       RowBox[{"(", 
        RowBox[{
         RowBox[{
          FractionBox["Dm", 
           RowBox[{"(", 
            RowBox[{"Dm", "+", "dm"}], ")"}]], 
          FractionBox["1", "2"]}], "+", 
         RowBox[{
          FractionBox["dm", 
           RowBox[{"(", 
            RowBox[{"Dm", "+", "dm"}], ")"}]], "*", 
          RowBox[{"(", 
           FractionBox[
            RowBox[{"(", 
             RowBox[{"1", "-", "\[Delta]"}], ")"}], "2"], ")"}]}]}], 
        ")"}]}]}], ")"}]}]}], ";"}]}], "Input",
 CellChangeTimes->{{3.7814428720012417`*^9, 3.781442894240843*^9}, {
  3.814008329923299*^9, 3.8140083738240767`*^9}, {3.814019442825118*^9, 
  3.814019443216568*^9}, {3.81401966596344*^9, 3.814019666468874*^9}, {
  3.81407591820485*^9, 3.814075935309269*^9}, {3.814075978592771*^9, 
  3.814075985638269*^9}, {3.814079326819665*^9, 3.8140793341532917`*^9}, {
  3.8140794393433743`*^9, 3.814079615488195*^9}, {3.8140796473810863`*^9, 
  3.8140798375136547`*^9}},
 CellLabel->
  "In[1272]:=",ExpressionUUID->"c7291398-a127-4761-862e-ec5f4fc0601c"],

Cell["\<\
Before the next generation of adults, viability selection acts according to \
genotype. \
\>", "Text",
 CellChangeTimes->{{3.814019703490279*^9, 3.814019713199089*^9}, {
  3.814080207173728*^9, 
  3.8140802231575623`*^9}},ExpressionUUID->"3b8c5c61-7339-4028-8e8d-\
7a59c34f43d9"],

Cell[BoxData[{
 RowBox[{
  RowBox[{"DDfprime", "=", 
   RowBox[{"DDfjuveniles", " ", "*", " ", "1"}]}], 
  ";"}], "\[IndentingNewLine]", 
 RowBox[{
  RowBox[{"Ddfprime", "=", 
   RowBox[{"Ddfjuveniles", "*", 
    RowBox[{"(", 
     RowBox[{"1", " ", "-", " ", 
      RowBox[{"h", "*", "sf"}]}], ")"}]}]}], ";"}], "\[IndentingNewLine]", 
 RowBox[{
  RowBox[{"ddfprime", "=", 
   RowBox[{"ddfjuveniles", "*", 
    RowBox[{"(", 
     RowBox[{"1", " ", "-", " ", "sf"}], ")"}]}]}], 
  ";"}], "\[IndentingNewLine]", 
 RowBox[{
  RowBox[{"Dmprime", "=", 
   RowBox[{"Dmjuveniles", "*", "1"}]}], ";"}], "\[IndentingNewLine]", 
 RowBox[{
  RowBox[{"dmprime", "=", 
   RowBox[{"dmjuveniles", "*", 
    RowBox[{"(", 
     RowBox[{"1", " ", "-", " ", "sm"}], ")"}]}]}], ";"}]}], "Input",
 CellChangeTimes->{{3.814078581020732*^9, 3.8140785877471237`*^9}, {
  3.814080227714635*^9, 3.814080242013699*^9}},
 CellLabel->
  "In[1277]:=",ExpressionUUID->"3c17a47e-acfe-4409-983e-0716addff0f1"]
}, Closed]],

Cell[CellGroupData[{

Cell["Equilibrium allele frequency and population size", "Subsection",
 CellChangeTimes->{{3.781270227876593*^9, 
  3.781270234977231*^9}},ExpressionUUID->"6cec54af-8f67-495d-ae4b-\
902b41bf33b2"],

Cell["\<\
We first look at only the difference in allele frequency and inbreeding \
coefficient to calculate the polymorphic allele frequency equilibrium (when \
changes in allele frequency and inbreeding coefficient are 0)\
\>", "Text",
 CellChangeTimes->{{3.781444079790472*^9, 3.781444104553514*^9}, {
  3.781457247488077*^9, 3.781457276171872*^9}, {3.781887727883533*^9, 
  3.781887728058511*^9}, {3.81408030548451*^9, 
  3.8140803701820793`*^9}},ExpressionUUID->"73ee4ad7-1b79-49e0-94b4-\
98acb39dd9bd"],

Cell[CellGroupData[{

Cell[BoxData[{
 RowBox[{
  RowBox[{"solveme", "=", 
   RowBox[{"Simplify", "[", 
    RowBox[{
     RowBox[{"{", 
      RowBox[{
       RowBox[{
        RowBox[{"(", 
         FractionBox["dmprime", 
          RowBox[{"Dmprime", "+", "dmprime"}]], ")"}], "-", 
        FractionBox["dm", 
         RowBox[{"dm", "+", "Dm"}]]}], ",", 
       RowBox[{
        RowBox[{"(", 
         FractionBox[
          RowBox[{
           RowBox[{"2", " ", "ddfprime"}], "+", "Ddfprime"}], 
          RowBox[{"2", " ", 
           RowBox[{"(", 
            RowBox[{"ddfprime", "+", "Ddfprime", "+", "DDfprime"}], ")"}]}]], 
         ")"}], "-", 
        FractionBox[
         RowBox[{
          RowBox[{"2", " ", "ddf"}], "+", "Ddf"}], 
         RowBox[{"2", " ", 
          RowBox[{"(", 
           RowBox[{"ddf", "+", "Ddf", "+", "DDf"}], ")"}]}]]}], ",", 
       RowBox[{
        RowBox[{"(", 
         FractionBox[
          RowBox[{
           RowBox[{"4", " ", "ddfprime", " ", "DDfprime"}], " ", "-", 
           SuperscriptBox["Ddfprime", "2"]}], 
          RowBox[{
           RowBox[{"(", 
            RowBox[{
             RowBox[{"2", " ", "ddfprime"}], "+", "Ddfprime"}], ")"}], " ", 
           RowBox[{"(", 
            RowBox[{"Ddfprime", "+", 
             RowBox[{"2", " ", "DDfprime"}]}], ")"}]}]], ")"}], "-", 
        FractionBox[
         RowBox[{
          RowBox[{"4", " ", "ddf", " ", "DDf"}], "-", 
          SuperscriptBox["Ddf", "2"]}], 
         RowBox[{
          RowBox[{"(", 
           RowBox[{
            RowBox[{"2", " ", "ddf"}], "+", "Ddf"}], ")"}], " ", 
          RowBox[{"(", 
           RowBox[{"Ddf", "+", 
            RowBox[{"2", " ", "DDf"}]}], ")"}]}]]}]}], "}"}], "/.", "subeq"}],
     "]"}]}], ";"}], "\[IndentingNewLine]", 
 RowBox[{"equilibrium", "=", 
  RowBox[{
   RowBox[{"Solve", "[", 
    RowBox[{
     RowBox[{
      RowBox[{"solveme", "[", 
       RowBox[{"[", "1", "]"}], "]"}], "\[Equal]", 
      RowBox[{"{", 
       RowBox[{"0", ",", "0", ",", "0"}], "}"}]}], ",", 
     RowBox[{"{", 
      RowBox[{"pm", ",", "pf", ",", "F"}], "}"}]}], "]"}], "//", 
   "Simplify"}]}]}], "Input",
 CellChangeTimes->{{3.7814440517009563`*^9, 3.7814440553262253`*^9}, {
  3.781444152654667*^9, 3.781444154007221*^9}, {3.781457230234502*^9, 
  3.781457236246904*^9}, {3.7814572809999943`*^9, 3.7814573029327593`*^9}, {
  3.781871249129278*^9, 3.781871260774413*^9}, {3.781884886553536*^9, 
  3.7818848877054157`*^9}, {3.8140084592127047`*^9, 3.81400845966019*^9}, {
  3.8140802649485693`*^9, 3.814080283706593*^9}},
 CellLabel->
  "In[1282]:=",ExpressionUUID->"de30a10f-65b2-4f17-9a3d-9934a3ef9d10"],

Cell[BoxData[
 RowBox[{"{", 
  RowBox[{"{", 
   RowBox[{
    RowBox[{"pm", "\[Rule]", 
     FractionBox[
      RowBox[{
       RowBox[{"(", 
        RowBox[{
         RowBox[{"-", "1"}], "+", "sm"}], ")"}], " ", 
       RowBox[{"(", 
        RowBox[{"\[Delta]", "-", 
         RowBox[{"sm", " ", 
          RowBox[{"(", 
           RowBox[{"1", "+", "\[Delta]"}], ")"}]}], "+", 
         RowBox[{"h", " ", "sf", " ", 
          RowBox[{"(", 
           RowBox[{
            RowBox[{"-", "2"}], "+", "sm", "-", "\[Delta]", "+", 
            RowBox[{"sm", " ", "\[Delta]"}]}], ")"}]}]}], ")"}]}], 
      RowBox[{
       RowBox[{"2", " ", "sf", " ", 
        RowBox[{"(", 
         RowBox[{
          RowBox[{"-", "1"}], "+", "sm"}], ")"}], " ", 
        RowBox[{"(", 
         RowBox[{"1", "+", "\[Delta]"}], ")"}]}], "+", 
       RowBox[{"h", " ", "sf", " ", 
        RowBox[{"(", 
         RowBox[{
          RowBox[{"-", "2"}], "+", "sm"}], ")"}], " ", 
        RowBox[{"(", 
         RowBox[{
          RowBox[{"-", "2"}], "+", "sm", "-", "\[Delta]", "+", 
          RowBox[{"sm", " ", "\[Delta]"}]}], ")"}]}], "-", 
       RowBox[{"sm", " ", 
        RowBox[{"(", 
         RowBox[{"sm", "-", "\[Delta]", "+", 
          RowBox[{"sm", " ", "\[Delta]"}]}], ")"}]}]}]]}], ",", 
    RowBox[{"pf", "\[Rule]", 
     FractionBox[
      RowBox[{"\[Delta]", "-", 
       RowBox[{"sm", " ", 
        RowBox[{"(", 
         RowBox[{"1", "+", "\[Delta]"}], ")"}]}], "+", 
       RowBox[{"h", " ", "sf", " ", 
        RowBox[{"(", 
         RowBox[{
          RowBox[{"-", "2"}], "+", "sm", "-", "\[Delta]", "+", 
          RowBox[{"sm", " ", "\[Delta]"}]}], ")"}]}]}], 
      RowBox[{"2", " ", "sf", " ", 
       RowBox[{"(", 
        RowBox[{
         RowBox[{
          RowBox[{"-", 
           RowBox[{"(", 
            RowBox[{
             RowBox[{"-", "1"}], "+", "sm"}], ")"}]}], " ", 
          RowBox[{"(", 
           RowBox[{"1", "+", "\[Delta]"}], ")"}]}], "+", 
         RowBox[{"h", " ", 
          RowBox[{"(", 
           RowBox[{
            RowBox[{"-", "2"}], "+", "sm", "-", "\[Delta]", "+", 
            RowBox[{"sm", " ", "\[Delta]"}]}], ")"}]}]}], ")"}]}]]}], ",", 
    RowBox[{"F", "\[Rule]", 
     RowBox[{"-", 
      FractionBox[
       RowBox[{
        RowBox[{
         RowBox[{"-", "2"}], " ", "h", " ", 
         SuperscriptBox["sf", "2"], " ", 
         RowBox[{"(", 
          RowBox[{
           RowBox[{"-", "1"}], "+", "sm"}], ")"}], " ", 
         RowBox[{"(", 
          RowBox[{"1", "+", "\[Delta]"}], ")"}], " ", 
         RowBox[{"(", 
          RowBox[{
           RowBox[{"-", "2"}], "+", "sm", "-", "\[Delta]", "+", 
           RowBox[{"sm", " ", "\[Delta]"}]}], ")"}]}], "+", 
        RowBox[{
         SuperscriptBox["h", "2"], " ", 
         SuperscriptBox["sf", "2"], " ", 
         SuperscriptBox[
          RowBox[{"(", 
           RowBox[{
            RowBox[{"-", "2"}], "+", "sm", "-", "\[Delta]", "+", 
            RowBox[{"sm", " ", "\[Delta]"}]}], ")"}], "2"]}], "+", 
        RowBox[{
         RowBox[{"(", 
          RowBox[{"sm", "-", "\[Delta]", "+", 
           RowBox[{"sm", " ", "\[Delta]"}]}], ")"}], " ", 
         RowBox[{"(", 
          RowBox[{"\[Delta]", "+", 
           RowBox[{"2", " ", "sf", " ", 
            RowBox[{"(", 
             RowBox[{
              RowBox[{"-", "1"}], "+", "sm"}], ")"}], " ", 
            RowBox[{"(", 
             RowBox[{"1", "+", "\[Delta]"}], ")"}]}], "-", 
           RowBox[{"sm", " ", 
            RowBox[{"(", 
             RowBox[{"1", "+", "\[Delta]"}], ")"}]}]}], ")"}]}]}], 
       RowBox[{
        RowBox[{
         SuperscriptBox["h", "2"], " ", 
         SuperscriptBox["sf", "2"], " ", 
         SuperscriptBox[
          RowBox[{"(", 
           RowBox[{
            RowBox[{"-", "2"}], "+", "sm", "-", "\[Delta]", "+", 
            RowBox[{"sm", " ", "\[Delta]"}]}], ")"}], "2"]}], "+", 
        SuperscriptBox[
         RowBox[{"(", 
          RowBox[{"sm", "-", "\[Delta]", "+", 
           RowBox[{"sm", " ", "\[Delta]"}]}], ")"}], "2"], "-", 
        RowBox[{"2", " ", "sf", " ", 
         RowBox[{"(", 
          RowBox[{
           RowBox[{"2", " ", 
            RowBox[{"(", 
             RowBox[{
              RowBox[{"-", "1"}], "+", "sm"}], ")"}], " ", 
            RowBox[{"(", 
             RowBox[{"1", "+", "\[Delta]"}], ")"}]}], "+", 
           RowBox[{"h", " ", 
            SuperscriptBox[
             RowBox[{"(", 
              RowBox[{
               RowBox[{"-", "2"}], "+", "sm", "-", "\[Delta]", "+", 
               RowBox[{"sm", " ", "\[Delta]"}]}], ")"}], "2"]}]}], 
          ")"}]}]}]]}]}]}], "}"}], "}"}]], "Output",
 CellChangeTimes->{{3.781457286264279*^9, 3.781457329190282*^9}, {
   3.781871254806685*^9, 3.781871262324274*^9}, {3.781884895727561*^9, 
   3.7818849162745123`*^9}, 3.7818880903011427`*^9, 3.781952829551042*^9, 
   3.7819529052636023`*^9, 3.781953902827918*^9, 3.781953937303303*^9, 
   3.782130661513627*^9, 3.783078902950335*^9, 3.7851523839197206`*^9, 
   3.78516271606812*^9, 3.785164947308786*^9, 3.78516498428473*^9, 
   3.7855724309572067`*^9, 3.7855883933616667`*^9, 3.789732675249082*^9, 
   3.790078582641345*^9, 3.795174667458596*^9, 3.799753840262821*^9, 
   3.800007967624103*^9, 3.800268193924336*^9, 3.813928086895986*^9, 
   3.813928138768917*^9, 3.814007922595193*^9, 3.8140189041365137`*^9, 
   3.814019740940907*^9, 3.814080287792897*^9, 3.814189390123685*^9, 
   3.8141959553591633`*^9, 3.814196505062763*^9},
 CellLabel->
  "Out[1283]=",ExpressionUUID->"12fdd420-b47c-4fea-b67f-734a62e0777f"]
}, Open  ]],

Cell["The full system is described by", "Text",
 CellChangeTimes->{{3.781459868655086*^9, 3.781459873390123*^9}, {
  3.814080376144492*^9, 
  3.814080390555254*^9}},ExpressionUUID->"eb65e4ed-19e0-45dc-8f5c-\
864c8f1c51d6"],

Cell[BoxData[
 RowBox[{
  RowBox[{"solvemeFull", "=", 
   RowBox[{
    RowBox[{
     RowBox[{"{", 
      RowBox[{
       RowBox[{"DDfprime", "-", "DDf"}], ",", 
       RowBox[{"Ddfprime", "-", "Ddf"}], ",", 
       RowBox[{"ddfprime", "-", "ddf"}], ",", " ", 
       RowBox[{"Dmprime", "-", "Dm"}], ",", " ", 
       RowBox[{"dmprime", "-", "dm"}]}], "}"}], "/.", 
     RowBox[{"subeq", "[", 
      RowBox[{"[", "1", "]"}], "]"}]}], "//", "Simplify"}]}], ";"}]], "Input",
 CellChangeTimes->{{3.781460135237193*^9, 3.7814601398015003`*^9}},
 CellLabel->
  "In[1284]:=",ExpressionUUID->"c3fd749e-7bec-4cdf-8c2a-e20d7b5d47c5"],

Cell["\<\
We can then solve for the population sizes. First we find population sizes \
when drive alleles are absent (pm=pf=0) or fixed (pm=pf=1):\
\>", "Text",
 CellChangeTimes->{{3.781459884674746*^9, 3.7814598986599483`*^9}, {
  3.7814599811143436`*^9, 3.781460009186688*^9}, {3.814081251018674*^9, 
  3.8140813011507196`*^9}},ExpressionUUID->"5d4314a8-b012-4d04-b525-\
0aad925df173"],

Cell[CellGroupData[{

Cell[BoxData[{
 RowBox[{"noDrivePopSizeSol", "=", 
  RowBox[{
   RowBox[{"Solve", "[", 
    RowBox[{
     RowBox[{
      RowBox[{"(", 
       RowBox[{
        RowBox[{"solvemeFull", "/.", 
         RowBox[{"pm", "\[Rule]", "0"}]}], "/.", 
        RowBox[{"pf", "\[Rule]", "0"}]}], ")"}], "\[Equal]", 
      RowBox[{"{", 
       RowBox[{"0", ",", "0", ",", "0", ",", "0", ",", "0"}], "}"}]}], ",", 
     RowBox[{"{", 
      RowBox[{"maleNum", ",", "femaleNum"}], "}"}]}], "]"}], "[", 
   RowBox[{"[", "2", "]"}], "]"}]}], "\[IndentingNewLine]", 
 RowBox[{"fixedDrivePopSizeSol", " ", "=", " ", 
  RowBox[{
   RowBox[{
    RowBox[{"Solve", "[", 
     RowBox[{
      RowBox[{
       RowBox[{"(", 
        RowBox[{
         RowBox[{"solvemeFull", "/.", 
          RowBox[{"pm", "\[Rule]", "1"}]}], "/.", 
         RowBox[{"pf", "\[Rule]", "1"}]}], ")"}], "\[Equal]", 
       RowBox[{"{", 
        RowBox[{"0", ",", "0", ",", "0", ",", "0", ",", "0"}], "}"}]}], ",", 
      RowBox[{"{", 
       RowBox[{"maleNum", ",", "femaleNum"}], "}"}]}], "]"}], "[", 
    RowBox[{"[", "2", "]"}], "]"}], "//", 
   "Simplify"}]}], "\[IndentingNewLine]", 
 RowBox[{"NnoDrive", "=", 
  RowBox[{
   RowBox[{
    RowBox[{"(", 
     RowBox[{"maleNum", " ", "+", " ", "femaleNum"}], ")"}], "/.", 
    "noDrivePopSizeSol"}], "//", "FullSimplify"}]}], "\[IndentingNewLine]", 
 RowBox[{"Nfixed", " ", "=", " ", 
  RowBox[{
   RowBox[{
    RowBox[{"(", 
     RowBox[{"maleNum", " ", "+", " ", "femaleNum"}], ")"}], "/.", 
    "fixedDrivePopSizeSol"}], "//", "FullSimplify"}]}]}], "Input",
 CellChangeTimes->{{3.781460025346209*^9, 3.78146009015233*^9}, {
   3.7818886806627913`*^9, 3.781888696789563*^9}, {3.7851524020108128`*^9, 
   3.785152402131963*^9}, {3.785152987803329*^9, 3.78515307352833*^9}, 
   3.785153145067387*^9, {3.7851560785368853`*^9, 3.7851560869155483`*^9}, {
   3.7851567369931593`*^9, 3.7851567962332497`*^9}, {3.785246487946137*^9, 
   3.785246499036151*^9}, 3.800008009183523*^9, {3.813928108083933*^9, 
   3.813928108539372*^9}, {3.814008467638185*^9, 3.814008471911174*^9}, {
   3.8140804965358553`*^9, 3.814080500635491*^9}},
 CellLabel->
  "In[1285]:=",ExpressionUUID->"b95805b1-e618-48a4-86f1-166d4a8e9edf"],

Cell[BoxData[
 RowBox[{"{", 
  RowBox[{
   RowBox[{"maleNum", "\[Rule]", 
    FractionBox[
     RowBox[{
      RowBox[{"-", "2"}], "+", "b"}], 
     RowBox[{"2", " ", "b", " ", "\[Alpha]"}]]}], ",", 
   RowBox[{"femaleNum", "\[Rule]", 
    FractionBox[
     RowBox[{
      RowBox[{"-", "2"}], "+", "b"}], 
     RowBox[{"2", " ", "b", " ", "\[Alpha]"}]]}]}], "}"}]], "Output",
 CellChangeTimes->{
  3.781459915238161*^9, {3.7814600290773687`*^9, 3.781460090503332*^9}, 
   3.781884923662936*^9, 3.781888113508231*^9, {3.781888685197915*^9, 
   3.781888700731204*^9}, 3.781952829895693*^9, 3.781952905640814*^9, 
   3.781953903117161*^9, 3.78195393756216*^9, 3.782130661789319*^9, 
   3.783078903509358*^9, 3.785152384536742*^9, 3.785152417953383*^9, 
   3.7851529881784067`*^9, {3.785153018743541*^9, 3.7851530741198397`*^9}, 
   3.7851531453647327`*^9, 3.7851560876389647`*^9, {3.785156737317646*^9, 
   3.785156796838706*^9}, 3.785162716671402*^9, 3.785164795842664*^9, 
   3.785164948200588*^9, 3.785164984849872*^9, {3.78524649041514*^9, 
   3.785246500246359*^9}, 3.785572431461936*^9, 3.785588394012405*^9, 
   3.78973267580768*^9, 3.7900785834664507`*^9, 3.795174667946147*^9, 
   3.799753840607267*^9, 3.799753999535042*^9, 3.8000079682202597`*^9, 
   3.800268194300805*^9, {3.8139280872828007`*^9, 3.813928108980541*^9}, 
   3.8139281390629587`*^9, 3.814019762371874*^9, 3.814080397313016*^9, 
   3.814080501865264*^9, 3.814189390336687*^9, 3.814195955528944*^9, 
   3.814196505260497*^9},
 CellLabel->
  "Out[1285]=",ExpressionUUID->"7eea0df3-e312-4ff9-9003-fab15e21f3e1"],

Cell[BoxData[
 RowBox[{"{", 
  RowBox[{
   RowBox[{"maleNum", "\[Rule]", 
    RowBox[{"-", 
     FractionBox[
      RowBox[{
       RowBox[{"(", 
        RowBox[{
         RowBox[{"-", "1"}], "+", "sm"}], ")"}], " ", 
       RowBox[{"(", 
        RowBox[{
         RowBox[{"-", "1"}], "+", "\[Delta]"}], ")"}], " ", 
       RowBox[{"(", 
        RowBox[{"2", "+", 
         RowBox[{"b", " ", 
          RowBox[{"(", 
           RowBox[{
            RowBox[{"-", "1"}], "+", "sf"}], ")"}], " ", 
          RowBox[{"(", 
           RowBox[{"1", "+", "\[Delta]"}], ")"}]}]}], ")"}]}], 
      RowBox[{"b", " ", 
       RowBox[{"(", 
        RowBox[{
         RowBox[{"-", "1"}], "+", "sf"}], ")"}], " ", "\[Alpha]", " ", 
       RowBox[{"(", 
        RowBox[{"1", "+", "\[Delta]"}], ")"}], " ", 
       RowBox[{"(", 
        RowBox[{
         RowBox[{"-", "2"}], "+", "sf", "+", "sm", "+", 
         RowBox[{"sf", " ", "\[Delta]"}], "-", 
         RowBox[{"sm", " ", "\[Delta]"}]}], ")"}]}]]}]}], ",", 
   RowBox[{"femaleNum", "\[Rule]", 
    FractionBox[
     RowBox[{"2", "+", 
      RowBox[{"b", " ", 
       RowBox[{"(", 
        RowBox[{
         RowBox[{"-", "1"}], "+", "sf"}], ")"}], " ", 
       RowBox[{"(", 
        RowBox[{"1", "+", "\[Delta]"}], ")"}]}]}], 
     RowBox[{"b", " ", "\[Alpha]", " ", 
      RowBox[{"(", 
       RowBox[{
        RowBox[{"-", "2"}], "+", "sf", "+", "sm", "+", 
        RowBox[{"sf", " ", "\[Delta]"}], "-", 
        RowBox[{"sm", " ", "\[Delta]"}]}], ")"}]}]]}]}], "}"}]], "Output",
 CellChangeTimes->{
  3.781459915238161*^9, {3.7814600290773687`*^9, 3.781460090503332*^9}, 
   3.781884923662936*^9, 3.781888113508231*^9, {3.781888685197915*^9, 
   3.781888700731204*^9}, 3.781952829895693*^9, 3.781952905640814*^9, 
   3.781953903117161*^9, 3.78195393756216*^9, 3.782130661789319*^9, 
   3.783078903509358*^9, 3.785152384536742*^9, 3.785152417953383*^9, 
   3.7851529881784067`*^9, {3.785153018743541*^9, 3.7851530741198397`*^9}, 
   3.7851531453647327`*^9, 3.7851560876389647`*^9, {3.785156737317646*^9, 
   3.785156796838706*^9}, 3.785162716671402*^9, 3.785164795842664*^9, 
   3.785164948200588*^9, 3.785164984849872*^9, {3.78524649041514*^9, 
   3.785246500246359*^9}, 3.785572431461936*^9, 3.785588394012405*^9, 
   3.78973267580768*^9, 3.7900785834664507`*^9, 3.795174667946147*^9, 
   3.799753840607267*^9, 3.799753999535042*^9, 3.8000079682202597`*^9, 
   3.800268194300805*^9, {3.8139280872828007`*^9, 3.813928108980541*^9}, 
   3.8139281390629587`*^9, 3.814019762371874*^9, 3.814080397313016*^9, 
   3.814080501865264*^9, 3.814189390336687*^9, 3.814195955528944*^9, 
   3.814196505297574*^9},
 CellLabel->
  "Out[1286]=",ExpressionUUID->"d14df597-d291-4b5b-940a-27a3b5d1a11d"],

Cell[BoxData[
 FractionBox[
  RowBox[{
   RowBox[{"-", "2"}], "+", "b"}], 
  RowBox[{"b", " ", "\[Alpha]"}]]], "Output",
 CellChangeTimes->{
  3.781459915238161*^9, {3.7814600290773687`*^9, 3.781460090503332*^9}, 
   3.781884923662936*^9, 3.781888113508231*^9, {3.781888685197915*^9, 
   3.781888700731204*^9}, 3.781952829895693*^9, 3.781952905640814*^9, 
   3.781953903117161*^9, 3.78195393756216*^9, 3.782130661789319*^9, 
   3.783078903509358*^9, 3.785152384536742*^9, 3.785152417953383*^9, 
   3.7851529881784067`*^9, {3.785153018743541*^9, 3.7851530741198397`*^9}, 
   3.7851531453647327`*^9, 3.7851560876389647`*^9, {3.785156737317646*^9, 
   3.785156796838706*^9}, 3.785162716671402*^9, 3.785164795842664*^9, 
   3.785164948200588*^9, 3.785164984849872*^9, {3.78524649041514*^9, 
   3.785246500246359*^9}, 3.785572431461936*^9, 3.785588394012405*^9, 
   3.78973267580768*^9, 3.7900785834664507`*^9, 3.795174667946147*^9, 
   3.799753840607267*^9, 3.799753999535042*^9, 3.8000079682202597`*^9, 
   3.800268194300805*^9, {3.8139280872828007`*^9, 3.813928108980541*^9}, 
   3.8139281390629587`*^9, 3.814019762371874*^9, 3.814080397313016*^9, 
   3.814080501865264*^9, 3.814189390336687*^9, 3.814195955528944*^9, 
   3.8141965052985153`*^9},
 CellLabel->
  "Out[1287]=",ExpressionUUID->"f3698d69-2c83-45f0-823e-0e01b1544ba1"],

Cell[BoxData[
 FractionBox[
  RowBox[{"1", "+", 
   FractionBox["2", 
    RowBox[{"b", " ", 
     RowBox[{"(", 
      RowBox[{
       RowBox[{"-", "1"}], "+", "sf"}], ")"}], " ", 
     RowBox[{"(", 
      RowBox[{"1", "+", "\[Delta]"}], ")"}]}]]}], "\[Alpha]"]], "Output",
 CellChangeTimes->{
  3.781459915238161*^9, {3.7814600290773687`*^9, 3.781460090503332*^9}, 
   3.781884923662936*^9, 3.781888113508231*^9, {3.781888685197915*^9, 
   3.781888700731204*^9}, 3.781952829895693*^9, 3.781952905640814*^9, 
   3.781953903117161*^9, 3.78195393756216*^9, 3.782130661789319*^9, 
   3.783078903509358*^9, 3.785152384536742*^9, 3.785152417953383*^9, 
   3.7851529881784067`*^9, {3.785153018743541*^9, 3.7851530741198397`*^9}, 
   3.7851531453647327`*^9, 3.7851560876389647`*^9, {3.785156737317646*^9, 
   3.785156796838706*^9}, 3.785162716671402*^9, 3.785164795842664*^9, 
   3.785164948200588*^9, 3.785164984849872*^9, {3.78524649041514*^9, 
   3.785246500246359*^9}, 3.785572431461936*^9, 3.785588394012405*^9, 
   3.78973267580768*^9, 3.7900785834664507`*^9, 3.795174667946147*^9, 
   3.799753840607267*^9, 3.799753999535042*^9, 3.8000079682202597`*^9, 
   3.800268194300805*^9, {3.8139280872828007`*^9, 3.813928108980541*^9}, 
   3.8139281390629587`*^9, 3.814019762371874*^9, 3.814080397313016*^9, 
   3.814080501865264*^9, 3.814189390336687*^9, 3.814195955528944*^9, 
   3.8141965053440847`*^9},
 CellLabel->
  "Out[1288]=",ExpressionUUID->"1b762892-0378-4e0c-b8d4-7449844f0bb0"]
}, Open  ]],

Cell["\<\
Now for the case where a polymorphic equilibrium has been reached\
\>", "Text",
 CellChangeTimes->{{3.781460153288809*^9, 
  3.7814601650929117`*^9}},ExpressionUUID->"a2ebe6e0-9a3a-4805-9370-\
fa17194a141a"],

Cell[CellGroupData[{

Cell[BoxData[{
 RowBox[{
  RowBox[{"simpleSolvemeFull", "=", 
   RowBox[{"Simplify", "[", 
    RowBox[{"(", "solvemeFull", ")"}], "]"}]}], ";"}], "\[IndentingNewLine]", 
 RowBox[{"equilPopSizeSol", "=", 
  RowBox[{"Simplify", "[", 
   RowBox[{"Solve", "[", 
    RowBox[{
     RowBox[{
      RowBox[{"Flatten", "[", 
       RowBox[{"Simplify", "[", 
        RowBox[{"(", 
         RowBox[{"simpleSolvemeFull", "/.", "equilibrium"}], ")"}], "]"}], 
       "]"}], "\[Equal]", 
      RowBox[{"{", 
       RowBox[{"0", ",", "0", ",", "0", ",", "0", ",", "0"}], "}"}]}], ",", 
     " ", 
     RowBox[{"{", 
      RowBox[{"femaleNum", ",", "maleNum"}], "}"}]}], "]"}], 
   "]"}]}], "\[IndentingNewLine]", 
 RowBox[{
  RowBox[{
   RowBox[{"(", 
    RowBox[{"maleNum", " ", "+", " ", "femaleNum"}], ")"}], "/.", 
   "equilPopSizeSol"}], "//", "FullSimplify"}]}], "Input",
 CellChangeTimes->{
  3.7814601807633467`*^9, {3.781460242908415*^9, 3.781460297045966*^9}, {
   3.78146042104764*^9, 3.7814604314131813`*^9}, {3.7818849387750196`*^9, 
   3.781884939816983*^9}, {3.785242212047227*^9, 3.785242352509232*^9}, {
   3.8140084788313217`*^9, 3.8140084793896217`*^9}},
 CellLabel->
  "In[1289]:=",ExpressionUUID->"bafee36e-ee18-4c2c-add2-b705d82ee33d"],

Cell[BoxData[
 RowBox[{"{", 
  RowBox[{
   RowBox[{"{", 
    RowBox[{
     RowBox[{"femaleNum", "\[Rule]", "0"}], ",", 
     RowBox[{"maleNum", "\[Rule]", "0"}]}], "}"}], ",", 
   RowBox[{"{", 
    RowBox[{
     RowBox[{"femaleNum", "\[Rule]", 
      RowBox[{
       RowBox[{"(", 
        RowBox[{"sf", " ", 
         RowBox[{"(", 
          RowBox[{
           RowBox[{
            RowBox[{"-", 
             RowBox[{"(", 
              RowBox[{
               RowBox[{"-", "1"}], "+", "sm"}], ")"}]}], " ", 
            RowBox[{"(", 
             RowBox[{"1", "+", "\[Delta]"}], ")"}]}], "+", 
           RowBox[{"h", " ", 
            RowBox[{"(", 
             RowBox[{
              RowBox[{"-", "2"}], "+", "sm", "-", "\[Delta]", "+", 
              RowBox[{"sm", " ", "\[Delta]"}]}], ")"}]}]}], ")"}], " ", 
         RowBox[{"(", 
          RowBox[{
           RowBox[{"b", " ", 
            SuperscriptBox["h", "2"], " ", 
            SuperscriptBox["sf", "2"], " ", 
            SuperscriptBox[
             RowBox[{"(", 
              RowBox[{
               RowBox[{"-", "2"}], "+", "sm", "-", "\[Delta]", "+", 
               RowBox[{"sm", " ", "\[Delta]"}]}], ")"}], "2"]}], "+", 
           RowBox[{
            RowBox[{"(", 
             RowBox[{"sm", "-", "\[Delta]", "+", 
              RowBox[{"sm", " ", "\[Delta]"}]}], ")"}], " ", 
            RowBox[{"(", 
             RowBox[{
              RowBox[{
               RowBox[{"-", "b"}], " ", "\[Delta]"}], "+", 
              RowBox[{"sm", " ", 
               RowBox[{"(", 
                RowBox[{
                 RowBox[{"-", "4"}], "+", "b", "+", 
                 RowBox[{"b", " ", "\[Delta]"}]}], ")"}]}]}], ")"}]}], "-", 
           RowBox[{"2", " ", "sf", " ", 
            RowBox[{"(", 
             RowBox[{
              RowBox[{
               RowBox[{"-", "4"}], " ", 
               RowBox[{"(", 
                RowBox[{
                 RowBox[{"-", "1"}], "+", "sm"}], ")"}], " ", 
               RowBox[{"(", 
                RowBox[{"1", "+", "\[Delta]"}], ")"}]}], "+", 
              RowBox[{"2", " ", "b", " ", 
               RowBox[{"(", 
                RowBox[{
                 RowBox[{"-", "1"}], "+", "sm"}], ")"}], " ", 
               RowBox[{"(", 
                RowBox[{"1", "+", "\[Delta]"}], ")"}]}], "-", 
              RowBox[{"2", " ", "h", " ", 
               RowBox[{"(", 
                RowBox[{
                 RowBox[{"-", "2"}], "+", "sm"}], ")"}], " ", 
               RowBox[{"(", 
                RowBox[{
                 RowBox[{"-", "2"}], "+", "sm", "-", "\[Delta]", "+", 
                 RowBox[{"sm", " ", "\[Delta]"}]}], ")"}]}], "+", 
              RowBox[{"b", " ", "h", " ", 
               SuperscriptBox[
                RowBox[{"(", 
                 RowBox[{
                  RowBox[{"-", "2"}], "+", "sm", "-", "\[Delta]", "+", 
                  RowBox[{"sm", " ", "\[Delta]"}]}], ")"}], "2"]}]}], 
             ")"}]}]}], ")"}]}], ")"}], "/", 
       RowBox[{"(", 
        RowBox[{"b", " ", "\[Alpha]", " ", 
         RowBox[{"(", 
          RowBox[{
           RowBox[{
            RowBox[{"-", "sm"}], " ", 
            RowBox[{"(", 
             RowBox[{
              RowBox[{"sm", " ", 
               RowBox[{"(", 
                RowBox[{
                 RowBox[{"-", "1"}], "+", "\[Delta]"}], ")"}]}], "-", 
              "\[Delta]"}], ")"}], " ", 
            SuperscriptBox[
             RowBox[{"(", 
              RowBox[{"sm", "-", "\[Delta]", "+", 
               RowBox[{"sm", " ", "\[Delta]"}]}], ")"}], "2"]}], "+", 
           RowBox[{
            SuperscriptBox["h", "2"], " ", 
            SuperscriptBox["sf", "3"], " ", 
            SuperscriptBox[
             RowBox[{"(", 
              RowBox[{
               RowBox[{"-", "2"}], "+", "sm", "-", "\[Delta]", "+", 
               RowBox[{"sm", " ", "\[Delta]"}]}], ")"}], "2"], " ", 
            RowBox[{"(", 
             RowBox[{
              RowBox[{
               RowBox[{"-", 
                RowBox[{"(", 
                 RowBox[{
                  RowBox[{"-", "1"}], "+", "sm"}], ")"}]}], " ", 
               RowBox[{"(", 
                RowBox[{"1", "+", "\[Delta]"}], ")"}]}], "+", 
              RowBox[{"h", " ", 
               RowBox[{"(", 
                RowBox[{
                 RowBox[{"-", "2"}], "+", "sm", "-", "\[Delta]", "+", 
                 RowBox[{"sm", " ", "\[Delta]"}]}], ")"}]}]}], ")"}]}], "+", 
           RowBox[{
            SuperscriptBox["sf", "2"], " ", 
            RowBox[{"(", 
             RowBox[{
              RowBox[{"8", " ", 
               SuperscriptBox[
                RowBox[{"(", 
                 RowBox[{
                  RowBox[{"-", "1"}], "+", "sm"}], ")"}], "2"], " ", 
               SuperscriptBox[
                RowBox[{"(", 
                 RowBox[{"1", "+", "\[Delta]"}], ")"}], "2"]}], "+", 
              RowBox[{"2", " ", "h", " ", 
               RowBox[{"(", 
                RowBox[{"8", "-", 
                 RowBox[{"11", " ", "sm"}], "+", 
                 RowBox[{"3", " ", 
                  SuperscriptBox["sm", "2"]}]}], ")"}], " ", 
               RowBox[{"(", 
                RowBox[{"1", "+", "\[Delta]"}], ")"}], " ", 
               RowBox[{"(", 
                RowBox[{
                 RowBox[{"-", "2"}], "+", "sm", "-", "\[Delta]", "+", 
                 RowBox[{"sm", " ", "\[Delta]"}]}], ")"}]}], "-", 
              RowBox[{
               SuperscriptBox["h", "2"], " ", 
               RowBox[{"(", 
                RowBox[{
                 RowBox[{"-", "8"}], "-", 
                 RowBox[{"sm", " ", 
                  RowBox[{"(", 
                   RowBox[{
                    RowBox[{"-", "6"}], "+", "\[Delta]"}], ")"}]}], "+", 
                 RowBox[{
                  SuperscriptBox["sm", "2"], " ", 
                  RowBox[{"(", 
                   RowBox[{
                    RowBox[{"-", "1"}], "+", "\[Delta]"}], ")"}]}]}], ")"}], 
               " ", 
               SuperscriptBox[
                RowBox[{"(", 
                 RowBox[{
                  RowBox[{"-", "2"}], "+", "sm", "-", "\[Delta]", "+", 
                  RowBox[{"sm", " ", "\[Delta]"}]}], ")"}], "2"]}]}], ")"}]}],
            "+", 
           RowBox[{"sf", " ", 
            RowBox[{"(", 
             RowBox[{"sm", "-", "\[Delta]", "+", 
              RowBox[{"sm", " ", "\[Delta]"}]}], ")"}], " ", 
            RowBox[{"(", 
             RowBox[{
              RowBox[{"2", " ", "h", " ", 
               SuperscriptBox["sm", "3"], " ", 
               RowBox[{"(", 
                RowBox[{
                 RowBox[{"-", "1"}], "+", 
                 SuperscriptBox["\[Delta]", "2"]}], ")"}]}], "+", 
              RowBox[{"\[Delta]", " ", 
               RowBox[{"(", 
                RowBox[{"1", "+", "\[Delta]", "-", 
                 RowBox[{"h", " ", 
                  RowBox[{"(", 
                   RowBox[{"2", "+", "\[Delta]"}], ")"}]}]}], ")"}]}], "+", 
              RowBox[{
               SuperscriptBox["sm", "2"], " ", 
               RowBox[{"(", 
                RowBox[{
                 RowBox[{"-", "5"}], "-", 
                 RowBox[{"4", " ", "\[Delta]"}], "+", 
                 SuperscriptBox["\[Delta]", "2"], "+", 
                 RowBox[{"h", " ", 
                  RowBox[{"(", 
                   RowBox[{"9", "-", 
                    RowBox[{"5", " ", 
                    SuperscriptBox["\[Delta]", "2"]}]}], ")"}]}]}], ")"}]}], 
              "+", 
              RowBox[{"sm", " ", 
               RowBox[{"(", 
                RowBox[{"5", "+", 
                 RowBox[{"3", " ", "\[Delta]"}], "-", 
                 RowBox[{"2", " ", 
                  SuperscriptBox["\[Delta]", "2"]}], "+", 
                 RowBox[{"2", " ", "h", " ", 
                  RowBox[{"(", 
                   RowBox[{
                    RowBox[{"-", "5"}], "+", "\[Delta]", "+", 
                    RowBox[{"2", " ", 
                    SuperscriptBox["\[Delta]", "2"]}]}], ")"}]}]}], ")"}]}]}],
              ")"}]}]}], ")"}]}], ")"}]}]}], ",", 
     RowBox[{"maleNum", "\[Rule]", 
      RowBox[{"-", 
       RowBox[{"(", 
        RowBox[{
         RowBox[{"(", 
          RowBox[{
           RowBox[{"(", 
            RowBox[{
             RowBox[{"b", " ", 
              SuperscriptBox["h", "2"], " ", 
              SuperscriptBox["sf", "2"], " ", 
              SuperscriptBox[
               RowBox[{"(", 
                RowBox[{
                 RowBox[{"-", "2"}], "+", "sm", "-", "\[Delta]", "+", 
                 RowBox[{"sm", " ", "\[Delta]"}]}], ")"}], "2"]}], "+", 
             RowBox[{
              RowBox[{"(", 
               RowBox[{"sm", "-", "\[Delta]", "+", 
                RowBox[{"sm", " ", "\[Delta]"}]}], ")"}], " ", 
              RowBox[{"(", 
               RowBox[{
                RowBox[{
                 RowBox[{"-", "b"}], " ", "\[Delta]"}], "+", 
                RowBox[{"sm", " ", 
                 RowBox[{"(", 
                  RowBox[{
                   RowBox[{"-", "4"}], "+", "b", "+", 
                   RowBox[{"b", " ", "\[Delta]"}]}], ")"}]}]}], ")"}]}], "-", 
             RowBox[{"2", " ", "sf", " ", 
              RowBox[{"(", 
               RowBox[{
                RowBox[{
                 RowBox[{"-", "4"}], " ", 
                 RowBox[{"(", 
                  RowBox[{
                   RowBox[{"-", "1"}], "+", "sm"}], ")"}], " ", 
                 RowBox[{"(", 
                  RowBox[{"1", "+", "\[Delta]"}], ")"}]}], "+", 
                RowBox[{"2", " ", "b", " ", 
                 RowBox[{"(", 
                  RowBox[{
                   RowBox[{"-", "1"}], "+", "sm"}], ")"}], " ", 
                 RowBox[{"(", 
                  RowBox[{"1", "+", "\[Delta]"}], ")"}]}], "-", 
                RowBox[{"2", " ", "h", " ", 
                 RowBox[{"(", 
                  RowBox[{
                   RowBox[{"-", "2"}], "+", "sm"}], ")"}], " ", 
                 RowBox[{"(", 
                  RowBox[{
                   RowBox[{"-", "2"}], "+", "sm", "-", "\[Delta]", "+", 
                   RowBox[{"sm", " ", "\[Delta]"}]}], ")"}]}], "+", 
                RowBox[{"b", " ", "h", " ", 
                 SuperscriptBox[
                  RowBox[{"(", 
                   RowBox[{
                    RowBox[{"-", "2"}], "+", "sm", "-", "\[Delta]", "+", 
                    RowBox[{"sm", " ", "\[Delta]"}]}], ")"}], "2"]}]}], 
               ")"}]}]}], ")"}], " ", 
           RowBox[{"(", 
            RowBox[{
             RowBox[{"sm", " ", 
              RowBox[{"(", 
               RowBox[{
                RowBox[{"sm", " ", 
                 RowBox[{"(", 
                  RowBox[{
                   RowBox[{"-", "1"}], "+", "\[Delta]"}], ")"}]}], "-", 
                "\[Delta]"}], ")"}], " ", 
              SuperscriptBox[
               RowBox[{"(", 
                RowBox[{"sm", "-", "\[Delta]", "+", 
                 RowBox[{"sm", " ", "\[Delta]"}]}], ")"}], "2"]}], "+", 
             RowBox[{
              SuperscriptBox["sf", "2"], " ", 
              RowBox[{"(", 
               RowBox[{
                RowBox[{
                 RowBox[{"-", "4"}], " ", 
                 SuperscriptBox[
                  RowBox[{"(", 
                   RowBox[{
                    RowBox[{"-", "1"}], "+", "sm"}], ")"}], "2"], " ", 
                 SuperscriptBox[
                  RowBox[{"(", 
                   RowBox[{"1", "+", "\[Delta]"}], ")"}], "2"]}], "+", 
                RowBox[{
                 SuperscriptBox["h", "2"], " ", 
                 RowBox[{"(", 
                  RowBox[{
                   RowBox[{"-", "2"}], "+", "sm"}], ")"}], " ", 
                 RowBox[{"(", 
                  RowBox[{"2", "+", 
                   RowBox[{"sm", " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{"-", "1"}], "+", "\[Delta]"}], ")"}]}], "-", 
                   "\[Delta]"}], ")"}], " ", 
                 SuperscriptBox[
                  RowBox[{"(", 
                   RowBox[{
                    RowBox[{"-", "2"}], "+", "sm", "-", "\[Delta]", "+", 
                    RowBox[{"sm", " ", "\[Delta]"}]}], ")"}], "2"]}], "+", 
                RowBox[{"2", " ", "h", " ", 
                 RowBox[{"(", 
                  RowBox[{
                   RowBox[{"-", "1"}], "+", "sm"}], ")"}], " ", 
                 RowBox[{"(", 
                  RowBox[{"1", "+", "\[Delta]"}], ")"}], " ", 
                 RowBox[{"(", 
                  RowBox[{
                   RowBox[{"-", "8"}], "-", 
                   RowBox[{"2", " ", "\[Delta]"}], "+", 
                   SuperscriptBox["\[Delta]", "2"], "+", 
                   RowBox[{"sm", " ", 
                    RowBox[{"(", 
                    RowBox[{"8", "+", 
                    RowBox[{"3", " ", "\[Delta]"}], "-", 
                    RowBox[{"2", " ", 
                    SuperscriptBox["\[Delta]", "2"]}]}], ")"}]}], "+", 
                   RowBox[{
                    SuperscriptBox["sm", "2"], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{"-", "2"}], "-", "\[Delta]", "+", 
                    SuperscriptBox["\[Delta]", "2"]}], ")"}]}]}], ")"}]}]}], 
               ")"}]}], "-", 
             RowBox[{"2", " ", "sf", " ", 
              RowBox[{"(", 
               RowBox[{"sm", "-", "\[Delta]", "+", 
                RowBox[{"sm", " ", "\[Delta]"}]}], ")"}], " ", 
              RowBox[{"(", 
               RowBox[{
                RowBox[{"h", " ", 
                 SuperscriptBox["sm", "3"], " ", 
                 RowBox[{"(", 
                  RowBox[{
                   RowBox[{"-", "1"}], "+", 
                   SuperscriptBox["\[Delta]", "2"]}], ")"}]}], "+", 
                RowBox[{"\[Delta]", " ", 
                 RowBox[{"(", 
                  RowBox[{"1", "+", "\[Delta]", "-", 
                   RowBox[{"h", " ", 
                    RowBox[{"(", 
                    RowBox[{"2", "+", "\[Delta]"}], ")"}]}]}], ")"}]}], "-", 
                RowBox[{
                 SuperscriptBox["sm", "2"], " ", 
                 RowBox[{"(", 
                  RowBox[{"2", "+", "\[Delta]", "-", 
                   SuperscriptBox["\[Delta]", "2"], "+", 
                   RowBox[{"h", " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{"-", "4"}], "+", "\[Delta]", "+", 
                    RowBox[{"3", " ", 
                    SuperscriptBox["\[Delta]", "2"]}]}], ")"}]}]}], ")"}]}], 
                "+", 
                RowBox[{"sm", " ", 
                 RowBox[{"(", 
                  RowBox[{"2", "-", 
                   RowBox[{"2", " ", 
                    SuperscriptBox["\[Delta]", "2"]}], "+", 
                   RowBox[{"h", " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{"-", "4"}], "+", 
                    RowBox[{"3", " ", "\[Delta]"}], "+", 
                    RowBox[{"3", " ", 
                    SuperscriptBox["\[Delta]", "2"]}]}], ")"}]}]}], ")"}]}]}],
                ")"}]}]}], ")"}]}], ")"}], "/", 
         RowBox[{"(", 
          RowBox[{"b", " ", "\[Alpha]", " ", 
           RowBox[{"(", 
            RowBox[{
             RowBox[{
              RowBox[{"-", "sm"}], " ", 
              RowBox[{"(", 
               RowBox[{
                RowBox[{"sm", " ", 
                 RowBox[{"(", 
                  RowBox[{
                   RowBox[{"-", "1"}], "+", "\[Delta]"}], ")"}]}], "-", 
                "\[Delta]"}], ")"}], " ", 
              SuperscriptBox[
               RowBox[{"(", 
                RowBox[{"sm", "-", "\[Delta]", "+", 
                 RowBox[{"sm", " ", "\[Delta]"}]}], ")"}], "4"]}], "+", 
             RowBox[{
              SuperscriptBox["h", "4"], " ", 
              SuperscriptBox["sf", "5"], " ", 
              SuperscriptBox[
               RowBox[{"(", 
                RowBox[{
                 RowBox[{"-", "2"}], "+", "sm", "-", "\[Delta]", "+", 
                 RowBox[{"sm", " ", "\[Delta]"}]}], ")"}], "4"], " ", 
              RowBox[{"(", 
               RowBox[{
                RowBox[{
                 RowBox[{"-", 
                  RowBox[{"(", 
                   RowBox[{
                    RowBox[{"-", "1"}], "+", "sm"}], ")"}]}], " ", 
                 RowBox[{"(", 
                  RowBox[{"1", "+", "\[Delta]"}], ")"}]}], "+", 
                RowBox[{"h", " ", 
                 RowBox[{"(", 
                  RowBox[{
                   RowBox[{"-", "2"}], "+", "sm", "-", "\[Delta]", "+", 
                   RowBox[{"sm", " ", "\[Delta]"}]}], ")"}]}]}], ")"}]}], "-", 
             RowBox[{
              SuperscriptBox["h", "2"], " ", 
              SuperscriptBox["sf", "4"], " ", 
              SuperscriptBox[
               RowBox[{"(", 
                RowBox[{
                 RowBox[{"-", "2"}], "+", "sm", "-", "\[Delta]", "+", 
                 RowBox[{"sm", " ", "\[Delta]"}]}], ")"}], "2"], " ", 
              RowBox[{"(", 
               RowBox[{
                RowBox[{
                 RowBox[{"-", "12"}], " ", 
                 SuperscriptBox[
                  RowBox[{"(", 
                   RowBox[{
                    RowBox[{"-", "1"}], "+", "sm"}], ")"}], "2"], " ", 
                 SuperscriptBox[
                  RowBox[{"(", 
                   RowBox[{"1", "+", "\[Delta]"}], ")"}], "2"]}], "+", 
                RowBox[{
                 SuperscriptBox["h", "2"], " ", 
                 SuperscriptBox[
                  RowBox[{"(", 
                   RowBox[{
                    RowBox[{"-", "2"}], "+", "sm", "-", "\[Delta]", "+", 
                    RowBox[{"sm", " ", "\[Delta]"}]}], ")"}], "2"], " ", 
                 RowBox[{"(", 
                  RowBox[{
                   RowBox[{
                    SuperscriptBox["sm", "2"], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{"-", "1"}], "+", "\[Delta]"}], ")"}]}], "-", 
                   RowBox[{"2", " ", 
                    RowBox[{"(", 
                    RowBox[{"6", "+", "\[Delta]"}], ")"}]}], "+", 
                   RowBox[{"sm", " ", 
                    RowBox[{"(", 
                    RowBox[{"8", "+", "\[Delta]"}], ")"}]}]}], ")"}]}], "-", 
                RowBox[{"2", " ", "h", " ", 
                 RowBox[{"(", 
                  RowBox[{
                   RowBox[{"-", "1"}], "+", "sm"}], ")"}], " ", 
                 RowBox[{"(", 
                  RowBox[{"1", "+", "\[Delta]"}], ")"}], " ", 
                 RowBox[{"(", 
                  RowBox[{"24", "+", 
                   RowBox[{"14", " ", "\[Delta]"}], "+", 
                   SuperscriptBox["\[Delta]", "2"], "+", 
                   RowBox[{
                    SuperscriptBox["sm", "2"], " ", 
                    RowBox[{"(", 
                    RowBox[{"4", "+", 
                    RowBox[{"5", " ", "\[Delta]"}], "+", 
                    SuperscriptBox["\[Delta]", "2"]}], ")"}]}], "-", 
                   RowBox[{"sm", " ", 
                    RowBox[{"(", 
                    RowBox[{"20", "+", 
                    RowBox[{"19", " ", "\[Delta]"}], "+", 
                    RowBox[{"2", " ", 
                    SuperscriptBox["\[Delta]", "2"]}]}], ")"}]}]}], ")"}]}]}],
                ")"}]}], "+", 
             RowBox[{"2", " ", 
              SuperscriptBox["sf", "3"], " ", 
              RowBox[{"(", 
               RowBox[{
                RowBox[{
                 RowBox[{"-", "16"}], " ", 
                 SuperscriptBox[
                  RowBox[{"(", 
                   RowBox[{
                    RowBox[{"-", "1"}], "+", "sm"}], ")"}], "3"], " ", 
                 SuperscriptBox[
                  RowBox[{"(", 
                   RowBox[{"1", "+", "\[Delta]"}], ")"}], "3"]}], "-", 
                RowBox[{
                 SuperscriptBox["h", "2"], " ", 
                 RowBox[{"(", 
                  RowBox[{
                   RowBox[{"-", "1"}], "+", "sm"}], ")"}], " ", 
                 RowBox[{"(", 
                  RowBox[{"1", "+", "\[Delta]"}], ")"}], " ", 
                 SuperscriptBox[
                  RowBox[{"(", 
                   RowBox[{
                    RowBox[{"-", "2"}], "+", "sm", "-", "\[Delta]", "+", 
                    RowBox[{"sm", " ", "\[Delta]"}]}], ")"}], "2"], " ", 
                 RowBox[{"(", 
                  RowBox[{
                   RowBox[{"16", " ", 
                    RowBox[{"(", 
                    RowBox[{"3", "+", "\[Delta]"}], ")"}]}], "+", 
                   RowBox[{
                    SuperscriptBox["sm", "2"], " ", 
                    RowBox[{"(", 
                    RowBox[{"11", "+", 
                    RowBox[{"7", " ", "\[Delta]"}]}], ")"}]}], "-", 
                   RowBox[{"sm", " ", 
                    RowBox[{"(", 
                    RowBox[{"40", "+", 
                    RowBox[{"23", " ", "\[Delta]"}]}], ")"}]}]}], ")"}]}], 
                "+", 
                RowBox[{
                 SuperscriptBox["h", "3"], " ", 
                 SuperscriptBox[
                  RowBox[{"(", 
                   RowBox[{
                    RowBox[{"-", "2"}], "+", "sm", "-", "\[Delta]", "+", 
                    RowBox[{"sm", " ", "\[Delta]"}]}], ")"}], "3"], " ", 
                 RowBox[{"(", 
                  RowBox[{
                   RowBox[{"8", " ", 
                    RowBox[{"(", 
                    RowBox[{"2", "+", "\[Delta]"}], ")"}]}], "+", 
                   RowBox[{
                    SuperscriptBox["sm", "2"], " ", 
                    RowBox[{"(", 
                    RowBox[{"11", "+", 
                    RowBox[{"7", " ", "\[Delta]"}], "-", 
                    RowBox[{"4", " ", 
                    SuperscriptBox["\[Delta]", "2"]}]}], ")"}]}], "+", 
                   RowBox[{"2", " ", 
                    SuperscriptBox["sm", "3"], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{"-", "1"}], "+", 
                    SuperscriptBox["\[Delta]", "2"]}], ")"}]}], "+", 
                   RowBox[{"sm", " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{"-", "20"}], "-", 
                    RowBox[{"15", " ", "\[Delta]"}], "+", 
                    RowBox[{"2", " ", 
                    SuperscriptBox["\[Delta]", "2"]}]}], ")"}]}]}], ")"}]}], 
                "-", 
                RowBox[{"4", " ", "h", " ", 
                 SuperscriptBox[
                  RowBox[{"(", 
                   RowBox[{
                    RowBox[{"-", "1"}], "+", "sm"}], ")"}], "2"], " ", 
                 SuperscriptBox[
                  RowBox[{"(", 
                   RowBox[{"1", "+", "\[Delta]"}], ")"}], "2"], " ", 
                 RowBox[{"(", 
                  RowBox[{
                   RowBox[{"2", " ", 
                    RowBox[{"(", 
                    RowBox[{"12", "+", 
                    RowBox[{"8", " ", "\[Delta]"}], "+", 
                    SuperscriptBox["\[Delta]", "2"]}], ")"}]}], "+", 
                   RowBox[{
                    SuperscriptBox["sm", "2"], " ", 
                    RowBox[{"(", 
                    RowBox[{"5", "+", 
                    RowBox[{"7", " ", "\[Delta]"}], "+", 
                    RowBox[{"2", " ", 
                    SuperscriptBox["\[Delta]", "2"]}]}], ")"}]}], "-", 
                   RowBox[{"sm", " ", 
                    RowBox[{"(", 
                    RowBox[{"22", "+", 
                    RowBox[{"23", " ", "\[Delta]"}], "+", 
                    RowBox[{"4", " ", 
                    SuperscriptBox["\[Delta]", "2"]}]}], ")"}]}]}], ")"}]}]}],
                ")"}]}], "+", 
             RowBox[{"sf", " ", 
              SuperscriptBox[
               RowBox[{"(", 
                RowBox[{"sm", "-", "\[Delta]", "+", 
                 RowBox[{"sm", " ", "\[Delta]"}]}], ")"}], "2"], " ", 
              RowBox[{"(", 
               RowBox[{
                RowBox[{"4", " ", "h", " ", 
                 SuperscriptBox["sm", "4"], " ", 
                 RowBox[{"(", 
                  RowBox[{
                   RowBox[{"-", "1"}], "+", "\[Delta]"}], ")"}], " ", 
                 SuperscriptBox[
                  RowBox[{"(", 
                   RowBox[{"1", "+", "\[Delta]"}], ")"}], "2"]}], "+", 
                RowBox[{"sm", " ", 
                 SuperscriptBox["\[Delta]", "2"], " ", 
                 RowBox[{"(", 
                  RowBox[{"3", "-", 
                   RowBox[{"13", " ", "h"}], "+", 
                   RowBox[{"3", " ", "\[Delta]"}], "-", 
                   RowBox[{"7", " ", "h", " ", "\[Delta]"}]}], ")"}]}], "+", 
                RowBox[{
                 SuperscriptBox["\[Delta]", "2"], " ", 
                 RowBox[{"(", 
                  RowBox[{
                   RowBox[{"-", "1"}], "-", "\[Delta]", "+", 
                   RowBox[{"h", " ", 
                    RowBox[{"(", 
                    RowBox[{"2", "+", "\[Delta]"}], ")"}]}]}], ")"}]}], "+", 
                RowBox[{"3", " ", 
                 SuperscriptBox["sm", "2"], " ", 
                 RowBox[{"(", 
                  RowBox[{"1", "+", "\[Delta]"}], ")"}], " ", 
                 RowBox[{"(", 
                  RowBox[{"3", "-", 
                   SuperscriptBox["\[Delta]", "2"], "+", 
                   RowBox[{"h", " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{"-", "6"}], "+", 
                    RowBox[{"3", " ", "\[Delta]"}], "+", 
                    RowBox[{"5", " ", 
                    SuperscriptBox["\[Delta]", "2"]}]}], ")"}]}]}], ")"}]}], 
                "-", 
                RowBox[{
                 SuperscriptBox["sm", "3"], " ", 
                 RowBox[{"(", 
                  RowBox[{"1", "+", "\[Delta]"}], ")"}], " ", 
                 RowBox[{"(", 
                  RowBox[{"9", "-", 
                   SuperscriptBox["\[Delta]", "2"], "+", 
                   RowBox[{"h", " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{"-", "17"}], "+", 
                    RowBox[{"4", " ", "\[Delta]"}], "+", 
                    RowBox[{"13", " ", 
                    SuperscriptBox["\[Delta]", "2"]}]}], ")"}]}]}], ")"}]}]}],
                ")"}]}], "-", 
             RowBox[{"2", " ", 
              SuperscriptBox["sf", "2"], " ", 
              RowBox[{"(", 
               RowBox[{"sm", "-", "\[Delta]", "+", 
                RowBox[{"sm", " ", "\[Delta]"}]}], ")"}], " ", 
              RowBox[{"(", 
               RowBox[{
                RowBox[{"3", " ", 
                 SuperscriptBox["h", "2"], " ", 
                 SuperscriptBox["sm", "5"], " ", 
                 RowBox[{"(", 
                  RowBox[{
                   RowBox[{"-", "1"}], "+", "\[Delta]"}], ")"}], " ", 
                 SuperscriptBox[
                  RowBox[{"(", 
                   RowBox[{"1", "+", "\[Delta]"}], ")"}], "3"]}], "-", 
                RowBox[{"h", " ", 
                 SuperscriptBox["sm", "4"], " ", 
                 SuperscriptBox[
                  RowBox[{"(", 
                   RowBox[{"1", "+", "\[Delta]"}], ")"}], "2"], " ", 
                 RowBox[{"(", 
                  RowBox[{"12", "+", 
                   RowBox[{"3", " ", "\[Delta]"}], "-", 
                   SuperscriptBox["\[Delta]", "2"], "+", 
                   RowBox[{"h", " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{"-", "24"}], "+", 
                    RowBox[{"3", " ", "\[Delta]"}], "+", 
                    RowBox[{"13", " ", 
                    SuperscriptBox["\[Delta]", "2"]}]}], ")"}]}]}], ")"}]}], 
                "+", 
                RowBox[{"\[Delta]", " ", 
                 RowBox[{"(", 
                  RowBox[{
                   RowBox[{"2", " ", 
                    SuperscriptBox[
                    RowBox[{"(", 
                    RowBox[{"1", "+", "\[Delta]"}], ")"}], "2"]}], "-", 
                   RowBox[{
                    SuperscriptBox["h", "2"], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{"-", "2"}], "+", "\[Delta]"}], ")"}], " ", 
                    SuperscriptBox[
                    RowBox[{"(", 
                    RowBox[{"2", "+", "\[Delta]"}], ")"}], "2"]}], "+", 
                   RowBox[{"h", " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{"-", "8"}], "-", 
                    RowBox[{"10", " ", "\[Delta]"}], "-", 
                    SuperscriptBox["\[Delta]", "2"], "+", 
                    SuperscriptBox["\[Delta]", "3"]}], ")"}]}]}], ")"}]}], 
                "+", 
                RowBox[{
                 SuperscriptBox["sm", "3"], " ", 
                 RowBox[{"(", 
                  RowBox[{"1", "+", "\[Delta]"}], ")"}], " ", 
                 RowBox[{"(", 
                  RowBox[{
                   RowBox[{
                    RowBox[{"-", "2"}], " ", 
                    RowBox[{"(", 
                    RowBox[{"7", "+", 
                    RowBox[{"8", " ", "\[Delta]"}], "+", 
                    SuperscriptBox["\[Delta]", "2"]}], ")"}]}], "+", 
                   RowBox[{"h", " ", 
                    RowBox[{"(", 
                    RowBox[{"64", "+", 
                    RowBox[{"68", " ", "\[Delta]"}], "+", 
                    RowBox[{"8", " ", 
                    SuperscriptBox["\[Delta]", "2"]}], "-", 
                    RowBox[{"4", " ", 
                    SuperscriptBox["\[Delta]", "3"]}]}], ")"}]}], "+", 
                   RowBox[{
                    SuperscriptBox["h", "2"], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{"-", "74"}], "-", 
                    RowBox[{"55", " ", "\[Delta]"}], "+", 
                    RowBox[{"31", " ", 
                    SuperscriptBox["\[Delta]", "2"]}], "+", 
                    RowBox[{"22", " ", 
                    SuperscriptBox["\[Delta]", "3"]}]}], ")"}]}]}], ")"}]}], 
                "+", 
                RowBox[{
                 SuperscriptBox["sm", "2"], " ", 
                 RowBox[{"(", 
                  RowBox[{
                   RowBox[{"2", " ", 
                    SuperscriptBox[
                    RowBox[{"(", 
                    RowBox[{"1", "+", "\[Delta]"}], ")"}], "2"], " ", 
                    RowBox[{"(", 
                    RowBox[{"14", "+", 
                    RowBox[{"3", " ", "\[Delta]"}]}], ")"}]}], "+", 
                   RowBox[{
                    SuperscriptBox["h", "2"], " ", 
                    RowBox[{"(", 
                    RowBox[{"104", "+", 
                    RowBox[{"166", " ", "\[Delta]"}], "+", 
                    RowBox[{"37", " ", 
                    SuperscriptBox["\[Delta]", "2"]}], "-", 
                    RowBox[{"45", " ", 
                    SuperscriptBox["\[Delta]", "3"]}], "-", 
                    RowBox[{"18", " ", 
                    SuperscriptBox["\[Delta]", "4"]}]}], ")"}]}], "+", 
                   RowBox[{"3", " ", "h", " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{"-", "36"}], "-", 
                    RowBox[{"69", " ", "\[Delta]"}], "-", 
                    RowBox[{"37", " ", 
                    SuperscriptBox["\[Delta]", "2"]}], "-", 
                    RowBox[{"2", " ", 
                    SuperscriptBox["\[Delta]", "3"]}], "+", 
                    RowBox[{"2", " ", 
                    SuperscriptBox["\[Delta]", "4"]}]}], ")"}]}]}], ")"}]}], 
                "+", 
                RowBox[{"sm", " ", 
                 RowBox[{"(", 
                  RowBox[{
                   RowBox[{
                    RowBox[{"-", "2"}], " ", 
                    SuperscriptBox[
                    RowBox[{"(", 
                    RowBox[{"1", "+", "\[Delta]"}], ")"}], "2"], " ", 
                    RowBox[{"(", 
                    RowBox[{"7", "+", 
                    RowBox[{"3", " ", "\[Delta]"}]}], ")"}]}], "+", 
                   RowBox[{"h", " ", 
                    RowBox[{"(", 
                    RowBox[{"56", "+", 
                    RowBox[{"110", " ", "\[Delta]"}], "+", 
                    RowBox[{"62", " ", 
                    SuperscriptBox["\[Delta]", "2"]}], "+", 
                    RowBox[{"4", " ", 
                    SuperscriptBox["\[Delta]", "3"]}], "-", 
                    RowBox[{"4", " ", 
                    SuperscriptBox["\[Delta]", "4"]}]}], ")"}]}], "+", 
                   RowBox[{
                    SuperscriptBox["h", "2"], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{"-", "56"}], "-", 
                    RowBox[{"84", " ", "\[Delta]"}], "-", 
                    RowBox[{"22", " ", 
                    SuperscriptBox["\[Delta]", "2"]}], "+", 
                    RowBox[{"17", " ", 
                    SuperscriptBox["\[Delta]", "3"]}], "+", 
                    RowBox[{"7", " ", 
                    SuperscriptBox["\[Delta]", "4"]}]}], ")"}]}]}], ")"}]}]}],
                ")"}]}]}], ")"}]}], ")"}]}], ")"}]}]}]}], "}"}]}], 
  "}"}]], "Output",
 CellChangeTimes->{{3.781460169945362*^9, 3.781460181056621*^9}, {
   3.781460263051641*^9, 3.781460298327677*^9}, {3.7814604223927298`*^9, 
   3.7814604321227903`*^9}, {3.781884936146596*^9, 3.781884950383851*^9}, 
   3.781888143776909*^9, 3.781888779399761*^9, 3.781952838781995*^9, 
   3.781952912022429*^9, 3.781953909942618*^9, 3.78195394409758*^9, 
   3.7821306694949303`*^9, 3.783078910864347*^9, 3.785152392323344*^9, 
   3.785162725603114*^9, 3.785164902578751*^9, 3.785164953595057*^9, 
   3.785164992909616*^9, 3.785165144424354*^9, {3.785242217151107*^9, 
   3.785242356886571*^9}, 3.78557243807008*^9, 3.785588400664415*^9, 
   3.789732682930251*^9, 3.7900785884225082`*^9, 3.795174674139629*^9, 
   3.799753847213955*^9, 3.7997540084461193`*^9, 3.800007973864676*^9, 
   3.800268200382759*^9, 3.813928149298214*^9, 3.814019806603877*^9, 
   3.8140813237764463`*^9, 3.814189397961047*^9, 3.814195962054716*^9, 
   3.81419651136869*^9},
 CellLabel->
  "Out[1290]=",ExpressionUUID->"9eae98eb-9b60-4173-be0b-3b853ab230b0"],

Cell[BoxData[
 RowBox[{"{", 
  RowBox[{"0", ",", 
   RowBox[{
    RowBox[{"(", 
     RowBox[{
      RowBox[{"b", " ", 
       SuperscriptBox["h", "2"], " ", 
       SuperscriptBox["sf", "2"], " ", 
       SuperscriptBox[
        RowBox[{"(", 
         RowBox[{
          RowBox[{"-", "2"}], "+", "sm", "+", 
          RowBox[{
           RowBox[{"(", 
            RowBox[{
             RowBox[{"-", "1"}], "+", "sm"}], ")"}], " ", "\[Delta]"}]}], 
         ")"}], "2"]}], "+", 
      RowBox[{
       RowBox[{"(", 
        RowBox[{"sm", "+", 
         RowBox[{
          RowBox[{"(", 
           RowBox[{
            RowBox[{"-", "1"}], "+", "sm"}], ")"}], " ", "\[Delta]"}]}], 
        ")"}], " ", 
       RowBox[{"(", 
        RowBox[{
         RowBox[{
          RowBox[{"(", 
           RowBox[{
            RowBox[{"-", "4"}], "+", "b"}], ")"}], " ", "sm"}], "+", 
         RowBox[{"b", " ", 
          RowBox[{"(", 
           RowBox[{
            RowBox[{"-", "1"}], "+", "sm"}], ")"}], " ", "\[Delta]"}]}], 
        ")"}]}], "-", 
      RowBox[{"2", " ", "sf", " ", 
       RowBox[{"(", 
        RowBox[{
         RowBox[{
          RowBox[{"-", "4"}], " ", 
          RowBox[{"(", 
           RowBox[{
            RowBox[{"-", "1"}], "+", "sm"}], ")"}], " ", 
          RowBox[{"(", 
           RowBox[{"1", "+", "\[Delta]"}], ")"}]}], "+", 
         RowBox[{"2", " ", "b", " ", 
          RowBox[{"(", 
           RowBox[{
            RowBox[{"-", "1"}], "+", "sm"}], ")"}], " ", 
          RowBox[{"(", 
           RowBox[{"1", "+", "\[Delta]"}], ")"}]}], "-", 
         RowBox[{"2", " ", "h", " ", 
          RowBox[{"(", 
           RowBox[{
            RowBox[{"-", "2"}], "+", "sm"}], ")"}], " ", 
          RowBox[{"(", 
           RowBox[{
            RowBox[{"-", "2"}], "+", "sm", "+", 
            RowBox[{
             RowBox[{"(", 
              RowBox[{
               RowBox[{"-", "1"}], "+", "sm"}], ")"}], " ", "\[Delta]"}]}], 
           ")"}]}], "+", 
         RowBox[{"b", " ", "h", " ", 
          SuperscriptBox[
           RowBox[{"(", 
            RowBox[{
             RowBox[{"-", "2"}], "+", "sm", "+", 
             RowBox[{
              RowBox[{"(", 
               RowBox[{
                RowBox[{"-", "1"}], "+", "sm"}], ")"}], " ", "\[Delta]"}]}], 
            ")"}], "2"]}]}], ")"}]}]}], ")"}], "/", 
    RowBox[{"(", 
     RowBox[{"b", " ", "\[Alpha]", " ", 
      RowBox[{"(", 
       RowBox[{
        RowBox[{
         RowBox[{"-", "4"}], " ", "sf", " ", 
         RowBox[{"(", 
          RowBox[{
           RowBox[{"-", "1"}], "+", "sm"}], ")"}], " ", 
         RowBox[{"(", 
          RowBox[{"1", "+", "\[Delta]"}], ")"}]}], "-", 
        RowBox[{"2", " ", "h", " ", "sf", " ", 
         SuperscriptBox[
          RowBox[{"(", 
           RowBox[{
            RowBox[{"-", "2"}], "+", "sm", "+", 
            RowBox[{
             RowBox[{"(", 
              RowBox[{
               RowBox[{"-", "1"}], "+", "sm"}], ")"}], " ", "\[Delta]"}]}], 
           ")"}], "2"]}], "+", 
        RowBox[{
         SuperscriptBox["h", "2"], " ", 
         SuperscriptBox["sf", "2"], " ", 
         SuperscriptBox[
          RowBox[{"(", 
           RowBox[{
            RowBox[{"-", "2"}], "+", "sm", "+", 
            RowBox[{
             RowBox[{"(", 
              RowBox[{
               RowBox[{"-", "1"}], "+", "sm"}], ")"}], " ", "\[Delta]"}]}], 
           ")"}], "2"]}], "+", 
        SuperscriptBox[
         RowBox[{"(", 
          RowBox[{"sm", "+", 
           RowBox[{
            RowBox[{"(", 
             RowBox[{
              RowBox[{"-", "1"}], "+", "sm"}], ")"}], " ", "\[Delta]"}]}], 
          ")"}], "2"]}], ")"}]}], ")"}]}]}], "}"}]], "Output",
 CellChangeTimes->{{3.781460169945362*^9, 3.781460181056621*^9}, {
   3.781460263051641*^9, 3.781460298327677*^9}, {3.7814604223927298`*^9, 
   3.7814604321227903`*^9}, {3.781884936146596*^9, 3.781884950383851*^9}, 
   3.781888143776909*^9, 3.781888779399761*^9, 3.781952838781995*^9, 
   3.781952912022429*^9, 3.781953909942618*^9, 3.78195394409758*^9, 
   3.7821306694949303`*^9, 3.783078910864347*^9, 3.785152392323344*^9, 
   3.785162725603114*^9, 3.785164902578751*^9, 3.785164953595057*^9, 
   3.785164992909616*^9, 3.785165144424354*^9, {3.785242217151107*^9, 
   3.785242356886571*^9}, 3.78557243807008*^9, 3.785588400664415*^9, 
   3.789732682930251*^9, 3.7900785884225082`*^9, 3.795174674139629*^9, 
   3.799753847213955*^9, 3.7997540084461193`*^9, 3.800007973864676*^9, 
   3.800268200382759*^9, 3.813928149298214*^9, 3.814019806603877*^9, 
   3.8140813237764463`*^9, 3.814189397961047*^9, 3.814195962054716*^9, 
   3.814196512406578*^9},
 CellLabel->
  "Out[1291]=",ExpressionUUID->"5e2aab2e-0478-4fbb-89c8-bda8cb3001c5"]
}, Open  ]],

Cell["\<\
Verify this is an equilibrium by seeing if the difference equations are 0 at \
this point\
\>", "Text",
 CellChangeTimes->{{3.7814604537964993`*^9, 
  3.781460476007682*^9}},ExpressionUUID->"bb02e358-4fa4-428e-afcb-\
5f3933e40e86"],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"Simplify", "[", 
  RowBox[{
   RowBox[{
    RowBox[{"solvemeFull", "/.", "subeq"}], "/.", "equilibrium"}], "/.", 
   RowBox[{"equilPopSizeSol", "[", 
    RowBox[{"[", "2", "]"}], "]"}]}], "]"}]], "Input",
 CellChangeTimes->{{3.781460345153535*^9, 3.7814603578425694`*^9}, {
  3.781884993039777*^9, 3.781884994226761*^9}, {3.8140084876620626`*^9, 
  3.814008488188561*^9}},
 CellLabel->
  "In[1292]:=",ExpressionUUID->"5fd12814-a562-43f7-bff3-28f08afc2467"],

Cell[BoxData[
 RowBox[{"{", 
  RowBox[{"{", 
   RowBox[{"{", 
    RowBox[{"0", ",", "0", ",", "0", ",", "0", ",", "0"}], "}"}], "}"}], 
  "}"}]], "Output",
 CellChangeTimes->{{3.781460334037698*^9, 3.7814603581871443`*^9}, 
   3.781884996207065*^9, 3.781888151904563*^9, 3.781888815502838*^9, 
   3.7819528391580257`*^9, 3.78195291236128*^9, 3.781953910237885*^9, 
   3.7819539443843307`*^9, 3.782130669814548*^9, 3.7830789112925663`*^9, 
   3.785152392831718*^9, 3.7851627262218847`*^9, 3.785164912505734*^9, 
   3.785164954157558*^9, 3.785164993514278*^9, 3.78557243998604*^9, 
   3.785588402949622*^9, 3.789732685214398*^9, 3.790078589989867*^9, 
   3.795174675666205*^9, 3.800007975600398*^9, 3.800268201899416*^9, 
   3.813928155463936*^9, 3.814019811394556*^9, 3.814081332227298*^9, 
   3.814189399423977*^9, 3.8141959633935623`*^9, 3.814196512640952*^9},
 CellLabel->
  "Out[1292]=",ExpressionUUID->"72e30847-6871-4d15-9421-a242db0bbb9d"]
}, Open  ]],

Cell["\<\
Define the \[CurlyPhi] and \[Psi] terms and show these can be used to express \
the within-sex allele frequencies.\
\>", "Text",
 CellChangeTimes->{{3.8000080598283653`*^9, 3.800008145978733*^9}, {
  3.8140813483445797`*^9, 
  3.814081359244545*^9}},ExpressionUUID->"31a4b22f-1b38-4dbb-b93e-\
621a9cd9e860"],

Cell[CellGroupData[{

Cell[BoxData[{
 RowBox[{
  RowBox[{"\[Phi]", " ", "=", " ", 
   RowBox[{"\[Delta]", " ", "-", " ", 
    RowBox[{"(", 
     RowBox[{"sm", "+", 
      RowBox[{"sm", " ", "\[Delta]"}], "+", 
      RowBox[{"h", " ", "sf", " ", 
       RowBox[{"(", 
        RowBox[{"2", "+", "\[Delta]", "-", 
         RowBox[{"sm", " ", 
          RowBox[{"(", 
           RowBox[{"1", "+", "\[Delta]"}], ")"}]}]}], ")"}]}]}], ")"}]}]}], 
  ";"}], "\[IndentingNewLine]", 
 RowBox[{
  RowBox[{"\[CurlyPhi]", "=", " ", 
   RowBox[{
    RowBox[{"sm", " ", 
     RowBox[{"(", 
      RowBox[{"1", "+", "\[Delta]"}], ")"}]}], "-", 
    RowBox[{"(", 
     RowBox[{"\[Delta]", "+", 
      RowBox[{"2", " ", "sf", " ", 
       RowBox[{"(", 
        RowBox[{
         RowBox[{"-", "1"}], "+", "sm"}], ")"}], " ", 
       RowBox[{"(", 
        RowBox[{"1", "+", "\[Delta]"}], ")"}]}], "+", 
      RowBox[{"h", " ", "sf", " ", 
       RowBox[{"(", 
        RowBox[{"2", "+", "\[Delta]", "-", 
         RowBox[{"sm", " ", 
          RowBox[{"(", 
           RowBox[{"1", "+", "\[Delta]"}], ")"}]}]}], ")"}]}]}], ")"}]}]}], 
  ";"}], "\[IndentingNewLine]", 
 RowBox[{
  RowBox[{"pfeq", "  ", "=", " ", 
   RowBox[{
    RowBox[{"\[Phi]", "/", 
     RowBox[{"(", 
      RowBox[{"\[Phi]", "+", "\[CurlyPhi]"}], ")"}]}], "//", 
    "FullSimplify"}]}], ";"}], "\[IndentingNewLine]", 
 RowBox[{
  RowBox[{
   RowBox[{"pfeq", "-", " ", "pf"}], "/.", 
   RowBox[{"equilibrium", "[", 
    RowBox[{"[", "1", "]"}], "]"}]}], " ", "//", 
  "Factor"}], "\[IndentingNewLine]", 
 RowBox[{
  RowBox[{"pmeq", "  ", "=", " ", 
   RowBox[{
    RowBox[{
     RowBox[{"(", 
      RowBox[{"1", "-", "sm"}], ")"}], 
     RowBox[{"\[Phi]", "/", 
      RowBox[{"(", 
       RowBox[{
        RowBox[{
         RowBox[{"(", 
          RowBox[{"1", "-", "sm"}], ")"}], "\[Phi]"}], "+", "\[CurlyPhi]"}], 
       ")"}]}]}], "//", "Simplify"}]}], ";"}], "\[IndentingNewLine]", 
 RowBox[{
  RowBox[{
   RowBox[{"pmeq", " ", "-", " ", "pm"}], "/.", 
   RowBox[{"equilibrium", "[", 
    RowBox[{"[", "1", "]"}], "]"}]}], " ", "//", "Factor"}]}], "Input",
 CellChangeTimes->{{3.785165011572194*^9, 3.7851650578862867`*^9}, {
  3.785165092314415*^9, 3.785165123216474*^9}, {3.785165839522647*^9, 
  3.785165968268564*^9}, {3.785588461811692*^9, 3.7855884676833487`*^9}, {
  3.785588514305647*^9, 3.785588514538514*^9}, {3.800008246803669*^9, 
  3.800008356026331*^9}, {3.800008520562923*^9, 3.80000854999323*^9}},
 CellLabel->
  "In[1293]:=",ExpressionUUID->"76e79208-a384-4c72-9517-ab46e19f04b3"],

Cell[BoxData["0"], "Output",
 CellChangeTimes->{
  3.785572440343376*^9, 3.785588403438594*^9, 3.7855884681264563`*^9, 
   3.785588516159246*^9, 3.789732685580689*^9, 3.790078590392893*^9, 
   3.79517467591718*^9, 3.8000079759268837`*^9, 3.800008276208026*^9, {
   3.800008310167995*^9, 3.800008356897874*^9}, {3.8000085249642057`*^9, 
   3.800008550752894*^9}, 3.800268202153144*^9, 3.813928166782549*^9, 
   3.814019811583646*^9, 3.814081338041605*^9, 3.8141893995826473`*^9, 
   3.8141959635283127`*^9, 3.8141965127469997`*^9},
 CellLabel->
  "Out[1296]=",ExpressionUUID->"ca5b5032-efd0-4446-b419-be905d55a573"],

Cell[BoxData["0"], "Output",
 CellChangeTimes->{
  3.785572440343376*^9, 3.785588403438594*^9, 3.7855884681264563`*^9, 
   3.785588516159246*^9, 3.789732685580689*^9, 3.790078590392893*^9, 
   3.79517467591718*^9, 3.8000079759268837`*^9, 3.800008276208026*^9, {
   3.800008310167995*^9, 3.800008356897874*^9}, {3.8000085249642057`*^9, 
   3.800008550752894*^9}, 3.800268202153144*^9, 3.813928166782549*^9, 
   3.814019811583646*^9, 3.814081338041605*^9, 3.8141893995826473`*^9, 
   3.8141959635283127`*^9, 3.814196512778748*^9},
 CellLabel->
  "Out[1298]=",ExpressionUUID->"6488e1bd-1c6f-4547-bf26-b7b41c73425f"]
}, Open  ]],

Cell["\<\
Then find the equilibrium population size and polymorphism and show it can be \
written with b = b* as in the manuscript\
\>", "Text",
 CellChangeTimes->{{3.800008557765349*^9, 3.800008607425334*^9}, {
  3.800008776745749*^9, 
  3.800008778657819*^9}},ExpressionUUID->"5ff8ff81-da59-4f3f-b306-\
d25ac6466a2d"],

Cell[CellGroupData[{

Cell[BoxData[{
 RowBox[{"Neq", " ", "=", " ", 
  RowBox[{
   RowBox[{"Simplify", "[", 
    RowBox[{"(", 
     RowBox[{
      RowBox[{"(", 
       RowBox[{"femaleNum", "+", "maleNum"}], ")"}], "/.", 
      RowBox[{"equilPopSizeSol", "[", 
       RowBox[{"[", "2", "]"}], "]"}]}], ")"}], "]"}], "//", 
   "Simplify"}]}], "\[IndentingNewLine]", 
 RowBox[{
  RowBox[{"Factor", "[", 
   RowBox[{
    RowBox[{"(", 
     FractionBox[
      RowBox[{"(", 
       RowBox[{
        RowBox[{"b", 
         RowBox[{"(", 
          RowBox[{"1", " ", "+", "  ", 
           RowBox[{"\[Phi]", " ", 
            RowBox[{"pfeq", "/", "2"}]}]}], ")"}], 
         RowBox[{"(", 
          RowBox[{
           RowBox[{"(", 
            RowBox[{"1", "-", "pmeq"}], ")"}], "/", 
           RowBox[{"(", 
            RowBox[{"1", "-", "pfeq"}], ")"}]}], ")"}]}], " ", "-", " ", 
        "2"}], ")"}], 
      RowBox[{" ", 
       RowBox[{"\[Alpha]", " ", "b", 
        RowBox[{"(", 
         RowBox[{"1", "+", 
          RowBox[{"\[Phi]", " ", 
           RowBox[{"pfeq", "/", "2"}]}]}], ")"}], 
        RowBox[{"(", 
         RowBox[{
          RowBox[{"(", 
           RowBox[{"1", "-", "pmeq"}], ")"}], "/", 
          RowBox[{"(", 
           RowBox[{"1", "-", "pfeq"}], ")"}]}], ")"}]}]}]], ")"}], " ", "-", 
    " ", "Neq"}], "]"}], "//", "Simplify"}], "\[IndentingNewLine]", 
 RowBox[{
  RowBox[{"bstar", "=", 
   RowBox[{"Factor", "[", 
    RowBox[{"(", 
     RowBox[{"b", 
      RowBox[{"(", 
       RowBox[{"1", "+", 
        FractionBox[
         RowBox[{"\[Phi]", " ", "pfeq"}], "2"]}], ")"}], 
      RowBox[{"(", 
       FractionBox[
        RowBox[{"1", "-", "pmeq"}], 
        RowBox[{"1", "-", "pfeq"}]], ")"}]}], ")"}], "]"}]}], 
  ";"}], "\[IndentingNewLine]", 
 RowBox[{"Factor", "[", 
  RowBox[{
   FractionBox[
    RowBox[{"bstar", "-", "2"}], 
    RowBox[{"bstar", " ", "\[Alpha]"}]], "-", "Neq"}], "]"}]}], "Input",
 CellChangeTimes->{
  3.8000085763622026`*^9, {3.814189487914071*^9, 3.814189593773863*^9}},
 CellLabel->
  "In[1299]:=",ExpressionUUID->"27c4ce68-c42f-4132-bce1-f9f2f51adf48"],

Cell[BoxData[
 RowBox[{
  RowBox[{"(", 
   RowBox[{
    RowBox[{"b", " ", 
     SuperscriptBox["h", "2"], " ", 
     SuperscriptBox["sf", "2"], " ", 
     SuperscriptBox[
      RowBox[{"(", 
       RowBox[{
        RowBox[{"-", "2"}], "+", "sm", "-", "\[Delta]", "+", 
        RowBox[{"sm", " ", "\[Delta]"}]}], ")"}], "2"]}], "+", 
    RowBox[{
     RowBox[{"(", 
      RowBox[{"sm", "-", "\[Delta]", "+", 
       RowBox[{"sm", " ", "\[Delta]"}]}], ")"}], " ", 
     RowBox[{"(", 
      RowBox[{
       RowBox[{
        RowBox[{"-", "b"}], " ", "\[Delta]"}], "+", 
       RowBox[{"sm", " ", 
        RowBox[{"(", 
         RowBox[{
          RowBox[{"-", "4"}], "+", "b", "+", 
          RowBox[{"b", " ", "\[Delta]"}]}], ")"}]}]}], ")"}]}], "-", 
    RowBox[{"2", " ", "sf", " ", 
     RowBox[{"(", 
      RowBox[{
       RowBox[{
        RowBox[{"-", "4"}], " ", 
        RowBox[{"(", 
         RowBox[{
          RowBox[{"-", "1"}], "+", "sm"}], ")"}], " ", 
        RowBox[{"(", 
         RowBox[{"1", "+", "\[Delta]"}], ")"}]}], "+", 
       RowBox[{"2", " ", "b", " ", 
        RowBox[{"(", 
         RowBox[{
          RowBox[{"-", "1"}], "+", "sm"}], ")"}], " ", 
        RowBox[{"(", 
         RowBox[{"1", "+", "\[Delta]"}], ")"}]}], "-", 
       RowBox[{"2", " ", "h", " ", 
        RowBox[{"(", 
         RowBox[{
          RowBox[{"-", "2"}], "+", "sm"}], ")"}], " ", 
        RowBox[{"(", 
         RowBox[{
          RowBox[{"-", "2"}], "+", "sm", "-", "\[Delta]", "+", 
          RowBox[{"sm", " ", "\[Delta]"}]}], ")"}]}], "+", 
       RowBox[{"b", " ", "h", " ", 
        SuperscriptBox[
         RowBox[{"(", 
          RowBox[{
           RowBox[{"-", "2"}], "+", "sm", "-", "\[Delta]", "+", 
           RowBox[{"sm", " ", "\[Delta]"}]}], ")"}], "2"]}]}], ")"}]}]}], 
   ")"}], "/", 
  RowBox[{"(", 
   RowBox[{"b", " ", "\[Alpha]", " ", 
    RowBox[{"(", 
     RowBox[{
      RowBox[{
       SuperscriptBox["h", "2"], " ", 
       SuperscriptBox["sf", "2"], " ", 
       SuperscriptBox[
        RowBox[{"(", 
         RowBox[{
          RowBox[{"-", "2"}], "+", "sm", "-", "\[Delta]", "+", 
          RowBox[{"sm", " ", "\[Delta]"}]}], ")"}], "2"]}], "+", 
      SuperscriptBox[
       RowBox[{"(", 
        RowBox[{"sm", "-", "\[Delta]", "+", 
         RowBox[{"sm", " ", "\[Delta]"}]}], ")"}], "2"], "-", 
      RowBox[{"2", " ", "sf", " ", 
       RowBox[{"(", 
        RowBox[{
         RowBox[{"2", " ", 
          RowBox[{"(", 
           RowBox[{
            RowBox[{"-", "1"}], "+", "sm"}], ")"}], " ", 
          RowBox[{"(", 
           RowBox[{"1", "+", "\[Delta]"}], ")"}]}], "+", 
         RowBox[{"h", " ", 
          SuperscriptBox[
           RowBox[{"(", 
            RowBox[{
             RowBox[{"-", "2"}], "+", "sm", "-", "\[Delta]", "+", 
             RowBox[{"sm", " ", "\[Delta]"}]}], ")"}], "2"]}]}], ")"}]}]}], 
     ")"}]}], ")"}]}]], "Output",
 CellChangeTimes->{
  3.8141895393379183`*^9, {3.814189590062188*^9, 3.814189594198102*^9}, 
   3.814195963603006*^9, 3.814196512845458*^9},
 CellLabel->
  "Out[1299]=",ExpressionUUID->"dac1b305-824a-4147-b6cf-b4bb658a7b9e"],

Cell[BoxData["0"], "Output",
 CellChangeTimes->{
  3.8141895393379183`*^9, {3.814189590062188*^9, 3.814189594198102*^9}, 
   3.814195963603006*^9, 3.8141965128755493`*^9},
 CellLabel->
  "Out[1300]=",ExpressionUUID->"797da42e-b36e-4881-9932-422c3de6e91d"],

Cell[BoxData["0"], "Output",
 CellChangeTimes->{
  3.8141895393379183`*^9, {3.814189590062188*^9, 3.814189594198102*^9}, 
   3.814195963603006*^9, 3.8141965128762197`*^9},
 CellLabel->
  "Out[1302]=",ExpressionUUID->"dcf1802f-61aa-4082-8057-dc2085b1b51e"]
}, Open  ]],

Cell[TextData[{
 "Now, we show that b* is increased by a factor that is the mean fitness of \
females at the polymorphic equilibrium (pm, pf) relative to the female \
population size without drive (",
 Cell[BoxData[
  FractionBox[
   RowBox[{"b", "-", "2"}], 
   RowBox[{"2", " ", "b", " ", "\[Alpha]"}]]],
  CellChangeTimes->{3.814190738615601*^9},ExpressionUUID->
  "3d6b35f9-3bea-48a7-b024-afff266f0a05"],
 ")."
}], "Text",
 CellChangeTimes->{{3.814190171721497*^9, 3.814190219954615*^9}, {
  3.814190278959606*^9, 3.814190336767393*^9}, {3.814190584791339*^9, 
  3.814190593645845*^9}, {3.8141907113643923`*^9, 3.814190788009817*^9}, {
  3.814190896975758*^9, 
  3.814190912769594*^9}},ExpressionUUID->"61bb9a29-74e9-418c-bb86-\
9b818c625f55"],

Cell[CellGroupData[{

Cell[BoxData[{
 RowBox[{
  RowBox[{"meanF", "=", 
   RowBox[{"Simplify", "[", 
    RowBox[{
     RowBox[{
      RowBox[{
       RowBox[{
        RowBox[{"(", 
         RowBox[{
          RowBox[{"DDfjuveniles", " ", "*", " ", "1"}], "+", 
          RowBox[{"Ddfjuveniles", "*", 
           RowBox[{"(", 
            RowBox[{"1", " ", "-", " ", 
             RowBox[{"h", "*", "sf"}]}], ")"}]}], "+", 
          RowBox[{"ddfjuveniles", "*", 
           RowBox[{"(", 
            RowBox[{"1", " ", "-", " ", "sf"}], ")"}]}]}], ")"}], "/", 
        "femaleNum"}], "/.", "subeq"}], "/.", "equilibrium"}], "/.", 
     "noDrivePopSizeSol"}], "]"}]}], ";"}], "\[IndentingNewLine]", 
 RowBox[{
  RowBox[{
   RowBox[{"b", " ", "meanF"}], "-", "bstar"}], "//", "Simplify"}]}], "Input",
 CellChangeTimes->{{3.8141896448704567`*^9, 3.814189820729373*^9}, {
  3.8141898511822987`*^9, 3.814189879192739*^9}, {3.814189918569771*^9, 
  3.814189942103026*^9}, {3.814189986426216*^9, 3.814190037308394*^9}, {
  3.814190146380966*^9, 3.81419014680739*^9}, {3.814190399968783*^9, 
  3.8141904516626167`*^9}, {3.814190798180624*^9, 3.814190885714787*^9}, {
  3.814190939704075*^9, 3.814190966004036*^9}},
 CellLabel->
  "In[1303]:=",ExpressionUUID->"b212b1df-35b8-4272-996f-97cfee5f7bd3"],

Cell[BoxData[
 RowBox[{"{", 
  RowBox[{"{", "0", "}"}], "}"}]], "Output",
 CellChangeTimes->{{3.814189854842723*^9, 3.814189879641571*^9}, {
   3.814189923555196*^9, 3.8141899423968687`*^9}, {3.8141899884691*^9, 
   3.814190037666664*^9}, 3.814190147700367*^9, {3.814190413440177*^9, 
   3.814190452004284*^9}, {3.814190855232069*^9, 3.8141908863993*^9}, {
   3.8141909400207157`*^9, 3.814190966335051*^9}, 3.814195963816349*^9, 
   3.814196513008129*^9},
 CellLabel->
  "Out[1304]=",ExpressionUUID->"5829f78e-11c2-4920-a939-57a2e567a899"]
}, Open  ]],

Cell["\<\
Finally, we also show that, when drive is fixed, the population size can be \
given by b~, as presented in the text. \
\>", "Text",
 CellChangeTimes->{{3.8141909935330677`*^9, 3.8141910162673264`*^9}, {
  3.81419108775981*^9, 
  3.8141910990239153`*^9}},ExpressionUUID->"819ae32d-3371-438e-83b2-\
ce5a625c27cb"],

Cell[CellGroupData[{

Cell[BoxData[{
 RowBox[{
  RowBox[{"bTilde", "=", 
   RowBox[{"b", 
    RowBox[{"(", 
     RowBox[{"1", "-", "sf"}], ")"}], " ", 
    RowBox[{"(", 
     RowBox[{"1", "+", "\[Delta]"}], ")"}]}]}], ";"}], "\[IndentingNewLine]", 
 RowBox[{"Factor", "[", 
  RowBox[{
   RowBox[{"(", 
    FractionBox[
     RowBox[{"bTilde", "-", "2"}], 
     RowBox[{"bTilde", " ", "\[Alpha]"}]], ")"}], "-", "Nfixed"}], 
  "]"}]}], "Input",
 CellChangeTimes->{{3.7855916876071978`*^9, 3.78559185789961*^9}, 
   3.788865029685322*^9, 3.795174722196233*^9, {3.814191034674275*^9, 
   3.8141910801439447`*^9}},
 CellLabel->
  "In[1305]:=",ExpressionUUID->"2ef5f460-c768-47ac-90cc-5ed41d836f6f"],

Cell[BoxData["0"], "Output",
 CellChangeTimes->{3.814191080432168*^9, 3.81419596382331*^9, 
  3.81419651301431*^9},
 CellLabel->
  "Out[1306]=",ExpressionUUID->"62c3fbaf-d7de-4955-bff1-28e455f6fe54"]
}, Open  ]]
}, Closed]],

Cell[CellGroupData[{

Cell["Stability", "Subsection",
 CellChangeTimes->{{3.7812702047944727`*^9, 
  3.781270206443592*^9}},ExpressionUUID->"c7603756-d652-48a5-8d25-\
7c22c8140a38"],

Cell["\<\
There are at least two immediately apparent equilibria: when drive is absent \
from the population and when drive is fixed in the population. 

We first find the Jacobian for the system:\
\>", "Text",
 CellChangeTimes->{{3.781272256024481*^9, 3.781272395504654*^9}, {
  3.781273708553084*^9, 3.7812737092006273`*^9}, {3.781273897437092*^9, 
  3.781273908742497*^9}, {3.800007455744604*^9, 3.800007456204402*^9}, {
  3.814092010155294*^9, 
  3.814092011347086*^9}},ExpressionUUID->"069fa669-e6f5-414b-8bb6-\
0768dbae7863"],

Cell[BoxData[{
 RowBox[{
  RowBox[{"recursions", "=", 
   RowBox[{"{", 
    RowBox[{
    "Dmprime", ",", " ", "dmprime", ",", " ", "DDfprime", ",", " ", 
     "Ddfprime", ",", " ", "ddfprime"}], "}"}]}], 
  ";"}], "\[IndentingNewLine]", 
 RowBox[{
  RowBox[{"jacob", "=", 
   RowBox[{"Simplify", "[", 
    RowBox[{"Transpose", "[", 
     RowBox[{"{", 
      RowBox[{
       RowBox[{"D", "[", 
        RowBox[{"recursions", ",", "Dm"}], "]"}], ",", " ", 
       RowBox[{"D", "[", 
        RowBox[{"recursions", ",", "dm"}], "]"}], ",", " ", 
       RowBox[{"D", "[", 
        RowBox[{"recursions", ",", " ", "DDf"}], "]"}], ",", " ", 
       RowBox[{"D", "[", 
        RowBox[{"recursions", ",", " ", "Ddf"}], "]"}], ",", " ", 
       RowBox[{"D", "[", 
        RowBox[{"recursions", ",", " ", "ddf"}], "]"}]}], "}"}], "]"}], 
    "]"}]}], ";"}]}], "Input",
 CellChangeTimes->{{3.7812738954655247`*^9, 3.781273895657804*^9}},
 CellLabel->
  "In[1307]:=",ExpressionUUID->"0311a470-b3b5-4128-9aee-344bd1c02d18"],

Cell["\<\
And then we assess the conditions under which the drive-absent and \
drive-fixed equilibria are stable/instable. \
\>", "Text",
 CellChangeTimes->{
  3.8140187547665863`*^9},ExpressionUUID->"849fe3c7-f7a5-4759-a193-\
b19237b95a9c"],

Cell[CellGroupData[{

Cell["drive absent (pm=pf=0) equilibrium", "Subsubsection",
 CellChangeTimes->{{3.781273712670188*^9, 3.7812737328189983`*^9}, 
   3.814018761654243*^9},ExpressionUUID->"79b91767-43e9-45e7-8f2d-\
900ada457e51"],

Cell["\<\
To evaluate the stability of the equilibrium where drive is absent we \
consider the eigenvalues of the Jacobian at this point (substituting pf->0, \
pm->0, and the equilibrium population size without drive). The equilibrium is \
unstable whenever it has an eigenvalue with absolute value > 1.\
\>", "Text",
 CellChangeTimes->{{3.781273918397637*^9, 3.781273991244995*^9}, {
  3.781275775595746*^9, 3.781275807826384*^9}, {3.814003868404478*^9, 
  3.814003889059251*^9}},ExpressionUUID->"bb9328f9-fea8-4cba-ada5-\
68fd7946ad96"],

Cell[CellGroupData[{

Cell[BoxData[{
 RowBox[{
  RowBox[{
   RowBox[{"Simplify", "[", 
    RowBox[{
     RowBox[{
      RowBox[{
       RowBox[{"jacob", "/.", "subeq"}], "/.", 
       RowBox[{"pf", "\[Rule]", "0"}]}], "/.", 
      RowBox[{"pm", "\[Rule]", "0"}]}], "/.", "noDrivePopSizeSol"}], "]"}], 
   "[", 
   RowBox[{"[", "1", "]"}], "]"}], ";"}], "\[IndentingNewLine]", 
 RowBox[{"charpolynomial0", "=", 
  RowBox[{"Factor", "[", 
   RowBox[{"Det", "[", 
    RowBox[{"%", "-", 
     RowBox[{"\[Lambda]", "*", 
      RowBox[{"IdentityMatrix", "[", "5", "]"}]}]}], "]"}], "]"}]}]}], "Input",
 CellChangeTimes->{
  3.781274002989327*^9, {3.813930917830587*^9, 3.813931059531245*^9}, {
   3.814003803361516*^9, 3.814003804854281*^9}, {3.8140085533840113`*^9, 
   3.814008553741354*^9}},
 CellLabel->
  "In[1309]:=",ExpressionUUID->"7f47ae55-1386-43af-8177-ce1b2387ab73"],

Cell[BoxData[
 RowBox[{
  RowBox[{"-", 
   FractionBox["1", "4"]}], " ", 
  SuperscriptBox["\[Lambda]", "2"], " ", 
  RowBox[{"(", 
   RowBox[{
    RowBox[{"-", "4"}], "+", "b", "+", 
    RowBox[{"2", " ", "\[Lambda]"}]}], ")"}], " ", 
  RowBox[{"(", 
   RowBox[{
    RowBox[{"-", "1"}], "+", 
    RowBox[{"h", " ", "sf"}], "+", "sm", "-", 
    RowBox[{"h", " ", "sf", " ", "sm"}], "-", "\[Delta]", "+", 
    RowBox[{"h", " ", "sf", " ", "\[Delta]"}], "+", 
    RowBox[{"sm", " ", "\[Delta]"}], "-", 
    RowBox[{"h", " ", "sf", " ", "sm", " ", "\[Delta]"}], "-", "\[Lambda]", 
    "+", 
    RowBox[{"h", " ", "sf", " ", "\[Lambda]"}], "+", 
    RowBox[{"2", " ", 
     SuperscriptBox["\[Lambda]", "2"]}]}], ")"}]}]], "Output",
 CellChangeTimes->{
  3.7812740033403482`*^9, 3.7814573218417273`*^9, 3.7818848609506083`*^9, 
   3.781952821819541*^9, 3.78195289751086*^9, {3.7819538958219957`*^9, 
   3.781953915448163*^9}, 3.7821306544665728`*^9, 3.782747157952016*^9, 
   3.7830788934301643`*^9, 3.785162683931014*^9, 3.785164973948011*^9, 
   3.789732648885466*^9, 3.795174657109274*^9, 3.799320669052409*^9, 
   3.799340142599029*^9, 3.799756425128694*^9, 3.800007655671279*^9, 
   3.8002681846876783`*^9, 3.813930901799224*^9, {3.8139309343336077`*^9, 
   3.8139310598724957`*^9}, 3.8140037587039423`*^9, 3.814003805149556*^9, 
   3.814081435651245*^9, 3.814189400309782*^9, 3.814195964352592*^9, 
   3.814196513430963*^9},
 CellLabel->
  "Out[1310]=",ExpressionUUID->"1548c55b-67b8-42b5-b844-9a327c3f3c8c"]
}, Open  ]],

Cell[TextData[{
 "The factor ",
 Cell[BoxData[
  RowBox[{
   RowBox[{"-", 
    FractionBox["1", "4"]}], " ", 
   SuperscriptBox["\[Lambda]", "2"], " ", 
   RowBox[{"(", 
    RowBox[{
     RowBox[{"-", "4"}], "+", "b", "+", 
     RowBox[{"2", " ", "\[Lambda]"}]}], ")"}]}]],
  CellChangeTimes->{
   3.7812740033403482`*^9, 3.7814573218417273`*^9, 3.7818848609506083`*^9, 
    3.781952821819541*^9, 3.78195289751086*^9, {3.7819538958219957`*^9, 
    3.781953915448163*^9}, 3.7821306544665728`*^9, 3.782747157952016*^9, 
    3.7830788934301643`*^9, 3.785162683931014*^9, 3.785164973948011*^9, 
    3.789732648885466*^9, 3.795174657109274*^9, 3.799320669052409*^9, 
    3.799340142599029*^9, 3.799756425128694*^9, 3.800007655671279*^9, 
    3.8002681846876783`*^9, 3.813930901799224*^9, {3.8139309343336077`*^9, 
    3.8139310598724957`*^9}, 3.8140037587039423`*^9, 3.814003805149556*^9},
  ExpressionUUID->"31396fd5-6a6b-4b6e-8acd-d50aeddeb136"],
 " gives two 0 eigenvalues and and eigenvalue of ",
 Cell[BoxData[
  FractionBox[
   RowBox[{"4", "-", "b"}], "2"]],
  CellChangeTimes->{3.814003986297804*^9},ExpressionUUID->
  "4f8b776c-b2ae-4d20-8b16-36637e57572d"],
 ", corresponding to the change in overall population size. Thus, we can see \
that the demographic dynamics are orthogonal to the evolutionary dynamics. \n\
\nHere,  we divide through by this factor to simplify the equation and \
evaluate the evolutionary dynamics. We then write the characteristic \
polynomial in the form\n",
 Cell[BoxData[
  RowBox[{
   SuperscriptBox["a\[Lambda]", "2"], "+", "b\[Lambda]", "+", "c"}]],
  CellChangeTimes->{
   3.7812740321927767`*^9, 3.781274073867482*^9, 3.7814285214279013`*^9, 
    3.781457321850852*^9, 3.78188486097983*^9, 3.781952821828101*^9, 
    3.7819528975187283`*^9, {3.781953895876915*^9, 3.7819539154956713`*^9}, 
    3.7821306544745502`*^9, 3.782747157961987*^9, 3.7830788934387407`*^9, 
    3.785162684364028*^9, 3.785164974025779*^9, 3.789732649060177*^9, 
    3.79517465722147*^9, 3.799320669259963*^9, 3.799340142805358*^9, 
    3.799756425312364*^9, 3.800007655751677*^9, 3.80026818474804*^9, 
    3.8139309588718987`*^9, 3.813931061212427*^9, 3.814004119936627*^9},
  ExpressionUUID->"812065ab-c20f-4f95-b025-9f98980df433"],
 " \nfinding the a, b, and c coefficients. "
}], "Text",
 CellChangeTimes->{{3.8140039558834753`*^9, 3.8140040343097057`*^9}, {
  3.8140040782287703`*^9, 3.814004106285927*^9}, {3.814004210438005*^9, 
  3.814004280376747*^9}},ExpressionUUID->"c6b7c149-1d9d-486e-8d79-\
e0b69ffdc7ac"],

Cell[CellGroupData[{

Cell[BoxData[{
 RowBox[{
  RowBox[{"Collect", "[", 
   RowBox[{
    RowBox[{"charpolynomial0", "/", 
     RowBox[{"(", 
      RowBox[{
       RowBox[{"-", 
        FractionBox["1", "4"]}], " ", 
       SuperscriptBox["\[Lambda]", "2"], " ", 
       RowBox[{"(", 
        RowBox[{
         RowBox[{"-", "4"}], "+", "b", "+", 
         RowBox[{"2", " ", "\[Lambda]"}]}], ")"}]}], ")"}]}], ",", 
    "\[Lambda]", ",", "Simplify"}], "]"}], ";"}], "\[IndentingNewLine]", 
 RowBox[{
  RowBox[{"Coefficient", "[", 
   RowBox[{"%", ",", 
    RowBox[{"\[Lambda]", "^", "2"}]}], "]"}], ";"}], "\[IndentingNewLine]", 
 RowBox[{
  RowBox[{"charpolySimpleACoef0", "=", 
   RowBox[{"%%", "/", "%"}]}], " ", 
  RowBox[{"(*", 
   RowBox[{
   "simplify", " ", "so", " ", "that", " ", "the", " ", "a", " ", 
    "coefficient", " ", "is", " ", "1"}], "*)"}]}], "\[IndentingNewLine]", 
 RowBox[{
  RowBox[{"acoef0", "=", 
   RowBox[{"Coefficient", "[", 
    RowBox[{"charpolySimpleACoef0", ",", 
     RowBox[{"\[Lambda]", "^", "2"}]}], "]"}]}], ";"}], "\[IndentingNewLine]", 
 RowBox[{
  RowBox[{"bcoef0", "=", 
   RowBox[{"Coefficient", "[", 
    RowBox[{"charpolySimpleACoef0", ",", "\[Lambda]"}], "]"}]}], 
  ";"}], "\[IndentingNewLine]", 
 RowBox[{
  RowBox[{"ccoef0", "=", 
   RowBox[{
    RowBox[{"charpolySimpleACoef0", " ", "-", " ", 
     RowBox[{"bcoef0", "*", "\[Lambda]"}], "-", 
     RowBox[{"acoef0", "*", 
      RowBox[{"\[Lambda]", "^", "2"}]}]}], " ", "//", " ", "Simplify"}]}], 
  ";"}]}], "Input",
 CellChangeTimes->{{3.7812740703585978`*^9, 3.7812740731784153`*^9}, {
   3.781428516566597*^9, 3.781428520312667*^9}, {3.814003772626279*^9, 
   3.814003773254695*^9}, 3.814004117518854*^9, {3.814004306769434*^9, 
   3.814004317051339*^9}},
 CellLabel->
  "In[1311]:=",ExpressionUUID->"a062c177-c1b5-4aea-89af-f51112f7b4e6"],

Cell[BoxData[
 RowBox[{
  FractionBox["1", "2"], " ", 
  RowBox[{"(", 
   RowBox[{
    RowBox[{
     RowBox[{"-", 
      RowBox[{"(", 
       RowBox[{
        RowBox[{"-", "1"}], "+", 
        RowBox[{"h", " ", "sf"}]}], ")"}]}], " ", 
     RowBox[{"(", 
      RowBox[{
       RowBox[{"-", "1"}], "+", "sm"}], ")"}], " ", 
     RowBox[{"(", 
      RowBox[{"1", "+", "\[Delta]"}], ")"}]}], "+", 
    RowBox[{
     RowBox[{"(", 
      RowBox[{
       RowBox[{"-", "1"}], "+", 
       RowBox[{"h", " ", "sf"}]}], ")"}], " ", "\[Lambda]"}], "+", 
    RowBox[{"2", " ", 
     SuperscriptBox["\[Lambda]", "2"]}]}], ")"}]}]], "Output",
 CellChangeTimes->{
  3.7812740321927767`*^9, 3.781274073867482*^9, 3.7814285214279013`*^9, 
   3.781457321850852*^9, 3.78188486097983*^9, 3.781952821828101*^9, 
   3.7819528975187283`*^9, {3.781953895876915*^9, 3.7819539154956713`*^9}, 
   3.7821306544745502`*^9, 3.782747157961987*^9, 3.7830788934387407`*^9, 
   3.785162684364028*^9, 3.785164974025779*^9, 3.789732649060177*^9, 
   3.79517465722147*^9, 3.799320669259963*^9, 3.799340142805358*^9, 
   3.799756425312364*^9, 3.800007655751677*^9, 3.80026818474804*^9, 
   3.8139309588718987`*^9, 3.813931061212427*^9, 3.814004119936627*^9, 
   3.814081440403721*^9, 3.814189400323086*^9, 3.814195964362994*^9, 
   3.81419651349531*^9},
 CellLabel->
  "Out[1313]=",ExpressionUUID->"ddb9ee80-7091-4e9d-a6fd-550af5402d74"]
}, Open  ]],

Cell["\<\
The discriminant must be non-negative for the eigenvalues to be real\
\>", "Text",
 CellChangeTimes->{{3.781428281916833*^9, 
  3.781428302970292*^9}},ExpressionUUID->"3f84c290-93d1-4128-a2e3-\
b4d2a996e2ab"],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{
  RowBox[{
   RowBox[{"bcoef0", "^", "2"}], "-", 
   RowBox[{"4", "acoef0", " ", "ccoef0"}]}], " ", "//", " ", 
  "FullSimplify"}]], "Input",
 CellChangeTimes->{{3.781428313265222*^9, 3.781428316168023*^9}, {
  3.781428426467711*^9, 3.7814285260999413`*^9}},
 CellLabel->
  "In[1317]:=",ExpressionUUID->"bc064fba-63b8-47ae-aa2c-a70d69dcf620"],

Cell[BoxData[
 RowBox[{
  FractionBox["1", "4"], " ", 
  RowBox[{"(", 
   RowBox[{
    RowBox[{"-", "1"}], "+", 
    RowBox[{"h", " ", "sf"}]}], ")"}], " ", 
  RowBox[{"(", 
   RowBox[{
    RowBox[{"-", "1"}], "+", 
    RowBox[{"h", " ", "sf"}], "+", 
    RowBox[{"8", " ", 
     RowBox[{"(", 
      RowBox[{
       RowBox[{"-", "1"}], "+", "sm"}], ")"}], " ", 
     RowBox[{"(", 
      RowBox[{"1", "+", "\[Delta]"}], ")"}]}]}], ")"}]}]], "Output",
 CellChangeTimes->{
  3.781428316942605*^9, {3.781428429373097*^9, 3.781428527133175*^9}, 
   3.7814573219234743`*^9, 3.781884861028933*^9, 3.7819528218734293`*^9, 
   3.781952897577923*^9, {3.781953895926058*^9, 3.781953915547658*^9}, 
   3.782130654518096*^9, 3.7827471580481977`*^9, 3.7830788935329447`*^9, 
   3.785162684709345*^9, 3.785164974150434*^9, 3.789732649230698*^9, 
   3.795174657383437*^9, 3.799320669605383*^9, 3.7993401431414948`*^9, 
   3.799756425497038*^9, 3.80000765591433*^9, 3.80026818485813*^9, 
   3.8139310639623957`*^9, 3.8140041234837303`*^9, 3.814081441766411*^9, 
   3.81418940036903*^9, 3.814195964403165*^9, 3.814196513541647*^9},
 CellLabel->
  "Out[1317]=",ExpressionUUID->"d26856b4-bb3f-4050-a6da-2ac9f7950954"]
}, Open  ]],

Cell[TextData[{
 "We can see that this is always true as the two terms in the product are \
always \[LessEqual] 0.\n\nThe conditions for the above characteristic \
polynomial ",
 Cell[BoxData[
  FormBox["p", TraditionalForm]],ExpressionUUID->
  "34e078e3-2304-422d-b56b-a4b98b1d698e"],
 "  to have roots between -1 and 1 are:\n1. ",
 Cell[BoxData[
  FormBox[
   RowBox[{
    RowBox[{"p", "(", "1", ")"}], " ", ">", " ", "0"}], TraditionalForm]],
  ExpressionUUID->"00646320-05e2-4013-bfbd-752e22d32adb"],
 "\n2. ",
 Cell[BoxData[
  FormBox[
   RowBox[{
    RowBox[{"p", "(", 
     RowBox[{"-", "1"}], ")"}], " ", ">", " ", "0"}], TraditionalForm]],
  ExpressionUUID->"9fdc6528-2117-44b5-8237-d38de3dcadc9"],
 "\n3.  ",
 Cell[BoxData[
  FormBox[
   RowBox[{
    RowBox[{"-", "2"}], " ", "<", " ", "b", " ", "<", " ", "2"}], 
   TraditionalForm]],ExpressionUUID->"f7a04489-da9f-4f9c-896d-49db1e0b2a38"],
 "\nWhen any of these are false, the largest eigenvalue is ",
 Cell[BoxData[
  FormBox[
   RowBox[{">", 
    RowBox[{"1", "\:201a", " "}]}], TraditionalForm]],ExpressionUUID->
  "aec931b3-8545-4140-a197-83a29ea86e37"],
 "and the equilibrium is unstable. Conditions 2 and 3 are always true:"
}], "Text",
 CellChangeTimes->{{3.781428661248453*^9, 3.781428712666779*^9}, {
  3.781428780656464*^9, 3.781428782114583*^9}, {3.781429946730907*^9, 
  3.7814299482572327`*^9}, {3.7814360173472347`*^9, 3.781436101352002*^9}, {
  3.781436149893372*^9, 3.7814362551794662`*^9}, {3.781436330451419*^9, 
  3.781436341934223*^9}, {3.781436393282946*^9, 3.781436444944702*^9}, {
  3.781436680869691*^9, 3.78143669857165*^9}, {3.814004641338129*^9, 
  3.814004642068164*^9}},ExpressionUUID->"1cf4ddfc-43a4-471a-9b4e-\
dfe5df17d9c2"],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"FullSimplify", "[", 
  RowBox[{
   RowBox[{
    RowBox[{"(", 
     RowBox[{"charpolySimpleACoef0", "/.", 
      RowBox[{"\[Lambda]", "\[Rule]", 
       RowBox[{"-", "1"}]}]}], " ", ")"}], ">", " ", "0"}], ",", "  ", 
   RowBox[{"{", 
    RowBox[{
     RowBox[{"1", ">", " ", "\[Delta]", ">", "0"}], ",", 
     RowBox[{"1", ">", "h", ">", "0"}], ",", 
     RowBox[{"1", ">", "sm", ">", "0"}], ",", 
     RowBox[{"0", "<", "sf", "<", "1"}]}], "}"}]}], "]"}]], "Input",
 CellChangeTimes->{{3.781436736869988*^9, 3.781436739183576*^9}},
 CellLabel->
  "In[1318]:=",ExpressionUUID->"ba5dece9-5a47-4e41-9ea8-ddec4b0874db"],

Cell[BoxData["True"], "Output",
 CellChangeTimes->{
  3.781436739697763*^9, 3.78145732216691*^9, 3.781884861062943*^9, 
   3.7819528219341993`*^9, 3.781952897819153*^9, {3.781953895978655*^9, 
   3.781953915596232*^9}, 3.782130654549685*^9, 3.782747158425377*^9, 
   3.7830788938290977`*^9, 3.785162685017118*^9, 3.785164974273959*^9, 
   3.789732649397847*^9, 3.795174657691847*^9, 3.799320669976574*^9, 
   3.799340143764106*^9, 3.799756425685603*^9, 3.8000076560762587`*^9, 
   3.800268185146523*^9, 3.814004596139834*^9, 3.814081444299515*^9, 
   3.814189400417293*^9, 3.814195964452488*^9, 3.814196513599567*^9},
 CellLabel->
  "Out[1318]=",ExpressionUUID->"cb214e2d-51de-436c-bc6b-e11e1e1135d9"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"FullSimplify", "[", 
  RowBox[{
   RowBox[{
    RowBox[{"-", "2"}], "<", "bcoef0", "<", "2"}], ",", 
   RowBox[{"{", 
    RowBox[{
     RowBox[{"1", ">", " ", "\[Delta]", ">", "0"}], ",", 
     RowBox[{"1", ">", "h", ">", "0"}], ",", 
     RowBox[{"1", ">", "sm", ">", "0"}], ",", 
     RowBox[{"0", "<", "sf", "<", "1"}]}], "}"}]}], "]"}]], "Input",
 CellChangeTimes->{{3.781436745619834*^9, 3.781436768260647*^9}},
 CellLabel->
  "In[1319]:=",ExpressionUUID->"7fde27ef-5868-4523-8056-b480154601d9"],

Cell[BoxData["True"], "Output",
 CellChangeTimes->{
  3.7814367687176857`*^9, 3.781457322173564*^9, 3.781884861071455*^9, 
   3.7819528219793453`*^9, 3.7819528978269367`*^9, {3.7819538959919233`*^9, 
   3.781953915604405*^9}, 3.782130654581544*^9, 3.782747158434576*^9, 
   3.78307889402479*^9, 3.785162685768771*^9, 3.785164974470911*^9, 
   3.789732649484159*^9, 3.795174657881214*^9, 3.7993206700828447`*^9, 
   3.799340144112084*^9, 3.799756425778799*^9, 3.8000076561625147`*^9, 
   3.800268185279215*^9, 3.814004597392589*^9, 3.8140814447417097`*^9, 
   3.8141894004264803`*^9, 3.814195964465844*^9, 3.8141965136058702`*^9},
 CellLabel->
  "Out[1319]=",ExpressionUUID->"81f42865-d0a0-433e-9086-66a6cba4e3da"]
}, Open  ]],

Cell["\<\
But condition 1 fails when the following inequality is true, giving us the \
instability condition.\
\>", "Text",
 CellChangeTimes->{{3.7814367744575577`*^9, 
  3.7814368080369053`*^9}},ExpressionUUID->"ce2389da-627b-495a-8e96-\
9e6424cc0855"],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"unstabConditions0", " ", "=", " ", 
  RowBox[{"FullSimplify", "[", 
   RowBox[{
    RowBox[{
     RowBox[{"(", 
      RowBox[{"charpolySimpleACoef0", "/.", 
       RowBox[{"\[Lambda]", "\[Rule]", "1"}]}], " ", ")"}], "<", " ", "0"}], 
    ",", "  ", 
    RowBox[{"{", 
     RowBox[{
      RowBox[{"1", ">", " ", "\[Delta]", ">", "0"}], ",", 
      RowBox[{"1", ">", "h", ">", "0"}], ",", 
      RowBox[{"1", ">", "sm", ">", "0"}], ",", 
      RowBox[{"0", "<", "sf", "<", "1"}]}], "}"}]}], "]"}]}]], "Input",
 CellChangeTimes->{{3.781429951183529*^9, 3.7814301179474297`*^9}, {
   3.7814302541642017`*^9, 3.781430298190481*^9}, {3.7814303383056297`*^9, 
   3.781430361497563*^9}, {3.7814312736199713`*^9, 3.78143127584548*^9}, {
   3.781436466857931*^9, 3.781436560917033*^9}, {3.781436631602688*^9, 
   3.781436644697918*^9}, 3.8140046082541723`*^9, {3.814004682873187*^9, 
   3.814004713133438*^9}},
 CellLabel->
  "In[1320]:=",ExpressionUUID->"5530f374-81a1-4254-82c7-fd3e53be49c9"],

Cell[BoxData[
 RowBox[{
  RowBox[{"sm", "+", 
   RowBox[{"sm", " ", "\[Delta]"}], "+", 
   RowBox[{"h", " ", "sf", " ", 
    RowBox[{"(", 
     RowBox[{"2", "+", "\[Delta]", "-", 
      RowBox[{"sm", " ", 
       RowBox[{"(", 
        RowBox[{"1", "+", "\[Delta]"}], ")"}]}]}], ")"}]}]}], "<", 
  "\[Delta]"}]], "Output",
 CellChangeTimes->{
  3.7814300190036163`*^9, {3.781430050891747*^9, 3.781430119181625*^9}, {
   3.781430258842821*^9, 3.7814302987538567`*^9}, {3.781430357589197*^9, 
   3.7814303619226294`*^9}, 3.781431276347818*^9, {3.781436471134864*^9, 
   3.7814365067294188`*^9}, 3.781436668610888*^9, 3.7814573222715*^9, 
   3.781884861204401*^9, 3.781952822089045*^9, 3.781952897921483*^9, {
   3.781953896105276*^9, 3.781953915747797*^9}, 3.782130654684663*^9, 
   3.782747158540249*^9, 3.783078894199912*^9, 3.7851626859809427`*^9, 
   3.785164974645007*^9, 3.789732649767761*^9, 3.795174658046834*^9, 
   3.799320670294146*^9, 3.7993401443157*^9, 3.79975642605756*^9, 
   3.800007656424817*^9, 3.800268185436696*^9, 3.8139311136920567`*^9, {
   3.814004601006237*^9, 3.8140046096271353`*^9}, {3.814004683951476*^9, 
   3.814004713521501*^9}, 3.814081445877571*^9, 3.814189400591201*^9, 
   3.814195964565126*^9, 3.81419651370522*^9},
 CellLabel->
  "Out[1320]=",ExpressionUUID->"669ee6e1-5491-45d7-bae6-5675aaeb1f83"]
}, Open  ]],

Cell["which can be re-written using the w fitness notation", "Text",
 CellChangeTimes->{{3.814004650674572*^9, 
  3.814004710715838*^9}},ExpressionUUID->"a7fe70a1-d54b-44b0-8921-\
4c0b2d57c4b4"],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"Simplify", "[", 
  RowBox[{"unstabConditions0", ",", " ", 
   RowBox[{"{", 
    RowBox[{
     RowBox[{"wXdY", " ", "\[Equal]", " ", 
      RowBox[{"1", " ", "-", " ", "sm"}]}], ",", 
     RowBox[{"wXdX", " ", "\[Equal]", " ", 
      RowBox[{"1", " ", "-", " ", 
       RowBox[{"h", " ", "sf"}]}]}], ",", 
     RowBox[{"wXdXd", " ", "\[Equal]", " ", 
      RowBox[{"1", " ", "-", " ", "sf"}]}]}], "}"}]}], "]"}]], "Input",
 CellChangeTimes->{{3.782747173005354*^9, 3.782747212585889*^9}, {
  3.782748801098444*^9, 3.7827488122111607`*^9}, {3.783078875883896*^9, 
  3.783078876006201*^9}},
 CellLabel->
  "In[1321]:=",ExpressionUUID->"b37fc450-b899-412b-9f90-3bbcfa772767"],

Cell[BoxData[
 RowBox[{
  RowBox[{"wXdX", "+", 
   RowBox[{"wXdX", " ", "wXdY", " ", 
    RowBox[{"(", 
     RowBox[{"1", "+", "\[Delta]"}], ")"}]}]}], ">", "2"}]], "Output",
 CellChangeTimes->{{3.7827471930133467`*^9, 3.782747213122994*^9}, {
   3.782748802357192*^9, 3.782748813081629*^9}, {3.7830788773278503`*^9, 
   3.783078894293872*^9}, 3.78516268619658*^9, 3.7851649747693663`*^9, 
   3.789732650076954*^9, 3.795174658156262*^9, 3.799320670556473*^9, 
   3.799340144519253*^9, 3.799756426506791*^9, 3.8000076569522038`*^9, 
   3.800268185528966*^9, 3.814004612665675*^9, 3.8140047147797937`*^9, 
   3.81408144704889*^9, 3.8141894006410093`*^9, 3.814195964607699*^9, 
   3.814196513752049*^9},
 CellLabel->
  "Out[1321]=",ExpressionUUID->"edce6e00-2e91-4b7c-906f-fb16ad73cca3"]
}, Open  ]]
}, Closed]],

Cell[CellGroupData[{

Cell["drive fixed (pm=pf=1) equilibrium", "Subsubsection",
 CellChangeTimes->{{3.781273735316922*^9, 3.781273740747805*^9}, 
   3.814018765219983*^9},ExpressionUUID->"532130bb-df22-432d-be64-\
3fff631f92c5"],

Cell["\<\
We proceed similarly to before, but now considering the other equilibrium:\
\>", "Text",
 CellChangeTimes->{{3.7814369207715807`*^9, 
  3.781436938486001*^9}},ExpressionUUID->"27fb58c4-df16-4b7d-8d6e-\
ff8caf4127a2"],

Cell[CellGroupData[{

Cell[BoxData[{
 RowBox[{
  RowBox[{
   RowBox[{"Simplify", "[", 
    RowBox[{
     RowBox[{
      RowBox[{
       RowBox[{"jacob", "/.", "subeq"}], "/.", 
       RowBox[{"pf", "\[Rule]", "1"}]}], "/.", 
      RowBox[{"pm", "\[Rule]", "1"}]}], "/.", "fixedDrivePopSizeSol"}], "]"}],
    "[", 
   RowBox[{"[", "1", "]"}], "]"}], ";"}], "\[IndentingNewLine]", 
 RowBox[{"charpolynomial1", "=", 
  RowBox[{"Factor", "[", 
   RowBox[{"Det", "[", 
    RowBox[{"%", "-", 
     RowBox[{"\[Lambda]", "*", 
      RowBox[{"IdentityMatrix", "[", "5", "]"}]}]}], "]"}], "]"}]}]}], "Input",
 CellChangeTimes->{{3.814004729064748*^9, 3.814004743901525*^9}, {
  3.8140047795551653`*^9, 3.814004791431684*^9}, {3.814005039784062*^9, 
  3.814005043586812*^9}, {3.814008492634818*^9, 3.814008493127555*^9}},
 CellLabel->
  "In[1322]:=",ExpressionUUID->"e6b57997-0174-4c54-ae73-1a8b8e77e142"],

Cell[BoxData[
 RowBox[{
  FractionBox["1", 
   RowBox[{"4", " ", 
    RowBox[{"(", 
     RowBox[{
      RowBox[{"-", "1"}], "+", "sf"}], ")"}], " ", 
    RowBox[{"(", 
     RowBox[{
      RowBox[{"-", "1"}], "+", "sm"}], ")"}], " ", 
    RowBox[{"(", 
     RowBox[{"1", "+", "\[Delta]"}], ")"}]}]], 
  RowBox[{
   RowBox[{"(", 
    RowBox[{"4", "-", "b", "+", 
     RowBox[{"b", " ", "sf"}], "-", 
     RowBox[{"b", " ", "\[Delta]"}], "+", 
     RowBox[{"b", " ", "sf", " ", "\[Delta]"}], "-", 
     RowBox[{"2", " ", "\[Lambda]"}]}], ")"}], " ", 
   SuperscriptBox["\[Lambda]", "2"], " ", 
   RowBox[{"(", 
    RowBox[{
     RowBox[{"-", "1"}], "+", 
     RowBox[{"h", " ", "sf"}], "-", "\[Lambda]", "+", 
     RowBox[{"h", " ", "sf", " ", "\[Lambda]"}], "+", 
     RowBox[{"sm", " ", "\[Lambda]"}], "-", 
     RowBox[{"h", " ", "sf", " ", "sm", " ", "\[Lambda]"}], "-", 
     RowBox[{"\[Delta]", " ", "\[Lambda]"}], "+", 
     RowBox[{"h", " ", "sf", " ", "\[Delta]", " ", "\[Lambda]"}], "+", 
     RowBox[{"sm", " ", "\[Delta]", " ", "\[Lambda]"}], "-", 
     RowBox[{"h", " ", "sf", " ", "sm", " ", "\[Delta]", " ", "\[Lambda]"}], 
     "+", 
     RowBox[{"2", " ", 
      SuperscriptBox["\[Lambda]", "2"]}], "-", 
     RowBox[{"2", " ", "sf", " ", 
      SuperscriptBox["\[Lambda]", "2"]}], "-", 
     RowBox[{"2", " ", "sm", " ", 
      SuperscriptBox["\[Lambda]", "2"]}], "+", 
     RowBox[{"2", " ", "sf", " ", "sm", " ", 
      SuperscriptBox["\[Lambda]", "2"]}], "+", 
     RowBox[{"2", " ", "\[Delta]", " ", 
      SuperscriptBox["\[Lambda]", "2"]}], "-", 
     RowBox[{"2", " ", "sf", " ", "\[Delta]", " ", 
      SuperscriptBox["\[Lambda]", "2"]}], "-", 
     RowBox[{"2", " ", "sm", " ", "\[Delta]", " ", 
      SuperscriptBox["\[Lambda]", "2"]}], "+", 
     RowBox[{"2", " ", "sf", " ", "sm", " ", "\[Delta]", " ", 
      SuperscriptBox["\[Lambda]", "2"]}]}], ")"}]}]}]], "Output",
 CellChangeTimes->{{3.814004786393588*^9, 3.814004791710847*^9}, 
   3.8140050447489643`*^9, 3.814005210911131*^9, 3.8140814605227623`*^9, 
   3.8141894010039062`*^9, 3.8141959649253273`*^9, 3.814196514063834*^9},
 CellLabel->
  "Out[1323]=",ExpressionUUID->"14349bd1-4d87-4bac-9dec-516f8ee4fa33"]
}, Open  ]],

Cell[TextData[{
 "The factor ",
 Cell[BoxData[
  FractionBox[
   RowBox[{
    RowBox[{"(", 
     RowBox[{"4", "-", "b", "+", 
      RowBox[{"b", " ", "sf"}], "-", 
      RowBox[{"b", " ", "\[Delta]"}], "+", 
      RowBox[{"b", " ", "sf", " ", "\[Delta]"}], "-", 
      RowBox[{"2", " ", "\[Lambda]"}]}], ")"}], " ", 
    SuperscriptBox["\[Lambda]", "2"], " "}], 
   RowBox[{"4", " ", 
    RowBox[{"(", 
     RowBox[{
      RowBox[{"-", "1"}], "+", "sf"}], ")"}], " ", 
    RowBox[{"(", 
     RowBox[{
      RowBox[{"-", "1"}], "+", "sm"}], ")"}], " ", 
    RowBox[{"(", 
     RowBox[{"1", "+", "\[Delta]"}], ")"}]}]]],
  CellChangeTimes->{
   3.7812740033403482`*^9, 3.7814573218417273`*^9, 3.7818848609506083`*^9, 
    3.781952821819541*^9, 3.78195289751086*^9, {3.7819538958219957`*^9, 
    3.781953915448163*^9}, 3.7821306544665728`*^9, 3.782747157952016*^9, 
    3.7830788934301643`*^9, 3.785162683931014*^9, 3.785164973948011*^9, 
    3.789732648885466*^9, 3.795174657109274*^9, 3.799320669052409*^9, 
    3.799340142599029*^9, 3.799756425128694*^9, 3.800007655671279*^9, 
    3.8002681846876783`*^9, 3.813930901799224*^9, {3.8139309343336077`*^9, 
    3.8139310598724957`*^9}, 3.8140037587039423`*^9, 3.814003805149556*^9},
  ExpressionUUID->"f0450e44-f8c2-4998-8163-46a32703ac33"],
 " gives two 0 eigenvalues and and eigenvalue of ",
 Cell[BoxData[
  RowBox[{"2", "+", 
   RowBox[{
    FractionBox["1", "2"], " ", "b", " ", 
    RowBox[{"(", 
     RowBox[{
      RowBox[{"-", "1"}], "+", "sf"}], ")"}], " ", 
    RowBox[{"(", 
     RowBox[{"1", "+", "\[Delta]"}], ")"}]}]}]],
  CellChangeTimes->{3.814003986297804*^9},ExpressionUUID->
  "6df63515-868a-41a0-bc99-34f7f0a4f844"],
 ", corresponding to the stability of the ecological equilibrium population \
size. Thus, we can see that the demographic dynamics are orthogonal to the \
evolutionary dynamics. \n\nHere,  we divide through by this factor to \
simplify the equation and evaluate the evolutionary dynamics. We then write \
the characteristic polynomial in the form\n",
 Cell[BoxData[
  RowBox[{
   SuperscriptBox["a\[Lambda]", "2"], "+", "b\[Lambda]", "+", "c"}]],
  CellChangeTimes->{
   3.7812740321927767`*^9, 3.781274073867482*^9, 3.7814285214279013`*^9, 
    3.781457321850852*^9, 3.78188486097983*^9, 3.781952821828101*^9, 
    3.7819528975187283`*^9, {3.781953895876915*^9, 3.7819539154956713`*^9}, 
    3.7821306544745502`*^9, 3.782747157961987*^9, 3.7830788934387407`*^9, 
    3.785162684364028*^9, 3.785164974025779*^9, 3.789732649060177*^9, 
    3.79517465722147*^9, 3.799320669259963*^9, 3.799340142805358*^9, 
    3.799756425312364*^9, 3.800007655751677*^9, 3.80026818474804*^9, 
    3.8139309588718987`*^9, 3.813931061212427*^9, 3.814004119936627*^9},
  ExpressionUUID->"fb6a3c2f-e0b2-4314-881c-853754ff0d06"],
 " \nfinding the a, b, and c coefficients. "
}], "Text",
 CellChangeTimes->{{3.814004985097933*^9, 3.814004991560975*^9}, {
  3.814005062086405*^9, 3.814005085289956*^9}, {3.814005142323441*^9, 
  3.8140051616191607`*^9}},ExpressionUUID->"8cc660ae-b589-4086-aacc-\
522dd5b2a3cf"],

Cell[CellGroupData[{

Cell[BoxData[{
 RowBox[{"Simplify", "[", 
  RowBox[{"charpolynomial1", "/", 
   RowBox[{"(", 
    FractionBox[
     RowBox[{
      RowBox[{"(", 
       RowBox[{"4", "-", "b", "+", 
        RowBox[{"b", " ", "sf"}], "-", 
        RowBox[{"b", " ", "\[Delta]"}], "+", 
        RowBox[{"b", " ", "sf", " ", "\[Delta]"}], "-", 
        RowBox[{"2", " ", "\[Lambda]"}]}], ")"}], " ", 
      SuperscriptBox["\[Lambda]", "2"], " "}], 
     RowBox[{"4", " ", 
      RowBox[{"(", 
       RowBox[{
        RowBox[{"-", "1"}], "+", "sf"}], ")"}], " ", 
      RowBox[{"(", 
       RowBox[{
        RowBox[{"-", "1"}], "+", "sm"}], ")"}], " ", 
      RowBox[{"(", 
       RowBox[{"1", "+", "\[Delta]"}], ")"}]}]], ")"}]}], 
  "]"}], "\[IndentingNewLine]", 
 RowBox[{"Collect", "[", 
  RowBox[{"%", ",", "\[Lambda]", ",", "Simplify"}], "]"}]}], "Input",
 CellChangeTimes->{{3.814005280882255*^9, 3.814005284996513*^9}},
 CellLabel->
  "In[1324]:=",ExpressionUUID->"e5f02f23-dcee-4bc6-91ee-8d5917ca99af"],

Cell[BoxData[
 RowBox[{
  RowBox[{"-", "1"}], "+", 
  RowBox[{
   RowBox[{"(", 
    RowBox[{
     RowBox[{"-", "1"}], "+", "sm"}], ")"}], " ", 
   RowBox[{"(", 
    RowBox[{"1", "+", "\[Delta]"}], ")"}], " ", "\[Lambda]"}], "+", 
  RowBox[{"2", " ", 
   RowBox[{"(", 
    RowBox[{
     RowBox[{"-", "1"}], "+", "sf"}], ")"}], " ", 
   RowBox[{"(", 
    RowBox[{
     RowBox[{"-", "1"}], "+", "sm"}], ")"}], " ", 
   RowBox[{"(", 
    RowBox[{"1", "+", "\[Delta]"}], ")"}], " ", 
   SuperscriptBox["\[Lambda]", "2"]}], "+", 
  RowBox[{"h", " ", 
   RowBox[{"(", 
    RowBox[{"sf", "-", 
     RowBox[{"sf", " ", 
      RowBox[{"(", 
       RowBox[{
        RowBox[{"-", "1"}], "+", "sm"}], ")"}], " ", 
      RowBox[{"(", 
       RowBox[{"1", "+", "\[Delta]"}], ")"}], " ", "\[Lambda]"}]}], 
    ")"}]}]}]], "Output",
 CellChangeTimes->{{3.814005268764485*^9, 3.814005285448605*^9}, 
   3.814081462523357*^9, 3.814189401040678*^9, 3.814195964969289*^9, 
   3.814196514110498*^9},
 CellLabel->
  "Out[1324]=",ExpressionUUID->"e26f52ca-46cb-4df1-b9cf-e718b0ae0acc"],

Cell[BoxData[
 RowBox[{
  RowBox[{"-", "1"}], "+", 
  RowBox[{"h", " ", "sf"}], "+", 
  RowBox[{
   RowBox[{"(", 
    RowBox[{"1", "-", 
     RowBox[{"h", " ", "sf"}]}], ")"}], " ", 
   RowBox[{"(", 
    RowBox[{
     RowBox[{"-", "1"}], "+", "sm"}], ")"}], " ", 
   RowBox[{"(", 
    RowBox[{"1", "+", "\[Delta]"}], ")"}], " ", "\[Lambda]"}], "+", 
  RowBox[{"2", " ", 
   RowBox[{"(", 
    RowBox[{
     RowBox[{"-", "1"}], "+", "sf"}], ")"}], " ", 
   RowBox[{"(", 
    RowBox[{
     RowBox[{"-", "1"}], "+", "sm"}], ")"}], " ", 
   RowBox[{"(", 
    RowBox[{"1", "+", "\[Delta]"}], ")"}], " ", 
   SuperscriptBox["\[Lambda]", "2"]}]}]], "Output",
 CellChangeTimes->{{3.814005268764485*^9, 3.814005285448605*^9}, 
   3.814081462523357*^9, 3.814189401040678*^9, 3.814195964969289*^9, 
   3.814196514111409*^9},
 CellLabel->
  "Out[1325]=",ExpressionUUID->"5fe05b59-72a1-4ede-a803-68b8ba53b75b"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[{
 RowBox[{
  RowBox[{"Simplify", "[", 
   RowBox[{"charpolynomial1", "/", 
    RowBox[{"(", 
     FractionBox[
      RowBox[{
       RowBox[{"(", 
        RowBox[{"4", "-", "b", "+", 
         RowBox[{"b", " ", "sf"}], "-", 
         RowBox[{"b", " ", "\[Delta]"}], "+", 
         RowBox[{"b", " ", "sf", " ", "\[Delta]"}], "-", 
         RowBox[{"2", " ", "\[Lambda]"}]}], ")"}], " ", 
       SuperscriptBox["\[Lambda]", "2"], " "}], 
      RowBox[{"4", " ", 
       RowBox[{"(", 
        RowBox[{
         RowBox[{"-", "1"}], "+", "sf"}], ")"}], " ", 
       RowBox[{"(", 
        RowBox[{
         RowBox[{"-", "1"}], "+", "sm"}], ")"}], " ", 
       RowBox[{"(", 
        RowBox[{"1", "+", "\[Delta]"}], ")"}]}]], ")"}]}], "]"}], 
  ";"}], "\[IndentingNewLine]", 
 RowBox[{
  RowBox[{"Collect", "[", 
   RowBox[{"%", ",", "\[Lambda]", ",", "Simplify"}], "]"}], 
  ";"}], "\[IndentingNewLine]", 
 RowBox[{
  RowBox[{"Coefficient", "[", 
   RowBox[{"%", ",", 
    RowBox[{"\[Lambda]", "^", "2"}]}], "]"}], ";"}], "\[IndentingNewLine]", 
 RowBox[{"charpolySimpleACoef1", "=", 
  RowBox[{"Collect", "[", 
   RowBox[{
    RowBox[{"%%", "/", "%"}], ",", "\[Lambda]", ",", "Simplify"}], 
   "]"}]}], "\[IndentingNewLine]", 
 RowBox[{
  RowBox[{"acoef1", "=", 
   RowBox[{"Coefficient", "[", 
    RowBox[{"charpolySimpleACoef1", ",", 
     RowBox[{"\[Lambda]", "^", "2"}]}], "]"}]}], ";"}], "\[IndentingNewLine]", 
 RowBox[{
  RowBox[{"bcoef1", "=", 
   RowBox[{"Coefficient", "[", 
    RowBox[{"charpolySimpleACoef1", ",", "\[Lambda]"}], "]"}]}], 
  ";"}], "\[IndentingNewLine]", 
 RowBox[{
  RowBox[{"ccoef1", "=", 
   RowBox[{"charpolySimpleACoef1", "-", " ", 
    RowBox[{"bcoef1", "*", "\[Lambda]"}], "-", 
    RowBox[{"acoef1", "*", 
     RowBox[{"\[Lambda]", "^", "2"}]}]}]}], ";"}]}], "Input",
 CellChangeTimes->{{3.814005188468329*^9, 3.814005191684287*^9}, {
  3.814005253330809*^9, 3.814005259347432*^9}, {3.8140052975628242`*^9, 
  3.8140053332696457`*^9}},
 CellLabel->
  "In[1326]:=",ExpressionUUID->"4f2268e3-8b21-42ca-9c21-5f26150d9928"],

Cell[BoxData[
 RowBox[{
  FractionBox[
   RowBox[{
    RowBox[{"-", "1"}], "+", 
    RowBox[{"h", " ", "sf"}]}], 
   RowBox[{"2", " ", 
    RowBox[{"(", 
     RowBox[{
      RowBox[{"-", "1"}], "+", "sf"}], ")"}], " ", 
    RowBox[{"(", 
     RowBox[{
      RowBox[{"-", "1"}], "+", "sm"}], ")"}], " ", 
    RowBox[{"(", 
     RowBox[{"1", "+", "\[Delta]"}], ")"}]}]], "+", 
  FractionBox[
   RowBox[{
    RowBox[{"(", 
     RowBox[{"1", "-", 
      RowBox[{"h", " ", "sf"}]}], ")"}], " ", "\[Lambda]"}], 
   RowBox[{"2", " ", 
    RowBox[{"(", 
     RowBox[{
      RowBox[{"-", "1"}], "+", "sf"}], ")"}]}]], "+", 
  SuperscriptBox["\[Lambda]", "2"]}]], "Output",
 CellChangeTimes->{3.81400519422432*^9, 3.814005259847539*^9, 
  3.8140053004626207`*^9, 3.814005333624386*^9, 3.814081463662334*^9, 
  3.814189401055848*^9, 3.8141959649815197`*^9, 3.814196514120184*^9},
 CellLabel->
  "Out[1329]=",ExpressionUUID->"f7de4613-1dc1-4969-b75c-fda2256b921f"]
}, Open  ]],

Cell["\<\
The discriminant is nonnegative and so the eigenvalues are real:\
\>", "Text",
 CellChangeTimes->{{3.7814370117088013`*^9, 3.7814370178709927`*^9}, {
  3.7814370976392593`*^9, 
  3.781437109782298*^9}},ExpressionUUID->"f9cb2b4e-b36a-47ac-9f6c-\
6aabd0033d26"],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{
  RowBox[{"bcoef1", "^", "2"}], "-", 
  RowBox[{"4", "acoef1", " ", "ccoef1"}]}]], "Input",
 CellLabel->
  "In[1333]:=",ExpressionUUID->"47a59791-140b-49c7-ad27-d57f3a116d92"],

Cell[BoxData[
 RowBox[{
  FractionBox[
   SuperscriptBox[
    RowBox[{"(", 
     RowBox[{"1", "-", 
      RowBox[{"h", " ", "sf"}]}], ")"}], "2"], 
   RowBox[{"4", " ", 
    SuperscriptBox[
     RowBox[{"(", 
      RowBox[{
       RowBox[{"-", "1"}], "+", "sf"}], ")"}], "2"]}]], "-", 
  FractionBox[
   RowBox[{"2", " ", 
    RowBox[{"(", 
     RowBox[{
      RowBox[{"-", "1"}], "+", 
      RowBox[{"h", " ", "sf"}]}], ")"}]}], 
   RowBox[{
    RowBox[{"(", 
     RowBox[{
      RowBox[{"-", "1"}], "+", "sf"}], ")"}], " ", 
    RowBox[{"(", 
     RowBox[{
      RowBox[{"-", "1"}], "+", "sm"}], ")"}], " ", 
    RowBox[{"(", 
     RowBox[{"1", "+", "\[Delta]"}], ")"}]}]]}]], "Output",
 CellChangeTimes->{3.81400537457533*^9, 3.81408146557857*^9, 
  3.814189401089308*^9, 3.814195965024002*^9, 3.8141965141679697`*^9},
 CellLabel->
  "Out[1333]=",ExpressionUUID->"fdfd170c-057e-4e00-84ac-17c8422a404d"]
}, Open  ]],

Cell[TextData[{
 "The conditions for the above characteristic polynomial ",
 Cell[BoxData[
  FormBox["p", TraditionalForm]],ExpressionUUID->
  "78b037a2-7d51-4c17-b3e4-906ac4e9a748"],
 "  to have roots between -1 and 1 are:\n1. ",
 Cell[BoxData[
  FormBox[
   RowBox[{
    RowBox[{"p", "(", "1", ")"}], " ", ">", " ", "0"}], TraditionalForm]],
  ExpressionUUID->"2f3770b3-9338-4e60-b602-37c0494df127"],
 "\n2. ",
 Cell[BoxData[
  FormBox[
   RowBox[{
    RowBox[{"p", "(", 
     RowBox[{"-", "1"}], ")"}], " ", ">", " ", "0"}], TraditionalForm]],
  ExpressionUUID->"ba8ca884-2a2f-4d52-81bc-9f514af16708"],
 "\n3.  ",
 Cell[BoxData[
  FormBox[
   RowBox[{
    RowBox[{"-", "2"}], " ", "<", " ", "b", " "}], TraditionalForm]],
  ExpressionUUID->"3f97142f-3c14-4977-87af-b720f5590eb5"],
 "\n4.  ",
 Cell[BoxData[
  FormBox[
   RowBox[{"b", " ", "<", " ", "2", " "}], TraditionalForm]],ExpressionUUID->
  "c56727ab-21a7-487d-b455-8d61f4f242d6"],
 "\nWhen any of these are false, the largest eigenvalue is ",
 Cell[BoxData[
  FormBox[
   RowBox[{">", 
    RowBox[{"1", "\:201a", " "}]}], TraditionalForm]],ExpressionUUID->
  "be21dc3f-b8cf-41b3-b539-aaeaca5f04c2"],
 "and the equilibrium is unstable. \n\nNow checking these conditions for \
instability by reversing the inequalities required for stability"
}], "Text",
 CellChangeTimes->{{3.7814371420798693`*^9, 3.781437152698497*^9}, {
  3.814005459930065*^9, 
  3.814005579369294*^9}},ExpressionUUID->"ea849c1f-d447-44c8-a831-\
dda615a758e0"],

Cell[CellGroupData[{

Cell[BoxData[{
 RowBox[{"r1", " ", "=", " ", 
  RowBox[{"FullSimplify", "[", 
   RowBox[{
    RowBox[{
     RowBox[{"(", 
      RowBox[{"charpolySimpleACoef1", "/.", 
       RowBox[{"\[Lambda]", "\[Rule]", "1"}]}], " ", ")"}], "<", " ", "0"}], 
    ",", "  ", 
    RowBox[{"{", 
     RowBox[{
      RowBox[{"1", ">", " ", "\[Delta]", ">", "0"}], ",", 
      RowBox[{"1", ">", "h", ">", "0"}], ",", 
      RowBox[{"1", ">", "sm", ">", "0"}], ",", 
      RowBox[{"0", "<", "sf", "<", "1"}]}], "}"}]}], 
   "]"}]}], "\[IndentingNewLine]", 
 RowBox[{"r2", " ", "=", 
  RowBox[{"FullSimplify", "[", 
   RowBox[{
    RowBox[{
     RowBox[{"(", 
      RowBox[{"charpolySimpleACoef1", "/.", 
       RowBox[{"\[Lambda]", "\[Rule]", 
        RowBox[{"-", "1"}]}]}], " ", ")"}], "<", " ", "0"}], ",", "  ", 
    RowBox[{"{", 
     RowBox[{
      RowBox[{"1", ">", " ", "\[Delta]", ">", "0"}], ",", 
      RowBox[{"1", ">", "h", ">", "0"}], ",", 
      RowBox[{"1", ">", "sm", ">", "0"}], ",", 
      RowBox[{"0", "<", "sf", "<", "1"}]}], "}"}]}], 
   "]"}]}], "\[IndentingNewLine]", 
 RowBox[{"r3", "=", 
  RowBox[{"FullSimplify", "[", 
   RowBox[{
    RowBox[{"bcoef1", "<", 
     RowBox[{"-", "2"}]}], ",", 
    RowBox[{"{", 
     RowBox[{
      RowBox[{"1", ">", " ", "\[Delta]", ">", "0"}], ",", 
      RowBox[{"1", ">", "h", ">", "0"}], ",", 
      RowBox[{"1", ">", "sm", ">", "0"}], ",", 
      RowBox[{"0", "<", "sf", "<", "1"}]}], "}"}]}], 
   "]"}]}], "\[IndentingNewLine]", 
 RowBox[{"r4", "=", 
  RowBox[{"FullSimplify", "[", 
   RowBox[{
    RowBox[{"bcoef1", ">", "2"}], ",", 
    RowBox[{"{", 
     RowBox[{
      RowBox[{"1", ">", " ", "\[Delta]", ">", "0"}], ",", 
      RowBox[{"1", ">", "h", ">", "0"}], ",", 
      RowBox[{"1", ">", "sm", ">", "0"}], ",", 
      RowBox[{"0", "<", "sf", "<", "1"}]}], "}"}]}], "]"}]}]}], "Input",
 CellChangeTimes->{{3.781437173950416*^9, 3.78143720186456*^9}, {
  3.781437293015765*^9, 3.781437392444491*^9}, {3.781437548946312*^9, 
  3.781437594361197*^9}, {3.781439033702999*^9, 3.781439042554227*^9}, {
  3.7814390879120607`*^9, 3.781439133547997*^9}, {3.7814394461550627`*^9, 
  3.78143945760787*^9}},
 CellLabel->
  "In[1334]:=",ExpressionUUID->"6851f3e7-d9e9-4279-9edf-bd03416640c2"],

Cell[BoxData[
 RowBox[{
  RowBox[{"\[Delta]", "+", 
   RowBox[{"2", " ", "sf", " ", 
    RowBox[{"(", 
     RowBox[{
      RowBox[{"-", "1"}], "+", "sm"}], ")"}], " ", 
    RowBox[{"(", 
     RowBox[{"1", "+", "\[Delta]"}], ")"}]}], "+", 
   RowBox[{"h", " ", "sf", " ", 
    RowBox[{"(", 
     RowBox[{"2", "+", "\[Delta]", "-", 
      RowBox[{"sm", " ", 
       RowBox[{"(", 
        RowBox[{"1", "+", "\[Delta]"}], ")"}]}]}], ")"}]}]}], "<", 
  RowBox[{"sm", " ", 
   RowBox[{"(", 
    RowBox[{"1", "+", "\[Delta]"}], ")"}]}]}]], "Output",
 CellChangeTimes->{{3.781437175151822*^9, 3.781437202530854*^9}, {
   3.781437295782474*^9, 3.781437352236356*^9}, 3.7814374035892963`*^9, 
   3.781437613171399*^9, 3.781439043535357*^9, {3.781439104876899*^9, 
   3.7814391344053497`*^9}, {3.7814394535688467`*^9, 3.7814394581207047`*^9}, 
   3.781457322620882*^9, 3.78188486163474*^9, 3.781952822497775*^9, 
   3.781952898261024*^9, {3.781953896487258*^9, 3.781953916168281*^9}, 
   3.782130655037745*^9, 3.7827471589942493`*^9, 3.783078894796163*^9, 
   3.785162687556292*^9, 3.785164975528594*^9, 3.7897326510301743`*^9, 
   3.795174658845468*^9, 3.799320671848111*^9, 3.7993401455539513`*^9, 
   3.7997564277832108`*^9, 3.800007657980575*^9, 3.800268186111622*^9, 
   3.8140055881642313`*^9, 3.814081468356298*^9, 3.8141894011950912`*^9, 
   3.8141959652989473`*^9, 3.814196514437024*^9},
 CellLabel->
  "Out[1334]=",ExpressionUUID->"a39878f8-78a0-48a9-a1a2-8548eff096cc"],

Cell[BoxData[
 RowBox[{
  RowBox[{"2", "+", 
   RowBox[{"3", " ", "\[Delta]"}], "+", 
   RowBox[{"sf", " ", 
    RowBox[{"(", 
     RowBox[{
      RowBox[{"-", "2"}], "-", 
      RowBox[{
       RowBox[{"(", 
        RowBox[{"2", "+", "h"}], ")"}], " ", "\[Delta]"}], "+", 
      RowBox[{
       RowBox[{"(", 
        RowBox[{"2", "+", "h"}], ")"}], " ", "sm", " ", 
       RowBox[{"(", 
        RowBox[{"1", "+", "\[Delta]"}], ")"}]}]}], ")"}]}]}], "<", 
  RowBox[{"3", " ", "sm", " ", 
   RowBox[{"(", 
    RowBox[{"1", "+", "\[Delta]"}], ")"}]}]}]], "Output",
 CellChangeTimes->{{3.781437175151822*^9, 3.781437202530854*^9}, {
   3.781437295782474*^9, 3.781437352236356*^9}, 3.7814374035892963`*^9, 
   3.781437613171399*^9, 3.781439043535357*^9, {3.781439104876899*^9, 
   3.7814391344053497`*^9}, {3.7814394535688467`*^9, 3.7814394581207047`*^9}, 
   3.781457322620882*^9, 3.78188486163474*^9, 3.781952822497775*^9, 
   3.781952898261024*^9, {3.781953896487258*^9, 3.781953916168281*^9}, 
   3.782130655037745*^9, 3.7827471589942493`*^9, 3.783078894796163*^9, 
   3.785162687556292*^9, 3.785164975528594*^9, 3.7897326510301743`*^9, 
   3.795174658845468*^9, 3.799320671848111*^9, 3.7993401455539513`*^9, 
   3.7997564277832108`*^9, 3.800007657980575*^9, 3.800268186111622*^9, 
   3.8140055881642313`*^9, 3.814081468356298*^9, 3.8141894011950912`*^9, 
   3.8141959652989473`*^9, 3.8141965146853523`*^9},
 CellLabel->
  "Out[1335]=",ExpressionUUID->"0d326dd4-b466-4acf-a0dc-aa8076598e89"],

Cell[BoxData[
 RowBox[{
  RowBox[{"3", "+", 
   RowBox[{"h", " ", "sf"}]}], "<", 
  RowBox[{"4", " ", "sf"}]}]], "Output",
 CellChangeTimes->{{3.781437175151822*^9, 3.781437202530854*^9}, {
   3.781437295782474*^9, 3.781437352236356*^9}, 3.7814374035892963`*^9, 
   3.781437613171399*^9, 3.781439043535357*^9, {3.781439104876899*^9, 
   3.7814391344053497`*^9}, {3.7814394535688467`*^9, 3.7814394581207047`*^9}, 
   3.781457322620882*^9, 3.78188486163474*^9, 3.781952822497775*^9, 
   3.781952898261024*^9, {3.781953896487258*^9, 3.781953916168281*^9}, 
   3.782130655037745*^9, 3.7827471589942493`*^9, 3.783078894796163*^9, 
   3.785162687556292*^9, 3.785164975528594*^9, 3.7897326510301743`*^9, 
   3.795174658845468*^9, 3.799320671848111*^9, 3.7993401455539513`*^9, 
   3.7997564277832108`*^9, 3.800007657980575*^9, 3.800268186111622*^9, 
   3.8140055881642313`*^9, 3.814081468356298*^9, 3.8141894011950912`*^9, 
   3.8141959652989473`*^9, 3.81419651473297*^9},
 CellLabel->
  "Out[1336]=",ExpressionUUID->"9d98d3a1-1f76-4752-88d8-b1b992cab3f0"],

Cell[BoxData["False"], "Output",
 CellChangeTimes->{{3.781437175151822*^9, 3.781437202530854*^9}, {
   3.781437295782474*^9, 3.781437352236356*^9}, 3.7814374035892963`*^9, 
   3.781437613171399*^9, 3.781439043535357*^9, {3.781439104876899*^9, 
   3.7814391344053497`*^9}, {3.7814394535688467`*^9, 3.7814394581207047`*^9}, 
   3.781457322620882*^9, 3.78188486163474*^9, 3.781952822497775*^9, 
   3.781952898261024*^9, {3.781953896487258*^9, 3.781953916168281*^9}, 
   3.782130655037745*^9, 3.7827471589942493`*^9, 3.783078894796163*^9, 
   3.785162687556292*^9, 3.785164975528594*^9, 3.7897326510301743`*^9, 
   3.795174658845468*^9, 3.799320671848111*^9, 3.7993401455539513`*^9, 
   3.7997564277832108`*^9, 3.800007657980575*^9, 3.800268186111622*^9, 
   3.8140055881642313`*^9, 3.814081468356298*^9, 3.8141894011950912`*^9, 
   3.8141959652989473`*^9, 3.8141965147338753`*^9},
 CellLabel->
  "Out[1337]=",ExpressionUUID->"ccb76f38-d119-4c6c-abc9-5087a899c08a"]
}, Open  ]],

Cell[TextData[{
 "Region r3 is independent of drive and relates only to female heterozygote \
and drive homozygote fitness (it can be rewritten as ",
 Cell[BoxData[
  FormBox[
   RowBox[{
    RowBox[{"1", "-", "s_f"}], " ", "<", " ", 
    RowBox[{
     RowBox[{"(", 
      RowBox[{"1", "-", "hs_f"}], ")"}], "/", "4"}]}], TraditionalForm]],
  ExpressionUUID->"3565beb5-f898-43d1-a701-a48c03571554"],
 ").\n\nThe region defined by r1 is included in all the other regions, so we \
need only satisfy this inequality"
}], "Text",
 CellChangeTimes->{{3.814005601784459*^9, 
  3.8140056112211018`*^9}},ExpressionUUID->"9af9eb1e-7c80-466b-af4a-\
62179fc49259"],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"Reduce", "[", 
  RowBox[{"{", 
   RowBox[{
    RowBox[{"!", "r1"}], ",", "r2", ",", "r3", ",", 
    RowBox[{"1", ">", " ", "\[Delta]", ">", "0"}], ",", 
    RowBox[{"1", ">", "h", ">", "0"}], ",", 
    RowBox[{"1", ">", "sm", ">", "0"}], ",", 
    RowBox[{"0", "<", "sf", "<", "1"}]}], "}"}], "]"}]], "Input",
 CellChangeTimes->{{3.781438491275654*^9, 3.7814385705794687`*^9}, {
  3.781438609675919*^9, 3.781438609925259*^9}, {3.781439361820053*^9, 
  3.781439383480029*^9}, {3.7818742029198303`*^9, 3.7818742196134233`*^9}},
 CellLabel->
  "In[1338]:=",ExpressionUUID->"491fb21d-1290-499e-87eb-f450adfd4779"],

Cell[BoxData["False"], "Output",
 CellChangeTimes->{
  3.7814385244255323`*^9, {3.78143855490099*^9, 3.781438571329862*^9}, 
   3.781438610236524*^9, {3.781439344792201*^9, 3.781439389623969*^9}, 
   3.7814394716890707`*^9, 3.781440562046509*^9, 3.781457323045916*^9, {
   3.781874206107297*^9, 3.781874221613144*^9}, 3.781884862070531*^9, 
   3.7819528229217377`*^9, 3.781952898586425*^9, {3.7819538968695173`*^9, 
   3.781953916535781*^9}, 3.782130655424851*^9, 3.7827471595772257`*^9, 
   3.783078895331085*^9, 3.7851626882942743`*^9, 3.785164976226273*^9, 
   3.7897326518393*^9, 3.79517465934138*^9, 3.799320672547058*^9, 
   3.799340146185979*^9, 3.799756428376923*^9, 3.8000076587001953`*^9, 
   3.800268186632812*^9, 3.814005589787862*^9, 3.8140814700973053`*^9, 
   3.814189401503531*^9, 3.814195965662497*^9, 3.814196514810627*^9},
 CellLabel->
  "Out[1338]=",ExpressionUUID->"e62d4d91-a610-4f42-8f94-af359b527e44"]
}, Open  ]],

Cell["Thus, the instability conditions are:", "Text",
 CellChangeTimes->{{3.814005633840632*^9, 
  3.8140056425480337`*^9}},ExpressionUUID->"82cebce6-a58c-41d6-9fe7-\
68f3c37b6f42"],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"unstabConditions1", " ", "=", " ", "r1"}]], "Input",
 CellChangeTimes->{{3.781440566940112*^9, 3.7814405727307663`*^9}},
 CellLabel->
  "In[1339]:=",ExpressionUUID->"eba6001c-7ecc-487c-bc04-abc003cfbda8"],

Cell[BoxData[
 RowBox[{
  RowBox[{"\[Delta]", "+", 
   RowBox[{"2", " ", "sf", " ", 
    RowBox[{"(", 
     RowBox[{
      RowBox[{"-", "1"}], "+", "sm"}], ")"}], " ", 
    RowBox[{"(", 
     RowBox[{"1", "+", "\[Delta]"}], ")"}]}], "+", 
   RowBox[{"h", " ", "sf", " ", 
    RowBox[{"(", 
     RowBox[{"2", "+", "\[Delta]", "-", 
      RowBox[{"sm", " ", 
       RowBox[{"(", 
        RowBox[{"1", "+", "\[Delta]"}], ")"}]}]}], ")"}]}]}], "<", 
  RowBox[{"sm", " ", 
   RowBox[{"(", 
    RowBox[{"1", "+", "\[Delta]"}], ")"}]}]}]], "Output",
 CellChangeTimes->{
  3.781440573583838*^9, 3.7814573230570507`*^9, 3.781884862079309*^9, 
   3.781952822929488*^9, 3.7819528985964413`*^9, {3.7819538968833323`*^9, 
   3.78195391654261*^9}, 3.7821306554345818`*^9, 3.782747159587644*^9, 
   3.78307889534017*^9, 3.785162688501986*^9, 3.78516497629982*^9, 
   3.7897326519250937`*^9, 3.7951746594067907`*^9, 3.799320672638276*^9, 
   3.799340146500635*^9, 3.799756428598815*^9, 3.8000076589437943`*^9, 
   3.800268186690694*^9, 3.814005590612234*^9, 3.8140814712053947`*^9, 
   3.8141894015172462`*^9, 3.814195965669839*^9, 3.814196514816902*^9},
 CellLabel->
  "Out[1339]=",ExpressionUUID->"463bdf59-0013-4c54-bd28-cf8aa8c64cf9"]
}, Open  ]],

Cell["which can be re-written using the w fitness notation", "Text",
 CellChangeTimes->{{3.78187228235487*^9, 3.781872375904385*^9}, 
   3.814005627241395*^9},ExpressionUUID->"65db19cc-62c1-4302-9fc4-\
09cc92a8709c"],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"FullSimplify", "[", 
  RowBox[{"unstabConditions1", ",", 
   RowBox[{"{", 
    RowBox[{
     RowBox[{"wXdY", "\[Equal]", 
      RowBox[{"1", "-", "sm"}]}], ",", 
     RowBox[{"wXdX", " ", "\[Equal]", " ", 
      RowBox[{"1", " ", "-", " ", 
       RowBox[{"h", " ", "sf"}]}]}], ",", 
     RowBox[{"wXdXd", " ", "\[Equal]", " ", 
      RowBox[{"1", " ", "-", " ", "sf"}]}]}], "}"}]}], "]"}]], "Input",
 CellChangeTimes->{{3.783082107345252*^9, 3.7830821633262053`*^9}},
 CellLabel->
  "In[1340]:=",ExpressionUUID->"de8f62a2-64e2-404d-aae8-663b919d7c2d"],

Cell[BoxData[
 RowBox[{
  RowBox[{"2", " ", "wXdXd", " ", "wXdY", " ", 
   RowBox[{"(", 
    RowBox[{"1", "+", "\[Delta]"}], ")"}]}], "<", 
  RowBox[{"wXdX", " ", 
   RowBox[{"(", 
    RowBox[{"1", "+", "wXdY", "+", 
     RowBox[{"wXdY", " ", "\[Delta]"}]}], ")"}]}]}]], "Output",
 CellChangeTimes->{{3.783082123089921*^9, 3.7830821639965973`*^9}, 
   3.785162688703109*^9, 3.785164976483889*^9, 3.789732652155534*^9, 
   3.795174659581051*^9, 3.799320672863186*^9, 3.799340146798864*^9, 
   3.79975642881137*^9, 3.800007659149609*^9, 3.800268186858501*^9, 
   3.814081472383049*^9, 3.814189401559244*^9, 3.8141959657852297`*^9, 
   3.814196514932007*^9},
 CellLabel->
  "Out[1340]=",ExpressionUUID->"d70e4ebf-e872-4ddc-8d19-edee238d7459"]
}, Open  ]]
}, Closed]]
}, Closed]],

Cell[CellGroupData[{

Cell["Critical birth rate", "Subsection",
 CellChangeTimes->{{3.781270243104228*^9, 
  3.781270244885517*^9}},ExpressionUUID->"84474009-2768-4dd8-97c9-\
3a4b7276fabc"],

Cell["\<\
The critical birth rate is the lowest birth rate for which a population still \
exists (equilibrium population size, N, bigger than 0). 

We first find the critical birth rate when drive is absent, polymorphic, or \
fixed. \
\>", "Text",
 CellChangeTimes->{{3.813930468388834*^9, 3.8139304720109*^9}, {
  3.8139306445371513`*^9, 
  3.813930713094428*^9}},ExpressionUUID->"a6911daa-5507-424e-90e6-\
cb28c25dcca6"],

Cell[CellGroupData[{

Cell[BoxData[{
 RowBox[{"bcritNoDrive", "=", 
  RowBox[{"b", "/.", 
   RowBox[{
    RowBox[{"Solve", "[", 
     RowBox[{
      RowBox[{
       RowBox[{"Simplify", "[", 
        RowBox[{
         RowBox[{"(", 
          RowBox[{"maleNum", " ", "+", " ", "femaleNum"}], ")"}], "/.", 
         "noDrivePopSizeSol"}], "]"}], "\[Equal]", "0"}], ",", "b"}], "]"}], 
    "[", 
    RowBox[{"[", "1", "]"}], "]"}]}]}], "\[IndentingNewLine]", 
 RowBox[{"bcritEquil", "=", 
  RowBox[{"b", "/.", 
   RowBox[{
    RowBox[{"Solve", "[", 
     RowBox[{
      RowBox[{"Neq", "\[Equal]", "0"}], ",", "b"}], "]"}], "[", 
    RowBox[{"[", "1", "]"}], "]"}]}]}], "\[IndentingNewLine]", 
 RowBox[{"bcritFixed", "=", 
  RowBox[{"b", "/.", 
   RowBox[{
    RowBox[{"Solve", "[", 
     RowBox[{
      RowBox[{"Nfixed", "\[Equal]", "0"}], ",", "b"}], "]"}], "[", 
    RowBox[{"[", "1", "]"}], "]"}]}]}]}], "Input",
 CellChangeTimes->{{3.813928398062742*^9, 3.813928561860159*^9}},
 CellLabel->
  "In[1341]:=",ExpressionUUID->"a84301ba-cf0e-4539-a1f4-955b7eb24eff"],

Cell[BoxData["2"], "Output",
 CellChangeTimes->{
  3.813928401523972*^9, {3.813928434003956*^9, 3.813928529425839*^9}, 
   3.813928562308132*^9, 3.814081785262821*^9, 3.814189401571858*^9, 
   3.814195965795162*^9, 3.8141965149389544`*^9},
 CellLabel->
  "Out[1341]=",ExpressionUUID->"a9fcffa3-1ba9-4eb2-8060-c38425e13d68"],

Cell[BoxData[
 RowBox[{"-", 
  RowBox[{"(", 
   RowBox[{
    RowBox[{"(", 
     RowBox[{"4", " ", 
      RowBox[{"(", 
       RowBox[{
        RowBox[{
         RowBox[{"-", "2"}], " ", "sf"}], "+", 
        RowBox[{"4", " ", "h", " ", "sf"}], "+", 
        RowBox[{"2", " ", "sf", " ", "sm"}], "-", 
        RowBox[{"4", " ", "h", " ", "sf", " ", "sm"}], "-", 
        SuperscriptBox["sm", "2"], "+", 
        RowBox[{"h", " ", "sf", " ", 
         SuperscriptBox["sm", "2"]}], "-", 
        RowBox[{"2", " ", "sf", " ", "\[Delta]"}], "+", 
        RowBox[{"2", " ", "h", " ", "sf", " ", "\[Delta]"}], "+", 
        RowBox[{"sm", " ", "\[Delta]"}], "+", 
        RowBox[{"2", " ", "sf", " ", "sm", " ", "\[Delta]"}], "-", 
        RowBox[{"3", " ", "h", " ", "sf", " ", "sm", " ", "\[Delta]"}], "-", 
        RowBox[{
         SuperscriptBox["sm", "2"], " ", "\[Delta]"}], "+", 
        RowBox[{"h", " ", "sf", " ", 
         SuperscriptBox["sm", "2"], " ", "\[Delta]"}]}], ")"}]}], ")"}], "/", 
    RowBox[{"(", 
     RowBox[{
      RowBox[{"4", " ", "sf"}], "-", 
      RowBox[{"8", " ", "h", " ", "sf"}], "+", 
      RowBox[{"4", " ", 
       SuperscriptBox["h", "2"], " ", 
       SuperscriptBox["sf", "2"]}], "-", 
      RowBox[{"4", " ", "sf", " ", "sm"}], "+", 
      RowBox[{"8", " ", "h", " ", "sf", " ", "sm"}], "-", 
      RowBox[{"4", " ", 
       SuperscriptBox["h", "2"], " ", 
       SuperscriptBox["sf", "2"], " ", "sm"}], "+", 
      SuperscriptBox["sm", "2"], "-", 
      RowBox[{"2", " ", "h", " ", "sf", " ", 
       SuperscriptBox["sm", "2"]}], "+", 
      RowBox[{
       SuperscriptBox["h", "2"], " ", 
       SuperscriptBox["sf", "2"], " ", 
       SuperscriptBox["sm", "2"]}], "+", 
      RowBox[{"4", " ", "sf", " ", "\[Delta]"}], "-", 
      RowBox[{"8", " ", "h", " ", "sf", " ", "\[Delta]"}], "+", 
      RowBox[{"4", " ", 
       SuperscriptBox["h", "2"], " ", 
       SuperscriptBox["sf", "2"], " ", "\[Delta]"}], "-", 
      RowBox[{"2", " ", "sm", " ", "\[Delta]"}], "-", 
      RowBox[{"4", " ", "sf", " ", "sm", " ", "\[Delta]"}], "+", 
      RowBox[{"12", " ", "h", " ", "sf", " ", "sm", " ", "\[Delta]"}], "-", 
      RowBox[{"6", " ", 
       SuperscriptBox["h", "2"], " ", 
       SuperscriptBox["sf", "2"], " ", "sm", " ", "\[Delta]"}], "+", 
      RowBox[{"2", " ", 
       SuperscriptBox["sm", "2"], " ", "\[Delta]"}], "-", 
      RowBox[{"4", " ", "h", " ", "sf", " ", 
       SuperscriptBox["sm", "2"], " ", "\[Delta]"}], "+", 
      RowBox[{"2", " ", 
       SuperscriptBox["h", "2"], " ", 
       SuperscriptBox["sf", "2"], " ", 
       SuperscriptBox["sm", "2"], " ", "\[Delta]"}], "+", 
      SuperscriptBox["\[Delta]", "2"], "-", 
      RowBox[{"2", " ", "h", " ", "sf", " ", 
       SuperscriptBox["\[Delta]", "2"]}], "+", 
      RowBox[{
       SuperscriptBox["h", "2"], " ", 
       SuperscriptBox["sf", "2"], " ", 
       SuperscriptBox["\[Delta]", "2"]}], "-", 
      RowBox[{"2", " ", "sm", " ", 
       SuperscriptBox["\[Delta]", "2"]}], "+", 
      RowBox[{"4", " ", "h", " ", "sf", " ", "sm", " ", 
       SuperscriptBox["\[Delta]", "2"]}], "-", 
      RowBox[{"2", " ", 
       SuperscriptBox["h", "2"], " ", 
       SuperscriptBox["sf", "2"], " ", "sm", " ", 
       SuperscriptBox["\[Delta]", "2"]}], "+", 
      RowBox[{
       SuperscriptBox["sm", "2"], " ", 
       SuperscriptBox["\[Delta]", "2"]}], "-", 
      RowBox[{"2", " ", "h", " ", "sf", " ", 
       SuperscriptBox["sm", "2"], " ", 
       SuperscriptBox["\[Delta]", "2"]}], "+", 
      RowBox[{
       SuperscriptBox["h", "2"], " ", 
       SuperscriptBox["sf", "2"], " ", 
       SuperscriptBox["sm", "2"], " ", 
       SuperscriptBox["\[Delta]", "2"]}]}], ")"}]}], ")"}]}]], "Output",
 CellChangeTimes->{
  3.813928401523972*^9, {3.813928434003956*^9, 3.813928529425839*^9}, 
   3.813928562308132*^9, 3.814081785262821*^9, 3.814189401571858*^9, 
   3.814195965795162*^9, 3.81419651498464*^9},
 CellLabel->
  "Out[1342]=",ExpressionUUID->"096912d8-e96d-4233-b417-bf0eaf30a9b1"],

Cell[BoxData[
 RowBox[{"-", 
  FractionBox["2", 
   RowBox[{
    RowBox[{"(", 
     RowBox[{
      RowBox[{"-", "1"}], "+", "sf"}], ")"}], " ", 
    RowBox[{"(", 
     RowBox[{"1", "+", "\[Delta]"}], ")"}]}]]}]], "Output",
 CellChangeTimes->{
  3.813928401523972*^9, {3.813928434003956*^9, 3.813928529425839*^9}, 
   3.813928562308132*^9, 3.814081785262821*^9, 3.814189401571858*^9, 
   3.814195965795162*^9, 3.814196514987307*^9},
 CellLabel->
  "Out[1343]=",ExpressionUUID->"f5964664-00d2-4470-b2f2-61f5cc56846d"]
}, Open  ]],

Cell["\<\
Given that drive is polymorphic (\[Phi]>0,\[CurlyPhi]>0), the critical birth \
rate is always smaller than it would have been without drive
Given that drive reaches fixation (\[Phi]>0,\[CurlyPhi]<0), the critical \
birth rate is always smaller than it would have been without drive\
\>", "Text",
 CellChangeTimes->{{3.813930720878367*^9, 
  3.813930768858816*^9}},ExpressionUUID->"a1a39302-2e9c-467b-853f-\
9039b002746c"],

Cell[CellGroupData[{

Cell[BoxData[{
 RowBox[{"Reduce", "[", 
  RowBox[{"{", 
   RowBox[{
    RowBox[{"bcritNoDrive", "<", "bcritEquil"}], ",", " ", 
    RowBox[{"0", "<", "sf", "<", "1"}], ",", 
    RowBox[{"0", "<", "h", "<", "1"}], ",", 
    RowBox[{"0", "<", "sm", "<", "1"}], ",", 
    RowBox[{"0", "<", "b"}], ",", " ", 
    RowBox[{"0", "<", "\[Delta]", "<", "1"}], ",", " ", 
    RowBox[{"\[Phi]", ">", "0"}], ",", 
    RowBox[{"\[CurlyPhi]", ">", "0"}]}], "}"}], "]"}], "\[IndentingNewLine]", 
 RowBox[{"Reduce", "[", 
  RowBox[{"{", 
   RowBox[{
    RowBox[{"bcritNoDrive", "<", "bcritFixed"}], ",", " ", 
    RowBox[{"0", "<", "sf", "<", "1"}], ",", 
    RowBox[{"0", "<", "h", "<", "1"}], ",", 
    RowBox[{"0", "<", "sm", "<", "1"}], ",", 
    RowBox[{"0", "<", "b"}], ",", " ", 
    RowBox[{"0", "<", "\[Delta]", "<", "1"}], ",", " ", 
    RowBox[{"\[Phi]", ">", "0"}], ",", 
    RowBox[{"\[CurlyPhi]", "<", "0"}]}], "}"}], "]"}]}], "Input",
 CellChangeTimes->{{3.81392856865447*^9, 3.81392863122628*^9}, 
   3.813928679185483*^9, {3.81393002322504*^9, 3.813930051570941*^9}, {
   3.813930082291155*^9, 3.813930177188037*^9}},
 CellLabel->
  "In[1344]:=",ExpressionUUID->"8dfa1109-9f28-44f2-ab69-f549df3c952a"],

Cell[BoxData["False"], "Output",
 CellChangeTimes->{{3.8139286049658337`*^9, 3.813928627951355*^9}, 
   3.813928684723576*^9, {3.8139300258073263`*^9, 3.8139300537471743`*^9}, {
   3.81393010421644*^9, 3.813930128247485*^9}, {3.813930158987651*^9, 
   3.8139301784794807`*^9}, 3.814081786639469*^9, 3.814189401790324*^9, 
   3.814195965961636*^9, 3.814196515124393*^9},
 CellLabel->
  "Out[1344]=",ExpressionUUID->"3e8a93d7-6ba5-4512-a7b9-d5cc32dd4360"],

Cell[BoxData["False"], "Output",
 CellChangeTimes->{{3.8139286049658337`*^9, 3.813928627951355*^9}, 
   3.813928684723576*^9, {3.8139300258073263`*^9, 3.8139300537471743`*^9}, {
   3.81393010421644*^9, 3.813930128247485*^9}, {3.813930158987651*^9, 
   3.8139301784794807`*^9}, 3.814081786639469*^9, 3.814189401790324*^9, 
   3.814195965961636*^9, 3.814196515159802*^9},
 CellLabel->
  "Out[1345]=",ExpressionUUID->"b886918b-da0f-4148-a97d-80a922096c2c"]
}, Open  ]],

Cell["\<\
The following functions define population sizes in these different cases for \
plotting (the same as the expressions derived above). \
\>", "Text",
 CellChangeTimes->{{3.7814619363838*^9, 3.7814619713578672`*^9}, {
  3.8000093523264723`*^9, 3.800009438490551*^9}, {3.813930793248259*^9, 
  3.813930824297714*^9}},ExpressionUUID->"92f7d12e-62d8-49ec-bd54-\
9a11abb6bf3e"],

Cell[BoxData[{
 RowBox[{
  RowBox[{"func", "[", 
   RowBox[{"{", 
    RowBox[{
    "sf_", ",", " ", "sm_", ",", " ", "h_", ",", " ", "\[Delta]_", ",", 
     "\[Alpha]_", ",", " ", "c_", ",", " ", "b_"}], "}"}], "]"}], ":=", 
  "\[IndentingNewLine]", 
  RowBox[{"If", "[", "\[IndentingNewLine]", 
   RowBox[{
    RowBox[{
     RowBox[{
      RowBox[{
       RowBox[{"(", 
        RowBox[{"sf", " ", 
         RowBox[{"(", 
          RowBox[{
           RowBox[{
            RowBox[{"-", 
             RowBox[{"(", 
              RowBox[{
               RowBox[{"-", "1"}], "+", "sm"}], ")"}]}], " ", 
            RowBox[{"(", 
             RowBox[{"1", "+", "\[Delta]"}], ")"}]}], "+", 
           RowBox[{"h", " ", 
            RowBox[{"(", 
             RowBox[{
              RowBox[{"-", "2"}], "+", "sm", "-", "\[Delta]", "+", 
              RowBox[{"sm", " ", "\[Delta]"}]}], ")"}]}]}], ")"}], " ", 
         RowBox[{"(", 
          RowBox[{
           RowBox[{"b", " ", 
            SuperscriptBox["h", "2"], " ", 
            SuperscriptBox["sf", "2"], " ", 
            SuperscriptBox[
             RowBox[{"(", 
              RowBox[{
               RowBox[{"-", "2"}], "+", "sm", "-", "\[Delta]", "+", 
               RowBox[{"sm", " ", "\[Delta]"}]}], ")"}], "2"]}], "+", 
           RowBox[{
            RowBox[{"(", 
             RowBox[{"sm", "-", "\[Delta]", "+", 
              RowBox[{"sm", " ", "\[Delta]"}]}], ")"}], " ", 
            RowBox[{"(", 
             RowBox[{
              RowBox[{
               RowBox[{"-", "b"}], " ", "\[Delta]"}], "+", 
              RowBox[{"sm", " ", 
               RowBox[{"(", 
                RowBox[{
                 RowBox[{"-", "4"}], "+", "b", "+", 
                 RowBox[{"b", " ", "\[Delta]"}]}], ")"}]}]}], ")"}]}], "-", 
           RowBox[{"2", " ", "sf", " ", 
            RowBox[{"(", 
             RowBox[{
              RowBox[{
               RowBox[{"-", "4"}], " ", 
               RowBox[{"(", 
                RowBox[{
                 RowBox[{"-", "1"}], "+", "sm"}], ")"}], " ", 
               RowBox[{"(", 
                RowBox[{"1", "+", "\[Delta]"}], ")"}]}], "+", 
              RowBox[{"2", " ", "b", " ", 
               RowBox[{"(", 
                RowBox[{
                 RowBox[{"-", "1"}], "+", "sm"}], ")"}], " ", 
               RowBox[{"(", 
                RowBox[{"1", "+", "\[Delta]"}], ")"}]}], "-", 
              RowBox[{"2", " ", "h", " ", 
               RowBox[{"(", 
                RowBox[{
                 RowBox[{"-", "2"}], "+", "sm"}], ")"}], " ", 
               RowBox[{"(", 
                RowBox[{
                 RowBox[{"-", "2"}], "+", "sm", "-", "\[Delta]", "+", 
                 RowBox[{"sm", " ", "\[Delta]"}]}], ")"}]}], "+", 
              RowBox[{"b", " ", "h", " ", 
               SuperscriptBox[
                RowBox[{"(", 
                 RowBox[{
                  RowBox[{"-", "2"}], "+", "sm", "-", "\[Delta]", "+", 
                  RowBox[{"sm", " ", "\[Delta]"}]}], ")"}], "2"]}]}], 
             ")"}]}]}], ")"}]}], ")"}], "/", 
       RowBox[{"(", 
        RowBox[{"b", " ", "\[Alpha]", " ", 
         RowBox[{"(", 
          RowBox[{
           RowBox[{
            RowBox[{"-", "sm"}], " ", 
            RowBox[{"(", 
             RowBox[{
              RowBox[{"sm", " ", 
               RowBox[{"(", 
                RowBox[{
                 RowBox[{"-", "1"}], "+", "\[Delta]"}], ")"}]}], "-", 
              "\[Delta]"}], ")"}], " ", 
            SuperscriptBox[
             RowBox[{"(", 
              RowBox[{"sm", "-", "\[Delta]", "+", 
               RowBox[{"sm", " ", "\[Delta]"}]}], ")"}], "2"]}], "+", 
           RowBox[{
            SuperscriptBox["h", "2"], " ", 
            SuperscriptBox["sf", "3"], " ", 
            SuperscriptBox[
             RowBox[{"(", 
              RowBox[{
               RowBox[{"-", "2"}], "+", "sm", "-", "\[Delta]", "+", 
               RowBox[{"sm", " ", "\[Delta]"}]}], ")"}], "2"], " ", 
            RowBox[{"(", 
             RowBox[{
              RowBox[{
               RowBox[{"-", 
                RowBox[{"(", 
                 RowBox[{
                  RowBox[{"-", "1"}], "+", "sm"}], ")"}]}], " ", 
               RowBox[{"(", 
                RowBox[{"1", "+", "\[Delta]"}], ")"}]}], "+", 
              RowBox[{"h", " ", 
               RowBox[{"(", 
                RowBox[{
                 RowBox[{"-", "2"}], "+", "sm", "-", "\[Delta]", "+", 
                 RowBox[{"sm", " ", "\[Delta]"}]}], ")"}]}]}], ")"}]}], "+", 
           RowBox[{
            SuperscriptBox["sf", "2"], " ", 
            RowBox[{"(", 
             RowBox[{
              RowBox[{"8", " ", 
               SuperscriptBox[
                RowBox[{"(", 
                 RowBox[{
                  RowBox[{"-", "1"}], "+", "sm"}], ")"}], "2"], " ", 
               SuperscriptBox[
                RowBox[{"(", 
                 RowBox[{"1", "+", "\[Delta]"}], ")"}], "2"]}], "+", 
              RowBox[{"2", " ", "h", " ", 
               RowBox[{"(", 
                RowBox[{"8", "-", 
                 RowBox[{"11", " ", "sm"}], "+", 
                 RowBox[{"3", " ", 
                  SuperscriptBox["sm", "2"]}]}], ")"}], " ", 
               RowBox[{"(", 
                RowBox[{"1", "+", "\[Delta]"}], ")"}], " ", 
               RowBox[{"(", 
                RowBox[{
                 RowBox[{"-", "2"}], "+", "sm", "-", "\[Delta]", "+", 
                 RowBox[{"sm", " ", "\[Delta]"}]}], ")"}]}], "-", 
              RowBox[{
               SuperscriptBox["h", "2"], " ", 
               RowBox[{"(", 
                RowBox[{
                 RowBox[{"-", "8"}], "-", 
                 RowBox[{"sm", " ", 
                  RowBox[{"(", 
                   RowBox[{
                    RowBox[{"-", "6"}], "+", "\[Delta]"}], ")"}]}], "+", 
                 RowBox[{
                  SuperscriptBox["sm", "2"], " ", 
                  RowBox[{"(", 
                   RowBox[{
                    RowBox[{"-", "1"}], "+", "\[Delta]"}], ")"}]}]}], ")"}], 
               " ", 
               SuperscriptBox[
                RowBox[{"(", 
                 RowBox[{
                  RowBox[{"-", "2"}], "+", "sm", "-", "\[Delta]", "+", 
                  RowBox[{"sm", " ", "\[Delta]"}]}], ")"}], "2"]}]}], ")"}]}],
            "+", 
           RowBox[{"sf", " ", 
            RowBox[{"(", 
             RowBox[{"sm", "-", "\[Delta]", "+", 
              RowBox[{"sm", " ", "\[Delta]"}]}], ")"}], " ", 
            RowBox[{"(", 
             RowBox[{
              RowBox[{"2", " ", "h", " ", 
               SuperscriptBox["sm", "3"], " ", 
               RowBox[{"(", 
                RowBox[{
                 RowBox[{"-", "1"}], "+", 
                 SuperscriptBox["\[Delta]", "2"]}], ")"}]}], "+", 
              RowBox[{"\[Delta]", " ", 
               RowBox[{"(", 
                RowBox[{"1", "+", "\[Delta]", "-", 
                 RowBox[{"h", " ", 
                  RowBox[{"(", 
                   RowBox[{"2", "+", "\[Delta]"}], ")"}]}]}], ")"}]}], "+", 
              RowBox[{
               SuperscriptBox["sm", "2"], " ", 
               RowBox[{"(", 
                RowBox[{
                 RowBox[{"-", "5"}], "-", 
                 RowBox[{"4", " ", "\[Delta]"}], "+", 
                 SuperscriptBox["\[Delta]", "2"], "+", 
                 RowBox[{"h", " ", 
                  RowBox[{"(", 
                   RowBox[{"9", "-", 
                    RowBox[{"5", " ", 
                    SuperscriptBox["\[Delta]", "2"]}]}], ")"}]}]}], ")"}]}], 
              "+", 
              RowBox[{"sm", " ", 
               RowBox[{"(", 
                RowBox[{"5", "+", 
                 RowBox[{"3", " ", "\[Delta]"}], "-", 
                 RowBox[{"2", " ", 
                  SuperscriptBox["\[Delta]", "2"]}], "+", 
                 RowBox[{"2", " ", "h", " ", 
                  RowBox[{"(", 
                   RowBox[{
                    RowBox[{"-", "5"}], "+", "\[Delta]", "+", 
                    RowBox[{"2", " ", 
                    SuperscriptBox["\[Delta]", "2"]}]}], ")"}]}]}], ")"}]}]}],
              ")"}]}]}], ")"}]}], ")"}]}], "-", 
      RowBox[{
       RowBox[{"(", 
        RowBox[{
         RowBox[{"(", 
          RowBox[{
           RowBox[{"b", " ", 
            SuperscriptBox["h", "2"], " ", 
            SuperscriptBox["sf", "2"], " ", 
            SuperscriptBox[
             RowBox[{"(", 
              RowBox[{
               RowBox[{"-", "2"}], "+", "sm", "-", "\[Delta]", "+", 
               RowBox[{"sm", " ", "\[Delta]"}]}], ")"}], "2"]}], "+", 
           RowBox[{
            RowBox[{"(", 
             RowBox[{"sm", "-", "\[Delta]", "+", 
              RowBox[{"sm", " ", "\[Delta]"}]}], ")"}], " ", 
            RowBox[{"(", 
             RowBox[{
              RowBox[{
               RowBox[{"-", "b"}], " ", "\[Delta]"}], "+", 
              RowBox[{"sm", " ", 
               RowBox[{"(", 
                RowBox[{
                 RowBox[{"-", "4"}], "+", "b", "+", 
                 RowBox[{"b", " ", "\[Delta]"}]}], ")"}]}]}], ")"}]}], "-", 
           RowBox[{"2", " ", "sf", " ", 
            RowBox[{"(", 
             RowBox[{
              RowBox[{
               RowBox[{"-", "4"}], " ", 
               RowBox[{"(", 
                RowBox[{
                 RowBox[{"-", "1"}], "+", "sm"}], ")"}], " ", 
               RowBox[{"(", 
                RowBox[{"1", "+", "\[Delta]"}], ")"}]}], "+", 
              RowBox[{"2", " ", "b", " ", 
               RowBox[{"(", 
                RowBox[{
                 RowBox[{"-", "1"}], "+", "sm"}], ")"}], " ", 
               RowBox[{"(", 
                RowBox[{"1", "+", "\[Delta]"}], ")"}]}], "-", 
              RowBox[{"2", " ", "h", " ", 
               RowBox[{"(", 
                RowBox[{
                 RowBox[{"-", "2"}], "+", "sm"}], ")"}], " ", 
               RowBox[{"(", 
                RowBox[{
                 RowBox[{"-", "2"}], "+", "sm", "-", "\[Delta]", "+", 
                 RowBox[{"sm", " ", "\[Delta]"}]}], ")"}]}], "+", 
              RowBox[{"b", " ", "h", " ", 
               SuperscriptBox[
                RowBox[{"(", 
                 RowBox[{
                  RowBox[{"-", "2"}], "+", "sm", "-", "\[Delta]", "+", 
                  RowBox[{"sm", " ", "\[Delta]"}]}], ")"}], "2"]}]}], 
             ")"}]}]}], ")"}], " ", 
         RowBox[{"(", 
          RowBox[{
           RowBox[{"sm", " ", 
            RowBox[{"(", 
             RowBox[{
              RowBox[{"sm", " ", 
               RowBox[{"(", 
                RowBox[{
                 RowBox[{"-", "1"}], "+", "\[Delta]"}], ")"}]}], "-", 
              "\[Delta]"}], ")"}], " ", 
            SuperscriptBox[
             RowBox[{"(", 
              RowBox[{"sm", "-", "\[Delta]", "+", 
               RowBox[{"sm", " ", "\[Delta]"}]}], ")"}], "2"]}], "+", 
           RowBox[{
            SuperscriptBox["sf", "2"], " ", 
            RowBox[{"(", 
             RowBox[{
              RowBox[{
               RowBox[{"-", "4"}], " ", 
               SuperscriptBox[
                RowBox[{"(", 
                 RowBox[{
                  RowBox[{"-", "1"}], "+", "sm"}], ")"}], "2"], " ", 
               SuperscriptBox[
                RowBox[{"(", 
                 RowBox[{"1", "+", "\[Delta]"}], ")"}], "2"]}], "+", 
              RowBox[{
               SuperscriptBox["h", "2"], " ", 
               RowBox[{"(", 
                RowBox[{
                 RowBox[{"-", "2"}], "+", "sm"}], ")"}], " ", 
               RowBox[{"(", 
                RowBox[{"2", "+", 
                 RowBox[{"sm", " ", 
                  RowBox[{"(", 
                   RowBox[{
                    RowBox[{"-", "1"}], "+", "\[Delta]"}], ")"}]}], "-", 
                 "\[Delta]"}], ")"}], " ", 
               SuperscriptBox[
                RowBox[{"(", 
                 RowBox[{
                  RowBox[{"-", "2"}], "+", "sm", "-", "\[Delta]", "+", 
                  RowBox[{"sm", " ", "\[Delta]"}]}], ")"}], "2"]}], "+", 
              RowBox[{"2", " ", "h", " ", 
               RowBox[{"(", 
                RowBox[{
                 RowBox[{"-", "1"}], "+", "sm"}], ")"}], " ", 
               RowBox[{"(", 
                RowBox[{"1", "+", "\[Delta]"}], ")"}], " ", 
               RowBox[{"(", 
                RowBox[{
                 RowBox[{"-", "8"}], "-", 
                 RowBox[{"2", " ", "\[Delta]"}], "+", 
                 SuperscriptBox["\[Delta]", "2"], "+", 
                 RowBox[{"sm", " ", 
                  RowBox[{"(", 
                   RowBox[{"8", "+", 
                    RowBox[{"3", " ", "\[Delta]"}], "-", 
                    RowBox[{"2", " ", 
                    SuperscriptBox["\[Delta]", "2"]}]}], ")"}]}], "+", 
                 RowBox[{
                  SuperscriptBox["sm", "2"], " ", 
                  RowBox[{"(", 
                   RowBox[{
                    RowBox[{"-", "2"}], "-", "\[Delta]", "+", 
                    SuperscriptBox["\[Delta]", "2"]}], ")"}]}]}], ")"}]}]}], 
             ")"}]}], "-", 
           RowBox[{"2", " ", "sf", " ", 
            RowBox[{"(", 
             RowBox[{"sm", "-", "\[Delta]", "+", 
              RowBox[{"sm", " ", "\[Delta]"}]}], ")"}], " ", 
            RowBox[{"(", 
             RowBox[{
              RowBox[{"h", " ", 
               SuperscriptBox["sm", "3"], " ", 
               RowBox[{"(", 
                RowBox[{
                 RowBox[{"-", "1"}], "+", 
                 SuperscriptBox["\[Delta]", "2"]}], ")"}]}], "+", 
              RowBox[{"\[Delta]", " ", 
               RowBox[{"(", 
                RowBox[{"1", "+", "\[Delta]", "-", 
                 RowBox[{"h", " ", 
                  RowBox[{"(", 
                   RowBox[{"2", "+", "\[Delta]"}], ")"}]}]}], ")"}]}], "-", 
              RowBox[{
               SuperscriptBox["sm", "2"], " ", 
               RowBox[{"(", 
                RowBox[{"2", "+", "\[Delta]", "-", 
                 SuperscriptBox["\[Delta]", "2"], "+", 
                 RowBox[{"h", " ", 
                  RowBox[{"(", 
                   RowBox[{
                    RowBox[{"-", "4"}], "+", "\[Delta]", "+", 
                    RowBox[{"3", " ", 
                    SuperscriptBox["\[Delta]", "2"]}]}], ")"}]}]}], ")"}]}], 
              "+", 
              RowBox[{"sm", " ", 
               RowBox[{"(", 
                RowBox[{"2", "-", 
                 RowBox[{"2", " ", 
                  SuperscriptBox["\[Delta]", "2"]}], "+", 
                 RowBox[{"h", " ", 
                  RowBox[{"(", 
                   RowBox[{
                    RowBox[{"-", "4"}], "+", 
                    RowBox[{"3", " ", "\[Delta]"}], "+", 
                    RowBox[{"3", " ", 
                    SuperscriptBox["\[Delta]", "2"]}]}], ")"}]}]}], ")"}]}]}],
              ")"}]}]}], ")"}]}], ")"}], "/", 
       RowBox[{"(", 
        RowBox[{"b", " ", "\[Alpha]", " ", 
         RowBox[{"(", 
          RowBox[{
           RowBox[{
            RowBox[{"-", "sm"}], " ", 
            RowBox[{"(", 
             RowBox[{
              RowBox[{"sm", " ", 
               RowBox[{"(", 
                RowBox[{
                 RowBox[{"-", "1"}], "+", "\[Delta]"}], ")"}]}], "-", 
              "\[Delta]"}], ")"}], " ", 
            SuperscriptBox[
             RowBox[{"(", 
              RowBox[{"sm", "-", "\[Delta]", "+", 
               RowBox[{"sm", " ", "\[Delta]"}]}], ")"}], "4"]}], "+", 
           RowBox[{
            SuperscriptBox["h", "4"], " ", 
            SuperscriptBox["sf", "5"], " ", 
            SuperscriptBox[
             RowBox[{"(", 
              RowBox[{
               RowBox[{"-", "2"}], "+", "sm", "-", "\[Delta]", "+", 
               RowBox[{"sm", " ", "\[Delta]"}]}], ")"}], "4"], " ", 
            RowBox[{"(", 
             RowBox[{
              RowBox[{
               RowBox[{"-", 
                RowBox[{"(", 
                 RowBox[{
                  RowBox[{"-", "1"}], "+", "sm"}], ")"}]}], " ", 
               RowBox[{"(", 
                RowBox[{"1", "+", "\[Delta]"}], ")"}]}], "+", 
              RowBox[{"h", " ", 
               RowBox[{"(", 
                RowBox[{
                 RowBox[{"-", "2"}], "+", "sm", "-", "\[Delta]", "+", 
                 RowBox[{"sm", " ", "\[Delta]"}]}], ")"}]}]}], ")"}]}], "-", 
           RowBox[{
            SuperscriptBox["h", "2"], " ", 
            SuperscriptBox["sf", "4"], " ", 
            SuperscriptBox[
             RowBox[{"(", 
              RowBox[{
               RowBox[{"-", "2"}], "+", "sm", "-", "\[Delta]", "+", 
               RowBox[{"sm", " ", "\[Delta]"}]}], ")"}], "2"], " ", 
            RowBox[{"(", 
             RowBox[{
              RowBox[{
               RowBox[{"-", "12"}], " ", 
               SuperscriptBox[
                RowBox[{"(", 
                 RowBox[{
                  RowBox[{"-", "1"}], "+", "sm"}], ")"}], "2"], " ", 
               SuperscriptBox[
                RowBox[{"(", 
                 RowBox[{"1", "+", "\[Delta]"}], ")"}], "2"]}], "+", 
              RowBox[{
               SuperscriptBox["h", "2"], " ", 
               SuperscriptBox[
                RowBox[{"(", 
                 RowBox[{
                  RowBox[{"-", "2"}], "+", "sm", "-", "\[Delta]", "+", 
                  RowBox[{"sm", " ", "\[Delta]"}]}], ")"}], "2"], " ", 
               RowBox[{"(", 
                RowBox[{
                 RowBox[{
                  SuperscriptBox["sm", "2"], " ", 
                  RowBox[{"(", 
                   RowBox[{
                    RowBox[{"-", "1"}], "+", "\[Delta]"}], ")"}]}], "-", 
                 RowBox[{"2", " ", 
                  RowBox[{"(", 
                   RowBox[{"6", "+", "\[Delta]"}], ")"}]}], "+", 
                 RowBox[{"sm", " ", 
                  RowBox[{"(", 
                   RowBox[{"8", "+", "\[Delta]"}], ")"}]}]}], ")"}]}], "-", 
              RowBox[{"2", " ", "h", " ", 
               RowBox[{"(", 
                RowBox[{
                 RowBox[{"-", "1"}], "+", "sm"}], ")"}], " ", 
               RowBox[{"(", 
                RowBox[{"1", "+", "\[Delta]"}], ")"}], " ", 
               RowBox[{"(", 
                RowBox[{"24", "+", 
                 RowBox[{"14", " ", "\[Delta]"}], "+", 
                 SuperscriptBox["\[Delta]", "2"], "+", 
                 RowBox[{
                  SuperscriptBox["sm", "2"], " ", 
                  RowBox[{"(", 
                   RowBox[{"4", "+", 
                    RowBox[{"5", " ", "\[Delta]"}], "+", 
                    SuperscriptBox["\[Delta]", "2"]}], ")"}]}], "-", 
                 RowBox[{"sm", " ", 
                  RowBox[{"(", 
                   RowBox[{"20", "+", 
                    RowBox[{"19", " ", "\[Delta]"}], "+", 
                    RowBox[{"2", " ", 
                    SuperscriptBox["\[Delta]", "2"]}]}], ")"}]}]}], ")"}]}]}],
              ")"}]}], "+", 
           RowBox[{"2", " ", 
            SuperscriptBox["sf", "3"], " ", 
            RowBox[{"(", 
             RowBox[{
              RowBox[{
               RowBox[{"-", "16"}], " ", 
               SuperscriptBox[
                RowBox[{"(", 
                 RowBox[{
                  RowBox[{"-", "1"}], "+", "sm"}], ")"}], "3"], " ", 
               SuperscriptBox[
                RowBox[{"(", 
                 RowBox[{"1", "+", "\[Delta]"}], ")"}], "3"]}], "-", 
              RowBox[{
               SuperscriptBox["h", "2"], " ", 
               RowBox[{"(", 
                RowBox[{
                 RowBox[{"-", "1"}], "+", "sm"}], ")"}], " ", 
               RowBox[{"(", 
                RowBox[{"1", "+", "\[Delta]"}], ")"}], " ", 
               SuperscriptBox[
                RowBox[{"(", 
                 RowBox[{
                  RowBox[{"-", "2"}], "+", "sm", "-", "\[Delta]", "+", 
                  RowBox[{"sm", " ", "\[Delta]"}]}], ")"}], "2"], " ", 
               RowBox[{"(", 
                RowBox[{
                 RowBox[{"16", " ", 
                  RowBox[{"(", 
                   RowBox[{"3", "+", "\[Delta]"}], ")"}]}], "+", 
                 RowBox[{
                  SuperscriptBox["sm", "2"], " ", 
                  RowBox[{"(", 
                   RowBox[{"11", "+", 
                    RowBox[{"7", " ", "\[Delta]"}]}], ")"}]}], "-", 
                 RowBox[{"sm", " ", 
                  RowBox[{"(", 
                   RowBox[{"40", "+", 
                    RowBox[{"23", " ", "\[Delta]"}]}], ")"}]}]}], ")"}]}], 
              "+", 
              RowBox[{
               SuperscriptBox["h", "3"], " ", 
               SuperscriptBox[
                RowBox[{"(", 
                 RowBox[{
                  RowBox[{"-", "2"}], "+", "sm", "-", "\[Delta]", "+", 
                  RowBox[{"sm", " ", "\[Delta]"}]}], ")"}], "3"], " ", 
               RowBox[{"(", 
                RowBox[{
                 RowBox[{"8", " ", 
                  RowBox[{"(", 
                   RowBox[{"2", "+", "\[Delta]"}], ")"}]}], "+", 
                 RowBox[{
                  SuperscriptBox["sm", "2"], " ", 
                  RowBox[{"(", 
                   RowBox[{"11", "+", 
                    RowBox[{"7", " ", "\[Delta]"}], "-", 
                    RowBox[{"4", " ", 
                    SuperscriptBox["\[Delta]", "2"]}]}], ")"}]}], "+", 
                 RowBox[{"2", " ", 
                  SuperscriptBox["sm", "3"], " ", 
                  RowBox[{"(", 
                   RowBox[{
                    RowBox[{"-", "1"}], "+", 
                    SuperscriptBox["\[Delta]", "2"]}], ")"}]}], "+", 
                 RowBox[{"sm", " ", 
                  RowBox[{"(", 
                   RowBox[{
                    RowBox[{"-", "20"}], "-", 
                    RowBox[{"15", " ", "\[Delta]"}], "+", 
                    RowBox[{"2", " ", 
                    SuperscriptBox["\[Delta]", "2"]}]}], ")"}]}]}], ")"}]}], 
              "-", 
              RowBox[{"4", " ", "h", " ", 
               SuperscriptBox[
                RowBox[{"(", 
                 RowBox[{
                  RowBox[{"-", "1"}], "+", "sm"}], ")"}], "2"], " ", 
               SuperscriptBox[
                RowBox[{"(", 
                 RowBox[{"1", "+", "\[Delta]"}], ")"}], "2"], " ", 
               RowBox[{"(", 
                RowBox[{
                 RowBox[{"2", " ", 
                  RowBox[{"(", 
                   RowBox[{"12", "+", 
                    RowBox[{"8", " ", "\[Delta]"}], "+", 
                    SuperscriptBox["\[Delta]", "2"]}], ")"}]}], "+", 
                 RowBox[{
                  SuperscriptBox["sm", "2"], " ", 
                  RowBox[{"(", 
                   RowBox[{"5", "+", 
                    RowBox[{"7", " ", "\[Delta]"}], "+", 
                    RowBox[{"2", " ", 
                    SuperscriptBox["\[Delta]", "2"]}]}], ")"}]}], "-", 
                 RowBox[{"sm", " ", 
                  RowBox[{"(", 
                   RowBox[{"22", "+", 
                    RowBox[{"23", " ", "\[Delta]"}], "+", 
                    RowBox[{"4", " ", 
                    SuperscriptBox["\[Delta]", "2"]}]}], ")"}]}]}], ")"}]}]}],
              ")"}]}], "+", 
           RowBox[{"sf", " ", 
            SuperscriptBox[
             RowBox[{"(", 
              RowBox[{"sm", "-", "\[Delta]", "+", 
               RowBox[{"sm", " ", "\[Delta]"}]}], ")"}], "2"], " ", 
            RowBox[{"(", 
             RowBox[{
              RowBox[{"4", " ", "h", " ", 
               SuperscriptBox["sm", "4"], " ", 
               RowBox[{"(", 
                RowBox[{
                 RowBox[{"-", "1"}], "+", "\[Delta]"}], ")"}], " ", 
               SuperscriptBox[
                RowBox[{"(", 
                 RowBox[{"1", "+", "\[Delta]"}], ")"}], "2"]}], "+", 
              RowBox[{"sm", " ", 
               SuperscriptBox["\[Delta]", "2"], " ", 
               RowBox[{"(", 
                RowBox[{"3", "-", 
                 RowBox[{"13", " ", "h"}], "+", 
                 RowBox[{"3", " ", "\[Delta]"}], "-", 
                 RowBox[{"7", " ", "h", " ", "\[Delta]"}]}], ")"}]}], "+", 
              RowBox[{
               SuperscriptBox["\[Delta]", "2"], " ", 
               RowBox[{"(", 
                RowBox[{
                 RowBox[{"-", "1"}], "-", "\[Delta]", "+", 
                 RowBox[{"h", " ", 
                  RowBox[{"(", 
                   RowBox[{"2", "+", "\[Delta]"}], ")"}]}]}], ")"}]}], "+", 
              RowBox[{"3", " ", 
               SuperscriptBox["sm", "2"], " ", 
               RowBox[{"(", 
                RowBox[{"1", "+", "\[Delta]"}], ")"}], " ", 
               RowBox[{"(", 
                RowBox[{"3", "-", 
                 SuperscriptBox["\[Delta]", "2"], "+", 
                 RowBox[{"h", " ", 
                  RowBox[{"(", 
                   RowBox[{
                    RowBox[{"-", "6"}], "+", 
                    RowBox[{"3", " ", "\[Delta]"}], "+", 
                    RowBox[{"5", " ", 
                    SuperscriptBox["\[Delta]", "2"]}]}], ")"}]}]}], ")"}]}], 
              "-", 
              RowBox[{
               SuperscriptBox["sm", "3"], " ", 
               RowBox[{"(", 
                RowBox[{"1", "+", "\[Delta]"}], ")"}], " ", 
               RowBox[{"(", 
                RowBox[{"9", "-", 
                 SuperscriptBox["\[Delta]", "2"], "+", 
                 RowBox[{"h", " ", 
                  RowBox[{"(", 
                   RowBox[{
                    RowBox[{"-", "17"}], "+", 
                    RowBox[{"4", " ", "\[Delta]"}], "+", 
                    RowBox[{"13", " ", 
                    SuperscriptBox["\[Delta]", "2"]}]}], ")"}]}]}], ")"}]}]}],
              ")"}]}], "-", 
           RowBox[{"2", " ", 
            SuperscriptBox["sf", "2"], " ", 
            RowBox[{"(", 
             RowBox[{"sm", "-", "\[Delta]", "+", 
              RowBox[{"sm", " ", "\[Delta]"}]}], ")"}], " ", 
            RowBox[{"(", 
             RowBox[{
              RowBox[{"3", " ", 
               SuperscriptBox["h", "2"], " ", 
               SuperscriptBox["sm", "5"], " ", 
               RowBox[{"(", 
                RowBox[{
                 RowBox[{"-", "1"}], "+", "\[Delta]"}], ")"}], " ", 
               SuperscriptBox[
                RowBox[{"(", 
                 RowBox[{"1", "+", "\[Delta]"}], ")"}], "3"]}], "-", 
              RowBox[{"h", " ", 
               SuperscriptBox["sm", "4"], " ", 
               SuperscriptBox[
                RowBox[{"(", 
                 RowBox[{"1", "+", "\[Delta]"}], ")"}], "2"], " ", 
               RowBox[{"(", 
                RowBox[{"12", "+", 
                 RowBox[{"3", " ", "\[Delta]"}], "-", 
                 SuperscriptBox["\[Delta]", "2"], "+", 
                 RowBox[{"h", " ", 
                  RowBox[{"(", 
                   RowBox[{
                    RowBox[{"-", "24"}], "+", 
                    RowBox[{"3", " ", "\[Delta]"}], "+", 
                    RowBox[{"13", " ", 
                    SuperscriptBox["\[Delta]", "2"]}]}], ")"}]}]}], ")"}]}], 
              "+", 
              RowBox[{"\[Delta]", " ", 
               RowBox[{"(", 
                RowBox[{
                 RowBox[{"2", " ", 
                  SuperscriptBox[
                   RowBox[{"(", 
                    RowBox[{"1", "+", "\[Delta]"}], ")"}], "2"]}], "-", 
                 RowBox[{
                  SuperscriptBox["h", "2"], " ", 
                  RowBox[{"(", 
                   RowBox[{
                    RowBox[{"-", "2"}], "+", "\[Delta]"}], ")"}], " ", 
                  SuperscriptBox[
                   RowBox[{"(", 
                    RowBox[{"2", "+", "\[Delta]"}], ")"}], "2"]}], "+", 
                 RowBox[{"h", " ", 
                  RowBox[{"(", 
                   RowBox[{
                    RowBox[{"-", "8"}], "-", 
                    RowBox[{"10", " ", "\[Delta]"}], "-", 
                    SuperscriptBox["\[Delta]", "2"], "+", 
                    SuperscriptBox["\[Delta]", "3"]}], ")"}]}]}], ")"}]}], 
              "+", 
              RowBox[{
               SuperscriptBox["sm", "3"], " ", 
               RowBox[{"(", 
                RowBox[{"1", "+", "\[Delta]"}], ")"}], " ", 
               RowBox[{"(", 
                RowBox[{
                 RowBox[{
                  RowBox[{"-", "2"}], " ", 
                  RowBox[{"(", 
                   RowBox[{"7", "+", 
                    RowBox[{"8", " ", "\[Delta]"}], "+", 
                    SuperscriptBox["\[Delta]", "2"]}], ")"}]}], "+", 
                 RowBox[{"h", " ", 
                  RowBox[{"(", 
                   RowBox[{"64", "+", 
                    RowBox[{"68", " ", "\[Delta]"}], "+", 
                    RowBox[{"8", " ", 
                    SuperscriptBox["\[Delta]", "2"]}], "-", 
                    RowBox[{"4", " ", 
                    SuperscriptBox["\[Delta]", "3"]}]}], ")"}]}], "+", 
                 RowBox[{
                  SuperscriptBox["h", "2"], " ", 
                  RowBox[{"(", 
                   RowBox[{
                    RowBox[{"-", "74"}], "-", 
                    RowBox[{"55", " ", "\[Delta]"}], "+", 
                    RowBox[{"31", " ", 
                    SuperscriptBox["\[Delta]", "2"]}], "+", 
                    RowBox[{"22", " ", 
                    SuperscriptBox["\[Delta]", "3"]}]}], ")"}]}]}], ")"}]}], 
              "+", 
              RowBox[{
               SuperscriptBox["sm", "2"], " ", 
               RowBox[{"(", 
                RowBox[{
                 RowBox[{"2", " ", 
                  SuperscriptBox[
                   RowBox[{"(", 
                    RowBox[{"1", "+", "\[Delta]"}], ")"}], "2"], " ", 
                  RowBox[{"(", 
                   RowBox[{"14", "+", 
                    RowBox[{"3", " ", "\[Delta]"}]}], ")"}]}], "+", 
                 RowBox[{
                  SuperscriptBox["h", "2"], " ", 
                  RowBox[{"(", 
                   RowBox[{"104", "+", 
                    RowBox[{"166", " ", "\[Delta]"}], "+", 
                    RowBox[{"37", " ", 
                    SuperscriptBox["\[Delta]", "2"]}], "-", 
                    RowBox[{"45", " ", 
                    SuperscriptBox["\[Delta]", "3"]}], "-", 
                    RowBox[{"18", " ", 
                    SuperscriptBox["\[Delta]", "4"]}]}], ")"}]}], "+", 
                 RowBox[{"3", " ", "h", " ", 
                  RowBox[{"(", 
                   RowBox[{
                    RowBox[{"-", "36"}], "-", 
                    RowBox[{"69", " ", "\[Delta]"}], "-", 
                    RowBox[{"37", " ", 
                    SuperscriptBox["\[Delta]", "2"]}], "-", 
                    RowBox[{"2", " ", 
                    SuperscriptBox["\[Delta]", "3"]}], "+", 
                    RowBox[{"2", " ", 
                    SuperscriptBox["\[Delta]", "4"]}]}], ")"}]}]}], ")"}]}], 
              "+", 
              RowBox[{"sm", " ", 
               RowBox[{"(", 
                RowBox[{
                 RowBox[{
                  RowBox[{"-", "2"}], " ", 
                  SuperscriptBox[
                   RowBox[{"(", 
                    RowBox[{"1", "+", "\[Delta]"}], ")"}], "2"], " ", 
                  RowBox[{"(", 
                   RowBox[{"7", "+", 
                    RowBox[{"3", " ", "\[Delta]"}]}], ")"}]}], "+", 
                 RowBox[{"h", " ", 
                  RowBox[{"(", 
                   RowBox[{"56", "+", 
                    RowBox[{"110", " ", "\[Delta]"}], "+", 
                    RowBox[{"62", " ", 
                    SuperscriptBox["\[Delta]", "2"]}], "+", 
                    RowBox[{"4", " ", 
                    SuperscriptBox["\[Delta]", "3"]}], "-", 
                    RowBox[{"4", " ", 
                    SuperscriptBox["\[Delta]", "4"]}]}], ")"}]}], "+", 
                 RowBox[{
                  SuperscriptBox["h", "2"], " ", 
                  RowBox[{"(", 
                   RowBox[{
                    RowBox[{"-", "56"}], "-", 
                    RowBox[{"84", " ", "\[Delta]"}], "-", 
                    RowBox[{"22", " ", 
                    SuperscriptBox["\[Delta]", "2"]}], "+", 
                    RowBox[{"17", " ", 
                    SuperscriptBox["\[Delta]", "3"]}], "+", 
                    RowBox[{"7", " ", 
                    SuperscriptBox["\[Delta]", "4"]}]}], ")"}]}]}], ")"}]}]}],
              ")"}]}]}], ")"}]}], ")"}]}]}], "<", "0"}], ",", "Null", ",", 
    RowBox[{
     RowBox[{
      RowBox[{"(", 
       RowBox[{"sf", " ", 
        RowBox[{"(", 
         RowBox[{
          RowBox[{
           RowBox[{"-", 
            RowBox[{"(", 
             RowBox[{
              RowBox[{"-", "1"}], "+", "sm"}], ")"}]}], " ", 
           RowBox[{"(", 
            RowBox[{"1", "+", "\[Delta]"}], ")"}]}], "+", 
          RowBox[{"h", " ", 
           RowBox[{"(", 
            RowBox[{
             RowBox[{"-", "2"}], "+", "sm", "-", "\[Delta]", "+", 
             RowBox[{"sm", " ", "\[Delta]"}]}], ")"}]}]}], ")"}], " ", 
        RowBox[{"(", 
         RowBox[{
          RowBox[{"b", " ", 
           SuperscriptBox["h", "2"], " ", 
           SuperscriptBox["sf", "2"], " ", 
           SuperscriptBox[
            RowBox[{"(", 
             RowBox[{
              RowBox[{"-", "2"}], "+", "sm", "-", "\[Delta]", "+", 
              RowBox[{"sm", " ", "\[Delta]"}]}], ")"}], "2"]}], "+", 
          RowBox[{
           RowBox[{"(", 
            RowBox[{"sm", "-", "\[Delta]", "+", 
             RowBox[{"sm", " ", "\[Delta]"}]}], ")"}], " ", 
           RowBox[{"(", 
            RowBox[{
             RowBox[{
              RowBox[{"-", "b"}], " ", "\[Delta]"}], "+", 
             RowBox[{"sm", " ", 
              RowBox[{"(", 
               RowBox[{
                RowBox[{"-", "4"}], "+", "b", "+", 
                RowBox[{"b", " ", "\[Delta]"}]}], ")"}]}]}], ")"}]}], "-", 
          RowBox[{"2", " ", "sf", " ", 
           RowBox[{"(", 
            RowBox[{
             RowBox[{
              RowBox[{"-", "4"}], " ", 
              RowBox[{"(", 
               RowBox[{
                RowBox[{"-", "1"}], "+", "sm"}], ")"}], " ", 
              RowBox[{"(", 
               RowBox[{"1", "+", "\[Delta]"}], ")"}]}], "+", 
             RowBox[{"2", " ", "b", " ", 
              RowBox[{"(", 
               RowBox[{
                RowBox[{"-", "1"}], "+", "sm"}], ")"}], " ", 
              RowBox[{"(", 
               RowBox[{"1", "+", "\[Delta]"}], ")"}]}], "-", 
             RowBox[{"2", " ", "h", " ", 
              RowBox[{"(", 
               RowBox[{
                RowBox[{"-", "2"}], "+", "sm"}], ")"}], " ", 
              RowBox[{"(", 
               RowBox[{
                RowBox[{"-", "2"}], "+", "sm", "-", "\[Delta]", "+", 
                RowBox[{"sm", " ", "\[Delta]"}]}], ")"}]}], "+", 
             RowBox[{"b", " ", "h", " ", 
              SuperscriptBox[
               RowBox[{"(", 
                RowBox[{
                 RowBox[{"-", "2"}], "+", "sm", "-", "\[Delta]", "+", 
                 RowBox[{"sm", " ", "\[Delta]"}]}], ")"}], "2"]}]}], 
            ")"}]}]}], ")"}]}], ")"}], "/", 
      RowBox[{"(", 
       RowBox[{"b", " ", "\[Alpha]", " ", 
        RowBox[{"(", 
         RowBox[{
          RowBox[{
           RowBox[{"-", "sm"}], " ", 
           RowBox[{"(", 
            RowBox[{
             RowBox[{"sm", " ", 
              RowBox[{"(", 
               RowBox[{
                RowBox[{"-", "1"}], "+", "\[Delta]"}], ")"}]}], "-", 
             "\[Delta]"}], ")"}], " ", 
           SuperscriptBox[
            RowBox[{"(", 
             RowBox[{"sm", "-", "\[Delta]", "+", 
              RowBox[{"sm", " ", "\[Delta]"}]}], ")"}], "2"]}], "+", 
          RowBox[{
           SuperscriptBox["h", "2"], " ", 
           SuperscriptBox["sf", "3"], " ", 
           SuperscriptBox[
            RowBox[{"(", 
             RowBox[{
              RowBox[{"-", "2"}], "+", "sm", "-", "\[Delta]", "+", 
              RowBox[{"sm", " ", "\[Delta]"}]}], ")"}], "2"], " ", 
           RowBox[{"(", 
            RowBox[{
             RowBox[{
              RowBox[{"-", 
               RowBox[{"(", 
                RowBox[{
                 RowBox[{"-", "1"}], "+", "sm"}], ")"}]}], " ", 
              RowBox[{"(", 
               RowBox[{"1", "+", "\[Delta]"}], ")"}]}], "+", 
             RowBox[{"h", " ", 
              RowBox[{"(", 
               RowBox[{
                RowBox[{"-", "2"}], "+", "sm", "-", "\[Delta]", "+", 
                RowBox[{"sm", " ", "\[Delta]"}]}], ")"}]}]}], ")"}]}], "+", 
          RowBox[{
           SuperscriptBox["sf", "2"], " ", 
           RowBox[{"(", 
            RowBox[{
             RowBox[{"8", " ", 
              SuperscriptBox[
               RowBox[{"(", 
                RowBox[{
                 RowBox[{"-", "1"}], "+", "sm"}], ")"}], "2"], " ", 
              SuperscriptBox[
               RowBox[{"(", 
                RowBox[{"1", "+", "\[Delta]"}], ")"}], "2"]}], "+", 
             RowBox[{"2", " ", "h", " ", 
              RowBox[{"(", 
               RowBox[{"8", "-", 
                RowBox[{"11", " ", "sm"}], "+", 
                RowBox[{"3", " ", 
                 SuperscriptBox["sm", "2"]}]}], ")"}], " ", 
              RowBox[{"(", 
               RowBox[{"1", "+", "\[Delta]"}], ")"}], " ", 
              RowBox[{"(", 
               RowBox[{
                RowBox[{"-", "2"}], "+", "sm", "-", "\[Delta]", "+", 
                RowBox[{"sm", " ", "\[Delta]"}]}], ")"}]}], "-", 
             RowBox[{
              SuperscriptBox["h", "2"], " ", 
              RowBox[{"(", 
               RowBox[{
                RowBox[{"-", "8"}], "-", 
                RowBox[{"sm", " ", 
                 RowBox[{"(", 
                  RowBox[{
                   RowBox[{"-", "6"}], "+", "\[Delta]"}], ")"}]}], "+", 
                RowBox[{
                 SuperscriptBox["sm", "2"], " ", 
                 RowBox[{"(", 
                  RowBox[{
                   RowBox[{"-", "1"}], "+", "\[Delta]"}], ")"}]}]}], ")"}], 
              " ", 
              SuperscriptBox[
               RowBox[{"(", 
                RowBox[{
                 RowBox[{"-", "2"}], "+", "sm", "-", "\[Delta]", "+", 
                 RowBox[{"sm", " ", "\[Delta]"}]}], ")"}], "2"]}]}], ")"}]}], 
          "+", 
          RowBox[{"sf", " ", 
           RowBox[{"(", 
            RowBox[{"sm", "-", "\[Delta]", "+", 
             RowBox[{"sm", " ", "\[Delta]"}]}], ")"}], " ", 
           RowBox[{"(", 
            RowBox[{
             RowBox[{"2", " ", "h", " ", 
              SuperscriptBox["sm", "3"], " ", 
              RowBox[{"(", 
               RowBox[{
                RowBox[{"-", "1"}], "+", 
                SuperscriptBox["\[Delta]", "2"]}], ")"}]}], "+", 
             RowBox[{"\[Delta]", " ", 
              RowBox[{"(", 
               RowBox[{"1", "+", "\[Delta]", "-", 
                RowBox[{"h", " ", 
                 RowBox[{"(", 
                  RowBox[{"2", "+", "\[Delta]"}], ")"}]}]}], ")"}]}], "+", 
             RowBox[{
              SuperscriptBox["sm", "2"], " ", 
              RowBox[{"(", 
               RowBox[{
                RowBox[{"-", "5"}], "-", 
                RowBox[{"4", " ", "\[Delta]"}], "+", 
                SuperscriptBox["\[Delta]", "2"], "+", 
                RowBox[{"h", " ", 
                 RowBox[{"(", 
                  RowBox[{"9", "-", 
                   RowBox[{"5", " ", 
                    SuperscriptBox["\[Delta]", "2"]}]}], ")"}]}]}], ")"}]}], 
             "+", 
             RowBox[{"sm", " ", 
              RowBox[{"(", 
               RowBox[{"5", "+", 
                RowBox[{"3", " ", "\[Delta]"}], "-", 
                RowBox[{"2", " ", 
                 SuperscriptBox["\[Delta]", "2"]}], "+", 
                RowBox[{"2", " ", "h", " ", 
                 RowBox[{"(", 
                  RowBox[{
                   RowBox[{"-", "5"}], "+", "\[Delta]", "+", 
                   RowBox[{"2", " ", 
                    SuperscriptBox["\[Delta]", "2"]}]}], ")"}]}]}], ")"}]}]}],
             ")"}]}]}], ")"}]}], ")"}]}], "-", 
     RowBox[{
      RowBox[{"(", 
       RowBox[{
        RowBox[{"(", 
         RowBox[{
          RowBox[{"b", " ", 
           SuperscriptBox["h", "2"], " ", 
           SuperscriptBox["sf", "2"], " ", 
           SuperscriptBox[
            RowBox[{"(", 
             RowBox[{
              RowBox[{"-", "2"}], "+", "sm", "-", "\[Delta]", "+", 
              RowBox[{"sm", " ", "\[Delta]"}]}], ")"}], "2"]}], "+", 
          RowBox[{
           RowBox[{"(", 
            RowBox[{"sm", "-", "\[Delta]", "+", 
             RowBox[{"sm", " ", "\[Delta]"}]}], ")"}], " ", 
           RowBox[{"(", 
            RowBox[{
             RowBox[{
              RowBox[{"-", "b"}], " ", "\[Delta]"}], "+", 
             RowBox[{"sm", " ", 
              RowBox[{"(", 
               RowBox[{
                RowBox[{"-", "4"}], "+", "b", "+", 
                RowBox[{"b", " ", "\[Delta]"}]}], ")"}]}]}], ")"}]}], "-", 
          RowBox[{"2", " ", "sf", " ", 
           RowBox[{"(", 
            RowBox[{
             RowBox[{
              RowBox[{"-", "4"}], " ", 
              RowBox[{"(", 
               RowBox[{
                RowBox[{"-", "1"}], "+", "sm"}], ")"}], " ", 
              RowBox[{"(", 
               RowBox[{"1", "+", "\[Delta]"}], ")"}]}], "+", 
             RowBox[{"2", " ", "b", " ", 
              RowBox[{"(", 
               RowBox[{
                RowBox[{"-", "1"}], "+", "sm"}], ")"}], " ", 
              RowBox[{"(", 
               RowBox[{"1", "+", "\[Delta]"}], ")"}]}], "-", 
             RowBox[{"2", " ", "h", " ", 
              RowBox[{"(", 
               RowBox[{
                RowBox[{"-", "2"}], "+", "sm"}], ")"}], " ", 
              RowBox[{"(", 
               RowBox[{
                RowBox[{"-", "2"}], "+", "sm", "-", "\[Delta]", "+", 
                RowBox[{"sm", " ", "\[Delta]"}]}], ")"}]}], "+", 
             RowBox[{"b", " ", "h", " ", 
              SuperscriptBox[
               RowBox[{"(", 
                RowBox[{
                 RowBox[{"-", "2"}], "+", "sm", "-", "\[Delta]", "+", 
                 RowBox[{"sm", " ", "\[Delta]"}]}], ")"}], "2"]}]}], 
            ")"}]}]}], ")"}], " ", 
        RowBox[{"(", 
         RowBox[{
          RowBox[{"sm", " ", 
           RowBox[{"(", 
            RowBox[{
             RowBox[{"sm", " ", 
              RowBox[{"(", 
               RowBox[{
                RowBox[{"-", "1"}], "+", "\[Delta]"}], ")"}]}], "-", 
             "\[Delta]"}], ")"}], " ", 
           SuperscriptBox[
            RowBox[{"(", 
             RowBox[{"sm", "-", "\[Delta]", "+", 
              RowBox[{"sm", " ", "\[Delta]"}]}], ")"}], "2"]}], "+", 
          RowBox[{
           SuperscriptBox["sf", "2"], " ", 
           RowBox[{"(", 
            RowBox[{
             RowBox[{
              RowBox[{"-", "4"}], " ", 
              SuperscriptBox[
               RowBox[{"(", 
                RowBox[{
                 RowBox[{"-", "1"}], "+", "sm"}], ")"}], "2"], " ", 
              SuperscriptBox[
               RowBox[{"(", 
                RowBox[{"1", "+", "\[Delta]"}], ")"}], "2"]}], "+", 
             RowBox[{
              SuperscriptBox["h", "2"], " ", 
              RowBox[{"(", 
               RowBox[{
                RowBox[{"-", "2"}], "+", "sm"}], ")"}], " ", 
              RowBox[{"(", 
               RowBox[{"2", "+", 
                RowBox[{"sm", " ", 
                 RowBox[{"(", 
                  RowBox[{
                   RowBox[{"-", "1"}], "+", "\[Delta]"}], ")"}]}], "-", 
                "\[Delta]"}], ")"}], " ", 
              SuperscriptBox[
               RowBox[{"(", 
                RowBox[{
                 RowBox[{"-", "2"}], "+", "sm", "-", "\[Delta]", "+", 
                 RowBox[{"sm", " ", "\[Delta]"}]}], ")"}], "2"]}], "+", 
             RowBox[{"2", " ", "h", " ", 
              RowBox[{"(", 
               RowBox[{
                RowBox[{"-", "1"}], "+", "sm"}], ")"}], " ", 
              RowBox[{"(", 
               RowBox[{"1", "+", "\[Delta]"}], ")"}], " ", 
              RowBox[{"(", 
               RowBox[{
                RowBox[{"-", "8"}], "-", 
                RowBox[{"2", " ", "\[Delta]"}], "+", 
                SuperscriptBox["\[Delta]", "2"], "+", 
                RowBox[{"sm", " ", 
                 RowBox[{"(", 
                  RowBox[{"8", "+", 
                   RowBox[{"3", " ", "\[Delta]"}], "-", 
                   RowBox[{"2", " ", 
                    SuperscriptBox["\[Delta]", "2"]}]}], ")"}]}], "+", 
                RowBox[{
                 SuperscriptBox["sm", "2"], " ", 
                 RowBox[{"(", 
                  RowBox[{
                   RowBox[{"-", "2"}], "-", "\[Delta]", "+", 
                   SuperscriptBox["\[Delta]", "2"]}], ")"}]}]}], ")"}]}]}], 
            ")"}]}], "-", 
          RowBox[{"2", " ", "sf", " ", 
           RowBox[{"(", 
            RowBox[{"sm", "-", "\[Delta]", "+", 
             RowBox[{"sm", " ", "\[Delta]"}]}], ")"}], " ", 
           RowBox[{"(", 
            RowBox[{
             RowBox[{"h", " ", 
              SuperscriptBox["sm", "3"], " ", 
              RowBox[{"(", 
               RowBox[{
                RowBox[{"-", "1"}], "+", 
                SuperscriptBox["\[Delta]", "2"]}], ")"}]}], "+", 
             RowBox[{"\[Delta]", " ", 
              RowBox[{"(", 
               RowBox[{"1", "+", "\[Delta]", "-", 
                RowBox[{"h", " ", 
                 RowBox[{"(", 
                  RowBox[{"2", "+", "\[Delta]"}], ")"}]}]}], ")"}]}], "-", 
             RowBox[{
              SuperscriptBox["sm", "2"], " ", 
              RowBox[{"(", 
               RowBox[{"2", "+", "\[Delta]", "-", 
                SuperscriptBox["\[Delta]", "2"], "+", 
                RowBox[{"h", " ", 
                 RowBox[{"(", 
                  RowBox[{
                   RowBox[{"-", "4"}], "+", "\[Delta]", "+", 
                   RowBox[{"3", " ", 
                    SuperscriptBox["\[Delta]", "2"]}]}], ")"}]}]}], ")"}]}], 
             "+", 
             RowBox[{"sm", " ", 
              RowBox[{"(", 
               RowBox[{"2", "-", 
                RowBox[{"2", " ", 
                 SuperscriptBox["\[Delta]", "2"]}], "+", 
                RowBox[{"h", " ", 
                 RowBox[{"(", 
                  RowBox[{
                   RowBox[{"-", "4"}], "+", 
                   RowBox[{"3", " ", "\[Delta]"}], "+", 
                   RowBox[{"3", " ", 
                    SuperscriptBox["\[Delta]", "2"]}]}], ")"}]}]}], ")"}]}]}],
             ")"}]}]}], ")"}]}], ")"}], "/", 
      RowBox[{"(", 
       RowBox[{"b", " ", "\[Alpha]", " ", 
        RowBox[{"(", 
         RowBox[{
          RowBox[{
           RowBox[{"-", "sm"}], " ", 
           RowBox[{"(", 
            RowBox[{
             RowBox[{"sm", " ", 
              RowBox[{"(", 
               RowBox[{
                RowBox[{"-", "1"}], "+", "\[Delta]"}], ")"}]}], "-", 
             "\[Delta]"}], ")"}], " ", 
           SuperscriptBox[
            RowBox[{"(", 
             RowBox[{"sm", "-", "\[Delta]", "+", 
              RowBox[{"sm", " ", "\[Delta]"}]}], ")"}], "4"]}], "+", 
          RowBox[{
           SuperscriptBox["h", "4"], " ", 
           SuperscriptBox["sf", "5"], " ", 
           SuperscriptBox[
            RowBox[{"(", 
             RowBox[{
              RowBox[{"-", "2"}], "+", "sm", "-", "\[Delta]", "+", 
              RowBox[{"sm", " ", "\[Delta]"}]}], ")"}], "4"], " ", 
           RowBox[{"(", 
            RowBox[{
             RowBox[{
              RowBox[{"-", 
               RowBox[{"(", 
                RowBox[{
                 RowBox[{"-", "1"}], "+", "sm"}], ")"}]}], " ", 
              RowBox[{"(", 
               RowBox[{"1", "+", "\[Delta]"}], ")"}]}], "+", 
             RowBox[{"h", " ", 
              RowBox[{"(", 
               RowBox[{
                RowBox[{"-", "2"}], "+", "sm", "-", "\[Delta]", "+", 
                RowBox[{"sm", " ", "\[Delta]"}]}], ")"}]}]}], ")"}]}], "-", 
          RowBox[{
           SuperscriptBox["h", "2"], " ", 
           SuperscriptBox["sf", "4"], " ", 
           SuperscriptBox[
            RowBox[{"(", 
             RowBox[{
              RowBox[{"-", "2"}], "+", "sm", "-", "\[Delta]", "+", 
              RowBox[{"sm", " ", "\[Delta]"}]}], ")"}], "2"], " ", 
           RowBox[{"(", 
            RowBox[{
             RowBox[{
              RowBox[{"-", "12"}], " ", 
              SuperscriptBox[
               RowBox[{"(", 
                RowBox[{
                 RowBox[{"-", "1"}], "+", "sm"}], ")"}], "2"], " ", 
              SuperscriptBox[
               RowBox[{"(", 
                RowBox[{"1", "+", "\[Delta]"}], ")"}], "2"]}], "+", 
             RowBox[{
              SuperscriptBox["h", "2"], " ", 
              SuperscriptBox[
               RowBox[{"(", 
                RowBox[{
                 RowBox[{"-", "2"}], "+", "sm", "-", "\[Delta]", "+", 
                 RowBox[{"sm", " ", "\[Delta]"}]}], ")"}], "2"], " ", 
              RowBox[{"(", 
               RowBox[{
                RowBox[{
                 SuperscriptBox["sm", "2"], " ", 
                 RowBox[{"(", 
                  RowBox[{
                   RowBox[{"-", "1"}], "+", "\[Delta]"}], ")"}]}], "-", 
                RowBox[{"2", " ", 
                 RowBox[{"(", 
                  RowBox[{"6", "+", "\[Delta]"}], ")"}]}], "+", 
                RowBox[{"sm", " ", 
                 RowBox[{"(", 
                  RowBox[{"8", "+", "\[Delta]"}], ")"}]}]}], ")"}]}], "-", 
             RowBox[{"2", " ", "h", " ", 
              RowBox[{"(", 
               RowBox[{
                RowBox[{"-", "1"}], "+", "sm"}], ")"}], " ", 
              RowBox[{"(", 
               RowBox[{"1", "+", "\[Delta]"}], ")"}], " ", 
              RowBox[{"(", 
               RowBox[{"24", "+", 
                RowBox[{"14", " ", "\[Delta]"}], "+", 
                SuperscriptBox["\[Delta]", "2"], "+", 
                RowBox[{
                 SuperscriptBox["sm", "2"], " ", 
                 RowBox[{"(", 
                  RowBox[{"4", "+", 
                   RowBox[{"5", " ", "\[Delta]"}], "+", 
                   SuperscriptBox["\[Delta]", "2"]}], ")"}]}], "-", 
                RowBox[{"sm", " ", 
                 RowBox[{"(", 
                  RowBox[{"20", "+", 
                   RowBox[{"19", " ", "\[Delta]"}], "+", 
                   RowBox[{"2", " ", 
                    SuperscriptBox["\[Delta]", "2"]}]}], ")"}]}]}], ")"}]}]}],
             ")"}]}], "+", 
          RowBox[{"2", " ", 
           SuperscriptBox["sf", "3"], " ", 
           RowBox[{"(", 
            RowBox[{
             RowBox[{
              RowBox[{"-", "16"}], " ", 
              SuperscriptBox[
               RowBox[{"(", 
                RowBox[{
                 RowBox[{"-", "1"}], "+", "sm"}], ")"}], "3"], " ", 
              SuperscriptBox[
               RowBox[{"(", 
                RowBox[{"1", "+", "\[Delta]"}], ")"}], "3"]}], "-", 
             RowBox[{
              SuperscriptBox["h", "2"], " ", 
              RowBox[{"(", 
               RowBox[{
                RowBox[{"-", "1"}], "+", "sm"}], ")"}], " ", 
              RowBox[{"(", 
               RowBox[{"1", "+", "\[Delta]"}], ")"}], " ", 
              SuperscriptBox[
               RowBox[{"(", 
                RowBox[{
                 RowBox[{"-", "2"}], "+", "sm", "-", "\[Delta]", "+", 
                 RowBox[{"sm", " ", "\[Delta]"}]}], ")"}], "2"], " ", 
              RowBox[{"(", 
               RowBox[{
                RowBox[{"16", " ", 
                 RowBox[{"(", 
                  RowBox[{"3", "+", "\[Delta]"}], ")"}]}], "+", 
                RowBox[{
                 SuperscriptBox["sm", "2"], " ", 
                 RowBox[{"(", 
                  RowBox[{"11", "+", 
                   RowBox[{"7", " ", "\[Delta]"}]}], ")"}]}], "-", 
                RowBox[{"sm", " ", 
                 RowBox[{"(", 
                  RowBox[{"40", "+", 
                   RowBox[{"23", " ", "\[Delta]"}]}], ")"}]}]}], ")"}]}], "+", 
             RowBox[{
              SuperscriptBox["h", "3"], " ", 
              SuperscriptBox[
               RowBox[{"(", 
                RowBox[{
                 RowBox[{"-", "2"}], "+", "sm", "-", "\[Delta]", "+", 
                 RowBox[{"sm", " ", "\[Delta]"}]}], ")"}], "3"], " ", 
              RowBox[{"(", 
               RowBox[{
                RowBox[{"8", " ", 
                 RowBox[{"(", 
                  RowBox[{"2", "+", "\[Delta]"}], ")"}]}], "+", 
                RowBox[{
                 SuperscriptBox["sm", "2"], " ", 
                 RowBox[{"(", 
                  RowBox[{"11", "+", 
                   RowBox[{"7", " ", "\[Delta]"}], "-", 
                   RowBox[{"4", " ", 
                    SuperscriptBox["\[Delta]", "2"]}]}], ")"}]}], "+", 
                RowBox[{"2", " ", 
                 SuperscriptBox["sm", "3"], " ", 
                 RowBox[{"(", 
                  RowBox[{
                   RowBox[{"-", "1"}], "+", 
                   SuperscriptBox["\[Delta]", "2"]}], ")"}]}], "+", 
                RowBox[{"sm", " ", 
                 RowBox[{"(", 
                  RowBox[{
                   RowBox[{"-", "20"}], "-", 
                   RowBox[{"15", " ", "\[Delta]"}], "+", 
                   RowBox[{"2", " ", 
                    SuperscriptBox["\[Delta]", "2"]}]}], ")"}]}]}], ")"}]}], 
             "-", 
             RowBox[{"4", " ", "h", " ", 
              SuperscriptBox[
               RowBox[{"(", 
                RowBox[{
                 RowBox[{"-", "1"}], "+", "sm"}], ")"}], "2"], " ", 
              SuperscriptBox[
               RowBox[{"(", 
                RowBox[{"1", "+", "\[Delta]"}], ")"}], "2"], " ", 
              RowBox[{"(", 
               RowBox[{
                RowBox[{"2", " ", 
                 RowBox[{"(", 
                  RowBox[{"12", "+", 
                   RowBox[{"8", " ", "\[Delta]"}], "+", 
                   SuperscriptBox["\[Delta]", "2"]}], ")"}]}], "+", 
                RowBox[{
                 SuperscriptBox["sm", "2"], " ", 
                 RowBox[{"(", 
                  RowBox[{"5", "+", 
                   RowBox[{"7", " ", "\[Delta]"}], "+", 
                   RowBox[{"2", " ", 
                    SuperscriptBox["\[Delta]", "2"]}]}], ")"}]}], "-", 
                RowBox[{"sm", " ", 
                 RowBox[{"(", 
                  RowBox[{"22", "+", 
                   RowBox[{"23", " ", "\[Delta]"}], "+", 
                   RowBox[{"4", " ", 
                    SuperscriptBox["\[Delta]", "2"]}]}], ")"}]}]}], ")"}]}]}],
             ")"}]}], "+", 
          RowBox[{"sf", " ", 
           SuperscriptBox[
            RowBox[{"(", 
             RowBox[{"sm", "-", "\[Delta]", "+", 
              RowBox[{"sm", " ", "\[Delta]"}]}], ")"}], "2"], " ", 
           RowBox[{"(", 
            RowBox[{
             RowBox[{"4", " ", "h", " ", 
              SuperscriptBox["sm", "4"], " ", 
              RowBox[{"(", 
               RowBox[{
                RowBox[{"-", "1"}], "+", "\[Delta]"}], ")"}], " ", 
              SuperscriptBox[
               RowBox[{"(", 
                RowBox[{"1", "+", "\[Delta]"}], ")"}], "2"]}], "+", 
             RowBox[{"sm", " ", 
              SuperscriptBox["\[Delta]", "2"], " ", 
              RowBox[{"(", 
               RowBox[{"3", "-", 
                RowBox[{"13", " ", "h"}], "+", 
                RowBox[{"3", " ", "\[Delta]"}], "-", 
                RowBox[{"7", " ", "h", " ", "\[Delta]"}]}], ")"}]}], "+", 
             RowBox[{
              SuperscriptBox["\[Delta]", "2"], " ", 
              RowBox[{"(", 
               RowBox[{
                RowBox[{"-", "1"}], "-", "\[Delta]", "+", 
                RowBox[{"h", " ", 
                 RowBox[{"(", 
                  RowBox[{"2", "+", "\[Delta]"}], ")"}]}]}], ")"}]}], "+", 
             RowBox[{"3", " ", 
              SuperscriptBox["sm", "2"], " ", 
              RowBox[{"(", 
               RowBox[{"1", "+", "\[Delta]"}], ")"}], " ", 
              RowBox[{"(", 
               RowBox[{"3", "-", 
                SuperscriptBox["\[Delta]", "2"], "+", 
                RowBox[{"h", " ", 
                 RowBox[{"(", 
                  RowBox[{
                   RowBox[{"-", "6"}], "+", 
                   RowBox[{"3", " ", "\[Delta]"}], "+", 
                   RowBox[{"5", " ", 
                    SuperscriptBox["\[Delta]", "2"]}]}], ")"}]}]}], ")"}]}], 
             "-", 
             RowBox[{
              SuperscriptBox["sm", "3"], " ", 
              RowBox[{"(", 
               RowBox[{"1", "+", "\[Delta]"}], ")"}], " ", 
              RowBox[{"(", 
               RowBox[{"9", "-", 
                SuperscriptBox["\[Delta]", "2"], "+", 
                RowBox[{"h", " ", 
                 RowBox[{"(", 
                  RowBox[{
                   RowBox[{"-", "17"}], "+", 
                   RowBox[{"4", " ", "\[Delta]"}], "+", 
                   RowBox[{"13", " ", 
                    SuperscriptBox["\[Delta]", "2"]}]}], ")"}]}]}], ")"}]}]}],
             ")"}]}], "-", 
          RowBox[{"2", " ", 
           SuperscriptBox["sf", "2"], " ", 
           RowBox[{"(", 
            RowBox[{"sm", "-", "\[Delta]", "+", 
             RowBox[{"sm", " ", "\[Delta]"}]}], ")"}], " ", 
           RowBox[{"(", 
            RowBox[{
             RowBox[{"3", " ", 
              SuperscriptBox["h", "2"], " ", 
              SuperscriptBox["sm", "5"], " ", 
              RowBox[{"(", 
               RowBox[{
                RowBox[{"-", "1"}], "+", "\[Delta]"}], ")"}], " ", 
              SuperscriptBox[
               RowBox[{"(", 
                RowBox[{"1", "+", "\[Delta]"}], ")"}], "3"]}], "-", 
             RowBox[{"h", " ", 
              SuperscriptBox["sm", "4"], " ", 
              SuperscriptBox[
               RowBox[{"(", 
                RowBox[{"1", "+", "\[Delta]"}], ")"}], "2"], " ", 
              RowBox[{"(", 
               RowBox[{"12", "+", 
                RowBox[{"3", " ", "\[Delta]"}], "-", 
                SuperscriptBox["\[Delta]", "2"], "+", 
                RowBox[{"h", " ", 
                 RowBox[{"(", 
                  RowBox[{
                   RowBox[{"-", "24"}], "+", 
                   RowBox[{"3", " ", "\[Delta]"}], "+", 
                   RowBox[{"13", " ", 
                    SuperscriptBox["\[Delta]", "2"]}]}], ")"}]}]}], ")"}]}], 
             "+", 
             RowBox[{"\[Delta]", " ", 
              RowBox[{"(", 
               RowBox[{
                RowBox[{"2", " ", 
                 SuperscriptBox[
                  RowBox[{"(", 
                   RowBox[{"1", "+", "\[Delta]"}], ")"}], "2"]}], "-", 
                RowBox[{
                 SuperscriptBox["h", "2"], " ", 
                 RowBox[{"(", 
                  RowBox[{
                   RowBox[{"-", "2"}], "+", "\[Delta]"}], ")"}], " ", 
                 SuperscriptBox[
                  RowBox[{"(", 
                   RowBox[{"2", "+", "\[Delta]"}], ")"}], "2"]}], "+", 
                RowBox[{"h", " ", 
                 RowBox[{"(", 
                  RowBox[{
                   RowBox[{"-", "8"}], "-", 
                   RowBox[{"10", " ", "\[Delta]"}], "-", 
                   SuperscriptBox["\[Delta]", "2"], "+", 
                   SuperscriptBox["\[Delta]", "3"]}], ")"}]}]}], ")"}]}], "+", 
             RowBox[{
              SuperscriptBox["sm", "3"], " ", 
              RowBox[{"(", 
               RowBox[{"1", "+", "\[Delta]"}], ")"}], " ", 
              RowBox[{"(", 
               RowBox[{
                RowBox[{
                 RowBox[{"-", "2"}], " ", 
                 RowBox[{"(", 
                  RowBox[{"7", "+", 
                   RowBox[{"8", " ", "\[Delta]"}], "+", 
                   SuperscriptBox["\[Delta]", "2"]}], ")"}]}], "+", 
                RowBox[{"h", " ", 
                 RowBox[{"(", 
                  RowBox[{"64", "+", 
                   RowBox[{"68", " ", "\[Delta]"}], "+", 
                   RowBox[{"8", " ", 
                    SuperscriptBox["\[Delta]", "2"]}], "-", 
                   RowBox[{"4", " ", 
                    SuperscriptBox["\[Delta]", "3"]}]}], ")"}]}], "+", 
                RowBox[{
                 SuperscriptBox["h", "2"], " ", 
                 RowBox[{"(", 
                  RowBox[{
                   RowBox[{"-", "74"}], "-", 
                   RowBox[{"55", " ", "\[Delta]"}], "+", 
                   RowBox[{"31", " ", 
                    SuperscriptBox["\[Delta]", "2"]}], "+", 
                   RowBox[{"22", " ", 
                    SuperscriptBox["\[Delta]", "3"]}]}], ")"}]}]}], ")"}]}], 
             "+", 
             RowBox[{
              SuperscriptBox["sm", "2"], " ", 
              RowBox[{"(", 
               RowBox[{
                RowBox[{"2", " ", 
                 SuperscriptBox[
                  RowBox[{"(", 
                   RowBox[{"1", "+", "\[Delta]"}], ")"}], "2"], " ", 
                 RowBox[{"(", 
                  RowBox[{"14", "+", 
                   RowBox[{"3", " ", "\[Delta]"}]}], ")"}]}], "+", 
                RowBox[{
                 SuperscriptBox["h", "2"], " ", 
                 RowBox[{"(", 
                  RowBox[{"104", "+", 
                   RowBox[{"166", " ", "\[Delta]"}], "+", 
                   RowBox[{"37", " ", 
                    SuperscriptBox["\[Delta]", "2"]}], "-", 
                   RowBox[{"45", " ", 
                    SuperscriptBox["\[Delta]", "3"]}], "-", 
                   RowBox[{"18", " ", 
                    SuperscriptBox["\[Delta]", "4"]}]}], ")"}]}], "+", 
                RowBox[{"3", " ", "h", " ", 
                 RowBox[{"(", 
                  RowBox[{
                   RowBox[{"-", "36"}], "-", 
                   RowBox[{"69", " ", "\[Delta]"}], "-", 
                   RowBox[{"37", " ", 
                    SuperscriptBox["\[Delta]", "2"]}], "-", 
                   RowBox[{"2", " ", 
                    SuperscriptBox["\[Delta]", "3"]}], "+", 
                   RowBox[{"2", " ", 
                    SuperscriptBox["\[Delta]", "4"]}]}], ")"}]}]}], ")"}]}], 
             "+", 
             RowBox[{"sm", " ", 
              RowBox[{"(", 
               RowBox[{
                RowBox[{
                 RowBox[{"-", "2"}], " ", 
                 SuperscriptBox[
                  RowBox[{"(", 
                   RowBox[{"1", "+", "\[Delta]"}], ")"}], "2"], " ", 
                 RowBox[{"(", 
                  RowBox[{"7", "+", 
                   RowBox[{"3", " ", "\[Delta]"}]}], ")"}]}], "+", 
                RowBox[{"h", " ", 
                 RowBox[{"(", 
                  RowBox[{"56", "+", 
                   RowBox[{"110", " ", "\[Delta]"}], "+", 
                   RowBox[{"62", " ", 
                    SuperscriptBox["\[Delta]", "2"]}], "+", 
                   RowBox[{"4", " ", 
                    SuperscriptBox["\[Delta]", "3"]}], "-", 
                   RowBox[{"4", " ", 
                    SuperscriptBox["\[Delta]", "4"]}]}], ")"}]}], "+", 
                RowBox[{
                 SuperscriptBox["h", "2"], " ", 
                 RowBox[{"(", 
                  RowBox[{
                   RowBox[{"-", "56"}], "-", 
                   RowBox[{"84", " ", "\[Delta]"}], "-", 
                   RowBox[{"22", " ", 
                    SuperscriptBox["\[Delta]", "2"]}], "+", 
                   RowBox[{"17", " ", 
                    SuperscriptBox["\[Delta]", "3"]}], "+", 
                   RowBox[{"7", " ", 
                    SuperscriptBox["\[Delta]", "4"]}]}], ")"}]}]}], ")"}]}]}],
             ")"}]}]}], ")"}]}], ")"}]}]}]}], "]"}]}], "\[IndentingNewLine]", 
 RowBox[{
  RowBox[{"func2", "[", 
   RowBox[{"{", 
    RowBox[{"b_", ",", "\[Alpha]_"}], "}"}], "]"}], ":=", 
  RowBox[{"If", "[", 
   RowBox[{
    RowBox[{
     FractionBox[
      RowBox[{
       RowBox[{"-", "2"}], "+", "b"}], 
      RowBox[{"b", " ", "\[Alpha]"}]], ">", "0"}], ",", 
    FractionBox[
     RowBox[{
      RowBox[{"-", "2"}], "+", "b"}], 
     RowBox[{"b", " ", "\[Alpha]"}]], ",", "Null"}], 
   "]"}]}], "\[IndentingNewLine]", 
 RowBox[{
  RowBox[{"func3", "[", 
   RowBox[{"{", 
    RowBox[{
    "sf_", ",", " ", "sm_", ",", " ", "h_", ",", " ", "\[Delta]_", ",", 
     "\[Alpha]_", ",", " ", "c_", ",", " ", "b_"}], "}"}], "]"}], ":=", 
  RowBox[{"If", "[", 
   RowBox[{
    RowBox[{
     RowBox[{
      FractionBox[
       RowBox[{"2", "+", 
        RowBox[{"b", " ", 
         RowBox[{"(", 
          RowBox[{
           RowBox[{"-", "1"}], "+", "sf"}], ")"}], " ", 
         RowBox[{"(", 
          RowBox[{"1", "+", "\[Delta]"}], ")"}]}]}], 
       RowBox[{"b", " ", "\[Alpha]", " ", 
        RowBox[{"(", 
         RowBox[{
          RowBox[{"-", "2"}], "+", "sf", "+", "sm", "+", 
          RowBox[{"sf", " ", "\[Delta]"}], "-", 
          RowBox[{"sm", " ", "\[Delta]"}]}], ")"}]}]], "-", 
      FractionBox[
       RowBox[{
        RowBox[{"(", 
         RowBox[{
          RowBox[{"-", "1"}], "+", "sm"}], ")"}], " ", 
        RowBox[{"(", 
         RowBox[{
          RowBox[{"-", "1"}], "+", "\[Delta]"}], ")"}], " ", 
        RowBox[{"(", 
         RowBox[{"2", "+", 
          RowBox[{"b", " ", 
           RowBox[{"(", 
            RowBox[{
             RowBox[{"-", "1"}], "+", "sf"}], ")"}], " ", 
           RowBox[{"(", 
            RowBox[{"1", "+", "\[Delta]"}], ")"}]}]}], ")"}]}], 
       RowBox[{"b", " ", 
        RowBox[{"(", 
         RowBox[{
          RowBox[{"-", "1"}], "+", "sf"}], ")"}], " ", "\[Alpha]", " ", 
        RowBox[{"(", 
         RowBox[{"1", "+", "\[Delta]"}], ")"}], " ", 
        RowBox[{"(", 
         RowBox[{
          RowBox[{"-", "2"}], "+", "sf", "+", "sm", "+", 
          RowBox[{"sf", " ", "\[Delta]"}], "-", 
          RowBox[{"sm", " ", "\[Delta]"}]}], ")"}]}]]}], ">", "0"}], ",", 
    RowBox[{
     FractionBox[
      RowBox[{"2", "+", 
       RowBox[{"b", " ", 
        RowBox[{"(", 
         RowBox[{
          RowBox[{"-", "1"}], "+", "sf"}], ")"}], " ", 
        RowBox[{"(", 
         RowBox[{"1", "+", "\[Delta]"}], ")"}]}]}], 
      RowBox[{"b", " ", "\[Alpha]", " ", 
       RowBox[{"(", 
        RowBox[{
         RowBox[{"-", "2"}], "+", "sf", "+", "sm", "+", 
         RowBox[{"sf", " ", "\[Delta]"}], "-", 
         RowBox[{"sm", " ", "\[Delta]"}]}], ")"}]}]], "-", 
     FractionBox[
      RowBox[{
       RowBox[{"(", 
        RowBox[{
         RowBox[{"-", "1"}], "+", "sm"}], ")"}], " ", 
       RowBox[{"(", 
        RowBox[{
         RowBox[{"-", "1"}], "+", "\[Delta]"}], ")"}], " ", 
       RowBox[{"(", 
        RowBox[{"2", "+", 
         RowBox[{"b", " ", 
          RowBox[{"(", 
           RowBox[{
            RowBox[{"-", "1"}], "+", "sf"}], ")"}], " ", 
          RowBox[{"(", 
           RowBox[{"1", "+", "\[Delta]"}], ")"}]}]}], ")"}]}], 
      RowBox[{"b", " ", 
       RowBox[{"(", 
        RowBox[{
         RowBox[{"-", "1"}], "+", "sf"}], ")"}], " ", "\[Alpha]", " ", 
       RowBox[{"(", 
        RowBox[{"1", "+", "\[Delta]"}], ")"}], " ", 
       RowBox[{"(", 
        RowBox[{
         RowBox[{"-", "2"}], "+", "sf", "+", "sm", "+", 
         RowBox[{"sf", " ", "\[Delta]"}], "-", 
         RowBox[{"sm", " ", "\[Delta]"}]}], ")"}]}]]}], ",", "Null"}], 
   "]"}]}]}], "Input",
 CellChangeTimes->{{3.7840487451446867`*^9, 3.784048798096971*^9}, {
  3.7855723745168943`*^9, 3.785572375583008*^9}, {3.7855724553828163`*^9, 
  3.785572456146434*^9}, {3.800009315045189*^9, 3.800009315347846*^9}, {
  3.813930849580433*^9, 3.8139308505473623`*^9}},
 CellLabel->
  "In[1346]:=",ExpressionUUID->"e2a5396b-cee0-4f52-8545-db86db08f2f0"],

Cell["Produce plot", "Text",
 CellChangeTimes->{{3.813930855200848*^9, 
  3.8139308567422743`*^9}},ExpressionUUID->"d1f15f6d-3212-4137-81ef-\
697918aa6d83"],

Cell[CellGroupData[{

Cell[BoxData[{
 RowBox[{
  RowBox[{"testParameters", " ", "=", " ", 
   RowBox[{"{", 
    RowBox[{
     RowBox[{"h", "\[Rule]", "0.1"}], ",", 
     RowBox[{"\[Alpha]", "\[Rule]", 
      FractionBox["1", "1000"]}], ",", 
     RowBox[{"\[Delta]", "\[Rule]", "1"}], ",", 
     RowBox[{"c", "\[Rule]", "1"}], ",", 
     RowBox[{"sm", "\[Rule]", "0"}], ",", 
     RowBox[{"sf", "\[Rule]", "0.8"}]}], "}"}]}], 
  ";"}], "\[IndentingNewLine]", 
 RowBox[{
  RowBox[{"weakDriveParams", " ", "=", " ", 
   RowBox[{"{", 
    RowBox[{
     RowBox[{"h", "\[Rule]", "0"}], ",", 
     RowBox[{"\[Alpha]", "\[Rule]", 
      RowBox[{"10", "^", 
       RowBox[{"(", 
        RowBox[{"-", "3"}], ")"}]}]}], ",", 
     RowBox[{"\[Delta]", "\[Rule]", "0.25"}], ",", 
     RowBox[{"c", "\[Rule]", "1"}], ",", 
     RowBox[{"sm", "\[Rule]", "0"}], ",", 
     RowBox[{"sf", "\[Rule]", "0"}]}], "}"}]}], ";"}], "\[IndentingNewLine]", 
 RowBox[{"Plot", "[", 
  RowBox[{
   RowBox[{"{", 
    RowBox[{
     RowBox[{"func", "[", 
      RowBox[{"(", 
       RowBox[{
        RowBox[{"{", 
         RowBox[{
         "sf", ",", " ", "sm", ",", " ", "h", ",", " ", "\[Delta]", ",", 
          "\[Alpha]", ",", " ", "c", ",", " ", "b"}], "}"}], "/.", 
        "testParameters"}], ")"}], "]"}], ",", 
     RowBox[{"func3", "[", 
      RowBox[{"(", 
       RowBox[{
        RowBox[{"{", 
         RowBox[{
         "sf", ",", " ", "sm", ",", " ", "h", ",", " ", "\[Delta]", ",", 
          "\[Alpha]", ",", " ", "c", ",", " ", "b"}], "}"}], "/.", 
        "weakDriveParams"}], ")"}], "]"}], ",", 
     RowBox[{"func2", "[", 
      RowBox[{"(", 
       RowBox[{
        RowBox[{"{", 
         RowBox[{"b", ",", "\[Alpha]"}], "}"}], "/.", "testParameters"}], 
       ")"}], "]"}]}], "}"}], ",", 
   RowBox[{"{", 
    RowBox[{"b", ",", "1", ",", "5"}], "}"}], ",", 
   RowBox[{"AxesLabel", "\[Rule]", 
    RowBox[{"{", 
     RowBox[{"b", ",", "N"}], "}"}]}], ",", 
   RowBox[{"PlotLegends", "\[Rule]", 
    RowBox[{"{", 
     RowBox[{
     "\"\<strong drive at equilibrium\>\"", ",", 
      "\"\<weak drive at fixation\>\"", ",", "\"\<wild-type\>\""}], "}"}]}]}],
   "]"}]}], "Input",
 CellChangeTimes->{{3.7897328085491533`*^9, 3.789732811887224*^9}, {
   3.789732845930471*^9, 3.789732929152852*^9}, {3.789733220604877*^9, 
   3.78973323877112*^9}, {3.789733443007872*^9, 3.789733452694234*^9}, 
   3.789733498313501*^9, {3.789734583327293*^9, 3.789734583467265*^9}, {
   3.790078635126957*^9, 3.790078724087949*^9}, {3.790078756175313*^9, 
   3.790078761302916*^9}, {3.790083741063793*^9, 3.790083748936092*^9}, {
   3.7900844385563087`*^9, 3.7900844386899347`*^9}, {3.800008950016754*^9, 
   3.80000895067909*^9}, {3.800009207797346*^9, 3.800009279281457*^9}, 
   3.800009319162566*^9},
 CellLabel->
  "In[1349]:=",ExpressionUUID->"a22582de-5b0f-4ce0-a46a-bc9a52795dd9"],

Cell[BoxData[
 TemplateBox[{GraphicsBox[{{{{}, {}, 
       TagBox[{
         Directive[
          Opacity[1.], 
          RGBColor[0.368417, 0.506779, 0.709798], 
          AbsoluteThickness[1.6]], 
         LineBox[CompressedData["
1:eJwVjn0wlAkAxhdpfcV633ylkrs92bpw4tTqvA85Kt+VmL0IW9cltSxFd8Lu
a+/U9XWX2+lCajIu5c4sey7nI+Sjayokg0hluvXRZsm2Rbu5vT+eeeaZ+c1v
HuckwfZ9hgwGY6s+/3d6WYDv81daqtKl0VBCXaQq7wYmqGe0lKVrR6Y8eZSK
vBa8if9WSyW4aJnlSjWllmyzf6jVUmuUnLaOCAYu8MPUMNRRQnHERdPVi7HJ
P7K7iqmjqnt2tyzXmkOijS44TeiokbryjOEuEj9/fWWu1E5H5dn0X+qU2OJS
tzK5ermOiuWm6pq5Dqgto8P6XfS8TXXksysrMR4qI5x9dVTGeSuJoJqN0BKL
EjlfRzGGe5rZe9xB+N2WDcl11NApm7oTHlyU5noMcWI+UClvjiXPBgbjYMCy
/fxfFijFI6fNnEXRsL7zRrszk4HqmiLVy73xePteoEq2N4Dbpw9zHL/gY8eb
uH3MTgNgxZ4iRch+FJzpzRQKDeEe3Vc4p01Gp5y3PpIwwljSuvbTnoeRfTla
KL9lhEtdn9Wn16XC1yGJ92PiIoS19NnPsIVIjRltH2cZQ+Z6QuomTUeO9+T9
pY3GkCtCDllNZkB62Si3Jm4xbnoRafKlR8HuOVczYM4EkTNpsSooEw1n3H9K
/YOJ1wvFHOGOLKxTRpv3h5lgo59PeHfYMSy1mjZpmzcB8huORkR8C+NbRVUf
F5viXXtSEG/9d0jxqh80DTEDL/yJI2dlNiq3Jl1u0pjhgWg61nEuG5qXy1gK
qTmkEyZef9cfR+2OYKPCQAtoOnq5FpIc5G2RlMoVFhg5Ef+7cEMuPnqybeCq
ZAlOmvibDTzLha9PfqGDpyVuBvZFjQTnIW5l15xBryVctAqZoTQPmrunR1NF
VhhYu03lOZaHUEN3LodgQem51mHMXQQzFS/KIIkF8gBflpohAt2E68U3WWDR
Tbte1Isg2lkVk2Zqjb1WnpWLtCJw7/cHmSZaY3pX3OLnfmI8tQnMN/nLGm0x
W9wNc8Qw/WSjrT+TwOwPPi87asWYsh2f2BRPIGtsNp39WozYLidCWEUg0Vg1
VbKaRlVMUYHlAoHsg73Z4TwaLa3dOBxNokLg5X3kLI1Y9ZkwZjkJTrD/yW9u
0eh4EOAh+I2EIqXB1beZhleFhui/RiLgkH3FkhYaVvHxA+U3SCw8PBVR3arn
O9fxg2Qk4kodvebbaXgX3cuSNJD4vPhwYv49GkSAWZlRL4ndRUeOiQdp5C1v
+v7gIxJl91MSoh7TmNKkHejtI+FnzlqzaojGPzcG3a4OkPhSmr6vcVjP21TU
BYyQcDaLO655SkM1HtwlmiCx/1e2D19BY/ft97KJSRI3GttCPMZo3C2pKoxS
6v8Opm/W6Xf5djues4rE8Qy7YekEjfgGxYtmNYlHoU41d5Q07kkvdrpqSJzf
Ovqs8BUNblr49XNvSagt/p1JmKJh41IrSJgnUe7W2vhORUPMSN5+5z2JDZXX
stqmaUw/XuHtoSOhdBuzPTej9//ZY3fhAwl2K33hq9d6/1nJ/MICiQJBvs5l
lsZ/5QQZWA==
          "]]}, Annotation[#, "Charting`Private`Tag$97410#1"]& ], 
       TagBox[{
         Directive[
          Opacity[1.], 
          RGBColor[0.880722, 0.611041, 0.142051], 
          AbsoluteThickness[1.6]], 
         LineBox[CompressedData["
1:eJwV0nk41AkcBvBxjivH/BA9rmm3dFmSSVH7e6kQ1TrCNhsKLaEclS5lxvgV
2Rldpototcoo1i2Jjs0xhS2JNtSqh8Iazz6uGcfs7B/f530+z/M+719fZmis
7z5lGo3mo7j/06DsyKv716Sk8PXcfbPiNLKNHn7xX4XtFiVFCZYWkR9qlX/z
yZWSb01zmDEDw+TYgVvVv9+Wkg69W4W19BmSxoRYT6RwFu1hqLYqrM6ekrRW
SMnWayK5Rbo+9vhNr/cUS8nEpznDZcnmiFO/sq2wXUp2pR/2nd1tBc4DVohG
p5QsynwWckjKRJ7lodTGD1Ky0jqkJsRuCf4e/qfNdUJKfiQclTcKVyI09VOY
s5WMvFoUK7FnOiC8so2/KlFGLq69m93uDvSMV7JsT8pIb7PArf75gL9DTu9q
jozk+mawa+YBt/JoG8dzMrLn9fC0caULrEs1Wl1uykjNHiv3dVabMHTPVSew
UUZaXDBXNxvbgvjbVee4xjOkU8QgfSTRC6cv3Tz7tmqGDPZel+XN8AM165/G
Z8yRvHRBhkdPEHKT7d4vD5wn19YOLH1+NgzRrosiwrLk5PQzbY2E4ggYNE/M
7jxKg/kua9OqN9GYmomVRJkoQV2pUMc7MhZ+E0H76E1KkDQE1IiT45Em6Dia
kKCMPmbltQnmITRVsNd4M1TQ/GJN30n+YSTl+SdUNKjgrcMw2Th3BM6moeyM
vapguNorxZNHERfY//yLvhrE5XKO/ZljOM0aajV8pIbaS56s6bLjEOapJJcH
qaN2SvUWHp7At6/Ol3dr0yEoIvXTWk6iTmB7Ia6Yjk0Xl/n0VifBZsRfu2u7
BnJClXtjsk7BUG9M4w+ZBkYCuuO8jp+GWsONkm+yNSF+YONiuikZMQ4P32l6
aeGipctgvCYH97aG5tVPaiHc0SXJN4qDyeFF+gNCbQjdnH6NfcxBlZ+7yuXN
OgjivC7lmXLB8aByKwZ08GZk9sXnSC4W93p251ML0JOoFhNRzoWzY+plU3td
XE8Yl7XQUhBk0S5V6tDFLx7Se2NbUjAp5vfHcfWgI2dzW/gp2KZs67ScoY+K
uJ5VVm0p0JKwfZRC9eHJypxv1OKBVw9Rdo0+NuRGqit78sDdWRIYr2kAtx+a
eU9SeHBq7XLT3GsAeyfxiZFHPHww2pyqUW2AocAX1uIJHjSXrDd2oTMQUGP3
XH1ZKkaNv3zdEMzANktRaebuVPzYbslIKGFgdGq1YX9GKkoCb6TpyhnI1ho+
/6gmFU+e/omD/gTeh/OrGgcV/XHBdnoBARvn6bssBoXGNle72DsE1Hfd0R5T
2KFwktF1l0A9v9hPRFDQCw7uLigiYGzSc8nCSNFvsglzKyVQx/brVDOhwLrx
8hhVRyBSKNrbbk4p/kTrtkoHgWbPiKYdyylwzOrPRL8hcOXnBR/VVlAYnYzf
39FJgLa/oa9O4Zaid9/ldxOwXx0kWLFK0TcqfODaR6D/s2m1qi0FyRf3du5X
AiYHw9ZUOFDY/Wym9OuQYl/ALtnPoiDOKbnsM0IgwKrSwHIthQLfhWymhECB
qyg93ZFCcN3A58fjBEYHVroHOVF4KbzetGySgOFs56SBMwWn+B2i81MEysp1
+Y0KGy2tit0jI+BxxuWA7UYKKbQo3+YZAgk/uVV/UnjsL3OW3RyBY+6SwSvf
K/YrXy28Ok+gON9F7kUq9jMpmVxOYKDJdU6u8H8+lEzI
          "]]}, Annotation[#, "Charting`Private`Tag$97410#2"]& ], 
       TagBox[{
         Directive[
          Opacity[1.], 
          RGBColor[0.560181, 0.691569, 0.194885], 
          AbsoluteThickness[1.6]], 
         LineBox[CompressedData["
1:eJwVkXs01AkUx38MeTbG/AYhNKy2x1bY2GLw9SipFOJwZqNQFNZjKtPkOb9S
elBOk80jtBhRZ7EeK2mQWnRS65jyOImsFZOTdmMq0uzsH/fc8z3ncz733nPZ
4fH+h1UJgtijrP+7i6lHsQlBIKnTbGf413xX7vPUlnJl5o5WjMyZPHSV88u8
gmgEroQ0iH5iyVxzTXqkWssInH35Y+Xp7I+u3QdYs3E6BPwCTnp3eC7DIVXH
FDZdyd8cW7FSqANF+QEtKYPArkfDUcMhDDjIqq22GirnaVdJ+q4ZoOy4WzCN
TaB0KsnZa9ECLkZRfzdaEdALW7WL483G8N1LvCOrCdzUDbeLFFmCoRi41Lue
QPa0/eW4tdZIOx/X8bMDgZkl6wtBe9aBW1y47rvdBFwDulaPuNlBv3v+SwCf
gE3y+cWVezkYq20LGBIQ8Ow+PJX9Fwe1+efvhKQo9ymoF/bxneEbY7Y/Ukjg
g8813b4SF+TQt93jXyTQHl0+PpQEaAeKBAUlBKZ2GEovtbtDbdxOPtpFgF3b
8U7g4YWPi/Gz0StUcJkSW3Sr+WLffMhhjS4VVLXV5c9/DUZWTj+fx1NFtdST
mBQdRFcD93tfJg28VOHtRvYhpJQG8hraaDBSe1IS3xIFJ+Nw7sUwNayvWKIK
XkUjIWj80RRDHclu6e5penFIs5f1su6ro3LLDfem9ATkldLS60OWQWy/vYfx
OBHf9F2pH9TRAC9vL0tmegytOZtyE37VgLSJVyHddhwbZgJ1Bnw0sW3RndXN
PQGW3nvNhwuaEMfSfYr5SVBvK6yxKtKC8xpB7ugxPmI33xvS2qUN2VPLB0X8
k7jjHV4qkWsr/1nm8PSoAPK3JozJPB3wa3MNjNxOoWmfF03kqYs3yOH0WCQj
Y0dmScOkLkjfT0tOk8mwHNk5WJa5HNZVy3Oo0hQ4/XBGZGxHB2Ue8eFTWCpC
zJ99VumnI9Q+vnaIngb54+zxBKEeilIjOvkP0rBbdZPjWiYDncbfJvhGpEN7
luunEs6Ak3pkyj8L6TgtQXVRMwM6XapZ1O4MCANqghK19GEc0TKReDUDjr0D
27XC9NHenZIfNZKBUQPPM5q/62Nig6SOYyWElvVWQzcNJmImU5I0w4V4Zzg1
zQllwqKSVVxfJkTwMwsmr4YJ2QmX/OFXQtQEFWbRFUx49FrPzxhS6HjwJ+IC
SZASm0PNfhSC53J8NMQkPns7kJ1nKfzx1N0mvpKE16Zc+p1zFDZXyZkDt0hs
tC2YEmVR0AsNHRTfJmFi3r8l8oKS79oQsb2ORLTtrZ0aORTsC5+czGwlocox
cPS4RoHprl1O6ydxysxN8EsZhYyVkrMxUhLO7Yzgc+UU3skTj/Y/J+HqEW8Y
W0Gh5/bQxrJBEmmpc7b2lUreoOqu+ysSpqFjE4+qKcxOeT0TTpMo9ubgdR2F
/Z2LddMyEsl2tqYPf6Pw+EaNyG+GROwRyxfiegpifyMue5bE3n7WTEwjhdDW
yYn2ORJVY/Lrc80UnuQVdK2RK+9Jvnj8xV0Kjol7qq98JFF60N+muYWCweqm
+IMLJF6+yd2R3EqBIqL9uxeVvn+NKvbfp/B+2MzeZolEoPHb184Spb+xz+j6
VxLXdZkK8zal/3LmgkJBYlRw9YtCmf8D2h4fFw==
          "]]}, Annotation[#, "Charting`Private`Tag$97410#3"]& ]}}, {}}, {
    DisplayFunction -> Identity, Ticks -> {Automatic, Automatic}, 
     AxesOrigin -> {1., 0}, FrameTicks -> {{Automatic, 
        Charting`ScaledFrameTicks[{Identity, Identity}]}, {Automatic, 
        Charting`ScaledFrameTicks[{Identity, Identity}]}}, 
     GridLines -> {None, None}, DisplayFunction -> Identity, 
     PlotRangePadding -> {{
        Scaled[0.02], 
        Scaled[0.02]}, {
        Scaled[0.05], 
        Scaled[0.05]}}, PlotRangeClipping -> True, ImagePadding -> All, 
     DisplayFunction -> Identity, AspectRatio -> 
     NCache[GoldenRatio^(-1), 0.6180339887498948], Axes -> {True, True}, 
     AxesLabel -> {
       FormBox[
        TagBox["b", HoldForm], TraditionalForm], 
       FormBox["N", TraditionalForm]}, AxesOrigin -> {1., 0}, DisplayFunction :> 
     Identity, Frame -> {{False, False}, {False, False}}, 
     FrameLabel -> {{None, None}, {None, None}}, 
     FrameTicks -> {{Automatic, Automatic}, {Automatic, Automatic}}, 
     GridLines -> {None, None}, GridLinesStyle -> Directive[
       GrayLevel[0.5, 0.4]], 
     Method -> {
      "DefaultBoundaryStyle" -> Automatic, 
       "DefaultGraphicsInteraction" -> {
        "Version" -> 1.2, "TrackMousePosition" -> {True, False}, 
         "Effects" -> {
          "Highlight" -> {"ratio" -> 2}, "HighlightPoint" -> {"ratio" -> 2}, 
           "Droplines" -> {
            "freeformCursorMode" -> True, 
             "placement" -> {"x" -> "All", "y" -> "None"}}}}, 
       "DefaultMeshStyle" -> AbsolutePointSize[6], "ScalingFunctions" -> None,
        "CoordinatesToolOptions" -> {"DisplayFunction" -> ({
           (Identity[#]& )[
            Part[#, 1]], 
           (Identity[#]& )[
            Part[#, 2]]}& ), "CopiedValueFunction" -> ({
           (Identity[#]& )[
            Part[#, 1]], 
           (Identity[#]& )[
            Part[#, 2]]}& )}}, PlotRange -> {{1, 5}, {0., 679.9999947755102}},
      PlotRangeClipping -> True, PlotRangePadding -> {{
        Scaled[0.02], 
        Scaled[0.02]}, {
        Scaled[0.02], 
        Scaled[0.02]}}, Ticks -> {Automatic, Automatic}}],FormBox[
    FormBox[
     TemplateBox[{
      "\"strong drive at equilibrium\"", "\"weak drive at fixation\"", 
       "\"wild-type\""}, "LineLegend", DisplayFunction -> (FormBox[
        StyleBox[
         StyleBox[
          PaneBox[
           TagBox[
            GridBox[{{
               TagBox[
                GridBox[{{
                   GraphicsBox[{{
                    Directive[
                    EdgeForm[
                    Directive[
                    Opacity[0.3], 
                    GrayLevel[0]]], 
                    PointSize[0.5], 
                    Opacity[1.], 
                    RGBColor[0.368417, 0.506779, 0.709798], 
                    AbsoluteThickness[1.6]], {
                    LineBox[{{0, 10}, {20, 10}}]}}, {
                    Directive[
                    EdgeForm[
                    Directive[
                    Opacity[0.3], 
                    GrayLevel[0]]], 
                    PointSize[0.5], 
                    Opacity[1.], 
                    RGBColor[0.368417, 0.506779, 0.709798], 
                    AbsoluteThickness[1.6]], {}}}, AspectRatio -> Full, 
                    ImageSize -> {20, 10}, PlotRangePadding -> None, 
                    ImagePadding -> Automatic, 
                    BaselinePosition -> (Scaled[0.1] -> Baseline)], #}, {
                   GraphicsBox[{{
                    Directive[
                    EdgeForm[
                    Directive[
                    Opacity[0.3], 
                    GrayLevel[0]]], 
                    PointSize[0.5], 
                    Opacity[1.], 
                    RGBColor[0.880722, 0.611041, 0.142051], 
                    AbsoluteThickness[1.6]], {
                    LineBox[{{0, 10}, {20, 10}}]}}, {
                    Directive[
                    EdgeForm[
                    Directive[
                    Opacity[0.3], 
                    GrayLevel[0]]], 
                    PointSize[0.5], 
                    Opacity[1.], 
                    RGBColor[0.880722, 0.611041, 0.142051], 
                    AbsoluteThickness[1.6]], {}}}, AspectRatio -> Full, 
                    ImageSize -> {20, 10}, PlotRangePadding -> None, 
                    ImagePadding -> Automatic, 
                    BaselinePosition -> (Scaled[0.1] -> Baseline)], #2}, {
                   GraphicsBox[{{
                    Directive[
                    EdgeForm[
                    Directive[
                    Opacity[0.3], 
                    GrayLevel[0]]], 
                    PointSize[0.5], 
                    Opacity[1.], 
                    RGBColor[0.560181, 0.691569, 0.194885], 
                    AbsoluteThickness[1.6]], {
                    LineBox[{{0, 10}, {20, 10}}]}}, {
                    Directive[
                    EdgeForm[
                    Directive[
                    Opacity[0.3], 
                    GrayLevel[0]]], 
                    PointSize[0.5], 
                    Opacity[1.], 
                    RGBColor[0.560181, 0.691569, 0.194885], 
                    AbsoluteThickness[1.6]], {}}}, AspectRatio -> Full, 
                    ImageSize -> {20, 10}, PlotRangePadding -> None, 
                    ImagePadding -> Automatic, 
                    BaselinePosition -> (Scaled[0.1] -> Baseline)], #3}}, 
                 GridBoxAlignment -> {
                  "Columns" -> {Center, Left}, "Rows" -> {{Baseline}}}, 
                 AutoDelete -> False, 
                 GridBoxDividers -> {
                  "Columns" -> {{False}}, "Rows" -> {{False}}}, 
                 GridBoxItemSize -> {"Columns" -> {{All}}, "Rows" -> {{All}}},
                  GridBoxSpacings -> {
                  "Columns" -> {{0.5}}, "Rows" -> {{0.8}}}], "Grid"]}}, 
             GridBoxAlignment -> {"Columns" -> {{Left}}, "Rows" -> {{Top}}}, 
             AutoDelete -> False, 
             GridBoxItemSize -> {
              "Columns" -> {{Automatic}}, "Rows" -> {{Automatic}}}, 
             GridBoxSpacings -> {"Columns" -> {{1}}, "Rows" -> {{0}}}], 
            "Grid"], Alignment -> Left, AppearanceElements -> None, 
           ImageMargins -> {{5, 5}, {5, 5}}, ImageSizeAction -> 
           "ResizeToFit"], LineIndent -> 0, StripOnInput -> False], {
         FontFamily -> "Arial"}, Background -> Automatic, StripOnInput -> 
         False], TraditionalForm]& ), 
      InterpretationFunction :> (RowBox[{"LineLegend", "[", 
         RowBox[{
           RowBox[{"{", 
             RowBox[{
               RowBox[{"Directive", "[", 
                 RowBox[{
                   RowBox[{"Opacity", "[", "1.`", "]"}], ",", 
                   InterpretationBox[
                    ButtonBox[
                    TooltipBox[
                    GraphicsBox[{{
                    GrayLevel[0], 
                    RectangleBox[{0, 0}]}, {
                    GrayLevel[0], 
                    RectangleBox[{1, -1}]}, {
                    RGBColor[0.368417, 0.506779, 0.709798], 
                    RectangleBox[{0, -1}, {2, 1}]}}, DefaultBaseStyle -> 
                    "ColorSwatchGraphics", AspectRatio -> 1, Frame -> True, 
                    FrameStyle -> 
                    RGBColor[
                    0.24561133333333335`, 0.3378526666666667, 
                    0.4731986666666667], FrameTicks -> None, PlotRangePadding -> 
                    None, ImageSize -> 
                    Dynamic[{
                    Automatic, 1.35 CurrentValue["FontCapHeight"]/
                    AbsoluteCurrentValue[Magnification]}]], 
                    StyleBox[
                    RowBox[{"RGBColor", "[", 
                    RowBox[{"0.368417`", ",", "0.506779`", ",", "0.709798`"}],
                     "]"}], NumberMarks -> False]], Appearance -> None, 
                    BaseStyle -> {}, BaselinePosition -> Baseline, 
                    DefaultBaseStyle -> {}, ButtonFunction :> 
                    With[{Typeset`box$ = EvaluationBox[]}, 
                    If[
                    Not[
                    AbsoluteCurrentValue["Deployed"]], 
                    SelectionMove[Typeset`box$, All, Expression]; 
                    FrontEnd`Private`$ColorSelectorInitialAlpha = 1; 
                    FrontEnd`Private`$ColorSelectorInitialColor = 
                    RGBColor[0.368417, 0.506779, 0.709798]; 
                    FrontEnd`Private`$ColorSelectorUseMakeBoxes = True; 
                    MathLink`CallFrontEnd[
                    FrontEnd`AttachCell[Typeset`box$, 
                    FrontEndResource["RGBColorValueSelector"], {
                    0, {Left, Bottom}}, {Left, Top}, 
                    "ClosingActions" -> {
                    "SelectionDeparture", "ParentChanged", 
                    "EvaluatorQuit"}]]]], BaseStyle -> Inherited, Evaluator -> 
                    Automatic, Method -> "Preemptive"], 
                    RGBColor[0.368417, 0.506779, 0.709798], Editable -> False,
                     Selectable -> False], ",", 
                   RowBox[{"AbsoluteThickness", "[", "1.6`", "]"}]}], "]"}], 
               ",", 
               RowBox[{"Directive", "[", 
                 RowBox[{
                   RowBox[{"Opacity", "[", "1.`", "]"}], ",", 
                   InterpretationBox[
                    ButtonBox[
                    TooltipBox[
                    GraphicsBox[{{
                    GrayLevel[0], 
                    RectangleBox[{0, 0}]}, {
                    GrayLevel[0], 
                    RectangleBox[{1, -1}]}, {
                    RGBColor[0.880722, 0.611041, 0.142051], 
                    RectangleBox[{0, -1}, {2, 1}]}}, DefaultBaseStyle -> 
                    "ColorSwatchGraphics", AspectRatio -> 1, Frame -> True, 
                    FrameStyle -> 
                    RGBColor[
                    0.587148, 0.40736066666666665`, 0.09470066666666668], 
                    FrameTicks -> None, PlotRangePadding -> None, ImageSize -> 
                    Dynamic[{
                    Automatic, 1.35 CurrentValue["FontCapHeight"]/
                    AbsoluteCurrentValue[Magnification]}]], 
                    StyleBox[
                    RowBox[{"RGBColor", "[", 
                    RowBox[{"0.880722`", ",", "0.611041`", ",", "0.142051`"}],
                     "]"}], NumberMarks -> False]], Appearance -> None, 
                    BaseStyle -> {}, BaselinePosition -> Baseline, 
                    DefaultBaseStyle -> {}, ButtonFunction :> 
                    With[{Typeset`box$ = EvaluationBox[]}, 
                    If[
                    Not[
                    AbsoluteCurrentValue["Deployed"]], 
                    SelectionMove[Typeset`box$, All, Expression]; 
                    FrontEnd`Private`$ColorSelectorInitialAlpha = 1; 
                    FrontEnd`Private`$ColorSelectorInitialColor = 
                    RGBColor[0.880722, 0.611041, 0.142051]; 
                    FrontEnd`Private`$ColorSelectorUseMakeBoxes = True; 
                    MathLink`CallFrontEnd[
                    FrontEnd`AttachCell[Typeset`box$, 
                    FrontEndResource["RGBColorValueSelector"], {
                    0, {Left, Bottom}}, {Left, Top}, 
                    "ClosingActions" -> {
                    "SelectionDeparture", "ParentChanged", 
                    "EvaluatorQuit"}]]]], BaseStyle -> Inherited, Evaluator -> 
                    Automatic, Method -> "Preemptive"], 
                    RGBColor[0.880722, 0.611041, 0.142051], Editable -> False,
                     Selectable -> False], ",", 
                   RowBox[{"AbsoluteThickness", "[", "1.6`", "]"}]}], "]"}], 
               ",", 
               RowBox[{"Directive", "[", 
                 RowBox[{
                   RowBox[{"Opacity", "[", "1.`", "]"}], ",", 
                   InterpretationBox[
                    ButtonBox[
                    TooltipBox[
                    GraphicsBox[{{
                    GrayLevel[0], 
                    RectangleBox[{0, 0}]}, {
                    GrayLevel[0], 
                    RectangleBox[{1, -1}]}, {
                    RGBColor[0.560181, 0.691569, 0.194885], 
                    RectangleBox[{0, -1}, {2, 1}]}}, DefaultBaseStyle -> 
                    "ColorSwatchGraphics", AspectRatio -> 1, Frame -> True, 
                    FrameStyle -> 
                    RGBColor[
                    0.37345400000000006`, 0.461046, 0.12992333333333334`], 
                    FrameTicks -> None, PlotRangePadding -> None, ImageSize -> 
                    Dynamic[{
                    Automatic, 1.35 CurrentValue["FontCapHeight"]/
                    AbsoluteCurrentValue[Magnification]}]], 
                    StyleBox[
                    RowBox[{"RGBColor", "[", 
                    RowBox[{"0.560181`", ",", "0.691569`", ",", "0.194885`"}],
                     "]"}], NumberMarks -> False]], Appearance -> None, 
                    BaseStyle -> {}, BaselinePosition -> Baseline, 
                    DefaultBaseStyle -> {}, ButtonFunction :> 
                    With[{Typeset`box$ = EvaluationBox[]}, 
                    If[
                    Not[
                    AbsoluteCurrentValue["Deployed"]], 
                    SelectionMove[Typeset`box$, All, Expression]; 
                    FrontEnd`Private`$ColorSelectorInitialAlpha = 1; 
                    FrontEnd`Private`$ColorSelectorInitialColor = 
                    RGBColor[0.560181, 0.691569, 0.194885]; 
                    FrontEnd`Private`$ColorSelectorUseMakeBoxes = True; 
                    MathLink`CallFrontEnd[
                    FrontEnd`AttachCell[Typeset`box$, 
                    FrontEndResource["RGBColorValueSelector"], {
                    0, {Left, Bottom}}, {Left, Top}, 
                    "ClosingActions" -> {
                    "SelectionDeparture", "ParentChanged", 
                    "EvaluatorQuit"}]]]], BaseStyle -> Inherited, Evaluator -> 
                    Automatic, Method -> "Preemptive"], 
                    RGBColor[0.560181, 0.691569, 0.194885], Editable -> False,
                     Selectable -> False], ",", 
                   RowBox[{"AbsoluteThickness", "[", "1.6`", "]"}]}], "]"}]}],
              "}"}], ",", 
           RowBox[{"{", 
             RowBox[{#, ",", #2, ",", #3}], "}"}], ",", 
           RowBox[{"LegendMarkers", "\[Rule]", "None"}], ",", 
           RowBox[{"LabelStyle", "\[Rule]", 
             RowBox[{"{", "}"}]}], ",", 
           RowBox[{"LegendLayout", "\[Rule]", "\"Column\""}]}], "]"}]& ), 
      Editable -> True], TraditionalForm], TraditionalForm]},
  "Legended",
  DisplayFunction->(GridBox[{{
      TagBox[
       ItemBox[
        PaneBox[
         TagBox[#, "SkipImageSizeLevel"], Alignment -> {Center, Baseline}, 
         BaselinePosition -> Baseline], DefaultBaseStyle -> "Labeled"], 
       "SkipImageSizeLevel"], 
      ItemBox[#2, DefaultBaseStyle -> "LabeledLabel"]}}, 
    GridBoxAlignment -> {"Columns" -> {{Center}}, "Rows" -> {{Center}}}, 
    AutoDelete -> False, GridBoxItemSize -> Automatic, 
    BaselinePosition -> {1, 1}]& ),
  Editable->True,
  InterpretationFunction->(RowBox[{"Legended", "[", 
     RowBox[{#, ",", 
       RowBox[{"Placed", "[", 
         RowBox[{#2, ",", "After"}], "]"}]}], "]"}]& )]], "Output",
 CellChangeTimes->{{3.789732862523242*^9, 3.7897328683434677`*^9}, {
   3.789732901600594*^9, 3.789732906917636*^9}, {3.789733026250841*^9, 
   3.789733042026977*^9}, 3.789733080539659*^9, {3.78973322157124*^9, 
   3.789733239211602*^9}, {3.7897334440408583`*^9, 3.7897334533115797`*^9}, 
   3.789733506292223*^9, 3.789734584216867*^9, {3.79007861463617*^9, 
   3.7900787247000027`*^9}, 3.790078762784197*^9, 3.7900844394608393`*^9, 
   3.795174683469287*^9, 3.800008967069524*^9, 3.80000903406283*^9, {
   3.800009214155257*^9, 3.800009229717184*^9}, {3.800009263015492*^9, 
   3.8000092848896008`*^9}, {3.800009321365273*^9, 3.8000093348560658`*^9}, 
   3.8002682059851093`*^9, {3.813930831728199*^9, 3.8139308414845877`*^9}, 
   3.8140817894275427`*^9, 3.8141894022019873`*^9, 3.81419596621653*^9, 
   3.81419651533729*^9},
 CellLabel->
  "Out[1351]=",ExpressionUUID->"7e6aa397-65c5-41f0-ac8c-0258435f0c59"]
}, Open  ]]
}, Closed]]
}, Closed]],

Cell[CellGroupData[{

Cell[TextData[{
 "Allele Frequency and Population Size Dynamics - females mate with two males \
(",
 Cell[BoxData[
  FormBox[
   RowBox[{
    SubscriptBox["\[Lambda]", "f"], "=", "2"}], TraditionalForm]],
  FormatType->"TraditionalForm",ExpressionUUID->
  "e69cc557-ca21-40f2-b4fd-ff1622f9eaea"],
 ")"
}], "Section",
 CellChangeTimes->{{3.781463311644246*^9, 3.781463316031954*^9}, {
  3.81401133197186*^9, 3.814011335758626*^9}, {3.814018589445348*^9, 
  3.814018627458487*^9}},ExpressionUUID->"a3d9aa4f-4097-4134-9b5c-\
9a4fe1d5ffb9"],

Cell[CellGroupData[{

Cell["Recursion equations", "Subsection",
 CellChangeTimes->{{3.781874873824012*^9, 
  3.781874878613985*^9}},ExpressionUUID->"d6a593b5-fc9a-4ff0-b568-\
77f6abb26a87"],

Cell["\<\
In this section, each female mates with two males. The probability of \
choosing each male genotype is assumed to be random such that it occurs in \
proportion to the proportion of males with each genotype. These \
recursion/difference equations track the mating combinations of female and \
two male genotypes that can produce offspring of each genotype. 
\
\>", "Text",
 CellChangeTimes->{{3.814082740615233*^9, 
  3.81408288606927*^9}},ExpressionUUID->"69f9f8a0-3224-4ce1-95a7-\
20e790020c64"],

Cell[BoxData[{
 RowBox[{
  RowBox[{"DDfjuveniles2", " ", "=", 
   RowBox[{"b", "*", 
    RowBox[{"(", 
     RowBox[{"1", "-", 
      RowBox[{"\[Alpha]", " ", "totAdults"}]}], ")"}], 
    RowBox[{"(", 
     RowBox[{
      RowBox[{"DDf", 
       RowBox[{"(", 
        RowBox[{
         FractionBox[
          RowBox[{" ", 
           RowBox[{"Dm", "^", "2"}]}], 
          RowBox[{"2", 
           RowBox[{
            RowBox[{"(", 
             RowBox[{"Dm", " ", "+", " ", "dm"}], ")"}], "^", "2"}]}]], " ", 
         "+", 
         RowBox[{"2", "  ", 
          FractionBox[
           RowBox[{"Dm", " ", "dm"}], 
           RowBox[{
            RowBox[{"(", 
             RowBox[{"Dm", " ", "+", " ", "dm"}], ")"}], "^", "2"}]], 
          RowBox[{"(", 
           FractionBox[
            FractionBox["1", "2"], 
            RowBox[{"(", 
             RowBox[{"1", "+", "c"}], ")"}]], ")"}]}]}], ")"}]}], "+", 
      "\[IndentingNewLine]", 
      RowBox[{
       FractionBox["Ddf", "2"], 
       RowBox[{"(", 
        RowBox[{
         RowBox[{
          FractionBox[
           RowBox[{" ", 
            RowBox[{"Dm", "^", "2"}]}], 
           RowBox[{
            RowBox[{"(", 
             RowBox[{"Dm", " ", "+", " ", "dm"}], ")"}], "^", "2"}]], " ", 
          RowBox[{"(", 
           FractionBox["1", "2"], ")"}]}], "+", 
         RowBox[{
          FractionBox[
           RowBox[{"2", "Dm", " ", "dm"}], 
           RowBox[{
            RowBox[{"(", 
             RowBox[{"Dm", " ", "+", " ", "dm"}], ")"}], "^", "2"}]], " ", 
          RowBox[{"(", 
           FractionBox[
            FractionBox["1", "2"], 
            RowBox[{"(", 
             RowBox[{"1", "+", "c"}], ")"}]], ")"}]}]}], ")"}]}]}], ")"}]}]}],
   " ", ";"}], "\[IndentingNewLine]", 
 RowBox[{
  RowBox[{
   RowBox[{"Ddfjuveniles2", " ", "=", 
    RowBox[{"b", "*", 
     RowBox[{"(", 
      RowBox[{"1", "-", 
       RowBox[{"\[Alpha]", " ", "totAdults"}]}], ")"}], 
     RowBox[{"(", 
      RowBox[{
       RowBox[{"DDf", 
        RowBox[{"(", 
         RowBox[{
          RowBox[{
           FractionBox[
            RowBox[{"dm", "^", "2"}], 
            RowBox[{
             RowBox[{"(", 
              RowBox[{"Dm", "+", "dm"}], ")"}], "^", "2"}]], 
           RowBox[{"(", 
            FractionBox[
             RowBox[{"(", 
              RowBox[{"1", "+", "\[Delta]"}], ")"}], "2"], ")"}]}], " ", "+", 
          RowBox[{
           FractionBox[
            RowBox[{" ", 
             RowBox[{"2", " ", "dm", " ", "Dm"}]}], 
            RowBox[{
             RowBox[{"(", 
              RowBox[{"Dm", "+", "dm"}], ")"}], "^", "2"}]], " ", 
           RowBox[{"(", 
            FractionBox[
             RowBox[{"c", 
              RowBox[{"(", 
               FractionBox[
                RowBox[{"1", "+", "\[Delta]"}], "2"], ")"}]}], 
             RowBox[{"(", 
              RowBox[{"1", "+", "c"}], ")"}]], ")"}]}]}], ")"}]}], " ", "+", 
       "\[IndentingNewLine]", 
       RowBox[{
        FractionBox["Ddf", "2"], 
        RowBox[{"(", 
         RowBox[{
          RowBox[{
           FractionBox[
            RowBox[{"Dm", "^", "2"}], 
            RowBox[{
             RowBox[{"(", 
              RowBox[{"Dm", "+", "dm"}], ")"}], "^", "2"}]], " ", 
           RowBox[{"(", 
            FractionBox[
             RowBox[{" ", "1"}], "2"], " ", ")"}]}], "+", "  ", 
          RowBox[{
           FractionBox[
            RowBox[{"dm", "^", "2"}], 
            RowBox[{
             RowBox[{"(", 
              RowBox[{"Dm", "+", "dm"}], ")"}], "^", "2"}]], 
           RowBox[{"(", 
            FractionBox[
             RowBox[{"(", 
              RowBox[{"1", "+", "\[Delta]"}], ")"}], "2"], ")"}]}], " ", "+", 
          RowBox[{
           FractionBox[
            RowBox[{"2", "  ", "Dm", " ", "dm"}], 
            RowBox[{
             RowBox[{"(", 
              RowBox[{"Dm", "+", "dm"}], ")"}], "^", "2"}]], " ", 
           RowBox[{"(", 
            RowBox[{"(", 
             FractionBox[
              RowBox[{
               FractionBox["1", "2"], "+", 
               RowBox[{"c", 
                RowBox[{"(", 
                 FractionBox[
                  RowBox[{"1", "+", "\[Delta]"}], "2"], ")"}]}]}], 
              RowBox[{"(", 
               RowBox[{"1", "+", "c"}], ")"}]], ")"}], ")"}]}]}], " ", 
         ")"}]}], "+", "\[IndentingNewLine]", 
       RowBox[{"ddf", 
        RowBox[{"(", 
         RowBox[{
          RowBox[{
           FractionBox[
            RowBox[{" ", 
             RowBox[{"Dm", "^", "2"}]}], 
            RowBox[{
             RowBox[{"(", 
              RowBox[{"Dm", "+", "dm"}], ")"}], "^", "2"}]], " ", 
           FractionBox["1", "2"]}], " ", "+", " ", 
          RowBox[{
           FractionBox[
            RowBox[{"2", " ", "Dm", " ", "dm"}], 
            RowBox[{
             RowBox[{"(", 
              RowBox[{"Dm", "+", "dm"}], ")"}], "^", "2"}]], " ", 
           FractionBox[
            FractionBox["1", "2"], 
            RowBox[{"(", 
             RowBox[{"1", "+", "c"}], ")"}]]}]}], ")"}]}]}], ")"}]}]}], ";"}],
   " "}], "\[IndentingNewLine]", 
 RowBox[{
  RowBox[{"ddfjuveniles2", "=", 
   RowBox[{"b", "*", 
    RowBox[{"(", 
     RowBox[{"1", "-", 
      RowBox[{"\[Alpha]", " ", "totAdults"}]}], ")"}], 
    RowBox[{"(", 
     RowBox[{
      RowBox[{
       FractionBox["Ddf", "2"], 
       RowBox[{"(", 
        RowBox[{
         RowBox[{
          FractionBox[
           RowBox[{"dm", "^", "2"}], 
           RowBox[{
            RowBox[{"(", 
             RowBox[{"Dm", "+", "dm"}], ")"}], "^", "2"}]], " ", 
          RowBox[{"(", 
           FractionBox[
            RowBox[{"(", 
             RowBox[{"1", "+", "\[Delta]"}], ")"}], "2"], ")"}]}], " ", "+", 
         " ", 
         RowBox[{
          FractionBox[
           RowBox[{"2", "Dm", " ", "dm"}], 
           RowBox[{
            RowBox[{"(", 
             RowBox[{"Dm", "+", "dm"}], ")"}], "^", "2"}]], " ", 
          RowBox[{"(", 
           FractionBox[
            RowBox[{"c", 
             RowBox[{"(", 
              FractionBox[
               RowBox[{"1", "+", "\[Delta]"}], "2"], ")"}]}], 
            RowBox[{
             RowBox[{"(", 
              RowBox[{"1", "+", "c"}], ")"}], " "}]], ")"}]}]}], ")"}]}], " ",
       "+", "\[IndentingNewLine]", 
      RowBox[{"ddf", 
       RowBox[{"(", 
        RowBox[{
         RowBox[{
          FractionBox[
           RowBox[{"dm", "^", "2"}], 
           RowBox[{
            RowBox[{"(", 
             RowBox[{"Dm", "+", "dm"}], ")"}], "^", "2"}]], " ", 
          RowBox[{"(", 
           FractionBox[
            RowBox[{"(", 
             RowBox[{"1", "+", "\[Delta]"}], ")"}], "2"], ")"}]}], " ", "+", 
         RowBox[{
          FractionBox[
           RowBox[{"2", " ", "Dm", " ", "dm"}], 
           RowBox[{
            RowBox[{"(", 
             RowBox[{"Dm", "+", "dm"}], ")"}], "^", "2"}]], " ", 
          RowBox[{"(", 
           FractionBox[
            RowBox[{"c", 
             RowBox[{"(", 
              FractionBox[
               RowBox[{"1", "+", "\[Delta]"}], "2"], ")"}]}], 
            RowBox[{"(", 
             RowBox[{"1", "+", "c"}], ")"}]], ")"}]}]}], ")"}]}]}], ")"}]}]}],
   ";"}], "\[IndentingNewLine]", 
 RowBox[{
  RowBox[{"Dmjuveniles2", " ", "=", 
   RowBox[{"b", "*", 
    RowBox[{"(", 
     RowBox[{"1", "-", 
      RowBox[{"\[Alpha]", " ", "totAdults"}]}], ")"}], 
    RowBox[{"(", 
     RowBox[{
      RowBox[{"DDf", 
       RowBox[{"(", 
        RowBox[{
         RowBox[{
          FractionBox[
           RowBox[{"Dm", "^", "2"}], 
           RowBox[{
            RowBox[{"(", 
             RowBox[{"Dm", "+", "dm"}], ")"}], "^", "2"}]], " ", 
          FractionBox[
           RowBox[{" ", "1"}], "2"]}], " ", "+", " ", 
         RowBox[{
          FractionBox[
           RowBox[{"dm", "^", "2"}], 
           RowBox[{
            RowBox[{"(", 
             RowBox[{"Dm", "+", "dm"}], ")"}], "^", "2"}]], " ", 
          RowBox[{"(", 
           FractionBox[
            RowBox[{"(", 
             RowBox[{"1", "-", "\[Delta]"}], ")"}], "2"], ")"}]}], " ", "+", 
         " ", 
         RowBox[{
          FractionBox[
           RowBox[{"2", " ", "Dm", " ", "dm"}], 
           RowBox[{
            RowBox[{"(", 
             RowBox[{"Dm", "+", "dm"}], ")"}], "^", "2"}]], " ", 
          RowBox[{"(", 
           FractionBox[
            RowBox[{
             FractionBox["1", "2"], "+", 
             RowBox[{"c", 
              RowBox[{"(", " ", 
               FractionBox[
                RowBox[{"1", "-", "\[Delta]"}], "2"], ")"}]}]}], 
            RowBox[{"(", 
             RowBox[{"1", "+", "c"}], ")"}]], ")"}]}]}], ")"}]}], " ", "+", 
      "\[IndentingNewLine]", 
      RowBox[{
       FractionBox["Ddf", "2"], 
       RowBox[{"(", 
        RowBox[{
         RowBox[{
          FractionBox[
           RowBox[{"Dm", "^", "2"}], 
           RowBox[{
            RowBox[{"(", 
             RowBox[{"Dm", "+", "dm"}], ")"}], "^", "2"}]], 
          RowBox[{"(", 
           FractionBox[
            RowBox[{" ", "1"}], "2"], ")"}]}], " ", "+", 
         RowBox[{
          FractionBox[
           RowBox[{"dm", "^", "2"}], 
           RowBox[{
            RowBox[{"(", 
             RowBox[{"Dm", "+", "dm"}], ")"}], "^", "2"}]], 
          RowBox[{"(", " ", 
           RowBox[{"(", 
            FractionBox[
             RowBox[{"(", 
              RowBox[{"1", "-", "\[Delta]"}], ")"}], "2"], ")"}], " ", 
           ")"}]}], "+", " ", 
         RowBox[{
          FractionBox[
           RowBox[{"2", " ", "Dm", " ", "dm"}], 
           RowBox[{
            RowBox[{"(", 
             RowBox[{"Dm", "+", "dm"}], ")"}], "^", "2"}]], " ", 
          RowBox[{"(", 
           FractionBox[
            RowBox[{
             FractionBox["1", "2"], "+", 
             RowBox[{"c", 
              RowBox[{"(", 
               FractionBox[
                RowBox[{"1", "-", "\[Delta]"}], "2"], ")"}]}]}], 
            RowBox[{"(", 
             RowBox[{"1", "+", "c"}], ")"}]], ")"}]}]}], ")"}]}]}], ")"}]}]}],
   ";"}], "\[IndentingNewLine]", 
 RowBox[{
  RowBox[{"dmjuveniles2", " ", "=", 
   RowBox[{"b", "*", 
    RowBox[{"(", 
     RowBox[{"1", "-", 
      RowBox[{"\[Alpha]", " ", "totAdults"}]}], ")"}], 
    RowBox[{"(", 
     RowBox[{
      RowBox[{
       FractionBox["Ddf", "2"], 
       RowBox[{"(", 
        RowBox[{
         RowBox[{
          FractionBox[
           RowBox[{"Dm", "^", "2"}], 
           RowBox[{
            RowBox[{"(", 
             RowBox[{"Dm", "+", "dm"}], ")"}], "^", "2"}]], 
          RowBox[{"(", 
           FractionBox[
            RowBox[{" ", "1"}], "2"], ")"}]}], " ", "+", 
         RowBox[{
          FractionBox[
           RowBox[{"dm", "^", "2"}], 
           RowBox[{
            RowBox[{"(", 
             RowBox[{"Dm", "+", "dm"}], ")"}], "^", "2"}]], 
          RowBox[{"(", 
           FractionBox[
            RowBox[{"(", 
             RowBox[{"1", "-", "\[Delta]"}], ")"}], "2"], ")"}]}], " ", "+", 
         " ", 
         RowBox[{
          FractionBox[
           RowBox[{"2", " ", "Dm", " ", "dm", " "}], 
           RowBox[{
            RowBox[{"(", 
             RowBox[{"Dm", "+", "dm"}], ")"}], "^", "2"}]], 
          RowBox[{"(", 
           FractionBox[
            RowBox[{
             FractionBox["1", "2"], "+", " ", 
             RowBox[{"c", 
              RowBox[{"(", 
               FractionBox[
                RowBox[{"1", "-", "\[Delta]"}], "2"], ")"}]}]}], 
            RowBox[{"(", 
             RowBox[{"1", "+", "c"}], ")"}]], ")"}]}]}], " ", ")"}]}], "+", 
      " ", "\[IndentingNewLine]", 
      RowBox[{"ddf", 
       RowBox[{"(", 
        RowBox[{
         RowBox[{
          FractionBox[
           RowBox[{"Dm", "^", "2"}], 
           RowBox[{
            RowBox[{"(", 
             RowBox[{"Dm", "+", "dm"}], ")"}], "^", "2"}]], 
          RowBox[{"(", 
           FractionBox["1", "2"], ")"}]}], " ", "+", 
         RowBox[{
          FractionBox[
           RowBox[{"dm", "^", "2"}], 
           RowBox[{
            RowBox[{"(", 
             RowBox[{"Dm", "+", "dm"}], ")"}], "^", "2"}]], " ", 
          RowBox[{"(", 
           FractionBox[
            RowBox[{"(", 
             RowBox[{"1", "-", "\[Delta]"}], ")"}], "2"], ")"}]}], " ", "+", 
         RowBox[{
          FractionBox[
           RowBox[{"2", " ", "Dm", " ", "dm", " "}], 
           RowBox[{
            RowBox[{"(", 
             RowBox[{"Dm", "+", "dm"}], ")"}], "^", "2"}]], 
          RowBox[{"(", 
           FractionBox[
            RowBox[{
             FractionBox["1", "2"], "+", 
             RowBox[{"c", 
              RowBox[{"(", " ", 
               FractionBox[
                RowBox[{"1", "-", "\[Delta]"}], "2"], ")"}]}]}], 
            RowBox[{"(", 
             RowBox[{"1", "+", "c"}], ")"}]], ")"}]}]}], " ", ")"}]}]}], 
     ")"}]}]}], ";"}]}], "Input",
 CellChangeTimes->{{3.781874918016214*^9, 3.781874951890514*^9}, {
   3.781875698598501*^9, 3.78187575402771*^9}, {3.781883716099955*^9, 
   3.781883758543943*^9}, 3.781961176952222*^9, {3.814006553046831*^9, 
   3.814006553430848*^9}, 3.814007837944531*^9, {3.814008284582357*^9, 
   3.814008332361627*^9}, 3.814075935343079*^9, {3.814075980488933*^9, 
   3.8140759804948893`*^9}, {3.814076984817071*^9, 3.814077013953087*^9}, {
   3.814077044964316*^9, 3.814077073009499*^9}, {3.814078523321558*^9, 
   3.814078524018896*^9}, {3.8140785664193373`*^9, 3.8140785750479307`*^9}, {
   3.814082704904151*^9, 3.8140827243784637`*^9}, {3.814082895233481*^9, 
   3.8140833100983257`*^9}, {3.814083353027583*^9, 3.814083462625531*^9}, {
   3.8140836262247887`*^9, 3.814083667303977*^9}, {3.814083708175639*^9, 
   3.814083733106092*^9}, {3.8140838005147743`*^9, 3.814083961307704*^9}, {
   3.814084240368881*^9, 3.814084314710822*^9}, {3.814084514824009*^9, 
   3.814084517145873*^9}, {3.814084581274603*^9, 3.814084593241027*^9}, {
   3.814084630355673*^9, 3.814084636508759*^9}},
 CellLabel->
  "In[1352]:=",ExpressionUUID->"49a453be-289a-46d8-a5bb-2cff568a2eac"],

Cell["\<\
Before the next generation of adults, viability selection acts according to \
genotype. \
\>", "Text",
 CellChangeTimes->{
  3.814084633192759*^9},ExpressionUUID->"84e39e22-f4cc-4793-a45a-\
e8a087e71a1e"],

Cell[BoxData[{
 RowBox[{
  RowBox[{"DDfprime2", "=", 
   RowBox[{"DDfjuveniles2", "*", " ", "1"}]}], ";"}], "\[IndentingNewLine]", 
 RowBox[{
  RowBox[{"Ddfprime2", "=", 
   RowBox[{"Ddfjuveniles2", "*", 
    RowBox[{"(", 
     RowBox[{"1", " ", "-", " ", 
      RowBox[{"h", "*", "sf"}]}], ")"}]}]}], ";"}], "\[IndentingNewLine]", 
 RowBox[{
  RowBox[{"ddfprime2", "=", 
   RowBox[{"ddfjuveniles2", "*", 
    RowBox[{"(", 
     RowBox[{"1", " ", "-", " ", "sf"}], ")"}]}]}], 
  ";"}], "\[IndentingNewLine]", 
 RowBox[{
  RowBox[{"Dmprime2", "=", 
   RowBox[{"Dmjuveniles2", "*", "1"}]}], ";"}], "\[IndentingNewLine]", 
 RowBox[{
  RowBox[{"dmprime2", "=", 
   RowBox[{"dmjuveniles2", "*", 
    RowBox[{"(", 
     RowBox[{"1", " ", "-", " ", "sm"}], ")"}]}]}], ";"}]}], "Input",
 CellLabel->
  "In[1357]:=",ExpressionUUID->"c9182c0f-7901-46d8-8bbb-a2f0e2c978ec"]
}, Closed]],

Cell[CellGroupData[{

Cell["Equilibrium allele frequency and population size", "Subsection",
 CellChangeTimes->{{3.7818749850842857`*^9, 
  3.781874995173348*^9}},ExpressionUUID->"8c3a6cc1-cc9d-4c72-b273-\
ef5ed5c3ee86"],

Cell["\<\
The eco-evolutionary dynamics are at equilibrium when the difference \
equations (solvemeFull) are equal to 0 (no change). We use subeq to convert \
the difference equations so that they track the genotype frequency in males \
(pm), females (pf), inbreeding coefficient (F), and the number of males \
(maleNum) and females (femaleNum). \
\>", "Text",
 CellChangeTimes->{{3.814007613282021*^9, 3.814007656119938*^9}, {
  3.814007994963559*^9, 3.8140080779300632`*^9}, {3.814008624908166*^9, 
  3.814008629557644*^9}},ExpressionUUID->"3fa6a303-336e-49b9-ba10-\
d0f9b7c8282d"],

Cell[BoxData[
 RowBox[{
  RowBox[{"solvemeFull", "=", 
   RowBox[{
    RowBox[{
     RowBox[{"{", 
      RowBox[{
       RowBox[{"DDfprime2", "-", "DDf"}], ",", 
       RowBox[{"Ddfprime2", "-", "Ddf"}], ",", 
       RowBox[{"ddfprime2", "-", "ddf"}], ",", " ", 
       RowBox[{"Dmprime2", "-", "Dm"}], ",", " ", 
       RowBox[{"dmprime2", "-", "dm"}]}], "}"}], "/.", 
     RowBox[{"subeq", "[", 
      RowBox[{"[", "1", "]"}], "]"}]}], "//", "Simplify"}]}], ";"}]], "Input",
 CellChangeTimes->{
  3.814007660711735*^9, {3.814007962027532*^9, 3.814007980946431*^9}, {
   3.814008499992148*^9, 3.8140085169206333`*^9}},
 CellLabel->
  "In[1362]:=",ExpressionUUID->"797883cf-eaed-4f58-8766-21be4be1bc4a"],

Cell["\<\
When drive is absent (pm=0, pf=0) or fixed (pm=1, pf=1), the allele \
frequencies cannot change. We can get the equilibrium population size \
(maleNum and femaleNum) for these evolutionary equilibria. \
\>", "Text",
 CellChangeTimes->{{3.814008155520987*^9, 3.8140082177096167`*^9}, {
  3.814008646664893*^9, 
  3.8140086479141006`*^9}},ExpressionUUID->"1f0fa615-a1a8-4a49-8f28-\
362abf728df4"],

Cell[CellGroupData[{

Cell[BoxData[{
 RowBox[{"noDrivePopSizeSol2", "=", 
  RowBox[{"Solve", "[", 
   RowBox[{
    RowBox[{
     RowBox[{"(", 
      RowBox[{
       RowBox[{"solvemeFull", "/.", 
        RowBox[{"pm", "\[Rule]", "0"}]}], "/.", 
       RowBox[{"pf", "\[Rule]", "0"}]}], ")"}], "\[Equal]", 
     RowBox[{"{", 
      RowBox[{"0", ",", "0", ",", "0", ",", "0", ",", "0"}], "}"}]}], ",", 
    RowBox[{"{", 
     RowBox[{"maleNum", ",", "femaleNum"}], "}"}]}], 
   "]"}]}], "\[IndentingNewLine]", 
 RowBox[{"fixedDrivePopSizeSol2", " ", "=", " ", 
  RowBox[{"Solve", "[", 
   RowBox[{
    RowBox[{
     RowBox[{"(", 
      RowBox[{
       RowBox[{"solvemeFull", "/.", 
        RowBox[{"pm", "\[Rule]", "1"}]}], "/.", 
       RowBox[{"pf", "\[Rule]", "1"}]}], ")"}], "\[Equal]", 
     RowBox[{"{", 
      RowBox[{"0", ",", "0", ",", "0", ",", "0", ",", "0"}], "}"}]}], ",", 
    RowBox[{"{", 
     RowBox[{"maleNum", ",", "femaleNum"}], "}"}]}], "]"}]}]}], "Input",
 CellChangeTimes->{{3.781883653232218*^9, 3.781883685637601*^9}, {
   3.781950010576271*^9, 3.7819502188531847`*^9}, {3.78195028325841*^9, 
   3.781950300097888*^9}, {3.78195045726005*^9, 3.781950472913829*^9}, {
   3.781961389029669*^9, 3.7819614047128553`*^9}, 3.781961746977097*^9, {
   3.800012174367577*^9, 3.800012179664307*^9}, {3.814005814117457*^9, 
   3.8140058327809057`*^9}, {3.814006629646102*^9, 3.8140066347172003`*^9}, {
   3.814007663064589*^9, 3.814007663894669*^9}, {3.81400855941399*^9, 
   3.814008559838789*^9}},
 CellLabel->
  "In[1363]:=",ExpressionUUID->"2fd847cc-8c70-4b60-852d-932e9004fbd7"],

Cell[BoxData[
 RowBox[{"{", 
  RowBox[{
   RowBox[{"{", 
    RowBox[{
     RowBox[{"maleNum", "\[Rule]", "0"}], ",", 
     RowBox[{"femaleNum", "\[Rule]", "0"}]}], "}"}], ",", 
   RowBox[{"{", 
    RowBox[{
     RowBox[{"maleNum", "\[Rule]", 
      FractionBox[
       RowBox[{
        RowBox[{"-", "2"}], "+", "b"}], 
       RowBox[{"2", " ", "b", " ", "\[Alpha]"}]]}], ",", 
     RowBox[{"femaleNum", "\[Rule]", 
      FractionBox[
       RowBox[{
        RowBox[{"-", "2"}], "+", "b"}], 
       RowBox[{"2", " ", "b", " ", "\[Alpha]"}]]}]}], "}"}]}], 
  "}"}]], "Output",
 CellChangeTimes->{
  3.800268331335506*^9, 3.81400652721336*^9, {3.814006618732967*^9, 
   3.814006635123662*^9}, 3.814007114055028*^9, 3.814007664388343*^9, 
   3.81400798381371*^9, 3.81400821956627*^9, {3.814008512705503*^9, 
   3.814008519393764*^9}, 3.8140091267207317`*^9, {3.814009414364393*^9, 
   3.8140094177191896`*^9}, 3.814084653233059*^9, 3.814196515663265*^9},
 CellLabel->
  "Out[1363]=",ExpressionUUID->"112ee71e-507d-4a44-92c1-285475b460fc"],

Cell[BoxData[
 RowBox[{"{", 
  RowBox[{
   RowBox[{"{", 
    RowBox[{
     RowBox[{"maleNum", "\[Rule]", "0"}], ",", 
     RowBox[{"femaleNum", "\[Rule]", "0"}]}], "}"}], ",", 
   RowBox[{"{", 
    RowBox[{
     RowBox[{"maleNum", "\[Rule]", 
      FractionBox[
       RowBox[{
        RowBox[{"(", 
         RowBox[{
          RowBox[{"-", "1"}], "+", "sm"}], ")"}], " ", 
        RowBox[{"(", 
         RowBox[{
          RowBox[{"-", "1"}], "+", "\[Delta]"}], ")"}], " ", 
        RowBox[{"(", 
         RowBox[{"2", "-", "b", "+", 
          RowBox[{"b", " ", "sf"}], "-", 
          RowBox[{"b", " ", "\[Delta]"}], "+", 
          RowBox[{"b", " ", "sf", " ", "\[Delta]"}]}], ")"}]}], 
       RowBox[{"b", " ", 
        RowBox[{"(", 
         RowBox[{
          RowBox[{"-", "1"}], "+", "sf"}], ")"}], " ", "\[Alpha]", " ", 
        RowBox[{"(", 
         RowBox[{"1", "+", "\[Delta]"}], ")"}], " ", 
        RowBox[{"(", 
         RowBox[{"2", "-", "sf", "-", "sm", "-", 
          RowBox[{"sf", " ", "\[Delta]"}], "+", 
          RowBox[{"sm", " ", "\[Delta]"}]}], ")"}]}]]}], ",", 
     RowBox[{"femaleNum", "\[Rule]", 
      FractionBox[
       RowBox[{"2", "-", "b", "+", 
        RowBox[{"b", " ", "sf"}], "-", 
        RowBox[{"b", " ", "\[Delta]"}], "+", 
        RowBox[{"b", " ", "sf", " ", "\[Delta]"}]}], 
       RowBox[{"b", " ", "\[Alpha]", " ", 
        RowBox[{"(", 
         RowBox[{
          RowBox[{"-", "2"}], "+", "sf", "+", "sm", "+", 
          RowBox[{"sf", " ", "\[Delta]"}], "-", 
          RowBox[{"sm", " ", "\[Delta]"}]}], ")"}]}]]}]}], "}"}]}], 
  "}"}]], "Output",
 CellChangeTimes->{
  3.800268331335506*^9, 3.81400652721336*^9, {3.814006618732967*^9, 
   3.814006635123662*^9}, 3.814007114055028*^9, 3.814007664388343*^9, 
   3.81400798381371*^9, 3.81400821956627*^9, {3.814008512705503*^9, 
   3.814008519393764*^9}, 3.8140091267207317`*^9, {3.814009414364393*^9, 
   3.8140094177191896`*^9}, 3.814084653233059*^9, 3.814196515710176*^9},
 CellLabel->
  "Out[1364]=",ExpressionUUID->"1ff45ad7-bfc0-40a6-8796-741664783276"]
}, Open  ]],

Cell["\<\
We also expect drive can reach an intermediate equilibrium allele frequency. \
We make it easier to solve for this equilibrium allele frequency by focussing \
on the difference in allele frequency and allele frequency between \
generations. \
\>", "Text",
 CellChangeTimes->{{3.8140091610875177`*^9, 
  3.8140092442461643`*^9}},ExpressionUUID->"3ee33cf4-9caa-438a-91a8-\
e6c61ff9bfce"],

Cell[BoxData[
 RowBox[{
  RowBox[{"solveme", "=", 
   RowBox[{"Simplify", "[", 
    RowBox[{
     RowBox[{
      RowBox[{"{", 
       RowBox[{
        RowBox[{
         RowBox[{"(", 
          FractionBox["dmprime2", 
           RowBox[{"Dmprime2", "+", "dmprime2"}]], ")"}], "-", "pm"}], ",", 
        RowBox[{
         RowBox[{"(", 
          FractionBox[
           RowBox[{
            RowBox[{"2", " ", "ddfprime2"}], "+", "Ddfprime2"}], 
           RowBox[{"2", " ", 
            RowBox[{"(", 
             RowBox[{"ddfprime2", "+", "Ddfprime2", "+", "DDfprime2"}], 
             ")"}]}]], ")"}], "-", "pf"}], ",", 
        RowBox[{
         RowBox[{"(", 
          FractionBox[
           RowBox[{
            RowBox[{"4", " ", "ddfprime2", " ", "DDfprime2"}], " ", "-", 
            SuperscriptBox["Ddfprime2", "2"]}], 
           RowBox[{
            RowBox[{"(", 
             RowBox[{
              RowBox[{"2", " ", "ddfprime2"}], "+", "Ddfprime2"}], ")"}], " ", 
            RowBox[{"(", 
             RowBox[{"Ddfprime2", "+", 
              RowBox[{"2", " ", "DDfprime2"}]}], ")"}]}]], ")"}], "-", 
         "F"}]}], "}"}], "/.", 
      RowBox[{"{", 
       RowBox[{
        RowBox[{"pm", "\[Rule]", 
         FractionBox["dm", 
          RowBox[{"dm", "+", "Dm"}]]}], ",", 
        RowBox[{"pf", "\[Rule]", 
         FractionBox[
          RowBox[{
           RowBox[{"2", " ", "ddf"}], "+", "Ddf"}], 
          RowBox[{"2", " ", 
           RowBox[{"(", 
            RowBox[{"ddf", "+", "Ddf", "+", "DDf"}], ")"}]}]]}], ",", 
        RowBox[{"F", "\[Rule]", 
         FractionBox[
          RowBox[{
           RowBox[{"-", 
            SuperscriptBox["Ddf", "2"]}], "+", 
           RowBox[{"4", " ", "ddf", " ", "DDf"}]}], 
          RowBox[{
           RowBox[{"(", 
            RowBox[{
             RowBox[{"2", " ", "ddf"}], "+", "Ddf"}], ")"}], " ", 
           RowBox[{"(", 
            RowBox[{"Ddf", "+", 
             RowBox[{"2", " ", "DDf"}]}], ")"}]}]]}]}], "}"}]}], "/.", 
     "subeq"}], "]"}]}], ";"}]], "Input",
 CellChangeTimes->{{3.781875008764103*^9, 3.7818750765542803`*^9}, 
   3.781875124347249*^9, {3.781875300683763*^9, 3.781875369095936*^9}, 
   3.781875932900531*^9, {3.781877093213214*^9, 3.78187709654655*^9}, {
   3.781877573705399*^9, 3.781877573959036*^9}, {3.7818843060919533`*^9, 
   3.781884321527413*^9}, {3.8140070581388903`*^9, 3.814007079486866*^9}, {
   3.8140082567724743`*^9, 3.8140082720497427`*^9}, {3.814008531389463*^9, 
   3.8140085319173737`*^9}, {3.814009248476453*^9, 3.814009248825746*^9}},
 CellLabel->
  "In[1365]:=",ExpressionUUID->"1c43e52d-6fae-4d7c-8290-b58af3761bb2"],

Cell["\<\
However, we are only able to get a result for the polymorphic equilibrium \
when we further assume that there is no selection in adult males (sm=0). \
\>", "Text",
 CellChangeTimes->{{3.8140092507145643`*^9, 
  3.814009313049942*^9}},ExpressionUUID->"ac86ebce-ccb5-4e46-895b-\
f4f370499b0b"],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"equilibrium", "=", 
  RowBox[{
   RowBox[{"Solve", "[", 
    RowBox[{
     RowBox[{
      RowBox[{"(", 
       RowBox[{
        RowBox[{"solveme", "[", 
         RowBox[{"[", "1", "]"}], "]"}], "/.", 
        RowBox[{"sm", "\[Rule]", "0"}]}], ")"}], "\[Equal]", 
      RowBox[{"{", 
       RowBox[{"0", ",", "0", ",", "0"}], "}"}]}], ",", 
     RowBox[{"{", 
      RowBox[{"pm", ",", "pf", ",", "F"}], "}"}]}], "]"}], "//", 
   "Simplify"}]}]], "Input",
 CellChangeTimes->{{3.781875133272497*^9, 3.781875133336864*^9}, {
   3.7818756691961737`*^9, 3.781875671668582*^9}, 3.7818758813381*^9, {
   3.8140075485293694`*^9, 3.814007554648246*^9}},
 CellLabel->
  "In[1366]:=",ExpressionUUID->"24e38814-b48b-4912-8a57-eba6bbb961af"],

Cell[BoxData[
 RowBox[{"{", 
  RowBox[{
   RowBox[{"{", 
    RowBox[{
     RowBox[{"pm", "\[Rule]", 
      RowBox[{"-", 
       RowBox[{
        FractionBox["1", 
         RowBox[{"4", " ", 
          RowBox[{"(", 
           RowBox[{
            RowBox[{"-", "1"}], "+", "c"}], ")"}], " ", "sf", " ", 
          RowBox[{"(", 
           RowBox[{
            RowBox[{"-", "1"}], "-", "\[Delta]", "+", 
            RowBox[{"h", " ", 
             RowBox[{"(", 
              RowBox[{"2", "+", "\[Delta]"}], ")"}]}]}], ")"}]}]], 
        RowBox[{"(", 
         RowBox[{
          RowBox[{"4", " ", "c", " ", "sf"}], "-", 
          RowBox[{"8", " ", "c", " ", "h", " ", "sf"}], "-", "\[Delta]", "+", 
          RowBox[{"c", " ", "\[Delta]"}], "+", 
          RowBox[{"4", " ", "c", " ", "sf", " ", "\[Delta]"}], "+", 
          RowBox[{"h", " ", "sf", " ", "\[Delta]"}], "-", 
          RowBox[{"5", " ", "c", " ", "h", " ", "sf", " ", "\[Delta]"}], "+", 
          SqrtBox[
           RowBox[{
            RowBox[{
             RowBox[{"-", "8"}], " ", 
             RowBox[{"(", 
              RowBox[{
               RowBox[{"-", "1"}], "+", "c"}], ")"}], " ", "sf", " ", 
             RowBox[{"(", 
              RowBox[{
               RowBox[{"-", "1"}], "-", "\[Delta]", "+", 
               RowBox[{"h", " ", 
                RowBox[{"(", 
                 RowBox[{"2", "+", "\[Delta]"}], ")"}]}]}], ")"}], " ", 
             RowBox[{"(", 
              RowBox[{"1", "+", 
               RowBox[{"h", " ", "sf"}], "+", 
               RowBox[{"c", " ", 
                RowBox[{"(", 
                 RowBox[{
                  RowBox[{"-", "1"}], "-", 
                  RowBox[{"2", " ", "\[Delta]"}], "+", 
                  RowBox[{"h", " ", "sf", " ", 
                   RowBox[{"(", 
                    RowBox[{"3", "+", 
                    RowBox[{"2", " ", "\[Delta]"}]}], ")"}]}]}], ")"}]}]}], 
              ")"}]}], "+", 
            SuperscriptBox[
             RowBox[{"(", 
              RowBox[{
               RowBox[{
                RowBox[{"(", 
                 RowBox[{
                  RowBox[{"-", "1"}], "+", 
                  RowBox[{"h", " ", "sf"}]}], ")"}], " ", "\[Delta]"}], "+", 
               RowBox[{"c", " ", 
                RowBox[{"(", 
                 RowBox[{"\[Delta]", "+", 
                  RowBox[{"sf", " ", 
                   RowBox[{"(", 
                    RowBox[{"4", "-", 
                    RowBox[{"8", " ", "h"}], "+", 
                    RowBox[{"4", " ", "\[Delta]"}], "-", 
                    RowBox[{"5", " ", "h", " ", "\[Delta]"}]}], ")"}]}]}], 
                 ")"}]}]}], ")"}], "2"]}]]}], ")"}]}]}]}], ",", 
     RowBox[{"pf", "\[Rule]", 
      RowBox[{"-", 
       RowBox[{
        FractionBox["1", 
         RowBox[{"4", " ", 
          RowBox[{"(", 
           RowBox[{
            RowBox[{"-", "1"}], "+", "c"}], ")"}], " ", "sf", " ", 
          RowBox[{"(", 
           RowBox[{
            RowBox[{"-", "1"}], "-", "\[Delta]", "+", 
            RowBox[{"h", " ", 
             RowBox[{"(", 
              RowBox[{"2", "+", "\[Delta]"}], ")"}]}]}], ")"}]}]], 
        RowBox[{"(", 
         RowBox[{
          RowBox[{"4", " ", "c", " ", "sf"}], "-", 
          RowBox[{"8", " ", "c", " ", "h", " ", "sf"}], "-", "\[Delta]", "+", 
          RowBox[{"c", " ", "\[Delta]"}], "+", 
          RowBox[{"4", " ", "c", " ", "sf", " ", "\[Delta]"}], "+", 
          RowBox[{"h", " ", "sf", " ", "\[Delta]"}], "-", 
          RowBox[{"5", " ", "c", " ", "h", " ", "sf", " ", "\[Delta]"}], "+", 
          SqrtBox[
           RowBox[{
            RowBox[{
             RowBox[{"-", "8"}], " ", 
             RowBox[{"(", 
              RowBox[{
               RowBox[{"-", "1"}], "+", "c"}], ")"}], " ", "sf", " ", 
             RowBox[{"(", 
              RowBox[{
               RowBox[{"-", "1"}], "-", "\[Delta]", "+", 
               RowBox[{"h", " ", 
                RowBox[{"(", 
                 RowBox[{"2", "+", "\[Delta]"}], ")"}]}]}], ")"}], " ", 
             RowBox[{"(", 
              RowBox[{"1", "+", 
               RowBox[{"h", " ", "sf"}], "+", 
               RowBox[{"c", " ", 
                RowBox[{"(", 
                 RowBox[{
                  RowBox[{"-", "1"}], "-", 
                  RowBox[{"2", " ", "\[Delta]"}], "+", 
                  RowBox[{"h", " ", "sf", " ", 
                   RowBox[{"(", 
                    RowBox[{"3", "+", 
                    RowBox[{"2", " ", "\[Delta]"}]}], ")"}]}]}], ")"}]}]}], 
              ")"}]}], "+", 
            SuperscriptBox[
             RowBox[{"(", 
              RowBox[{
               RowBox[{
                RowBox[{"(", 
                 RowBox[{
                  RowBox[{"-", "1"}], "+", 
                  RowBox[{"h", " ", "sf"}]}], ")"}], " ", "\[Delta]"}], "+", 
               RowBox[{"c", " ", 
                RowBox[{"(", 
                 RowBox[{"\[Delta]", "+", 
                  RowBox[{"sf", " ", 
                   RowBox[{"(", 
                    RowBox[{"4", "-", 
                    RowBox[{"8", " ", "h"}], "+", 
                    RowBox[{"4", " ", "\[Delta]"}], "-", 
                    RowBox[{"5", " ", "h", " ", "\[Delta]"}]}], ")"}]}]}], 
                 ")"}]}]}], ")"}], "2"]}]]}], ")"}]}]}]}], ",", 
     RowBox[{"F", "\[Rule]", 
      RowBox[{
       RowBox[{"(", 
        RowBox[{
         RowBox[{
          RowBox[{"-", "2"}], " ", "sf"}], "+", 
         RowBox[{"4", " ", "c", " ", "sf"}], "-", 
         RowBox[{"2", " ", 
          SuperscriptBox["c", "2"], " ", "sf"}], "+", 
         RowBox[{"4", " ", "h", " ", "sf"}], "-", 
         RowBox[{"8", " ", "c", " ", "h", " ", "sf"}], "+", 
         RowBox[{"4", " ", 
          SuperscriptBox["c", "2"], " ", "h", " ", "sf"}], "+", 
         RowBox[{"8", " ", 
          SuperscriptBox["sf", "2"]}], "+", 
         RowBox[{"4", " ", "c", " ", 
          SuperscriptBox["sf", "2"]}], "+", 
         RowBox[{"4", " ", 
          SuperscriptBox["c", "2"], " ", 
          SuperscriptBox["sf", "2"]}], "-", 
         RowBox[{"18", " ", "h", " ", 
          SuperscriptBox["sf", "2"]}], "-", 
         RowBox[{"36", " ", "c", " ", "h", " ", 
          SuperscriptBox["sf", "2"]}], "-", 
         RowBox[{"10", " ", 
          SuperscriptBox["c", "2"], " ", "h", " ", 
          SuperscriptBox["sf", "2"]}], "+", 
         RowBox[{"8", " ", 
          SuperscriptBox["h", "2"], " ", 
          SuperscriptBox["sf", "2"]}], "+", 
         RowBox[{"48", " ", "c", " ", 
          SuperscriptBox["h", "2"], " ", 
          SuperscriptBox["sf", "2"]}], "+", 
         RowBox[{"8", " ", 
          SuperscriptBox["c", "2"], " ", 
          SuperscriptBox["h", "2"], " ", 
          SuperscriptBox["sf", "2"]}], "+", 
         RowBox[{"8", " ", "h", " ", 
          SuperscriptBox["sf", "3"]}], "+", 
         RowBox[{"24", " ", "c", " ", "h", " ", 
          SuperscriptBox["sf", "3"]}], "-", 
         RowBox[{"14", " ", 
          SuperscriptBox["h", "2"], " ", 
          SuperscriptBox["sf", "3"]}], "-", 
         RowBox[{"36", " ", "c", " ", 
          SuperscriptBox["h", "2"], " ", 
          SuperscriptBox["sf", "3"]}], "-", 
         RowBox[{"14", " ", 
          SuperscriptBox["c", "2"], " ", 
          SuperscriptBox["h", "2"], " ", 
          SuperscriptBox["sf", "3"]}], "+", 
         RowBox[{"4", " ", 
          SuperscriptBox["h", "3"], " ", 
          SuperscriptBox["sf", "3"]}], "-", 
         RowBox[{"8", " ", "c", " ", 
          SuperscriptBox["h", "3"], " ", 
          SuperscriptBox["sf", "3"]}], "+", 
         RowBox[{"4", " ", 
          SuperscriptBox["c", "2"], " ", 
          SuperscriptBox["h", "3"], " ", 
          SuperscriptBox["sf", "3"]}], "+", 
         RowBox[{"4", " ", "c", " ", 
          SuperscriptBox["h", "2"], " ", 
          SuperscriptBox["sf", "4"]}], "+", 
         RowBox[{"12", " ", 
          SuperscriptBox["c", "2"], " ", 
          SuperscriptBox["h", "2"], " ", 
          SuperscriptBox["sf", "4"]}], "+", 
         RowBox[{"2", " ", 
          SuperscriptBox["h", "3"], " ", 
          SuperscriptBox["sf", "4"]}], "+", 
         RowBox[{"4", " ", "c", " ", 
          SuperscriptBox["h", "3"], " ", 
          SuperscriptBox["sf", "4"]}], "-", 
         RowBox[{"6", " ", 
          SuperscriptBox["c", "2"], " ", 
          SuperscriptBox["h", "3"], " ", 
          SuperscriptBox["sf", "4"]}], "-", 
         RowBox[{"9", " ", "sf", " ", "\[Delta]"}], "+", 
         RowBox[{"10", " ", "c", " ", "sf", " ", "\[Delta]"}], "-", 
         RowBox[{
          SuperscriptBox["c", "2"], " ", "sf", " ", "\[Delta]"}], "+", 
         RowBox[{"12", " ", "h", " ", "sf", " ", "\[Delta]"}], "-", 
         RowBox[{"8", " ", "c", " ", "h", " ", "sf", " ", "\[Delta]"}], "-", 
         RowBox[{"4", " ", 
          SuperscriptBox["c", "2"], " ", "h", " ", "sf", " ", "\[Delta]"}], 
         "+", 
         RowBox[{"8", " ", 
          SuperscriptBox["sf", "2"], " ", "\[Delta]"}], "-", 
         RowBox[{"4", " ", "c", " ", 
          SuperscriptBox["sf", "2"], " ", "\[Delta]"}], "+", 
         RowBox[{"4", " ", 
          SuperscriptBox["c", "2"], " ", 
          SuperscriptBox["sf", "2"], " ", "\[Delta]"}], "-", 
         RowBox[{"13", " ", "h", " ", 
          SuperscriptBox["sf", "2"], " ", "\[Delta]"}], "-", 
         RowBox[{"38", " ", "c", " ", "h", " ", 
          SuperscriptBox["sf", "2"], " ", "\[Delta]"}], "+", 
         RowBox[{"3", " ", 
          SuperscriptBox["c", "2"], " ", "h", " ", 
          SuperscriptBox["sf", "2"], " ", "\[Delta]"}], "+", 
         RowBox[{"48", " ", "c", " ", 
          SuperscriptBox["h", "2"], " ", 
          SuperscriptBox["sf", "2"], " ", "\[Delta]"}], "+", 
         RowBox[{"16", " ", 
          SuperscriptBox["c", "2"], " ", 
          SuperscriptBox["h", "2"], " ", 
          SuperscriptBox["sf", "2"], " ", "\[Delta]"}], "+", 
         RowBox[{"8", " ", "h", " ", 
          SuperscriptBox["sf", "3"], " ", "\[Delta]"}], "+", 
         RowBox[{"32", " ", "c", " ", "h", " ", 
          SuperscriptBox["sf", "3"], " ", "\[Delta]"}], "-", 
         RowBox[{"8", " ", 
          SuperscriptBox["c", "2"], " ", "h", " ", 
          SuperscriptBox["sf", "3"], " ", "\[Delta]"}], "-", 
         RowBox[{"11", " ", 
          SuperscriptBox["h", "2"], " ", 
          SuperscriptBox["sf", "3"], " ", "\[Delta]"}], "-", 
         RowBox[{"42", " ", "c", " ", 
          SuperscriptBox["h", "2"], " ", 
          SuperscriptBox["sf", "3"], " ", "\[Delta]"}], "-", 
         RowBox[{"27", " ", 
          SuperscriptBox["c", "2"], " ", 
          SuperscriptBox["h", "2"], " ", 
          SuperscriptBox["sf", "3"], " ", "\[Delta]"}], "+", 
         RowBox[{"4", " ", 
          SuperscriptBox["h", "3"], " ", 
          SuperscriptBox["sf", "3"], " ", "\[Delta]"}], "-", 
         RowBox[{"8", " ", "c", " ", 
          SuperscriptBox["h", "3"], " ", 
          SuperscriptBox["sf", "3"], " ", "\[Delta]"}], "+", 
         RowBox[{"4", " ", 
          SuperscriptBox["c", "2"], " ", 
          SuperscriptBox["h", "3"], " ", 
          SuperscriptBox["sf", "3"], " ", "\[Delta]"}], "+", 
         RowBox[{"4", " ", "c", " ", 
          SuperscriptBox["h", "2"], " ", 
          SuperscriptBox["sf", "4"], " ", "\[Delta]"}], "+", 
         RowBox[{"20", " ", 
          SuperscriptBox["c", "2"], " ", 
          SuperscriptBox["h", "2"], " ", 
          SuperscriptBox["sf", "4"], " ", "\[Delta]"}], "+", 
         RowBox[{
          SuperscriptBox["h", "3"], " ", 
          SuperscriptBox["sf", "4"], " ", "\[Delta]"}], "+", 
         RowBox[{"6", " ", "c", " ", 
          SuperscriptBox["h", "3"], " ", 
          SuperscriptBox["sf", "4"], " ", "\[Delta]"}], "-", 
         RowBox[{"7", " ", 
          SuperscriptBox["c", "2"], " ", 
          SuperscriptBox["h", "3"], " ", 
          SuperscriptBox["sf", "4"], " ", "\[Delta]"}], "+", 
         SuperscriptBox["\[Delta]", "2"], "-", 
         RowBox[{"2", " ", "c", " ", 
          SuperscriptBox["\[Delta]", "2"]}], "+", 
         RowBox[{
          SuperscriptBox["c", "2"], " ", 
          SuperscriptBox["\[Delta]", "2"]}], "-", 
         RowBox[{"2", " ", "sf", " ", 
          SuperscriptBox["\[Delta]", "2"]}], "+", 
         RowBox[{"10", " ", "c", " ", "sf", " ", 
          SuperscriptBox["\[Delta]", "2"]}], "-", 
         RowBox[{"h", " ", "sf", " ", 
          SuperscriptBox["\[Delta]", "2"]}], "-", 
         RowBox[{"6", " ", "c", " ", "h", " ", "sf", " ", 
          SuperscriptBox["\[Delta]", "2"]}], "-", 
         RowBox[{"9", " ", 
          SuperscriptBox["c", "2"], " ", "h", " ", "sf", " ", 
          SuperscriptBox["\[Delta]", "2"]}], "-", 
         RowBox[{"8", " ", "c", " ", 
          SuperscriptBox["sf", "2"], " ", 
          SuperscriptBox["\[Delta]", "2"]}], "+", 
         RowBox[{"4", " ", "h", " ", 
          SuperscriptBox["sf", "2"], " ", 
          SuperscriptBox["\[Delta]", "2"]}], "-", 
         RowBox[{"2", " ", "c", " ", "h", " ", 
          SuperscriptBox["sf", "2"], " ", 
          SuperscriptBox["\[Delta]", "2"]}], "+", 
         RowBox[{"14", " ", 
          SuperscriptBox["c", "2"], " ", "h", " ", 
          SuperscriptBox["sf", "2"], " ", 
          SuperscriptBox["\[Delta]", "2"]}], "-", 
         RowBox[{
          SuperscriptBox["h", "2"], " ", 
          SuperscriptBox["sf", "2"], " ", 
          SuperscriptBox["\[Delta]", "2"]}], "+", 
         RowBox[{"10", " ", "c", " ", 
          SuperscriptBox["h", "2"], " ", 
          SuperscriptBox["sf", "2"], " ", 
          SuperscriptBox["\[Delta]", "2"]}], "+", 
         RowBox[{"7", " ", 
          SuperscriptBox["c", "2"], " ", 
          SuperscriptBox["h", "2"], " ", 
          SuperscriptBox["sf", "2"], " ", 
          SuperscriptBox["\[Delta]", "2"]}], "+", 
         RowBox[{"8", " ", "c", " ", "h", " ", 
          SuperscriptBox["sf", "3"], " ", 
          SuperscriptBox["\[Delta]", "2"]}], "-", 
         RowBox[{"8", " ", 
          SuperscriptBox["c", "2"], " ", "h", " ", 
          SuperscriptBox["sf", "3"], " ", 
          SuperscriptBox["\[Delta]", "2"]}], "-", 
         RowBox[{"2", " ", 
          SuperscriptBox["h", "2"], " ", 
          SuperscriptBox["sf", "3"], " ", 
          SuperscriptBox["\[Delta]", "2"]}], "-", 
         RowBox[{"10", " ", "c", " ", 
          SuperscriptBox["h", "2"], " ", 
          SuperscriptBox["sf", "3"], " ", 
          SuperscriptBox["\[Delta]", "2"]}], "-", 
         RowBox[{"12", " ", 
          SuperscriptBox["c", "2"], " ", 
          SuperscriptBox["h", "2"], " ", 
          SuperscriptBox["sf", "3"], " ", 
          SuperscriptBox["\[Delta]", "2"]}], "+", 
         RowBox[{
          SuperscriptBox["h", "3"], " ", 
          SuperscriptBox["sf", "3"], " ", 
          SuperscriptBox["\[Delta]", "2"]}], "-", 
         RowBox[{"2", " ", "c", " ", 
          SuperscriptBox["h", "3"], " ", 
          SuperscriptBox["sf", "3"], " ", 
          SuperscriptBox["\[Delta]", "2"]}], "+", 
         RowBox[{
          SuperscriptBox["c", "2"], " ", 
          SuperscriptBox["h", "3"], " ", 
          SuperscriptBox["sf", "3"], " ", 
          SuperscriptBox["\[Delta]", "2"]}], "+", 
         RowBox[{"8", " ", 
          SuperscriptBox["c", "2"], " ", 
          SuperscriptBox["h", "2"], " ", 
          SuperscriptBox["sf", "4"], " ", 
          SuperscriptBox["\[Delta]", "2"]}], "+", 
         RowBox[{"2", " ", "c", " ", 
          SuperscriptBox["h", "3"], " ", 
          SuperscriptBox["sf", "4"], " ", 
          SuperscriptBox["\[Delta]", "2"]}], "-", 
         RowBox[{"2", " ", 
          SuperscriptBox["c", "2"], " ", 
          SuperscriptBox["h", "3"], " ", 
          SuperscriptBox["sf", "4"], " ", 
          SuperscriptBox["\[Delta]", "2"]}], "+", 
         RowBox[{"3", " ", "sf", " ", 
          SqrtBox[
           RowBox[{
            RowBox[{
             RowBox[{"-", "8"}], " ", 
             RowBox[{"(", 
              RowBox[{
               RowBox[{"-", "1"}], "+", "c"}], ")"}], " ", "sf", " ", 
             RowBox[{"(", 
              RowBox[{
               RowBox[{"-", "1"}], "-", "\[Delta]", "+", 
               RowBox[{"h", " ", 
                RowBox[{"(", 
                 RowBox[{"2", "+", "\[Delta]"}], ")"}]}]}], ")"}], " ", 
             RowBox[{"(", 
              RowBox[{"1", "+", 
               RowBox[{"h", " ", "sf"}], "+", 
               RowBox[{"c", " ", 
                RowBox[{"(", 
                 RowBox[{
                  RowBox[{"-", "1"}], "-", 
                  RowBox[{"2", " ", "\[Delta]"}], "+", 
                  RowBox[{"h", " ", "sf", " ", 
                   RowBox[{"(", 
                    RowBox[{"3", "+", 
                    RowBox[{"2", " ", "\[Delta]"}]}], ")"}]}]}], ")"}]}]}], 
              ")"}]}], "+", 
            SuperscriptBox[
             RowBox[{"(", 
              RowBox[{
               RowBox[{
                RowBox[{"(", 
                 RowBox[{
                  RowBox[{"-", "1"}], "+", 
                  RowBox[{"h", " ", "sf"}]}], ")"}], " ", "\[Delta]"}], "+", 
               RowBox[{"c", " ", 
                RowBox[{"(", 
                 RowBox[{"\[Delta]", "+", 
                  RowBox[{"sf", " ", 
                   RowBox[{"(", 
                    RowBox[{"4", "-", 
                    RowBox[{"8", " ", "h"}], "+", 
                    RowBox[{"4", " ", "\[Delta]"}], "-", 
                    RowBox[{"5", " ", "h", " ", "\[Delta]"}]}], ")"}]}]}], 
                 ")"}]}]}], ")"}], "2"]}]]}], "+", 
         RowBox[{"c", " ", "sf", " ", 
          SqrtBox[
           RowBox[{
            RowBox[{
             RowBox[{"-", "8"}], " ", 
             RowBox[{"(", 
              RowBox[{
               RowBox[{"-", "1"}], "+", "c"}], ")"}], " ", "sf", " ", 
             RowBox[{"(", 
              RowBox[{
               RowBox[{"-", "1"}], "-", "\[Delta]", "+", 
               RowBox[{"h", " ", 
                RowBox[{"(", 
                 RowBox[{"2", "+", "\[Delta]"}], ")"}]}]}], ")"}], " ", 
             RowBox[{"(", 
              RowBox[{"1", "+", 
               RowBox[{"h", " ", "sf"}], "+", 
               RowBox[{"c", " ", 
                RowBox[{"(", 
                 RowBox[{
                  RowBox[{"-", "1"}], "-", 
                  RowBox[{"2", " ", "\[Delta]"}], "+", 
                  RowBox[{"h", " ", "sf", " ", 
                   RowBox[{"(", 
                    RowBox[{"3", "+", 
                    RowBox[{"2", " ", "\[Delta]"}]}], ")"}]}]}], ")"}]}]}], 
              ")"}]}], "+", 
            SuperscriptBox[
             RowBox[{"(", 
              RowBox[{
               RowBox[{
                RowBox[{"(", 
                 RowBox[{
                  RowBox[{"-", "1"}], "+", 
                  RowBox[{"h", " ", "sf"}]}], ")"}], " ", "\[Delta]"}], "+", 
               RowBox[{"c", " ", 
                RowBox[{"(", 
                 RowBox[{"\[Delta]", "+", 
                  RowBox[{"sf", " ", 
                   RowBox[{"(", 
                    RowBox[{"4", "-", 
                    RowBox[{"8", " ", "h"}], "+", 
                    RowBox[{"4", " ", "\[Delta]"}], "-", 
                    RowBox[{"5", " ", "h", " ", "\[Delta]"}]}], ")"}]}]}], 
                 ")"}]}]}], ")"}], "2"]}]]}], "-", 
         RowBox[{"4", " ", "h", " ", "sf", " ", 
          SqrtBox[
           RowBox[{
            RowBox[{
             RowBox[{"-", "8"}], " ", 
             RowBox[{"(", 
              RowBox[{
               RowBox[{"-", "1"}], "+", "c"}], ")"}], " ", "sf", " ", 
             RowBox[{"(", 
              RowBox[{
               RowBox[{"-", "1"}], "-", "\[Delta]", "+", 
               RowBox[{"h", " ", 
                RowBox[{"(", 
                 RowBox[{"2", "+", "\[Delta]"}], ")"}]}]}], ")"}], " ", 
             RowBox[{"(", 
              RowBox[{"1", "+", 
               RowBox[{"h", " ", "sf"}], "+", 
               RowBox[{"c", " ", 
                RowBox[{"(", 
                 RowBox[{
                  RowBox[{"-", "1"}], "-", 
                  RowBox[{"2", " ", "\[Delta]"}], "+", 
                  RowBox[{"h", " ", "sf", " ", 
                   RowBox[{"(", 
                    RowBox[{"3", "+", 
                    RowBox[{"2", " ", "\[Delta]"}]}], ")"}]}]}], ")"}]}]}], 
              ")"}]}], "+", 
            SuperscriptBox[
             RowBox[{"(", 
              RowBox[{
               RowBox[{
                RowBox[{"(", 
                 RowBox[{
                  RowBox[{"-", "1"}], "+", 
                  RowBox[{"h", " ", "sf"}]}], ")"}], " ", "\[Delta]"}], "+", 
               RowBox[{"c", " ", 
                RowBox[{"(", 
                 RowBox[{"\[Delta]", "+", 
                  RowBox[{"sf", " ", 
                   RowBox[{"(", 
                    RowBox[{"4", "-", 
                    RowBox[{"8", " ", "h"}], "+", 
                    RowBox[{"4", " ", "\[Delta]"}], "-", 
                    RowBox[{"5", " ", "h", " ", "\[Delta]"}]}], ")"}]}]}], 
                 ")"}]}]}], ")"}], "2"]}]]}], "-", 
         RowBox[{"4", " ", "c", " ", "h", " ", "sf", " ", 
          SqrtBox[
           RowBox[{
            RowBox[{
             RowBox[{"-", "8"}], " ", 
             RowBox[{"(", 
              RowBox[{
               RowBox[{"-", "1"}], "+", "c"}], ")"}], " ", "sf", " ", 
             RowBox[{"(", 
              RowBox[{
               RowBox[{"-", "1"}], "-", "\[Delta]", "+", 
               RowBox[{"h", " ", 
                RowBox[{"(", 
                 RowBox[{"2", "+", "\[Delta]"}], ")"}]}]}], ")"}], " ", 
             RowBox[{"(", 
              RowBox[{"1", "+", 
               RowBox[{"h", " ", "sf"}], "+", 
               RowBox[{"c", " ", 
                RowBox[{"(", 
                 RowBox[{
                  RowBox[{"-", "1"}], "-", 
                  RowBox[{"2", " ", "\[Delta]"}], "+", 
                  RowBox[{"h", " ", "sf", " ", 
                   RowBox[{"(", 
                    RowBox[{"3", "+", 
                    RowBox[{"2", " ", "\[Delta]"}]}], ")"}]}]}], ")"}]}]}], 
              ")"}]}], "+", 
            SuperscriptBox[
             RowBox[{"(", 
              RowBox[{
               RowBox[{
                RowBox[{"(", 
                 RowBox[{
                  RowBox[{"-", "1"}], "+", 
                  RowBox[{"h", " ", "sf"}]}], ")"}], " ", "\[Delta]"}], "+", 
               RowBox[{"c", " ", 
                RowBox[{"(", 
                 RowBox[{"\[Delta]", "+", 
                  RowBox[{"sf", " ", 
                   RowBox[{"(", 
                    RowBox[{"4", "-", 
                    RowBox[{"8", " ", "h"}], "+", 
                    RowBox[{"4", " ", "\[Delta]"}], "-", 
                    RowBox[{"5", " ", "h", " ", "\[Delta]"}]}], ")"}]}]}], 
                 ")"}]}]}], ")"}], "2"]}]]}], "-", 
         RowBox[{"2", " ", "h", " ", 
          SuperscriptBox["sf", "2"], " ", 
          SqrtBox[
           RowBox[{
            RowBox[{
             RowBox[{"-", "8"}], " ", 
             RowBox[{"(", 
              RowBox[{
               RowBox[{"-", "1"}], "+", "c"}], ")"}], " ", "sf", " ", 
             RowBox[{"(", 
              RowBox[{
               RowBox[{"-", "1"}], "-", "\[Delta]", "+", 
               RowBox[{"h", " ", 
                RowBox[{"(", 
                 RowBox[{"2", "+", "\[Delta]"}], ")"}]}]}], ")"}], " ", 
             RowBox[{"(", 
              RowBox[{"1", "+", 
               RowBox[{"h", " ", "sf"}], "+", 
               RowBox[{"c", " ", 
                RowBox[{"(", 
                 RowBox[{
                  RowBox[{"-", "1"}], "-", 
                  RowBox[{"2", " ", "\[Delta]"}], "+", 
                  RowBox[{"h", " ", "sf", " ", 
                   RowBox[{"(", 
                    RowBox[{"3", "+", 
                    RowBox[{"2", " ", "\[Delta]"}]}], ")"}]}]}], ")"}]}]}], 
              ")"}]}], "+", 
            SuperscriptBox[
             RowBox[{"(", 
              RowBox[{
               RowBox[{
                RowBox[{"(", 
                 RowBox[{
                  RowBox[{"-", "1"}], "+", 
                  RowBox[{"h", " ", "sf"}]}], ")"}], " ", "\[Delta]"}], "+", 
               RowBox[{"c", " ", 
                RowBox[{"(", 
                 RowBox[{"\[Delta]", "+", 
                  RowBox[{"sf", " ", 
                   RowBox[{"(", 
                    RowBox[{"4", "-", 
                    RowBox[{"8", " ", "h"}], "+", 
                    RowBox[{"4", " ", "\[Delta]"}], "-", 
                    RowBox[{"5", " ", "h", " ", "\[Delta]"}]}], ")"}]}]}], 
                 ")"}]}]}], ")"}], "2"]}]]}], "+", 
         RowBox[{"2", " ", "c", " ", "h", " ", 
          SuperscriptBox["sf", "2"], " ", 
          SqrtBox[
           RowBox[{
            RowBox[{
             RowBox[{"-", "8"}], " ", 
             RowBox[{"(", 
              RowBox[{
               RowBox[{"-", "1"}], "+", "c"}], ")"}], " ", "sf", " ", 
             RowBox[{"(", 
              RowBox[{
               RowBox[{"-", "1"}], "-", "\[Delta]", "+", 
               RowBox[{"h", " ", 
                RowBox[{"(", 
                 RowBox[{"2", "+", "\[Delta]"}], ")"}]}]}], ")"}], " ", 
             RowBox[{"(", 
              RowBox[{"1", "+", 
               RowBox[{"h", " ", "sf"}], "+", 
               RowBox[{"c", " ", 
                RowBox[{"(", 
                 RowBox[{
                  RowBox[{"-", "1"}], "-", 
                  RowBox[{"2", " ", "\[Delta]"}], "+", 
                  RowBox[{"h", " ", "sf", " ", 
                   RowBox[{"(", 
                    RowBox[{"3", "+", 
                    RowBox[{"2", " ", "\[Delta]"}]}], ")"}]}]}], ")"}]}]}], 
              ")"}]}], "+", 
            SuperscriptBox[
             RowBox[{"(", 
              RowBox[{
               RowBox[{
                RowBox[{"(", 
                 RowBox[{
                  RowBox[{"-", "1"}], "+", 
                  RowBox[{"h", " ", "sf"}]}], ")"}], " ", "\[Delta]"}], "+", 
               RowBox[{"c", " ", 
                RowBox[{"(", 
                 RowBox[{"\[Delta]", "+", 
                  RowBox[{"sf", " ", 
                   RowBox[{"(", 
                    RowBox[{"4", "-", 
                    RowBox[{"8", " ", "h"}], "+", 
                    RowBox[{"4", " ", "\[Delta]"}], "-", 
                    RowBox[{"5", " ", "h", " ", "\[Delta]"}]}], ")"}]}]}], 
                 ")"}]}]}], ")"}], "2"]}]]}], "+", 
         RowBox[{"4", " ", 
          SuperscriptBox["h", "2"], " ", 
          SuperscriptBox["sf", "2"], " ", 
          SqrtBox[
           RowBox[{
            RowBox[{
             RowBox[{"-", "8"}], " ", 
             RowBox[{"(", 
              RowBox[{
               RowBox[{"-", "1"}], "+", "c"}], ")"}], " ", "sf", " ", 
             RowBox[{"(", 
              RowBox[{
               RowBox[{"-", "1"}], "-", "\[Delta]", "+", 
               RowBox[{"h", " ", 
                RowBox[{"(", 
                 RowBox[{"2", "+", "\[Delta]"}], ")"}]}]}], ")"}], " ", 
             RowBox[{"(", 
              RowBox[{"1", "+", 
               RowBox[{"h", " ", "sf"}], "+", 
               RowBox[{"c", " ", 
                RowBox[{"(", 
                 RowBox[{
                  RowBox[{"-", "1"}], "-", 
                  RowBox[{"2", " ", "\[Delta]"}], "+", 
                  RowBox[{"h", " ", "sf", " ", 
                   RowBox[{"(", 
                    RowBox[{"3", "+", 
                    RowBox[{"2", " ", "\[Delta]"}]}], ")"}]}]}], ")"}]}]}], 
              ")"}]}], "+", 
            SuperscriptBox[
             RowBox[{"(", 
              RowBox[{
               RowBox[{
                RowBox[{"(", 
                 RowBox[{
                  RowBox[{"-", "1"}], "+", 
                  RowBox[{"h", " ", "sf"}]}], ")"}], " ", "\[Delta]"}], "+", 
               RowBox[{"c", " ", 
                RowBox[{"(", 
                 RowBox[{"\[Delta]", "+", 
                  RowBox[{"sf", " ", 
                   RowBox[{"(", 
                    RowBox[{"4", "-", 
                    RowBox[{"8", " ", "h"}], "+", 
                    RowBox[{"4", " ", "\[Delta]"}], "-", 
                    RowBox[{"5", " ", "h", " ", "\[Delta]"}]}], ")"}]}]}], 
                 ")"}]}]}], ")"}], "2"]}]]}], "+", 
         RowBox[{"4", " ", "c", " ", 
          SuperscriptBox["h", "2"], " ", 
          SuperscriptBox["sf", "2"], " ", 
          SqrtBox[
           RowBox[{
            RowBox[{
             RowBox[{"-", "8"}], " ", 
             RowBox[{"(", 
              RowBox[{
               RowBox[{"-", "1"}], "+", "c"}], ")"}], " ", "sf", " ", 
             RowBox[{"(", 
              RowBox[{
               RowBox[{"-", "1"}], "-", "\[Delta]", "+", 
               RowBox[{"h", " ", 
                RowBox[{"(", 
                 RowBox[{"2", "+", "\[Delta]"}], ")"}]}]}], ")"}], " ", 
             RowBox[{"(", 
              RowBox[{"1", "+", 
               RowBox[{"h", " ", "sf"}], "+", 
               RowBox[{"c", " ", 
                RowBox[{"(", 
                 RowBox[{
                  RowBox[{"-", "1"}], "-", 
                  RowBox[{"2", " ", "\[Delta]"}], "+", 
                  RowBox[{"h", " ", "sf", " ", 
                   RowBox[{"(", 
                    RowBox[{"3", "+", 
                    RowBox[{"2", " ", "\[Delta]"}]}], ")"}]}]}], ")"}]}]}], 
              ")"}]}], "+", 
            SuperscriptBox[
             RowBox[{"(", 
              RowBox[{
               RowBox[{
                RowBox[{"(", 
                 RowBox[{
                  RowBox[{"-", "1"}], "+", 
                  RowBox[{"h", " ", "sf"}]}], ")"}], " ", "\[Delta]"}], "+", 
               RowBox[{"c", " ", 
                RowBox[{"(", 
                 RowBox[{"\[Delta]", "+", 
                  RowBox[{"sf", " ", 
                   RowBox[{"(", 
                    RowBox[{"4", "-", 
                    RowBox[{"8", " ", "h"}], "+", 
                    RowBox[{"4", " ", "\[Delta]"}], "-", 
                    RowBox[{"5", " ", "h", " ", "\[Delta]"}]}], ")"}]}]}], 
                 ")"}]}]}], ")"}], "2"]}]]}], "-", 
         RowBox[{
          SuperscriptBox["h", "2"], " ", 
          SuperscriptBox["sf", "3"], " ", 
          SqrtBox[
           RowBox[{
            RowBox[{
             RowBox[{"-", "8"}], " ", 
             RowBox[{"(", 
              RowBox[{
               RowBox[{"-", "1"}], "+", "c"}], ")"}], " ", "sf", " ", 
             RowBox[{"(", 
              RowBox[{
               RowBox[{"-", "1"}], "-", "\[Delta]", "+", 
               RowBox[{"h", " ", 
                RowBox[{"(", 
                 RowBox[{"2", "+", "\[Delta]"}], ")"}]}]}], ")"}], " ", 
             RowBox[{"(", 
              RowBox[{"1", "+", 
               RowBox[{"h", " ", "sf"}], "+", 
               RowBox[{"c", " ", 
                RowBox[{"(", 
                 RowBox[{
                  RowBox[{"-", "1"}], "-", 
                  RowBox[{"2", " ", "\[Delta]"}], "+", 
                  RowBox[{"h", " ", "sf", " ", 
                   RowBox[{"(", 
                    RowBox[{"3", "+", 
                    RowBox[{"2", " ", "\[Delta]"}]}], ")"}]}]}], ")"}]}]}], 
              ")"}]}], "+", 
            SuperscriptBox[
             RowBox[{"(", 
              RowBox[{
               RowBox[{
                RowBox[{"(", 
                 RowBox[{
                  RowBox[{"-", "1"}], "+", 
                  RowBox[{"h", " ", "sf"}]}], ")"}], " ", "\[Delta]"}], "+", 
               RowBox[{"c", " ", 
                RowBox[{"(", 
                 RowBox[{"\[Delta]", "+", 
                  RowBox[{"sf", " ", 
                   RowBox[{"(", 
                    RowBox[{"4", "-", 
                    RowBox[{"8", " ", "h"}], "+", 
                    RowBox[{"4", " ", "\[Delta]"}], "-", 
                    RowBox[{"5", " ", "h", " ", "\[Delta]"}]}], ")"}]}]}], 
                 ")"}]}]}], ")"}], "2"]}]]}], "-", 
         RowBox[{"3", " ", "c", " ", 
          SuperscriptBox["h", "2"], " ", 
          SuperscriptBox["sf", "3"], " ", 
          SqrtBox[
           RowBox[{
            RowBox[{
             RowBox[{"-", "8"}], " ", 
             RowBox[{"(", 
              RowBox[{
               RowBox[{"-", "1"}], "+", "c"}], ")"}], " ", "sf", " ", 
             RowBox[{"(", 
              RowBox[{
               RowBox[{"-", "1"}], "-", "\[Delta]", "+", 
               RowBox[{"h", " ", 
                RowBox[{"(", 
                 RowBox[{"2", "+", "\[Delta]"}], ")"}]}]}], ")"}], " ", 
             RowBox[{"(", 
              RowBox[{"1", "+", 
               RowBox[{"h", " ", "sf"}], "+", 
               RowBox[{"c", " ", 
                RowBox[{"(", 
                 RowBox[{
                  RowBox[{"-", "1"}], "-", 
                  RowBox[{"2", " ", "\[Delta]"}], "+", 
                  RowBox[{"h", " ", "sf", " ", 
                   RowBox[{"(", 
                    RowBox[{"3", "+", 
                    RowBox[{"2", " ", "\[Delta]"}]}], ")"}]}]}], ")"}]}]}], 
              ")"}]}], "+", 
            SuperscriptBox[
             RowBox[{"(", 
              RowBox[{
               RowBox[{
                RowBox[{"(", 
                 RowBox[{
                  RowBox[{"-", "1"}], "+", 
                  RowBox[{"h", " ", "sf"}]}], ")"}], " ", "\[Delta]"}], "+", 
               RowBox[{"c", " ", 
                RowBox[{"(", 
                 RowBox[{"\[Delta]", "+", 
                  RowBox[{"sf", " ", 
                   RowBox[{"(", 
                    RowBox[{"4", "-", 
                    RowBox[{"8", " ", "h"}], "+", 
                    RowBox[{"4", " ", "\[Delta]"}], "-", 
                    RowBox[{"5", " ", "h", " ", "\[Delta]"}]}], ")"}]}]}], 
                 ")"}]}]}], ")"}], "2"]}]]}], "-", 
         RowBox[{"\[Delta]", " ", 
          SqrtBox[
           RowBox[{
            RowBox[{
             RowBox[{"-", "8"}], " ", 
             RowBox[{"(", 
              RowBox[{
               RowBox[{"-", "1"}], "+", "c"}], ")"}], " ", "sf", " ", 
             RowBox[{"(", 
              RowBox[{
               RowBox[{"-", "1"}], "-", "\[Delta]", "+", 
               RowBox[{"h", " ", 
                RowBox[{"(", 
                 RowBox[{"2", "+", "\[Delta]"}], ")"}]}]}], ")"}], " ", 
             RowBox[{"(", 
              RowBox[{"1", "+", 
               RowBox[{"h", " ", "sf"}], "+", 
               RowBox[{"c", " ", 
                RowBox[{"(", 
                 RowBox[{
                  RowBox[{"-", "1"}], "-", 
                  RowBox[{"2", " ", "\[Delta]"}], "+", 
                  RowBox[{"h", " ", "sf", " ", 
                   RowBox[{"(", 
                    RowBox[{"3", "+", 
                    RowBox[{"2", " ", "\[Delta]"}]}], ")"}]}]}], ")"}]}]}], 
              ")"}]}], "+", 
            SuperscriptBox[
             RowBox[{"(", 
              RowBox[{
               RowBox[{
                RowBox[{"(", 
                 RowBox[{
                  RowBox[{"-", "1"}], "+", 
                  RowBox[{"h", " ", "sf"}]}], ")"}], " ", "\[Delta]"}], "+", 
               RowBox[{"c", " ", 
                RowBox[{"(", 
                 RowBox[{"\[Delta]", "+", 
                  RowBox[{"sf", " ", 
                   RowBox[{"(", 
                    RowBox[{"4", "-", 
                    RowBox[{"8", " ", "h"}], "+", 
                    RowBox[{"4", " ", "\[Delta]"}], "-", 
                    RowBox[{"5", " ", "h", " ", "\[Delta]"}]}], ")"}]}]}], 
                 ")"}]}]}], ")"}], "2"]}]]}], "+", 
         RowBox[{"c", " ", "\[Delta]", " ", 
          SqrtBox[
           RowBox[{
            RowBox[{
             RowBox[{"-", "8"}], " ", 
             RowBox[{"(", 
              RowBox[{
               RowBox[{"-", "1"}], "+", "c"}], ")"}], " ", "sf", " ", 
             RowBox[{"(", 
              RowBox[{
               RowBox[{"-", "1"}], "-", "\[Delta]", "+", 
               RowBox[{"h", " ", 
                RowBox[{"(", 
                 RowBox[{"2", "+", "\[Delta]"}], ")"}]}]}], ")"}], " ", 
             RowBox[{"(", 
              RowBox[{"1", "+", 
               RowBox[{"h", " ", "sf"}], "+", 
               RowBox[{"c", " ", 
                RowBox[{"(", 
                 RowBox[{
                  RowBox[{"-", "1"}], "-", 
                  RowBox[{"2", " ", "\[Delta]"}], "+", 
                  RowBox[{"h", " ", "sf", " ", 
                   RowBox[{"(", 
                    RowBox[{"3", "+", 
                    RowBox[{"2", " ", "\[Delta]"}]}], ")"}]}]}], ")"}]}]}], 
              ")"}]}], "+", 
            SuperscriptBox[
             RowBox[{"(", 
              RowBox[{
               RowBox[{
                RowBox[{"(", 
                 RowBox[{
                  RowBox[{"-", "1"}], "+", 
                  RowBox[{"h", " ", "sf"}]}], ")"}], " ", "\[Delta]"}], "+", 
               RowBox[{"c", " ", 
                RowBox[{"(", 
                 RowBox[{"\[Delta]", "+", 
                  RowBox[{"sf", " ", 
                   RowBox[{"(", 
                    RowBox[{"4", "-", 
                    RowBox[{"8", " ", "h"}], "+", 
                    RowBox[{"4", " ", "\[Delta]"}], "-", 
                    RowBox[{"5", " ", "h", " ", "\[Delta]"}]}], ")"}]}]}], 
                 ")"}]}]}], ")"}], "2"]}]]}], "+", 
         RowBox[{"2", " ", "sf", " ", "\[Delta]", " ", 
          SqrtBox[
           RowBox[{
            RowBox[{
             RowBox[{"-", "8"}], " ", 
             RowBox[{"(", 
              RowBox[{
               RowBox[{"-", "1"}], "+", "c"}], ")"}], " ", "sf", " ", 
             RowBox[{"(", 
              RowBox[{
               RowBox[{"-", "1"}], "-", "\[Delta]", "+", 
               RowBox[{"h", " ", 
                RowBox[{"(", 
                 RowBox[{"2", "+", "\[Delta]"}], ")"}]}]}], ")"}], " ", 
             RowBox[{"(", 
              RowBox[{"1", "+", 
               RowBox[{"h", " ", "sf"}], "+", 
               RowBox[{"c", " ", 
                RowBox[{"(", 
                 RowBox[{
                  RowBox[{"-", "1"}], "-", 
                  RowBox[{"2", " ", "\[Delta]"}], "+", 
                  RowBox[{"h", " ", "sf", " ", 
                   RowBox[{"(", 
                    RowBox[{"3", "+", 
                    RowBox[{"2", " ", "\[Delta]"}]}], ")"}]}]}], ")"}]}]}], 
              ")"}]}], "+", 
            SuperscriptBox[
             RowBox[{"(", 
              RowBox[{
               RowBox[{
                RowBox[{"(", 
                 RowBox[{
                  RowBox[{"-", "1"}], "+", 
                  RowBox[{"h", " ", "sf"}]}], ")"}], " ", "\[Delta]"}], "+", 
               RowBox[{"c", " ", 
                RowBox[{"(", 
                 RowBox[{"\[Delta]", "+", 
                  RowBox[{"sf", " ", 
                   RowBox[{"(", 
                    RowBox[{"4", "-", 
                    RowBox[{"8", " ", "h"}], "+", 
                    RowBox[{"4", " ", "\[Delta]"}], "-", 
                    RowBox[{"5", " ", "h", " ", "\[Delta]"}]}], ")"}]}]}], 
                 ")"}]}]}], ")"}], "2"]}]]}], "-", 
         RowBox[{"4", " ", "c", " ", "h", " ", "sf", " ", "\[Delta]", " ", 
          SqrtBox[
           RowBox[{
            RowBox[{
             RowBox[{"-", "8"}], " ", 
             RowBox[{"(", 
              RowBox[{
               RowBox[{"-", "1"}], "+", "c"}], ")"}], " ", "sf", " ", 
             RowBox[{"(", 
              RowBox[{
               RowBox[{"-", "1"}], "-", "\[Delta]", "+", 
               RowBox[{"h", " ", 
                RowBox[{"(", 
                 RowBox[{"2", "+", "\[Delta]"}], ")"}]}]}], ")"}], " ", 
             RowBox[{"(", 
              RowBox[{"1", "+", 
               RowBox[{"h", " ", "sf"}], "+", 
               RowBox[{"c", " ", 
                RowBox[{"(", 
                 RowBox[{
                  RowBox[{"-", "1"}], "-", 
                  RowBox[{"2", " ", "\[Delta]"}], "+", 
                  RowBox[{"h", " ", "sf", " ", 
                   RowBox[{"(", 
                    RowBox[{"3", "+", 
                    RowBox[{"2", " ", "\[Delta]"}]}], ")"}]}]}], ")"}]}]}], 
              ")"}]}], "+", 
            SuperscriptBox[
             RowBox[{"(", 
              RowBox[{
               RowBox[{
                RowBox[{"(", 
                 RowBox[{
                  RowBox[{"-", "1"}], "+", 
                  RowBox[{"h", " ", "sf"}]}], ")"}], " ", "\[Delta]"}], "+", 
               RowBox[{"c", " ", 
                RowBox[{"(", 
                 RowBox[{"\[Delta]", "+", 
                  RowBox[{"sf", " ", 
                   RowBox[{"(", 
                    RowBox[{"4", "-", 
                    RowBox[{"8", " ", "h"}], "+", 
                    RowBox[{"4", " ", "\[Delta]"}], "-", 
                    RowBox[{"5", " ", "h", " ", "\[Delta]"}]}], ")"}]}]}], 
                 ")"}]}]}], ")"}], "2"]}]]}], "-", 
         RowBox[{"2", " ", "h", " ", 
          SuperscriptBox["sf", "2"], " ", "\[Delta]", " ", 
          SqrtBox[
           RowBox[{
            RowBox[{
             RowBox[{"-", "8"}], " ", 
             RowBox[{"(", 
              RowBox[{
               RowBox[{"-", "1"}], "+", "c"}], ")"}], " ", "sf", " ", 
             RowBox[{"(", 
              RowBox[{
               RowBox[{"-", "1"}], "-", "\[Delta]", "+", 
               RowBox[{"h", " ", 
                RowBox[{"(", 
                 RowBox[{"2", "+", "\[Delta]"}], ")"}]}]}], ")"}], " ", 
             RowBox[{"(", 
              RowBox[{"1", "+", 
               RowBox[{"h", " ", "sf"}], "+", 
               RowBox[{"c", " ", 
                RowBox[{"(", 
                 RowBox[{
                  RowBox[{"-", "1"}], "-", 
                  RowBox[{"2", " ", "\[Delta]"}], "+", 
                  RowBox[{"h", " ", "sf", " ", 
                   RowBox[{"(", 
                    RowBox[{"3", "+", 
                    RowBox[{"2", " ", "\[Delta]"}]}], ")"}]}]}], ")"}]}]}], 
              ")"}]}], "+", 
            SuperscriptBox[
             RowBox[{"(", 
              RowBox[{
               RowBox[{
                RowBox[{"(", 
                 RowBox[{
                  RowBox[{"-", "1"}], "+", 
                  RowBox[{"h", " ", "sf"}]}], ")"}], " ", "\[Delta]"}], "+", 
               RowBox[{"c", " ", 
                RowBox[{"(", 
                 RowBox[{"\[Delta]", "+", 
                  RowBox[{"sf", " ", 
                   RowBox[{"(", 
                    RowBox[{"4", "-", 
                    RowBox[{"8", " ", "h"}], "+", 
                    RowBox[{"4", " ", "\[Delta]"}], "-", 
                    RowBox[{"5", " ", "h", " ", "\[Delta]"}]}], ")"}]}]}], 
                 ")"}]}]}], ")"}], "2"]}]]}], "+", 
         RowBox[{"2", " ", "c", " ", "h", " ", 
          SuperscriptBox["sf", "2"], " ", "\[Delta]", " ", 
          SqrtBox[
           RowBox[{
            RowBox[{
             RowBox[{"-", "8"}], " ", 
             RowBox[{"(", 
              RowBox[{
               RowBox[{"-", "1"}], "+", "c"}], ")"}], " ", "sf", " ", 
             RowBox[{"(", 
              RowBox[{
               RowBox[{"-", "1"}], "-", "\[Delta]", "+", 
               RowBox[{"h", " ", 
                RowBox[{"(", 
                 RowBox[{"2", "+", "\[Delta]"}], ")"}]}]}], ")"}], " ", 
             RowBox[{"(", 
              RowBox[{"1", "+", 
               RowBox[{"h", " ", "sf"}], "+", 
               RowBox[{"c", " ", 
                RowBox[{"(", 
                 RowBox[{
                  RowBox[{"-", "1"}], "-", 
                  RowBox[{"2", " ", "\[Delta]"}], "+", 
                  RowBox[{"h", " ", "sf", " ", 
                   RowBox[{"(", 
                    RowBox[{"3", "+", 
                    RowBox[{"2", " ", "\[Delta]"}]}], ")"}]}]}], ")"}]}]}], 
              ")"}]}], "+", 
            SuperscriptBox[
             RowBox[{"(", 
              RowBox[{
               RowBox[{
                RowBox[{"(", 
                 RowBox[{
                  RowBox[{"-", "1"}], "+", 
                  RowBox[{"h", " ", "sf"}]}], ")"}], " ", "\[Delta]"}], "+", 
               RowBox[{"c", " ", 
                RowBox[{"(", 
                 RowBox[{"\[Delta]", "+", 
                  RowBox[{"sf", " ", 
                   RowBox[{"(", 
                    RowBox[{"4", "-", 
                    RowBox[{"8", " ", "h"}], "+", 
                    RowBox[{"4", " ", "\[Delta]"}], "-", 
                    RowBox[{"5", " ", "h", " ", "\[Delta]"}]}], ")"}]}]}], 
                 ")"}]}]}], ")"}], "2"]}]]}], "+", 
         RowBox[{
          SuperscriptBox["h", "2"], " ", 
          SuperscriptBox["sf", "2"], " ", "\[Delta]", " ", 
          SqrtBox[
           RowBox[{
            RowBox[{
             RowBox[{"-", "8"}], " ", 
             RowBox[{"(", 
              RowBox[{
               RowBox[{"-", "1"}], "+", "c"}], ")"}], " ", "sf", " ", 
             RowBox[{"(", 
              RowBox[{
               RowBox[{"-", "1"}], "-", "\[Delta]", "+", 
               RowBox[{"h", " ", 
                RowBox[{"(", 
                 RowBox[{"2", "+", "\[Delta]"}], ")"}]}]}], ")"}], " ", 
             RowBox[{"(", 
              RowBox[{"1", "+", 
               RowBox[{"h", " ", "sf"}], "+", 
               RowBox[{"c", " ", 
                RowBox[{"(", 
                 RowBox[{
                  RowBox[{"-", "1"}], "-", 
                  RowBox[{"2", " ", "\[Delta]"}], "+", 
                  RowBox[{"h", " ", "sf", " ", 
                   RowBox[{"(", 
                    RowBox[{"3", "+", 
                    RowBox[{"2", " ", "\[Delta]"}]}], ")"}]}]}], ")"}]}]}], 
              ")"}]}], "+", 
            SuperscriptBox[
             RowBox[{"(", 
              RowBox[{
               RowBox[{
                RowBox[{"(", 
                 RowBox[{
                  RowBox[{"-", "1"}], "+", 
                  RowBox[{"h", " ", "sf"}]}], ")"}], " ", "\[Delta]"}], "+", 
               RowBox[{"c", " ", 
                RowBox[{"(", 
                 RowBox[{"\[Delta]", "+", 
                  RowBox[{"sf", " ", 
                   RowBox[{"(", 
                    RowBox[{"4", "-", 
                    RowBox[{"8", " ", "h"}], "+", 
                    RowBox[{"4", " ", "\[Delta]"}], "-", 
                    RowBox[{"5", " ", "h", " ", "\[Delta]"}]}], ")"}]}]}], 
                 ")"}]}]}], ")"}], "2"]}]]}], "+", 
         RowBox[{"3", " ", "c", " ", 
          SuperscriptBox["h", "2"], " ", 
          SuperscriptBox["sf", "2"], " ", "\[Delta]", " ", 
          SqrtBox[
           RowBox[{
            RowBox[{
             RowBox[{"-", "8"}], " ", 
             RowBox[{"(", 
              RowBox[{
               RowBox[{"-", "1"}], "+", "c"}], ")"}], " ", "sf", " ", 
             RowBox[{"(", 
              RowBox[{
               RowBox[{"-", "1"}], "-", "\[Delta]", "+", 
               RowBox[{"h", " ", 
                RowBox[{"(", 
                 RowBox[{"2", "+", "\[Delta]"}], ")"}]}]}], ")"}], " ", 
             RowBox[{"(", 
              RowBox[{"1", "+", 
               RowBox[{"h", " ", "sf"}], "+", 
               RowBox[{"c", " ", 
                RowBox[{"(", 
                 RowBox[{
                  RowBox[{"-", "1"}], "-", 
                  RowBox[{"2", " ", "\[Delta]"}], "+", 
                  RowBox[{"h", " ", "sf", " ", 
                   RowBox[{"(", 
                    RowBox[{"3", "+", 
                    RowBox[{"2", " ", "\[Delta]"}]}], ")"}]}]}], ")"}]}]}], 
              ")"}]}], "+", 
            SuperscriptBox[
             RowBox[{"(", 
              RowBox[{
               RowBox[{
                RowBox[{"(", 
                 RowBox[{
                  RowBox[{"-", "1"}], "+", 
                  RowBox[{"h", " ", "sf"}]}], ")"}], " ", "\[Delta]"}], "+", 
               RowBox[{"c", " ", 
                RowBox[{"(", 
                 RowBox[{"\[Delta]", "+", 
                  RowBox[{"sf", " ", 
                   RowBox[{"(", 
                    RowBox[{"4", "-", 
                    RowBox[{"8", " ", "h"}], "+", 
                    RowBox[{"4", " ", "\[Delta]"}], "-", 
                    RowBox[{"5", " ", "h", " ", "\[Delta]"}]}], ")"}]}]}], 
                 ")"}]}]}], ")"}], "2"]}]]}], "-", 
         RowBox[{"2", " ", "c", " ", 
          SuperscriptBox["h", "2"], " ", 
          SuperscriptBox["sf", "3"], " ", "\[Delta]", " ", 
          SqrtBox[
           RowBox[{
            RowBox[{
             RowBox[{"-", "8"}], " ", 
             RowBox[{"(", 
              RowBox[{
               RowBox[{"-", "1"}], "+", "c"}], ")"}], " ", "sf", " ", 
             RowBox[{"(", 
              RowBox[{
               RowBox[{"-", "1"}], "-", "\[Delta]", "+", 
               RowBox[{"h", " ", 
                RowBox[{"(", 
                 RowBox[{"2", "+", "\[Delta]"}], ")"}]}]}], ")"}], " ", 
             RowBox[{"(", 
              RowBox[{"1", "+", 
               RowBox[{"h", " ", "sf"}], "+", 
               RowBox[{"c", " ", 
                RowBox[{"(", 
                 RowBox[{
                  RowBox[{"-", "1"}], "-", 
                  RowBox[{"2", " ", "\[Delta]"}], "+", 
                  RowBox[{"h", " ", "sf", " ", 
                   RowBox[{"(", 
                    RowBox[{"3", "+", 
                    RowBox[{"2", " ", "\[Delta]"}]}], ")"}]}]}], ")"}]}]}], 
              ")"}]}], "+", 
            SuperscriptBox[
             RowBox[{"(", 
              RowBox[{
               RowBox[{
                RowBox[{"(", 
                 RowBox[{
                  RowBox[{"-", "1"}], "+", 
                  RowBox[{"h", " ", "sf"}]}], ")"}], " ", "\[Delta]"}], "+", 
               RowBox[{"c", " ", 
                RowBox[{"(", 
                 RowBox[{"\[Delta]", "+", 
                  RowBox[{"sf", " ", 
                   RowBox[{"(", 
                    RowBox[{"4", "-", 
                    RowBox[{"8", " ", "h"}], "+", 
                    RowBox[{"4", " ", "\[Delta]"}], "-", 
                    RowBox[{"5", " ", "h", " ", "\[Delta]"}]}], ")"}]}]}], 
                 ")"}]}]}], ")"}], "2"]}]]}]}], ")"}], "/", 
       RowBox[{"(", 
        RowBox[{"2", " ", 
         RowBox[{"(", 
          RowBox[{
           RowBox[{
            SuperscriptBox[
             RowBox[{"(", 
              RowBox[{
               RowBox[{"-", "1"}], "+", "c"}], ")"}], "2"], " ", 
            SuperscriptBox["\[Delta]", "2"]}], "+", 
           RowBox[{
            SuperscriptBox["h", "3"], " ", 
            SuperscriptBox["sf", "4"], " ", 
            SuperscriptBox[
             RowBox[{"(", 
              RowBox[{"1", "+", 
               RowBox[{"c", " ", 
                RowBox[{"(", 
                 RowBox[{"3", "+", 
                  RowBox[{"2", " ", "\[Delta]"}]}], ")"}]}]}], ")"}], "2"]}], 
           "+", 
           RowBox[{"sf", " ", 
            RowBox[{"(", 
             RowBox[{"1", "-", 
              RowBox[{"2", " ", "\[Delta]"}], "+", 
              RowBox[{"h", " ", 
               RowBox[{"(", 
                RowBox[{
                 RowBox[{"-", "2"}], "+", 
                 RowBox[{"2", " ", "\[Delta]"}], "-", 
                 RowBox[{"3", " ", 
                  SuperscriptBox["\[Delta]", "2"]}]}], ")"}]}], "+", 
              RowBox[{"2", " ", "c", " ", 
               RowBox[{"(", 
                RowBox[{
                 RowBox[{"-", "1"}], "+", 
                 RowBox[{"2", " ", 
                  SuperscriptBox["\[Delta]", "2"]}], "+", 
                 RowBox[{"h", " ", 
                  RowBox[{"(", 
                   RowBox[{"2", "+", 
                    RowBox[{"2", " ", "\[Delta]"}], "+", 
                    SuperscriptBox["\[Delta]", "2"]}], ")"}]}]}], ")"}]}], 
              "+", 
              RowBox[{
               SuperscriptBox["c", "2"], " ", 
               RowBox[{"(", 
                RowBox[{"1", "+", 
                 RowBox[{"2", " ", "\[Delta]"}], "-", 
                 RowBox[{"h", " ", 
                  RowBox[{"(", 
                   RowBox[{"2", "+", 
                    RowBox[{"6", " ", "\[Delta]"}], "+", 
                    RowBox[{"7", " ", 
                    SuperscriptBox["\[Delta]", "2"]}]}], ")"}]}]}], ")"}]}]}],
              ")"}]}], "-", 
           RowBox[{"h", " ", 
            SuperscriptBox["sf", "3"], " ", 
            RowBox[{"(", 
             RowBox[{
              RowBox[{"2", " ", "c", " ", 
               RowBox[{"(", 
                RowBox[{
                 RowBox[{
                  RowBox[{"-", "4"}], " ", 
                  RowBox[{"(", 
                   RowBox[{"1", "+", "\[Delta]"}], ")"}]}], "+", 
                 RowBox[{"h", " ", 
                  RowBox[{"(", 
                   RowBox[{"3", "-", 
                    RowBox[{"2", " ", 
                    SuperscriptBox["\[Delta]", "2"]}]}], ")"}]}], "+", 
                 RowBox[{
                  SuperscriptBox["h", "2"], " ", 
                  RowBox[{"(", 
                   RowBox[{"6", "+", 
                    RowBox[{"6", " ", "\[Delta]"}], "+", 
                    SuperscriptBox["\[Delta]", "2"]}], ")"}]}]}], ")"}]}], 
              "+", 
              RowBox[{"h", " ", 
               RowBox[{"(", 
                RowBox[{
                 RowBox[{"-", "7"}], "-", 
                 RowBox[{"6", " ", "\[Delta]"}], "+", 
                 RowBox[{"h", " ", 
                  RowBox[{"(", 
                   RowBox[{"10", "+", 
                    RowBox[{"6", " ", "\[Delta]"}], "+", 
                    SuperscriptBox["\[Delta]", "2"]}], ")"}]}]}], ")"}]}], 
              "+", 
              RowBox[{
               SuperscriptBox["c", "2"], " ", 
               RowBox[{"(", 
                RowBox[{
                 RowBox[{
                  RowBox[{"-", "8"}], " ", 
                  RowBox[{"(", 
                   RowBox[{"1", "+", "\[Delta]"}], ")"}]}], "+", 
                 RowBox[{
                  SuperscriptBox["h", "2"], " ", 
                  RowBox[{"(", 
                   RowBox[{"10", "+", 
                    RowBox[{"14", " ", "\[Delta]"}], "+", 
                    RowBox[{"5", " ", 
                    SuperscriptBox["\[Delta]", "2"]}]}], ")"}]}], "+", 
                 RowBox[{"h", " ", 
                  RowBox[{"(", 
                   RowBox[{"17", "+", 
                    RowBox[{"22", " ", "\[Delta]"}], "+", 
                    RowBox[{"8", " ", 
                    SuperscriptBox["\[Delta]", "2"]}]}], ")"}]}]}], ")"}]}]}],
              ")"}]}], "+", 
           RowBox[{
            SuperscriptBox["sf", "2"], " ", 
            RowBox[{"(", 
             RowBox[{
              RowBox[{"8", " ", 
               RowBox[{"(", 
                RowBox[{"1", "+", "\[Delta]"}], ")"}]}], "-", 
              RowBox[{"5", " ", "h", " ", 
               RowBox[{"(", 
                RowBox[{"5", "+", 
                 RowBox[{"4", " ", "\[Delta]"}]}], ")"}]}], "+", 
              RowBox[{
               SuperscriptBox["h", "2"], " ", 
               RowBox[{"(", 
                RowBox[{"20", "+", 
                 RowBox[{"12", " ", "\[Delta]"}], "+", 
                 RowBox[{"3", " ", 
                  SuperscriptBox["\[Delta]", "2"]}]}], ")"}]}], "+", 
              RowBox[{"2", " ", "c", " ", 
               RowBox[{"(", 
                RowBox[{
                 RowBox[{"4", " ", 
                  RowBox[{"(", 
                   RowBox[{"1", "+", "\[Delta]"}], ")"}]}], "+", 
                 RowBox[{
                  SuperscriptBox["h", "2"], " ", 
                  RowBox[{"(", 
                   RowBox[{"12", "+", 
                    RowBox[{"12", " ", "\[Delta]"}], "+", 
                    SuperscriptBox["\[Delta]", "2"]}], ")"}]}], "-", 
                 RowBox[{"h", " ", 
                  RowBox[{"(", 
                   RowBox[{"15", "+", 
                    RowBox[{"18", " ", "\[Delta]"}], "+", 
                    RowBox[{"4", " ", 
                    SuperscriptBox["\[Delta]", "2"]}]}], ")"}]}]}], ")"}]}], 
              "+", 
              RowBox[{
               SuperscriptBox["c", "2"], " ", "h", " ", 
               RowBox[{"(", 
                RowBox[{
                 RowBox[{"-", "9"}], "-", 
                 RowBox[{"8", " ", "\[Delta]"}], "+", 
                 RowBox[{"4", " ", 
                  SuperscriptBox["\[Delta]", "2"]}], "+", 
                 RowBox[{"h", " ", 
                  RowBox[{"(", 
                   RowBox[{"20", "+", 
                    RowBox[{"28", " ", "\[Delta]"}], "+", 
                    RowBox[{"11", " ", 
                    SuperscriptBox["\[Delta]", "2"]}]}], ")"}]}]}], ")"}]}]}],
              ")"}]}]}], ")"}]}], ")"}]}]}]}], "}"}], ",", 
   RowBox[{"{", 
    RowBox[{
     RowBox[{"pm", "\[Rule]", 
      RowBox[{
       FractionBox["1", 
        RowBox[{"4", " ", 
         RowBox[{"(", 
          RowBox[{
           RowBox[{"-", "1"}], "+", "c"}], ")"}], " ", "sf", " ", 
         RowBox[{"(", 
          RowBox[{
           RowBox[{"-", "1"}], "-", "\[Delta]", "+", 
           RowBox[{"h", " ", 
            RowBox[{"(", 
             RowBox[{"2", "+", "\[Delta]"}], ")"}]}]}], ")"}]}]], 
       RowBox[{"(", 
        RowBox[{
         RowBox[{
          RowBox[{"-", "4"}], " ", "c", " ", "sf"}], "+", 
         RowBox[{"8", " ", "c", " ", "h", " ", "sf"}], "+", "\[Delta]", "-", 
         RowBox[{"c", " ", "\[Delta]"}], "-", 
         RowBox[{"4", " ", "c", " ", "sf", " ", "\[Delta]"}], "-", 
         RowBox[{"h", " ", "sf", " ", "\[Delta]"}], "+", 
         RowBox[{"5", " ", "c", " ", "h", " ", "sf", " ", "\[Delta]"}], "+", 
         SqrtBox[
          RowBox[{
           RowBox[{
            RowBox[{"-", "8"}], " ", 
            RowBox[{"(", 
             RowBox[{
              RowBox[{"-", "1"}], "+", "c"}], ")"}], " ", "sf", " ", 
            RowBox[{"(", 
             RowBox[{
              RowBox[{"-", "1"}], "-", "\[Delta]", "+", 
              RowBox[{"h", " ", 
               RowBox[{"(", 
                RowBox[{"2", "+", "\[Delta]"}], ")"}]}]}], ")"}], " ", 
            RowBox[{"(", 
             RowBox[{"1", "+", 
              RowBox[{"h", " ", "sf"}], "+", 
              RowBox[{"c", " ", 
               RowBox[{"(", 
                RowBox[{
                 RowBox[{"-", "1"}], "-", 
                 RowBox[{"2", " ", "\[Delta]"}], "+", 
                 RowBox[{"h", " ", "sf", " ", 
                  RowBox[{"(", 
                   RowBox[{"3", "+", 
                    RowBox[{"2", " ", "\[Delta]"}]}], ")"}]}]}], ")"}]}]}], 
             ")"}]}], "+", 
           SuperscriptBox[
            RowBox[{"(", 
             RowBox[{
              RowBox[{
               RowBox[{"(", 
                RowBox[{
                 RowBox[{"-", "1"}], "+", 
                 RowBox[{"h", " ", "sf"}]}], ")"}], " ", "\[Delta]"}], "+", 
              RowBox[{"c", " ", 
               RowBox[{"(", 
                RowBox[{"\[Delta]", "+", 
                 RowBox[{"sf", " ", 
                  RowBox[{"(", 
                   RowBox[{"4", "-", 
                    RowBox[{"8", " ", "h"}], "+", 
                    RowBox[{"4", " ", "\[Delta]"}], "-", 
                    RowBox[{"5", " ", "h", " ", "\[Delta]"}]}], ")"}]}]}], 
                ")"}]}]}], ")"}], "2"]}]]}], ")"}]}]}], ",", 
     RowBox[{"pf", "\[Rule]", 
      RowBox[{
       FractionBox["1", 
        RowBox[{"4", " ", 
         RowBox[{"(", 
          RowBox[{
           RowBox[{"-", "1"}], "+", "c"}], ")"}], " ", "sf", " ", 
         RowBox[{"(", 
          RowBox[{
           RowBox[{"-", "1"}], "-", "\[Delta]", "+", 
           RowBox[{"h", " ", 
            RowBox[{"(", 
             RowBox[{"2", "+", "\[Delta]"}], ")"}]}]}], ")"}]}]], 
       RowBox[{"(", 
        RowBox[{
         RowBox[{
          RowBox[{"-", "4"}], " ", "c", " ", "sf"}], "+", 
         RowBox[{"8", " ", "c", " ", "h", " ", "sf"}], "+", "\[Delta]", "-", 
         RowBox[{"c", " ", "\[Delta]"}], "-", 
         RowBox[{"4", " ", "c", " ", "sf", " ", "\[Delta]"}], "-", 
         RowBox[{"h", " ", "sf", " ", "\[Delta]"}], "+", 
         RowBox[{"5", " ", "c", " ", "h", " ", "sf", " ", "\[Delta]"}], "+", 
         SqrtBox[
          RowBox[{
           RowBox[{
            RowBox[{"-", "8"}], " ", 
            RowBox[{"(", 
             RowBox[{
              RowBox[{"-", "1"}], "+", "c"}], ")"}], " ", "sf", " ", 
            RowBox[{"(", 
             RowBox[{
              RowBox[{"-", "1"}], "-", "\[Delta]", "+", 
              RowBox[{"h", " ", 
               RowBox[{"(", 
                RowBox[{"2", "+", "\[Delta]"}], ")"}]}]}], ")"}], " ", 
            RowBox[{"(", 
             RowBox[{"1", "+", 
              RowBox[{"h", " ", "sf"}], "+", 
              RowBox[{"c", " ", 
               RowBox[{"(", 
                RowBox[{
                 RowBox[{"-", "1"}], "-", 
                 RowBox[{"2", " ", "\[Delta]"}], "+", 
                 RowBox[{"h", " ", "sf", " ", 
                  RowBox[{"(", 
                   RowBox[{"3", "+", 
                    RowBox[{"2", " ", "\[Delta]"}]}], ")"}]}]}], ")"}]}]}], 
             ")"}]}], "+", 
           SuperscriptBox[
            RowBox[{"(", 
             RowBox[{
              RowBox[{
               RowBox[{"(", 
                RowBox[{
                 RowBox[{"-", "1"}], "+", 
                 RowBox[{"h", " ", "sf"}]}], ")"}], " ", "\[Delta]"}], "+", 
              RowBox[{"c", " ", 
               RowBox[{"(", 
                RowBox[{"\[Delta]", "+", 
                 RowBox[{"sf", " ", 
                  RowBox[{"(", 
                   RowBox[{"4", "-", 
                    RowBox[{"8", " ", "h"}], "+", 
                    RowBox[{"4", " ", "\[Delta]"}], "-", 
                    RowBox[{"5", " ", "h", " ", "\[Delta]"}]}], ")"}]}]}], 
                ")"}]}]}], ")"}], "2"]}]]}], ")"}]}]}], ",", 
     RowBox[{"F", "\[Rule]", 
      RowBox[{
       RowBox[{"(", 
        RowBox[{
         RowBox[{
          RowBox[{"-", "2"}], " ", "sf"}], "+", 
         RowBox[{"4", " ", "c", " ", "sf"}], "-", 
         RowBox[{"2", " ", 
          SuperscriptBox["c", "2"], " ", "sf"}], "+", 
         RowBox[{"4", " ", "h", " ", "sf"}], "-", 
         RowBox[{"8", " ", "c", " ", "h", " ", "sf"}], "+", 
         RowBox[{"4", " ", 
          SuperscriptBox["c", "2"], " ", "h", " ", "sf"}], "+", 
         RowBox[{"8", " ", 
          SuperscriptBox["sf", "2"]}], "+", 
         RowBox[{"4", " ", "c", " ", 
          SuperscriptBox["sf", "2"]}], "+", 
         RowBox[{"4", " ", 
          SuperscriptBox["c", "2"], " ", 
          SuperscriptBox["sf", "2"]}], "-", 
         RowBox[{"18", " ", "h", " ", 
          SuperscriptBox["sf", "2"]}], "-", 
         RowBox[{"36", " ", "c", " ", "h", " ", 
          SuperscriptBox["sf", "2"]}], "-", 
         RowBox[{"10", " ", 
          SuperscriptBox["c", "2"], " ", "h", " ", 
          SuperscriptBox["sf", "2"]}], "+", 
         RowBox[{"8", " ", 
          SuperscriptBox["h", "2"], " ", 
          SuperscriptBox["sf", "2"]}], "+", 
         RowBox[{"48", " ", "c", " ", 
          SuperscriptBox["h", "2"], " ", 
          SuperscriptBox["sf", "2"]}], "+", 
         RowBox[{"8", " ", 
          SuperscriptBox["c", "2"], " ", 
          SuperscriptBox["h", "2"], " ", 
          SuperscriptBox["sf", "2"]}], "+", 
         RowBox[{"8", " ", "h", " ", 
          SuperscriptBox["sf", "3"]}], "+", 
         RowBox[{"24", " ", "c", " ", "h", " ", 
          SuperscriptBox["sf", "3"]}], "-", 
         RowBox[{"14", " ", 
          SuperscriptBox["h", "2"], " ", 
          SuperscriptBox["sf", "3"]}], "-", 
         RowBox[{"36", " ", "c", " ", 
          SuperscriptBox["h", "2"], " ", 
          SuperscriptBox["sf", "3"]}], "-", 
         RowBox[{"14", " ", 
          SuperscriptBox["c", "2"], " ", 
          SuperscriptBox["h", "2"], " ", 
          SuperscriptBox["sf", "3"]}], "+", 
         RowBox[{"4", " ", 
          SuperscriptBox["h", "3"], " ", 
          SuperscriptBox["sf", "3"]}], "-", 
         RowBox[{"8", " ", "c", " ", 
          SuperscriptBox["h", "3"], " ", 
          SuperscriptBox["sf", "3"]}], "+", 
         RowBox[{"4", " ", 
          SuperscriptBox["c", "2"], " ", 
          SuperscriptBox["h", "3"], " ", 
          SuperscriptBox["sf", "3"]}], "+", 
         RowBox[{"4", " ", "c", " ", 
          SuperscriptBox["h", "2"], " ", 
          SuperscriptBox["sf", "4"]}], "+", 
         RowBox[{"12", " ", 
          SuperscriptBox["c", "2"], " ", 
          SuperscriptBox["h", "2"], " ", 
          SuperscriptBox["sf", "4"]}], "+", 
         RowBox[{"2", " ", 
          SuperscriptBox["h", "3"], " ", 
          SuperscriptBox["sf", "4"]}], "+", 
         RowBox[{"4", " ", "c", " ", 
          SuperscriptBox["h", "3"], " ", 
          SuperscriptBox["sf", "4"]}], "-", 
         RowBox[{"6", " ", 
          SuperscriptBox["c", "2"], " ", 
          SuperscriptBox["h", "3"], " ", 
          SuperscriptBox["sf", "4"]}], "-", 
         RowBox[{"9", " ", "sf", " ", "\[Delta]"}], "+", 
         RowBox[{"10", " ", "c", " ", "sf", " ", "\[Delta]"}], "-", 
         RowBox[{
          SuperscriptBox["c", "2"], " ", "sf", " ", "\[Delta]"}], "+", 
         RowBox[{"12", " ", "h", " ", "sf", " ", "\[Delta]"}], "-", 
         RowBox[{"8", " ", "c", " ", "h", " ", "sf", " ", "\[Delta]"}], "-", 
         RowBox[{"4", " ", 
          SuperscriptBox["c", "2"], " ", "h", " ", "sf", " ", "\[Delta]"}], 
         "+", 
         RowBox[{"8", " ", 
          SuperscriptBox["sf", "2"], " ", "\[Delta]"}], "-", 
         RowBox[{"4", " ", "c", " ", 
          SuperscriptBox["sf", "2"], " ", "\[Delta]"}], "+", 
         RowBox[{"4", " ", 
          SuperscriptBox["c", "2"], " ", 
          SuperscriptBox["sf", "2"], " ", "\[Delta]"}], "-", 
         RowBox[{"13", " ", "h", " ", 
          SuperscriptBox["sf", "2"], " ", "\[Delta]"}], "-", 
         RowBox[{"38", " ", "c", " ", "h", " ", 
          SuperscriptBox["sf", "2"], " ", "\[Delta]"}], "+", 
         RowBox[{"3", " ", 
          SuperscriptBox["c", "2"], " ", "h", " ", 
          SuperscriptBox["sf", "2"], " ", "\[Delta]"}], "+", 
         RowBox[{"48", " ", "c", " ", 
          SuperscriptBox["h", "2"], " ", 
          SuperscriptBox["sf", "2"], " ", "\[Delta]"}], "+", 
         RowBox[{"16", " ", 
          SuperscriptBox["c", "2"], " ", 
          SuperscriptBox["h", "2"], " ", 
          SuperscriptBox["sf", "2"], " ", "\[Delta]"}], "+", 
         RowBox[{"8", " ", "h", " ", 
          SuperscriptBox["sf", "3"], " ", "\[Delta]"}], "+", 
         RowBox[{"32", " ", "c", " ", "h", " ", 
          SuperscriptBox["sf", "3"], " ", "\[Delta]"}], "-", 
         RowBox[{"8", " ", 
          SuperscriptBox["c", "2"], " ", "h", " ", 
          SuperscriptBox["sf", "3"], " ", "\[Delta]"}], "-", 
         RowBox[{"11", " ", 
          SuperscriptBox["h", "2"], " ", 
          SuperscriptBox["sf", "3"], " ", "\[Delta]"}], "-", 
         RowBox[{"42", " ", "c", " ", 
          SuperscriptBox["h", "2"], " ", 
          SuperscriptBox["sf", "3"], " ", "\[Delta]"}], "-", 
         RowBox[{"27", " ", 
          SuperscriptBox["c", "2"], " ", 
          SuperscriptBox["h", "2"], " ", 
          SuperscriptBox["sf", "3"], " ", "\[Delta]"}], "+", 
         RowBox[{"4", " ", 
          SuperscriptBox["h", "3"], " ", 
          SuperscriptBox["sf", "3"], " ", "\[Delta]"}], "-", 
         RowBox[{"8", " ", "c", " ", 
          SuperscriptBox["h", "3"], " ", 
          SuperscriptBox["sf", "3"], " ", "\[Delta]"}], "+", 
         RowBox[{"4", " ", 
          SuperscriptBox["c", "2"], " ", 
          SuperscriptBox["h", "3"], " ", 
          SuperscriptBox["sf", "3"], " ", "\[Delta]"}], "+", 
         RowBox[{"4", " ", "c", " ", 
          SuperscriptBox["h", "2"], " ", 
          SuperscriptBox["sf", "4"], " ", "\[Delta]"}], "+", 
         RowBox[{"20", " ", 
          SuperscriptBox["c", "2"], " ", 
          SuperscriptBox["h", "2"], " ", 
          SuperscriptBox["sf", "4"], " ", "\[Delta]"}], "+", 
         RowBox[{
          SuperscriptBox["h", "3"], " ", 
          SuperscriptBox["sf", "4"], " ", "\[Delta]"}], "+", 
         RowBox[{"6", " ", "c", " ", 
          SuperscriptBox["h", "3"], " ", 
          SuperscriptBox["sf", "4"], " ", "\[Delta]"}], "-", 
         RowBox[{"7", " ", 
          SuperscriptBox["c", "2"], " ", 
          SuperscriptBox["h", "3"], " ", 
          SuperscriptBox["sf", "4"], " ", "\[Delta]"}], "+", 
         SuperscriptBox["\[Delta]", "2"], "-", 
         RowBox[{"2", " ", "c", " ", 
          SuperscriptBox["\[Delta]", "2"]}], "+", 
         RowBox[{
          SuperscriptBox["c", "2"], " ", 
          SuperscriptBox["\[Delta]", "2"]}], "-", 
         RowBox[{"2", " ", "sf", " ", 
          SuperscriptBox["\[Delta]", "2"]}], "+", 
         RowBox[{"10", " ", "c", " ", "sf", " ", 
          SuperscriptBox["\[Delta]", "2"]}], "-", 
         RowBox[{"h", " ", "sf", " ", 
          SuperscriptBox["\[Delta]", "2"]}], "-", 
         RowBox[{"6", " ", "c", " ", "h", " ", "sf", " ", 
          SuperscriptBox["\[Delta]", "2"]}], "-", 
         RowBox[{"9", " ", 
          SuperscriptBox["c", "2"], " ", "h", " ", "sf", " ", 
          SuperscriptBox["\[Delta]", "2"]}], "-", 
         RowBox[{"8", " ", "c", " ", 
          SuperscriptBox["sf", "2"], " ", 
          SuperscriptBox["\[Delta]", "2"]}], "+", 
         RowBox[{"4", " ", "h", " ", 
          SuperscriptBox["sf", "2"], " ", 
          SuperscriptBox["\[Delta]", "2"]}], "-", 
         RowBox[{"2", " ", "c", " ", "h", " ", 
          SuperscriptBox["sf", "2"], " ", 
          SuperscriptBox["\[Delta]", "2"]}], "+", 
         RowBox[{"14", " ", 
          SuperscriptBox["c", "2"], " ", "h", " ", 
          SuperscriptBox["sf", "2"], " ", 
          SuperscriptBox["\[Delta]", "2"]}], "-", 
         RowBox[{
          SuperscriptBox["h", "2"], " ", 
          SuperscriptBox["sf", "2"], " ", 
          SuperscriptBox["\[Delta]", "2"]}], "+", 
         RowBox[{"10", " ", "c", " ", 
          SuperscriptBox["h", "2"], " ", 
          SuperscriptBox["sf", "2"], " ", 
          SuperscriptBox["\[Delta]", "2"]}], "+", 
         RowBox[{"7", " ", 
          SuperscriptBox["c", "2"], " ", 
          SuperscriptBox["h", "2"], " ", 
          SuperscriptBox["sf", "2"], " ", 
          SuperscriptBox["\[Delta]", "2"]}], "+", 
         RowBox[{"8", " ", "c", " ", "h", " ", 
          SuperscriptBox["sf", "3"], " ", 
          SuperscriptBox["\[Delta]", "2"]}], "-", 
         RowBox[{"8", " ", 
          SuperscriptBox["c", "2"], " ", "h", " ", 
          SuperscriptBox["sf", "3"], " ", 
          SuperscriptBox["\[Delta]", "2"]}], "-", 
         RowBox[{"2", " ", 
          SuperscriptBox["h", "2"], " ", 
          SuperscriptBox["sf", "3"], " ", 
          SuperscriptBox["\[Delta]", "2"]}], "-", 
         RowBox[{"10", " ", "c", " ", 
          SuperscriptBox["h", "2"], " ", 
          SuperscriptBox["sf", "3"], " ", 
          SuperscriptBox["\[Delta]", "2"]}], "-", 
         RowBox[{"12", " ", 
          SuperscriptBox["c", "2"], " ", 
          SuperscriptBox["h", "2"], " ", 
          SuperscriptBox["sf", "3"], " ", 
          SuperscriptBox["\[Delta]", "2"]}], "+", 
         RowBox[{
          SuperscriptBox["h", "3"], " ", 
          SuperscriptBox["sf", "3"], " ", 
          SuperscriptBox["\[Delta]", "2"]}], "-", 
         RowBox[{"2", " ", "c", " ", 
          SuperscriptBox["h", "3"], " ", 
          SuperscriptBox["sf", "3"], " ", 
          SuperscriptBox["\[Delta]", "2"]}], "+", 
         RowBox[{
          SuperscriptBox["c", "2"], " ", 
          SuperscriptBox["h", "3"], " ", 
          SuperscriptBox["sf", "3"], " ", 
          SuperscriptBox["\[Delta]", "2"]}], "+", 
         RowBox[{"8", " ", 
          SuperscriptBox["c", "2"], " ", 
          SuperscriptBox["h", "2"], " ", 
          SuperscriptBox["sf", "4"], " ", 
          SuperscriptBox["\[Delta]", "2"]}], "+", 
         RowBox[{"2", " ", "c", " ", 
          SuperscriptBox["h", "3"], " ", 
          SuperscriptBox["sf", "4"], " ", 
          SuperscriptBox["\[Delta]", "2"]}], "-", 
         RowBox[{"2", " ", 
          SuperscriptBox["c", "2"], " ", 
          SuperscriptBox["h", "3"], " ", 
          SuperscriptBox["sf", "4"], " ", 
          SuperscriptBox["\[Delta]", "2"]}], "-", 
         RowBox[{"3", " ", "sf", " ", 
          SqrtBox[
           RowBox[{
            RowBox[{
             RowBox[{"-", "8"}], " ", 
             RowBox[{"(", 
              RowBox[{
               RowBox[{"-", "1"}], "+", "c"}], ")"}], " ", "sf", " ", 
             RowBox[{"(", 
              RowBox[{
               RowBox[{"-", "1"}], "-", "\[Delta]", "+", 
               RowBox[{"h", " ", 
                RowBox[{"(", 
                 RowBox[{"2", "+", "\[Delta]"}], ")"}]}]}], ")"}], " ", 
             RowBox[{"(", 
              RowBox[{"1", "+", 
               RowBox[{"h", " ", "sf"}], "+", 
               RowBox[{"c", " ", 
                RowBox[{"(", 
                 RowBox[{
                  RowBox[{"-", "1"}], "-", 
                  RowBox[{"2", " ", "\[Delta]"}], "+", 
                  RowBox[{"h", " ", "sf", " ", 
                   RowBox[{"(", 
                    RowBox[{"3", "+", 
                    RowBox[{"2", " ", "\[Delta]"}]}], ")"}]}]}], ")"}]}]}], 
              ")"}]}], "+", 
            SuperscriptBox[
             RowBox[{"(", 
              RowBox[{
               RowBox[{
                RowBox[{"(", 
                 RowBox[{
                  RowBox[{"-", "1"}], "+", 
                  RowBox[{"h", " ", "sf"}]}], ")"}], " ", "\[Delta]"}], "+", 
               RowBox[{"c", " ", 
                RowBox[{"(", 
                 RowBox[{"\[Delta]", "+", 
                  RowBox[{"sf", " ", 
                   RowBox[{"(", 
                    RowBox[{"4", "-", 
                    RowBox[{"8", " ", "h"}], "+", 
                    RowBox[{"4", " ", "\[Delta]"}], "-", 
                    RowBox[{"5", " ", "h", " ", "\[Delta]"}]}], ")"}]}]}], 
                 ")"}]}]}], ")"}], "2"]}]]}], "-", 
         RowBox[{"c", " ", "sf", " ", 
          SqrtBox[
           RowBox[{
            RowBox[{
             RowBox[{"-", "8"}], " ", 
             RowBox[{"(", 
              RowBox[{
               RowBox[{"-", "1"}], "+", "c"}], ")"}], " ", "sf", " ", 
             RowBox[{"(", 
              RowBox[{
               RowBox[{"-", "1"}], "-", "\[Delta]", "+", 
               RowBox[{"h", " ", 
                RowBox[{"(", 
                 RowBox[{"2", "+", "\[Delta]"}], ")"}]}]}], ")"}], " ", 
             RowBox[{"(", 
              RowBox[{"1", "+", 
               RowBox[{"h", " ", "sf"}], "+", 
               RowBox[{"c", " ", 
                RowBox[{"(", 
                 RowBox[{
                  RowBox[{"-", "1"}], "-", 
                  RowBox[{"2", " ", "\[Delta]"}], "+", 
                  RowBox[{"h", " ", "sf", " ", 
                   RowBox[{"(", 
                    RowBox[{"3", "+", 
                    RowBox[{"2", " ", "\[Delta]"}]}], ")"}]}]}], ")"}]}]}], 
              ")"}]}], "+", 
            SuperscriptBox[
             RowBox[{"(", 
              RowBox[{
               RowBox[{
                RowBox[{"(", 
                 RowBox[{
                  RowBox[{"-", "1"}], "+", 
                  RowBox[{"h", " ", "sf"}]}], ")"}], " ", "\[Delta]"}], "+", 
               RowBox[{"c", " ", 
                RowBox[{"(", 
                 RowBox[{"\[Delta]", "+", 
                  RowBox[{"sf", " ", 
                   RowBox[{"(", 
                    RowBox[{"4", "-", 
                    RowBox[{"8", " ", "h"}], "+", 
                    RowBox[{"4", " ", "\[Delta]"}], "-", 
                    RowBox[{"5", " ", "h", " ", "\[Delta]"}]}], ")"}]}]}], 
                 ")"}]}]}], ")"}], "2"]}]]}], "+", 
         RowBox[{"4", " ", "h", " ", "sf", " ", 
          SqrtBox[
           RowBox[{
            RowBox[{
             RowBox[{"-", "8"}], " ", 
             RowBox[{"(", 
              RowBox[{
               RowBox[{"-", "1"}], "+", "c"}], ")"}], " ", "sf", " ", 
             RowBox[{"(", 
              RowBox[{
               RowBox[{"-", "1"}], "-", "\[Delta]", "+", 
               RowBox[{"h", " ", 
                RowBox[{"(", 
                 RowBox[{"2", "+", "\[Delta]"}], ")"}]}]}], ")"}], " ", 
             RowBox[{"(", 
              RowBox[{"1", "+", 
               RowBox[{"h", " ", "sf"}], "+", 
               RowBox[{"c", " ", 
                RowBox[{"(", 
                 RowBox[{
                  RowBox[{"-", "1"}], "-", 
                  RowBox[{"2", " ", "\[Delta]"}], "+", 
                  RowBox[{"h", " ", "sf", " ", 
                   RowBox[{"(", 
                    RowBox[{"3", "+", 
                    RowBox[{"2", " ", "\[Delta]"}]}], ")"}]}]}], ")"}]}]}], 
              ")"}]}], "+", 
            SuperscriptBox[
             RowBox[{"(", 
              RowBox[{
               RowBox[{
                RowBox[{"(", 
                 RowBox[{
                  RowBox[{"-", "1"}], "+", 
                  RowBox[{"h", " ", "sf"}]}], ")"}], " ", "\[Delta]"}], "+", 
               RowBox[{"c", " ", 
                RowBox[{"(", 
                 RowBox[{"\[Delta]", "+", 
                  RowBox[{"sf", " ", 
                   RowBox[{"(", 
                    RowBox[{"4", "-", 
                    RowBox[{"8", " ", "h"}], "+", 
                    RowBox[{"4", " ", "\[Delta]"}], "-", 
                    RowBox[{"5", " ", "h", " ", "\[Delta]"}]}], ")"}]}]}], 
                 ")"}]}]}], ")"}], "2"]}]]}], "+", 
         RowBox[{"4", " ", "c", " ", "h", " ", "sf", " ", 
          SqrtBox[
           RowBox[{
            RowBox[{
             RowBox[{"-", "8"}], " ", 
             RowBox[{"(", 
              RowBox[{
               RowBox[{"-", "1"}], "+", "c"}], ")"}], " ", "sf", " ", 
             RowBox[{"(", 
              RowBox[{
               RowBox[{"-", "1"}], "-", "\[Delta]", "+", 
               RowBox[{"h", " ", 
                RowBox[{"(", 
                 RowBox[{"2", "+", "\[Delta]"}], ")"}]}]}], ")"}], " ", 
             RowBox[{"(", 
              RowBox[{"1", "+", 
               RowBox[{"h", " ", "sf"}], "+", 
               RowBox[{"c", " ", 
                RowBox[{"(", 
                 RowBox[{
                  RowBox[{"-", "1"}], "-", 
                  RowBox[{"2", " ", "\[Delta]"}], "+", 
                  RowBox[{"h", " ", "sf", " ", 
                   RowBox[{"(", 
                    RowBox[{"3", "+", 
                    RowBox[{"2", " ", "\[Delta]"}]}], ")"}]}]}], ")"}]}]}], 
              ")"}]}], "+", 
            SuperscriptBox[
             RowBox[{"(", 
              RowBox[{
               RowBox[{
                RowBox[{"(", 
                 RowBox[{
                  RowBox[{"-", "1"}], "+", 
                  RowBox[{"h", " ", "sf"}]}], ")"}], " ", "\[Delta]"}], "+", 
               RowBox[{"c", " ", 
                RowBox[{"(", 
                 RowBox[{"\[Delta]", "+", 
                  RowBox[{"sf", " ", 
                   RowBox[{"(", 
                    RowBox[{"4", "-", 
                    RowBox[{"8", " ", "h"}], "+", 
                    RowBox[{"4", " ", "\[Delta]"}], "-", 
                    RowBox[{"5", " ", "h", " ", "\[Delta]"}]}], ")"}]}]}], 
                 ")"}]}]}], ")"}], "2"]}]]}], "+", 
         RowBox[{"2", " ", "h", " ", 
          SuperscriptBox["sf", "2"], " ", 
          SqrtBox[
           RowBox[{
            RowBox[{
             RowBox[{"-", "8"}], " ", 
             RowBox[{"(", 
              RowBox[{
               RowBox[{"-", "1"}], "+", "c"}], ")"}], " ", "sf", " ", 
             RowBox[{"(", 
              RowBox[{
               RowBox[{"-", "1"}], "-", "\[Delta]", "+", 
               RowBox[{"h", " ", 
                RowBox[{"(", 
                 RowBox[{"2", "+", "\[Delta]"}], ")"}]}]}], ")"}], " ", 
             RowBox[{"(", 
              RowBox[{"1", "+", 
               RowBox[{"h", " ", "sf"}], "+", 
               RowBox[{"c", " ", 
                RowBox[{"(", 
                 RowBox[{
                  RowBox[{"-", "1"}], "-", 
                  RowBox[{"2", " ", "\[Delta]"}], "+", 
                  RowBox[{"h", " ", "sf", " ", 
                   RowBox[{"(", 
                    RowBox[{"3", "+", 
                    RowBox[{"2", " ", "\[Delta]"}]}], ")"}]}]}], ")"}]}]}], 
              ")"}]}], "+", 
            SuperscriptBox[
             RowBox[{"(", 
              RowBox[{
               RowBox[{
                RowBox[{"(", 
                 RowBox[{
                  RowBox[{"-", "1"}], "+", 
                  RowBox[{"h", " ", "sf"}]}], ")"}], " ", "\[Delta]"}], "+", 
               RowBox[{"c", " ", 
                RowBox[{"(", 
                 RowBox[{"\[Delta]", "+", 
                  RowBox[{"sf", " ", 
                   RowBox[{"(", 
                    RowBox[{"4", "-", 
                    RowBox[{"8", " ", "h"}], "+", 
                    RowBox[{"4", " ", "\[Delta]"}], "-", 
                    RowBox[{"5", " ", "h", " ", "\[Delta]"}]}], ")"}]}]}], 
                 ")"}]}]}], ")"}], "2"]}]]}], "-", 
         RowBox[{"2", " ", "c", " ", "h", " ", 
          SuperscriptBox["sf", "2"], " ", 
          SqrtBox[
           RowBox[{
            RowBox[{
             RowBox[{"-", "8"}], " ", 
             RowBox[{"(", 
              RowBox[{
               RowBox[{"-", "1"}], "+", "c"}], ")"}], " ", "sf", " ", 
             RowBox[{"(", 
              RowBox[{
               RowBox[{"-", "1"}], "-", "\[Delta]", "+", 
               RowBox[{"h", " ", 
                RowBox[{"(", 
                 RowBox[{"2", "+", "\[Delta]"}], ")"}]}]}], ")"}], " ", 
             RowBox[{"(", 
              RowBox[{"1", "+", 
               RowBox[{"h", " ", "sf"}], "+", 
               RowBox[{"c", " ", 
                RowBox[{"(", 
                 RowBox[{
                  RowBox[{"-", "1"}], "-", 
                  RowBox[{"2", " ", "\[Delta]"}], "+", 
                  RowBox[{"h", " ", "sf", " ", 
                   RowBox[{"(", 
                    RowBox[{"3", "+", 
                    RowBox[{"2", " ", "\[Delta]"}]}], ")"}]}]}], ")"}]}]}], 
              ")"}]}], "+", 
            SuperscriptBox[
             RowBox[{"(", 
              RowBox[{
               RowBox[{
                RowBox[{"(", 
                 RowBox[{
                  RowBox[{"-", "1"}], "+", 
                  RowBox[{"h", " ", "sf"}]}], ")"}], " ", "\[Delta]"}], "+", 
               RowBox[{"c", " ", 
                RowBox[{"(", 
                 RowBox[{"\[Delta]", "+", 
                  RowBox[{"sf", " ", 
                   RowBox[{"(", 
                    RowBox[{"4", "-", 
                    RowBox[{"8", " ", "h"}], "+", 
                    RowBox[{"4", " ", "\[Delta]"}], "-", 
                    RowBox[{"5", " ", "h", " ", "\[Delta]"}]}], ")"}]}]}], 
                 ")"}]}]}], ")"}], "2"]}]]}], "-", 
         RowBox[{"4", " ", 
          SuperscriptBox["h", "2"], " ", 
          SuperscriptBox["sf", "2"], " ", 
          SqrtBox[
           RowBox[{
            RowBox[{
             RowBox[{"-", "8"}], " ", 
             RowBox[{"(", 
              RowBox[{
               RowBox[{"-", "1"}], "+", "c"}], ")"}], " ", "sf", " ", 
             RowBox[{"(", 
              RowBox[{
               RowBox[{"-", "1"}], "-", "\[Delta]", "+", 
               RowBox[{"h", " ", 
                RowBox[{"(", 
                 RowBox[{"2", "+", "\[Delta]"}], ")"}]}]}], ")"}], " ", 
             RowBox[{"(", 
              RowBox[{"1", "+", 
               RowBox[{"h", " ", "sf"}], "+", 
               RowBox[{"c", " ", 
                RowBox[{"(", 
                 RowBox[{
                  RowBox[{"-", "1"}], "-", 
                  RowBox[{"2", " ", "\[Delta]"}], "+", 
                  RowBox[{"h", " ", "sf", " ", 
                   RowBox[{"(", 
                    RowBox[{"3", "+", 
                    RowBox[{"2", " ", "\[Delta]"}]}], ")"}]}]}], ")"}]}]}], 
              ")"}]}], "+", 
            SuperscriptBox[
             RowBox[{"(", 
              RowBox[{
               RowBox[{
                RowBox[{"(", 
                 RowBox[{
                  RowBox[{"-", "1"}], "+", 
                  RowBox[{"h", " ", "sf"}]}], ")"}], " ", "\[Delta]"}], "+", 
               RowBox[{"c", " ", 
                RowBox[{"(", 
                 RowBox[{"\[Delta]", "+", 
                  RowBox[{"sf", " ", 
                   RowBox[{"(", 
                    RowBox[{"4", "-", 
                    RowBox[{"8", " ", "h"}], "+", 
                    RowBox[{"4", " ", "\[Delta]"}], "-", 
                    RowBox[{"5", " ", "h", " ", "\[Delta]"}]}], ")"}]}]}], 
                 ")"}]}]}], ")"}], "2"]}]]}], "-", 
         RowBox[{"4", " ", "c", " ", 
          SuperscriptBox["h", "2"], " ", 
          SuperscriptBox["sf", "2"], " ", 
          SqrtBox[
           RowBox[{
            RowBox[{
             RowBox[{"-", "8"}], " ", 
             RowBox[{"(", 
              RowBox[{
               RowBox[{"-", "1"}], "+", "c"}], ")"}], " ", "sf", " ", 
             RowBox[{"(", 
              RowBox[{
               RowBox[{"-", "1"}], "-", "\[Delta]", "+", 
               RowBox[{"h", " ", 
                RowBox[{"(", 
                 RowBox[{"2", "+", "\[Delta]"}], ")"}]}]}], ")"}], " ", 
             RowBox[{"(", 
              RowBox[{"1", "+", 
               RowBox[{"h", " ", "sf"}], "+", 
               RowBox[{"c", " ", 
                RowBox[{"(", 
                 RowBox[{
                  RowBox[{"-", "1"}], "-", 
                  RowBox[{"2", " ", "\[Delta]"}], "+", 
                  RowBox[{"h", " ", "sf", " ", 
                   RowBox[{"(", 
                    RowBox[{"3", "+", 
                    RowBox[{"2", " ", "\[Delta]"}]}], ")"}]}]}], ")"}]}]}], 
              ")"}]}], "+", 
            SuperscriptBox[
             RowBox[{"(", 
              RowBox[{
               RowBox[{
                RowBox[{"(", 
                 RowBox[{
                  RowBox[{"-", "1"}], "+", 
                  RowBox[{"h", " ", "sf"}]}], ")"}], " ", "\[Delta]"}], "+", 
               RowBox[{"c", " ", 
                RowBox[{"(", 
                 RowBox[{"\[Delta]", "+", 
                  RowBox[{"sf", " ", 
                   RowBox[{"(", 
                    RowBox[{"4", "-", 
                    RowBox[{"8", " ", "h"}], "+", 
                    RowBox[{"4", " ", "\[Delta]"}], "-", 
                    RowBox[{"5", " ", "h", " ", "\[Delta]"}]}], ")"}]}]}], 
                 ")"}]}]}], ")"}], "2"]}]]}], "+", 
         RowBox[{
          SuperscriptBox["h", "2"], " ", 
          SuperscriptBox["sf", "3"], " ", 
          SqrtBox[
           RowBox[{
            RowBox[{
             RowBox[{"-", "8"}], " ", 
             RowBox[{"(", 
              RowBox[{
               RowBox[{"-", "1"}], "+", "c"}], ")"}], " ", "sf", " ", 
             RowBox[{"(", 
              RowBox[{
               RowBox[{"-", "1"}], "-", "\[Delta]", "+", 
               RowBox[{"h", " ", 
                RowBox[{"(", 
                 RowBox[{"2", "+", "\[Delta]"}], ")"}]}]}], ")"}], " ", 
             RowBox[{"(", 
              RowBox[{"1", "+", 
               RowBox[{"h", " ", "sf"}], "+", 
               RowBox[{"c", " ", 
                RowBox[{"(", 
                 RowBox[{
                  RowBox[{"-", "1"}], "-", 
                  RowBox[{"2", " ", "\[Delta]"}], "+", 
                  RowBox[{"h", " ", "sf", " ", 
                   RowBox[{"(", 
                    RowBox[{"3", "+", 
                    RowBox[{"2", " ", "\[Delta]"}]}], ")"}]}]}], ")"}]}]}], 
              ")"}]}], "+", 
            SuperscriptBox[
             RowBox[{"(", 
              RowBox[{
               RowBox[{
                RowBox[{"(", 
                 RowBox[{
                  RowBox[{"-", "1"}], "+", 
                  RowBox[{"h", " ", "sf"}]}], ")"}], " ", "\[Delta]"}], "+", 
               RowBox[{"c", " ", 
                RowBox[{"(", 
                 RowBox[{"\[Delta]", "+", 
                  RowBox[{"sf", " ", 
                   RowBox[{"(", 
                    RowBox[{"4", "-", 
                    RowBox[{"8", " ", "h"}], "+", 
                    RowBox[{"4", " ", "\[Delta]"}], "-", 
                    RowBox[{"5", " ", "h", " ", "\[Delta]"}]}], ")"}]}]}], 
                 ")"}]}]}], ")"}], "2"]}]]}], "+", 
         RowBox[{"3", " ", "c", " ", 
          SuperscriptBox["h", "2"], " ", 
          SuperscriptBox["sf", "3"], " ", 
          SqrtBox[
           RowBox[{
            RowBox[{
             RowBox[{"-", "8"}], " ", 
             RowBox[{"(", 
              RowBox[{
               RowBox[{"-", "1"}], "+", "c"}], ")"}], " ", "sf", " ", 
             RowBox[{"(", 
              RowBox[{
               RowBox[{"-", "1"}], "-", "\[Delta]", "+", 
               RowBox[{"h", " ", 
                RowBox[{"(", 
                 RowBox[{"2", "+", "\[Delta]"}], ")"}]}]}], ")"}], " ", 
             RowBox[{"(", 
              RowBox[{"1", "+", 
               RowBox[{"h", " ", "sf"}], "+", 
               RowBox[{"c", " ", 
                RowBox[{"(", 
                 RowBox[{
                  RowBox[{"-", "1"}], "-", 
                  RowBox[{"2", " ", "\[Delta]"}], "+", 
                  RowBox[{"h", " ", "sf", " ", 
                   RowBox[{"(", 
                    RowBox[{"3", "+", 
                    RowBox[{"2", " ", "\[Delta]"}]}], ")"}]}]}], ")"}]}]}], 
              ")"}]}], "+", 
            SuperscriptBox[
             RowBox[{"(", 
              RowBox[{
               RowBox[{
                RowBox[{"(", 
                 RowBox[{
                  RowBox[{"-", "1"}], "+", 
                  RowBox[{"h", " ", "sf"}]}], ")"}], " ", "\[Delta]"}], "+", 
               RowBox[{"c", " ", 
                RowBox[{"(", 
                 RowBox[{"\[Delta]", "+", 
                  RowBox[{"sf", " ", 
                   RowBox[{"(", 
                    RowBox[{"4", "-", 
                    RowBox[{"8", " ", "h"}], "+", 
                    RowBox[{"4", " ", "\[Delta]"}], "-", 
                    RowBox[{"5", " ", "h", " ", "\[Delta]"}]}], ")"}]}]}], 
                 ")"}]}]}], ")"}], "2"]}]]}], "+", 
         RowBox[{"\[Delta]", " ", 
          SqrtBox[
           RowBox[{
            RowBox[{
             RowBox[{"-", "8"}], " ", 
             RowBox[{"(", 
              RowBox[{
               RowBox[{"-", "1"}], "+", "c"}], ")"}], " ", "sf", " ", 
             RowBox[{"(", 
              RowBox[{
               RowBox[{"-", "1"}], "-", "\[Delta]", "+", 
               RowBox[{"h", " ", 
                RowBox[{"(", 
                 RowBox[{"2", "+", "\[Delta]"}], ")"}]}]}], ")"}], " ", 
             RowBox[{"(", 
              RowBox[{"1", "+", 
               RowBox[{"h", " ", "sf"}], "+", 
               RowBox[{"c", " ", 
                RowBox[{"(", 
                 RowBox[{
                  RowBox[{"-", "1"}], "-", 
                  RowBox[{"2", " ", "\[Delta]"}], "+", 
                  RowBox[{"h", " ", "sf", " ", 
                   RowBox[{"(", 
                    RowBox[{"3", "+", 
                    RowBox[{"2", " ", "\[Delta]"}]}], ")"}]}]}], ")"}]}]}], 
              ")"}]}], "+", 
            SuperscriptBox[
             RowBox[{"(", 
              RowBox[{
               RowBox[{
                RowBox[{"(", 
                 RowBox[{
                  RowBox[{"-", "1"}], "+", 
                  RowBox[{"h", " ", "sf"}]}], ")"}], " ", "\[Delta]"}], "+", 
               RowBox[{"c", " ", 
                RowBox[{"(", 
                 RowBox[{"\[Delta]", "+", 
                  RowBox[{"sf", " ", 
                   RowBox[{"(", 
                    RowBox[{"4", "-", 
                    RowBox[{"8", " ", "h"}], "+", 
                    RowBox[{"4", " ", "\[Delta]"}], "-", 
                    RowBox[{"5", " ", "h", " ", "\[Delta]"}]}], ")"}]}]}], 
                 ")"}]}]}], ")"}], "2"]}]]}], "-", 
         RowBox[{"c", " ", "\[Delta]", " ", 
          SqrtBox[
           RowBox[{
            RowBox[{
             RowBox[{"-", "8"}], " ", 
             RowBox[{"(", 
              RowBox[{
               RowBox[{"-", "1"}], "+", "c"}], ")"}], " ", "sf", " ", 
             RowBox[{"(", 
              RowBox[{
               RowBox[{"-", "1"}], "-", "\[Delta]", "+", 
               RowBox[{"h", " ", 
                RowBox[{"(", 
                 RowBox[{"2", "+", "\[Delta]"}], ")"}]}]}], ")"}], " ", 
             RowBox[{"(", 
              RowBox[{"1", "+", 
               RowBox[{"h", " ", "sf"}], "+", 
               RowBox[{"c", " ", 
                RowBox[{"(", 
                 RowBox[{
                  RowBox[{"-", "1"}], "-", 
                  RowBox[{"2", " ", "\[Delta]"}], "+", 
                  RowBox[{"h", " ", "sf", " ", 
                   RowBox[{"(", 
                    RowBox[{"3", "+", 
                    RowBox[{"2", " ", "\[Delta]"}]}], ")"}]}]}], ")"}]}]}], 
              ")"}]}], "+", 
            SuperscriptBox[
             RowBox[{"(", 
              RowBox[{
               RowBox[{
                RowBox[{"(", 
                 RowBox[{
                  RowBox[{"-", "1"}], "+", 
                  RowBox[{"h", " ", "sf"}]}], ")"}], " ", "\[Delta]"}], "+", 
               RowBox[{"c", " ", 
                RowBox[{"(", 
                 RowBox[{"\[Delta]", "+", 
                  RowBox[{"sf", " ", 
                   RowBox[{"(", 
                    RowBox[{"4", "-", 
                    RowBox[{"8", " ", "h"}], "+", 
                    RowBox[{"4", " ", "\[Delta]"}], "-", 
                    RowBox[{"5", " ", "h", " ", "\[Delta]"}]}], ")"}]}]}], 
                 ")"}]}]}], ")"}], "2"]}]]}], "-", 
         RowBox[{"2", " ", "sf", " ", "\[Delta]", " ", 
          SqrtBox[
           RowBox[{
            RowBox[{
             RowBox[{"-", "8"}], " ", 
             RowBox[{"(", 
              RowBox[{
               RowBox[{"-", "1"}], "+", "c"}], ")"}], " ", "sf", " ", 
             RowBox[{"(", 
              RowBox[{
               RowBox[{"-", "1"}], "-", "\[Delta]", "+", 
               RowBox[{"h", " ", 
                RowBox[{"(", 
                 RowBox[{"2", "+", "\[Delta]"}], ")"}]}]}], ")"}], " ", 
             RowBox[{"(", 
              RowBox[{"1", "+", 
               RowBox[{"h", " ", "sf"}], "+", 
               RowBox[{"c", " ", 
                RowBox[{"(", 
                 RowBox[{
                  RowBox[{"-", "1"}], "-", 
                  RowBox[{"2", " ", "\[Delta]"}], "+", 
                  RowBox[{"h", " ", "sf", " ", 
                   RowBox[{"(", 
                    RowBox[{"3", "+", 
                    RowBox[{"2", " ", "\[Delta]"}]}], ")"}]}]}], ")"}]}]}], 
              ")"}]}], "+", 
            SuperscriptBox[
             RowBox[{"(", 
              RowBox[{
               RowBox[{
                RowBox[{"(", 
                 RowBox[{
                  RowBox[{"-", "1"}], "+", 
                  RowBox[{"h", " ", "sf"}]}], ")"}], " ", "\[Delta]"}], "+", 
               RowBox[{"c", " ", 
                RowBox[{"(", 
                 RowBox[{"\[Delta]", "+", 
                  RowBox[{"sf", " ", 
                   RowBox[{"(", 
                    RowBox[{"4", "-", 
                    RowBox[{"8", " ", "h"}], "+", 
                    RowBox[{"4", " ", "\[Delta]"}], "-", 
                    RowBox[{"5", " ", "h", " ", "\[Delta]"}]}], ")"}]}]}], 
                 ")"}]}]}], ")"}], "2"]}]]}], "+", 
         RowBox[{"4", " ", "c", " ", "h", " ", "sf", " ", "\[Delta]", " ", 
          SqrtBox[
           RowBox[{
            RowBox[{
             RowBox[{"-", "8"}], " ", 
             RowBox[{"(", 
              RowBox[{
               RowBox[{"-", "1"}], "+", "c"}], ")"}], " ", "sf", " ", 
             RowBox[{"(", 
              RowBox[{
               RowBox[{"-", "1"}], "-", "\[Delta]", "+", 
               RowBox[{"h", " ", 
                RowBox[{"(", 
                 RowBox[{"2", "+", "\[Delta]"}], ")"}]}]}], ")"}], " ", 
             RowBox[{"(", 
              RowBox[{"1", "+", 
               RowBox[{"h", " ", "sf"}], "+", 
               RowBox[{"c", " ", 
                RowBox[{"(", 
                 RowBox[{
                  RowBox[{"-", "1"}], "-", 
                  RowBox[{"2", " ", "\[Delta]"}], "+", 
                  RowBox[{"h", " ", "sf", " ", 
                   RowBox[{"(", 
                    RowBox[{"3", "+", 
                    RowBox[{"2", " ", "\[Delta]"}]}], ")"}]}]}], ")"}]}]}], 
              ")"}]}], "+", 
            SuperscriptBox[
             RowBox[{"(", 
              RowBox[{
               RowBox[{
                RowBox[{"(", 
                 RowBox[{
                  RowBox[{"-", "1"}], "+", 
                  RowBox[{"h", " ", "sf"}]}], ")"}], " ", "\[Delta]"}], "+", 
               RowBox[{"c", " ", 
                RowBox[{"(", 
                 RowBox[{"\[Delta]", "+", 
                  RowBox[{"sf", " ", 
                   RowBox[{"(", 
                    RowBox[{"4", "-", 
                    RowBox[{"8", " ", "h"}], "+", 
                    RowBox[{"4", " ", "\[Delta]"}], "-", 
                    RowBox[{"5", " ", "h", " ", "\[Delta]"}]}], ")"}]}]}], 
                 ")"}]}]}], ")"}], "2"]}]]}], "+", 
         RowBox[{"2", " ", "h", " ", 
          SuperscriptBox["sf", "2"], " ", "\[Delta]", " ", 
          SqrtBox[
           RowBox[{
            RowBox[{
             RowBox[{"-", "8"}], " ", 
             RowBox[{"(", 
              RowBox[{
               RowBox[{"-", "1"}], "+", "c"}], ")"}], " ", "sf", " ", 
             RowBox[{"(", 
              RowBox[{
               RowBox[{"-", "1"}], "-", "\[Delta]", "+", 
               RowBox[{"h", " ", 
                RowBox[{"(", 
                 RowBox[{"2", "+", "\[Delta]"}], ")"}]}]}], ")"}], " ", 
             RowBox[{"(", 
              RowBox[{"1", "+", 
               RowBox[{"h", " ", "sf"}], "+", 
               RowBox[{"c", " ", 
                RowBox[{"(", 
                 RowBox[{
                  RowBox[{"-", "1"}], "-", 
                  RowBox[{"2", " ", "\[Delta]"}], "+", 
                  RowBox[{"h", " ", "sf", " ", 
                   RowBox[{"(", 
                    RowBox[{"3", "+", 
                    RowBox[{"2", " ", "\[Delta]"}]}], ")"}]}]}], ")"}]}]}], 
              ")"}]}], "+", 
            SuperscriptBox[
             RowBox[{"(", 
              RowBox[{
               RowBox[{
                RowBox[{"(", 
                 RowBox[{
                  RowBox[{"-", "1"}], "+", 
                  RowBox[{"h", " ", "sf"}]}], ")"}], " ", "\[Delta]"}], "+", 
               RowBox[{"c", " ", 
                RowBox[{"(", 
                 RowBox[{"\[Delta]", "+", 
                  RowBox[{"sf", " ", 
                   RowBox[{"(", 
                    RowBox[{"4", "-", 
                    RowBox[{"8", " ", "h"}], "+", 
                    RowBox[{"4", " ", "\[Delta]"}], "-", 
                    RowBox[{"5", " ", "h", " ", "\[Delta]"}]}], ")"}]}]}], 
                 ")"}]}]}], ")"}], "2"]}]]}], "-", 
         RowBox[{"2", " ", "c", " ", "h", " ", 
          SuperscriptBox["sf", "2"], " ", "\[Delta]", " ", 
          SqrtBox[
           RowBox[{
            RowBox[{
             RowBox[{"-", "8"}], " ", 
             RowBox[{"(", 
              RowBox[{
               RowBox[{"-", "1"}], "+", "c"}], ")"}], " ", "sf", " ", 
             RowBox[{"(", 
              RowBox[{
               RowBox[{"-", "1"}], "-", "\[Delta]", "+", 
               RowBox[{"h", " ", 
                RowBox[{"(", 
                 RowBox[{"2", "+", "\[Delta]"}], ")"}]}]}], ")"}], " ", 
             RowBox[{"(", 
              RowBox[{"1", "+", 
               RowBox[{"h", " ", "sf"}], "+", 
               RowBox[{"c", " ", 
                RowBox[{"(", 
                 RowBox[{
                  RowBox[{"-", "1"}], "-", 
                  RowBox[{"2", " ", "\[Delta]"}], "+", 
                  RowBox[{"h", " ", "sf", " ", 
                   RowBox[{"(", 
                    RowBox[{"3", "+", 
                    RowBox[{"2", " ", "\[Delta]"}]}], ")"}]}]}], ")"}]}]}], 
              ")"}]}], "+", 
            SuperscriptBox[
             RowBox[{"(", 
              RowBox[{
               RowBox[{
                RowBox[{"(", 
                 RowBox[{
                  RowBox[{"-", "1"}], "+", 
                  RowBox[{"h", " ", "sf"}]}], ")"}], " ", "\[Delta]"}], "+", 
               RowBox[{"c", " ", 
                RowBox[{"(", 
                 RowBox[{"\[Delta]", "+", 
                  RowBox[{"sf", " ", 
                   RowBox[{"(", 
                    RowBox[{"4", "-", 
                    RowBox[{"8", " ", "h"}], "+", 
                    RowBox[{"4", " ", "\[Delta]"}], "-", 
                    RowBox[{"5", " ", "h", " ", "\[Delta]"}]}], ")"}]}]}], 
                 ")"}]}]}], ")"}], "2"]}]]}], "-", 
         RowBox[{
          SuperscriptBox["h", "2"], " ", 
          SuperscriptBox["sf", "2"], " ", "\[Delta]", " ", 
          SqrtBox[
           RowBox[{
            RowBox[{
             RowBox[{"-", "8"}], " ", 
             RowBox[{"(", 
              RowBox[{
               RowBox[{"-", "1"}], "+", "c"}], ")"}], " ", "sf", " ", 
             RowBox[{"(", 
              RowBox[{
               RowBox[{"-", "1"}], "-", "\[Delta]", "+", 
               RowBox[{"h", " ", 
                RowBox[{"(", 
                 RowBox[{"2", "+", "\[Delta]"}], ")"}]}]}], ")"}], " ", 
             RowBox[{"(", 
              RowBox[{"1", "+", 
               RowBox[{"h", " ", "sf"}], "+", 
               RowBox[{"c", " ", 
                RowBox[{"(", 
                 RowBox[{
                  RowBox[{"-", "1"}], "-", 
                  RowBox[{"2", " ", "\[Delta]"}], "+", 
                  RowBox[{"h", " ", "sf", " ", 
                   RowBox[{"(", 
                    RowBox[{"3", "+", 
                    RowBox[{"2", " ", "\[Delta]"}]}], ")"}]}]}], ")"}]}]}], 
              ")"}]}], "+", 
            SuperscriptBox[
             RowBox[{"(", 
              RowBox[{
               RowBox[{
                RowBox[{"(", 
                 RowBox[{
                  RowBox[{"-", "1"}], "+", 
                  RowBox[{"h", " ", "sf"}]}], ")"}], " ", "\[Delta]"}], "+", 
               RowBox[{"c", " ", 
                RowBox[{"(", 
                 RowBox[{"\[Delta]", "+", 
                  RowBox[{"sf", " ", 
                   RowBox[{"(", 
                    RowBox[{"4", "-", 
                    RowBox[{"8", " ", "h"}], "+", 
                    RowBox[{"4", " ", "\[Delta]"}], "-", 
                    RowBox[{"5", " ", "h", " ", "\[Delta]"}]}], ")"}]}]}], 
                 ")"}]}]}], ")"}], "2"]}]]}], "-", 
         RowBox[{"3", " ", "c", " ", 
          SuperscriptBox["h", "2"], " ", 
          SuperscriptBox["sf", "2"], " ", "\[Delta]", " ", 
          SqrtBox[
           RowBox[{
            RowBox[{
             RowBox[{"-", "8"}], " ", 
             RowBox[{"(", 
              RowBox[{
               RowBox[{"-", "1"}], "+", "c"}], ")"}], " ", "sf", " ", 
             RowBox[{"(", 
              RowBox[{
               RowBox[{"-", "1"}], "-", "\[Delta]", "+", 
               RowBox[{"h", " ", 
                RowBox[{"(", 
                 RowBox[{"2", "+", "\[Delta]"}], ")"}]}]}], ")"}], " ", 
             RowBox[{"(", 
              RowBox[{"1", "+", 
               RowBox[{"h", " ", "sf"}], "+", 
               RowBox[{"c", " ", 
                RowBox[{"(", 
                 RowBox[{
                  RowBox[{"-", "1"}], "-", 
                  RowBox[{"2", " ", "\[Delta]"}], "+", 
                  RowBox[{"h", " ", "sf", " ", 
                   RowBox[{"(", 
                    RowBox[{"3", "+", 
                    RowBox[{"2", " ", "\[Delta]"}]}], ")"}]}]}], ")"}]}]}], 
              ")"}]}], "+", 
            SuperscriptBox[
             RowBox[{"(", 
              RowBox[{
               RowBox[{
                RowBox[{"(", 
                 RowBox[{
                  RowBox[{"-", "1"}], "+", 
                  RowBox[{"h", " ", "sf"}]}], ")"}], " ", "\[Delta]"}], "+", 
               RowBox[{"c", " ", 
                RowBox[{"(", 
                 RowBox[{"\[Delta]", "+", 
                  RowBox[{"sf", " ", 
                   RowBox[{"(", 
                    RowBox[{"4", "-", 
                    RowBox[{"8", " ", "h"}], "+", 
                    RowBox[{"4", " ", "\[Delta]"}], "-", 
                    RowBox[{"5", " ", "h", " ", "\[Delta]"}]}], ")"}]}]}], 
                 ")"}]}]}], ")"}], "2"]}]]}], "+", 
         RowBox[{"2", " ", "c", " ", 
          SuperscriptBox["h", "2"], " ", 
          SuperscriptBox["sf", "3"], " ", "\[Delta]", " ", 
          SqrtBox[
           RowBox[{
            RowBox[{
             RowBox[{"-", "8"}], " ", 
             RowBox[{"(", 
              RowBox[{
               RowBox[{"-", "1"}], "+", "c"}], ")"}], " ", "sf", " ", 
             RowBox[{"(", 
              RowBox[{
               RowBox[{"-", "1"}], "-", "\[Delta]", "+", 
               RowBox[{"h", " ", 
                RowBox[{"(", 
                 RowBox[{"2", "+", "\[Delta]"}], ")"}]}]}], ")"}], " ", 
             RowBox[{"(", 
              RowBox[{"1", "+", 
               RowBox[{"h", " ", "sf"}], "+", 
               RowBox[{"c", " ", 
                RowBox[{"(", 
                 RowBox[{
                  RowBox[{"-", "1"}], "-", 
                  RowBox[{"2", " ", "\[Delta]"}], "+", 
                  RowBox[{"h", " ", "sf", " ", 
                   RowBox[{"(", 
                    RowBox[{"3", "+", 
                    RowBox[{"2", " ", "\[Delta]"}]}], ")"}]}]}], ")"}]}]}], 
              ")"}]}], "+", 
            SuperscriptBox[
             RowBox[{"(", 
              RowBox[{
               RowBox[{
                RowBox[{"(", 
                 RowBox[{
                  RowBox[{"-", "1"}], "+", 
                  RowBox[{"h", " ", "sf"}]}], ")"}], " ", "\[Delta]"}], "+", 
               RowBox[{"c", " ", 
                RowBox[{"(", 
                 RowBox[{"\[Delta]", "+", 
                  RowBox[{"sf", " ", 
                   RowBox[{"(", 
                    RowBox[{"4", "-", 
                    RowBox[{"8", " ", "h"}], "+", 
                    RowBox[{"4", " ", "\[Delta]"}], "-", 
                    RowBox[{"5", " ", "h", " ", "\[Delta]"}]}], ")"}]}]}], 
                 ")"}]}]}], ")"}], "2"]}]]}]}], ")"}], "/", 
       RowBox[{"(", 
        RowBox[{"2", " ", 
         RowBox[{"(", 
          RowBox[{
           RowBox[{
            SuperscriptBox[
             RowBox[{"(", 
              RowBox[{
               RowBox[{"-", "1"}], "+", "c"}], ")"}], "2"], " ", 
            SuperscriptBox["\[Delta]", "2"]}], "+", 
           RowBox[{
            SuperscriptBox["h", "3"], " ", 
            SuperscriptBox["sf", "4"], " ", 
            SuperscriptBox[
             RowBox[{"(", 
              RowBox[{"1", "+", 
               RowBox[{"c", " ", 
                RowBox[{"(", 
                 RowBox[{"3", "+", 
                  RowBox[{"2", " ", "\[Delta]"}]}], ")"}]}]}], ")"}], "2"]}], 
           "+", 
           RowBox[{"sf", " ", 
            RowBox[{"(", 
             RowBox[{"1", "-", 
              RowBox[{"2", " ", "\[Delta]"}], "+", 
              RowBox[{"h", " ", 
               RowBox[{"(", 
                RowBox[{
                 RowBox[{"-", "2"}], "+", 
                 RowBox[{"2", " ", "\[Delta]"}], "-", 
                 RowBox[{"3", " ", 
                  SuperscriptBox["\[Delta]", "2"]}]}], ")"}]}], "+", 
              RowBox[{"2", " ", "c", " ", 
               RowBox[{"(", 
                RowBox[{
                 RowBox[{"-", "1"}], "+", 
                 RowBox[{"2", " ", 
                  SuperscriptBox["\[Delta]", "2"]}], "+", 
                 RowBox[{"h", " ", 
                  RowBox[{"(", 
                   RowBox[{"2", "+", 
                    RowBox[{"2", " ", "\[Delta]"}], "+", 
                    SuperscriptBox["\[Delta]", "2"]}], ")"}]}]}], ")"}]}], 
              "+", 
              RowBox[{
               SuperscriptBox["c", "2"], " ", 
               RowBox[{"(", 
                RowBox[{"1", "+", 
                 RowBox[{"2", " ", "\[Delta]"}], "-", 
                 RowBox[{"h", " ", 
                  RowBox[{"(", 
                   RowBox[{"2", "+", 
                    RowBox[{"6", " ", "\[Delta]"}], "+", 
                    RowBox[{"7", " ", 
                    SuperscriptBox["\[Delta]", "2"]}]}], ")"}]}]}], ")"}]}]}],
              ")"}]}], "-", 
           RowBox[{"h", " ", 
            SuperscriptBox["sf", "3"], " ", 
            RowBox[{"(", 
             RowBox[{
              RowBox[{"2", " ", "c", " ", 
               RowBox[{"(", 
                RowBox[{
                 RowBox[{
                  RowBox[{"-", "4"}], " ", 
                  RowBox[{"(", 
                   RowBox[{"1", "+", "\[Delta]"}], ")"}]}], "+", 
                 RowBox[{"h", " ", 
                  RowBox[{"(", 
                   RowBox[{"3", "-", 
                    RowBox[{"2", " ", 
                    SuperscriptBox["\[Delta]", "2"]}]}], ")"}]}], "+", 
                 RowBox[{
                  SuperscriptBox["h", "2"], " ", 
                  RowBox[{"(", 
                   RowBox[{"6", "+", 
                    RowBox[{"6", " ", "\[Delta]"}], "+", 
                    SuperscriptBox["\[Delta]", "2"]}], ")"}]}]}], ")"}]}], 
              "+", 
              RowBox[{"h", " ", 
               RowBox[{"(", 
                RowBox[{
                 RowBox[{"-", "7"}], "-", 
                 RowBox[{"6", " ", "\[Delta]"}], "+", 
                 RowBox[{"h", " ", 
                  RowBox[{"(", 
                   RowBox[{"10", "+", 
                    RowBox[{"6", " ", "\[Delta]"}], "+", 
                    SuperscriptBox["\[Delta]", "2"]}], ")"}]}]}], ")"}]}], 
              "+", 
              RowBox[{
               SuperscriptBox["c", "2"], " ", 
               RowBox[{"(", 
                RowBox[{
                 RowBox[{
                  RowBox[{"-", "8"}], " ", 
                  RowBox[{"(", 
                   RowBox[{"1", "+", "\[Delta]"}], ")"}]}], "+", 
                 RowBox[{
                  SuperscriptBox["h", "2"], " ", 
                  RowBox[{"(", 
                   RowBox[{"10", "+", 
                    RowBox[{"14", " ", "\[Delta]"}], "+", 
                    RowBox[{"5", " ", 
                    SuperscriptBox["\[Delta]", "2"]}]}], ")"}]}], "+", 
                 RowBox[{"h", " ", 
                  RowBox[{"(", 
                   RowBox[{"17", "+", 
                    RowBox[{"22", " ", "\[Delta]"}], "+", 
                    RowBox[{"8", " ", 
                    SuperscriptBox["\[Delta]", "2"]}]}], ")"}]}]}], ")"}]}]}],
              ")"}]}], "+", 
           RowBox[{
            SuperscriptBox["sf", "2"], " ", 
            RowBox[{"(", 
             RowBox[{
              RowBox[{"8", " ", 
               RowBox[{"(", 
                RowBox[{"1", "+", "\[Delta]"}], ")"}]}], "-", 
              RowBox[{"5", " ", "h", " ", 
               RowBox[{"(", 
                RowBox[{"5", "+", 
                 RowBox[{"4", " ", "\[Delta]"}]}], ")"}]}], "+", 
              RowBox[{
               SuperscriptBox["h", "2"], " ", 
               RowBox[{"(", 
                RowBox[{"20", "+", 
                 RowBox[{"12", " ", "\[Delta]"}], "+", 
                 RowBox[{"3", " ", 
                  SuperscriptBox["\[Delta]", "2"]}]}], ")"}]}], "+", 
              RowBox[{"2", " ", "c", " ", 
               RowBox[{"(", 
                RowBox[{
                 RowBox[{"4", " ", 
                  RowBox[{"(", 
                   RowBox[{"1", "+", "\[Delta]"}], ")"}]}], "+", 
                 RowBox[{
                  SuperscriptBox["h", "2"], " ", 
                  RowBox[{"(", 
                   RowBox[{"12", "+", 
                    RowBox[{"12", " ", "\[Delta]"}], "+", 
                    SuperscriptBox["\[Delta]", "2"]}], ")"}]}], "-", 
                 RowBox[{"h", " ", 
                  RowBox[{"(", 
                   RowBox[{"15", "+", 
                    RowBox[{"18", " ", "\[Delta]"}], "+", 
                    RowBox[{"4", " ", 
                    SuperscriptBox["\[Delta]", "2"]}]}], ")"}]}]}], ")"}]}], 
              "+", 
              RowBox[{
               SuperscriptBox["c", "2"], " ", "h", " ", 
               RowBox[{"(", 
                RowBox[{
                 RowBox[{"-", "9"}], "-", 
                 RowBox[{"8", " ", "\[Delta]"}], "+", 
                 RowBox[{"4", " ", 
                  SuperscriptBox["\[Delta]", "2"]}], "+", 
                 RowBox[{"h", " ", 
                  RowBox[{"(", 
                   RowBox[{"20", "+", 
                    RowBox[{"28", " ", "\[Delta]"}], "+", 
                    RowBox[{"11", " ", 
                    SuperscriptBox["\[Delta]", "2"]}]}], ")"}]}]}], ")"}]}]}],
              ")"}]}]}], ")"}]}], ")"}]}]}]}], "}"}]}], "}"}]], "Output",
 CellChangeTimes->{
  3.781875105173994*^9, 3.781875170162109*^9, {3.781875672576749*^9, 
   3.781875676961214*^9}, 3.781875825351769*^9, 3.78187591736909*^9, 
   3.781876477606761*^9, 3.781877569043977*^9, 3.781877675927492*^9, 
   3.781884304067277*^9, 3.7819497132392693`*^9, 3.781961305696556*^9, 
   3.785575757732452*^9, 3.78559449935441*^9, 3.8000118795670223`*^9, 
   3.800268330874473*^9, 3.814006516690022*^9, 3.8140067104966097`*^9, 
   3.8140075449729233`*^9, 3.814007588987846*^9, 3.814084697471445*^9, 
   3.814196549180375*^9},
 CellLabel->
  "Out[1366]=",ExpressionUUID->"3ba05cf0-15cb-4431-b1a8-b31ed054edce"]
}, Open  ]]
}, Closed]],

Cell[CellGroupData[{

Cell["Stability", "Subsection",
 CellChangeTimes->{{3.781465326981051*^9, 
  3.781465329029233*^9}},ExpressionUUID->"bd99752f-e3a9-4507-a41e-\
16f72d13af99"],

Cell["\<\
There are at least two immediately apparent equilibria: when drive is absent \
from the population and when drive is fixed in the population. 

We first find the Jacobian for the system:\
\>", "Text",
 CellChangeTimes->{3.814018732143256*^9, 
  3.814092016447557*^9},ExpressionUUID->"2e906936-97d9-44e1-a935-\
2be5e7061234"],

Cell[BoxData[{
 RowBox[{
  RowBox[{"recursions2", "=", 
   RowBox[{"{", 
    RowBox[{
    "Dmprime2", ",", " ", "dmprime2", ",", " ", "DDfprime2", ",", " ", 
     "Ddfprime2", ",", " ", "ddfprime2"}], "}"}]}], 
  ";"}], "\[IndentingNewLine]", 
 RowBox[{
  RowBox[{"jacob2", "=", 
   RowBox[{"Simplify", "[", 
    RowBox[{"Transpose", "[", 
     RowBox[{"{", 
      RowBox[{
       RowBox[{"D", "[", 
        RowBox[{"recursions2", ",", "Dm"}], "]"}], ",", " ", 
       RowBox[{"D", "[", 
        RowBox[{"recursions2", ",", "dm"}], "]"}], ",", " ", 
       RowBox[{"D", "[", 
        RowBox[{"recursions2", ",", " ", "DDf"}], "]"}], ",", " ", 
       RowBox[{"D", "[", 
        RowBox[{"recursions2", ",", " ", "Ddf"}], "]"}], ",", " ", 
       RowBox[{"D", "[", 
        RowBox[{"recursions2", ",", " ", "ddf"}], "]"}]}], "}"}], "]"}], 
    "]"}]}], ";"}]}], "Input",
 CellChangeTimes->{{3.781465345288929*^9, 3.781465345456649*^9}, {
  3.781511024565343*^9, 3.7815110446207*^9}, {3.781511354597693*^9, 
  3.7815113827574787`*^9}},
 CellLabel->
  "In[1367]:=",ExpressionUUID->"f15f9af6-18ce-4a44-acc5-bbe08526fe99"],

Cell["\<\
And then we assess the conditions under which the drive-absent and \
drive-fixed equilibria are stable/instable. \
\>", "Text",
 CellChangeTimes->{
  3.814018737227765*^9},ExpressionUUID->"fe4e26a8-f4e0-41b2-b4dc-\
01f75780b3bf"],

Cell[CellGroupData[{

Cell["drive absent (pm=pf=0) equilibrium", "Subsubsection",
 CellChangeTimes->{{3.7814653542305307`*^9, 3.7814653590889*^9}, 
   3.814018723199863*^9},ExpressionUUID->"ddc39688-4044-459a-8251-\
86bb830d06bd"],

Cell["\<\
To evaluate the stability of the equilibrium where drive is absent we \
consider the eigenvalues of the Jacobian at this point (substituting pf->0, \
pm->0, and the equilibrium population size without drive). The equilibrium is \
unstable whenever it has an eigenvalue with absolute value > 1.\
\>", "Text",
 CellChangeTimes->{{3.7814654517132387`*^9, 3.781465455407995*^9}, 
   3.814084704719413*^9},ExpressionUUID->"299b84b6-b272-4a8b-b5a7-\
5add72f98c62"],

Cell[CellGroupData[{

Cell[BoxData[{
 RowBox[{
  RowBox[{"Simplify", "[", 
   RowBox[{
    RowBox[{
     RowBox[{
      RowBox[{"jacob2", "/.", 
       RowBox[{"subeq", "[", 
        RowBox[{"[", "1", "]"}], "]"}]}], "/.", 
      RowBox[{"pf", "\[Rule]", "0"}]}], "/.", 
     RowBox[{"pm", "\[Rule]", "0"}]}], "/.", 
    RowBox[{"noDrivePopSizeSol2", "[", 
     RowBox[{"[", "2", "]"}], "]"}]}], "]"}], ";"}], "\[IndentingNewLine]", 
 RowBox[{"charpolynomial02", "=", 
  RowBox[{"Factor", "[", 
   RowBox[{"Det", "[", 
    RowBox[{"%", "-", 
     RowBox[{"\[Lambda]", "*", 
      RowBox[{"IdentityMatrix", "[", "5", "]"}]}]}], "]"}], "]"}]}]}], "Input",
 CellChangeTimes->{{3.814009388479999*^9, 3.8140094011923647`*^9}, {
   3.814009435990284*^9, 3.814009472662134*^9}, {3.814009517855748*^9, 
   3.8140095446179943`*^9}, 3.8140095798181334`*^9, {3.814009653572708*^9, 
   3.8140096651062803`*^9}, {3.814009729946822*^9, 3.8140097521423693`*^9}},
 CellLabel->
  "In[1369]:=",ExpressionUUID->"ce804ada-232a-456f-84c5-36048946ac34"],

Cell[BoxData[
 RowBox[{"-", 
  FractionBox[
   RowBox[{
    SuperscriptBox["\[Lambda]", "2"], " ", 
    RowBox[{"(", 
     RowBox[{
      RowBox[{"-", "4"}], "+", "b", "+", 
      RowBox[{"2", " ", "\[Lambda]"}]}], ")"}], " ", 
    RowBox[{"(", 
     RowBox[{
      RowBox[{
       RowBox[{"-", "2"}], " ", "c"}], "+", 
      RowBox[{"2", " ", "c", " ", "h", " ", "sf"}], "+", 
      RowBox[{"2", " ", "c", " ", "sm"}], "-", 
      RowBox[{"2", " ", "c", " ", "h", " ", "sf", " ", "sm"}], "-", 
      RowBox[{"2", " ", "c", " ", "\[Delta]"}], "+", 
      RowBox[{"2", " ", "c", " ", "h", " ", "sf", " ", "\[Delta]"}], "+", 
      RowBox[{"2", " ", "c", " ", "sm", " ", "\[Delta]"}], "-", 
      RowBox[{
      "2", " ", "c", " ", "h", " ", "sf", " ", "sm", " ", "\[Delta]"}], "-", 
      "\[Lambda]", "-", 
      RowBox[{"c", " ", "\[Lambda]"}], "+", 
      RowBox[{"h", " ", "sf", " ", "\[Lambda]"}], "+", 
      RowBox[{"c", " ", "h", " ", "sf", " ", "\[Lambda]"}], "+", 
      RowBox[{"2", " ", 
       SuperscriptBox["\[Lambda]", "2"]}], "+", 
      RowBox[{"2", " ", "c", " ", 
       SuperscriptBox["\[Lambda]", "2"]}]}], ")"}]}], 
   RowBox[{"4", " ", 
    RowBox[{"(", 
     RowBox[{"1", "+", "c"}], ")"}]}]]}]], "Output",
 CellChangeTimes->{{3.814009396578875*^9, 3.814009473040482*^9}, {
   3.814009522575335*^9, 3.814009546081353*^9}, 3.814009580161955*^9, {
   3.814009653911927*^9, 3.8140096653633537`*^9}, {3.814009731293776*^9, 
   3.814009752479535*^9}, 3.814084699204679*^9, 3.814196550697547*^9},
 CellLabel->
  "Out[1370]=",ExpressionUUID->"b8275a37-31a0-4c71-ae64-965dc064d9ae"]
}, Open  ]],

Cell[TextData[{
 "The factor ",
 Cell[BoxData[
  RowBox[{"-", 
   FractionBox[
    RowBox[{
     SuperscriptBox["\[Lambda]", "2"], " ", 
     RowBox[{"(", 
      RowBox[{
       RowBox[{"-", "4"}], "+", "b", "+", 
       RowBox[{"2", " ", "\[Lambda]"}]}], ")"}], " "}], 
    RowBox[{"4", " ", 
     RowBox[{"(", 
      RowBox[{"1", "+", "c"}], ")"}]}]]}]],
  CellChangeTimes->{
   3.7812740033403482`*^9, 3.7814573218417273`*^9, 3.7818848609506083`*^9, 
    3.781952821819541*^9, 3.78195289751086*^9, {3.7819538958219957`*^9, 
    3.781953915448163*^9}, 3.7821306544665728`*^9, 3.782747157952016*^9, 
    3.7830788934301643`*^9, 3.785162683931014*^9, 3.785164973948011*^9, 
    3.789732648885466*^9, 3.795174657109274*^9, 3.799320669052409*^9, 
    3.799340142599029*^9, 3.799756425128694*^9, 3.800007655671279*^9, 
    3.8002681846876783`*^9, 3.813930901799224*^9, {3.8139309343336077`*^9, 
    3.8139310598724957`*^9}, 3.8140037587039423`*^9, 3.814003805149556*^9},
  ExpressionUUID->"9e3c4050-daca-4ca4-a643-2ef63379f5fa"],
 " gives two 0 eigenvalues and and eigenvalue of ",
 Cell[BoxData[
  FractionBox[
   RowBox[{"4", "-", "b"}], "2"]],
  CellChangeTimes->{3.814003986297804*^9},ExpressionUUID->
  "9e0166f5-4454-44cf-92ea-4e0bc32f4df0"],
 ", corresponding to the change in overall population size. Thus, we can see \
that the demographic dynamics are orthogonal to the evolutionary dynamics. \n\
\nHere,  we divide through by this factor to simplify the equation and \
evaluate the evolutionary dynamics. We then write the characteristic \
polynomial in the form\n",
 Cell[BoxData[
  RowBox[{
   SuperscriptBox["a\[Lambda]", "2"], "+", "b\[Lambda]", "+", "c"}]],
  CellChangeTimes->{
   3.7812740321927767`*^9, 3.781274073867482*^9, 3.7814285214279013`*^9, 
    3.781457321850852*^9, 3.78188486097983*^9, 3.781952821828101*^9, 
    3.7819528975187283`*^9, {3.781953895876915*^9, 3.7819539154956713`*^9}, 
    3.7821306544745502`*^9, 3.782747157961987*^9, 3.7830788934387407`*^9, 
    3.785162684364028*^9, 3.785164974025779*^9, 3.789732649060177*^9, 
    3.79517465722147*^9, 3.799320669259963*^9, 3.799340142805358*^9, 
    3.799756425312364*^9, 3.800007655751677*^9, 3.80026818474804*^9, 
    3.8139309588718987`*^9, 3.813931061212427*^9, 3.814004119936627*^9},
  ExpressionUUID->"3837d71a-80a8-49b4-9b06-ea6326a5d79a"],
 " \nfinding the a, b, and c coefficients. "
}], "Text",
 CellChangeTimes->{{3.814010329691909*^9, 
  3.81401033968956*^9}},ExpressionUUID->"3314a751-f981-4828-bc7d-\
4866dc84ed1b"],

Cell[CellGroupData[{

Cell[BoxData[{
 RowBox[{
  RowBox[{"Collect", "[", 
   RowBox[{
    RowBox[{"charpolynomial02", "/", 
     RowBox[{"(", 
      RowBox[{"-", 
       FractionBox[
        RowBox[{
         SuperscriptBox["\[Lambda]", "2"], " ", 
         RowBox[{"(", 
          RowBox[{
           RowBox[{"-", "4"}], "+", "b", "+", 
           RowBox[{"2", " ", "\[Lambda]"}]}], ")"}], " "}], 
        RowBox[{"4", " ", 
         RowBox[{"(", 
          RowBox[{"1", "+", "c"}], ")"}]}]]}], ")"}]}], ",", "\[Lambda]", ",",
     "Simplify"}], "]"}], ";"}], "\[IndentingNewLine]", 
 RowBox[{
  RowBox[{"Coefficient", "[", 
   RowBox[{"%", ",", 
    RowBox[{"\[Lambda]", "^", "2"}]}], "]"}], ";"}], "\[IndentingNewLine]", 
 RowBox[{"charpolySimpleACoef02", "=", 
  RowBox[{"Collect", "[", 
   RowBox[{
    RowBox[{"%%", "/", "%"}], ",", "\[Lambda]", ",", "Simplify"}], 
   "]"}]}], "\[IndentingNewLine]", 
 RowBox[{
  RowBox[{"acoef02", "=", 
   RowBox[{"Coefficient", "[", 
    RowBox[{"charpolySimpleACoef02", ",", 
     RowBox[{"\[Lambda]", "^", "2"}]}], "]"}]}], ";"}], "\[IndentingNewLine]", 
 RowBox[{
  RowBox[{"bcoef02", "=", 
   RowBox[{"Coefficient", "[", 
    RowBox[{"charpolySimpleACoef02", ",", "\[Lambda]"}], "]"}]}], 
  ";"}], "\[IndentingNewLine]", 
 RowBox[{
  RowBox[{"ccoef02", "=", 
   RowBox[{
    RowBox[{"charpolySimpleACoef02", " ", "-", " ", 
     RowBox[{"bcoef02", "*", "\[Lambda]"}], "-", 
     RowBox[{"acoef02", "*", 
      RowBox[{"\[Lambda]", "^", "2"}]}]}], " ", "//", "Simplify"}]}], 
  ";"}]}], "Input",
 CellChangeTimes->{{3.781465386858316*^9, 3.781465388026968*^9}, {
  3.781465437030558*^9, 3.781465440011166*^9}, {3.814010383487996*^9, 
  3.8140104313523197`*^9}},
 CellLabel->
  "In[1371]:=",ExpressionUUID->"f2985155-9fb4-430b-a258-99e2fb6a1626"],

Cell[BoxData[
 RowBox[{
  RowBox[{"-", 
   FractionBox[
    RowBox[{"c", " ", 
     RowBox[{"(", 
      RowBox[{
       RowBox[{"-", "1"}], "+", 
       RowBox[{"h", " ", "sf"}]}], ")"}], " ", 
     RowBox[{"(", 
      RowBox[{
       RowBox[{"-", "1"}], "+", "sm"}], ")"}], " ", 
     RowBox[{"(", 
      RowBox[{"1", "+", "\[Delta]"}], ")"}]}], 
    RowBox[{"1", "+", "c"}]]}], "+", 
  RowBox[{
   FractionBox["1", "2"], " ", 
   RowBox[{"(", 
    RowBox[{
     RowBox[{"-", "1"}], "+", 
     RowBox[{"h", " ", "sf"}]}], ")"}], " ", "\[Lambda]"}], "+", 
  SuperscriptBox["\[Lambda]", "2"]}]], "Output",
 CellChangeTimes->{{3.781465383147938*^9, 3.781465388446659*^9}, 
   3.781465441743964*^9, 3.781954038460005*^9, 3.781960483937896*^9, {
   3.783348419284086*^9, 3.783348423689898*^9}, 3.784050027570841*^9, 
   3.785570189009104*^9, 3.7952685727830067`*^9, 3.800009515393708*^9, 
   3.800268244980092*^9, {3.814010406744915*^9, 3.814010431746543*^9}, 
   3.814084710728756*^9, 3.8141965507454*^9},
 CellLabel->
  "Out[1373]=",ExpressionUUID->"883882bc-7fea-4519-8d44-997b117b938f"]
}, Open  ]],

Cell["Now check the discriminant", "Text",
 CellChangeTimes->{{3.7814654589666853`*^9, 
  3.781465462010653*^9}},ExpressionUUID->"57f608a2-425f-4a96-acf0-\
62ca320654b2"],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{
  RowBox[{
   RowBox[{"bcoef02", "^", "2"}], "-", 
   RowBox[{"4", "acoef02", " ", "ccoef02"}]}], " ", "//", 
  "FullSimplify"}]], "Input",
 CellLabel->
  "In[1377]:=",ExpressionUUID->"cd15c18a-b09c-410c-b1f2-430be7bad0f9"],

Cell[BoxData[
 RowBox[{
  FractionBox["1", "4"], " ", 
  RowBox[{"(", 
   RowBox[{
    RowBox[{"-", "1"}], "+", 
    RowBox[{"h", " ", "sf"}]}], ")"}], " ", 
  RowBox[{"(", 
   RowBox[{
    RowBox[{"-", "1"}], "+", 
    RowBox[{"h", " ", "sf"}], "+", 
    FractionBox[
     RowBox[{"16", " ", "c", " ", 
      RowBox[{"(", 
       RowBox[{
        RowBox[{"-", "1"}], "+", "sm"}], ")"}], " ", 
      RowBox[{"(", 
       RowBox[{"1", "+", "\[Delta]"}], ")"}]}], 
     RowBox[{"1", "+", "c"}]]}], ")"}]}]], "Output",
 CellChangeTimes->{
  3.781465463847497*^9, 3.781954041615614*^9, 3.781960485991297*^9, {
   3.7833484193916273`*^9, 3.783348423774037*^9}, 3.7840500277604837`*^9, 
   3.7855701892119904`*^9, 3.795268572979253*^9, 3.8000095155902557`*^9, 
   3.800268245120281*^9, 3.814010444128042*^9, 3.8140847128046722`*^9, 
   3.814196550848081*^9},
 CellLabel->
  "Out[1377]=",ExpressionUUID->"f51e1940-4940-4cd9-91cd-22503cacfd8b"]
}, Open  ]],

Cell[TextData[{
 "We can see that this is always positive as the two terms in the product are \
always \[LessEqual] 0.\n\nThe conditions for the above characteristic \
polynomial ",
 Cell[BoxData[
  FormBox["p", TraditionalForm]],ExpressionUUID->
  "51e4640b-3682-4ba3-8d13-e7338d605e25"],
 "  to have roots between -1 and 1 are:\n1. ",
 Cell[BoxData[
  FormBox[
   RowBox[{
    RowBox[{"p", "(", "1", ")"}], " ", ">", " ", "0"}], TraditionalForm]],
  ExpressionUUID->"dd563b43-f0cb-48c0-a829-a1ea810c92a0"],
 "\n2. ",
 Cell[BoxData[
  FormBox[
   RowBox[{
    RowBox[{"p", "(", 
     RowBox[{"-", "1"}], ")"}], " ", ">", " ", "0"}], TraditionalForm]],
  ExpressionUUID->"6b438e43-4c7d-4965-80b2-94c57348f3a0"],
 "\n3.  ",
 Cell[BoxData[
  FormBox[
   RowBox[{
    RowBox[{"-", "2"}], " ", "<", " ", "b", " ", "<", " ", "2"}], 
   TraditionalForm]],ExpressionUUID->"6b580dce-6b20-4d34-bfb5-1c9ec1662811"],
 "\nWhen any of these are false, the largest eigenvalue is ",
 Cell[BoxData[
  FormBox[
   RowBox[{">", 
    RowBox[{"1", "\:201a", " "}]}], TraditionalForm]],ExpressionUUID->
  "87538b96-b4d0-4369-9a7f-e3273efed38f"],
 "and the equilibrium is unstable. Condition 3 is always true:"
}], "Text",
 CellChangeTimes->{{3.781465492088834*^9, 3.781465506160551*^9}, {
  3.814010450688242*^9, 3.814010451992012*^9}, {3.81401091773033*^9, 
  3.814010947224269*^9}},ExpressionUUID->"88ee105b-b6e1-4def-a77f-\
522399abc6bf"],

Cell[CellGroupData[{

Cell[BoxData[{
 RowBox[{"FullSimplify", "[", 
  RowBox[{
   RowBox[{
    RowBox[{"(", 
     RowBox[{"charpolySimpleACoef02", "/.", 
      RowBox[{"\[Lambda]", "\[Rule]", "1"}]}], " ", ")"}], "<", " ", "0"}], 
   ",", "  ", 
   RowBox[{"{", 
    RowBox[{
     RowBox[{"1", ">", " ", "\[Delta]", ">", "0"}], ",", 
     RowBox[{"1", ">", "h", ">", "0"}], ",", 
     RowBox[{"1", ">", "sm", ">", "0"}], ",", 
     RowBox[{"0", "<", "sf", "<", "1"}]}], "}"}]}], 
  "]"}], "\[IndentingNewLine]", 
 RowBox[{"FullSimplify", "[", 
  RowBox[{
   RowBox[{
    RowBox[{"(", 
     RowBox[{"charpolySimpleACoef02", "/.", 
      RowBox[{"\[Lambda]", "\[Rule]", 
       RowBox[{"-", "1"}]}]}], " ", ")"}], "<", " ", "0"}], ",", "  ", 
   RowBox[{"{", 
    RowBox[{
     RowBox[{"1", ">", " ", "\[Delta]", ">", "0"}], ",", 
     RowBox[{"1", ">", "h", ">", "0"}], ",", 
     RowBox[{"1", ">", "sm", ">", "0"}], ",", 
     RowBox[{"0", "<", "sf", "<", "1"}]}], "}"}]}], 
  "]"}], "\[IndentingNewLine]", 
 RowBox[{"FullSimplify", "[", 
  RowBox[{
   RowBox[{
    RowBox[{"-", "2"}], "<", "bcoef02", "<", "2"}], ",", 
   RowBox[{"{", 
    RowBox[{
     RowBox[{"1", ">", " ", "\[Delta]", ">=", "0"}], ",", 
     RowBox[{"1", ">", "h", ">", "0"}], ",", 
     RowBox[{"1", ">", "sm", ">", "0"}], ",", 
     RowBox[{"0", "<", "sf", "<", "1"}]}], "}"}]}], "]"}]}], "Input",
 CellChangeTimes->{{3.781465526360558*^9, 3.7814655367472467`*^9}, {
   3.781465581129383*^9, 3.781465591995639*^9}, {3.781465688280877*^9, 
   3.781465701486*^9}, 3.78146618447902*^9, 3.7815122676036797`*^9},
 CellLabel->
  "In[1378]:=",ExpressionUUID->"52b9dcb3-9b7e-4ec5-b7ae-59ca876cc397"],

Cell[BoxData[
 RowBox[{
  RowBox[{
   FractionBox["1", "2"], " ", 
   RowBox[{"(", 
    RowBox[{"1", "+", 
     RowBox[{"h", " ", "sf"}], "-", 
     FractionBox[
      RowBox[{"2", " ", "c", " ", 
       RowBox[{"(", 
        RowBox[{
         RowBox[{"-", "1"}], "+", 
         RowBox[{"h", " ", "sf"}]}], ")"}], " ", 
       RowBox[{"(", 
        RowBox[{
         RowBox[{"-", "1"}], "+", "sm"}], ")"}], " ", 
       RowBox[{"(", 
        RowBox[{"1", "+", "\[Delta]"}], ")"}]}], 
      RowBox[{"1", "+", "c"}]]}], ")"}]}], "<", "0"}]], "Output",
 CellChangeTimes->{{3.7814655286447268`*^9, 3.781465538365871*^9}, {
   3.781465582955743*^9, 3.781465592641117*^9}, 3.781465701977508*^9, 
   3.781466185013439*^9, 3.781512269203911*^9, 3.781954043802164*^9, 
   3.7819604884610863`*^9, {3.7833484196520147`*^9, 3.7833484238556013`*^9}, 
   3.784050028079029*^9, 3.785570189683228*^9, 3.7952685733076477`*^9, 
   3.8000095159012947`*^9, 3.800268245391749*^9, 3.814010457997061*^9, 
   3.8140847153656588`*^9, 3.814196551073114*^9},
 CellLabel->
  "Out[1378]=",ExpressionUUID->"4c98ecf0-81bb-42c5-a010-6b750005e07f"],

Cell[BoxData[
 RowBox[{
  RowBox[{
   FractionBox["1", "2"], " ", 
   RowBox[{"(", 
    RowBox[{"3", "-", 
     RowBox[{"h", " ", "sf"}], "-", 
     FractionBox[
      RowBox[{"2", " ", "c", " ", 
       RowBox[{"(", 
        RowBox[{
         RowBox[{"-", "1"}], "+", 
         RowBox[{"h", " ", "sf"}]}], ")"}], " ", 
       RowBox[{"(", 
        RowBox[{
         RowBox[{"-", "1"}], "+", "sm"}], ")"}], " ", 
       RowBox[{"(", 
        RowBox[{"1", "+", "\[Delta]"}], ")"}]}], 
      RowBox[{"1", "+", "c"}]]}], ")"}]}], "<", "0"}]], "Output",
 CellChangeTimes->{{3.7814655286447268`*^9, 3.781465538365871*^9}, {
   3.781465582955743*^9, 3.781465592641117*^9}, 3.781465701977508*^9, 
   3.781466185013439*^9, 3.781512269203911*^9, 3.781954043802164*^9, 
   3.7819604884610863`*^9, {3.7833484196520147`*^9, 3.7833484238556013`*^9}, 
   3.784050028079029*^9, 3.785570189683228*^9, 3.7952685733076477`*^9, 
   3.8000095159012947`*^9, 3.800268245391749*^9, 3.814010457997061*^9, 
   3.8140847153656588`*^9, 3.814196551293988*^9},
 CellLabel->
  "Out[1379]=",ExpressionUUID->"38f36a22-7bfa-409e-a9ad-068d6af4ee23"],

Cell[BoxData["True"], "Output",
 CellChangeTimes->{{3.7814655286447268`*^9, 3.781465538365871*^9}, {
   3.781465582955743*^9, 3.781465592641117*^9}, 3.781465701977508*^9, 
   3.781466185013439*^9, 3.781512269203911*^9, 3.781954043802164*^9, 
   3.7819604884610863`*^9, {3.7833484196520147`*^9, 3.7833484238556013`*^9}, 
   3.784050028079029*^9, 3.785570189683228*^9, 3.7952685733076477`*^9, 
   3.8000095159012947`*^9, 3.800268245391749*^9, 3.814010457997061*^9, 
   3.8140847153656588`*^9, 3.814196551298975*^9},
 CellLabel->
  "Out[1380]=",ExpressionUUID->"dcbc15c1-1d15-4e82-a93b-21a6302d6939"]
}, Open  ]],

Cell["\<\
The second of these inequalities implies the first and so we need only the \
first (only one condition is needed for instability).\
\>", "Text",
 CellChangeTimes->{{3.781465759480749*^9, 3.7814657696038733`*^9}, {
  3.781465837640523*^9, 3.7814658532489157`*^9}, {3.78187395266564*^9, 
  3.781873964963763*^9}},ExpressionUUID->"6b47a709-ea8f-4dcd-9dcf-\
4cc6acfed59b"],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"unstabConditions02", "=", 
  RowBox[{"FullSimplify", "[", 
   RowBox[{
    RowBox[{
     RowBox[{"(", 
      RowBox[{"charpolySimpleACoef02", "/.", 
       RowBox[{"\[Lambda]", "\[Rule]", "1"}]}], " ", ")"}], "<", " ", "0"}], 
    ",", "  ", 
    RowBox[{"{", 
     RowBox[{
      RowBox[{"1", ">", " ", "\[Delta]", ">", "0"}], ",", 
      RowBox[{"1", ">", "h", ">", "0"}], ",", 
      RowBox[{"1", ">", "sm", ">", "0"}], ",", 
      RowBox[{"0", "<", "sf", "<", "1"}]}], "}"}]}], "]"}]}]], "Input",
 CellChangeTimes->{{3.781465877162455*^9, 3.781465971554441*^9}, 
   3.781873788934682*^9},
 CellLabel->
  "In[1381]:=",ExpressionUUID->"ca6a9cb5-1fa5-45f3-b3f0-9b6ffdce575c"],

Cell[BoxData[
 RowBox[{
  RowBox[{
   FractionBox["1", "2"], " ", 
   RowBox[{"(", 
    RowBox[{"1", "+", 
     RowBox[{"h", " ", "sf"}], "-", 
     FractionBox[
      RowBox[{"2", " ", "c", " ", 
       RowBox[{"(", 
        RowBox[{
         RowBox[{"-", "1"}], "+", 
         RowBox[{"h", " ", "sf"}]}], ")"}], " ", 
       RowBox[{"(", 
        RowBox[{
         RowBox[{"-", "1"}], "+", "sm"}], ")"}], " ", 
       RowBox[{"(", 
        RowBox[{"1", "+", "\[Delta]"}], ")"}]}], 
      RowBox[{"1", "+", "c"}]]}], ")"}]}], "<", "0"}]], "Output",
 CellChangeTimes->{{3.781465886634276*^9, 3.781465972182002*^9}, 
   3.781873789761869*^9, 3.7819540469446173`*^9, 3.781960492622634*^9, {
   3.78334841995374*^9, 3.783348423946249*^9}, 3.784050028762014*^9, 
   3.78557019030302*^9, 3.795268573966165*^9, 3.8000095165324993`*^9, 
   3.800268245752823*^9, 3.814010464425288*^9, 3.8140847174236107`*^9, 
   3.814196551402801*^9},
 CellLabel->
  "Out[1381]=",ExpressionUUID->"44829f00-2da3-476e-8be2-82f3c37d989e"]
}, Open  ]],

Cell["which can be re-written using the w fitness notation", "Text",
 CellChangeTimes->{
  3.814010967375692*^9},ExpressionUUID->"30fbf15a-a200-4476-8ddb-\
0150815c1608"],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"Simplify", "[", 
  RowBox[{"unstabConditions02", ",", " ", 
   RowBox[{"{", 
    RowBox[{
     RowBox[{"wXdX", " ", "\[Equal]", " ", 
      RowBox[{"1", " ", "-", " ", 
       RowBox[{"h", " ", "sf"}]}]}], ",", " ", 
     RowBox[{"wXdXd", " ", "\[Equal]", " ", 
      RowBox[{"1", " ", "-", " ", "sf"}]}], ",", " ", 
     RowBox[{"wXdY", " ", "\[Equal]", " ", 
      RowBox[{"1", " ", "-", " ", "sm"}]}]}], "}"}]}], "]"}]], "Input",
 CellChangeTimes->{{3.7833569152996073`*^9, 3.7833569483092127`*^9}},
 CellLabel->
  "In[1382]:=",ExpressionUUID->"a57737c8-21c9-4cbd-bdf3-ef59defff834"],

Cell[BoxData[
 RowBox[{
  RowBox[{
   RowBox[{"(", 
    RowBox[{"1", "+", "c"}], ")"}], " ", 
   RowBox[{"(", 
    RowBox[{
     RowBox[{"-", "2"}], "+", "wXdX", "+", 
     RowBox[{"c", " ", 
      RowBox[{"(", 
       RowBox[{
        RowBox[{"-", "2"}], "+", "wXdX", "+", 
        RowBox[{"2", " ", "wXdX", " ", "wXdY", " ", 
         RowBox[{"(", 
          RowBox[{"1", "+", "\[Delta]"}], ")"}]}]}], ")"}]}]}], ")"}]}], ">", 
  "0"}]], "Output",
 CellChangeTimes->{{3.783356930939971*^9, 3.7833569486658287`*^9}, 
   3.7840500291045218`*^9, 3.785570190631494*^9, 3.7952685742774467`*^9, 
   3.800009516893228*^9, 3.800268245888525*^9, 3.814010466301713*^9, 
   3.814010968648089*^9, 3.8140847194302998`*^9, 3.8141965514942617`*^9},
 CellLabel->
  "Out[1382]=",ExpressionUUID->"f55c05fd-d07f-4418-8673-45d9c298035c"]
}, Open  ]]
}, Closed]],

Cell[CellGroupData[{

Cell["drive fixed (pm=pf=1) equilibrium", "Subsubsection",
 CellChangeTimes->{{3.7814659855542717`*^9, 3.781465989505756*^9}, 
   3.8140187152461357`*^9},ExpressionUUID->"0cf784ab-e1fd-40fd-aabe-\
51311338b372"],

Cell["\<\
We proceed similarly to before, but now considering the other equilibrium:\
\>", "Text",
 CellChangeTimes->{
  3.814010990066278*^9},ExpressionUUID->"5dffe054-dd46-4f7a-b6d9-\
eb3b1c467ab4"],

Cell[CellGroupData[{

Cell[BoxData[{
 RowBox[{
  RowBox[{"Simplify", "[", 
   RowBox[{
    RowBox[{
     RowBox[{
      RowBox[{"jacob2", "/.", 
       RowBox[{"subeq", "[", 
        RowBox[{"[", "1", "]"}], "]"}]}], "/.", 
      RowBox[{"pf", "\[Rule]", "1"}]}], "/.", 
     RowBox[{"pm", "\[Rule]", "1"}]}], "/.", 
    RowBox[{"fixedDrivePopSizeSol2", "[", 
     RowBox[{"[", "2", "]"}], "]"}]}], "]"}], ";"}], "\[IndentingNewLine]", 
 RowBox[{"charpolynomial12", "=", 
  RowBox[{"Factor", "[", 
   RowBox[{"Det", "[", 
    RowBox[{"%", "-", 
     RowBox[{"\[Lambda]", "*", 
      RowBox[{"IdentityMatrix", "[", "5", "]"}]}]}], "]"}], "]"}]}]}], "Input",
 CellChangeTimes->{{3.8140110467667093`*^9, 3.814011063131847*^9}},
 CellLabel->
  "In[1383]:=",ExpressionUUID->"56682ae3-a806-4c2e-8a48-cc078ee8ed44"],

Cell[BoxData[
 RowBox[{
  FractionBox["1", 
   RowBox[{"4", " ", 
    RowBox[{"(", 
     RowBox[{"1", "+", "c"}], ")"}], " ", 
    RowBox[{"(", 
     RowBox[{
      RowBox[{"-", "1"}], "+", "sf"}], ")"}], " ", 
    RowBox[{"(", 
     RowBox[{
      RowBox[{"-", "1"}], "+", "sm"}], ")"}], " ", 
    RowBox[{"(", 
     RowBox[{"1", "+", "\[Delta]"}], ")"}]}]], 
  RowBox[{
   RowBox[{"(", 
    RowBox[{"4", "-", "b", "+", 
     RowBox[{"b", " ", "sf"}], "-", 
     RowBox[{"b", " ", "\[Delta]"}], "+", 
     RowBox[{"b", " ", "sf", " ", "\[Delta]"}], "-", 
     RowBox[{"2", " ", "\[Lambda]"}]}], ")"}], " ", 
   SuperscriptBox["\[Lambda]", "2"], " ", 
   RowBox[{"(", 
    RowBox[{
     RowBox[{"-", "2"}], "+", 
     RowBox[{"2", " ", "h", " ", "sf"}], "-", "\[Lambda]", "-", 
     RowBox[{"c", " ", "\[Lambda]"}], "+", 
     RowBox[{"h", " ", "sf", " ", "\[Lambda]"}], "+", 
     RowBox[{"c", " ", "h", " ", "sf", " ", "\[Lambda]"}], "+", 
     RowBox[{"sm", " ", "\[Lambda]"}], "+", 
     RowBox[{"c", " ", "sm", " ", "\[Lambda]"}], "-", 
     RowBox[{"h", " ", "sf", " ", "sm", " ", "\[Lambda]"}], "-", 
     RowBox[{"c", " ", "h", " ", "sf", " ", "sm", " ", "\[Lambda]"}], "-", 
     RowBox[{"\[Delta]", " ", "\[Lambda]"}], "-", 
     RowBox[{"c", " ", "\[Delta]", " ", "\[Lambda]"}], "+", 
     RowBox[{"h", " ", "sf", " ", "\[Delta]", " ", "\[Lambda]"}], "+", 
     RowBox[{"c", " ", "h", " ", "sf", " ", "\[Delta]", " ", "\[Lambda]"}], 
     "+", 
     RowBox[{"sm", " ", "\[Delta]", " ", "\[Lambda]"}], "+", 
     RowBox[{"c", " ", "sm", " ", "\[Delta]", " ", "\[Lambda]"}], "-", 
     RowBox[{"h", " ", "sf", " ", "sm", " ", "\[Delta]", " ", "\[Lambda]"}], 
     "-", 
     RowBox[{
     "c", " ", "h", " ", "sf", " ", "sm", " ", "\[Delta]", " ", "\[Lambda]"}],
      "+", 
     RowBox[{"2", " ", 
      SuperscriptBox["\[Lambda]", "2"]}], "+", 
     RowBox[{"2", " ", "c", " ", 
      SuperscriptBox["\[Lambda]", "2"]}], "-", 
     RowBox[{"2", " ", "sf", " ", 
      SuperscriptBox["\[Lambda]", "2"]}], "-", 
     RowBox[{"2", " ", "c", " ", "sf", " ", 
      SuperscriptBox["\[Lambda]", "2"]}], "-", 
     RowBox[{"2", " ", "sm", " ", 
      SuperscriptBox["\[Lambda]", "2"]}], "-", 
     RowBox[{"2", " ", "c", " ", "sm", " ", 
      SuperscriptBox["\[Lambda]", "2"]}], "+", 
     RowBox[{"2", " ", "sf", " ", "sm", " ", 
      SuperscriptBox["\[Lambda]", "2"]}], "+", 
     RowBox[{"2", " ", "c", " ", "sf", " ", "sm", " ", 
      SuperscriptBox["\[Lambda]", "2"]}], "+", 
     RowBox[{"2", " ", "\[Delta]", " ", 
      SuperscriptBox["\[Lambda]", "2"]}], "+", 
     RowBox[{"2", " ", "c", " ", "\[Delta]", " ", 
      SuperscriptBox["\[Lambda]", "2"]}], "-", 
     RowBox[{"2", " ", "sf", " ", "\[Delta]", " ", 
      SuperscriptBox["\[Lambda]", "2"]}], "-", 
     RowBox[{"2", " ", "c", " ", "sf", " ", "\[Delta]", " ", 
      SuperscriptBox["\[Lambda]", "2"]}], "-", 
     RowBox[{"2", " ", "sm", " ", "\[Delta]", " ", 
      SuperscriptBox["\[Lambda]", "2"]}], "-", 
     RowBox[{"2", " ", "c", " ", "sm", " ", "\[Delta]", " ", 
      SuperscriptBox["\[Lambda]", "2"]}], "+", 
     RowBox[{"2", " ", "sf", " ", "sm", " ", "\[Delta]", " ", 
      SuperscriptBox["\[Lambda]", "2"]}], "+", 
     RowBox[{"2", " ", "c", " ", "sf", " ", "sm", " ", "\[Delta]", " ", 
      SuperscriptBox["\[Lambda]", "2"]}]}], ")"}]}]}]], "Output",
 CellChangeTimes->{3.8140110640091867`*^9, 3.814084724073814*^9, 
  3.814196552357511*^9},
 CellLabel->
  "Out[1384]=",ExpressionUUID->"c5a617cd-1425-41dd-91b9-09bdde328bbc"]
}, Open  ]],

Cell[TextData[{
 "The factor ",
 Cell[BoxData[
  FractionBox[
   RowBox[{
    RowBox[{"(", 
     RowBox[{"4", "-", "b", "+", 
      RowBox[{"b", " ", "sf"}], "-", 
      RowBox[{"b", " ", "\[Delta]"}], "+", 
      RowBox[{"b", " ", "sf", " ", "\[Delta]"}], "-", 
      RowBox[{"2", " ", "\[Lambda]"}]}], ")"}], " ", 
    SuperscriptBox["\[Lambda]", "2"]}], 
   RowBox[{"4", " ", 
    RowBox[{"(", 
     RowBox[{"1", "+", "c"}], ")"}], " ", 
    RowBox[{"(", 
     RowBox[{
      RowBox[{"-", "1"}], "+", "sf"}], ")"}], " ", 
    RowBox[{"(", 
     RowBox[{
      RowBox[{"-", "1"}], "+", "sm"}], ")"}], " ", 
    RowBox[{"(", 
     RowBox[{"1", "+", "\[Delta]"}], ")"}]}]]],
  CellChangeTimes->{
   3.7812740033403482`*^9, 3.7814573218417273`*^9, 3.7818848609506083`*^9, 
    3.781952821819541*^9, 3.78195289751086*^9, {3.7819538958219957`*^9, 
    3.781953915448163*^9}, 3.7821306544665728`*^9, 3.782747157952016*^9, 
    3.7830788934301643`*^9, 3.785162683931014*^9, 3.785164973948011*^9, 
    3.789732648885466*^9, 3.795174657109274*^9, 3.799320669052409*^9, 
    3.799340142599029*^9, 3.799756425128694*^9, 3.800007655671279*^9, 
    3.8002681846876783`*^9, 3.813930901799224*^9, {3.8139309343336077`*^9, 
    3.8139310598724957`*^9}, 3.8140037587039423`*^9, 3.814003805149556*^9},
  ExpressionUUID->"75708eb9-8704-47dc-8b68-3b6777691014"],
 " gives two 0 eigenvalues and and eigenvalue of ",
 Cell[BoxData[
  RowBox[{"2", "+", 
   RowBox[{
    FractionBox["1", "2"], " ", "b", " ", 
    RowBox[{"(", 
     RowBox[{
      RowBox[{"-", "1"}], "+", "sf"}], ")"}], " ", 
    RowBox[{"(", 
     RowBox[{"1", "+", "\[Delta]"}], ")"}]}]}]],
  CellChangeTimes->{3.814003986297804*^9},ExpressionUUID->
  "f474377d-fead-44b7-bf5d-f8c6cf558e99"],
 ", corresponding to the stability of the ecological equilibrium population \
size. Thus, we can see that the demographic dynamics are orthogonal to the \
evolutionary dynamics. \n\nHere,  we divide through by this factor to \
simplify the equation and evaluate the evolutionary dynamics. We then write \
the characteristic polynomial in the form\n",
 Cell[BoxData[
  RowBox[{
   SuperscriptBox["a\[Lambda]", "2"], "+", "b\[Lambda]", "+", "c"}]],
  CellChangeTimes->{
   3.7812740321927767`*^9, 3.781274073867482*^9, 3.7814285214279013`*^9, 
    3.781457321850852*^9, 3.78188486097983*^9, 3.781952821828101*^9, 
    3.7819528975187283`*^9, {3.781953895876915*^9, 3.7819539154956713`*^9}, 
    3.7821306544745502`*^9, 3.782747157961987*^9, 3.7830788934387407`*^9, 
    3.785162684364028*^9, 3.785164974025779*^9, 3.789732649060177*^9, 
    3.79517465722147*^9, 3.799320669259963*^9, 3.799340142805358*^9, 
    3.799756425312364*^9, 3.800007655751677*^9, 3.80026818474804*^9, 
    3.8139309588718987`*^9, 3.813931061212427*^9, 3.814004119936627*^9},
  ExpressionUUID->"4d2067d0-ccb1-47e5-bc32-1db8f70aeaed"],
 " \nfinding the a, b, and c coefficients. "
}], "Text",
 CellChangeTimes->{
  3.81401109022639*^9, {3.814011153607579*^9, 
   3.81401115360782*^9}},ExpressionUUID->"dce823c8-8f9a-4560-8b4f-\
352fa58efdac"],

Cell[CellGroupData[{

Cell[BoxData[{
 RowBox[{
  RowBox[{"Collect", "[", 
   RowBox[{
    RowBox[{"charpolynomial12", "/", 
     RowBox[{"(", 
      FractionBox[
       RowBox[{
        RowBox[{"(", 
         RowBox[{"4", "-", "b", "+", 
          RowBox[{"b", " ", "sf"}], "-", 
          RowBox[{"b", " ", "\[Delta]"}], "+", 
          RowBox[{"b", " ", "sf", " ", "\[Delta]"}], "-", 
          RowBox[{"2", " ", "\[Lambda]"}]}], ")"}], " ", 
        SuperscriptBox["\[Lambda]", "2"]}], 
       RowBox[{"4", " ", 
        RowBox[{"(", 
         RowBox[{"1", "+", "c"}], ")"}], " ", 
        RowBox[{"(", 
         RowBox[{
          RowBox[{"-", "1"}], "+", "sf"}], ")"}], " ", 
        RowBox[{"(", 
         RowBox[{
          RowBox[{"-", "1"}], "+", "sm"}], ")"}], " ", 
        RowBox[{"(", 
         RowBox[{"1", "+", "\[Delta]"}], ")"}]}]], ")"}]}], ",", "\[Lambda]", 
    ",", "Simplify"}], "]"}], ";"}], "\[IndentingNewLine]", 
 RowBox[{
  RowBox[{"Coefficient", "[", 
   RowBox[{"%", ",", 
    RowBox[{"\[Lambda]", "^", "2"}]}], "]"}], ";"}], "\[IndentingNewLine]", 
 RowBox[{"charpolySimpleACoef12", "=", 
  RowBox[{"Collect", "[", 
   RowBox[{
    RowBox[{"%%", "/", "%"}], ",", "\[Lambda]", ",", "Simplify"}], 
   "]"}]}], "\[IndentingNewLine]", 
 RowBox[{
  RowBox[{"acoef12", "=", 
   RowBox[{"Coefficient", "[", 
    RowBox[{"charpolySimpleACoef12", ",", 
     RowBox[{"\[Lambda]", "^", "2"}]}], "]"}]}], ";"}], "\[IndentingNewLine]", 
 RowBox[{
  RowBox[{"bcoef12", "=", 
   RowBox[{"Coefficient", "[", 
    RowBox[{"charpolySimpleACoef12", ",", "\[Lambda]"}], "]"}]}], 
  ";"}], "\[IndentingNewLine]", 
 RowBox[{
  RowBox[{"ccoef12", "=", 
   RowBox[{"charpolySimpleACoef12", "-", " ", 
    RowBox[{"bcoef12", "*", "\[Lambda]"}], "-", 
    RowBox[{"acoef12", "*", 
     RowBox[{"\[Lambda]", "^", "2"}]}]}]}], ";"}]}], "Input",
 CellChangeTimes->{{3.7814660158915987`*^9, 3.781466030367185*^9}, {
  3.781511388407653*^9, 3.781511406925776*^9}, {3.7815114853232603`*^9, 
  3.781511487165525*^9}, {3.8140110114581947`*^9, 3.814011035726304*^9}, {
  3.81401119347377*^9, 3.8140112094565897`*^9}},
 CellLabel->
  "In[1385]:=",ExpressionUUID->"dfa473d8-2b51-429d-b510-5c1e5c37f9aa"],

Cell[BoxData[
 RowBox[{
  FractionBox[
   RowBox[{
    RowBox[{"-", "1"}], "+", 
    RowBox[{"h", " ", "sf"}]}], 
   RowBox[{
    RowBox[{"(", 
     RowBox[{"1", "+", "c"}], ")"}], " ", 
    RowBox[{"(", 
     RowBox[{
      RowBox[{"-", "1"}], "+", "sf"}], ")"}], " ", 
    RowBox[{"(", 
     RowBox[{
      RowBox[{"-", "1"}], "+", "sm"}], ")"}], " ", 
    RowBox[{"(", 
     RowBox[{"1", "+", "\[Delta]"}], ")"}]}]], "+", 
  FractionBox[
   RowBox[{
    RowBox[{"(", 
     RowBox[{"1", "-", 
      RowBox[{"h", " ", "sf"}]}], ")"}], " ", "\[Lambda]"}], 
   RowBox[{"2", " ", 
    RowBox[{"(", 
     RowBox[{
      RowBox[{"-", "1"}], "+", "sf"}], ")"}]}]], "+", 
  SuperscriptBox["\[Lambda]", "2"]}]], "Output",
 CellChangeTimes->{{3.781465999356587*^9, 3.781466030701129*^9}, {
   3.781511391556911*^9, 3.7815114073942223`*^9}, 3.7815114881434813`*^9, 
   3.7815133913596582`*^9, 3.7819604953090477`*^9, 3.783348438199882*^9, 
   3.7840500300847263`*^9, 3.7855701908368073`*^9, 3.7952685744353533`*^9, 
   3.800009517092691*^9, 3.800268246008481*^9, 3.814011037187717*^9, {
   3.81401120075671*^9, 3.81401120992056*^9}, 3.814084725708302*^9, 
   3.814196552428441*^9},
 CellLabel->
  "Out[1387]=",ExpressionUUID->"99711265-4b94-4f5e-870c-d29f8ac76404"]
}, Open  ]],

Cell["\<\
The discriminant is nonnegative and so the eigenvalues are real:\
\>", "Text",
 CellChangeTimes->{{3.781466041455315*^9, 3.781466045839364*^9}, 
   3.8140112634898663`*^9},ExpressionUUID->"64fd5d32-57a1-45d1-8bc8-\
3b887885e303"],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{
  RowBox[{"bcoef12", "^", "2"}], "-", 
  RowBox[{"4", "acoef12", " ", "ccoef12"}]}]], "Input",
 CellChangeTimes->{{3.78146605752468*^9, 3.7814660576893463`*^9}},
 CellLabel->
  "In[1391]:=",ExpressionUUID->"8e4a5cd7-1809-43ec-a6c7-a8fb59e6e802"],

Cell[BoxData[
 RowBox[{
  FractionBox[
   SuperscriptBox[
    RowBox[{"(", 
     RowBox[{"1", "-", 
      RowBox[{"h", " ", "sf"}]}], ")"}], "2"], 
   RowBox[{"4", " ", 
    SuperscriptBox[
     RowBox[{"(", 
      RowBox[{
       RowBox[{"-", "1"}], "+", "sf"}], ")"}], "2"]}]], "-", 
  FractionBox[
   RowBox[{"4", " ", 
    RowBox[{"(", 
     RowBox[{
      RowBox[{"-", "1"}], "+", 
      RowBox[{"h", " ", "sf"}]}], ")"}]}], 
   RowBox[{
    RowBox[{"(", 
     RowBox[{"1", "+", "c"}], ")"}], " ", 
    RowBox[{"(", 
     RowBox[{
      RowBox[{"-", "1"}], "+", "sf"}], ")"}], " ", 
    RowBox[{"(", 
     RowBox[{
      RowBox[{"-", "1"}], "+", "sm"}], ")"}], " ", 
    RowBox[{"(", 
     RowBox[{"1", "+", "\[Delta]"}], ")"}]}]]}]], "Output",
 CellChangeTimes->{{3.78146605450915*^9, 3.7814660582264423`*^9}, 
   3.781513396654119*^9, 3.7819604977917757`*^9, 3.783348438423812*^9, 
   3.784050030288706*^9, 3.7855701909125433`*^9, 3.795268574496313*^9, 
   3.800009517154484*^9, 3.80026824614252*^9, 3.8140847275646143`*^9, 
   3.814196552436729*^9},
 CellLabel->
  "Out[1391]=",ExpressionUUID->"ac09b42e-a463-40b9-8809-0c1dfe5510da"]
}, Open  ]],

Cell[TextData[{
 "The conditions for the above characteristic polynomial ",
 Cell[BoxData[
  FormBox["p", TraditionalForm]],ExpressionUUID->
  "d636a5a7-a2c7-4272-b1ab-0f8ac1345d73"],
 "  to have roots between -1 and 1 are:\n1. ",
 Cell[BoxData[
  FormBox[
   RowBox[{
    RowBox[{"p", "(", "1", ")"}], " ", ">", " ", "0"}], TraditionalForm]],
  ExpressionUUID->"d264372d-3899-4857-8fe4-c948349048c8"],
 "\n2. ",
 Cell[BoxData[
  FormBox[
   RowBox[{
    RowBox[{"p", "(", 
     RowBox[{"-", "1"}], ")"}], " ", ">", " ", "0"}], TraditionalForm]],
  ExpressionUUID->"6fea94e4-6cc6-4f65-a3c5-8cadb2ffb726"],
 "\n3.  ",
 Cell[BoxData[
  FormBox[
   RowBox[{
    RowBox[{"-", "2"}], " ", "<", " ", "b", " "}], TraditionalForm]],
  ExpressionUUID->"751d54c2-ab11-4d07-ba81-ca22959e1d4c"],
 "\n4.  ",
 Cell[BoxData[
  FormBox[
   RowBox[{"b", " ", "<", " ", "2", " "}], TraditionalForm]],ExpressionUUID->
  "de60e943-19b8-4239-bd10-829ebe67cf68"],
 "\nWhen any of these are false, the largest eigenvalue is ",
 Cell[BoxData[
  FormBox[
   RowBox[{">", 
    RowBox[{"1", "\:201a", " "}]}], TraditionalForm]],ExpressionUUID->
  "2287487e-c4c0-4772-b521-6c8abfde88c8"],
 "and the equilibrium is unstable. \n\nNow checking these conditions for \
instability by reversing the inequalities required for stability"
}], "Text",
 CellChangeTimes->{{3.781466076377097*^9, 3.78146608440063*^9}, {
   3.781466192176076*^9, 3.781466195526973*^9}, {3.781960505405582*^9, 
   3.781960506689489*^9}, 
   3.814011272224868*^9},ExpressionUUID->"0010841b-1eca-4e62-9dfb-\
ed7eff11b6fe"],

Cell[CellGroupData[{

Cell[BoxData[{
 RowBox[{"r1", "=", 
  RowBox[{"FullSimplify", "[", 
   RowBox[{
    RowBox[{
     RowBox[{"(", 
      RowBox[{"charpolySimpleACoef12", "/.", 
       RowBox[{"\[Lambda]", "\[Rule]", "1"}]}], " ", ")"}], "<", " ", "0"}], 
    ",", "  ", 
    RowBox[{"{", 
     RowBox[{
      RowBox[{"1", ">", " ", "\[Delta]", ">", "0"}], ",", 
      RowBox[{"1", ">", "h", ">", "0"}], ",", 
      RowBox[{"1", ">", "sm", ">", "0"}], ",", 
      RowBox[{"0", "<", "sf", "<", "1"}]}], "}"}]}], 
   "]"}]}], "\[IndentingNewLine]", 
 RowBox[{"r2", "=", 
  RowBox[{"FullSimplify", "[", 
   RowBox[{
    RowBox[{
     RowBox[{"(", 
      RowBox[{"charpolySimpleACoef12", "/.", 
       RowBox[{"\[Lambda]", "\[Rule]", 
        RowBox[{"-", "1"}]}]}], " ", ")"}], "<", " ", "0"}], ",", "  ", 
    RowBox[{"{", 
     RowBox[{
      RowBox[{"1", ">", " ", "\[Delta]", ">", "0"}], ",", 
      RowBox[{"1", ">", "h", ">", "0"}], ",", 
      RowBox[{"1", ">", "sm", ">", "0"}], ",", 
      RowBox[{"0", "<", "sf", "<", "1"}]}], "}"}]}], 
   "]"}]}], "\[IndentingNewLine]", 
 RowBox[{"r3", "=", 
  RowBox[{"FullSimplify", "[", 
   RowBox[{
    RowBox[{"bcoef12", "<", 
     RowBox[{"-", "2"}]}], ",", 
    RowBox[{"{", 
     RowBox[{
      RowBox[{"1", ">", " ", "\[Delta]", ">", "0"}], ",", 
      RowBox[{"1", ">", "h", ">", "0"}], ",", 
      RowBox[{"1", ">", "sm", ">", "0"}], ",", 
      RowBox[{"0", "<", "sf", "<", "1"}]}], "}"}]}], 
   "]"}]}], "\[IndentingNewLine]", 
 RowBox[{"r4", "=", 
  RowBox[{"FullSimplify", "[", 
   RowBox[{
    RowBox[{"2", "<", "bcoef12"}], ",", 
    RowBox[{"{", 
     RowBox[{
      RowBox[{"1", ">", " ", "\[Delta]", ">", "0"}], ",", 
      RowBox[{"1", ">", "h", ">", "0"}], ",", 
      RowBox[{"1", ">", "sm", ">", "0"}], ",", 
      RowBox[{"0", "<", "sf", "<", "1"}]}], "}"}]}], "]"}]}]}], "Input",
 CellChangeTimes->{{3.781466198557598*^9, 3.781466244073738*^9}, {
  3.781513299834971*^9, 3.78151330877592*^9}},
 CellLabel->
  "In[1392]:=",ExpressionUUID->"866dbab8-0523-4787-88c5-77d23c472993"],

Cell[BoxData[
 RowBox[{
  RowBox[{"1", "+", 
   FractionBox[
    RowBox[{"1", "-", 
     RowBox[{"h", " ", "sf"}]}], 
    RowBox[{"2", " ", 
     RowBox[{"(", 
      RowBox[{
       RowBox[{"-", "1"}], "+", "sf"}], ")"}]}]], "+", 
   FractionBox[
    RowBox[{
     RowBox[{"-", "1"}], "+", 
     RowBox[{"h", " ", "sf"}]}], 
    RowBox[{
     RowBox[{"(", 
      RowBox[{"1", "+", "c"}], ")"}], " ", 
     RowBox[{"(", 
      RowBox[{
       RowBox[{"-", "1"}], "+", "sf"}], ")"}], " ", 
     RowBox[{"(", 
      RowBox[{
       RowBox[{"-", "1"}], "+", "sm"}], ")"}], " ", 
     RowBox[{"(", 
      RowBox[{"1", "+", "\[Delta]"}], ")"}]}]]}], "<", "0"}]], "Output",
 CellChangeTimes->{
  3.781466207213058*^9, 3.7814662447499647`*^9, {3.78151331072229*^9, 
   3.781513335444324*^9}, 3.781513400170166*^9, 3.781960509141514*^9, 
   3.7833484391219378`*^9, 3.7840500312745237`*^9, 3.785570191793244*^9, 
   3.7952685753436327`*^9, 3.80000951794683*^9, 3.80026824682135*^9, 
   3.814011276950014*^9, 3.814084729882031*^9, 3.8141965530930243`*^9},
 CellLabel->
  "Out[1392]=",ExpressionUUID->"2b73118d-1bcb-49b9-9740-0d3679e45fad"],

Cell[BoxData[
 RowBox[{
  RowBox[{"1", "+", 
   FractionBox[
    RowBox[{"1", "-", 
     RowBox[{"h", " ", "sf"}]}], 
    RowBox[{"2", "-", 
     RowBox[{"2", " ", "sf"}]}]], "+", 
   FractionBox[
    RowBox[{
     RowBox[{"-", "1"}], "+", 
     RowBox[{"h", " ", "sf"}]}], 
    RowBox[{
     RowBox[{"(", 
      RowBox[{"1", "+", "c"}], ")"}], " ", 
     RowBox[{"(", 
      RowBox[{
       RowBox[{"-", "1"}], "+", "sf"}], ")"}], " ", 
     RowBox[{"(", 
      RowBox[{
       RowBox[{"-", "1"}], "+", "sm"}], ")"}], " ", 
     RowBox[{"(", 
      RowBox[{"1", "+", "\[Delta]"}], ")"}]}]]}], "<", "0"}]], "Output",
 CellChangeTimes->{
  3.781466207213058*^9, 3.7814662447499647`*^9, {3.78151331072229*^9, 
   3.781513335444324*^9}, 3.781513400170166*^9, 3.781960509141514*^9, 
   3.7833484391219378`*^9, 3.7840500312745237`*^9, 3.785570191793244*^9, 
   3.7952685753436327`*^9, 3.80000951794683*^9, 3.80026824682135*^9, 
   3.814011276950014*^9, 3.814084729882031*^9, 3.814196553686109*^9},
 CellLabel->
  "Out[1393]=",ExpressionUUID->"5ac98a83-df50-4c17-bcb7-d7ab6108c8dc"],

Cell[BoxData[
 RowBox[{
  RowBox[{"3", "+", 
   RowBox[{"h", " ", "sf"}]}], "<", 
  RowBox[{"4", " ", "sf"}]}]], "Output",
 CellChangeTimes->{
  3.781466207213058*^9, 3.7814662447499647`*^9, {3.78151331072229*^9, 
   3.781513335444324*^9}, 3.781513400170166*^9, 3.781960509141514*^9, 
   3.7833484391219378`*^9, 3.7840500312745237`*^9, 3.785570191793244*^9, 
   3.7952685753436327`*^9, 3.80000951794683*^9, 3.80026824682135*^9, 
   3.814011276950014*^9, 3.814084729882031*^9, 3.814196553738007*^9},
 CellLabel->
  "Out[1394]=",ExpressionUUID->"43a08ec8-b4df-4dfb-9af8-8c150f512416"],

Cell[BoxData["False"], "Output",
 CellChangeTimes->{
  3.781466207213058*^9, 3.7814662447499647`*^9, {3.78151331072229*^9, 
   3.781513335444324*^9}, 3.781513400170166*^9, 3.781960509141514*^9, 
   3.7833484391219378`*^9, 3.7840500312745237`*^9, 3.785570191793244*^9, 
   3.7952685753436327`*^9, 3.80000951794683*^9, 3.80026824682135*^9, 
   3.814011276950014*^9, 3.814084729882031*^9, 3.8141965537391*^9},
 CellLabel->
  "Out[1395]=",ExpressionUUID->"07cbc167-45e6-47ce-9e30-dd5acfcd19e7"]
}, Open  ]],

Cell["\<\
The second inequality implies the first again. r3 only applies if \[Delta] or \
c is less than 1. Otherwise, it is covered by r1. \
\>", "Text",
 CellChangeTimes->{{3.781440514566372*^9, 3.781440560329487*^9}, {
  3.781513468075613*^9, 3.78151364185474*^9}, {3.7815137448951883`*^9, 
  3.781513750665613*^9}, {3.781515522495345*^9, 
  3.78151552384299*^9}},ExpressionUUID->"21c5f077-202a-4e67-be8c-\
e0c1e0eb41bc"],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"Reduce", "[", 
  RowBox[{"{", 
   RowBox[{
    RowBox[{"!", "r1"}], ",", "r3", ",", 
    RowBox[{
     RowBox[{"1", "==", " ", "\[Delta]"}], "||", 
     RowBox[{"1", "\[Equal]", "c"}]}], ",", 
    RowBox[{"0", "<", "\[Delta]", "\[LessEqual]", "1"}], ",", 
    RowBox[{"0", "<", "c", "\[LessEqual]", "1"}], ",", 
    RowBox[{"1", ">", "h", ">", "0"}], ",", 
    RowBox[{"1", ">", "sm", ">", "0"}], ",", 
    RowBox[{"0", "<", "sf", "<", "1"}]}], "}"}], "]"}]], "Input",
 CellChangeTimes->{{3.781438491275654*^9, 3.7814385705794687`*^9}, {
   3.781438609675919*^9, 3.781438609925259*^9}, {3.781439361820053*^9, 
   3.781439383480029*^9}, 3.781513324912135*^9, {3.781513413430835*^9, 
   3.7815134539688053`*^9}},
 CellLabel->
  "In[1396]:=",ExpressionUUID->"966610af-88a3-49f2-8723-ecdc07880cce"],

Cell[BoxData["False"], "Output",
 CellChangeTimes->{
  3.7814385244255323`*^9, {3.78143855490099*^9, 3.781438571329862*^9}, 
   3.781438610236524*^9, {3.781439344792201*^9, 3.781439389623969*^9}, 
   3.7814394716890707`*^9, 3.781440562046509*^9, 3.781457323045916*^9, {
   3.7815134086526213`*^9, 3.781513419403633*^9}, 3.78151345446651*^9, 
   3.7833484401216583`*^9, 3.784050032554224*^9, 3.785570192992031*^9, 
   3.795268576496269*^9, 3.800009519129652*^9, 3.8002682477844467`*^9, 
   3.814011286219734*^9, 3.814084731795144*^9, 3.81419655385225*^9},
 CellLabel->
  "Out[1396]=",ExpressionUUID->"813b299a-4a5b-48a8-8003-d91dc644c0ba"]
}, Open  ]],

Cell[TextData[{
 "If \[Delta]<1 or c<1, r3 shows that the drive-fixed equilibrium is always \
unstable when heterozygous females have at least 4 times higher survival \
probability than homozygous drive females: re-writing r3 as  ",
 Cell[BoxData[
  RowBox[{
   RowBox[{"4", 
    RowBox[{"(", 
     RowBox[{"1", "-", "sf"}], ")"}]}], "<", 
   RowBox[{"1", "-", 
    RowBox[{"h", " ", "sf"}]}]}]],
  CellChangeTimes->{{3.781513605956525*^9, 3.781513625762225*^9}},
  ExpressionUUID->"70206ea0-df6a-4289-ac86-35b1df9dacaa"],
 ". Even if r3 is not met, the drive-fixed equilibrium can still be unstable \
when the r1 condition is met. "
}], "Text",
 CellChangeTimes->{{3.781513755067279*^9, 3.781513782566339*^9}, {
  3.781513833350754*^9, 
  3.7815138689367247`*^9}},ExpressionUUID->"f34cac6b-8b66-4371-8621-\
80a37628061d"],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"unstabConditions12", " ", "=", " ", 
  RowBox[{"r1", "||", "r3"}]}]], "Input",
 CellChangeTimes->{{3.7818731353415947`*^9, 3.78187317241343*^9}, {
  3.781873218740693*^9, 3.781873219335464*^9}},
 CellLabel->
  "In[1397]:=",ExpressionUUID->"5e4271bd-a10a-4cce-9caf-6fc1a0aee066"],

Cell[BoxData[
 RowBox[{
  RowBox[{
   RowBox[{"1", "+", 
    FractionBox[
     RowBox[{"1", "-", 
      RowBox[{"h", " ", "sf"}]}], 
     RowBox[{"2", " ", 
      RowBox[{"(", 
       RowBox[{
        RowBox[{"-", "1"}], "+", "sf"}], ")"}]}]], "+", 
    FractionBox[
     RowBox[{
      RowBox[{"-", "1"}], "+", 
      RowBox[{"h", " ", "sf"}]}], 
     RowBox[{
      RowBox[{"(", 
       RowBox[{"1", "+", "c"}], ")"}], " ", 
      RowBox[{"(", 
       RowBox[{
        RowBox[{"-", "1"}], "+", "sf"}], ")"}], " ", 
      RowBox[{"(", 
       RowBox[{
        RowBox[{"-", "1"}], "+", "sm"}], ")"}], " ", 
      RowBox[{"(", 
       RowBox[{"1", "+", "\[Delta]"}], ")"}]}]]}], "<", "0"}], "||", 
  RowBox[{
   RowBox[{"3", "+", 
    RowBox[{"h", " ", "sf"}]}], "<", 
   RowBox[{"4", " ", "sf"}]}]}]], "Output",
 CellChangeTimes->{{3.781873149935636*^9, 3.781873172799984*^9}, 
   3.781873220134734*^9, 3.7819605344534063`*^9, 3.783348440138775*^9, 
   3.784050032765121*^9, 3.785570193196538*^9, 3.795268576688867*^9, 
   3.800009519338931*^9, 3.800268247829068*^9, 3.814011297261401*^9, 
   3.814084732717431*^9, 3.814196553859428*^9},
 CellLabel->
  "Out[1397]=",ExpressionUUID->"6369344b-a73e-4cb8-b820-bff32b89eecd"]
}, Open  ]]
}, Closed]]
}, Closed]]
}, Closed]],

Cell[CellGroupData[{

Cell[TextData[{
 "Allele Frequency and Population Size Dynamics - females mate with many \
males (",
 Cell[BoxData[
  FormBox[
   RowBox[{
    SubscriptBox["\[Lambda]", "f"], "\[Rule]", "\[Infinity]"}], 
   TraditionalForm]],
  FormatType->"TraditionalForm",ExpressionUUID->
  "615ed352-eaec-4035-93d6-8c8b7910067c"],
 ")"
}], "Section",
 CellChangeTimes->{{3.7820456022076*^9, 3.782045608920878*^9}, {
  3.7833522380024137`*^9, 3.783352238830532*^9}, {3.814018635305141*^9, 
  3.814018678482142*^9}},ExpressionUUID->"ae63bde1-64a0-45d8-a5df-\
efb083e37f75"],

Cell[CellGroupData[{

Cell["Recursion equations", "Subsection",
 CellChangeTimes->{{3.782045639539033*^9, 
  3.7820456446347*^9}},ExpressionUUID->"857e1452-59f6-4db3-9370-0c89dc3e3a7b"],

Cell["\<\
Here we assume that each female mates many times. Therefore, females \
effectively sample alleles from the total sperm produced by the male \
population (S). In the following recursion/difference equations, 
Xm are sperm with the wildtype-X allele
Xdm are sperm with the drive-X allele
Ym are sperm with the Y allele
Xf are eggs with the wildtype-X allele
Xdf are eggs with the drive-X allele
These gametes then be combined to produce the juveniles of the next \
generation. \
\>", "Text",
 CellChangeTimes->{{3.8140848260615377`*^9, 3.814084835013913*^9}, {
  3.814091360457943*^9, 3.814091552637352*^9}, {3.8140916312181187`*^9, 
  3.814091654156864*^9}},ExpressionUUID->"f5cc5e9e-2d93-4dde-82cb-\
9b3a43b74208"],

Cell[BoxData[{
 RowBox[{
  RowBox[{"S", " ", "=", " ", 
   RowBox[{"Dm", " ", "+", " ", 
    RowBox[{"c", " ", "dm"}]}]}], ";"}], "\[IndentingNewLine]", 
 RowBox[{
  RowBox[{"Xm", " ", "=", " ", 
   FractionBox["Dm", 
    RowBox[{"2", "S"}]]}], ";"}], "\[IndentingNewLine]", 
 RowBox[{
  RowBox[{"Ym", " ", "=", " ", 
   RowBox[{
    RowBox[{"(", 
     RowBox[{
      RowBox[{"Dm", "/", "2"}], " ", "+", " ", 
      RowBox[{"dm", 
       FractionBox[
        RowBox[{"c", 
         RowBox[{"(", 
          RowBox[{"1", "-", "\[Delta]"}], ")"}]}], "2"]}]}], ")"}], "/", 
    "S"}]}], ";"}], "\[IndentingNewLine]", 
 RowBox[{
  RowBox[{"Xdm", " ", "=", " ", 
   RowBox[{"c", 
    FractionBox[
     RowBox[{"(", 
      RowBox[{"1", "+", "\[Delta]"}], ")"}], "2"], 
    RowBox[{"dm", "/", "S"}]}]}], ";"}], "\[IndentingNewLine]", 
 RowBox[{
  RowBox[{"Xf", " ", "=", "  ", 
   RowBox[{"DDf", " ", "+", " ", 
    RowBox[{"Ddf", "/", "2"}]}]}], ";"}], "\[IndentingNewLine]", 
 RowBox[{
  RowBox[{
   RowBox[{"Xdf", " ", "=", " ", 
    RowBox[{
     RowBox[{"Ddf", "/", "2"}], " ", "+", " ", "ddf"}]}], ";"}], 
  "\[IndentingNewLine]"}], "\[IndentingNewLine]", 
 RowBox[{
  RowBox[{"DDfjuvenilesi", " ", "=", 
   RowBox[{"b", 
    RowBox[{"(", 
     RowBox[{"1", "-", 
      RowBox[{"\[Alpha]", " ", "totAdults"}]}], ")"}], " ", "Xm", " ", 
    "Xf"}]}], ";"}], "\[IndentingNewLine]", 
 RowBox[{
  RowBox[{"Ddfjuvenilesi", " ", "=", " ", 
   RowBox[{"b", 
    RowBox[{"(", 
     RowBox[{"1", "-", 
      RowBox[{"\[Alpha]", " ", "totAdults"}]}], ")"}], 
    RowBox[{"(", 
     RowBox[{
      RowBox[{"Xdm", " ", "Xf"}], " ", "+", " ", 
      RowBox[{"Xm", " ", "Xdf"}]}], ")"}]}]}], ";"}], "\[IndentingNewLine]", 
 RowBox[{
  RowBox[{"ddfjuvenilesi", " ", "=", " ", 
   RowBox[{"b", 
    RowBox[{"(", 
     RowBox[{"1", "-", 
      RowBox[{"\[Alpha]", " ", "totAdults"}]}], ")"}], "Xdm", " ", "Xdf"}]}], 
  ";"}], "\[IndentingNewLine]", 
 RowBox[{
  RowBox[{"Dmjuvenilesi", " ", "=", " ", 
   RowBox[{"b", 
    RowBox[{"(", 
     RowBox[{"1", "-", 
      RowBox[{"\[Alpha]", " ", "totAdults"}]}], ")"}], "Ym", " ", "Xf"}]}], 
  ";"}], "\[IndentingNewLine]", 
 RowBox[{
  RowBox[{"dmjuvenilesi", " ", "=", " ", 
   RowBox[{"b", 
    RowBox[{"(", 
     RowBox[{"1", "-", 
      RowBox[{"\[Alpha]", " ", "totAdults"}]}], ")"}], "Ym", " ", "Xdf"}]}], 
  ";"}]}], "Input",
 CellChangeTimes->{{3.782045656931858*^9, 3.7820456690248833`*^9}, {
  3.782045700537414*^9, 3.7820457212826433`*^9}, {3.782045795173643*^9, 
  3.782045829421994*^9}, {3.814008325163143*^9, 3.814008332363858*^9}, {
  3.814078539185297*^9, 3.8140785447397547`*^9}, {3.814091564454846*^9, 
  3.814091625188179*^9}, {3.814091660040991*^9, 3.814091662469996*^9}},
 CellLabel->
  "In[1398]:=",ExpressionUUID->"212cac65-fd16-437c-b577-9747db50f982"],

Cell["\<\
Before the next generation of adults, viability selection acts according to \
genotype. \
\>", "Text",
 CellChangeTimes->{
  3.814091621153105*^9},ExpressionUUID->"2fc5c8bb-b413-4c0a-80f7-\
0ac214eb88a3"],

Cell[BoxData[{
 RowBox[{
  RowBox[{"DDfprimei", "=", 
   RowBox[{"DDfjuvenilesi", "*", "1"}]}], ";"}], "\[IndentingNewLine]", 
 RowBox[{
  RowBox[{"Ddfprimei", "=", 
   RowBox[{"Ddfjuvenilesi", "*", 
    RowBox[{"(", 
     RowBox[{"1", " ", "-", " ", 
      RowBox[{"h", "*", "sf"}]}], ")"}]}]}], ";"}], "\[IndentingNewLine]", 
 RowBox[{
  RowBox[{"ddfprimei", "=", 
   RowBox[{"ddfjuvenilesi", "*", 
    RowBox[{"(", 
     RowBox[{"1", " ", "-", " ", "sf"}], ")"}]}]}], 
  ";"}], "\[IndentingNewLine]", 
 RowBox[{
  RowBox[{"Dmprimei", "=", 
   RowBox[{"Dmjuvenilesi", "*", "1"}]}], ";"}], "\[IndentingNewLine]", 
 RowBox[{
  RowBox[{"dmprimei", "=", 
   RowBox[{"dmjuvenilesi", "*", 
    RowBox[{"(", 
     RowBox[{"1", " ", "-", " ", "sm"}], ")"}]}]}], ";"}]}], "Input",
 CellLabel->
  "In[1409]:=",ExpressionUUID->"fbf2e3c1-0190-4b27-9424-bf3a77c868ba"]
}, Closed]],

Cell[CellGroupData[{

Cell["Equilibrium allele frequency and population size", "Subsection",
 CellChangeTimes->{{3.8000146920049686`*^9, 3.800014708486021*^9}, 
   3.8140171871723146`*^9},ExpressionUUID->"b849157a-262e-4cbf-9118-\
8d1e8c4b3334"],

Cell["\<\
We first look at only the difference in allele frequency and inbreeding \
coefficient to calculate the polymorphic allele frequency equilibrium (when \
changes in allele frequency and inbreeding coefficient are 0)\
\>", "Text",
 CellChangeTimes->{
  3.814091727710836*^9},ExpressionUUID->"9a26f5fd-a2fa-42e2-abf3-\
46cb5b5baeb8"],

Cell[CellGroupData[{

Cell[BoxData[{
 RowBox[{
  RowBox[{"solvemei", "=", 
   RowBox[{"Simplify", "[", 
    RowBox[{
     RowBox[{"{", 
      RowBox[{
       RowBox[{
        RowBox[{"(", 
         FractionBox["dmprimei", 
          RowBox[{"Dmprimei", "+", "dmprimei"}]], ")"}], "-", 
        FractionBox["dm", 
         RowBox[{"dm", "+", "Dm"}]]}], ",", 
       RowBox[{
        RowBox[{"(", 
         FractionBox[
          RowBox[{
           RowBox[{"2", " ", "ddfprimei"}], "+", "Ddfprimei"}], 
          RowBox[{"2", " ", 
           RowBox[{"(", 
            RowBox[{"ddfprimei", "+", "Ddfprimei", "+", "DDfprimei"}], 
            ")"}]}]], ")"}], "-", 
        FractionBox[
         RowBox[{
          RowBox[{"2", " ", "ddf"}], "+", "Ddf"}], 
         RowBox[{"2", " ", 
          RowBox[{"(", 
           RowBox[{"ddf", "+", "Ddf", "+", "DDf"}], ")"}]}]]}], ",", 
       RowBox[{
        RowBox[{"(", 
         FractionBox[
          RowBox[{
           RowBox[{"4", " ", "ddfprimei", " ", "DDfprimei"}], " ", "-", 
           SuperscriptBox["Ddfprimei", "2"]}], 
          RowBox[{
           RowBox[{"(", 
            RowBox[{
             RowBox[{"2", " ", "ddfprimei"}], "+", "Ddfprimei"}], ")"}], " ", 
           RowBox[{"(", 
            RowBox[{"Ddfprimei", "+", 
             RowBox[{"2", " ", "DDfprimei"}]}], ")"}]}]], ")"}], "-", 
        FractionBox[
         RowBox[{
          RowBox[{"4", " ", "ddf", " ", "DDf"}], "-", 
          SuperscriptBox["Ddf", "2"]}], 
         RowBox[{
          RowBox[{"(", 
           RowBox[{
            RowBox[{"2", " ", "ddf"}], "+", "Ddf"}], ")"}], " ", 
          RowBox[{"(", 
           RowBox[{"Ddf", "+", 
            RowBox[{"2", " ", "DDf"}]}], ")"}]}]]}]}], "}"}], "/.", "subeq"}],
     "]"}]}], ";"}], "\[IndentingNewLine]", 
 RowBox[{"equilibriumi", "=", 
  RowBox[{
   RowBox[{"Solve", "[", 
    RowBox[{
     RowBox[{
      RowBox[{"solvemei", "[", 
       RowBox[{"[", "1", "]"}], "]"}], "\[Equal]", 
      RowBox[{"{", 
       RowBox[{"0", ",", "0", ",", "0"}], "}"}]}], ",", 
     RowBox[{"{", 
      RowBox[{"pm", ",", "pf", ",", "F"}], "}"}]}], "]"}], "//", 
   "Simplify"}]}]}], "Input",
 CellChangeTimes->{{3.782045867513173*^9, 3.782045943984172*^9}, 
   3.782046197709968*^9, {3.782046238579432*^9, 3.7820462394894753`*^9}, 
   3.8000145432139883`*^9, {3.814008540975511*^9, 3.8140085414797564`*^9}, {
   3.814011361825158*^9, 3.8140113625212107`*^9}, {3.814091678828651*^9, 
   3.814091694475749*^9}, 3.814091745498659*^9},
 CellLabel->
  "In[1414]:=",ExpressionUUID->"915927a4-d9f8-450f-832c-1069351e9a1e"],

Cell[BoxData[
 RowBox[{"{", 
  RowBox[{"{", 
   RowBox[{
    RowBox[{"pm", "\[Rule]", 
     RowBox[{"-", 
      FractionBox[
       RowBox[{
        RowBox[{"(", 
         RowBox[{
          RowBox[{"-", "1"}], "+", "sm"}], ")"}], " ", 
        RowBox[{"(", 
         RowBox[{
          RowBox[{"-", "1"}], "-", 
          RowBox[{"h", " ", "sf"}], "+", 
          RowBox[{"c", " ", 
           RowBox[{"(", 
            RowBox[{
             RowBox[{"-", "1"}], "+", 
             RowBox[{"h", " ", "sf"}]}], ")"}], " ", 
           RowBox[{"(", 
            RowBox[{
             RowBox[{"-", "1"}], "+", "sm"}], ")"}], " ", 
           RowBox[{"(", 
            RowBox[{"1", "+", "\[Delta]"}], ")"}]}]}], ")"}]}], 
       RowBox[{
        RowBox[{"h", " ", "sf", " ", 
         RowBox[{"(", 
          RowBox[{
           RowBox[{"-", "2"}], "+", "sm"}], ")"}]}], "+", "sm", "-", 
        RowBox[{"c", " ", 
         RowBox[{"(", 
          RowBox[{
           RowBox[{"sf", " ", 
            RowBox[{"(", 
             RowBox[{"2", "+", 
              RowBox[{"h", " ", 
               RowBox[{"(", 
                RowBox[{
                 RowBox[{"-", "2"}], "+", "sm"}], ")"}]}]}], ")"}]}], "-", 
           "sm"}], ")"}], " ", 
         RowBox[{"(", 
          RowBox[{
           RowBox[{"-", "1"}], "+", "sm"}], ")"}], " ", 
         RowBox[{"(", 
          RowBox[{"1", "+", "\[Delta]"}], ")"}]}]}]]}]}], ",", 
    RowBox[{"pf", "\[Rule]", 
     FractionBox[
      RowBox[{
       RowBox[{"-", "1"}], "-", 
       RowBox[{"h", " ", "sf"}], "+", 
       RowBox[{"c", " ", 
        RowBox[{"(", 
         RowBox[{
          RowBox[{"-", "1"}], "+", 
          RowBox[{"h", " ", "sf"}]}], ")"}], " ", 
        RowBox[{"(", 
         RowBox[{
          RowBox[{"-", "1"}], "+", "sm"}], ")"}], " ", 
        RowBox[{"(", 
         RowBox[{"1", "+", "\[Delta]"}], ")"}]}]}], 
      RowBox[{"2", " ", "sf", " ", 
       RowBox[{"(", 
        RowBox[{
         RowBox[{"-", "h"}], "+", 
         RowBox[{"c", " ", 
          RowBox[{"(", 
           RowBox[{
            RowBox[{"-", "1"}], "+", "h"}], ")"}], " ", 
          RowBox[{"(", 
           RowBox[{
            RowBox[{"-", "1"}], "+", "sm"}], ")"}], " ", 
          RowBox[{"(", 
           RowBox[{"1", "+", "\[Delta]"}], ")"}]}]}], ")"}]}]]}], ",", 
    RowBox[{"F", "\[Rule]", 
     FractionBox[
      RowBox[{"1", "-", 
       RowBox[{
        SuperscriptBox["h", "2"], " ", 
        SuperscriptBox["sf", "2"]}], "+", 
       RowBox[{"2", " ", "c", " ", 
        RowBox[{"(", 
         RowBox[{"1", "-", "sf", "+", 
          RowBox[{
           RowBox[{"(", 
            RowBox[{
             RowBox[{"-", "1"}], "+", "h"}], ")"}], " ", "h", " ", 
           SuperscriptBox["sf", "2"]}]}], ")"}], " ", 
        RowBox[{"(", 
         RowBox[{
          RowBox[{"-", "1"}], "+", "sm"}], ")"}], " ", 
        RowBox[{"(", 
         RowBox[{"1", "+", "\[Delta]"}], ")"}]}], "-", 
       RowBox[{
        SuperscriptBox["c", "2"], " ", 
        RowBox[{"(", 
         RowBox[{
          RowBox[{"-", "1"}], "+", 
          RowBox[{"2", " ", "sf"}], "+", 
          RowBox[{
           RowBox[{"(", 
            RowBox[{
             RowBox[{"-", "2"}], "+", "h"}], ")"}], " ", "h", " ", 
           SuperscriptBox["sf", "2"]}]}], ")"}], " ", 
        SuperscriptBox[
         RowBox[{"(", 
          RowBox[{
           RowBox[{"-", "1"}], "+", "sm"}], ")"}], "2"], " ", 
        SuperscriptBox[
         RowBox[{"(", 
          RowBox[{"1", "+", "\[Delta]"}], ")"}], "2"]}]}], 
      RowBox[{
       SuperscriptBox[
        RowBox[{"(", 
         RowBox[{
          RowBox[{"-", "1"}], "+", 
          RowBox[{"h", " ", "sf"}]}], ")"}], "2"], "-", 
       RowBox[{"2", " ", "c", " ", 
        RowBox[{"(", 
         RowBox[{
          RowBox[{"-", "1"}], "-", 
          RowBox[{"2", " ", 
           RowBox[{"(", 
            RowBox[{
             RowBox[{"-", "1"}], "+", "h"}], ")"}], " ", "sf"}], "+", 
          RowBox[{
           SuperscriptBox["h", "2"], " ", 
           SuperscriptBox["sf", "2"]}]}], ")"}], " ", 
        RowBox[{"(", 
         RowBox[{
          RowBox[{"-", "1"}], "+", "sm"}], ")"}], " ", 
        RowBox[{"(", 
         RowBox[{"1", "+", "\[Delta]"}], ")"}]}], "+", 
       RowBox[{
        SuperscriptBox["c", "2"], " ", 
        SuperscriptBox[
         RowBox[{"(", 
          RowBox[{
           RowBox[{"-", "1"}], "+", 
           RowBox[{"h", " ", "sf"}]}], ")"}], "2"], " ", 
        SuperscriptBox[
         RowBox[{"(", 
          RowBox[{
           RowBox[{"-", "1"}], "+", "sm"}], ")"}], "2"], " ", 
        SuperscriptBox[
         RowBox[{"(", 
          RowBox[{"1", "+", "\[Delta]"}], ")"}], "2"]}]}]]}]}], "}"}], 
  "}"}]], "Output",
 CellChangeTimes->{3.814091758233905*^9, 3.814196563696995*^9},
 CellLabel->
  "Out[1415]=",ExpressionUUID->"b58b7436-e2d4-496c-81b0-c709638835e3"]
}, Open  ]],

Cell["The full system is described by", "Text",
 CellChangeTimes->{{3.8140917527276917`*^9, 
  3.8140917697854643`*^9}},ExpressionUUID->"6057c0d4-b727-46fe-9189-\
f0a38bbd200e"],

Cell[BoxData[
 RowBox[{
  RowBox[{"solvemeFulli", "=", 
   RowBox[{
    RowBox[{
     RowBox[{"{", 
      RowBox[{
       RowBox[{"DDfprimei", "-", "DDf"}], ",", 
       RowBox[{"Ddfprimei", "-", "Ddf"}], ",", 
       RowBox[{"ddfprimei", "-", "ddf"}], ",", " ", 
       RowBox[{"Dmprimei", "-", "Dm"}], ",", " ", 
       RowBox[{"dmprimei", "-", "dm"}]}], "}"}], "/.", 
     RowBox[{"subeq", "[", 
      RowBox[{"[", "1", "]"}], "]"}]}], "//", "Simplify"}]}], ";"}]], "Input",
 CellChangeTimes->{{3.782046292949939*^9, 3.78204630300189*^9}, {
  3.782046378445221*^9, 3.782046379380213*^9}, {3.814008565912324*^9, 
  3.814008566316824*^9}},
 CellLabel->
  "In[1416]:=",ExpressionUUID->"af9006e4-f765-4961-b4f8-c03ebc39bb15"],

Cell["\<\
We can then solve for the population sizes. First we find population sizes \
when drive alleles are absent (pm=pf=0) or fixed (pm=pf=1):\
\>", "Text",
 CellChangeTimes->{
  3.8140917832958117`*^9},ExpressionUUID->"fb77491e-bd41-415d-bd14-\
88d89493f794"],

Cell[CellGroupData[{

Cell[BoxData[{
 RowBox[{
  RowBox[{"noDrivePopSizeSoli", "=", 
   RowBox[{"Solve", "[", 
    RowBox[{
     RowBox[{
      RowBox[{"(", 
       RowBox[{
        RowBox[{"solvemeFulli", "/.", 
         RowBox[{"pm", "\[Rule]", "0"}]}], "/.", 
        RowBox[{"pf", "\[Rule]", "0"}]}], ")"}], "\[Equal]", 
      RowBox[{"{", 
       RowBox[{"0", ",", "0", ",", "0", ",", "0", ",", "0"}], "}"}]}], ",", 
     RowBox[{"{", 
      RowBox[{"maleNum", ",", "femaleNum"}], "}"}]}], "]"}]}], 
  ";"}], "\[IndentingNewLine]", 
 RowBox[{
  RowBox[{"fixedDrivePopSizeSoli", "=", 
   RowBox[{"Solve", "[", 
    RowBox[{
     RowBox[{
      RowBox[{"(", 
       RowBox[{
        RowBox[{"solvemeFulli", "/.", 
         RowBox[{"pm", "\[Rule]", "1"}]}], "/.", 
        RowBox[{"pf", "\[Rule]", "1"}]}], ")"}], "\[Equal]", 
      RowBox[{"{", 
       RowBox[{"0", ",", "0", ",", "0", ",", "0", ",", "0"}], "}"}]}], ",", 
     RowBox[{"{", 
      RowBox[{"maleNum", ",", "femaleNum"}], "}"}]}], "]"}]}], 
  ";"}], "\[IndentingNewLine]", 
 RowBox[{"NnoDrivei", "=", 
  RowBox[{
   RowBox[{
    RowBox[{"(", 
     RowBox[{"maleNum", " ", "+", " ", "femaleNum"}], ")"}], "/.", 
    RowBox[{"noDrivePopSizeSoli", "[", 
     RowBox[{"[", "2", "]"}], "]"}]}], "//", 
   "FullSimplify"}]}], "\[IndentingNewLine]", 
 RowBox[{"Nfixedi", "=", " ", 
  RowBox[{
   RowBox[{
    RowBox[{"(", 
     RowBox[{"maleNum", " ", "+", " ", "femaleNum"}], ")"}], "/.", 
    RowBox[{"fixedDrivePopSizeSoli", "[", 
     RowBox[{"[", "2", "]"}], "]"}]}], "//", "FullSimplify"}]}]}], "Input",
 CellChangeTimes->{{3.7820463137454157`*^9, 3.78204631865808*^9}, {
  3.814008547780476*^9, 3.814008548330002*^9}, {3.81401716236158*^9, 
  3.814017174104857*^9}, {3.814091790190922*^9, 3.814091819803329*^9}, {
  3.814091861811695*^9, 3.8140918829769297`*^9}},
 CellLabel->
  "In[1417]:=",ExpressionUUID->"ff3a0256-827a-4e99-9958-7f4408ab7370"],

Cell[BoxData[
 FractionBox[
  RowBox[{
   RowBox[{"-", "2"}], "+", "b"}], 
  RowBox[{"b", " ", "\[Alpha]"}]]], "Output",
 CellChangeTimes->{
  3.782046319625399*^9, 3.782046382284474*^9, 3.783352285597454*^9, 
   3.78557501139997*^9, 3.800014573508539*^9, 3.800269109738594*^9, 
   3.8140113874476233`*^9, {3.814017149234234*^9, 3.814017174453267*^9}, 
   3.814091699219639*^9, {3.8140918096062927`*^9, 3.814091820121717*^9}, {
   3.814091867013352*^9, 3.814091883835573*^9}, 3.814196563982563*^9},
 CellLabel->
  "Out[1419]=",ExpressionUUID->"2fe10b64-a2b9-402d-add5-ed7158ffe041"],

Cell[BoxData[
 FractionBox[
  RowBox[{"1", "+", 
   FractionBox["2", 
    RowBox[{"b", " ", 
     RowBox[{"(", 
      RowBox[{
       RowBox[{"-", "1"}], "+", "sf"}], ")"}], " ", 
     RowBox[{"(", 
      RowBox[{"1", "+", "\[Delta]"}], ")"}]}]]}], "\[Alpha]"]], "Output",
 CellChangeTimes->{
  3.782046319625399*^9, 3.782046382284474*^9, 3.783352285597454*^9, 
   3.78557501139997*^9, 3.800014573508539*^9, 3.800269109738594*^9, 
   3.8140113874476233`*^9, {3.814017149234234*^9, 3.814017174453267*^9}, 
   3.814091699219639*^9, {3.8140918096062927`*^9, 3.814091820121717*^9}, {
   3.814091867013352*^9, 3.814091883835573*^9}, 3.814196564028223*^9},
 CellLabel->
  "Out[1420]=",ExpressionUUID->"d9334b4c-9932-48e6-b6be-7e1617afdf26"]
}, Open  ]],

Cell["\<\
Now for the case where a polymorphic equilibrium has been reached\
\>", "Text",
 CellChangeTimes->{{3.800014722019763*^9, 3.800014729398663*^9}, 
   3.814091917394124*^9},ExpressionUUID->"7dbef16d-c4ba-4902-8cf2-\
180cdecaccb4"],

Cell[BoxData[
 RowBox[{
  RowBox[{"equilPopSizeSoli", "=", 
   RowBox[{"Simplify", "[", 
    RowBox[{"Solve", "[", 
     RowBox[{
      RowBox[{
       RowBox[{"Flatten", "[", 
        RowBox[{"Simplify", "[", 
         RowBox[{"(", 
          RowBox[{
           RowBox[{"solvemeFulli", "/.", "equilibriumi"}], "/.", 
           RowBox[{"sm", "\[Rule]", "0"}]}], ")"}], "]"}], "]"}], "\[Equal]", 
       RowBox[{"{", 
        RowBox[{"0", ",", "0", ",", "0", ",", "0", ",", "0"}], "}"}]}], ",", 
      " ", 
      RowBox[{"{", 
       RowBox[{"femaleNum", ",", "maleNum"}], "}"}]}], "]"}], "]"}]}], 
  ";"}]], "Input",
 CellChangeTimes->{{3.782046342000181*^9, 3.782046348335209*^9}, 
   3.7820464885591717`*^9, {3.782046557165715*^9, 3.7820465585190477`*^9}, 
   3.800014736838333*^9},
 CellLabel->
  "In[1421]:=",ExpressionUUID->"5d83063f-2a14-4474-9a7d-a9aab5e80959"],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"Neqi", " ", "=", 
  RowBox[{
   RowBox[{
    RowBox[{"(", 
     RowBox[{"femaleNum", "+", "maleNum"}], ")"}], "/.", 
    RowBox[{"equilPopSizeSol", "[", 
     RowBox[{"[", "2", "]"}], "]"}]}], "//", "Simplify"}]}]], "Input",
 CellChangeTimes->{
  3.7820465950934258`*^9, 3.782046796622093*^9, {3.814091931638362*^9, 
   3.814091938181066*^9}},
 CellLabel->
  "In[1422]:=",ExpressionUUID->"7eb4b35b-93f8-4eee-ae84-40debbf22b38"],

Cell[BoxData[
 RowBox[{
  RowBox[{"(", 
   RowBox[{
    RowBox[{"b", " ", 
     SuperscriptBox["h", "2"], " ", 
     SuperscriptBox["sf", "2"], " ", 
     SuperscriptBox[
      RowBox[{"(", 
       RowBox[{
        RowBox[{"-", "2"}], "+", "sm", "-", "\[Delta]", "+", 
        RowBox[{"sm", " ", "\[Delta]"}]}], ")"}], "2"]}], "+", 
    RowBox[{
     RowBox[{"(", 
      RowBox[{"sm", "-", "\[Delta]", "+", 
       RowBox[{"sm", " ", "\[Delta]"}]}], ")"}], " ", 
     RowBox[{"(", 
      RowBox[{
       RowBox[{
        RowBox[{"-", "b"}], " ", "\[Delta]"}], "+", 
       RowBox[{"sm", " ", 
        RowBox[{"(", 
         RowBox[{
          RowBox[{"-", "4"}], "+", "b", "+", 
          RowBox[{"b", " ", "\[Delta]"}]}], ")"}]}]}], ")"}]}], "-", 
    RowBox[{"2", " ", "sf", " ", 
     RowBox[{"(", 
      RowBox[{
       RowBox[{
        RowBox[{"-", "4"}], " ", 
        RowBox[{"(", 
         RowBox[{
          RowBox[{"-", "1"}], "+", "sm"}], ")"}], " ", 
        RowBox[{"(", 
         RowBox[{"1", "+", "\[Delta]"}], ")"}]}], "+", 
       RowBox[{"2", " ", "b", " ", 
        RowBox[{"(", 
         RowBox[{
          RowBox[{"-", "1"}], "+", "sm"}], ")"}], " ", 
        RowBox[{"(", 
         RowBox[{"1", "+", "\[Delta]"}], ")"}]}], "-", 
       RowBox[{"2", " ", "h", " ", 
        RowBox[{"(", 
         RowBox[{
          RowBox[{"-", "2"}], "+", "sm"}], ")"}], " ", 
        RowBox[{"(", 
         RowBox[{
          RowBox[{"-", "2"}], "+", "sm", "-", "\[Delta]", "+", 
          RowBox[{"sm", " ", "\[Delta]"}]}], ")"}]}], "+", 
       RowBox[{"b", " ", "h", " ", 
        SuperscriptBox[
         RowBox[{"(", 
          RowBox[{
           RowBox[{"-", "2"}], "+", "sm", "-", "\[Delta]", "+", 
           RowBox[{"sm", " ", "\[Delta]"}]}], ")"}], "2"]}]}], ")"}]}]}], 
   ")"}], "/", 
  RowBox[{"(", 
   RowBox[{"b", " ", "\[Alpha]", " ", 
    RowBox[{"(", 
     RowBox[{
      RowBox[{
       SuperscriptBox["h", "2"], " ", 
       SuperscriptBox["sf", "2"], " ", 
       SuperscriptBox[
        RowBox[{"(", 
         RowBox[{
          RowBox[{"-", "2"}], "+", "sm", "-", "\[Delta]", "+", 
          RowBox[{"sm", " ", "\[Delta]"}]}], ")"}], "2"]}], "+", 
      SuperscriptBox[
       RowBox[{"(", 
        RowBox[{"sm", "-", "\[Delta]", "+", 
         RowBox[{"sm", " ", "\[Delta]"}]}], ")"}], "2"], "-", 
      RowBox[{"2", " ", "sf", " ", 
       RowBox[{"(", 
        RowBox[{
         RowBox[{"2", " ", 
          RowBox[{"(", 
           RowBox[{
            RowBox[{"-", "1"}], "+", "sm"}], ")"}], " ", 
          RowBox[{"(", 
           RowBox[{"1", "+", "\[Delta]"}], ")"}]}], "+", 
         RowBox[{"h", " ", 
          SuperscriptBox[
           RowBox[{"(", 
            RowBox[{
             RowBox[{"-", "2"}], "+", "sm", "-", "\[Delta]", "+", 
             RowBox[{"sm", " ", "\[Delta]"}]}], ")"}], "2"]}]}], ")"}]}]}], 
     ")"}]}], ")"}]}]], "Output",
 CellChangeTimes->{
  3.782046598251156*^9, 3.782046797162363*^9, 3.783352290652995*^9, 
   3.785575017309208*^9, 3.800014744545074*^9, 3.800269114660803*^9, 
   3.814011393330263*^9, 3.814017153286552*^9, 3.814091708502544*^9, {
   3.814091934038047*^9, 3.814091938425481*^9}, 3.814196568387684*^9},
 CellLabel->
  "Out[1422]=",ExpressionUUID->"ce6fc66f-12f8-4453-b78e-a878a126492c"]
}, Open  ]]
}, Closed]],

Cell[CellGroupData[{

Cell["Stability", "Subsection",
 CellChangeTimes->{{3.781465326981051*^9, 
  3.781465329029233*^9}},ExpressionUUID->"dd05fb51-e055-4cb7-bb15-\
81c0cc3c5f9d"],

Cell["\<\
There are at least two immediately apparent equilibria: when drive is absent \
from the population and when drive is fixed in the population. 

We first find the Jacobian for the system:\
\>", "Text",
 CellChangeTimes->{3.814018214040639*^9, 
  3.814092021318809*^9},ExpressionUUID->"47972617-7c6d-4b93-9067-\
80c6647915f7"],

Cell[BoxData[{
 RowBox[{
  RowBox[{"recursionsi", "=", 
   RowBox[{"{", 
    RowBox[{
    "Dmprimei", ",", " ", "dmprimei", ",", " ", "DDfprimei", ",", " ", 
     "Ddfprimei", ",", " ", "ddfprimei"}], "}"}]}], 
  ";"}], "\[IndentingNewLine]", 
 RowBox[{
  RowBox[{"jacobi", "=", 
   RowBox[{"Simplify", "[", 
    RowBox[{"Transpose", "[", 
     RowBox[{"{", 
      RowBox[{
       RowBox[{"D", "[", 
        RowBox[{"recursionsi", ",", "Dm"}], "]"}], ",", " ", 
       RowBox[{"D", "[", 
        RowBox[{"recursionsi", ",", "dm"}], "]"}], ",", " ", 
       RowBox[{"D", "[", 
        RowBox[{"recursionsi", ",", " ", "DDf"}], "]"}], ",", " ", 
       RowBox[{"D", "[", 
        RowBox[{"recursionsi", ",", " ", "Ddf"}], "]"}], ",", " ", 
       RowBox[{"D", "[", 
        RowBox[{"recursionsi", ",", " ", "ddf"}], "]"}]}], "}"}], "]"}], 
    "]"}]}], ";"}]}], "Input",
 CellChangeTimes->{{3.781465345288929*^9, 3.781465345456649*^9}, {
  3.781511024565343*^9, 3.7815110446207*^9}, {3.781511354597693*^9, 
  3.7815113827574787`*^9}, {3.8140171991715717`*^9, 3.814017218805543*^9}},
 CellLabel->
  "In[1423]:=",ExpressionUUID->"5a43fb10-1189-4adb-b947-0359d89fbcc5"],

Cell["\<\
And then we assess the conditions under which the drive-absent and \
drive-fixed equilibria are stable/instable. \
\>", "Text",
 CellChangeTimes->{{3.814018217674459*^9, 
  3.814018251389496*^9}},ExpressionUUID->"c475972b-93ec-4c87-87a7-\
4ed38c09ad07"],

Cell[CellGroupData[{

Cell["drive absent (pm=pf=0) equilibrium", "Subsubsection",
 CellChangeTimes->{{3.7814653542305307`*^9, 3.7814653590889*^9}, {
  3.814018255247793*^9, 
  3.814018264000095*^9}},ExpressionUUID->"52a97988-5a75-4509-936e-\
06464b1a879c"],

Cell["Proceed as before:", "Text",
 CellChangeTimes->{{3.7814654517132387`*^9, 
  3.781465455407995*^9}},ExpressionUUID->"6e06c72a-772e-4df2-adc5-\
6c3ec47dee3f"],

Cell[CellGroupData[{

Cell[BoxData[{
 RowBox[{
  RowBox[{"Simplify", "[", 
   RowBox[{
    RowBox[{
     RowBox[{
      RowBox[{"jacobi", "/.", 
       RowBox[{"subeq", "[", 
        RowBox[{"[", "1", "]"}], "]"}]}], "/.", 
      RowBox[{"pf", "\[Rule]", "0"}]}], "/.", 
     RowBox[{"pm", "\[Rule]", "0"}]}], "/.", 
    RowBox[{"noDrivePopSizeSoli", "[", 
     RowBox[{"[", "2", "]"}], "]"}]}], "]"}], ";"}], "\[IndentingNewLine]", 
 RowBox[{"charpolynomial0i", "=", 
  RowBox[{"Factor", "[", 
   RowBox[{"Det", "[", 
    RowBox[{"%", "-", 
     RowBox[{"\[Lambda]", "*", 
      RowBox[{"IdentityMatrix", "[", "5", "]"}]}]}], "]"}], "]"}]}]}], "Input",
 CellChangeTimes->{{3.814009388479999*^9, 3.8140094011923647`*^9}, {
   3.814009435990284*^9, 3.814009472662134*^9}, {3.814009517855748*^9, 
   3.8140095446179943`*^9}, 3.8140095798181334`*^9, {3.814009653572708*^9, 
   3.8140096651062803`*^9}, {3.814009729946822*^9, 3.8140097521423693`*^9}, {
   3.814017226433319*^9, 3.81401723481985*^9}},
 CellLabel->
  "In[1425]:=",ExpressionUUID->"05134b4a-647c-45fc-a865-ba3d57c68168"],

Cell[BoxData[
 RowBox[{
  RowBox[{"-", 
   FractionBox["1", "4"]}], " ", 
  SuperscriptBox["\[Lambda]", "2"], " ", 
  RowBox[{"(", 
   RowBox[{
    RowBox[{"-", "4"}], "+", "b", "+", 
    RowBox[{"2", " ", "\[Lambda]"}]}], ")"}], " ", 
  RowBox[{"(", 
   RowBox[{
    RowBox[{"-", "c"}], "+", 
    RowBox[{"c", " ", "h", " ", "sf"}], "+", 
    RowBox[{"c", " ", "sm"}], "-", 
    RowBox[{"c", " ", "h", " ", "sf", " ", "sm"}], "-", 
    RowBox[{"c", " ", "\[Delta]"}], "+", 
    RowBox[{"c", " ", "h", " ", "sf", " ", "\[Delta]"}], "+", 
    RowBox[{"c", " ", "sm", " ", "\[Delta]"}], "-", 
    RowBox[{"c", " ", "h", " ", "sf", " ", "sm", " ", "\[Delta]"}], "-", 
    "\[Lambda]", "+", 
    RowBox[{"h", " ", "sf", " ", "\[Lambda]"}], "+", 
    RowBox[{"2", " ", 
     SuperscriptBox["\[Lambda]", "2"]}]}], ")"}]}]], "Output",
 CellChangeTimes->{{3.814009396578875*^9, 3.814009473040482*^9}, {
   3.814009522575335*^9, 3.814009546081353*^9}, 3.814009580161955*^9, {
   3.814009653911927*^9, 3.8140096653633537`*^9}, {3.814009731293776*^9, 
   3.814009752479535*^9}, 3.814017235251996*^9, 3.814091946417801*^9, 
   3.814196568896222*^9},
 CellLabel->
  "Out[1426]=",ExpressionUUID->"64713bff-66ef-408e-ac32-075cf301a5a9"]
}, Open  ]],

Cell[TextData[{
 "The factor ",
 Cell[BoxData[
  RowBox[{"-", 
   FractionBox[
    RowBox[{
     SuperscriptBox["\[Lambda]", "2"], " ", 
     RowBox[{"(", 
      RowBox[{
       RowBox[{"-", "4"}], "+", "b", "+", 
       RowBox[{"2", " ", "\[Lambda]"}]}], ")"}], " "}], 
    RowBox[{"4", " "}]]}]],
  CellChangeTimes->{
   3.7812740033403482`*^9, 3.7814573218417273`*^9, 3.7818848609506083`*^9, 
    3.781952821819541*^9, 3.78195289751086*^9, {3.7819538958219957`*^9, 
    3.781953915448163*^9}, 3.7821306544665728`*^9, 3.782747157952016*^9, 
    3.7830788934301643`*^9, 3.785162683931014*^9, 3.785164973948011*^9, 
    3.789732648885466*^9, 3.795174657109274*^9, 3.799320669052409*^9, 
    3.799340142599029*^9, 3.799756425128694*^9, 3.800007655671279*^9, 
    3.8002681846876783`*^9, 3.813930901799224*^9, {3.8139309343336077`*^9, 
    3.8139310598724957`*^9}, 3.8140037587039423`*^9, 3.814003805149556*^9},
  ExpressionUUID->"338684d7-87a9-4b1e-b765-f148437563ed"],
 " gives two 0 eigenvalues and and eigenvalue of ",
 Cell[BoxData[
  FractionBox[
   RowBox[{"4", "-", "b"}], "2"]],
  CellChangeTimes->{3.814003986297804*^9},ExpressionUUID->
  "fff42b1f-b3f1-4c36-9bc1-952a62bba69f"],
 ", corresponding to the change in overall population size. Thus, we can see \
that the demographic dynamics are orthogonal to the evolutionary dynamics. \n\
\nHere,  we divide through by this factor to simplify the equation and \
evaluate the evolutionary dynamics. We then write the characteristic \
polynomial in the form\n",
 Cell[BoxData[
  RowBox[{
   SuperscriptBox["a\[Lambda]", "2"], "+", "b\[Lambda]", "+", "c"}]],
  CellChangeTimes->{
   3.7812740321927767`*^9, 3.781274073867482*^9, 3.7814285214279013`*^9, 
    3.781457321850852*^9, 3.78188486097983*^9, 3.781952821828101*^9, 
    3.7819528975187283`*^9, {3.781953895876915*^9, 3.7819539154956713`*^9}, 
    3.7821306544745502`*^9, 3.782747157961987*^9, 3.7830788934387407`*^9, 
    3.785162684364028*^9, 3.785164974025779*^9, 3.789732649060177*^9, 
    3.79517465722147*^9, 3.799320669259963*^9, 3.799340142805358*^9, 
    3.799756425312364*^9, 3.800007655751677*^9, 3.80026818474804*^9, 
    3.8139309588718987`*^9, 3.813931061212427*^9, 3.814004119936627*^9},
  ExpressionUUID->"62e03f76-6f51-4229-8b76-53c646a3c31e"],
 " \nfinding the a, b, and c coefficients. "
}], "Text",
 CellChangeTimes->{{3.814010329691909*^9, 3.81401033968956*^9}, {
  3.8140172528489428`*^9, 
  3.814017260821107*^9}},ExpressionUUID->"cb3ec2b4-cd18-4d16-8684-\
5a33b61da948"],

Cell[CellGroupData[{

Cell[BoxData[{
 RowBox[{
  RowBox[{"Collect", "[", 
   RowBox[{
    RowBox[{"charpolynomial0i", "/", 
     RowBox[{"(", 
      RowBox[{
       RowBox[{"-", 
        FractionBox["1", "4"]}], " ", 
       SuperscriptBox["\[Lambda]", "2"], " ", 
       RowBox[{"(", 
        RowBox[{
         RowBox[{"-", "4"}], "+", "b", "+", 
         RowBox[{"2", " ", "\[Lambda]"}]}], ")"}]}], " ", ")"}]}], ",", 
    "\[Lambda]", ",", "Simplify"}], "]"}], ";"}], "\[IndentingNewLine]", 
 RowBox[{
  RowBox[{"Coefficient", "[", 
   RowBox[{"%", ",", 
    RowBox[{"\[Lambda]", "^", "2"}]}], "]"}], ";"}], "\[IndentingNewLine]", 
 RowBox[{"charpolySimpleACoef0i", "=", 
  RowBox[{"Collect", "[", 
   RowBox[{
    RowBox[{"%%", "/", "%"}], ",", "\[Lambda]", ",", "Simplify"}], 
   "]"}]}], "\[IndentingNewLine]", 
 RowBox[{
  RowBox[{"acoef0i", "=", 
   RowBox[{"Coefficient", "[", 
    RowBox[{"charpolySimpleACoef0i", ",", 
     RowBox[{"\[Lambda]", "^", "2"}]}], "]"}]}], ";"}], "\[IndentingNewLine]", 
 RowBox[{
  RowBox[{"bcoef0i", "=", 
   RowBox[{"Coefficient", "[", 
    RowBox[{"charpolySimpleACoef0i", ",", "\[Lambda]"}], "]"}]}], 
  ";"}], "\[IndentingNewLine]", 
 RowBox[{
  RowBox[{"ccoef0i", "=", 
   RowBox[{
    RowBox[{"charpolySimpleACoef0i", " ", "-", " ", 
     RowBox[{"bcoef0i", "*", "\[Lambda]"}], "-", 
     RowBox[{"acoef0i", "*", 
      RowBox[{"\[Lambda]", "^", "2"}]}]}], " ", "//", "Simplify"}]}], 
  ";"}]}], "Input",
 CellChangeTimes->{{3.781465386858316*^9, 3.781465388026968*^9}, {
  3.781465437030558*^9, 3.781465440011166*^9}, {3.814010383487996*^9, 
  3.8140104313523197`*^9}, {3.814017286391328*^9, 3.814017319542255*^9}, {
  3.814017370506489*^9, 3.8140173706172123`*^9}},
 CellLabel->
  "In[1427]:=",ExpressionUUID->"60c71e76-7525-4a3a-97d0-63b6944d2b63"],

Cell[BoxData[
 RowBox[{
  RowBox[{
   RowBox[{"-", 
    FractionBox["1", "2"]}], " ", "c", " ", 
   RowBox[{"(", 
    RowBox[{
     RowBox[{"-", "1"}], "+", 
     RowBox[{"h", " ", "sf"}]}], ")"}], " ", 
   RowBox[{"(", 
    RowBox[{
     RowBox[{"-", "1"}], "+", "sm"}], ")"}], " ", 
   RowBox[{"(", 
    RowBox[{"1", "+", "\[Delta]"}], ")"}]}], "+", 
  RowBox[{
   FractionBox["1", "2"], " ", 
   RowBox[{"(", 
    RowBox[{
     RowBox[{"-", "1"}], "+", 
     RowBox[{"h", " ", "sf"}]}], ")"}], " ", "\[Lambda]"}], "+", 
  SuperscriptBox["\[Lambda]", "2"]}]], "Output",
 CellChangeTimes->{{3.781465383147938*^9, 3.781465388446659*^9}, 
   3.781465441743964*^9, 3.781954038460005*^9, 3.781960483937896*^9, {
   3.783348419284086*^9, 3.783348423689898*^9}, 3.784050027570841*^9, 
   3.785570189009104*^9, 3.7952685727830067`*^9, 3.800009515393708*^9, 
   3.800268244980092*^9, {3.814010406744915*^9, 3.814010431746543*^9}, {
   3.8140172957484083`*^9, 3.8140173734647827`*^9}, 3.814091948062118*^9, 
   3.8141965689052067`*^9},
 CellLabel->
  "Out[1429]=",ExpressionUUID->"06bdfeba-862b-4c2b-8944-a32b7c8a8b12"]
}, Open  ]],

Cell["Now check the discriminant", "Text",
 CellChangeTimes->{{3.7814654589666853`*^9, 
  3.781465462010653*^9}},ExpressionUUID->"990c23e7-a271-4e5a-b33a-\
88dea689bc15"],

Cell[CellGroupData[{

Cell[BoxData["ccoef0i"], "Input",
 CellChangeTimes->{3.814017361622592*^9},
 CellLabel->
  "In[1433]:=",ExpressionUUID->"92bc3fda-d49d-41c0-a32b-6a3de94c7e02"],

Cell[BoxData[
 RowBox[{
  RowBox[{"-", 
   FractionBox["1", "2"]}], " ", "c", " ", 
  RowBox[{"(", 
   RowBox[{
    RowBox[{"-", "1"}], "+", 
    RowBox[{"h", " ", "sf"}]}], ")"}], " ", 
  RowBox[{"(", 
   RowBox[{
    RowBox[{"-", "1"}], "+", "sm"}], ")"}], " ", 
  RowBox[{"(", 
   RowBox[{"1", "+", "\[Delta]"}], ")"}]}]], "Output",
 CellChangeTimes->{{3.814017349981729*^9, 3.814017374371996*^9}, 
   3.814091949445006*^9, 3.814196568954535*^9},
 CellLabel->
  "Out[1433]=",ExpressionUUID->"15978260-aead-4eae-9ca7-cf3c57d812c7"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{
  RowBox[{
   RowBox[{"bcoef0i", "^", "2"}], "-", 
   RowBox[{"4", "acoef0i", " ", "ccoef0i"}]}], " ", "//", 
  "FullSimplify"}]], "Input",
 CellChangeTimes->{{3.814017325341064*^9, 3.814017379854299*^9}},
 CellLabel->
  "In[1434]:=",ExpressionUUID->"43d153fb-eb05-4992-b2bc-6bc01f67ef36"],

Cell[BoxData[
 RowBox[{
  FractionBox["1", "4"], " ", 
  RowBox[{"(", 
   RowBox[{
    RowBox[{"-", "1"}], "+", 
    RowBox[{"h", " ", "sf"}]}], ")"}], " ", 
  RowBox[{"(", 
   RowBox[{
    RowBox[{"-", "1"}], "+", 
    RowBox[{"h", " ", "sf"}], "+", 
    RowBox[{"8", " ", "c", " ", 
     RowBox[{"(", 
      RowBox[{
       RowBox[{"-", "1"}], "+", "sm"}], ")"}], " ", 
     RowBox[{"(", 
      RowBox[{"1", "+", "\[Delta]"}], ")"}]}]}], ")"}]}]], "Output",
 CellChangeTimes->{
  3.781465463847497*^9, 3.781954041615614*^9, 3.781960485991297*^9, {
   3.7833484193916273`*^9, 3.783348423774037*^9}, 3.7840500277604837`*^9, 
   3.7855701892119904`*^9, 3.795268572979253*^9, 3.8000095155902557`*^9, 
   3.800268245120281*^9, 3.814010444128042*^9, {3.8140173303015757`*^9, 
   3.814017380192358*^9}, 3.8140919501383877`*^9, 3.814196569008091*^9},
 CellLabel->
  "Out[1434]=",ExpressionUUID->"8d54243d-b6e6-4f83-80f3-a786c04bac3b"]
}, Open  ]],

Cell[TextData[{
 "We can see that this is always positive as the two terms in the product are \
always \[LessEqual] 0.\n\nThe conditions for the above characteristic \
polynomial ",
 Cell[BoxData[
  FormBox["p", TraditionalForm]],ExpressionUUID->
  "db78da60-1861-4cfe-9c4f-ba1070b2f2fb"],
 "  to have roots between -1 and 1 are:\n1. ",
 Cell[BoxData[
  FormBox[
   RowBox[{
    RowBox[{"p", "(", "1", ")"}], " ", ">", " ", "0"}], TraditionalForm]],
  ExpressionUUID->"540124b0-4bf5-46ae-ba97-47682f32b6c9"],
 "\n2. ",
 Cell[BoxData[
  FormBox[
   RowBox[{
    RowBox[{"p", "(", 
     RowBox[{"-", "1"}], ")"}], " ", ">", " ", "0"}], TraditionalForm]],
  ExpressionUUID->"38e38460-f669-48af-8bf9-d077d73acea4"],
 "\n3.  ",
 Cell[BoxData[
  FormBox[
   RowBox[{
    RowBox[{"-", "2"}], " ", "<", " ", "b", " ", "<", " ", "2"}], 
   TraditionalForm]],ExpressionUUID->"6ba76a76-46ef-470a-a904-6a0a6409fcc6"],
 "\nWhen any of these are false, the largest eigenvalue is ",
 Cell[BoxData[
  FormBox[
   RowBox[{">", 
    RowBox[{"1", "\:201a", " "}]}], TraditionalForm]],ExpressionUUID->
  "4614d629-e9a5-47ca-8319-361bc07cdd21"],
 "and the equilibrium is unstable. Condition 3 is always true:"
}], "Text",
 CellChangeTimes->{{3.781465492088834*^9, 3.781465506160551*^9}, {
  3.814010450688242*^9, 3.814010451992012*^9}, {3.81401091773033*^9, 
  3.814010947224269*^9}},ExpressionUUID->"d6b47106-1413-48c6-8b15-\
e893e39ce732"],

Cell[CellGroupData[{

Cell[BoxData[{
 RowBox[{"FullSimplify", "[", 
  RowBox[{
   RowBox[{
    RowBox[{"(", 
     RowBox[{"charpolySimpleACoef0i", "/.", 
      RowBox[{"\[Lambda]", "\[Rule]", "1"}]}], " ", ")"}], "<", " ", "0"}], 
   ",", "  ", 
   RowBox[{"{", 
    RowBox[{
     RowBox[{"1", ">", " ", "\[Delta]", ">", "0"}], ",", 
     RowBox[{"1", ">", "h", ">", "0"}], ",", 
     RowBox[{"1", ">", "sm", ">", "0"}], ",", 
     RowBox[{"0", "<", "sf", "<", "1"}]}], "}"}]}], 
  "]"}], "\[IndentingNewLine]", 
 RowBox[{"FullSimplify", "[", 
  RowBox[{
   RowBox[{
    RowBox[{"(", 
     RowBox[{"charpolySimpleACoef0i", "/.", 
      RowBox[{"\[Lambda]", "\[Rule]", 
       RowBox[{"-", "1"}]}]}], " ", ")"}], "<", " ", "0"}], ",", "  ", 
   RowBox[{"{", 
    RowBox[{
     RowBox[{"1", ">", " ", "\[Delta]", ">", "0"}], ",", 
     RowBox[{"1", ">", "h", ">", "0"}], ",", 
     RowBox[{"1", ">", "sm", ">", "0"}], ",", 
     RowBox[{"0", "<", "sf", "<", "1"}]}], "}"}]}], 
  "]"}], "\[IndentingNewLine]", 
 RowBox[{"FullSimplify", "[", 
  RowBox[{
   RowBox[{
    RowBox[{"-", "2"}], "<", "bcoef0i", "<", "2"}], ",", 
   RowBox[{"{", 
    RowBox[{
     RowBox[{"1", ">", " ", "\[Delta]", ">=", "0"}], ",", 
     RowBox[{"1", ">", "h", ">", "0"}], ",", 
     RowBox[{"1", ">", "sm", ">", "0"}], ",", 
     RowBox[{"0", "<", "sf", "<", "1"}]}], "}"}]}], "]"}]}], "Input",
 CellChangeTimes->{{3.781465526360558*^9, 3.7814655367472467`*^9}, {
   3.781465581129383*^9, 3.781465591995639*^9}, {3.781465688280877*^9, 
   3.781465701486*^9}, 3.78146618447902*^9, 3.7815122676036797`*^9, {
   3.814017397160585*^9, 3.814017401541716*^9}},
 CellLabel->
  "In[1435]:=",ExpressionUUID->"f81c340e-6f9a-495e-a4fb-f8522f1ec28b"],

Cell[BoxData[
 RowBox[{
  RowBox[{"1", "+", 
   RowBox[{"h", " ", "sf"}]}], "<", 
  RowBox[{"c", " ", 
   RowBox[{"(", 
    RowBox[{
     RowBox[{"-", "1"}], "+", 
     RowBox[{"h", " ", "sf"}]}], ")"}], " ", 
   RowBox[{"(", 
    RowBox[{
     RowBox[{"-", "1"}], "+", "sm"}], ")"}], " ", 
   RowBox[{"(", 
    RowBox[{"1", "+", "\[Delta]"}], ")"}]}]}]], "Output",
 CellChangeTimes->{{3.7814655286447268`*^9, 3.781465538365871*^9}, {
   3.781465582955743*^9, 3.781465592641117*^9}, 3.781465701977508*^9, 
   3.781466185013439*^9, 3.781512269203911*^9, 3.781954043802164*^9, 
   3.7819604884610863`*^9, {3.7833484196520147`*^9, 3.7833484238556013`*^9}, 
   3.784050028079029*^9, 3.785570189683228*^9, 3.7952685733076477`*^9, 
   3.8000095159012947`*^9, 3.800268245391749*^9, 3.814010457997061*^9, 
   3.8140174148280573`*^9, 3.814091952080904*^9, 3.814196569062356*^9},
 CellLabel->
  "Out[1435]=",ExpressionUUID->"a935214a-6eb2-4f38-a8f2-c4f423a3595f"],

Cell[BoxData[
 RowBox[{
  RowBox[{
   RowBox[{"h", " ", "sf"}], "+", 
   RowBox[{"c", " ", 
    RowBox[{"(", 
     RowBox[{
      RowBox[{"-", "1"}], "+", 
      RowBox[{"h", " ", "sf"}]}], ")"}], " ", 
    RowBox[{"(", 
     RowBox[{
      RowBox[{"-", "1"}], "+", "sm"}], ")"}], " ", 
    RowBox[{"(", 
     RowBox[{"1", "+", "\[Delta]"}], ")"}]}]}], ">", "3"}]], "Output",
 CellChangeTimes->{{3.7814655286447268`*^9, 3.781465538365871*^9}, {
   3.781465582955743*^9, 3.781465592641117*^9}, 3.781465701977508*^9, 
   3.781466185013439*^9, 3.781512269203911*^9, 3.781954043802164*^9, 
   3.7819604884610863`*^9, {3.7833484196520147`*^9, 3.7833484238556013`*^9}, 
   3.784050028079029*^9, 3.785570189683228*^9, 3.7952685733076477`*^9, 
   3.8000095159012947`*^9, 3.800268245391749*^9, 3.814010457997061*^9, 
   3.8140174148280573`*^9, 3.814091952080904*^9, 3.814196569122394*^9},
 CellLabel->
  "Out[1436]=",ExpressionUUID->"b31606ee-5624-4ecc-b4e4-3f5af30febb2"],

Cell[BoxData["True"], "Output",
 CellChangeTimes->{{3.7814655286447268`*^9, 3.781465538365871*^9}, {
   3.781465582955743*^9, 3.781465592641117*^9}, 3.781465701977508*^9, 
   3.781466185013439*^9, 3.781512269203911*^9, 3.781954043802164*^9, 
   3.7819604884610863`*^9, {3.7833484196520147`*^9, 3.7833484238556013`*^9}, 
   3.784050028079029*^9, 3.785570189683228*^9, 3.7952685733076477`*^9, 
   3.8000095159012947`*^9, 3.800268245391749*^9, 3.814010457997061*^9, 
   3.8140174148280573`*^9, 3.814091952080904*^9, 3.814196569123658*^9},
 CellLabel->
  "Out[1437]=",ExpressionUUID->"6af46331-b7a5-43eb-afc5-15a21d117d38"]
}, Open  ]],

Cell["\<\
The second of these inequalities implies the first and so we need only the \
first (only one condition is needed for instability).\
\>", "Text",
 CellChangeTimes->{{3.781465759480749*^9, 3.7814657696038733`*^9}, {
  3.781465837640523*^9, 3.7814658532489157`*^9}, {3.78187395266564*^9, 
  3.781873964963763*^9}},ExpressionUUID->"86fd9338-dd80-4568-b298-\
512a1d31063e"],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"unstabConditions0i", "=", 
  RowBox[{"FullSimplify", "[", 
   RowBox[{
    RowBox[{
     RowBox[{"(", 
      RowBox[{"charpolySimpleACoef0i", "/.", 
       RowBox[{"\[Lambda]", "\[Rule]", "1"}]}], " ", ")"}], "<", " ", "0"}], 
    ",", "  ", 
    RowBox[{"{", 
     RowBox[{
      RowBox[{"1", ">", " ", "\[Delta]", ">", "0"}], ",", 
      RowBox[{"1", ">", "h", ">", "0"}], ",", 
      RowBox[{"1", ">", "sm", ">", "0"}], ",", 
      RowBox[{"0", "<", "sf", "<", "1"}]}], "}"}]}], "]"}]}]], "Input",
 CellChangeTimes->{{3.781465877162455*^9, 3.781465971554441*^9}, 
   3.781873788934682*^9, {3.814017520555068*^9, 3.814017523140624*^9}},
 CellLabel->
  "In[1438]:=",ExpressionUUID->"8c892f84-4ed8-4152-959e-bd6460391a65"],

Cell[BoxData[
 RowBox[{
  RowBox[{"1", "+", 
   RowBox[{"h", " ", "sf"}]}], "<", 
  RowBox[{"c", " ", 
   RowBox[{"(", 
    RowBox[{
     RowBox[{"-", "1"}], "+", 
     RowBox[{"h", " ", "sf"}]}], ")"}], " ", 
   RowBox[{"(", 
    RowBox[{
     RowBox[{"-", "1"}], "+", "sm"}], ")"}], " ", 
   RowBox[{"(", 
    RowBox[{"1", "+", "\[Delta]"}], ")"}]}]}]], "Output",
 CellChangeTimes->{{3.781465886634276*^9, 3.781465972182002*^9}, 
   3.781873789761869*^9, 3.7819540469446173`*^9, 3.781960492622634*^9, {
   3.78334841995374*^9, 3.783348423946249*^9}, 3.784050028762014*^9, 
   3.78557019030302*^9, 3.795268573966165*^9, 3.8000095165324993`*^9, 
   3.800268245752823*^9, 3.814010464425288*^9, 3.814017523899661*^9, 
   3.814091953190713*^9, 3.8141965691753817`*^9},
 CellLabel->
  "Out[1438]=",ExpressionUUID->"ecea66e5-5689-4cc3-b8ac-c9532b465ef3"]
}, Open  ]],

Cell["which can be re-written using the w fitness notation", "Text",
 CellChangeTimes->{
  3.814010967375692*^9},ExpressionUUID->"d36a44fc-c0bc-4ed2-a7e9-\
1177e72d41c7"],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"Simplify", "[", 
  RowBox[{"unstabConditions0i", ",", " ", 
   RowBox[{"{", 
    RowBox[{
     RowBox[{"wXdX", " ", "\[Equal]", " ", 
      RowBox[{"1", " ", "-", " ", 
       RowBox[{"h", " ", "sf"}]}]}], ",", " ", 
     RowBox[{"wXdXd", " ", "\[Equal]", " ", 
      RowBox[{"1", " ", "-", " ", "sf"}]}], ",", " ", 
     RowBox[{"wXdY", " ", "\[Equal]", " ", 
      RowBox[{"1", " ", "-", " ", "sm"}]}]}], "}"}]}], "]"}]], "Input",
 CellChangeTimes->{{3.7833569152996073`*^9, 3.7833569483092127`*^9}, {
  3.814017527945406*^9, 3.81401752803433*^9}},
 CellLabel->
  "In[1439]:=",ExpressionUUID->"d61046b6-a41d-406a-926f-df53053cec8e"],

Cell[BoxData[
 RowBox[{
  RowBox[{"wXdX", "+", 
   RowBox[{"c", " ", "wXdX", " ", "wXdY", " ", 
    RowBox[{"(", 
     RowBox[{"1", "+", "\[Delta]"}], ")"}]}]}], ">", "2"}]], "Output",
 CellChangeTimes->{{3.783356930939971*^9, 3.7833569486658287`*^9}, 
   3.7840500291045218`*^9, 3.785570190631494*^9, 3.7952685742774467`*^9, 
   3.800009516893228*^9, 3.800268245888525*^9, 3.814010466301713*^9, 
   3.814010968648089*^9, 3.8140175282947598`*^9, 3.814091954218461*^9, 
   3.814196569239531*^9},
 CellLabel->
  "Out[1439]=",ExpressionUUID->"8747b626-48d0-46f3-97e6-e1eaa14bff1e"]
}, Open  ]]
}, Closed]],

Cell[CellGroupData[{

Cell["drive fixed (pm=pf=1) equilibrium", "Subsubsection",
 CellChangeTimes->{{3.7814659855542717`*^9, 3.781465989505756*^9}, {
  3.814018267841861*^9, 
  3.8140182726513033`*^9}},ExpressionUUID->"9448e96a-9b8d-4993-a76d-\
d262a1ec1243"],

Cell["\<\
We proceed similarly to before, but now considering the other equilibrium:\
\>", "Text",
 CellChangeTimes->{
  3.814010990066278*^9},ExpressionUUID->"5e5dbf94-c343-4c9f-a04b-\
a9a03660486d"],

Cell[CellGroupData[{

Cell[BoxData[{
 RowBox[{
  RowBox[{"Simplify", "[", 
   RowBox[{
    RowBox[{
     RowBox[{
      RowBox[{"jacobi", "/.", 
       RowBox[{"subeq", "[", 
        RowBox[{"[", "1", "]"}], "]"}]}], "/.", 
      RowBox[{"pf", "\[Rule]", "1"}]}], "/.", 
     RowBox[{"pm", "\[Rule]", "1"}]}], "/.", 
    RowBox[{"fixedDrivePopSizeSoli", "[", 
     RowBox[{"[", "2", "]"}], "]"}]}], "]"}], ";"}], "\[IndentingNewLine]", 
 RowBox[{"charpolynomial1i", "=", 
  RowBox[{"Factor", "[", 
   RowBox[{"Det", "[", 
    RowBox[{"%", "-", 
     RowBox[{"\[Lambda]", "*", 
      RowBox[{"IdentityMatrix", "[", "5", "]"}]}]}], "]"}], "]"}]}]}], "Input",
 CellChangeTimes->{{3.8140110467667093`*^9, 3.814011063131847*^9}, {
  3.8140175362249413`*^9, 3.814017543485701*^9}},
 CellLabel->
  "In[1440]:=",ExpressionUUID->"4d52337a-32db-4790-b7e5-808754491976"],

Cell[BoxData[
 RowBox[{
  FractionBox["1", 
   RowBox[{"4", " ", "c", " ", 
    RowBox[{"(", 
     RowBox[{
      RowBox[{"-", "1"}], "+", "sf"}], ")"}], " ", 
    RowBox[{"(", 
     RowBox[{
      RowBox[{"-", "1"}], "+", "sm"}], ")"}], " ", 
    RowBox[{"(", 
     RowBox[{"1", "+", "\[Delta]"}], ")"}]}]], 
  RowBox[{
   RowBox[{"(", 
    RowBox[{"4", "-", "b", "+", 
     RowBox[{"b", " ", "sf"}], "-", 
     RowBox[{"b", " ", "\[Delta]"}], "+", 
     RowBox[{"b", " ", "sf", " ", "\[Delta]"}], "-", 
     RowBox[{"2", " ", "\[Lambda]"}]}], ")"}], " ", 
   SuperscriptBox["\[Lambda]", "2"], " ", 
   RowBox[{"(", 
    RowBox[{
     RowBox[{"-", "1"}], "+", 
     RowBox[{"h", " ", "sf"}], "-", 
     RowBox[{"c", " ", "\[Lambda]"}], "+", 
     RowBox[{"c", " ", "h", " ", "sf", " ", "\[Lambda]"}], "+", 
     RowBox[{"c", " ", "sm", " ", "\[Lambda]"}], "-", 
     RowBox[{"c", " ", "h", " ", "sf", " ", "sm", " ", "\[Lambda]"}], "-", 
     RowBox[{"c", " ", "\[Delta]", " ", "\[Lambda]"}], "+", 
     RowBox[{"c", " ", "h", " ", "sf", " ", "\[Delta]", " ", "\[Lambda]"}], 
     "+", 
     RowBox[{"c", " ", "sm", " ", "\[Delta]", " ", "\[Lambda]"}], "-", 
     RowBox[{
     "c", " ", "h", " ", "sf", " ", "sm", " ", "\[Delta]", " ", "\[Lambda]"}],
      "+", 
     RowBox[{"2", " ", "c", " ", 
      SuperscriptBox["\[Lambda]", "2"]}], "-", 
     RowBox[{"2", " ", "c", " ", "sf", " ", 
      SuperscriptBox["\[Lambda]", "2"]}], "-", 
     RowBox[{"2", " ", "c", " ", "sm", " ", 
      SuperscriptBox["\[Lambda]", "2"]}], "+", 
     RowBox[{"2", " ", "c", " ", "sf", " ", "sm", " ", 
      SuperscriptBox["\[Lambda]", "2"]}], "+", 
     RowBox[{"2", " ", "c", " ", "\[Delta]", " ", 
      SuperscriptBox["\[Lambda]", "2"]}], "-", 
     RowBox[{"2", " ", "c", " ", "sf", " ", "\[Delta]", " ", 
      SuperscriptBox["\[Lambda]", "2"]}], "-", 
     RowBox[{"2", " ", "c", " ", "sm", " ", "\[Delta]", " ", 
      SuperscriptBox["\[Lambda]", "2"]}], "+", 
     RowBox[{"2", " ", "c", " ", "sf", " ", "sm", " ", "\[Delta]", " ", 
      SuperscriptBox["\[Lambda]", "2"]}]}], ")"}]}]}]], "Output",
 CellChangeTimes->{3.8140110640091867`*^9, 3.8140175442027082`*^9, 
  3.814091959446931*^9, 3.814196569617208*^9},
 CellLabel->
  "Out[1441]=",ExpressionUUID->"9845ca6f-247f-4c2f-a282-e60f91764288"]
}, Open  ]],

Cell[TextData[{
 "The factor ",
 Cell[BoxData[
  FractionBox[
   RowBox[{
    RowBox[{"(", 
     RowBox[{"4", "-", "b", "+", 
      RowBox[{"b", " ", "sf"}], "-", 
      RowBox[{"b", " ", "\[Delta]"}], "+", 
      RowBox[{"b", " ", "sf", " ", "\[Delta]"}], "-", 
      RowBox[{"2", " ", "\[Lambda]"}]}], ")"}], " ", 
    SuperscriptBox["\[Lambda]", "2"], " "}], 
   RowBox[{"4", " ", "c", " ", 
    RowBox[{"(", 
     RowBox[{
      RowBox[{"-", "1"}], "+", "sf"}], ")"}], " ", 
    RowBox[{"(", 
     RowBox[{
      RowBox[{"-", "1"}], "+", "sm"}], ")"}], " ", 
    RowBox[{"(", 
     RowBox[{"1", "+", "\[Delta]"}], ")"}]}]]],
  CellChangeTimes->{
   3.7812740033403482`*^9, 3.7814573218417273`*^9, 3.7818848609506083`*^9, 
    3.781952821819541*^9, 3.78195289751086*^9, {3.7819538958219957`*^9, 
    3.781953915448163*^9}, 3.7821306544665728`*^9, 3.782747157952016*^9, 
    3.7830788934301643`*^9, 3.785162683931014*^9, 3.785164973948011*^9, 
    3.789732648885466*^9, 3.795174657109274*^9, 3.799320669052409*^9, 
    3.799340142599029*^9, 3.799756425128694*^9, 3.800007655671279*^9, 
    3.8002681846876783`*^9, 3.813930901799224*^9, {3.8139309343336077`*^9, 
    3.8139310598724957`*^9}, 3.8140037587039423`*^9, 3.814003805149556*^9},
  ExpressionUUID->"e4a601ae-ed72-4dbd-b92f-ba86311c2557"],
 " gives two 0 eigenvalues and and eigenvalue of ",
 Cell[BoxData[
  RowBox[{"2", "+", 
   RowBox[{
    FractionBox["1", "2"], " ", "b", " ", 
    RowBox[{"(", 
     RowBox[{
      RowBox[{"-", "1"}], "+", "sf"}], ")"}], " ", 
    RowBox[{"(", 
     RowBox[{"1", "+", "\[Delta]"}], ")"}]}]}]],
  CellChangeTimes->{3.814003986297804*^9},ExpressionUUID->
  "e15c661f-31a3-4750-8e8a-f7d5119e502b"],
 ", corresponding to the stability of the ecological equilibrium population \
size. Thus, we can see that the demographic dynamics are orthogonal to the \
evolutionary dynamics. \n\nHere,  we divide through by this factor to \
simplify the equation and evaluate the evolutionary dynamics. We then write \
the characteristic polynomial in the form\n",
 Cell[BoxData[
  RowBox[{
   SuperscriptBox["a\[Lambda]", "2"], "+", "b\[Lambda]", "+", "c"}]],
  CellChangeTimes->{
   3.7812740321927767`*^9, 3.781274073867482*^9, 3.7814285214279013`*^9, 
    3.781457321850852*^9, 3.78188486097983*^9, 3.781952821828101*^9, 
    3.7819528975187283`*^9, {3.781953895876915*^9, 3.7819539154956713`*^9}, 
    3.7821306544745502`*^9, 3.782747157961987*^9, 3.7830788934387407`*^9, 
    3.785162684364028*^9, 3.785164974025779*^9, 3.789732649060177*^9, 
    3.79517465722147*^9, 3.799320669259963*^9, 3.799340142805358*^9, 
    3.799756425312364*^9, 3.800007655751677*^9, 3.80026818474804*^9, 
    3.8139309588718987`*^9, 3.813931061212427*^9, 3.814004119936627*^9},
  ExpressionUUID->"1bcb9fe9-dee0-46c4-b3cb-7713744c3470"],
 " \nfinding the a, b, and c coefficients. "
}], "Text",
 CellChangeTimes->{
  3.81401109022639*^9, {3.814011153607579*^9, 3.81401115360782*^9}, {
   3.814017550438505*^9, 
   3.8140175574448767`*^9}},ExpressionUUID->"75774f4d-26ff-42a6-b382-\
e34a5c2c7e24"],

Cell[CellGroupData[{

Cell[BoxData[{
 RowBox[{
  RowBox[{"Collect", "[", 
   RowBox[{
    RowBox[{"charpolynomial1i", "/", 
     RowBox[{"(", 
      FractionBox[
       RowBox[{
        RowBox[{"(", 
         RowBox[{"4", "-", "b", "+", 
          RowBox[{"b", " ", "sf"}], "-", 
          RowBox[{"b", " ", "\[Delta]"}], "+", 
          RowBox[{"b", " ", "sf", " ", "\[Delta]"}], "-", 
          RowBox[{"2", " ", "\[Lambda]"}]}], ")"}], " ", 
        SuperscriptBox["\[Lambda]", "2"], " "}], 
       RowBox[{"4", " ", "c", " ", 
        RowBox[{"(", 
         RowBox[{
          RowBox[{"-", "1"}], "+", "sf"}], ")"}], " ", 
        RowBox[{"(", 
         RowBox[{
          RowBox[{"-", "1"}], "+", "sm"}], ")"}], " ", 
        RowBox[{"(", 
         RowBox[{"1", "+", "\[Delta]"}], ")"}]}]], ")"}]}], ",", "\[Lambda]", 
    ",", "Simplify"}], "]"}], ";"}], "\[IndentingNewLine]", 
 RowBox[{
  RowBox[{"Coefficient", "[", 
   RowBox[{"%", ",", 
    RowBox[{"\[Lambda]", "^", "2"}]}], "]"}], ";"}], "\[IndentingNewLine]", 
 RowBox[{"charpolySimpleACoef1i", "=", 
  RowBox[{"Collect", "[", 
   RowBox[{
    RowBox[{"%%", "/", "%"}], ",", "\[Lambda]", ",", "Simplify"}], 
   "]"}]}], "\[IndentingNewLine]", 
 RowBox[{
  RowBox[{"acoef1i", "=", 
   RowBox[{"Coefficient", "[", 
    RowBox[{"charpolySimpleACoef1i", ",", 
     RowBox[{"\[Lambda]", "^", "2"}]}], "]"}]}], ";"}], "\[IndentingNewLine]", 
 RowBox[{
  RowBox[{"bcoef1i", "=", 
   RowBox[{"Coefficient", "[", 
    RowBox[{"charpolySimpleACoef1i", ",", "\[Lambda]"}], "]"}]}], 
  ";"}], "\[IndentingNewLine]", 
 RowBox[{
  RowBox[{"ccoef1i", "=", 
   RowBox[{"charpolySimpleACoef1i", "-", " ", 
    RowBox[{"bcoef1i", "*", "\[Lambda]"}], "-", 
    RowBox[{"acoef1i", "*", 
     RowBox[{"\[Lambda]", "^", "2"}]}]}]}], ";"}]}], "Input",
 CellChangeTimes->{{3.7814660158915987`*^9, 3.781466030367185*^9}, {
  3.781511388407653*^9, 3.781511406925776*^9}, {3.7815114853232603`*^9, 
  3.781511487165525*^9}, {3.8140110114581947`*^9, 3.814011035726304*^9}, {
  3.81401119347377*^9, 3.8140112094565897`*^9}, {3.814017596782649*^9, 
  3.814017619218994*^9}},
 CellLabel->
  "In[1442]:=",ExpressionUUID->"7e78625f-aa15-4880-88d4-b4d8b1e2b50b"],

Cell[BoxData[
 RowBox[{
  FractionBox[
   RowBox[{
    RowBox[{"-", "1"}], "+", 
    RowBox[{"h", " ", "sf"}]}], 
   RowBox[{"2", " ", "c", " ", 
    RowBox[{"(", 
     RowBox[{
      RowBox[{"-", "1"}], "+", "sf"}], ")"}], " ", 
    RowBox[{"(", 
     RowBox[{
      RowBox[{"-", "1"}], "+", "sm"}], ")"}], " ", 
    RowBox[{"(", 
     RowBox[{"1", "+", "\[Delta]"}], ")"}]}]], "+", 
  FractionBox[
   RowBox[{
    RowBox[{"(", 
     RowBox[{"1", "-", 
      RowBox[{"h", " ", "sf"}]}], ")"}], " ", "\[Lambda]"}], 
   RowBox[{"2", " ", 
    RowBox[{"(", 
     RowBox[{
      RowBox[{"-", "1"}], "+", "sf"}], ")"}]}]], "+", 
  SuperscriptBox["\[Lambda]", "2"]}]], "Output",
 CellChangeTimes->{{3.781465999356587*^9, 3.781466030701129*^9}, {
   3.781511391556911*^9, 3.7815114073942223`*^9}, 3.7815114881434813`*^9, 
   3.7815133913596582`*^9, 3.7819604953090477`*^9, 3.783348438199882*^9, 
   3.7840500300847263`*^9, 3.7855701908368073`*^9, 3.7952685744353533`*^9, 
   3.800009517092691*^9, 3.800268246008481*^9, 3.814011037187717*^9, {
   3.81401120075671*^9, 3.81401120992056*^9}, 3.8140176195500317`*^9, 
   3.8140919611332684`*^9, 3.814196569687551*^9},
 CellLabel->
  "Out[1444]=",ExpressionUUID->"1b7b201f-282d-4ee2-a0b4-1da70df94894"]
}, Open  ]],

Cell["\<\
The discriminant is nonnegative and so the eigenvalues are real:\
\>", "Text",
 CellChangeTimes->{{3.781466041455315*^9, 3.781466045839364*^9}, 
   3.8140112634898663`*^9},ExpressionUUID->"3e1c4ca1-1e86-4c4e-9d39-\
b51c02a3b3aa"],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{
  RowBox[{"bcoef1i", "^", "2"}], "-", 
  RowBox[{"4", "acoef1i", " ", "ccoef1i"}]}]], "Input",
 CellChangeTimes->{{3.78146605752468*^9, 3.7814660576893463`*^9}, {
  3.8140176241597137`*^9, 3.814017628306972*^9}},
 CellLabel->
  "In[1448]:=",ExpressionUUID->"70f2be14-e7b7-480a-ab97-979e3cd0984b"],

Cell[BoxData[
 RowBox[{
  FractionBox[
   SuperscriptBox[
    RowBox[{"(", 
     RowBox[{"1", "-", 
      RowBox[{"h", " ", "sf"}]}], ")"}], "2"], 
   RowBox[{"4", " ", 
    SuperscriptBox[
     RowBox[{"(", 
      RowBox[{
       RowBox[{"-", "1"}], "+", "sf"}], ")"}], "2"]}]], "-", 
  FractionBox[
   RowBox[{"2", " ", 
    RowBox[{"(", 
     RowBox[{
      RowBox[{"-", "1"}], "+", 
      RowBox[{"h", " ", "sf"}]}], ")"}]}], 
   RowBox[{"c", " ", 
    RowBox[{"(", 
     RowBox[{
      RowBox[{"-", "1"}], "+", "sf"}], ")"}], " ", 
    RowBox[{"(", 
     RowBox[{
      RowBox[{"-", "1"}], "+", "sm"}], ")"}], " ", 
    RowBox[{"(", 
     RowBox[{"1", "+", "\[Delta]"}], ")"}]}]]}]], "Output",
 CellChangeTimes->{{3.78146605450915*^9, 3.7814660582264423`*^9}, 
   3.781513396654119*^9, 3.7819604977917757`*^9, 3.783348438423812*^9, 
   3.784050030288706*^9, 3.7855701909125433`*^9, 3.795268574496313*^9, 
   3.800009517154484*^9, 3.80026824614252*^9, 3.814017628538315*^9, 
   3.814091962838134*^9, 3.814196569695698*^9},
 CellLabel->
  "Out[1448]=",ExpressionUUID->"a06bc303-9f62-469f-b0e2-1f194523403c"]
}, Open  ]],

Cell[TextData[{
 "The conditions for the above characteristic polynomial ",
 Cell[BoxData[
  FormBox["p", TraditionalForm]],ExpressionUUID->
  "04b9d021-7061-4b4f-b872-53b2e565660c"],
 "  to have roots between -1 and 1 are:\n1. ",
 Cell[BoxData[
  FormBox[
   RowBox[{
    RowBox[{"p", "(", "1", ")"}], " ", ">", " ", "0"}], TraditionalForm]],
  ExpressionUUID->"6f77e5dd-c580-48a9-b92b-41c3b4c9d2f4"],
 "\n2. ",
 Cell[BoxData[
  FormBox[
   RowBox[{
    RowBox[{"p", "(", 
     RowBox[{"-", "1"}], ")"}], " ", ">", " ", "0"}], TraditionalForm]],
  ExpressionUUID->"9c9c95f6-859b-43ae-a2c9-2a4769a1d292"],
 "\n3.  ",
 Cell[BoxData[
  FormBox[
   RowBox[{
    RowBox[{"-", "2"}], " ", "<", " ", "b", " "}], TraditionalForm]],
  ExpressionUUID->"d353e791-4ba0-4e2f-b9c7-d7522396328a"],
 "\n4.  ",
 Cell[BoxData[
  FormBox[
   RowBox[{"b", " ", "<", " ", "2", " "}], TraditionalForm]],ExpressionUUID->
  "3e9e9deb-4eea-4c30-85ed-037338e84180"],
 "\nWhen any of these are false, the largest eigenvalue is ",
 Cell[BoxData[
  FormBox[
   RowBox[{">", 
    RowBox[{"1", "\:201a", " "}]}], TraditionalForm]],ExpressionUUID->
  "992b7b89-d175-4101-825d-f54a377ae104"],
 "and the equilibrium is unstable. \n\nNow checking these conditions for \
instability by reversing the inequalities required for stability"
}], "Text",
 CellChangeTimes->{{3.781466076377097*^9, 3.78146608440063*^9}, {
   3.781466192176076*^9, 3.781466195526973*^9}, {3.781960505405582*^9, 
   3.781960506689489*^9}, 
   3.814011272224868*^9},ExpressionUUID->"b1e1aa0a-d623-4803-93b4-\
15419b2e8379"],

Cell[CellGroupData[{

Cell[BoxData[{
 RowBox[{"r1", "=", 
  RowBox[{"FullSimplify", "[", 
   RowBox[{
    RowBox[{
     RowBox[{"(", 
      RowBox[{"charpolySimpleACoef1i", "/.", 
       RowBox[{"\[Lambda]", "\[Rule]", "1"}]}], " ", ")"}], "<", " ", "0"}], 
    ",", "  ", 
    RowBox[{"{", 
     RowBox[{
      RowBox[{"1", ">", " ", "\[Delta]", ">", "0"}], ",", 
      RowBox[{"1", ">", "h", ">", "0"}], ",", 
      RowBox[{"1", ">", "sm", ">", "0"}], ",", 
      RowBox[{"0", "<", "sf", "<", "1"}]}], "}"}]}], 
   "]"}]}], "\[IndentingNewLine]", 
 RowBox[{"r2", "=", 
  RowBox[{"FullSimplify", "[", 
   RowBox[{
    RowBox[{
     RowBox[{"(", 
      RowBox[{"charpolySimpleACoef1i", "/.", 
       RowBox[{"\[Lambda]", "\[Rule]", 
        RowBox[{"-", "1"}]}]}], " ", ")"}], "<", " ", "0"}], ",", "  ", 
    RowBox[{"{", 
     RowBox[{
      RowBox[{"1", ">", " ", "\[Delta]", ">", "0"}], ",", 
      RowBox[{"1", ">", "h", ">", "0"}], ",", 
      RowBox[{"1", ">", "sm", ">", "0"}], ",", 
      RowBox[{"0", "<", "sf", "<", "1"}]}], "}"}]}], 
   "]"}]}], "\[IndentingNewLine]", 
 RowBox[{"r3", "=", 
  RowBox[{"FullSimplify", "[", 
   RowBox[{
    RowBox[{"bcoef1i", "<", 
     RowBox[{"-", "2"}]}], ",", 
    RowBox[{"{", 
     RowBox[{
      RowBox[{"1", ">", " ", "\[Delta]", ">", "0"}], ",", 
      RowBox[{"1", ">", "h", ">", "0"}], ",", 
      RowBox[{"1", ">", "sm", ">", "0"}], ",", 
      RowBox[{"0", "<", "sf", "<", "1"}]}], "}"}]}], 
   "]"}]}], "\[IndentingNewLine]", 
 RowBox[{"r4", "=", 
  RowBox[{"FullSimplify", "[", 
   RowBox[{
    RowBox[{"2", "<", "bcoef1i"}], ",", 
    RowBox[{"{", 
     RowBox[{
      RowBox[{"1", ">", " ", "\[Delta]", ">", "0"}], ",", 
      RowBox[{"1", ">", "h", ">", "0"}], ",", 
      RowBox[{"1", ">", "sm", ">", "0"}], ",", 
      RowBox[{"0", "<", "sf", "<", "1"}]}], "}"}]}], "]"}]}]}], "Input",
 CellChangeTimes->{{3.781466198557598*^9, 3.781466244073738*^9}, {
  3.781513299834971*^9, 3.78151330877592*^9}, {3.814017658989674*^9, 
  3.814017665107162*^9}},
 CellLabel->
  "In[1449]:=",ExpressionUUID->"7f52a622-8937-410a-88ec-95958fc17686"],

Cell[BoxData[
 RowBox[{
  RowBox[{"c", " ", 
   RowBox[{"(", 
    RowBox[{"1", "-", 
     RowBox[{"h", " ", "sf"}], "+", 
     RowBox[{"c", " ", 
      RowBox[{"(", 
       RowBox[{"1", "+", 
        RowBox[{
         RowBox[{"(", 
          RowBox[{
           RowBox[{"-", "2"}], "+", "h"}], ")"}], " ", "sf"}]}], ")"}], " ", 
      RowBox[{"(", 
       RowBox[{
        RowBox[{"-", "1"}], "+", "sm"}], ")"}], " ", 
      RowBox[{"(", 
       RowBox[{"1", "+", "\[Delta]"}], ")"}]}]}], ")"}]}], ">", 
  "0"}]], "Output",
 CellChangeTimes->{
  3.781466207213058*^9, 3.7814662447499647`*^9, {3.78151331072229*^9, 
   3.781513335444324*^9}, 3.781513400170166*^9, 3.781960509141514*^9, 
   3.7833484391219378`*^9, 3.7840500312745237`*^9, 3.785570191793244*^9, 
   3.7952685753436327`*^9, 3.80000951794683*^9, 3.80026824682135*^9, 
   3.814011276950014*^9, 3.814017665533905*^9, 3.814091965616982*^9, 
   3.8141965698031073`*^9},
 CellLabel->
  "Out[1449]=",ExpressionUUID->"954154d8-a97c-42ba-8afa-41f7a9da4d51"],

Cell[BoxData[
 RowBox[{
  RowBox[{"c", " ", 
   RowBox[{"(", 
    RowBox[{
     RowBox[{"-", "1"}], "+", 
     RowBox[{"h", " ", "sf"}], "+", 
     RowBox[{"c", " ", 
      RowBox[{"(", 
       RowBox[{
        RowBox[{"-", "3"}], "+", 
        RowBox[{
         RowBox[{"(", 
          RowBox[{"2", "+", "h"}], ")"}], " ", "sf"}]}], ")"}], " ", 
      RowBox[{"(", 
       RowBox[{
        RowBox[{"-", "1"}], "+", "sm"}], ")"}], " ", 
      RowBox[{"(", 
       RowBox[{"1", "+", "\[Delta]"}], ")"}]}]}], ")"}]}], "<", 
  "0"}]], "Output",
 CellChangeTimes->{
  3.781466207213058*^9, 3.7814662447499647`*^9, {3.78151331072229*^9, 
   3.781513335444324*^9}, 3.781513400170166*^9, 3.781960509141514*^9, 
   3.7833484391219378`*^9, 3.7840500312745237`*^9, 3.785570191793244*^9, 
   3.7952685753436327`*^9, 3.80000951794683*^9, 3.80026824682135*^9, 
   3.814011276950014*^9, 3.814017665533905*^9, 3.814091965616982*^9, 
   3.814196569889516*^9},
 CellLabel->
  "Out[1450]=",ExpressionUUID->"e5aa4eb8-c218-476b-ba22-904b8b4646b6"],

Cell[BoxData[
 RowBox[{
  RowBox[{"3", "+", 
   RowBox[{"h", " ", "sf"}]}], "<", 
  RowBox[{"4", " ", "sf"}]}]], "Output",
 CellChangeTimes->{
  3.781466207213058*^9, 3.7814662447499647`*^9, {3.78151331072229*^9, 
   3.781513335444324*^9}, 3.781513400170166*^9, 3.781960509141514*^9, 
   3.7833484391219378`*^9, 3.7840500312745237`*^9, 3.785570191793244*^9, 
   3.7952685753436327`*^9, 3.80000951794683*^9, 3.80026824682135*^9, 
   3.814011276950014*^9, 3.814017665533905*^9, 3.814091965616982*^9, 
   3.8141965699333277`*^9},
 CellLabel->
  "Out[1451]=",ExpressionUUID->"786a77f7-de8f-4b06-9fa4-31fda5552368"],

Cell[BoxData["False"], "Output",
 CellChangeTimes->{
  3.781466207213058*^9, 3.7814662447499647`*^9, {3.78151331072229*^9, 
   3.781513335444324*^9}, 3.781513400170166*^9, 3.781960509141514*^9, 
   3.7833484391219378`*^9, 3.7840500312745237`*^9, 3.785570191793244*^9, 
   3.7952685753436327`*^9, 3.80000951794683*^9, 3.80026824682135*^9, 
   3.814011276950014*^9, 3.814017665533905*^9, 3.814091965616982*^9, 
   3.814196569934585*^9},
 CellLabel->
  "Out[1452]=",ExpressionUUID->"c68377aa-5875-43ae-ac14-f3d44a58a099"]
}, Open  ]],

Cell["\<\
The second inequality implies the first again. r3 only applies if \[Delta] or \
c is less than 1. Otherwise, it is covered by r1. \
\>", "Text",
 CellChangeTimes->{{3.781440514566372*^9, 3.781440560329487*^9}, {
  3.781513468075613*^9, 3.78151364185474*^9}, {3.7815137448951883`*^9, 
  3.781513750665613*^9}, {3.781515522495345*^9, 
  3.78151552384299*^9}},ExpressionUUID->"2e9457fc-050d-4c87-833d-\
f09c218d7e7b"],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"Reduce", "[", 
  RowBox[{"{", 
   RowBox[{
    RowBox[{"!", "r1"}], ",", "r3", ",", 
    RowBox[{
     RowBox[{"1", "==", " ", "\[Delta]"}], "||", 
     RowBox[{"1", "\[Equal]", "c"}]}], ",", 
    RowBox[{"0", "<", "\[Delta]", "\[LessEqual]", "1"}], ",", 
    RowBox[{"0", "<", "c", "\[LessEqual]", "1"}], ",", 
    RowBox[{"1", ">", "h", ">", "0"}], ",", 
    RowBox[{"1", ">", "sm", ">", "0"}], ",", 
    RowBox[{"0", "<", "sf", "<", "1"}]}], "}"}], "]"}]], "Input",
 CellChangeTimes->{{3.781438491275654*^9, 3.7814385705794687`*^9}, {
   3.781438609675919*^9, 3.781438609925259*^9}, {3.781439361820053*^9, 
   3.781439383480029*^9}, 3.781513324912135*^9, {3.781513413430835*^9, 
   3.7815134539688053`*^9}},
 CellLabel->
  "In[1453]:=",ExpressionUUID->"ffcc99d6-a5f0-4ae8-a087-25cf2910c2ba"],

Cell[BoxData["False"], "Output",
 CellChangeTimes->{
  3.7814385244255323`*^9, {3.78143855490099*^9, 3.781438571329862*^9}, 
   3.781438610236524*^9, {3.781439344792201*^9, 3.781439389623969*^9}, 
   3.7814394716890707`*^9, 3.781440562046509*^9, 3.781457323045916*^9, {
   3.7815134086526213`*^9, 3.781513419403633*^9}, 3.78151345446651*^9, 
   3.7833484401216583`*^9, 3.784050032554224*^9, 3.785570192992031*^9, 
   3.795268576496269*^9, 3.800009519129652*^9, 3.8002682477844467`*^9, 
   3.814011286219734*^9, 3.814017694733781*^9, 3.8140919672756767`*^9, 
   3.814196569990034*^9},
 CellLabel->
  "Out[1453]=",ExpressionUUID->"a2fc5805-70d9-4a63-84aa-b77ff4115eba"]
}, Open  ]],

Cell[TextData[{
 "If \[Delta]<1 or c<1, r3 shows that the drive-fixed equilibrium is always \
unstable when heterozygous females have at least 4 times higher survival \
probability than homozygous drive females: re-writing r3 as  ",
 Cell[BoxData[
  RowBox[{
   RowBox[{"4", 
    RowBox[{"(", 
     RowBox[{"1", "-", "sf"}], ")"}]}], "<", 
   RowBox[{"1", "-", 
    RowBox[{"h", " ", "sf"}]}]}]],
  CellChangeTimes->{{3.781513605956525*^9, 3.781513625762225*^9}},
  ExpressionUUID->"99ee6d27-8a79-4d8a-a3af-b3cb9b62bba8"],
 ". Even if r3 is not met, the drive-fixed equilibrium can still be unstable \
when the r1 condition is met. "
}], "Text",
 CellChangeTimes->{{3.781513755067279*^9, 3.781513782566339*^9}, {
  3.781513833350754*^9, 
  3.7815138689367247`*^9}},ExpressionUUID->"fe1c83ee-be24-4e16-943d-\
a90f1fee73e1"],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"unstabConditions1i", " ", "=", " ", 
  RowBox[{"r1", "||", "r3"}]}]], "Input",
 CellChangeTimes->{{3.7818731353415947`*^9, 3.78187317241343*^9}, {
  3.781873218740693*^9, 3.781873219335464*^9}, {3.81401770207661*^9, 
  3.814017702174984*^9}},
 CellLabel->
  "In[1454]:=",ExpressionUUID->"8d8b0218-772f-4eeb-a34b-589827e37067"],

Cell[BoxData[
 RowBox[{
  RowBox[{
   RowBox[{"c", " ", 
    RowBox[{"(", 
     RowBox[{"1", "-", 
      RowBox[{"h", " ", "sf"}], "+", 
      RowBox[{"c", " ", 
       RowBox[{"(", 
        RowBox[{"1", "+", 
         RowBox[{
          RowBox[{"(", 
           RowBox[{
            RowBox[{"-", "2"}], "+", "h"}], ")"}], " ", "sf"}]}], ")"}], " ", 
       RowBox[{"(", 
        RowBox[{
         RowBox[{"-", "1"}], "+", "sm"}], ")"}], " ", 
       RowBox[{"(", 
        RowBox[{"1", "+", "\[Delta]"}], ")"}]}]}], ")"}]}], ">", "0"}], "||", 
  RowBox[{
   RowBox[{"3", "+", 
    RowBox[{"h", " ", "sf"}]}], "<", 
   RowBox[{"4", " ", "sf"}]}]}]], "Output",
 CellChangeTimes->{{3.781873149935636*^9, 3.781873172799984*^9}, 
   3.781873220134734*^9, 3.7819605344534063`*^9, 3.783348440138775*^9, 
   3.784050032765121*^9, 3.785570193196538*^9, 3.795268576688867*^9, 
   3.800009519338931*^9, 3.800268247829068*^9, 3.814011297261401*^9, {
   3.81401769837959*^9, 3.8140177027719*^9}, 3.814091968468122*^9, 
   3.814196570002014*^9},
 CellLabel->
  "Out[1454]=",ExpressionUUID->"47d051a0-b998-4322-b2d0-2bd0cfada6c4"]
}, Open  ]]
}, Closed]]
}, Closed]]
}, Closed]],

Cell[CellGroupData[{

Cell[TextData[StyleBox["Plots", "Subtitle"]], "Chapter",
 CellChangeTimes->{{3.783256596451075*^9, 
  3.7832565968784857`*^9}},ExpressionUUID->"9ba9e339-5f1d-42fd-8e6e-\
d3494987d8b8"],

Cell[BoxData[{
 RowBox[{
  RowBox[{"SetDirectory", "[", 
   RowBox[{"NotebookDirectory", "[", "]"}], "]"}], 
  ";"}], "\[IndentingNewLine]", 
 RowBox[{
  RowBox[{"ticks", " ", "=", " ", 
   RowBox[{"{", 
    RowBox[{
     RowBox[{"{", 
      RowBox[{"1", ",", "0"}], "}"}], ",", 
     RowBox[{"{", 
      RowBox[{"5", ",", "0.2"}], "}"}], ",", 
     RowBox[{"{", 
      RowBox[{"9", ",", "0.4"}], "}"}], ",", 
     RowBox[{"{", 
      RowBox[{"13", ",", "0.6"}], "}"}], ",", 
     RowBox[{"{", 
      RowBox[{"17", ",", "0.8"}], "}"}], ",", 
     RowBox[{"{", 
      RowBox[{"21", ",", "1"}], "}"}]}], "}"}]}], 
  ";"}], "\[IndentingNewLine]"}], "Input",
 CellChangeTimes->{{3.7832566001197557`*^9, 3.7832566212691174`*^9}, {
   3.7832569225841093`*^9, 3.7832569702125807`*^9}, {3.783257069393852*^9, 
   3.783257111864566*^9}, 3.783257156145256*^9, {3.7832572156575108`*^9, 
   3.783257251039603*^9}, {3.783257794194255*^9, 3.783257795868723*^9}, 
   3.8000184133909597`*^9, {3.809688012905837*^9, 3.809688043926581*^9}, {
   3.8096882095072517`*^9, 3.8096882713823433`*^9}, {3.8096883409964237`*^9, 
   3.8096883775984497`*^9}, {3.809688669077989*^9, 3.809688727943183*^9}, {
   3.809691687031938*^9, 3.8096916878842363`*^9}},
 CellLabel->
  "In[1508]:=",ExpressionUUID->"548e99b5-8fd9-4ae2-a7c2-64982b0bc649"],

Cell[CellGroupData[{

Cell["Boundaries", "Subsubsection",
 CellChangeTimes->{{3.78325780871565*^9, 3.78325783066441*^9}, {
  3.800290031690465*^9, 
  3.800290032937936*^9}},ExpressionUUID->"7d1cc74b-1f3d-4fbf-b304-\
bd7a55bca76f"],

Cell["One female mating, \[Delta] = 1, sm = 0", "Text",
 CellChangeTimes->{{3.783257841734631*^9, 
  3.7832578575294437`*^9}},ExpressionUUID->"fa3eebbc-479f-4bc3-bfb9-\
9231b76136a9"],

Cell[CellGroupData[{

Cell[BoxData[{
 RowBox[{
  RowBox[{"data", " ", "=", " ", 
   RowBox[{
   "Import", "[", "\"\<per_hmap_b3_sm0_fm1_mm20_delta1_c1.csv\>\"", "]"}]}], 
  ";"}], "\[IndentingNewLine]", 
 RowBox[{
  RowBox[{
   RowBox[{"b3sm0fm1mm20delta1c1HMap", " ", "=", " ", 
    RowBox[{"MatrixPlot", "[", 
     RowBox[{"data", ",", 
      RowBox[{"DataReversed", "\[Rule]", "True"}], ",", 
      RowBox[{"Frame", "\[Rule]", "True"}], ",", " ", 
      RowBox[{"FrameTicks", "\[Rule]", 
       RowBox[{"{", 
        RowBox[{
         RowBox[{"{", 
          RowBox[{"ticks", ",", "None"}], "}"}], ",", 
         RowBox[{"{", 
          RowBox[{"ticks", ",", "None"}], "}"}]}], "}"}]}], ",", " ", 
      RowBox[{"FrameTicksStyle", "\[Rule]", 
       RowBox[{"Directive", "[", "White", "]"}]}], ",", 
      RowBox[{"FrameStyle", "\[Rule]", 
       RowBox[{"Directive", "[", "White", "]"}]}], ",", 
      RowBox[{"PlotLegends", "\[Rule]", "legend"}], ",", 
      RowBox[{"ColorFunctionScaling", "\[Rule]", "False"}], ",", 
      RowBox[{"ColorFunction", "\[Rule]", 
       RowBox[{"(", 
        RowBox[{
         RowBox[{"Lighter", "[", 
          RowBox[{"Orange", ",", 
           RowBox[{"1", "-", "#"}]}], "]"}], "&"}], ")"}]}]}], "]"}]}], ";"}],
   "\[IndentingNewLine]"}], "\[IndentingNewLine]", 
 RowBox[{"sm0fm1delta1c1RPlot", " ", "=", " ", 
  RowBox[{"RegionPlot", "[", 
   RowBox[{
    RowBox[{"{", 
     RowBox[{
      RowBox[{"(", 
       RowBox[{"unstabConditions0", "/.", 
        RowBox[{"{", 
         RowBox[{
          RowBox[{"sm", "\[Rule]", "0"}], ",", 
          RowBox[{"\[Delta]", "\[Rule]", "1"}], ",", 
          RowBox[{"c", "\[Rule]", "1"}]}], "}"}]}], ")"}], ",", 
      RowBox[{
       RowBox[{"(", 
        RowBox[{"unstabConditions0", "/.", 
         RowBox[{"{", 
          RowBox[{
           RowBox[{"sm", "\[Rule]", "0"}], ",", 
           RowBox[{"\[Delta]", "\[Rule]", "1"}], ",", 
           RowBox[{"c", "\[Rule]", "1"}]}], "}"}]}], ")"}], "&&", 
       RowBox[{"(", 
        RowBox[{"unstabConditions1", "/.", 
         RowBox[{"{", 
          RowBox[{
           RowBox[{"sm", "\[Rule]", "0"}], ",", 
           RowBox[{"\[Delta]", "\[Rule]", "1"}], ",", 
           RowBox[{"c", "\[Rule]", "1"}]}], "}"}]}], ")"}]}]}], "}"}], ",", 
    RowBox[{"{", 
     RowBox[{"sf", ",", "0", ",", "1"}], "}"}], ",", 
    RowBox[{"{", 
     RowBox[{"h", ",", "0", ",", "1"}], "}"}], ",", 
    RowBox[{"PlotStyle", "\[Rule]", 
     RowBox[{"Opacity", "[", "0", "]"}]}]}], "]"}]}]}], "Input",
 CellChangeTimes->{{3.78325659328164*^9, 3.783256593829995*^9}, {
   3.783256975973332*^9, 3.783256988482939*^9}, {3.78325702198771*^9, 
   3.783257064586575*^9}, {3.783257117477017*^9, 3.7832572015504503`*^9}, {
   3.783257256529653*^9, 3.783257319022606*^9}, {3.7832573855015182`*^9, 
   3.783257536328903*^9}, {3.783257571544264*^9, 3.783257599005581*^9}, {
   3.7832577986718903`*^9, 3.783257799513914*^9}, 3.80967706574004*^9},
 CellLabel->
  "In[1510]:=",ExpressionUUID->"66571b6d-e08d-4aef-9b29-1c627f42dff1"],

Cell[BoxData[
 TemplateBox[{
  "Import","nffil",
   "\"File \\!\\(\\*RowBox[{\\\"\\\\\\\"per_hmap_b3_sm0_fm1_mm20_delta1_c1.\
csv\\\\\\\"\\\"}]\\) not found during \
\\!\\(\\*RowBox[{\\\"Import\\\"}]\\).\"",2,1510,41,32715148735184203895,
   "Local"},
  "MessageTemplate"]], "Message", "MSG",
 CellChangeTimes->{3.814197126416143*^9},
 CellLabel->
  "During evaluation of \
In[1510]:=",ExpressionUUID->"7a8bd29f-ad80-41f5-9ef3-b16cbab556a8"],

Cell[BoxData[
 TemplateBox[{
  "MatrixPlot","mat0",
   "\"Argument \\!\\(\\*RowBox[{\\\"$Failed\\\"}]\\) at position \
\\!\\(\\*RowBox[{\\\"1\\\"}]\\) is not a matrix.\"",2,1511,42,
   32715148735184203895,"Local"},
  "MessageTemplate"]], "Message", "MSG",
 CellChangeTimes->{3.814197126416808*^9},
 CellLabel->
  "During evaluation of \
In[1510]:=",ExpressionUUID->"079f19e2-f100-4fb6-9bd5-c8e5b9dc5480"],

Cell[BoxData[
 GraphicsBox[{GraphicsComplexBox[CompressedData["
1:eJzlmnl0zPf6x0eo2kZ/1LVMItq6YmjsEku1T2KI9l5iJJYrdGqp/DD2uhhL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    "], {{
      {Opacity[0], EdgeForm[None], 
       GraphicsGroupBox[{PolygonBox[CompressedData["
1:eJxNmHn8V1Max+8pSkkpVCIpJJKUSosoSaVFq2gPIaTs+26KjApZspMlJmMJ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          "]], PolygonBox[CompressedData["
1:eJwt08tLVVEUx/Gzo5l/gWTi1dQyLUNJDC9KoihBoIJIIKFQaJGm5ruHZqHV
7eGjgjTn4SMte6ulKXYdKImYNGnc3+DE72L/Bh8W556z9157rXVDtY3lDQeC
IMjFQdS5IPiNCvzj+SbxFTIRwgCWUYJSvEQEh3EC7XiIVOSgD39QjSPowhzy
EMa01ti3pzGKMRxHIt5iRu+myWeC2IZsZGEQf1GDBHTiCY7ijHLbxUXleANv
kIxDuIRNlGGe/R8Tm3BSe0xiSmeN836EeB/FyMA17OACYrCN28q9ClvO19Bq
V6lve5CkNVabu4hVjSzXR8rFcv6FNUR5Pme/45vOOosh3EIB0lCvM+2s/6xZ
If5U7axnH/FBvc1T7Z7pW6vhEn7orkXEYdxBoXryBZ+xx/t84j0sqhZWU5ud
bvXCZiiKdedzj8cGWlX786rFa5xSTewun5zfu0C9uIxj6onl8lV3t5xe4Dni
kIJanWF7RwI/K/16ZzOzgF7nZzFda1edn2XbY9b5GbLZydVsdCh3m5Gnzs9w
WDk9wHXnZ99m/h3ea631rJl4Rb2x/1ALGnS23fmq8zNTpZrabH13vvY2Y/uX
hmnu
          "]]}]}, 
      {Opacity[0], EdgeForm[None], 
       GraphicsGroupBox[{PolygonBox[CompressedData["
1:eJxNmHmYz1UUxmfOrVCRJWPfyVJkbI09MlqUGUI0USSppFWhon3TnlZt2je0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          "]], PolygonBox[CompressedData["
1:eJwtlOmLj1EUx5/fPUPW7Ip4MYksZc0WYwkx1pkh+zIzDGOY+Y2xjOxLIska
IgwisiQ1JSX5G8goa8g2sq9J+Hyd++LT+Z77PM957j3LzSxM55WFJEnyIAPq
W5K8ZGEtrIA8/ClQhS6FReglcAm9Gh7CQPw92OnQGt0IimOMpuhX2HWwCtri
/8bmw3p4BBWs3cBOE+g25s9OQRZ6KOxFF0BLdAGcRy+ExujlcBNdAp3Rv+Je
N2tP+O1gA/o0tEf/xRYGX/sBzcz3dhxamf9DsSvgM9TTP7CH4Sl0MY89L/hZ
SvGvxDP1RPeA7egdsBTdCTahL8N3aIK/ErsxeG6eQGXMUYn5O3qmHGuvZaxd
jXuui14G11UL+Ak5+CewW+ELNDCv3VF4Af3wd2GzYTB6EDxIJck2/GfQ3VzX
pnzvb9D7gp9B706Fk8G/GY6+ozzBfijH/4AtgmuQxk9gProahpjXVLVUDZ9D
H/yd2A7QG10bY2ntXoypWFnQF/0OeyD4Gcaan1FnOwLj8Meb61yYiJ4Ex9Cz
ob95DVW73TGXd+O/ldP3MEo5jWeYYN4T6gXF+AgZ5r1wCEabr0kfhDH42RY1
jETX6Nxxz5PNe0C1r4q9cjvmWj2jWt6PtVFNc/G7wZbgNa2JOa2OMZX7Xua9
pRoE9Cw4gz4LI9AzzHu9HOqg58K54DOj2ayEW8FndBj6dfDZSgev/Z+o1QMd
ef4t+GypZzPxP8Xe00wMMM+xcqs7QHvvaj4bOkORZgkuoBcH/3dD81nXHjS7
j4PfHZph3QVf47u6EzQ7xXAx+AwtUG+Z1/Z/TPM7SXeR3kmhZ8Ka4DXN117M
e0M5aG4+M5oVzVwL8xwqd/pmjmKb11I5fRtzqlyqp/8BHFSjSA==
          "]]}]}, {}, {}, {}, {}, {}}, {
      {RGBColor[0.368417, 0.506779, 0.709798], AbsoluteThickness[1.6], 
       LineBox[CompressedData["
1:eJwt0kk3FQAYx+GLikLoXtLSju9Q3wEfwdCcW0gUGYqSNM8TRZSSFBVlSlSX
SGhaWbdta+G551g857yr9/wXv6yCcH5JTCAQ6CWW6J1ODrvIo5gqWmmjjxEi
LLPCP/6zygZPkggSIp0MtpPJDhr4xSme0E09I0QHlXGfOn5yiMd84z0fGKaW
qyxzkHvc5SRLHKCTWd4xxFtquMIi+ynlDtX8YB8dXOIvl7nNCRbYyyNmGOQN
AxznIt/ZzRFa+cMFqujnNa/Yw0M+84VpInzlFpXMU0w7LfzmPDc5TDPnOMYL
XtJHEW1M8omPnGWK69ygiTNU8IxenlPIA8aZYIxGrhFmlNM8pZwuejjKHAXR
jVqopIhcdpJNiCDbSCOVFLaSTBKJbGEzCcSziY3Rvohb73UNscFaxQ==
        "]]}, 
      {RGBColor[0.880722, 0.611041, 0.142051], AbsoluteThickness[1.6], 
       LineBox[CompressedData["
1:eJwl0mdbjmEAx+H7qZQVRZkZFTLTEBkpNIWGTYooIpUR+ja+Ct/ASrR3IYSI
rNPhxXn8/td1PS+eF3dibXNlUygIgkc8DguCToeHOsOp8CDYyz62kcqA+ycs
s5/qTf2pK3Q5p9lPLjmMecvTX9qqv3Wc7XYa6TxzPqBvNEMzGbSfs9J+oU36
R1dpAmc4yyHeuj+o7zTQZtrtfJ3QLN3JsN3BGvu13tJIJu1EjeIes/nkbpdm
s5su52Tt0TsazZS9XhfwgIV8dXddb3CVBq6xgUaOc4JKvvvtHj3JD3ujbqLf
TtGtbGYLFZQz7W2R3qXXHiXGbiOWb85HtYxj9Dmv08tcoZZ66ljKF++3tVuT
dB73mc8SRtzP0bl8tg/rEUrpdF6rNVykmllcIo54wongo98WaQnFvHJerVVc
4DxhLGbI2zkNUcB75x1ayAe7RV9qDmn//jdxRDLt+x1njFFGGGaIQQbop49e
euimK/T/+/8Ln2Nc0Q==
        "]]}}}], {}},
  AspectRatio->1,
  Axes->{False, False},
  AxesLabel->{None, None},
  AxesOrigin->{Automatic, Automatic},
  DisplayFunction->Identity,
  Frame->{{True, True}, {True, True}},
  FrameLabel->{{None, None}, {None, None}},
  FrameTicks->{{Automatic, Automatic}, {Automatic, Automatic}},
  GridLinesStyle->Directive[
    GrayLevel[0.5, 0.4]],
  ImagePadding->All,
  Method->{
   "DefaultGraphicsInteraction" -> {
     "Version" -> 1.2, "TrackMousePosition" -> {True, False}, 
      "Effects" -> {
       "Highlight" -> {"ratio" -> 2}, "HighlightPoint" -> {"ratio" -> 2}, 
        "Droplines" -> {
         "freeformCursorMode" -> True, 
          "placement" -> {"x" -> "All", "y" -> "None"}}}}, "ScalingFunctions" -> 
    None, "TransparentPolygonMesh" -> True, "AxesInFront" -> True},
  PlotRange->{{0, 1}, {0, 1}},
  PlotRangeClipping->True,
  PlotRangePadding->{{
     Scaled[0.02], 
     Scaled[0.02]}, {
     Scaled[0.02], 
     Scaled[0.02]}},
  Ticks->{Automatic, Automatic}]], "Output",
 CellChangeTimes->{3.809677066558326*^9, 3.814005330007378*^9, 
  3.814196570583888*^9, 3.814197126681551*^9},
 CellLabel->
  "Out[1512]=",ExpressionUUID->"a5c51404-aaa5-435b-8aa5-dbddc3ab8e58"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[{
 RowBox[{
  RowBox[{"regionComparisonParameters", " ", "=", " ", 
   RowBox[{"{", 
    RowBox[{
     RowBox[{"sm", "\[Rule]", "0"}], ",", 
     RowBox[{"\[Delta]", "\[Rule]", "1"}], ",", 
     RowBox[{"c", "\[Rule]", "0.75"}]}], "}"}]}], 
  ";"}], "\[IndentingNewLine]", 
 RowBox[{
  RowBox[{"sm0fm20RPlot", " ", "=", " ", 
   RowBox[{"RegionPlot", "[", 
    RowBox[{
     RowBox[{"{", 
      RowBox[{
       RowBox[{"(", 
        RowBox[{"unstabConditions0i", "/.", "regionComparisonParameters"}], 
        ")"}], ",", 
       RowBox[{
        RowBox[{"(", 
         RowBox[{"unstabConditions0i", "/.", "regionComparisonParameters"}], 
         ")"}], "&&", 
        RowBox[{"(", 
         RowBox[{"unstabConditions1i", "/.", "regionComparisonParameters"}], 
         ")"}]}]}], "}"}], ",", 
     RowBox[{"{", 
      RowBox[{"sf", ",", "0", ",", "1"}], "}"}], ",", 
     RowBox[{"{", 
      RowBox[{"h", ",", "0", ",", "1"}], "}"}], ",", 
     RowBox[{"PlotStyle", "\[Rule]", 
      RowBox[{"Opacity", "[", "0", "]"}]}], ",", 
     RowBox[{"BoundaryStyle", "\[Rule]", "Dotted"}], ",", 
     RowBox[{"Epilog", "\[Rule]", 
      RowBox[{"{", 
       RowBox[{
        RowBox[{"Text", "[", 
         RowBox[{
          RowBox[{"Style", "[", 
           RowBox[{"\"\<drive invades\\nand fixes\>\"", ",", 
            RowBox[{"FontSize", "\[Rule]", "13"}], ",", "Black"}], "]"}], ",", 
          RowBox[{"{", 
           RowBox[{"0.03", ",", "0.56"}], "}"}], ",", 
          RowBox[{"{", 
           RowBox[{
            RowBox[{"-", "1"}], ",", "0"}], "}"}]}], "]"}], ",", 
        "\[IndentingNewLine]", 
        RowBox[{"Text", "[", 
         RowBox[{
          RowBox[{"Style", "[", 
           RowBox[{"\"\<drive doesn't invade\>\"", ",", 
            RowBox[{"FontSize", "\[Rule]", "15"}], ",", "Black"}], "]"}], ",", 
          RowBox[{"{", 
           RowBox[{"0.55", ",", "0.7"}], "}"}], ",", 
          RowBox[{"{", 
           RowBox[{
            RowBox[{"-", "1"}], ",", "0"}], "}"}]}], "]"}], ",", 
        RowBox[{"Text", "[", 
         RowBox[{
          RowBox[{"Style", "[", 
           RowBox[{"\"\<polymorphism\>\"", ",", 
            RowBox[{"FontSize", "\[Rule]", "15"}], ",", "Black"}], "]"}], ",", 
          RowBox[{"{", 
           RowBox[{"0.5", ",", "0.15"}], "}"}], ",", 
          RowBox[{"{", 
           RowBox[{
            RowBox[{"-", "1"}], ",", "0"}], "}"}]}], "]"}]}], "}"}]}]}], 
    "]"}]}], ";"}], "\[IndentingNewLine]", 
 RowBox[{
  RowBox[{"sm0fm2RPlot", " ", "=", " ", 
   RowBox[{"RegionPlot", "[", 
    RowBox[{
     RowBox[{"{", 
      RowBox[{
       RowBox[{"(", 
        RowBox[{"unstabConditions02", "/.", "regionComparisonParameters"}], 
        ")"}], ",", 
       RowBox[{
        RowBox[{"(", 
         RowBox[{"unstabConditions02", "/.", "regionComparisonParameters"}], 
         ")"}], "&&", 
        RowBox[{"(", 
         RowBox[{"unstabConditions12", "/.", "regionComparisonParameters"}], 
         ")"}]}]}], "}"}], ",", 
     RowBox[{"{", 
      RowBox[{"sf", ",", "0", ",", "1"}], "}"}], ",", 
     RowBox[{"{", 
      RowBox[{"h", ",", "0", ",", "1"}], "}"}], ",", 
     RowBox[{"PlotStyle", "\[Rule]", 
      RowBox[{"Opacity", "[", "0", "]"}]}], ",", 
     RowBox[{"BoundaryStyle", "\[Rule]", "Dashed"}]}], "]"}]}], 
  ";"}], "\[IndentingNewLine]", 
 RowBox[{
  RowBox[{"sm0fm1RPlot", " ", "=", " ", 
   RowBox[{"RegionPlot", "[", 
    RowBox[{
     RowBox[{"{", 
      RowBox[{
       RowBox[{"(", 
        RowBox[{"unstabConditions0", "/.", "regionComparisonParameters"}], 
        ")"}], ",", 
       RowBox[{
        RowBox[{"(", 
         RowBox[{"unstabConditions0", "/.", "regionComparisonParameters"}], 
         ")"}], "&&", 
        RowBox[{"(", 
         RowBox[{"unstabConditions1", "/.", "regionComparisonParameters"}], 
         ")"}]}]}], "}"}], ",", 
     RowBox[{"{", 
      RowBox[{"sf", ",", "0", ",", "1"}], "}"}], ",", 
     RowBox[{"{", 
      RowBox[{"h", ",", "0", ",", "1"}], "}"}], ",", 
     RowBox[{"PlotStyle", "\[Rule]", 
      RowBox[{"Opacity", "[", "0", "]"}]}]}], "]"}]}], 
  ";"}], "\[IndentingNewLine]", 
 RowBox[{
  RowBox[{"allregionsplot", " ", "=", " ", 
   RowBox[{"Labeled", "[", 
    RowBox[{
     RowBox[{"Legended", "[", 
      RowBox[{
       RowBox[{"Show", "[", 
        RowBox[{"{", 
         RowBox[{"sm0fm20RPlot", ",", "sm0fm2RPlot", ",", "sm0fm1RPlot"}], 
         "}"}], "]"}], ",", 
       RowBox[{"LineLegend", "[", 
        RowBox[{
         RowBox[{"{", 
          RowBox[{"Solid", ",", "Dashed", ",", "Dotted"}], "}"}], ",", 
         RowBox[{"{", 
          RowBox[{
          "\"\<\!\(\*SubscriptBox[\(\[Lambda]\), \(f\)]\) = 1\>\"", ",", 
           "\"\<\!\(\*SubscriptBox[\(\[Lambda]\), \(f\)]\) = 2\>\"", ",", 
           "\"\<\!\(\*SubscriptBox[\(\[Lambda]\), \(f\)]\) large\>\""}], 
          "}"}]}], "]"}]}], "]"}], ",", 
     RowBox[{"{", 
      RowBox[{
       RowBox[{
       "Text", "[", "\"\<\!\(\*SubscriptBox[\(s\), \(f\)]\)\>\"", "]"}], ",", 
       RowBox[{"Text", "[", "\"\<h\>\"", "]"}]}], "}"}], ",", 
     RowBox[{"{", 
      RowBox[{"Bottom", ",", " ", "Left"}], "}"}]}], "]"}]}], ";"}], "\n", 
 RowBox[{"Legended", "[", 
  RowBox[{
   RowBox[{"Labeled", "[", 
    RowBox[{
     RowBox[{"Show", "[", 
      RowBox[{"{", 
       RowBox[{"sm0fm20RPlot", ",", "sm0fm2RPlot", ",", "sm0fm1RPlot"}], 
       "}"}], "]"}], ",", 
     RowBox[{"{", 
      RowBox[{
       RowBox[{
       "Text", "[", "\"\<\!\(\*SubscriptBox[\(s\), \(f\)]\)\>\"", "]"}], ",", 
       RowBox[{"Text", "[", "\"\<h\>\"", "]"}]}], "}"}], ",", 
     RowBox[{"{", 
      RowBox[{"Bottom", ",", " ", "Left"}], "}"}]}], "]"}], ",", 
   RowBox[{"LineLegend", "[", 
    RowBox[{
     RowBox[{"{", 
      RowBox[{"Solid", ",", "Dashed", ",", "Dotted"}], "}"}], ",", 
     RowBox[{"{", 
      RowBox[{
      "\"\<\!\(\*SubscriptBox[\(\[Lambda]\), \(f\)]\) = 1\>\"", ",", 
       "\"\<\!\(\*SubscriptBox[\(\[Lambda]\), \(f\)]\) = 2\>\"", ",", 
       "\"\<\!\(\*SubscriptBox[\(\[Lambda]\), \(f\)]\) large\>\""}], "}"}]}], 
    "]"}]}], "]"}], "\n"}], "Input",
 CellChangeTimes->{{3.783419103244385*^9, 3.783419552861373*^9}, {
   3.783954459444705*^9, 3.7839545271015587`*^9}, {3.7839545746907873`*^9, 
   3.783954621635387*^9}, {3.783954710985065*^9, 3.7839547110843763`*^9}, {
   3.7839547484370747`*^9, 3.783954752372245*^9}, {3.783954783379318*^9, 
   3.7839547835798903`*^9}, {3.783955049331286*^9, 3.783955111318411*^9}, {
   3.7855709250136957`*^9, 3.785570925280274*^9}, {3.785571112776237*^9, 
   3.785571117262401*^9}, {3.785571313927066*^9, 3.785571316028226*^9}, {
   3.7855714786956873`*^9, 3.785571495973899*^9}, {3.785571653275477*^9, 
   3.785571702998274*^9}, {3.785571737198708*^9, 3.785571818626439*^9}, {
   3.7855718550450706`*^9, 3.78557190016722*^9}, 3.795268411926231*^9, 
   3.7952691583855047`*^9, {3.800015338171227*^9, 3.800015394223094*^9}},
 CellLabel->
  "In[1513]:=",ExpressionUUID->"4c9ca0f4-6cb8-42e6-bb4c-6dcdc4cfdea2"],

Cell[BoxData[
 TemplateBox[{TemplateBox[{
     GraphicsBox[{{
        GraphicsComplexBox[CompressedData["
1:eJzlWWtUk1fWThHvBj9RKxAUdLTGrqL1EmypukVspeogQau20mCZyqBBtNbq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         "], {{{
            EdgeForm[], 
            Directive[
             Opacity[0]], 
            GraphicsGroupBox[{
              PolygonBox[CompressedData["
1:eJxNmHv819Mdx8/JPbeIEFIh/FJSEqVURGQo5LYlKXILSUS1XKJcUrlkCplQ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               "]], 
              PolygonBox[CompressedData["
1:eJwt01dI1mEUx/HniBVJioZkVBRCCBIRYYQlUbTLhmllWxyULVdgi+guyIqW
DaOUBBu2F1lJKhURjRvv2ha0l0VJWNT3cM7Fh/N7+L/vf5znPIl5xZlFESGE
dESiREK4ggF4wnoddT0+kl9iEXkj2smvMY28F53kvygnr8UH8gtMJW/HT3IH
tpHPoDfesC6kPsBQPGW9gvoQKXjOegG1FrGIwxaMxxeufcc88gZ8I7fhNHkC
vgZ7xjDsEHt2HgrIF9Hfv3EixuBTsHuMIt/FIFxiPZJ6G0k4wjqNegfJqGM9
m3oY3RGF0biHwWji+lxqDXogGmNxH0PQyvXF1JPohQRMwjj8CNZj7W06/gTr
8T7ydAT8o0wW66n28jF2k/fgNzkfW8kV+Ex+hVPkerFr+sxd2CnWG/2P7tUU
/Aq2Z3XkY4jHI9a3qDPwjDyQulpsj+L8Gfqt2Yjxb67CLHRDJKq1j95b7dlK
sR5G+cw0Um94r/U/BzETXRCBVWI9jfaZbKE2+7v1pJ7Ace+tvnMTbvpexPgs
LRT7bazvbabPWpLvbZbY7Cb7LMwRm32dieGoFJv1MpSKnRk9K++Q4e+s79rV
373Kv12/IVWsh9q7imB7my/2X93jEeRD/t9K1muoZWJn432wvV8qlnUGjmK+
2N7qDF8gnxebhbfB9jJXLOueNpCv+r0TfRb1TOtZ1pnUs7jM303PpJ7F5WKz
rmdSZ1vPXFuwGc/BWfRBP1zDdbF764zotSXo67+5rHuPzcHOYJHYGdDZ1zNd
7PfUe+kZPkfdJNZrvcd+HBCbfb3Hf0kAm/k=
               "]]}]}, {
            EdgeForm[], 
            Directive[
             Opacity[0]], 
            GraphicsGroupBox[{
              PolygonBox[CompressedData["
1:eJxNmHd01UUQhZNZRNAgRYqAoQapoShgiNJLFEERpFlR7AUUREXsgL2gIBYQ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               "]], 
              PolygonBox[CompressedData["
1:eJwtlPlrz3Ecx7/ebz8QOUOOMHLkvhaTzZiZbWwaO9nWltyZ5IjkyhUiaqWV
tORPIFk5c9/NlauICGGOnOHx7PX64dHr9Xx/Pp/X5/1+He+k6pqCZSGRSBRB
c/iKqI6JxFHsdLgMndHzsclwATqi52HbwG0YgFaQ5ZCE3xeW4GfAVPxL2Lmw
E3LROZCJX6gY+Bexc6DSY93C1gSL2QhvoRhWwR/4DdOgBDJ5P0A5/g54pWfo
Xdg98AImordit0EP/Ctag3zohb6GTYeZ8AmqWDvi/yjCbwEV+AfgHcxA78Xu
g1PaO/ogdgz8VJxoz7LgEaSgN2A3wn344HtfF+ysrz0XOvMvj5HlZz4JI/l+
NbY/XIfnvtdF0JJnTdicYHu8Cj2jnU1nahvtG71bBXdhBGsrg8U8C+XoQ9hU
OA7t0dXYv83s3UbPvb7phL6HXROsJ+7AcNaKVGsYgj8UVijXMAV/MhTgb/dc
PcBmB8uZ/v1ZfRFsD2Xo9x6rVn2DPu29V+e5fILdHCynpegSmOBrg/EHQfdg
exgXrQbK/XrVCJ0Ma/GHwehoNVEttDYK3aAz+RlvKu+szQ7Ws5X4z9QDUA+T
0OnRzqoeS8VPgwX+zkMYiy71M3bRXmFLsHe+e4+qNzUD/aLNkGZnabBZ/IJN
CzaTFeinwWbzsPfqR+899ex532O9f6N/F0frXe2hNX6raLUug3MeU7GUwxMw
MNosK4fd8I9h82AhJND/gs2+atouWo+pt9QzXZX7aO9qxh57DZT7TZ7LrGh3
gXKqXHzzvSgnTV7zOu/pDvg3sIuD3Tk/gs24Zlt3SjbPY7TZ2K2Y+IWw33tM
d8FLf6Y7ocF7rNZr3CfaHaK7Q3fWLPUqpASL0TvaTGgWtAfdNRnRell3Tj5+
XrRe0p3wRrWJVjvNxBkYH+3sivkfZA6big==
               "]]}]}, {}, {}, {}, {}, {}}, {{
            Directive[
             RGBColor[0.368417, 0.506779, 0.709798], 
             AbsoluteThickness[1.6], 
             Dashing[{0, Small}]], 
            LineBox[CompressedData["
1:eJwt09dWDwAABvB/ysyIeAB33FtRNITMyiq7FA0kmWVkj4ZVlJEd0jBCsjIi
yXEjXoAXqBfo1zkufud8d9/N941Pz0veFhQIBLoYQH8ex0SiSCKDAoopoZIa
ammgmVbaaKeTn/yim9/84S+9xCmZSxFr+E4sx/nIfR5wl1ru0cYH7rCa26xi
K53EcIbTnCKVW0Qzh0JS+MYsjvGO99xkJTdYwRY6mMk+9rKcr0RxlLdcp4Zr
LCOXL0Syh90spZ0ZHCGbH+Swi2Q+M53DvOYNV0mimiVUkcgVIijmJS28oJVX
7GQxn5jGITbTRRY7KCCTTSyikstcYioHeUYzT8ngOdvJJ52NLOAiC6lgCgdo
Yh2NrCePeOppoIR5nOQxT9hAGgmUMp8yJnGWc5RzgfNMZj91POIha+kf82xO
0GNr/0gkkgmMJZwxjCaMUYxkBMMJZRhDGcJgBjGQEIL/f6UPFjlisw==
             "]]}, {
            Directive[
             RGBColor[0.880722, 0.611041, 0.142051], 
             AbsoluteThickness[1.6], 
             Dashing[{0, Small}]], 
            LineBox[CompressedData["
1:eJwl0+dbTgEYwOHzlswIWVH0RkR2yMpoWYkKlRXxmb/OjqxktvcQMjIyS0bF
7fLhvn7P85zz6b3eEy4/X3guFARBNbWGGvIjguACcyODYB5xPLDHa4/O14ea
oI90gT7TU/pcHxNjnso0LtoT9YlO11iqzJf4Q2Dv1UzNJosIckhiIWGSWUQe
fd5fpfW6Uhv0A0fN13StNutMncV983XGmiv1uH7RCTqeIxSTTpv7Bm3Xr5ww
39JJGs138xkd1K06mzmc5a7bNs3gqblM7+hk/ablOqCjdHPSflsnagmlbKHL
bZN26Ebt1M16U8dpIUUUcJhDjHjWyjHzDY3ST3paP+tvhunnIy2kebaedVTY
x2g+B9jPe7eD+lN/8ZZ3NLLavUnX6FWN1D28NO/WV/qDanLtu1hMjX2G7mUf
r+1vqGWJvU5T9LIu01SWsoLlXHEP6Q52sp0XbkPcY4q9TAv+/e6M+E/3000X
nXTQThuttNBME400UE9d6P938RdBcWkP
             "]]}}}], {}}, {
        GraphicsComplexBox[CompressedData["
1:eJzlWntUk1e2R7T4WEZFbLVQtFq10YpyqfGBj63kLu04FQODtD6aqYxFC4jW
sbbBttoRY9Varzat+IA21AcPMVqcqWK0PlrQCEgIEEBECAnytsbn2AqXWau/
k1k7K8vlUu+968o/Z/2+nO+cfX77+e3DwIiloe+4u7m5ze3o5vavsUt6eg/t
1JjJGH+6MDjkM3saAQef7bt0vT1L4Lf+K3+uTfaTwF7k2aq2GwS23U5VaMwF
Ak8J/uMgm6xI4F17m27INGaBb/+2+We1vVzgkLBR282KSoFLZi8erTFXCRzf
YXyiXm0ROCCjq4dNZhW46s3yWInNJvAXndLNMs1VgSceWjVFKa8XuGHe66lq
e4PACZ19e+u0TQJPy2yOMytaBL6uPFnT1naNXPHGR/AHDP6AwR8w+AMGf8Dg
Dxj8AYM/YPAHDP6AwR8w+AMGf8DgDxj8AYM/YPAHDP5c2RnnjY/gDxj8AYM/
YPAHDP6AwR8w+AMGf8DgDxj8AYM/YPAHDP6AwR8w+AMGf8Dgz5VfcjvjvPER
/AGDP2DwBwz+gMEfMPgDBn/A4A8Y/AGDP2DwBwz+gMEfMPgDBn/A4M9VHON+
ye2M88ZH8AcM/oDBHzD4AwZ/wOAPGPwBgz9g8AcM/oDBHzD4AwZ/wOAPGPy5
ivs8jnG/5HbGeeMj+AMGf8DgDxj8AYM/YPAHDP6AwR8w+AMGf8DgDxj8AYM/
V3mRx3ket7gfcrviPPERfAGDL2DwBQy+gMEXMPgCBl/A4AsYfLmqA3he43Ga
xx3uR9wu+Ln5iPMD4/zAOD8wzg+M8wPj/K7qGp6ned7hcZTHBW7nXI/8XHzE
+YBxPmCcz1UdxusKnid53OdxjPsltzuuBy43HyG/qzqR1z08j/O8xOMsjxvc
L7idcF65nLxu5XUYryt4nuRxn8cx7qfcbrmeed3M60Be1/A8zfMOj6M8TnC/
4XU6rzt5HcXrAp7neNzmcYn7Kf8u4HUur9t4HcLzKs8TPA7y7w5eR/O6kNc5
PG/zPMTjLP+u4XU6rzt5HcXrAp7n+HcSr/t5HcvrMl5n8LzJv7v4dwSvi3md
x+sWnof5dxz/LuF1Nq8beR30QtdwXbT0GsV+/f1FsyKfvDysK+zeDbQtOe6Y
Xl1Ipw8fuextaBT4h4NB32n1tdTyT59NElsxvTPXq7shuk7geW9ojQplDb0W
cmK+Ul5GXTsuD1RJbAJnpBcsluqqaU+KcqROWyH8Gri17V9/V8SIOASMeVgH
8QoY+0AOxA9XckFunAv+5uqc4AE8ibrGBW/gFTxDn5x36AOjq/dgx67kgB3i
d/ixq3MirrnijfPM9cL1iHWwLvaFHJCLnwNye4zeeF6muUbtRv6Cm1sBtfr3
26JXN9C7dy3NCqWJkstWnoiWtse9WfE/avUlNO1vxY3ehhoasH/IVrt3OSWU
dFpr97ZQXGt2hFx1WYx4jnl4D+tgXeyDfSEH5Fo2YskParuF8kJv/9PbcInO
rNkdrJRXkzRt+4Vo6RUx4jnm4T3oF+tixHPsI/K4es1ZhbJRyLX0rXXT5Kpm
YR/AIq8GhIxQSZpJ/3rvdVKdkeZ+duxtic2B+5dOelGntQoeIB/0jueYh30h
h6t9gbGPKzmgd6wr8tjveoDdQA68j/jH/RAY9q9YN32oxlxHffdkhqskxU5+
Dgz7v+m3L0yustHyX8OkhuhSl/OAMQ/vuZIDGHJArph+q6LsbQ79zugVulub
5vBfnHNMQfk39mwjtfoEjVVdcvgr8uSBiJffk88uppzBKxu93azCX1EnKAau
lOkSS8lfP2ClJMfhv902q+3eblXCf9ck3put7HtZ+O/bk73T9WnlQg/Pzeg2
TXOpROjpSP/B/oY2k9BjSdTNo/pCh/+Wv/uSUmV22DPOi+c7dk27Z61soKQ8
+knbo4jmTCxWSsOvkv+zU919+pnp2fc7vyF/0Ur6zzcskueUU9jzr6drR1no
xtrZmdrky2LEc8zDe1gH62If7As5INe2QMlG/YJmKpm/cYrujUI6eMjnI/Pe
ZvL6bWawxq+QEmdZWq1BdbS65yd7zWNKaNyNjKGaREe+631wnr9mUA15hpsa
ZcMrKCU4ZrpytiPfgVfwXBg+fou50JHvGuR/zlJPr6YP/H2/0/s74ir0VG1p
Gar7t/yHeXgPeoWeUwOG97KpHPkQckFO2AnsBnYFO5v5lV8HnyOO/IdzgwfY
Kex2y37zauleR/4Dj+AVPIJX2H2tcex1azv/Myo+7u7zulHkOYx4jnV3jlYu
l+915DfYXdTSgOnyvs1O9UlQl/093NrPgbj2ZZ18ksZ0SfACnsAreIZeoCf4
EfwKeoWe4YfwS9gF7ARyQc4pby3YJslxxLOyLB+FMqpKjIqeQ46r2+WIb4+m
EjfHiOdn686s12e0UNWCsCjlViPNf9amiza3UGTlplkalZGGeUX9Zk2pJ58+
0p12t2LaXtBjseRkPQUeuzVBd62IFi5TjlMF1pK26KZaQqVUObqi1ntXLQXv
OBYjN5vFc8zDc8zDutgH62IfyAU5IRfkRFyAH2ZX+ikl4ddoQvKcg+YAo8h/
hWHD1FJPo6ij3Aeq0swLTSKuJqnqwlSeJSIub7ufHqzaUSZGPM8sU3wgVZXR
5Jzj05VSm3gP64xZ5DPA1rF9HU+vEtmqOrEP9g33HDjY9oqJBtV39tK90ESW
0XtM6ruFtGxyl2cMi5vo3Z9ijDLfRgr6uFuhuspEH3qkehg+aqSMO6+ukO9y
5Gv3c6u2Sv5QSGGpmTetHi2k+8Ijw2w3UmDfAwfVV6/S9AH9x+s2OeSAXKOH
P7Nfu8tKv7hvXyt5uZwGH794py3ZSq8s+cMUjW+5OBfOGTnOs1m28hL1vhXx
jvTbGvps3okZKpMjvyPfx2XHdbYtrBHrYh+si33sr8XHSVPNFLHmi1rvnFoh
J+T2VI0dZVhXRKt85v5qDWoQPIAX8ABewAN4AQ/gBXYAu0Ce+TE9bZCtzaEH
6AV6gF6Qp7qfWPsnVXveyg4cpZQeca4vwg9/eVSb5qgvBpe+ssSuahHnwLlw
bvBwsOf9a2296sl4tH+DYk2x4Bm8K1afXCA9ZCNp0rogTVkpfTH5xGJpVjV5
xXXZbo6sFP48f3d6hrbdn5ueb/DT1VSQNKt1t36/haJrm7r6nCqk+GHhgcqT
TeI55uE9rIN1sQ/2hRyQE3I/bfebT9t95NN2f/i03fc9bfdz/9/uz/6v32/9
b98/Pe77oEe9n3nY+xIuB/oluorJz2607xQ4oU/xBZvsiPg+wHNgjI+7P4sR
3zHA+O550v1bV+u46ufy/i14QH4HRv4DRn4AL/C/B/VXMR9+8Lj7rXiP849z
Pqgfi/mcJ1f9Wd6H4Hp70v1W1NdYH/0wvP+w/VRX77nqr/J+KvTkPt5vg8Z8
SjwHxnuHrPev69VnxPyjV5eOGKPJEfOBuV9Bv8DgEfxBH9CT6L/8jk81VEfq
tLkU37o8cbN9o8A9fC0ZtbI9Qr5U72NTShUnBe428ZuzEtvPgm+M2B+8P+5+
MtaFHnh/GeeGHnEe+CvW4fnUVX8Y/GJ97M/9B78Lv/29nsB+kONJ95Ohb/gj
nsOesS/qO8gHfYr7EZYvEN8wH/YAuTCP2yXOgRHyIS4AI55AXujv1Zs9gm1L
Lji+714r6hijuCjk7z3gxVJ1rPP/RY3vXr9DPjX/ofvfST+cWCGfaXLqUz6u
fviJ1RdqFRFmp74rr2PqJw5vbeta7dQvv3EgMUQzs4BGpt9tU1eef+z97r5h
daGqV83Ux6+2Rra/yqn/HfEeReru59Knhpa7mi2HaU+nlD4hUXnkN65h6g33
f1DKxSv7pJvy6Vxb22F54inqti3hlvrkRXqhaYPH+qHZ9Fzksl8UTUbqdSX+
b9KBJvrL+fK5qsxC8jUdr1ccLX7k/rh+57od9vntcafzrQGGYZcf2C/f/k3I
UvmfS2j2XIOPYd2VJ94/lw/6+3l1QrtdnA0L0HSyUGp4ylZ7QRkdKHrlB3Xf
mkfur2dNvZ2tHlru1IcGnrOsQ4r0ap7wZ8ThmE1DqjWdf6R91QOUqrICp3oF
/XGFV9JK+SRHvQK/12tevSuTFj9y/zzBNjlSWW1yqkdwjoIzv92WlTrqEeg9
uHbNBnts5QP77Y0rn4mRTyoV85LfnnRcKy1zuk++fHlJlSzJ4tSff9j++Z9S
0lTytQU0rqLnAemuC1Q7foi3W4CRrJkj10s/bq8P9XltsvIC+mzQPT+f8PxH
7rff3v7JS7b1JrLOO2aXBZVS5PmXJusyTbQs96sPpadLH9iPnx2blmKOKqLq
ipGkOVFOoW2zvjMvN9OvKf+YL69rP7dvaHc3jzLKXdTxfcliy0P312OTDPuk
2/NF/WPVvfmaLTWfPgw1dUj2P00ld5I8QwZdpFsjLf1iFp6lFPf3v5J0ddQ/
L86KHGYbVUBBY+/EaWrOOfXrb/c8M1d12pGHjKctR/UVheT3fPxGyYQSuhM6
abDNx0Rjbw3bIGkreWB/P6FySp766xJhR19Omr7HXFtC7kcKX9aFV9FY3XMB
Pu654ntrRJ+hOXb1BVo1PFq+fvNy6vbMmLWqsgvkntGv6VbWl3TgnD5NOi+X
IvauTOqg2EuxdfKuMRNzKdsramzZt8l0I/7SDNukPJF/CxaMatZW5ZLvxYRm
3ZxD9P3qTz1icvOo7Ku/V+jqj9HNwIg6bcJFsoz+5K+6j3+mw6dDzmlPGYV/
zQz9z+uKFUYxvvX54lmazUbykeiz9AVG6n8qPUfbv5Byoz/yN+SZnvj9Qta8
ryfohji+/4yLxqeY04po/XMf/VG16BJltmlGGqqL6D/u9BlkuHWJ3L6Vntbm
F9PCexPyogMrH/p+Yp9yZ5eY3Y76D3XY96Nuv1z2djqtMN973m2Gcz34uPr/
7yQkvam6Y6KXOkcNNXxa5ui7ZMQe044sp/cW9kk1j3TUl2Pc485qPylyqi9x
f7Au8HO1tD0+j+jZ8hfdzPz/8fsDkzU6Rt6xlHp333pA+0G1033Cw94fBNum
NGrv5tHmIb+OvZF8nHrlTLwvCy6ks8M2VEbHFpFyyNJOPhXFZJFW+BrmVNJf
13pelwWUtftND5Nii4UMn8Zkq1eU0y99AoyKxBr6b2kfyVk=
         "], {{{
            EdgeForm[], 
            Directive[
             Opacity[0]], 
            GraphicsGroupBox[{
              PolygonBox[CompressedData["
1:eJxNmHv819Mdx89BQkUkkjsVKZcQFdKFpCgizdyzYaSwShYqIXLLJcIid5Ii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               "]], 
              PolygonBox[CompressedData["
1:eJwt091Pz2EYx/HvTWOszchCpQdFElqhVlFbY2sLqZbMc+mRKGEaThqVPIuQ
0DgQxbLCGlohmzlp49jf4NCskfe1+3Pw2n3v9/19r+u+7uv6xlU0FB+bEQTB
ZoTgmguCq/jHfhq32Heinf181mJ8wxZ08dsI62nMRiQ+Yhy/ebaA9SganI+V
xLoX25yPFYOX+IxNCMN2vMA6i8s7+1gLsRCxqEKL87HTsANfkaczdrHe1tns
nZu4oVjzWM/gAzKRhf0oQjji0IgmOKzCQZRiCeLRi36kqMZPzp8xVjUcwi5E
YQUqUKa7WY4x3YHVvhETOoPlzsE5VCIRoTiCA1iGv+Q7pZgWaxaaldNyzcED
7EE0VuMRHjpf+1o926q9/eceuvXuT+K3sZ5Asmq+g7vKNcnzVt2R3c1inEcd
VuoOL6jn1utF6MMzLMVc1bYba1TjffSo1l/EP8lajgTMxHH1IF49sdm8rtw2
owOaCZuFVFxWj7OU03p9FtnqucXeiQjlsF69Q4Z69h5PddYM3X2JarEePMYr
rMcU+S+yfkc+QnAJP1Dg/Ddhud4gXTmrnT+zndVm/DWGVOsGvMUTxbJ3RnEF
uarBZr0WfwI/8zXOf0Nh+m3Q+ZnK0Rm/6I7tbi2GfbuHdXb7hp87P5M2iymq
pV29zlctHZqNAs3GsM6arN7XOz+bNgP/AepLb80=
               "]]}]}, {
            EdgeForm[], 
            Directive[
             Opacity[0]], 
            GraphicsGroupBox[{
              PolygonBox[CompressedData["
1:eJxNmHmU12UVh2ferwuCKCii4gAzgIAjKLHvAoOoIEgoi0RZttiulrbaYmWI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               "]], 
              PolygonBox[CompressedData["
1:eJwtlPWLlVEQhr97xu5u0bWwk10D107UhVVEDIwFxcK127VQRLHFjlUEQfwD
FDvA7m7FRlEQBDHQ52XODw/zzv3ud+6ZmXduWk7uwCkhSZJsKADvSf7BDFgC
dSxJ6sE6dD94Ab9gKuyC63AOcmEulOC7xSEDPRm+QV/yg8QB0BXdBbLREyAd
3RZ2oAdDE3Qz2ITOgr+QTX6EOAJqoKvDavQaeAg/4lnboDPPrhJnw17oRn6N
OAf2wR94C4thOvyG15AH0+AUtOad7boLFDKvQXdfAB/gFsyHkdCb56VhIXo/
3IAL8ax58BOewaLYo058t6D5WXugj3kP1bsD8DKeoXeHQiWeV4QV6JXQCH2H
2B9yoBz5FeIsWBbvdglmxjumeG7mtQ6Dx9CGfCJxEJyGmua91MzaoZ/GWWjG
WeTPiUPgcJzFRdUdfCZNzWtSLZpZbXOPyBtrdU/yJ/G3DgXvVU/ID96zFuYz
06y2QC90D3Pv5EfvHCeWCe6hjuT1YT16NzQw96i8qc9uQ0PyMbFH8soD4ubg
nrkMVflsVTyzCvoecWPwz1qZe0pe2hq8tydgeezx2ViTakmHm5BmXqt6olnd
JW4IPrMK5OVhbDyjvfnMNeudUAvdwVyrxxnmnpKXtBOfg++kdnEUBPOZapbD
g//W1+gF/aZq/Q7jYs2PYk2qRZ6W9z9G72oHPsGXePZSKMt33wXfNf0HNCe/
H3wXdWY18pbms9IO10UfJY4O/h9xPnpGXskM7v1X0cvagWI8K2reu/FwDCqb
ezsFRdCFzXdPPdPunoFJwXdYu30yvqsdl5e7m++2PK1dfhN7kxe9kGnuFXmi
FLqk+d10ZmLeA9WunWlsfqdU9MR/XpucLQ==
               "]]}]}, {}, {}, {}, {}, {}}, {{
            Directive[
             RGBColor[0.368417, 0.506779, 0.709798], 
             AbsoluteThickness[1.6], 
             Dashing[{Small, Small}]], 
            LineBox[CompressedData["
1:eJwt0klXDQAYx+EbkdCtpA9gl3PyDfINUDdTZGwyRJklFE0qcwgZu6gMGQoZ
wpWpJKJFu1ZWrYQP4LnnWDznvLt38f/NySvNLokJBAK9TCJ6pzKX+YQopIx6
mgnTSQ8R+hlmlDF+Ms4vJvjNH/6S4ME8VlLAAV5zkEu08IFjvGI/vZRzlK+s
oIN22njJC3LIp4zn7OMiF2hghEZucZO9PGM5eeymhz2cp5kj/KCeG4R5yhOW
sZ5ddPGYbnZSxxBLWcd2vrODbdTyhSWc5DgniI4xTClbqWGQxaylmG9sYRPV
DBBiDe95wENaecRmPpPNWc5RxScyWU0f7+hkA9e5z0b6yaKJM1TykYWs4g23
ucM1IrzlEIcp5C5F3GMRpzhNBVe4ygJyuUw6QQo0k0UGacwmhVkkk0QiwWhb
zGQG04lnGnFMZQqxTP7f+j/0I2D6
             "]]}, {
            Directive[
             RGBColor[0.880722, 0.611041, 0.142051], 
             AbsoluteThickness[1.6], 
             Dashing[{Small, Small}]], 
            LineBox[CompressedData["
1:eJwl0llXTQEcxuF9ColGU2VIISRSyRwhc5kSDVetlut8B5+KMgspNGieCyFD
KkMSeVounvV7/2efi7PO2qlV14urQ0EQ1NButFEZFgQ32RgeBJvYwGbSyOMQ
W2nxnS3aqun6Wa/qF33FSnsVCdxyb9M2TdQkntk1LLBr9YyO6yJdyGG2s4MM
2j3L1Akt1UntYI29lnXccy/VKH7ax/UkJ4jlFHvYy256fWef/tIyndEBUuxU
muzHxNlP9ZL+0bfE23+1XOd0lI9UuMf0Ex8I3MXU282844r7vb5hlovuJxqu
IcIYcZfoax3mNxfcdfqSQS67h7SPac65H+oLTtv7OcB6DtLv8/P6SKM1hh92
od7X5xyzC9hFtztZc+mxi/SBRuoSpuyzekcbybd3ksVqjnCUb559p5Nsd5fm
6F2N0MV8tZfrCm7bDSyzb+i1+d9B7Pz/yYT3dIRhhhhkgH766KWHbrropCP0
/x3/B+9KYfU=
             "]]}}}], {}}, {
        GraphicsComplexBox[CompressedData["
1:eJzlmnl0zPf6x0eo2kZ/1LVMItq6YmjsEku1T2KI9l5iJJYrdGqp/DD2uhhL
qca0qpYytoswRC3BKErp0FhTI8gi882+TxaSaCdqZ256Tt+f7z3PnDl+zq/3
3Hvd/PM5rznf5Xne38/neZ7P88nr46aHTfBSKBSGmgrFr2Od2NiG5uAp72C8
cOWPQz537idw6Plm0z9znhL8/pprEY6AC4JfpUZPjU6bYMfdfVqTlCg4KPTP
bzgCbgjesru8KsAkCb77eOVFozND8JChnTZK2hzB9mETu5ukPMFRNXptsxoL
BHc9WLe2I6BIcN5fMqYpHQ7Bq2rFSgGmEsF9Di8I0mnKBN8cNXCf0XlT8KaX
Wza2mMsFhxytmC9pKwX/rDtT6HLdJk+68RH6gaEfGPqBoR8Y+oGhHxj6gaEf
GPqBoR8Y+oGhHxj6gaEfGPqBoR8Y+oGhn6d5xnXjI/QDQz8w9ANDPzD0A0M/
MPQDQz8w9ANDPzD0A0M/MPQDQz8w9ANDPzD087Qu+TzjuvER+oGhHxj6gaEf
GPqBoR8Y+oGhHxj6gaEfGPqBoR8Y+oGhHxj6gaGfpzjG1yWfZ1w3PkI/MPQD
Qz8w9ANDPzD0A0M/MPQDQz8w9ANDPzD0A0M/MPQDQz9PcZ/HMb4u+TzjuvER
+oGhHxj6gaEfGPqBoR8Y+oGhHxj6gaEfGPqBoR8Y+oGhn6c8yeM+j2N8XfJ5
xnXjI/QDQz8w9ANDPzD0A0M/MPQDQz8w9ANDPzD0A0M/MPTzVFfwPMnjPo9j
fF3yecZ14yP0A0M/MPQDQz8w9ANDPzD0A0M/MPQDQz9PdRavG3ge5HGdxym+
7vg84rrwEfqAoQ8Y+oChDxj6gKEPGPp4qit5ncTzPs9jPC7zOMPXDZ8H3G8+
wn8w/AfDfzD891Qn87qP1zE8L/M8w+MmjwN8XvPvyP3iI/wDwz9PdT6vW3kd
xusKnid53OdxjK9LPg/5d+F+8H0Ir6t5ncjrHp7HeV7icZbHDb5O+LzhOvN9
Eq/7eR3L6zJeZ/C8yfMAj2t8HfN5zfdtfB/C62peJ/K6h+dxnpd4nOVxhe8b
+T6I1/W8TuV1F68jeF7kcZ7HNb5v5fswvq/gdTKv+3gdw/MyzzN8n8z3fXwf
w+tyXmfyuonXATyv8X0532fyfRPfB/C6ltdpvO7g+36+j+X7Mr7P4HUzrwN5
XTP9/WUhGkMFVT7wXqF0pJJP3eEWvfq24FdrF812qm7Su0NOj9Zp0unEob67
zNZiitmr62gxZ1HdmrN6G5QOeur69S+XDsYmTlRb8sWI33Ed7sNz8Fy8B+/1
ZBd0wu+YJ3gO1hHeg7jJ7cLvuA734Tl4Lt6D90Z8fnKM0lFBFwZH/WC22mln
+pzTenUJzX96aZzGkE2b7LU+daoKxIjfcR3uw3POfnMsW2W7Jd4DFnnPuOS8
VneLZj0aqrbp02hCxKsNbPpSYScY63Z+p9yfXa4Suhp294HKlkmjRpiTtLpC
8R3AiNMz/KeeMDoL6NySraE6Tb7H68C4Dvd5sgMMO2CXJz/B8BN+Yx3ivqiu
Q/wNSll/zBMw4hbsUmdaItSWMmq1p81XTlWGmGdg5Cn47Zv29msWcxHdbN9d
ZdPniXm3duf8k1ZjsmCso9rdv7gcYLpNzWKODjcoU93mORh5506Hr4dqDA5S
7994Ra/OFe+FHXgv7ODPASN/YX4lfNpwuaSV1xN0gJ/QATqKePqbbtDVk59g
+Am/YQfsgh2wC37DfvgNHRCHDoxrO1MzTLYrKiKinU2S89KGvoFdLC9livdN
mfygStUxl5okj/zOOKVArGfouG5jQrChUK4Df25bP1HbT65X5rzuv8EaVUyT
F499SXEnyy0fRxaPCtINK6OJD5tO1gxLd8tPQ4dsvltUVU5fDr4TI0XZRXyY
37lzuLqygmrtmBkjXbPTu/GDGtqq1/3F9Nm1FPOrv9Oky/vNhgIx4ndch/vw
nMnDRln0bW5T2NhGK5S3Umlony4OVwc5Ph9cM6uP7stb5PtKL5UjPo1cC2Y3
sHUvJfut1zpZumXR3tApA3TD5Pi8JcYxQT28kCruN3ni+jpfjPj9h5/iD5kD
88T3OjXaN85YO0fEazwX7zkVfPeS0U9eV7ADdvk8eRBmaCeJdQo/4Be+O+YB
/IKf0HXpualHrFHyem8046MxunZynYDvtL7jgpdtL91yyysXry08afaT6yx8
96Uh211FkpwvMC8wT+A3dMA8w7wzLv9lgnpQoVt+we+Yp5i3+A74Lngv7IDd
8OPoB4tbmhIrqLH5fn+Tw06Jn6keFUVW12FD20/SZKbT7rGT9xnzb1Jy/09G
6yakU2RAv3YmvxJyhvivcC7MJvvxYwN0/g5KONpX6fDNEyN+x3W4D8/Fe/Bc
vAd2wK6dOssm88lyOhjeqIb3BYnemnc22vxeGVV92WKp0pVBPWvkL3C2KSZX
TJ9vzMnV+1CfwrcMfYrolwvBk9U5eWLE77gO9+E5eC7eg/ciLhnLmqxQlqWK
Eb//T8bAVo6SVOpgmGaQ2t2mmt10VwJWVNLkB4/ftI20i99xHfKQb+J7VvOa
NCpq0u2Y/n/LSbEizmKtk0YhXvbxyjfKaXdIJ4V3Pzkff32VkvR3MmnJ1O7N
LR3LaPpSn6YO70w6PHfkaPWZUlrdQxWp2Sjn42mmJUGG7fnUNaZhE1txdZ7V
jfez+OXTnfPrZyl/LKQZ50anaoPzxXUiH/92He7DdbgPdsAu2AG7YAfsgp/w
G37Cb/gJv6EbdPQtbdxTV7tS5JEjzeeVusIqqLVPcg/LAImu1Ru+SKqS62rk
mQifcWGGxRL9bdGBlACpjGqV+36q1GdQ/vbm8dph8j4EeeujIr98rSmDupfc
a23yk/P3pnVxWQFHi+jD4uN27aI8mpIfP1zTT96HIo/e94+2a1fIca33/qgA
U3Wciwx57br2spx3m+e88r2xUN6/mq9bW9uq8w7y0KqSPG+HJO8PvNaUDFbb
bpNy9LTxOkMq1Tf+sNC5x0HDImzetmW5NCazsFL1SgH17DKwtaW4QNgJuz/p
cP0X1R9zSLtlbZjm62K6nz1gj/VGFvXa3spkPVgidIFOnePf6m4ZnE45G+r/
wXL/ptAZusMO2IXn4j2wA3bBTtj933ae+t92/vmin0++6OeHL/r53ot+/vai
nYf9p51P/budF/2zz2t+7/OSZ51P8PvQT9jUJPWKI+CYYK9eHZabpDix38Pv
4Gf1AT31JzHyPqOnfiV+5/s33r/kfTQ8B3Yj/4ARr2EnX0+e+pcYoTPvZ/K+
5rP6m3gOtx/f53n7mdaBjZepLUlUpZ+zLdB0QvSDJt0vqNDqUsR96A+i3xOy
NPWWylbo1t96Vn8Relqy3vnDF86/iTgKho7flUz3DzTFC53BsCtyXdB6neay
uH7ahiPXJe01cT2YfwfECz6voCPWFxj7EvgJvfapTgalac8Ie+r12X5e6bgo
OO5mfqTFnCC+A/qgz9sfFv2+33T/Z/eLMeI6vBfft7rI9lEo5P0E9MB6wfzh
+edf1e/FesF8xXzm/vH4hvoa/r19uttDk3RY+IF5C39xHezGdfiuYPiJ8Xn7
0Xg+vjfmCb738/arcR3mA56DefCsfjD8gH3wg8+PqKeztq10fiHH+d/qQ9yH
9cLXKfzDiPt5vkF8xvOwHnmcxQg94C/0wPvhBxjfA3bgPXXq+SkC97r/39qY
rPByS6Nv6FLP0i42V7Lwt72/X6Tudorb/1kFJmZsd15KosSv5jZS+NiF/7yf
fnrxlWLtOLne4f113Oe16ePOlqmZbv1269LYCN10uX7g/Xc8t/z4wlDdsPz/
cz8efZqBP551arcm0uPhhWOnNt9Hu/ZtzdUnp9B7YY+U3rMT3frp37Yd3t8U
LNEri+81dWzPeGZ/fd2IkuaK4iTyuh5uU7Y4R8fvOC6ZVyXTNM3mRlOi4une
6siBpuAUivk+Y586M4GO3z44XletG8ZdTYc3VkSnUs/s/Jre/naKi44o1S6x
05wDH7tUe9Oeuz9/U3NKa5ieRl7Hx/Y2/TWbvmw208+2KJ2svQYGmi7mPrN/
37xjzBapev45Szt0NVXn40PBy8dr6meTVK/BCWO7wmf291cnmCzmc7kUt3XP
OLW/w63fX1s9J9zmui7W45mZVW0dNZJE/dY853E9bz+5vhH1TsKfVxj6x1No
uK6TbfoNt3ru6dnQcMMD++/ez997pFmavkmaWKcXnjiXKZPSxHrFfQuzTzV0
vJf7/+7/b/2lzTum8Ey53nJOGqEbkOV2npw8vNdqKbnguc8Lfu9+f9850U7t
4ep9ceKS4F3+B6hlYLvL5pAkmnpJtXJI4EnqPaGzUa1MottXK82BDb6lxe9/
+pHGlUK7S2a6Ahom0crY8be0P96gqp+6jjTsSqG1oa3/ZFp+g4p2huXrM5Ld
+v8LHiYcts6QqNneg3UUTTPdzgP6+rbT6f6aRn96fcsmaWm22/nA3NVPmzq2
ZNCGBofrO7zy3c4LKru2XeVMyKR9IS0H6XwKaF7z1Ja293PI98rxc/pORW7n
Cbzfij6wqD+f0U/n/eAbc/tVacPk+nPjtpZV2onV8byn170PqnbQ0ikFl8xr
kkX9mbikQxvHumTqFpNa87ND8STtHhErFSfTFz1HqIZ0tdGY3NH7pdkpov5U
HW28X4pOoe8SWsc5P77udr7hNaHhEespu5jXVY/ax0h37fTzw48PWPOq89CT
2FDD5nQRL5QPFFHKW+mUl2JeJ4Xn0bejvo+W6mVQtHbSOumjPLfzqEOWxjUC
30ikbe2H1EyKNdLrn9cs1ndNpIU/SZNr7lhD35wd8qM5Lknk05H5a1oojEnk
PfXo2sADp2lsj6Bl6vVJ1eu5RlDad2foUUyzYMvLydR5fZ3Jth7nSf9mx7re
/il0QTtvm+btBLqTVN7bskze783d9vBdU5dUOrF/yoe6shvUzDVnt9TbTuf3
D2iouCn9x53f3Mk4f9L4D+c3qJPq7H5pi/ViIe1d0badZb28n+y0sGmG1pRD
bwQv2mMeUUQrQ6J9bMpcKl9+6p7rUJHb+R3q686N1dfrvpn5jv3lFuuUH8j1
ddCHV2dq7sn/j4065Fv/VXul+TdeuPMbdcNWZQFHqvef8/ftNR8sFHH8LwGT
FjvbFFHp49fGaaRseT/aZtUM9SC5vkedGvcwKyfgk0Iq7V+ZoI+U+ywXwncW
6R8mUgtz1YLP9h+l4qCvHgSMu0Hllu4F+ohkatRszgCT0k7zPl/bw6KX3M6H
Zh4bVBVQKdHVzU86WqrzBj8velw8dKSuSybFBlZEW1fnu50fbV7YaLuzZzIN
ztqx2pB0gR6ea3XIeieNdhQk9Ddsz3E772lw9NWVzolZNHhhToVqV4Hb+c/f
Aflo3/A=
         "], {{{
            EdgeForm[], 
            Directive[
             Opacity[0]], 
            GraphicsGroupBox[{
              PolygonBox[CompressedData["
1:eJxNmHn8V1Max+8pSkkpVCIpJJKUSosoSaVFq2gPIaTs+26KjApZspMlJmMJ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               "]], 
              PolygonBox[CompressedData["
1:eJwt08tLVVEUx/Gzo5l/gWTi1dQyLUNJDC9KoihBoIJIIKFQaJGm5ruHZqHV
7eGjgjTn4SMte6ulKXYdKImYNGnc3+DE72L/Bh8W556z9157rXVDtY3lDQeC
IMjFQdS5IPiNCvzj+SbxFTIRwgCWUYJSvEQEh3EC7XiIVOSgD39QjSPowhzy
EMa01ti3pzGKMRxHIt5iRu+myWeC2IZsZGEQf1GDBHTiCY7ijHLbxUXleANv
kIxDuIRNlGGe/R8Tm3BSe0xiSmeN836EeB/FyMA17OACYrCN28q9ClvO19Bq
V6lve5CkNVabu4hVjSzXR8rFcv6FNUR5Pme/45vOOosh3EIB0lCvM+2s/6xZ
If5U7axnH/FBvc1T7Z7pW6vhEn7orkXEYdxBoXryBZ+xx/t84j0sqhZWU5ud
bvXCZiiKdedzj8cGWlX786rFa5xSTewun5zfu0C9uIxj6onl8lV3t5xe4Dni
kIJanWF7RwI/K/16ZzOzgF7nZzFda1edn2XbY9b5GbLZydVsdCh3m5Gnzs9w
WDk9wHXnZ99m/h3ea631rJl4Rb2x/1ALGnS23fmq8zNTpZrabH13vvY2Y/uX
hmnu
               "]]}]}, {
            EdgeForm[], 
            Directive[
             Opacity[0]], 
            GraphicsGroupBox[{
              PolygonBox[CompressedData["
1:eJxNmHmYz1UUxmfOrVCRJWPfyVJkbI09MlqUGUI0USSppFWhon3TnlZt2je0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               "]], 
              PolygonBox[CompressedData["
1:eJwtlOmLj1EUx5/fPUPW7Ip4MYksZc0WYwkx1pkh+zIzDGOY+Y2xjOxLIska
IgwisiQ1JSX5G8goa8g2sq9J+Hyd++LT+Z77PM957j3LzSxM55WFJEnyIAPq
W5K8ZGEtrIA8/ClQhS6FReglcAm9Gh7CQPw92OnQGt0IimOMpuhX2HWwCtri
/8bmw3p4BBWs3cBOE+g25s9OQRZ6KOxFF0BLdAGcRy+ExujlcBNdAp3Rv+Je
N2tP+O1gA/o0tEf/xRYGX/sBzcz3dhxamf9DsSvgM9TTP7CH4Sl0MY89L/hZ
SvGvxDP1RPeA7egdsBTdCTahL8N3aIK/ErsxeG6eQGXMUYn5O3qmHGuvZaxd
jXuui14G11UL+Ak5+CewW+ELNDCv3VF4Af3wd2GzYTB6EDxIJck2/GfQ3VzX
pnzvb9D7gp9B706Fk8G/GY6+ozzBfijH/4AtgmuQxk9gProahpjXVLVUDZ9D
H/yd2A7QG10bY2ntXoypWFnQF/0OeyD4Gcaan1FnOwLj8Meb61yYiJ4Ex9Cz
ob95DVW73TGXd+O/ldP3MEo5jWeYYN4T6gXF+AgZ5r1wCEabr0kfhDH42RY1
jETX6Nxxz5PNe0C1r4q9cjvmWj2jWt6PtVFNc/G7wZbgNa2JOa2OMZX7Xua9
pRoE9Cw4gz4LI9AzzHu9HOqg58K54DOj2ayEW8FndBj6dfDZSgev/Z+o1QMd
ef4t+GypZzPxP8Xe00wMMM+xcqs7QHvvaj4bOkORZgkuoBcH/3dD81nXHjS7
j4PfHZph3QVf47u6EzQ7xXAx+AwtUG+Z1/Z/TPM7SXeR3kmhZ8Ka4DXN117M
e0M5aG4+M5oVzVwL8xwqd/pmjmKb11I5fRtzqlyqp/8BHFSjSA==
               "]]}]}, {}, {}, {}, {}, {}}, {{
            Directive[
             RGBColor[0.368417, 0.506779, 0.709798], 
             AbsoluteThickness[1.6]], 
            LineBox[CompressedData["
1:eJwt0kk3FQAYx+GLikLoXtLSju9Q3wEfwdCcW0gUGYqSNM8TRZSSFBVlSlSX
SGhaWbdta+G551g857yr9/wXv6yCcH5JTCAQ6CWW6J1ODrvIo5gqWmmjjxEi
LLPCP/6zygZPkggSIp0MtpPJDhr4xSme0E09I0QHlXGfOn5yiMd84z0fGKaW
qyxzkHvc5SRLHKCTWd4xxFtquMIi+ynlDtX8YB8dXOIvl7nNCRbYyyNmGOQN
AxznIt/ZzRFa+cMFqujnNa/Yw0M+84VpInzlFpXMU0w7LfzmPDc5TDPnOMYL
XtJHEW1M8omPnGWK69ygiTNU8IxenlPIA8aZYIxGrhFmlNM8pZwuejjKHAXR
jVqopIhcdpJNiCDbSCOVFLaSTBKJbGEzCcSziY3Rvohb73UNscFaxQ==
             "]]}, {
            Directive[
             RGBColor[0.880722, 0.611041, 0.142051], 
             AbsoluteThickness[1.6]], 
            LineBox[CompressedData["
1:eJwl0mdbjmEAx+H7qZQVRZkZFTLTEBkpNIWGTYooIpUR+ja+Ct/ASrR3IYSI
rNPhxXn8/td1PS+eF3dibXNlUygIgkc8DguCToeHOsOp8CDYyz62kcqA+ycs
s5/qTf2pK3Q5p9lPLjmMecvTX9qqv3Wc7XYa6TxzPqBvNEMzGbSfs9J+oU36
R1dpAmc4yyHeuj+o7zTQZtrtfJ3QLN3JsN3BGvu13tJIJu1EjeIes/nkbpdm
s5su52Tt0TsazZS9XhfwgIV8dXddb3CVBq6xgUaOc4JKvvvtHj3JD3ujbqLf
TtGtbGYLFZQz7W2R3qXXHiXGbiOWb85HtYxj9Dmv08tcoZZ66ljKF++3tVuT
dB73mc8SRtzP0bl8tg/rEUrpdF6rNVykmllcIo54wongo98WaQnFvHJerVVc
4DxhLGbI2zkNUcB75x1ayAe7RV9qDmn//jdxRDLt+x1njFFGGGaIQQbop49e
euimK/T/+/8Ln2Nc0Q==
             "]]}}}], {}}}, {
      DisplayFunction -> Identity, PlotRangePadding -> {{
          Scaled[0.02], 
          Scaled[0.02]}, {
          Scaled[0.02], 
          Scaled[0.02]}}, PlotRangeClipping -> True, ImagePadding -> All, 
       DisplayFunction -> Identity, AspectRatio -> 1, Axes -> {False, False}, 
       AxesLabel -> {None, None}, AxesOrigin -> {Automatic, Automatic}, 
       DisplayFunction :> Identity, Epilog -> {
         InsetBox[
          FormBox[
           StyleBox["\"drive invades\\nand fixes\"", FontSize -> 13, 
            GrayLevel[0], StripOnInput -> False], TraditionalForm], {0.03, 
          0.56}, {-1, 0}], 
         InsetBox[
          FormBox[
           StyleBox["\"drive doesn't invade\"", FontSize -> 15, 
            GrayLevel[0], StripOnInput -> False], TraditionalForm], {0.55, 
          0.7}, {-1, 0}], 
         InsetBox[
          FormBox[
           StyleBox["\"polymorphism\"", FontSize -> 15, 
            GrayLevel[0], StripOnInput -> False], TraditionalForm], {0.5, 
          0.15}, {-1, 0}]}, Frame -> {{True, True}, {True, True}}, 
       FrameLabel -> {{None, None}, {None, None}}, 
       FrameTicks -> {{Automatic, Automatic}, {Automatic, Automatic}}, 
       GridLinesStyle -> Directive[
         GrayLevel[0.5, 0.4]], 
       Method -> {
        "DefaultGraphicsInteraction" -> {
          "Version" -> 1.2, "TrackMousePosition" -> {True, False}, 
           "Effects" -> {
            "Highlight" -> {"ratio" -> 2}, "HighlightPoint" -> {"ratio" -> 2},
              "Droplines" -> {
              "freeformCursorMode" -> True, 
               "placement" -> {"x" -> "All", "y" -> "None"}}}}, 
         "ScalingFunctions" -> None, "TransparentPolygonMesh" -> True, 
         "AxesInFront" -> True}, PlotRange -> {{0, 1}, {0, 1}}, 
       PlotRangeClipping -> True, PlotRangePadding -> {
         Scaled[0.02], 
         Scaled[0.02]}, Ticks -> {Automatic, Automatic}}], 
     InterpretationBox[
      Cell[
       BoxData[
        FormBox["\"\\!\\(\\*SubscriptBox[\\(s\\), \\(f\\)]\\)\"", TextForm]], 
       "InlineText"], 
      Text["\!\(\*SubscriptBox[\(s\), \(f\)]\)"]], 
     InterpretationBox[
      Cell[
       BoxData[
        FormBox["\"h\"", TextForm]], "InlineText"], 
      Text["h"]]}, "Labeled", DisplayFunction -> (GridBox[{{
        ItemBox[#3, DefaultBaseStyle -> "LabeledLabel"], 
        TagBox[
         ItemBox[
          PaneBox[
           TagBox[#, "SkipImageSizeLevel"], Alignment -> {Center, Baseline}, 
           BaselinePosition -> Baseline], DefaultBaseStyle -> "Labeled"], 
         "SkipImageSizeLevel"]}, {"", 
        ItemBox[#2, DefaultBaseStyle -> "LabeledLabel"]}}, 
      GridBoxAlignment -> {"Columns" -> {{Center}}, "Rows" -> {{Center}}}, 
      AutoDelete -> False, 
      GridBoxItemSize -> {
       "Columns" -> {{Automatic}}, "Rows" -> {{Automatic}}}, 
      BaselinePosition -> {1, 2}]& ), 
    InterpretationFunction -> (RowBox[{"Labeled", "[", 
       RowBox[{#, ",", 
         RowBox[{"{", 
           RowBox[{#2, ",", #3}], "}"}], ",", 
         RowBox[{"{", 
           RowBox[{"Bottom", ",", "Left"}], "}"}]}], "]"}]& )],
   TemplateBox[{
    "\"\\!\\(\\*SubscriptBox[\\(\[Lambda]\\), \\(f\\)]\\) = 1\"", 
     "\"\\!\\(\\*SubscriptBox[\\(\[Lambda]\\), \\(f\\)]\\) = 2\"", 
     "\"\\!\\(\\*SubscriptBox[\\(\[Lambda]\\), \\(f\\)]\\) large\""}, 
    "LineLegend", DisplayFunction -> (StyleBox[
      StyleBox[
       PaneBox[
        TagBox[
         GridBox[{{
            TagBox[
             GridBox[{{
                GraphicsBox[{{
                   Directive[
                    EdgeForm[
                    Directive[
                    Opacity[0.3], 
                    GrayLevel[0]]], 
                    PointSize[0.5], 
                    AbsoluteThickness[1.6]], {
                    LineBox[{{0, 10}, {40, 10}}]}}, {
                   Directive[
                    EdgeForm[
                    Directive[
                    Opacity[0.3], 
                    GrayLevel[0]]], 
                    PointSize[0.5], 
                    AbsoluteThickness[1.6]], {}}}, AspectRatio -> Full, 
                 ImageSize -> {40, 10}, PlotRangePadding -> None, 
                 ImagePadding -> Automatic, 
                 BaselinePosition -> (Scaled[0.1] -> Baseline)], #}, {
                GraphicsBox[{{
                   Directive[
                    EdgeForm[
                    Directive[
                    Opacity[0.3], 
                    GrayLevel[0]]], 
                    PointSize[0.5], 
                    AbsoluteThickness[1.6], 
                    Dashing[{Small, Small}]], {
                    LineBox[{{0, 10}, {40, 10}}]}}, {
                   Directive[
                    EdgeForm[
                    Directive[
                    Opacity[0.3], 
                    GrayLevel[0]]], 
                    PointSize[0.5], 
                    AbsoluteThickness[1.6], 
                    Dashing[{Small, Small}]], {}}}, AspectRatio -> Full, 
                 ImageSize -> {40, 10}, PlotRangePadding -> None, 
                 ImagePadding -> Automatic, 
                 BaselinePosition -> (Scaled[0.1] -> Baseline)], #2}, {
                GraphicsBox[{{
                   Directive[
                    EdgeForm[
                    Directive[
                    Opacity[0.3], 
                    GrayLevel[0]]], 
                    PointSize[0.5], 
                    AbsoluteThickness[1.6], 
                    Dashing[{0, Small}]], {
                    LineBox[{{0, 10}, {40, 10}}]}}, {
                   Directive[
                    EdgeForm[
                    Directive[
                    Opacity[0.3], 
                    GrayLevel[0]]], 
                    PointSize[0.5], 
                    AbsoluteThickness[1.6], 
                    Dashing[{0, Small}]], {}}}, AspectRatio -> Full, 
                 ImageSize -> {40, 10}, PlotRangePadding -> None, 
                 ImagePadding -> Automatic, 
                 BaselinePosition -> (Scaled[0.1] -> Baseline)], #3}}, 
              GridBoxAlignment -> {
               "Columns" -> {Center, Left}, "Rows" -> {{Baseline}}}, 
              AutoDelete -> False, 
              GridBoxDividers -> {
               "Columns" -> {{False}}, "Rows" -> {{False}}}, 
              GridBoxItemSize -> {"Columns" -> {{All}}, "Rows" -> {{All}}}, 
              GridBoxSpacings -> {"Columns" -> {{0.5}}, "Rows" -> {{0.8}}}], 
             "Grid"]}}, 
          GridBoxAlignment -> {"Columns" -> {{Left}}, "Rows" -> {{Top}}}, 
          AutoDelete -> False, 
          GridBoxItemSize -> {
           "Columns" -> {{Automatic}}, "Rows" -> {{Automatic}}}, 
          GridBoxSpacings -> {"Columns" -> {{1}}, "Rows" -> {{0}}}], "Grid"], 
        Alignment -> Left, AppearanceElements -> None, 
        ImageMargins -> {{5, 5}, {5, 5}}, ImageSizeAction -> "ResizeToFit"], 
       LineIndent -> 0, StripOnInput -> False], {FontFamily -> "Arial"}, 
      Background -> Automatic, StripOnInput -> False]& ), 
    InterpretationFunction :> (RowBox[{"LineLegend", "[", 
       RowBox[{
         RowBox[{"{", 
           RowBox[{"Solid", ",", 
             RowBox[{"Dashing", "[", 
               RowBox[{"{", 
                 RowBox[{"Small", ",", "Small"}], "}"}], "]"}], ",", 
             RowBox[{"Dashing", "[", 
               RowBox[{"{", 
                 RowBox[{"0", ",", "Small"}], "}"}], "]"}]}], "}"}], ",", 
         RowBox[{"{", 
           RowBox[{#, ",", #2, ",", #3}], "}"}]}], "]"}]& ), Editable -> 
    True]},
  "Legended",
  DisplayFunction->(GridBox[{{
      TagBox[
       ItemBox[
        PaneBox[
         TagBox[#, "SkipImageSizeLevel"], Alignment -> {Center, Baseline}, 
         BaselinePosition -> Baseline], DefaultBaseStyle -> "Labeled"], 
       "SkipImageSizeLevel"], 
      ItemBox[#2, DefaultBaseStyle -> "LabeledLabel"]}}, 
    GridBoxAlignment -> {"Columns" -> {{Center}}, "Rows" -> {{Center}}}, 
    AutoDelete -> False, GridBoxItemSize -> Automatic, 
    BaselinePosition -> {1, 1}]& ),
  Editable->True,
  InterpretationFunction->(RowBox[{"Legended", "[", 
     RowBox[{#, ",", 
       RowBox[{"Placed", "[", 
         RowBox[{#2, ",", "After"}], "]"}]}], "]"}]& )]], "Output",
 CellChangeTimes->{3.80001540826814*^9, 3.80026911807964*^9, 
  3.809677028181974*^9, 3.814005331203319*^9, 3.8141965712930937`*^9, 
  3.8141971275068417`*^9},
 CellLabel->
  "Out[1518]=",ExpressionUUID->"a32d642c-a66e-427c-bf8b-7eb833dd75f9"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"\[IndentingNewLine]", 
  RowBox[{
   RowBox[{
    RowBox[{"params1", " ", "=", " ", 
     RowBox[{"{", 
      RowBox[{
       RowBox[{"sm", "\[Rule]", "0"}], ",", 
       RowBox[{"\[Delta]", "\[Rule]", "1"}], ",", 
       RowBox[{"c", "\[Rule]", "1"}]}], "}"}]}], ";"}], "\[IndentingNewLine]", 
   RowBox[{
    RowBox[{"params08", " ", "=", " ", 
     RowBox[{"{", 
      RowBox[{
       RowBox[{"sm", "\[Rule]", "0"}], ",", 
       RowBox[{"\[Delta]", "\[Rule]", "1"}], ",", 
       RowBox[{"c", "\[Rule]", "0.8"}]}], "}"}]}], ";"}], 
   "\[IndentingNewLine]", 
   RowBox[{
    RowBox[{"params06", " ", "=", " ", 
     RowBox[{"{", 
      RowBox[{
       RowBox[{"sm", "\[Rule]", "0"}], ",", 
       RowBox[{"\[Delta]", "\[Rule]", "1"}], ",", 
       RowBox[{"c", "\[Rule]", "0.6"}]}], "}"}]}], ";"}], 
   "\[IndentingNewLine]", 
   RowBox[{
    RowBox[{"params05", " ", "=", " ", 
     RowBox[{"{", 
      RowBox[{
       RowBox[{"sm", "\[Rule]", "0"}], ",", 
       RowBox[{"\[Delta]", "\[Rule]", "1"}], ",", 
       RowBox[{"c", "\[Rule]", "0.5"}]}], "}"}]}], ";"}], 
   "\[IndentingNewLine]", 
   RowBox[{"boundaries1", " ", "=", " ", 
    RowBox[{"Show", "[", 
     RowBox[{
      RowBox[{"Labeled", "[", 
       RowBox[{
        RowBox[{"Show", "[", 
         RowBox[{"{", 
          RowBox[{
           RowBox[{"RegionPlot", "[", 
            RowBox[{
             RowBox[{"{", 
              RowBox[{
               RowBox[{"(", 
                RowBox[{"unstabConditions02", "/.", "params1"}], ")"}], ",", 
               RowBox[{
                RowBox[{"(", 
                 RowBox[{"unstabConditions02", "/.", "params1"}], ")"}], "&&", 
                RowBox[{"(", 
                 RowBox[{"unstabConditions12", "/.", "params1"}], ")"}]}]}], 
              "}"}], ",", 
             RowBox[{"{", 
              RowBox[{"sf", ",", "0", ",", "1"}], "}"}], ",", 
             RowBox[{"{", 
              RowBox[{"h", ",", "0", ",", "1"}], "}"}], ",", 
             RowBox[{"PlotStyle", "\[Rule]", 
              RowBox[{"Opacity", "[", "0", "]"}]}], ",", 
             RowBox[{"Epilog", "\[Rule]", 
              RowBox[{"{", 
               RowBox[{
                RowBox[{"Text", "[", 
                 RowBox[{
                  RowBox[{"Style", "[", 
                   RowBox[{
                   "\"\<\!\(\*StyleBox[\"c\",FontSlant->\"Italic\"]\)\!\(\*\
StyleBox[\" \",FontSlant->\"Italic\"]\)= 1\>\"", ",", 
                    RowBox[{"FontSize", "\[Rule]", "13"}], ",", "Black", ",", 
                    RowBox[{"Background", "\[Rule]", "White"}]}], "]"}], ",", 
                  RowBox[{"{", 
                   RowBox[{"0.63", ",", "0.55"}], "}"}], ",", 
                  RowBox[{"{", 
                   RowBox[{
                    RowBox[{"-", "1"}], ",", "0"}], "}"}]}], "]"}], ",", 
                RowBox[{"Text", "[", 
                 RowBox[{
                  RowBox[{"Style", "[", 
                   RowBox[{
                   "\"\<\!\(\*StyleBox[\"c\",FontSlant->\"Italic\"]\)\!\(\*\
StyleBox[\" \",FontSlant->\"Italic\"]\)= 0.8\>\"", ",", 
                    RowBox[{"FontSize", "\[Rule]", "13"}], ",", "Black", ",", 
                    RowBox[{"Background", "\[Rule]", "White"}]}], "]"}], ",", 
                  RowBox[{"{", 
                   RowBox[{"0.63", ",", "0.465"}], "}"}], ",", 
                  RowBox[{"{", 
                   RowBox[{
                    RowBox[{"-", "1"}], ",", "0"}], "}"}]}], "]"}], ",", 
                RowBox[{"Text", "[", 
                 RowBox[{
                  RowBox[{"Style", "[", 
                   RowBox[{
                   "\"\<\!\(\*StyleBox[\"c\",FontSlant->\"Italic\"]\)\!\(\*\
StyleBox[\" \",FontSlant->\"Italic\"]\)= 0.6\>\"", ",", 
                    RowBox[{"FontSize", "\[Rule]", "13"}], ",", "Black", ",", 
                    RowBox[{"Background", "\[Rule]", "White"}]}], "]"}], ",", 
                  RowBox[{"{", 
                   RowBox[{"0.63", ",", "0.34"}], "}"}], ",", 
                  RowBox[{"{", 
                   RowBox[{
                    RowBox[{"-", "1"}], ",", "0"}], "}"}]}], "]"}], ",", 
                RowBox[{"Text", "[", 
                 RowBox[{
                  RowBox[{"Style", "[", 
                   RowBox[{
                   "\"\<\!\(\*StyleBox[\"c\",FontSlant->\"Italic\"]\)\!\(\*\
StyleBox[\" \",FontSlant->\"Italic\"]\)= 0.5\>\"", ",", 
                    RowBox[{"FontSize", "\[Rule]", "13"}], ",", "Black", ",", 
                    RowBox[{"Background", "\[Rule]", "White"}]}], "]"}], ",", 
                  RowBox[{"{", 
                   RowBox[{"0.63", ",", "0.25"}], "}"}], ",", 
                  RowBox[{"{", 
                   RowBox[{
                    RowBox[{"-", "1"}], ",", "0"}], "}"}]}], "]"}], ",", 
                "\[IndentingNewLine]", 
                RowBox[{"Rotate", "[", 
                 RowBox[{
                  RowBox[{"Text", "[", 
                   RowBox[{
                    RowBox[{"Style", "[", 
                    RowBox[{"\"\<drive invades and fixes\>\"", ",", 
                    RowBox[{"FontSize", "\[Rule]", "15"}], ",", "Black"}], 
                    "]"}], ",", 
                    RowBox[{"{", 
                    RowBox[{
                    RowBox[{"-", "0.15"}], ",", "0.47"}], "}"}], ",", 
                    RowBox[{"{", 
                    RowBox[{
                    RowBox[{"-", "1"}], ",", "0"}], "}"}]}], "]"}], ",", 
                  RowBox[{"90", " ", "Degree"}]}], "]"}], ",", 
                "\[IndentingNewLine]", 
                RowBox[{"Text", "[", 
                 RowBox[{
                  RowBox[{"Style", "[", 
                   RowBox[{"\"\<drive doesn't invade\>\"", ",", 
                    RowBox[{"FontSize", "\[Rule]", "15"}], ",", "Black"}], 
                   "]"}], ",", 
                  RowBox[{"{", 
                   RowBox[{"0.55", ",", "0.75"}], "}"}], ",", 
                  RowBox[{"{", 
                   RowBox[{
                    RowBox[{"-", "1"}], ",", "0"}], "}"}]}], "]"}], ",", 
                RowBox[{"Text", "[", 
                 RowBox[{
                  RowBox[{"Style", "[", 
                   RowBox[{"\"\<polymorphism\>\"", ",", 
                    RowBox[{"FontSize", "\[Rule]", "15"}], ",", "Black"}], 
                   "]"}], ",", 
                  RowBox[{"{", 
                   RowBox[{"0.47", ",", "0.1"}], "}"}], ",", 
                  RowBox[{"{", 
                   RowBox[{
                    RowBox[{"-", "1"}], ",", "0"}], "}"}]}], "]"}]}], 
               "}"}]}]}], "]"}], ",", 
           RowBox[{"RegionPlot", "[", 
            RowBox[{
             RowBox[{"{", 
              RowBox[{
               RowBox[{"(", 
                RowBox[{"unstabConditions02", "/.", "params08"}], ")"}], ",", 
               RowBox[{
                RowBox[{"(", 
                 RowBox[{"unstabConditions02", "/.", "params08"}], ")"}], "&&", 
                RowBox[{"(", 
                 RowBox[{"unstabConditions12", "/.", "params08"}], ")"}]}]}], 
              "}"}], ",", 
             RowBox[{"{", 
              RowBox[{"sf", ",", "0", ",", "1"}], "}"}], ",", 
             RowBox[{"{", 
              RowBox[{"h", ",", "0", ",", "1"}], "}"}], ",", 
             RowBox[{"PlotStyle", "\[Rule]", 
              RowBox[{"Opacity", "[", "0", "]"}]}]}], "]"}], ",", 
           RowBox[{"RegionPlot", "[", 
            RowBox[{
             RowBox[{"{", 
              RowBox[{
               RowBox[{"(", 
                RowBox[{"unstabConditions02", "/.", "params06"}], ")"}], ",", 
               RowBox[{
                RowBox[{"(", 
                 RowBox[{"unstabConditions02", "/.", "params06"}], ")"}], "&&", 
                RowBox[{"(", 
                 RowBox[{"unstabConditions12", "/.", "params06"}], ")"}]}]}], 
              "}"}], ",", 
             RowBox[{"{", 
              RowBox[{"sf", ",", "0", ",", "1"}], "}"}], ",", 
             RowBox[{"{", 
              RowBox[{"h", ",", "0", ",", "1"}], "}"}], ",", 
             RowBox[{"PlotStyle", "\[Rule]", 
              RowBox[{"Opacity", "[", "0", "]"}]}]}], "]"}], ",", 
           RowBox[{"RegionPlot", "[", 
            RowBox[{
             RowBox[{"{", 
              RowBox[{
               RowBox[{"(", 
                RowBox[{"unstabConditions02", "/.", "params05"}], ")"}], ",", 
               RowBox[{
                RowBox[{"(", 
                 RowBox[{"unstabConditions02", "/.", "params05"}], ")"}], "&&", 
                RowBox[{"(", 
                 RowBox[{"unstabConditions12", "/.", "params05"}], ")"}]}]}], 
              "}"}], ",", 
             RowBox[{"{", 
              RowBox[{"sf", ",", "0", ",", "1"}], "}"}], ",", 
             RowBox[{"{", 
              RowBox[{"h", ",", "0", ",", "1"}], "}"}], ",", 
             RowBox[{"PlotStyle", "\[Rule]", 
              RowBox[{"Opacity", "[", "0", "]"}]}]}], "]"}]}], "}"}], "]"}], 
        ",", 
        RowBox[{"{", 
         RowBox[{
          RowBox[{
          "Text", "[", 
           "\"\<female selection \
\!\(\*StyleBox[\"(\",FontSlant->\"Italic\"]\)\!\(\*StyleBox[SubscriptBox[\"s\"\
, \"f\"],FontSlant->\"Italic\"]\))\>\"", "]"}], ",", 
          RowBox[{"Rotate", "[", 
           RowBox[{
            RowBox[{
            "Text", "[", 
             "\"\<dominance (\!\(\*StyleBox[\"h\",FontSlant->\"Italic\"]\))\>\
\"", "]"}], ",", 
            RowBox[{"90", " ", "Degree"}]}], "]"}], ",", 
          RowBox[{"Row", "[", 
           RowBox[{"{", 
            RowBox[{
             RowBox[{"Text", "[", 
              RowBox[{"Style", "[", 
               RowBox[{"\"\<A\>\"", ",", "Bold", ",", 
                RowBox[{"FontSize", "\[Rule]", "20"}]}], "]"}], " ", "]"}], 
             ",", 
             RowBox[{"Text", "[", 
              RowBox[{
              "Style", "[", 
               "\"\<   \!\(\*StyleBox[SubscriptBox[\"\[Lambda]\", RowBox[{\"f\
\", \"  \"}]],FontWeight->\"Bold\",FontSlant->\"Italic\"]\)\!\(\*StyleBox[\"=\
\",FontWeight->\"Bold\"]\)\!\(\*StyleBox[\" \
\",FontWeight->\"Bold\"]\)\!\(\*StyleBox[\"2\",FontWeight->\"Bold\"]\)\>\"", 
               "]"}], "]"}]}], "}"}], "]"}]}], "}"}], ",", 
        RowBox[{"{", 
         RowBox[{"Bottom", ",", "Left", ",", 
          RowBox[{"{", 
           RowBox[{"Top", ",", "Left"}], "}"}]}], "}"}]}], "]"}], ",", 
      RowBox[{"ImageSize", "\[Rule]", "Large"}]}], "]"}]}], 
   "\[IndentingNewLine]", 
   RowBox[{"boundaries2", " ", "=", 
    RowBox[{"Show", "[", 
     RowBox[{
      RowBox[{"Labeled", "[", 
       RowBox[{
        RowBox[{"Show", "[", 
         RowBox[{"{", 
          RowBox[{
           RowBox[{"RegionPlot", "[", 
            RowBox[{
             RowBox[{"{", 
              RowBox[{
               RowBox[{"(", 
                RowBox[{"unstabConditions0i", "/.", "params1"}], ")"}], ",", 
               RowBox[{
                RowBox[{"(", 
                 RowBox[{"unstabConditions0i", "/.", "params1"}], ")"}], "&&", 
                RowBox[{"(", 
                 RowBox[{"unstabConditions1i", "/.", "params1"}], ")"}]}]}], 
              "}"}], ",", 
             RowBox[{"{", 
              RowBox[{"sf", ",", "0", ",", "1"}], "}"}], ",", 
             RowBox[{"{", 
              RowBox[{"h", ",", "0", ",", "1"}], "}"}], ",", 
             RowBox[{"PlotStyle", "\[Rule]", 
              RowBox[{"Opacity", "[", "0", "]"}]}], ",", 
             RowBox[{"Epilog", "\[Rule]", 
              RowBox[{"{", 
               RowBox[{
                RowBox[{"Text", "[", 
                 RowBox[{
                  RowBox[{"Style", "[", 
                   RowBox[{
                   "\"\<\!\(\*StyleBox[\"c\",FontSlant->\"Italic\"]\)\!\(\*\
StyleBox[\" \",FontSlant->\"Italic\"]\)= 1\>\"", ",", 
                    RowBox[{"FontSize", "\[Rule]", "13"}], ",", "Black", ",", 
                    RowBox[{"Background", "\[Rule]", "White"}]}], "]"}], ",", 
                  RowBox[{"{", 
                   RowBox[{"0.63", ",", "0.55"}], "}"}], ",", 
                  RowBox[{"{", 
                   RowBox[{
                    RowBox[{"-", "1"}], ",", "0"}], "}"}]}], "]"}], ",", 
                RowBox[{"Text", "[", 
                 RowBox[{
                  RowBox[{"Style", "[", 
                   RowBox[{
                   "\"\<\!\(\*StyleBox[\"c\",FontSlant->\"Italic\"]\)\!\(\*\
StyleBox[\" \",FontSlant->\"Italic\"]\)= 0.8\>\"", ",", 
                    RowBox[{"FontSize", "\[Rule]", "13"}], ",", "Black", ",", 
                    RowBox[{"Background", "\[Rule]", "White"}]}], "]"}], ",", 
                  RowBox[{"{", 
                   RowBox[{"0.63", ",", "0.385"}], "}"}], ",", 
                  RowBox[{"{", 
                   RowBox[{
                    RowBox[{"-", "1"}], ",", "0"}], "}"}]}], "]"}], ",", 
                RowBox[{"Text", "[", 
                 RowBox[{
                  RowBox[{"Style", "[", 
                   RowBox[{
                   "\"\<\!\(\*StyleBox[\"c\",FontSlant->\"Italic\"]\)\!\(\*\
StyleBox[\" \",FontSlant->\"Italic\"]\)= 0.6\>\"", ",", 
                    RowBox[{"FontSize", "\[Rule]", "13"}], ",", "Black", ",", 
                    RowBox[{"Background", "\[Rule]", "White"}]}], "]"}], ",", 
                  RowBox[{"{", 
                   RowBox[{"0.63", ",", "0.165"}], "}"}], ",", 
                  RowBox[{"{", 
                   RowBox[{
                    RowBox[{"-", "1"}], ",", "0"}], "}"}]}], "]"}], ",", 
                "\[IndentingNewLine]", 
                RowBox[{"Text", "[", 
                 RowBox[{
                  RowBox[{"Style", "[", 
                   RowBox[{
                   "\"\<\!\(\*StyleBox[\"c\",FontSlant->\"Italic\"]\)\!\(\*\
StyleBox[\" \",FontSlant->\"Italic\"]\)= 0.5\>\"", ",", 
                    RowBox[{"FontSize", "\[Rule]", "13"}], ",", "Black", ",", 
                    RowBox[{"Background", "\[Rule]", "White"}]}], "]"}], ",", 
                  RowBox[{"{", 
                   RowBox[{"0.83", ",", "0.035"}], "}"}], ",", 
                  RowBox[{"{", 
                   RowBox[{
                    RowBox[{"-", "1"}], ",", "0"}], "}"}]}], "]"}], ",", 
                "\[IndentingNewLine]", 
                RowBox[{"Rotate", "[", 
                 RowBox[{
                  RowBox[{"Text", "[", 
                   RowBox[{
                    RowBox[{"Style", "[", 
                    RowBox[{"\"\<drive invades and fixes\>\"", ",", 
                    RowBox[{"FontSize", "\[Rule]", "15"}], ",", "Black"}], 
                    "]"}], ",", 
                    RowBox[{"{", 
                    RowBox[{
                    RowBox[{"-", "0.18"}], ",", "0.5"}], "}"}], ",", 
                    RowBox[{"{", 
                    RowBox[{
                    RowBox[{"-", "1"}], ",", "0"}], "}"}]}], "]"}], ",", 
                  RowBox[{"90", " ", "Degree"}]}], "]"}], ",", 
                "\[IndentingNewLine]", 
                RowBox[{"Text", "[", 
                 RowBox[{
                  RowBox[{"Style", "[", 
                   RowBox[{"\"\<drive doesn't invade\>\"", ",", 
                    RowBox[{"FontSize", "\[Rule]", "15"}], ",", "Black"}], 
                   "]"}], ",", 
                  RowBox[{"{", 
                   RowBox[{"0.55", ",", "0.75"}], "}"}], ",", 
                  RowBox[{"{", 
                   RowBox[{
                    RowBox[{"-", "1"}], ",", "0"}], "}"}]}], "]"}], ",", 
                RowBox[{"Text", "[", 
                 RowBox[{
                  RowBox[{"Style", "[", 
                   RowBox[{"\"\<polymorphism\>\"", ",", 
                    RowBox[{"FontSize", "\[Rule]", "15"}], ",", "Black"}], 
                   "]"}], ",", 
                  RowBox[{"{", 
                   RowBox[{"0.47", ",", "0.07"}], "}"}], ",", 
                  RowBox[{"{", 
                   RowBox[{
                    RowBox[{"-", "1"}], ",", "0"}], "}"}]}], "]"}]}], 
               "}"}]}]}], "]"}], ",", "\[IndentingNewLine]", 
           RowBox[{"RegionPlot", "[", 
            RowBox[{
             RowBox[{"{", 
              RowBox[{
               RowBox[{"(", 
                RowBox[{"unstabConditions0i", "/.", "params08"}], ")"}], ",", 
               RowBox[{
                RowBox[{"(", 
                 RowBox[{"unstabConditions0i", "/.", "params08"}], ")"}], "&&", 
                RowBox[{"(", 
                 RowBox[{"unstabConditions1i", "/.", "params08"}], ")"}]}]}], 
              "}"}], ",", 
             RowBox[{"{", 
              RowBox[{"sf", ",", "0", ",", "1"}], "}"}], ",", 
             RowBox[{"{", 
              RowBox[{"h", ",", "0", ",", "1"}], "}"}], ",", 
             RowBox[{"PlotStyle", "\[Rule]", 
              RowBox[{"Opacity", "[", "0", "]"}]}]}], "]"}], ",", 
           RowBox[{"RegionPlot", "[", 
            RowBox[{
             RowBox[{"{", 
              RowBox[{
               RowBox[{"(", 
                RowBox[{"unstabConditions0i", "/.", "params06"}], ")"}], ",", 
               RowBox[{
                RowBox[{"(", 
                 RowBox[{"unstabConditions0i", "/.", "params06"}], ")"}], "&&", 
                RowBox[{"(", 
                 RowBox[{"unstabConditions1i", "/.", "params06"}], ")"}]}]}], 
              "}"}], ",", 
             RowBox[{"{", 
              RowBox[{"sf", ",", "0", ",", "1"}], "}"}], ",", 
             RowBox[{"{", 
              RowBox[{"h", ",", "0", ",", "1"}], "}"}], ",", 
             RowBox[{"PlotStyle", "\[Rule]", 
              RowBox[{"Opacity", "[", "0", "]"}]}]}], "]"}], ",", 
           RowBox[{"RegionPlot", "[", 
            RowBox[{
             RowBox[{"{", 
              RowBox[{
               RowBox[{"(", 
                RowBox[{"unstabConditions0i", "/.", "params05"}], ")"}], ",", 
               RowBox[{
                RowBox[{"(", 
                 RowBox[{"unstabConditions0i", "/.", "params05"}], ")"}], "&&", 
                RowBox[{"(", 
                 RowBox[{"unstabConditions1i", "/.", "params05"}], ")"}]}]}], 
              "}"}], ",", 
             RowBox[{"{", 
              RowBox[{"sf", ",", "0", ",", "1"}], "}"}], ",", 
             RowBox[{"{", 
              RowBox[{"h", ",", "0", ",", "1"}], "}"}], ",", 
             RowBox[{"PlotStyle", "\[Rule]", 
              RowBox[{"Opacity", "[", "0", "]"}]}]}], "]"}]}], "}"}], "]"}], 
        ",", 
        RowBox[{"{", 
         RowBox[{
          RowBox[{
          "Text", "[", 
           "\"\<female selection \
\!\(\*StyleBox[\"(\",FontSlant->\"Italic\"]\)\!\(\*SubscriptBox[StyleBox[\"s\"\
,FontSlant->\"Italic\"], \"f\"]\))\>\"", "]"}], ",", 
          RowBox[{"Rotate", "[", 
           RowBox[{
            RowBox[{
            "Text", "[", 
             "\"\<dominance (\!\(\*StyleBox[\"h\",FontSlant->\"Italic\"]\))\>\
\"", "]"}], ",", 
            RowBox[{"90", " ", "Degree"}]}], "]"}], ",", 
          RowBox[{"Row", "[", 
           RowBox[{"{", 
            RowBox[{
             RowBox[{"Text", "[", 
              RowBox[{"Style", "[", 
               RowBox[{"\"\<B\>\"", ",", "Bold", ",", 
                RowBox[{"FontSize", "\[Rule]", "20"}]}], "]"}], " ", "]"}], 
             ",", 
             RowBox[{"Text", "[", 
              RowBox[{
              "Style", "[", 
               "\"\<   \!\(\*StyleBox[SubscriptBox[\"\[Lambda]\", RowBox[{\"f\
\", \"  \"}]],FontWeight->\"Bold\",FontSlant->\"Italic\"]\)\!\(\*StyleBox[\"\
\[Rule]\",FontWeight->\"Bold\",FontSlant->\"Italic\"]\)\!\(\*StyleBox[\" \
\",FontWeight->\"Bold\",FontSlant->\"Italic\"]\)\!\(\*StyleBox[\"\[Infinity]\"\
,FontWeight->\"Bold\",FontSlant->\"Italic\"]\)\>\"", "]"}], "]"}]}], "}"}], 
           "]"}]}], "}"}], ",", 
        RowBox[{"{", 
         RowBox[{"Bottom", ",", "Left", ",", 
          RowBox[{"{", 
           RowBox[{"Top", ",", "Left"}], "}"}]}], "}"}]}], "]"}], ",", 
      RowBox[{"ImageSize", "\[Rule]", "Large"}]}], "]"}]}]}]}]], "Input",
 CellChangeTimes->{{3.80967710943602*^9, 3.80967727695399*^9}, {
   3.8096773453319197`*^9, 3.8096774822999973`*^9}, {3.8096775475836353`*^9, 
   3.8096775547478228`*^9}, {3.8096777303567543`*^9, 3.809677744415505*^9}, {
   3.809677779825479*^9, 3.8096779209741087`*^9}, {3.809677957138678*^9, 
   3.809677986417451*^9}, {3.809678035569549*^9, 3.8096782305315237`*^9}, 
   3.8096783066332684`*^9, {3.809678341448812*^9, 3.809678414754534*^9}, {
   3.809678455757783*^9, 3.8096787692279863`*^9}, {3.809678877358776*^9, 
   3.809678914645728*^9}, {3.809679015930869*^9, 3.809679032533498*^9}, {
   3.8096790645189877`*^9, 3.8096791222897167`*^9}, {3.809679197211945*^9, 
   3.80967920247857*^9}, {3.809679248187942*^9, 3.809679354766959*^9}, {
   3.809679507043399*^9, 3.809679640750332*^9}, {3.809679762041155*^9, 
   3.809679806798585*^9}, {3.8096798443549337`*^9, 3.8096799045558243`*^9}, {
   3.809679969337083*^9, 3.809679969886499*^9}, {3.809680009523411*^9, 
   3.809680075721839*^9}, {3.8096801321322308`*^9, 3.809680156510552*^9}, {
   3.8096801962563972`*^9, 3.809680197500328*^9}, {3.809680257635291*^9, 
   3.809680413788992*^9}, {3.809680495126647*^9, 3.809680538935069*^9}, {
   3.80968061948205*^9, 3.80968062613911*^9}, {3.809680686240098*^9, 
   3.8096807548359327`*^9}, 3.809680933783175*^9, {3.809681204522635*^9, 
   3.809681506010157*^9}, {3.809681588266674*^9, 3.809681598540093*^9}, {
   3.809681670863538*^9, 3.809681700172694*^9}, {3.809681767969493*^9, 
   3.80968191110082*^9}, {3.809681967815984*^9, 3.809682026177568*^9}, {
   3.809682094771089*^9, 3.809682164546288*^9}, {3.809682208697507*^9, 
   3.809682295446869*^9}, {3.814005624397678*^9, 3.814005663073463*^9}, {
   3.8140057720907173`*^9, 3.814005872290097*^9}, {3.814005940912386*^9, 
   3.814005971193005*^9}, {3.814006042747965*^9, 3.8140060447521963`*^9}, {
   3.8140060777698298`*^9, 3.8140063008948708`*^9}, {3.814006425508295*^9, 
   3.814006459863667*^9}, {3.814007127159527*^9, 3.814007133221218*^9}, {
   3.81400716698982*^9, 3.8140071804281178`*^9}, {3.814011866611616*^9, 
   3.814011873992628*^9}, {3.8140119721242857`*^9, 3.814011987707789*^9}, {
   3.814085979865374*^9, 3.814085980930501*^9}, 3.814086037743071*^9},
 CellLabel->
  "In[1519]:=",ExpressionUUID->"8c2b3f1e-e249-476d-86d8-5c0252a58cd0"],

Cell[BoxData[
 TemplateBox[{GraphicsBox[{{
      GraphicsComplexBox[CompressedData["
1:eJzlmnl0zPf6x0eo2kZ/1LVMItq6YmjsEku1T2KI9l5iJJYrdGqp/DD2uhhL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       "], {{{
          EdgeForm[], 
          Directive[
           Opacity[0]], 
          GraphicsGroupBox[{
            PolygonBox[CompressedData["
1:eJxNmHn8V1Max+8pSkkpVCIpJJKUSosoSaVFq2gPIaTs+26KjApZspMlJmMJ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             "]], 
            PolygonBox[CompressedData["
1:eJwt08tLVVEUx/Gzo5l/gWTi1dQyLUNJDC9KoihBoIJIIKFQaJGm5ruHZqHV
7eGjgjTn4SMte6ulKXYdKImYNGnc3+DE72L/Bh8W556z9157rXVDtY3lDQeC
IMjFQdS5IPiNCvzj+SbxFTIRwgCWUYJSvEQEh3EC7XiIVOSgD39QjSPowhzy
EMa01ti3pzGKMRxHIt5iRu+myWeC2IZsZGEQf1GDBHTiCY7ijHLbxUXleANv
kIxDuIRNlGGe/R8Tm3BSe0xiSmeN836EeB/FyMA17OACYrCN28q9ClvO19Bq
V6lve5CkNVabu4hVjSzXR8rFcv6FNUR5Pme/45vOOosh3EIB0lCvM+2s/6xZ
If5U7axnH/FBvc1T7Z7pW6vhEn7orkXEYdxBoXryBZ+xx/t84j0sqhZWU5ud
bvXCZiiKdedzj8cGWlX786rFa5xSTewun5zfu0C9uIxj6onl8lV3t5xe4Dni
kIJanWF7RwI/K/16ZzOzgF7nZzFda1edn2XbY9b5GbLZydVsdCh3m5Gnzs9w
WDk9wHXnZ99m/h3ea631rJl4Rb2x/1ALGnS23fmq8zNTpZrabH13vvY2Y/uX
hmnu
             "]]}]}, {
          EdgeForm[], 
          Directive[
           Opacity[0]], 
          GraphicsGroupBox[{
            PolygonBox[CompressedData["
1:eJxNmHmYz1UUxmfOrVCRJWPfyVJkbI09MlqUGUI0USSppFWhon3TnlZt2je0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             "]], 
            PolygonBox[CompressedData["
1:eJwtlOmLj1EUx5/fPUPW7Ip4MYksZc0WYwkx1pkh+zIzDGOY+Y2xjOxLIska
IgwisiQ1JSX5G8goa8g2sq9J+Hyd++LT+Z77PM957j3LzSxM55WFJEnyIAPq
W5K8ZGEtrIA8/ClQhS6FReglcAm9Gh7CQPw92OnQGt0IimOMpuhX2HWwCtri
/8bmw3p4BBWs3cBOE+g25s9OQRZ6KOxFF0BLdAGcRy+ExujlcBNdAp3Rv+Je
N2tP+O1gA/o0tEf/xRYGX/sBzcz3dhxamf9DsSvgM9TTP7CH4Sl0MY89L/hZ
SvGvxDP1RPeA7egdsBTdCTahL8N3aIK/ErsxeG6eQGXMUYn5O3qmHGuvZaxd
jXuui14G11UL+Ak5+CewW+ELNDCv3VF4Af3wd2GzYTB6EDxIJck2/GfQ3VzX
pnzvb9D7gp9B706Fk8G/GY6+ozzBfijH/4AtgmuQxk9gProahpjXVLVUDZ9D
H/yd2A7QG10bY2ntXoypWFnQF/0OeyD4Gcaan1FnOwLj8Meb61yYiJ4Ex9Cz
ob95DVW73TGXd+O/ldP3MEo5jWeYYN4T6gXF+AgZ5r1wCEabr0kfhDH42RY1
jETX6Nxxz5PNe0C1r4q9cjvmWj2jWt6PtVFNc/G7wZbgNa2JOa2OMZX7Xua9
pRoE9Cw4gz4LI9AzzHu9HOqg58K54DOj2ayEW8FndBj6dfDZSgev/Z+o1QMd
ef4t+GypZzPxP8Xe00wMMM+xcqs7QHvvaj4bOkORZgkuoBcH/3dD81nXHjS7
j4PfHZph3QVf47u6EzQ7xXAx+AwtUG+Z1/Z/TPM7SXeR3kmhZ8Ka4DXN117M
e0M5aG4+M5oVzVwL8xwqd/pmjmKb11I5fRtzqlyqp/8BHFSjSA==
             "]]}]}, {}, {}, {}, {}, {}}, {{
          Directive[
           RGBColor[0.368417, 0.506779, 0.709798], 
           AbsoluteThickness[1.6]], 
          LineBox[CompressedData["
1:eJwt0kk3FQAYx+GLikLoXtLSju9Q3wEfwdCcW0gUGYqSNM8TRZSSFBVlSlSX
SGhaWbdta+G551g857yr9/wXv6yCcH5JTCAQ6CWW6J1ODrvIo5gqWmmjjxEi
LLPCP/6zygZPkggSIp0MtpPJDhr4xSme0E09I0QHlXGfOn5yiMd84z0fGKaW
qyxzkHvc5SRLHKCTWd4xxFtquMIi+ynlDtX8YB8dXOIvl7nNCRbYyyNmGOQN
AxznIt/ZzRFa+cMFqujnNa/Yw0M+84VpInzlFpXMU0w7LfzmPDc5TDPnOMYL
XtJHEW1M8omPnGWK69ygiTNU8IxenlPIA8aZYIxGrhFmlNM8pZwuejjKHAXR
jVqopIhcdpJNiCDbSCOVFLaSTBKJbGEzCcSziY3Rvohb73UNscFaxQ==
           "]]}, {
          Directive[
           RGBColor[0.880722, 0.611041, 0.142051], 
           AbsoluteThickness[1.6]], 
          LineBox[CompressedData["
1:eJwl0mdbjmEAx+H7qZQVRZkZFTLTEBkpNIWGTYooIpUR+ja+Ct/ASrR3IYSI
rNPhxXn8/td1PS+eF3dibXNlUygIgkc8DguCToeHOsOp8CDYyz62kcqA+ycs
s5/qTf2pK3Q5p9lPLjmMecvTX9qqv3Wc7XYa6TxzPqBvNEMzGbSfs9J+oU36
R1dpAmc4yyHeuj+o7zTQZtrtfJ3QLN3JsN3BGvu13tJIJu1EjeIes/nkbpdm
s5su52Tt0TsazZS9XhfwgIV8dXddb3CVBq6xgUaOc4JKvvvtHj3JD3ujbqLf
TtGtbGYLFZQz7W2R3qXXHiXGbiOWb85HtYxj9Dmv08tcoZZ66ljKF++3tVuT
dB73mc8SRtzP0bl8tg/rEUrpdF6rNVykmllcIo54wongo98WaQnFvHJerVVc
4DxhLGbI2zkNUcB75x1ayAe7RV9qDmn//jdxRDLt+x1njFFGGGaIQQbop49e
euimK/T/+/8Ln2Nc0Q==
           "]]}}}], {}}, {
      GraphicsComplexBox[CompressedData["
1:eJzlWntUlGUan9BFWR1avGWAWB4vgyFqOKREPOp4a1McCDFFx0uJtShsauZg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       "], {{{
          EdgeForm[], 
          Directive[
           Opacity[0]], 
          GraphicsGroupBox[{
            PolygonBox[CompressedData["
1:eJxNmHncl2MWxp87a5IS2YrwKpEUqQh5W7xKSaUikSTrEIoYUzS2KcqSspOl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             "]], 
            PolygonBox[CompressedData["
1:eJwt0+tPj2Ecx/H7ktMwRklOUVOTkSmNNltbzdaambEhcigzUgqFchg2Rc5n
Hjg9QB44G2ZSjq05rAeRh/4C/4P3Z9fnwWvXfrvv33Vf1/f7+WbVNK5sGJIk
SRmG4m1IkkMYg0k4iJvIxGhcxR+s9jv1eI5ZyEUdjmMCctCIDqQhD+vxGnNw
g+9VsvZhid7j9ybWnyhHL7/3sbZgFLKwET+wFF08v8Q6iFUYiQv4jRUYgXP4
heUYjla8wVzMw1q88Nlb2O8E62csQjEe4hmyEbAB31Gqe/P+R9YPfqY7V+Eb
SnCf52dYB7AMw1CNflTgL893sb7CbNfkPY6iEAuwA1uQgRmoRY17k+na17nW
6sFOvEQqUlyrZtdCNevBVkxBkc+6x2cr8d2aXKtS96bBe6lHJ7EN09zT27jl
3o1nPYXdmImJ2I/tmO4eHvaddJexuOuaqVb6z2M8wWSMw0Xcc+/7k/jtdd5b
Z3iATu/1j+dHXEPVriDELO11NpSpQe+pvZSZL7gTYhYXu9bvMN81b8MnTHVN
n7qGRc5EN45hoXumXirzee7ptRDPVOA7deGR76ZvaJau+OyaqcshZlTZTEcv
voaYlfwQs30+xNlQxpXtsyHOgjKurJ0OMZvKnLKmGdbsVrj3a0LslTKgWTgQ
4t6aCWW/PcRaaAauhzgTxa6BernZZ1NP/wMXt3B7
             "]]}]}, {
          EdgeForm[], 
          Directive[
           Opacity[0]], 
          GraphicsGroupBox[{
            PolygonBox[CompressedData["
1:eJxNmHl0lsUVxpM7giKbQZB9SQDZUXaJYZPIHhOwLpgKhmKrooJi61qXtlbF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             "]], 
            PolygonBox[CompressedData["
1:eJwtlPmLj1EUxr9zz9hlzQ+2NEhkGSLLkL3sxgzDkGGaMYuJGfuabBOyZ6kp
JaWmlJIkf4WQfc8SWbM1JMrncc4Pn85z7vu+995zz3PfnLK6wtqUyWTyIRu+
kLSwTGYN8RA8gEryRuIMGIDuD4XoDdAevRquosuhE7oWrqFXwD19R36YOAVG
o/NgN3oRPIJq8ovJ5+xtvqbW2gZfYTpjB5PPMQu9DE6j18MC9DPiRjgObchX
wRV0KTRHvyBuTl5TS/JvMfdayCHPhhL0VpiN7mxe2xHtCd0dtqCPJZ/rbnyr
OfvyrA/MRK+D+1BBfj75WAm6H2xPvucPMJl8H/EADEPPMV+rIp69J+5P/s4T
qGHsMnE+3IYy8rPEvFirlc4k+Zra6zw4mnzPBeh8qE4+pl5+j7NpjLk+EZcm
n7Oc/A5xLJyDleQ/o5eXolcdoCx5z6rQTdH7CzCQ/A1xF+yAUeTXiU1ZPjbO
3APq/R54DSPJq4hFyWvPyFNxBh9hqjwVe5QXvoQ35Ykf4Ql5QXuSF9tBQXJP
TkDfIA6FvfAWxjNWQyyGpzDIfK/N4CUMJ6+Ut+A55JLvjDMZgn4Xc2nslnxt
3svc5N5rDXOTe1C9+xW9VQ+1l1LGzsSe5N2b+i483JNnPcx7q7HH0QOdvWoa
jJ4E9eglYOhiOIE+ldzL6qF6J08/jB6pN9rTCPSrqE1n3o28q3nt8shnmEZe
l9xT6t1y8obo4UJ0kfna6tFE8zuou1cf7/6OXjZErX/jrFXzH/nYvPfqgf4V
Hc29rX/GGPM7p7umf8RidC/zf8fJ5LUmc62au5ifgWrfpB5qb+Z3WTVkyWvm
z/57CN3W/K7qTv4Dln6TDg==
             "]]}]}, {}, {}, {}, {}, {}}, {{
          Directive[
           RGBColor[0.368417, 0.506779, 0.709798], 
           AbsoluteThickness[1.6]], 
          LineBox[CompressedData["
1:eJwt0klTDgAcx/FHlC2SeAFuunfMe5BL2WmhiEhUyhpKiVAoWVLakRSFZEmF
VgrHbm5OTin6PDMdPjP/23/mN991iRmbDi4IBAIdhBC81xLNBuJIIZdiKqmn
nR4GGOMnU/ziN3+Y5i8zzPKP/8R4coY3nKaHU9xgjGQe84hWXvOKJPZxgpec
pIF6yvlBBS00k083iaRRTR6j7OYhH+mni+O8IJfrjLCLVLIZJ4dj1HGV71zj
NkcZZie1XGGSMqrIYogdPKCUCS5TyRG+sJ0aSvjGJW6RyWe2cZ/DFPOJLeyl
iK9cJINODvGcrdyjjwt8oJCb7OE953nGAZroIBjIIJu5w13OMkA8KfTyhDYa
ecs7CjjHfp6STjsJwS21kMxGYlnPGqJYTSSriGAlKwhnOctYyhIWE0Yoi1g4
3/IcpH1ifg==
           "]]}, {
          Directive[
           RGBColor[0.880722, 0.611041, 0.142051], 
           AbsoluteThickness[1.6]], 
          LineBox[CompressedData["
1:eJwl0mlXTlEAhuHzNpCZEqHoTYVEKmOhpEJSCaGMUTLV//MbVBLN84QQRRkr
11o+XOt+zt774wnXt1Q3h4IgeEGv0UNjRBC8JDcyCMY0R8f1sP7SozqhR3RS
f3PXbtVjOqX7NZNBu41t9nb+2JV6mSp2UM0J8jjOSfK5wl/vT+s7PaXvdZF7
drsW6Ac9qFkM26/Zab/RBxpFmGRqieYcM+5L9Yuu0ga67d2ayjrqWMst0rnN
HS7y3dsyndeN2kSfvVf3sZkW4ijnEtlU8NO7JzqgGbqJZmJJ4Ifzx9qve3Q9
z9nAVhacP9JeTdPVPGMNW/jmvFF7NEXnNEYf8tYeZYX9lJV89X1eZ/WCduku
vUkEN4gknhF31zVEMZ98H9ISPtv12qmJmsSSfV9f6RBX7ULOcIBr1LDsLtBp
LdKPelY7NIswcUQx7z+dYpIJxhljlBGGGWKQAfrpC/3/x/8B0k9a0w==
           "]]}}}], {}}, {
      GraphicsComplexBox[CompressedData["
1:eJzlWXtQ01cWThV1rQ1WrBWBCmqtseNqfYTWV48a13V8YKBIW8Ww0soCoaBW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       "], {{{
          EdgeForm[], 
          Directive[
           Opacity[0]], 
          GraphicsGroupBox[{
            PolygonBox[CompressedData["
1:eJxNl3kcl1MWh+8t+5JsIUuLRKVE2UoUlV0KETNJWSIVIaKaLGlBKpIh1Ahl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             "]], 
            PolygonBox[CompressedData["
1:eJwt1FlIVVEUxvG9xIokRUNUKgohBIkQMUSLKJrLBrW0zDTUaE7NoJHoLciK
JhuMSgrMBpuLbCCLioiGl96aLWieLCzCov6LtR5+rG9z7zn3nLXXvsnlVfmV
ESGEHESiWkK4gH54wnoldRU+kl9iFnkN2smvMZG8A53kv1hOXoEP5BeYQN6E
DvIvbCSfQBLesJ5PvY90PGW9kPoAGXjOeib1EGIRh/UYhS989h3N5NH4Guye
hWhAD0RjBlbjG5+3YQ75LPr6Ow7CZrFnK8dQ8h0MwDnWg6m3kIL9rIdQbyMV
jaynUfehO6IwBsPxKdhvDiPfxUC0sh5BvYc0PGJdQj2KBCRiLEbiR7Ae7yRP
QsA/yi7qbs/rME6sp9rLx8H2Igd/gu3JNvJ2/CZX4Dj5mFjW39yKLWLvrt/Z
QK7FZ/IrNJIPIx4PWd+kTsYzcn/qErE9ivNrtPfTEeN7UI88dEMkDmgfvbfa
s0ViPYzymblKveK91mv2YAq6IAKLfY+jfSZvUK/7s/WkHkGT91afuRXXvPcx
PkvFYt+N9b3N91lL8b2dKja7qT4LBWKzrzORiTqxWa/BUrEzo2flHXL9mfVZ
u/qz1/u76ztkifVQe1cbbK8qxK7VPcsm7/Vr61gvo9aInY33wWZhrljWmTiI
IrG90pk7Qz4tNgtvMZ5cJpZ/ooV80e+d7LOoZ1rPss6knsV5/mx6JvUsLhCb
9QyfZT1TbcFmejZOohf64BIui91bZ0Q/K0Vv/8553Xux2dUzWCl2BnT29UxX
+T31XnqmT1HXivVa76H/JfpO+i76n/IfONGb+Q==
             "]]}]}, {
          EdgeForm[], 
          Directive[
           Opacity[0]], 
          GraphicsGroupBox[{
            PolygonBox[CompressedData["
1:eJxNmHmUl2MUx6f7pLRLi9I6aDKlpkVNxbSnopSQNUtkX1IoirKUkF1ECJFs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             "]], 
            PolygonBox[CompressedData["
1:eJwtlPmLj1EUh79zzyRr1mQQBmNXxFiHGWOZwYxhFmYQWbLGJEtI1owQIqUk
f4GI7MQYjC1LtoQskbUQUmR7Pp37w9P5nPfe995z7znnpk4vL1wYEolEMSTD
E5xMSyS2YbfDPeiDvxFbAVehG/5q7Bo4ChPw92J7whFoh78U2x6OQ3f8tdh0
uAAt8BdgyzUH3QGWobPB0AGmoKfCZXgO42AF1GWsDsxCz4ZG6CYwFz0QXkIZ
/n7sUKiCxzAGlkAlPIJ80MEvQlvmLw4ecy10TZiBngkf4UNcSzE81D6Mb8Zu
hWvQFX9V8DO+h3eQFffsyNileLZCqIZXUa8MfpbqqHWmE1DKt33YftALfQeb
q3uGz/BJJPm37/AN0qAAhjH/adSlMZZs81wqpr9wCkpgIvRn7DR2PGyCX/BT
OYciOKgx5uxR3nQm9Os4VgY98MfCLnQxHIN65rnRGinma+jf+XAYapvfrWI+
BC3NayEFhqN/xNh3wD+4BSNgEvyBJOZMVuzQEN0AnnEf0/DPQjP8UdiM4Lk8
F/dWTpua14xqZR6ch+b4edjMqAeY34W+9UZ/iXe/AU5CDfO9lAPV6hn0nOA1
q9qujLWlGk+Yn0GxK+Y08zPqbKr5NuhUWKS9Yy1WxVpVTbbGb2V+N5qThx5t
XhvrVCPovub/qkcbm8esWNUTb4L3cEXM+QFINu8t3Zl6+XpcSz1dH/82dn08
011IN78r3emNmPPimBPVvt4EvQXqgc74XWA5uht0Mu9h9a6+ZeB/Df42bAn+
dlxR3QZ/Q4bgZ5n/qzcoF51j/nbshJHKrbnWP4PRg8xzlx9r9X7MvWpWvXQT
mxO8p9S7D+Le6mG9FS+Cv116M94G73H1tnqimPES2B3XKEIXmteevhWgfwfv
BfXAf/WUlYo=
             "]]}]}, {}, {}, {}, {}, {}}, {{
          Directive[
           RGBColor[0.368417, 0.506779, 0.709798], 
           AbsoluteThickness[1.6]], 
          LineBox[CompressedData["
1:eJwt09dWDwAABvB/ysyIeAB33FtRNITMyiq7FA0kmWVkj4ZVlJEd0jBCsjIi
yXEjXoAXqBfo1zkufud8d9/N941Pz0veFhQIBLoYQH8ex0SiSCKDAoopoZIa
ammgmVbaaKeTn/yim9/84S+9xCmZSxFr+E4sx/nIfR5wl1ru0cYH7rCa26xi
K53EcIbTnCKVW0Qzh0JS+MYsjvGO99xkJTdYwRY6mMk+9rKcr0RxlLdcp4Zr
LCOXL0Syh90spZ0ZHCGbH+Swi2Q+M53DvOYNV0mimiVUkcgVIijmJS28oJVX
7GQxn5jGITbTRRY7KCCTTSyikstcYioHeUYzT8ngOdvJJ52NLOAiC6lgCgdo
Yh2NrCePeOppoIR5nOQxT9hAGgmUMp8yJnGWc5RzgfNMZj91POIha+kf82xO
0GNr/0gkkgmMJZwxjCaMUYxkBMMJZRhDGcJgBjGQEIL/f6UPFjlisw==
           "]]}, {
          Directive[
           RGBColor[0.880722, 0.611041, 0.142051], 
           AbsoluteThickness[1.6]], 
          LineBox[CompressedData["
1:eJwl0+dejgEYwOHnLVtGKKOMkF1mRWRTlJVRfOwAcg6OxAHYKySjKLO9h5lI
IlRCydXPh+v3v+/7+fa87xOXdzo7PxQEQRlnw4Kg3HJGK5gSHgSFGqbhjOKc
fapGMpox3HQ7zxe+UsU690pdrz26Qaczg2lEE0US1Z7P0js6U4u0lhPmCzpb
7+o4Hc8ELtt/0Esbu9xadbf26x6NZwmLeei2VN/qKX2npUw0RzCJ6/Zf/OY9
B9za9aAO6iE9zBqe2Vfrc52sayk3DxOYQwyZs/UoRzjOMbrdX5BrLtA/ZJqz
SOSpfZU+1pX6RBP0hg6Qbs5gL/vZx2f3MnLM1/Qnrzlpf6Of6OIlr3jEMs9W
sJyr9j62m3ewjRb7Tv2onTTRTDEL3Ut0kV7R72wx1+lmrdcP3GaTPZU53LOP
1TS20mBv5D4x9gcaqxd1ns5nLnEs4JL7N1LMG0mmxt7BLf6O/L4j75MIhvyf
u2mjlRaaaaKRBuqpo5YaqqmikorQ/+/iHwj+bBQ=
           "]]}}}], {}}, {
      GraphicsComplexBox[CompressedData["
1:eJzlWX1Yk9cVT6lFxYYO0VY+/Nq0jX3EihVsnXrQtGorWuInOo3OiVOCYhmz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       "], {{{
          EdgeForm[], 
          Directive[
           Opacity[0]], 
          GraphicsGroupBox[{
            PolygonBox[CompressedData["
1:eJxNl3mYkFUVxu8RF3aQRURUhACFAEcEgUGRfZsRl2GRRUKQkG0ADRNhYhmU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             "]], 
            PolygonBox[CompressedData["
1:eJwtlHeIz3EYxz8PRyklXUodcvaWvffmHC77zMMpcifRiTp7dLZOVnH2kXFn
RGcre+8dEkLICCG83j3fP159Xt/f7zuf5/154lPSk9IKhRAGQwyMsRBOQxyc
5Lgi62EoDukct2ZdCq/wN9AJbwC38QJoic+HZ/gLaIxXh6v4QxiNr4cfeA5U
wo9ACcjgeCtrbbiO/4GOeGfzez+AsvghKAYJHFdgXQHv8VRIwrfBX/wfDMJ3
Q2EoAiPgFJTWc/XdrPXhFn4BEvBGcBc/DkPwdnAfvwgTzWuiWui3yfgU83Oz
YTg+yfxdM4PfO978Wj2jHL7QvDY9g9euMlwKXsP+eD24iZ+HDHwWPMbzIBev
CzfwEL1rNbgS/J0b4lXhMn4HhqmPUEp95Hgq61x4ih+ANniy+be/gxS8O3wJ
fo+1eGKUBfVwMb4EXuKJwbPQCl4Hz0RbfDm8Dd6TRXiW+bfqmgXmPVTvnsMW
PFOZw38Hz0KHqDfKRAF0g6Mcl4x63w8OBs+Astk++jZldKZ5hpQd1UxZbGH+
LGWyOT4PzgW/RzO8DjzB82ENPg1O4N+C13a1uavG2XgP+Ip/hpV4V/iEf4AN
+Eb4pSzDJjwHcoP/th/PN//vJ8w277F6q3c4hvdRRvHYqJbqkXqjmq7De5tf
+x1G4l3gY/AMlceXme/N5OB7WXtOey0uysLQaK8pE2kwwfzcs8F7MQMeBe+J
3mU83Av+TspeGfNzlUFlRTVRLZSZpngtOBP8HlWiHqp3WcGzMsA828qM3l01
VO30Der1OPPsquc1zGsSG9VwH+ve6F6quXo13Xy2qGe9zGuk2mjGpJrPBM0C
zZQ55nsoRHsgz3wPa+8WZd0B282zrRmy03zPaK8Ya1/YZe7K0CrWUebXKhNj
zWeOZo32sGaNZoJmgWaOZtFAiDGfSTVhs3n2r8EePN18lumcJtE5+k89+A8O
wsbJ
             "]]}]}, {
          EdgeForm[], 
          Directive[
           Opacity[0]], 
          GraphicsGroupBox[{
            PolygonBox[CompressedData["
1:eJxNmHds1XUUxfvuAxktlEIBAVuhslQQFFFRQQTFgQqCijhYKioiU0BRWTJM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             "]], 
            PolygonBox[CompressedData["
1:eJwt1OdrVUEQxuFz9yj2FkksiUossYGKvaGIXVRMiGKLaFRUrBAISPAPUFRs
2LH3KLELimLvBbGDXVQs2Csa4+9l5sPDnt2z957dmdlNz52ZNSNEUZSNUthK
5yiaow+uIi2Ooqm00/EVT5GK/tiMD2iGWjiGyswfSzsOS1CO/kiNYxd+YiAG
oybv7tF2xkS8wHcMQF2UIMGcTNp0vMcm/c6/+Qbv0NvHzmgdzJ9EOwWX0Zj+
HNpGvvYv6Od7KMQptMZQ7EVZ5g8PtuZn2Ia+6IEHWIYkdMEF7PPndijCebTF
CGzAK/RECj5hntaQsP/ciJXohRwsQmm+P0zrwCEcQRN0xTU05H2B9uVrqU8/
39e0HA3oz6ZNxho897VHeBQsx8ptN+xGKeZnB/vmR2zxXKoGfmEBilnvIH+u
x/xZPqbnExoLNvYN1RjL9Rxe95pQLbTHHpThfRutUznl+aI/5/neVnisCnxv
Bzx3+b63g7QZwfao3F7yvSrHL7EfozzmymWR50I5LU9/KW11n1OMfxiCkoTF
otBrQTH5iz+ak7A5r1GBOaM9Z3FsNarazMJnVGJsjNdYxdhyplzpN4c95zme
06q8rxLbt3VmdNau0E4LdubOIjW2/1bMbgSrcdV2RxzXXng/nralx34tbYdg
OTiJdT5XNX4O69EpWI3qLO4MlludSf12Plr4f9zFnWC1qLHHXkOqne64jVte
W62wEL+DnW3F7C3tXDQNdkZvok5sd0mGx2qV14Zipr0/DFbrioHuhhTmTAh2
R9TgOTm2b2ksKbY7Q3eFYlCb/v1gd81k7MAPj41qVHfDaX+nO0K1oBrM85rY
jtXB7rY07Su2O0p3U6bHbnGws64Y6iw/8VjoTP8HAg2XOA==
             "]]}]}, {}, {}, {}, {}, {}}, {{
          Directive[
           RGBColor[0.368417, 0.506779, 0.709798], 
           AbsoluteThickness[1.6]], 
          LineBox[CompressedData["
1:eJwt01dXDgAABuDPKCvJuktkpOzILiIjRGRTIiMyys6MkpXxiQ8hFEU0zWz1
G/pBns5x8ZzzXr1X7xubV5h1pEcgEOiiJ915OAmkkEk2+RRRTAnlVBAkRBXV
1FBHA4000Uwnk5TPoYSLvKKWiczmIOv4yQRO8ZGX1FDNC57TxgeesZYnZPKU
PXSwlwucZw1VJDCLc5zlMY+IZyYZPGQ1JzlBK7sJsYoHrCSPH4wniQJW8J04
jtNMC/dJp5Ll7OIb45jBUn6xjCUcI5e/7OQAabQzlqPk8Icd7GcxXxlDEYvI
5guxTGcbv9nOQoKkco/RFNLEFhrZSj6JbGIzKdxhAXcZxWG6h7SeT0QzlbeU
cYViGpjGOzZQzjzO8J6NfGYE15jPVZK5Tgw3qeAGt7nFSA5RzxvqyGIfk3nN
JS5zmlKmMJc220wmnmEMZQiDiWIQkQwkggH0px996UM4YfSm1/8v/QM4P1fV

           "]]}, {
          Directive[
           RGBColor[0.880722, 0.611041, 0.142051], 
           AbsoluteThickness[1.6]], 
          LineBox[CompressedData["
1:eJwl0ulWT1EYwOHzr5RKihIRDaTcjktwAVxHZR5DhUylMoco0TwPKhGNhsgY
JaF4rD486/e+e6/z5aydsWvPzt2hIAiqaDDUsyMsCHKZ5gPvKeEj9/jMeT5x
gS9UkxUeBPf1IiOMMkYeX7nEBJOcJp9+Bqhjne/Xk8yQfYMGhPhrD9cwNtFo
j9JKjdSr2kqG+YjG6E19w1vOcowXnGCb+y5dpbFMmd+RYy7Qc9zhLlW8pINU
9526Wfv0NlvNx3WYNrLt7drLLTLtR/U5N2gh3dkhPUMT2+3Neo2V5usarYf1
GRvN9ZqiDdpDMY94TDkR7ip0hR7UQZ5SwxZne7WQNeZZneMVa+1Fmqi1+oNS
XpPk7KGm6T79SRnzXGGB/fzmD79YYpEDPCHetwms5oG9m5PE2Wf0G+N85xSX
//9nZrzFCfropYduuuikg3baaKWFZppoDC2/539McXGr
           "]]}}}], {}}}, {DisplayFunction -> Identity, PlotRangePadding -> {{
        Scaled[0.02], 
        Scaled[0.02]}, {
        Scaled[0.02], 
        Scaled[0.02]}}, PlotRangeClipping -> True, ImagePadding -> All, 
     DisplayFunction -> Identity, AspectRatio -> 1, Axes -> {False, False}, 
     AxesLabel -> {None, None}, AxesOrigin -> {Automatic, Automatic}, 
     DisplayFunction :> Identity, Epilog -> {
       InsetBox[
        FormBox[
         StyleBox[
         "\"\\!\\(\\*StyleBox[\\\"c\\\",FontSlant->\\\"Italic\\\"]\\)\\!\\(\\*\
StyleBox[\\\" \\\",FontSlant->\\\"Italic\\\"]\\)= 1\"", FontSize -> 13, 
          GrayLevel[0], Background -> GrayLevel[1], StripOnInput -> False], 
         TraditionalForm], {0.63, 0.55}, {-1, 0}], 
       InsetBox[
        FormBox[
         StyleBox[
         "\"\\!\\(\\*StyleBox[\\\"c\\\",FontSlant->\\\"Italic\\\"]\\)\\!\\(\\*\
StyleBox[\\\" \\\",FontSlant->\\\"Italic\\\"]\\)= 0.8\"", FontSize -> 13, 
          GrayLevel[0], Background -> GrayLevel[1], StripOnInput -> False], 
         TraditionalForm], {0.63, 0.465}, {-1, 0}], 
       InsetBox[
        FormBox[
         StyleBox[
         "\"\\!\\(\\*StyleBox[\\\"c\\\",FontSlant->\\\"Italic\\\"]\\)\\!\\(\\*\
StyleBox[\\\" \\\",FontSlant->\\\"Italic\\\"]\\)= 0.6\"", FontSize -> 13, 
          GrayLevel[0], Background -> GrayLevel[1], StripOnInput -> False], 
         TraditionalForm], {0.63, 0.34}, {-1, 0}], 
       InsetBox[
        FormBox[
         StyleBox[
         "\"\\!\\(\\*StyleBox[\\\"c\\\",FontSlant->\\\"Italic\\\"]\\)\\!\\(\\*\
StyleBox[\\\" \\\",FontSlant->\\\"Italic\\\"]\\)= 0.5\"", FontSize -> 13, 
          GrayLevel[0], Background -> GrayLevel[1], StripOnInput -> False], 
         TraditionalForm], {0.63, 0.25}, {-1, 0}], 
       GeometricTransformationBox[
        InsetBox[
         FormBox[
          StyleBox["\"drive invades and fixes\"", FontSize -> 15, 
           GrayLevel[0], StripOnInput -> False], TraditionalForm], {-0.15, 
         0.47}, {-1, 0}], {{{0, -1}, {1, 0}}, Center}], 
       InsetBox[
        FormBox[
         StyleBox["\"drive doesn't invade\"", FontSize -> 15, 
          GrayLevel[0], StripOnInput -> False], TraditionalForm], {0.55, 
        0.75}, {-1, 0}], 
       InsetBox[
        FormBox[
         StyleBox["\"polymorphism\"", FontSize -> 15, 
          GrayLevel[0], StripOnInput -> False], TraditionalForm], {0.47, 
        0.1}, {-1, 0}]}, Frame -> {{True, True}, {True, True}}, 
     FrameLabel -> {{None, None}, {None, None}}, 
     FrameTicks -> {{Automatic, Automatic}, {Automatic, Automatic}}, 
     GridLinesStyle -> Directive[
       GrayLevel[0.5, 0.4]], 
     Method -> {
      "DefaultGraphicsInteraction" -> {
        "Version" -> 1.2, "TrackMousePosition" -> {True, False}, 
         "Effects" -> {
          "Highlight" -> {"ratio" -> 2}, "HighlightPoint" -> {"ratio" -> 2}, 
           "Droplines" -> {
            "freeformCursorMode" -> True, 
             "placement" -> {"x" -> "All", "y" -> "None"}}}}, 
       "ScalingFunctions" -> None, "TransparentPolygonMesh" -> True, 
       "AxesInFront" -> True}, PlotRange -> {{0, 1}, {0, 1}}, 
     PlotRangeClipping -> True, PlotRangePadding -> {
       Scaled[0.02], 
       Scaled[0.02]}, Ticks -> {Automatic, Automatic}}],InterpretationBox[
    Cell[
     BoxData[
      FormBox[
      "\"female selection \\!\\(\\*StyleBox[\\\"(\\\",FontSlant->\\\"Italic\\\
\"]\\)\\!\\(\\*StyleBox[SubscriptBox[\\\"s\\\", \
\\\"f\\\"],FontSlant->\\\"Italic\\\"]\\))\"", TextForm]], "InlineText"], 
    Text["female selection \
\!\(\*StyleBox[\"(\",FontSlant->\"Italic\"]\)\!\(\*StyleBox[SubscriptBox[\"s\"\
, \"f\"],FontSlant->\"Italic\"]\))"]],RotationBox[
    InterpretationBox[
     Cell[
      BoxData[
       FormBox[
       "\"dominance \
(\\!\\(\\*StyleBox[\\\"h\\\",FontSlant->\\\"Italic\\\"]\\))\"", TextForm]], 
      "InlineText"], 
     Text["dominance (\!\(\*StyleBox[\"h\",FontSlant->\"Italic\"]\))"]], 
    BoxRotation -> 1.5707963267948966`],TemplateBox[{
     InterpretationBox[
      Cell[
       BoxData[
        FormBox[
         StyleBox["\"A\"", Bold, FontSize -> 20, StripOnInput -> False], 
         TextForm]], "InlineText"], 
      Text[
       Style["A", Bold, FontSize -> 20]]], 
     InterpretationBox[
      Cell[
       BoxData[
        FormBox[
         StyleBox[
         "\"   \\!\\(\\*StyleBox[SubscriptBox[\\\"\[Lambda]\\\", \
RowBox[{\\\"f\\\", \\\"  \
\\\"}]],FontWeight->\\\"Bold\\\",FontSlant->\\\"Italic\\\"]\\)\\!\\(\\*\
StyleBox[\\\"=\\\",FontWeight->\\\"Bold\\\"]\\)\\!\\(\\*StyleBox[\\\" \
\\\",FontWeight->\\\"Bold\\\"]\\)\\!\\(\\*StyleBox[\\\"2\\\",FontWeight->\\\"\
Bold\\\"]\\)\"", StripOnInput -> False], TextForm]], "InlineText"], 
      Text[
       Style[
       "   \!\(\*StyleBox[SubscriptBox[\"\[Lambda]\", RowBox[{\"f\", \"  \
\"}]],FontWeight->\"Bold\",FontSlant->\"Italic\"]\)\!\(\*StyleBox[\"=\",\
FontWeight->\"Bold\"]\)\!\(\*StyleBox[\" \
\",FontWeight->\"Bold\"]\)\!\(\*StyleBox[\"2\",FontWeight->\"Bold\"]\)"]]]}, 
    "RowDefault"]},
  "Labeled",
  DisplayFunction->(GridBox[{{"", 
      ItemBox[#4, Alignment -> {Left, Inherited}, DefaultBaseStyle -> 
       "LabeledLabel"]}, {
      ItemBox[#3, DefaultBaseStyle -> "LabeledLabel"], 
      TagBox[
       ItemBox[
        PaneBox[
         TagBox[#, "SkipImageSizeLevel"], Alignment -> {Center, Baseline}, 
         BaselinePosition -> Baseline, ImageSize -> Large], DefaultBaseStyle -> 
        "Labeled"], "SkipImageSizeLevel"]}, {"", 
      ItemBox[#2, DefaultBaseStyle -> "LabeledLabel"]}}, 
    GridBoxAlignment -> {"Columns" -> {{Center}}, "Rows" -> {{Center}}}, 
    AutoDelete -> False, 
    GridBoxItemSize -> {"Columns" -> {{Automatic}}, "Rows" -> {{Automatic}}}, 
    BaselinePosition -> {2, 2}]& ),
  InterpretationFunction->(RowBox[{"Labeled", "[", 
     RowBox[{#, ",", 
       RowBox[{"{", 
         RowBox[{#2, ",", #3, ",", #4}], "}"}], ",", 
       RowBox[{"{", 
         RowBox[{"Bottom", ",", "Left", ",", 
           RowBox[{"{", 
             RowBox[{"Top", ",", "Left"}], "}"}]}], "}"}], ",", 
       RowBox[{"{", 
         RowBox[{"ImageSize", "\[Rule]", "Large"}], "}"}]}], 
     "]"}]& )]], "Output",
 CellChangeTimes->{{3.814006148431986*^9, 3.814006157968115*^9}, {
   3.8140061879904633`*^9, 3.8140062736784477`*^9}, {3.814006431513706*^9, 
   3.814006461889913*^9}, 3.814007135210734*^9, {3.814007168715795*^9, 
   3.814007181841032*^9}, 3.814011877087171*^9, 3.81401199119109*^9, 
   3.814085984015634*^9, 3.814086040256649*^9, 3.8141965724436617`*^9, 
   3.814197128695944*^9},
 CellLabel->
  "Out[1523]=",ExpressionUUID->"5d51ac5a-5023-4e5d-aa69-59f1a9a324be"],

Cell[BoxData[
 TemplateBox[{GraphicsBox[{{
      GraphicsComplexBox[CompressedData["
1:eJzlmnl0zPf6x0eo2kZ/1LVMItq6YmjsEku1T2KI9l5iJJYrdGqp/DD2uhhL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       "], {{{
          EdgeForm[], 
          Directive[
           Opacity[0]], 
          GraphicsGroupBox[{
            PolygonBox[CompressedData["
1:eJxNmHn8V1Max+8pSkkpVCIpJJKUSosoSaVFq2gPIaTs+26KjApZspMlJmMJ
KUqlUBShsRsjw9gNYx3GPO/Xed/Xtz8+Pd9zz3mWz3POPffzq+lRkwefXKUo
itnxT9Ww9QMpwLMzA1cFpgcaB7YI1Ag0LPLazQLbazcPNNJWCzQwBut2DFTX
fwfnGe8Z2M6cOxm7ZmDnwJaBWoFdA3UCWwd2CdR2vJvP6ga6B/YKtAq0CGwT
2DbQxHjEau7aeoHdtaxrap6tAi2thdo7mwO/veUJv/bWh98+8oFfGy19aquF
U2v9WNfOmvDf1/km1t7AvrbSkq+jnOlBJy017Rdo5lwXa4TT/lp6cEBgD3vc
wXrx6SZPch7ofEvntrK/pwZGB8YEesgfHofIE379rY+6essHfn209OlQLfl7
6ce6fnLAv6/zjEcEeppngLHZh4HyhN8wObDnQ+XA+HCfHRQ42HrZnyMdE/cw
4xFruGvhd4SWdSPNT72jtPA70Rz4HRsYFBgcGCcHOI3X0pujtPA4WnuYPe1j
b3paIz09xnm4jnW+r/vQW5/jzDkkcJK1UPvxPqMfJ9gjaj1FDsSY5Fr6Mdle
w/VknzGeaAz8pziP/2nWTV3rAssCywOPBO4I3Fnku2K8vM+wL4wvsD5in+U8
/ThbC+8Lnaf28+wvXM8NTHB8vs+IdZFr2ZOp8uTMXmJf4Hq5POFxmTwZ/8ln
p1j31YEZgYuNh/8048H7Cu3pgSu18JuuLe/IM+VHrHOs/XrjUtcsucHjGi29
uVZLD2bqx7rZ8sT/OucZ90u5d38O3GDsSwM3yRPeN/qM8Ryfwftm7VTnJ9mX
W3wG79vkDNdbfcZ4t8jbKGWuB4dtm3IPOI9di3zf0Jux7v1cOcP1XnnC4x6f
Mf6LdVPjfc7D+34t/ZvvPLU/IGf4zXOe8YM+I9ZDroXT49ZBzuZR744p83ss
cFfg7sCjngPGC3w2VztTDr3Cr33K/XjCeHB6UkvtC7XU/pSWGhdpqf1pa4X3
CvNR4+4Ru3HKtS8NPFzkd+tZLTWOifnDUu7Hcjngv8x5xovNQ47njE29q6yV
Glf6jPHzPqPeF7SL7MUM+b/oM2K/FFhijYdGLR1T5rTWWqnlZecZD/FccEe/
Yt3w5t7lO8d35XXro5a39Cde+S3kO/WG89TYInI2Sbmu1+SD/5vWh//fAmsc
fxJ4J/Bu4G1jr3NMTa8G/hHYoN/+EbtLyjk/8hmx/wj8EPgx8E9jvSPWGWuj
a+HxsZZ1n5rvvcC/tO8Hfg58Hfgm8J3rmP/K3Bud32i8b7Tk/1YLvy8CH8qj
gT3jm/5v54n7pfPEHRf8hgY+iN//CXxujF/M8a1cv9DvJ2v62trp2frAr64l
z++B7413YMQ+IGVOPxoD//85T2y+J9yrp9qL9e7nB1rORe2IsXmgWqB6yv7s
wxYpW4RqzUAKVAkcFOgeqKrP7+YcFb9HBuoFtgpsZlz0DroHbYTPL3KqkXLs
pE8N8+Dzm1y3MxY1NgxsHagbaBCo47jOJrVvraX2+voxTyo0JzoYPYu+KzUv
Fo2HhkUnohdLnY0+REfv4Bg9vaNjdHZjx+jvhp4LevSz+1zDeKwrdTzr0Mo7
m6+WvWFML37yzKKjm1trqaebux8/el6qmBNu6OwWroMnmhNtWlt+zVzX0rny
+3qldXFPcBegafhGnyBPtFNr/dGPHYqK/m4jv30dw6ed41Kv7yM//Epd3951
xEUP9rDGjuaAMxqyk7Vyt5VaG9tVHt2cQ0uio0bIg3gHFRVNT1ze1+76tHK+
u/X1tI42xurpOjTaUGtBNw4oKpq7j5z6Oqbufo7J198xvUBbonF5//hmrLSu
Aa4rNX1v+9jLOtDQA83d3b0ZJn80wSD5UyPfgeb6dJbfcH3ghv4sNfgRzh2g
32A5j3Adc2g0zkH5N0+XoqLXR8oZHTJGf3TjsUVFo48rKhp9fFHR6EfZo3H6
D9JvgjyPcR1zaJ3TrPF4c8CfsznRWtFWpe6eZN3wmOwcPR1dVDQ6d+Ep9myi
ceEzRZ9Rzk+xxjOsAw5nOmYOjXe5Pbq4yLoRHujKc631fMfkuMAxOS90XGp0
NG0fezlafhe5boLxzrF3Z1kHPC8xN/VOtSZ6cKlzk60Rjcr9xfu2tz2Ypg/f
GfTCc0VFi0+zd5cZq9TlzG0b908t7+XJgdmBVoGuKX/X35Azmq7U2tgZ1oSW
vL6oaPRZRUWjX2O+G11HD9CopTa/znUXG2O2Nd6kT8+UdTMautTic6wbTVpq
bSzata1aFW6lplxoz+YYl79d7izy32BwuMvxVGPcLL+7nYMD2nVuUdHZpe7G
3qsPOnS+POa5Ds4POL7WtfdYB3q21OAPum5WUdHR7dW2txQVrfykNaFbSw2O
Re/2Tllzo9fxX+A64j3u+A7XPlxUtHipzZ9w3WC13N+LihZ/yhrRrWjasSlr
68bq66flgz/nboU50LRLioouX2qela6jPvTtMnksc90CYyy3xlX64I9Wfamo
6G80bt+UtfVO6usXrXOJa9fY38XyWajfKvOtNS751zkm9yuOqeVVx9S93jH+
aK7XXIeuRbs28b15wdzo4Q3W8qZj8r+h/1r9Si3/lutWGnu9ud81B2P0X6kr
33Ou1ILMdVGrEp+zP989og407YfmQCeXmv0j56gFnVzq8Y9dR+5PnCPfp45L
jf6pcdHA6OPxKevmnVNeh+b91hxfuo6cXzlG56I9fy8q2r3U8l+7rtTx35jv
O+OinX/1N7HQVKUe/8HYTdXX5d8LP7mO2D87/ty135uHmKXG/8V131vjb8b+
w7q5R7lPt1H3JTVxqb+rq5erqr2bqbt/Ndbm6mxyVFFLE7f6Jtq8mpqausvY
hXp7C7V8zVTxr+ncZt7xtfSv7RhsZ93V1Ne1jVV3Ew1eavNdUtb3Zdx6ruO9
4r3jzNdWrxN335Tv5dud591gDX9bNNwkB7/5O2Cg7zF3yuiU/w7ZNdA6sEfK
/v0DO6T8/yqA7wPal7uf+4/vwWDPHe8Aa3dz/T5h90z5LAzTch6OlBu8hrs3
7EubsHs5t5e1wLldyv8/Qs6hxuC9487mXuR92y/l/6fgWeeU/w6nB2WseuZp
6Tloa7+ISb18hxsZY++U4za1fs4xnPHhe9fCHjUxRjvjdEj5W/LXIt833Avc
CTzrYG3Y/fzdxD6zto57Ut+e4sP3qFPK9+zqIu8Tvxf7vLMcedbJ3+RES/BN
qevZqmGeruZizw9137F9jf+B/X3fPR3ivhJzsPvbUd/V7iPvwJb2lN9V7Fk3
+9Yj5b97q7rX3B3/LfJ5INdn3is8L++WA13Tzd/MdzfOlvoN8xwc7u/PjHm4
+8bZ5nv5TJHtQPnOd8z93M9zii48OX4v8aw8G1iesiZCO6Jd0DKcnR5y4b0Y
kPJ5H2T8Z6x/uOd5iH1jLwbow/+LDtSHfejnc+ro75g1Z1vfWZ47tNntvnO9
PCPY3p433mne8+21o/zNOUY/wHtXn9ezN2Os+SH64VlmfrS+jxDbs8tej0j5
fcJvrL58c8bb//vCnm5P6llHXc8F/jU9T+NS5ayN03dN2Oc9i8vgKt8FYR9N
+XwfF3jMs9fInvBe8u4ekirPeup7iL95foT7wrnqY9/oZXv7SQ+PCfuwv4fr
w51xRuBBeZ0XmCOPswM3uY8j5Mj5fDUwM+V37vXAtSm/E0ezD9a2Iex1KZ+p
FYGV1nRh4BbzTg3ckPLd91Lgbs/Sy2HvkcvasPealxrvs85jrYMaHghcnvI9
90LgLnt4gvXtae9ftP/EPtH488JelvLZmRWY635NCXtjynfl9MCVKd+d06yb
mq8K3Ok5eTvs1SnfnZcE7jfmrYGT7MPSlP8uamVfV9vbGSn7t5DrJPkSY55x
2JNr3JeFgUUp3x2323d6/oR9pIeLqSvlO4UaJ1jnHSmfA84AXCfK96mU+05P
Hg/7ZMr37HMp+7Cevh5nb59OeW+o87bAaSl/324OnJryt5H9PN566MEG+7Aq
cEXK5/T/EdSFCg==
             "]], 
            PolygonBox[CompressedData["
1:eJwt08tLVVEUx/Gzo5l/gWTi1dQyLUNJDC9KoihBoIJIIKFQaJGm5ruHZqHV
7eGjgjTn4SMte6ulKXYdKImYNGnc3+DE72L/Bh8W556z9157rXVDtY3lDQeC
IMjFQdS5IPiNCvzj+SbxFTIRwgCWUYJSvEQEh3EC7XiIVOSgD39QjSPowhzy
EMa01ti3pzGKMRxHIt5iRu+myWeC2IZsZGEQf1GDBHTiCY7ijHLbxUXleANv
kIxDuIRNlGGe/R8Tm3BSe0xiSmeN836EeB/FyMA17OACYrCN28q9ClvO19Bq
V6lve5CkNVabu4hVjSzXR8rFcv6FNUR5Pme/45vOOosh3EIB0lCvM+2s/6xZ
If5U7axnH/FBvc1T7Z7pW6vhEn7orkXEYdxBoXryBZ+xx/t84j0sqhZWU5ud
bvXCZiiKdedzj8cGWlX786rFa5xSTewun5zfu0C9uIxj6onl8lV3t5xe4Dni
kIJanWF7RwI/K/16ZzOzgF7nZzFda1edn2XbY9b5GbLZydVsdCh3m5Gnzs9w
WDk9wHXnZ99m/h3ea631rJl4Rb2x/1ALGnS23fmq8zNTpZrabH13vvY2Y/uX
hmnu
             "]]}]}, {
          EdgeForm[], 
          Directive[
           Opacity[0]], 
          GraphicsGroupBox[{
            PolygonBox[CompressedData["
1:eJxNmHmYz1UUxmfOrVCRJWPfyVJkbI09MlqUGUI0USSppFWhon3TnlZt2je0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             "]], 
            PolygonBox[CompressedData["
1:eJwtlOmLj1EUx5/fPUPW7Ip4MYksZc0WYwkx1pkh+zIzDGOY+Y2xjOxLIska
IgwisiQ1JSX5G8goa8g2sq9J+Hyd++LT+Z77PM957j3LzSxM55WFJEnyIAPq
W5K8ZGEtrIA8/ClQhS6FReglcAm9Gh7CQPw92OnQGt0IimOMpuhX2HWwCtri
/8bmw3p4BBWs3cBOE+g25s9OQRZ6KOxFF0BLdAGcRy+ExujlcBNdAp3Rv+Je
N2tP+O1gA/o0tEf/xRYGX/sBzcz3dhxamf9DsSvgM9TTP7CH4Sl0MY89L/hZ
SvGvxDP1RPeA7egdsBTdCTahL8N3aIK/ErsxeG6eQGXMUYn5O3qmHGuvZaxd
jXuui14G11UL+Ak5+CewW+ELNDCv3VF4Af3wd2GzYTB6EDxIJck2/GfQ3VzX
pnzvb9D7gp9B706Fk8G/GY6+ozzBfijH/4AtgmuQxk9gProahpjXVLVUDZ9D
H/yd2A7QG10bY2ntXoypWFnQF/0OeyD4Gcaan1FnOwLj8Meb61yYiJ4Ex9Cz
ob95DVW73TGXd+O/ldP3MEo5jWeYYN4T6gXF+AgZ5r1wCEabr0kfhDH42RY1
jETX6Nxxz5PNe0C1r4q9cjvmWj2jWt6PtVFNc/G7wZbgNa2JOa2OMZX7Xua9
pRoE9Cw4gz4LI9AzzHu9HOqg58K54DOj2ayEW8FndBj6dfDZSgev/Z+o1QMd
ef4t+GypZzPxP8Xe00wMMM+xcqs7QHvvaj4bOkORZgkuoBcH/3dD81nXHjS7
j4PfHZph3QVf47u6EzQ7xXAx+AwtUG+Z1/Z/TPM7SXeR3kmhZ8Ka4DXN117M
e0M5aG4+M5oVzVwL8xwqd/pmjmKb11I5fRtzqlyqp/8BHFSjSA==
             "]]}]}, {}, {}, {}, {}, {}}, {{
          Directive[
           RGBColor[0.368417, 0.506779, 0.709798], 
           AbsoluteThickness[1.6]], 
          LineBox[CompressedData["
1:eJwt0kk3FQAYx+GLikLoXtLSju9Q3wEfwdCcW0gUGYqSNM8TRZSSFBVlSlSX
SGhaWbdta+G551g857yr9/wXv6yCcH5JTCAQ6CWW6J1ODrvIo5gqWmmjjxEi
LLPCP/6zygZPkggSIp0MtpPJDhr4xSme0E09I0QHlXGfOn5yiMd84z0fGKaW
qyxzkHvc5SRLHKCTWd4xxFtquMIi+ynlDtX8YB8dXOIvl7nNCRbYyyNmGOQN
AxznIt/ZzRFa+cMFqujnNa/Yw0M+84VpInzlFpXMU0w7LfzmPDc5TDPnOMYL
XtJHEW1M8omPnGWK69ygiTNU8IxenlPIA8aZYIxGrhFmlNM8pZwuejjKHAXR
jVqopIhcdpJNiCDbSCOVFLaSTBKJbGEzCcSziY3Rvohb73UNscFaxQ==
           "]]}, {
          Directive[
           RGBColor[0.880722, 0.611041, 0.142051], 
           AbsoluteThickness[1.6]], 
          LineBox[CompressedData["
1:eJwl0mdbjmEAx+H7qZQVRZkZFTLTEBkpNIWGTYooIpUR+ja+Ct/ASrR3IYSI
rNPhxXn8/td1PS+eF3dibXNlUygIgkc8DguCToeHOsOp8CDYyz62kcqA+ycs
s5/qTf2pK3Q5p9lPLjmMecvTX9qqv3Wc7XYa6TxzPqBvNEMzGbSfs9J+oU36
R1dpAmc4yyHeuj+o7zTQZtrtfJ3QLN3JsN3BGvu13tJIJu1EjeIes/nkbpdm
s5su52Tt0TsazZS9XhfwgIV8dXddb3CVBq6xgUaOc4JKvvvtHj3JD3ujbqLf
TtGtbGYLFZQz7W2R3qXXHiXGbiOWb85HtYxj9Dmv08tcoZZ66ljKF++3tVuT
dB73mc8SRtzP0bl8tg/rEUrpdF6rNVykmllcIo54wongo98WaQnFvHJerVVc
4DxhLGbI2zkNUcB75x1ayAe7RV9qDmn//jdxRDLt+x1njFFGGGaIQQbop49e
euimK/T/+/8Ln2Nc0Q==
           "]]}}}], {}}, {
      GraphicsComplexBox[CompressedData["
1:eJzlWXlUk1cWj5W6VOOGVg3ua6xYWzQuFb3YOHbR2mBRWyqxUrFiQAq1tGHc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       "], {{{
          EdgeForm[], 
          Directive[
           Opacity[0]], 
          GraphicsGroupBox[{
            PolygonBox[CompressedData["
1:eJxNmHncl2MWxp87u8iuUEivVirtUiRjGUQIIYpsSSUtIu0iyhIpDSFLSais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             "]], 
            PolygonBox[CompressedData["
1:eJwt0/dL1lEUx/F71PayjCapj0IQBEEQIoJto4iGmVY0xTIqkvYOWj9YUdEE
kzYVDWhPgkpo/d5/EEU0KYoKo96H8/nhxbn36/M83ns+55upbahcnpVSqkAO
xlpKTzAYTf6c+hRDcI79DeoS/GHdnroVW/CR/ScsYn0VA5CP86jFN/7WnToJ
w9HKvgN1CkYjC52wEzuwJsVnRrE+jT56VkNdgYTeGIFHGIg6Ho+hPsYg/x32
M6mzLc6WS52MkfjHviP1Nm7hLfu/uMf6Lt6n+B/j0IKhfi8eXaIuxA/WPagb
0WixbuHZMuozDFOP6qiV6KLvLKBORWf1ZDF1GrriF/tdFnfws3ejXsQBfGff
T9nMwecUGfldZ+FDijt7VnPxJUVm3pvpeJeiRystepKrnm/DdmTjK5+5qYw8
mww2YL5Fdv6Z3dijs77GIdaH0RY/2R+lHrOYjd/sT1lk5ln1QhWaLXrtZ/K7
T0Q79eCOMvJsinBfM+KzUYxX6rn3upz6EuuUTZmymqe7eGYvfG4sZreUuh9n
0Vc98yzGI0eZeFYT0EaZ7cNJ9NRMrKY+R4nFzO+lHkeeZmi9zlSmDDbpzOXK
cK3OVKoM/e7XUKgezLDomffKZ77JIpOM/sd1HFFv/Ttn1KNi3cnPtsri3fUz
+iwuxZsUM7nZ4h3NV2Z+18vorzv7u1uPAot3+CEeWGRdoF5dsJhF79kVHFT2
/hsnlGGRetaAaotZ9nfuP9rfgdI=
             "]]}]}, {
          EdgeForm[], 
          Directive[
           Opacity[0]], 
          GraphicsGroupBox[{
            PolygonBox[CompressedData["
1:eJxNmHe4z3UUx+/vfCoUl6yLyMgohbqurEQpSiIZUUkRbaU0tGlvRYqSor1F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             "]], 
            PolygonBox[CompressedData["
1:eJwtlPlrz3Ecx797v5HcRzS3sSVXbrYcM6YRm9kaiwlznxOJ3HLfpdRKKS35
CzT3lTP3ETJEiTRXrnJ7PHu9fnj0er4+78/xfj9fr9cnpbSsYFFIJBKFUAOa
x0SiG6znYjaMQGdDDnoSdESnwRp0FtyG1762Cfqw1g+2oEfBD/gJ06EAhrD2
hzgbdkAt8pqQgZ4Gp6A3+WZiKlyAYvJDehbGovMgH10Go9Gf/Nn5MJT8EnEG
7IS25G1gJXoVvIJbMAcmQh3WGsFidCZ8haewwPc8nLXLxJmwC+qS14OFfk/9
aO/Qs7rWAp0My9ErYAy6drS97YWAPkPs7nsYRp4VzetiOC6fyPcTk2EgOjfa
s6XQBN0YlqCXymP0FeIs2AMjySPMRe+Gt3AP5sEUaMhag2je6cxHoWm0d1Ul
mVcP0NuCefYRSrh2mJjne7upvgi2xyLy+6oDlEPXaDWQ9+qh1Gg1V63VM/no
lzo37PPaVnlvqMbjyR/5tw7ABPInxHFwUH1G/o6YDhVwEZpxbRmxmv1fI3Ym
L/JvqBbXieuC1aRXtJ5Vr26Ex+6xvFVPtUc/I26H1XBDtYpWG51Z3l/1NdXg
O3xTnSEXurBeqT7wb3aKNjOalbXBai/Pyr0H5O0H4qBgHveI1nPqtQ0wIFqN
Vdut0DdajVVbzdgdP5POopk55x4ecI9+wy/3ppB43j096N9Urathqtf8C0xm
/Yh73h99N9gs655W5D2j7U3/hNbo9Gh7Uw8MRj8P9i3N+HuvWYV7pF497b2o
nj0L7aLNZob36kP3Uj3bgbWUaL2pGr+AzGheqsYnoGW0WUsLNqsng822Zlb/
gmM6c7B/Qg73vpHvwWak0ntyn9fsH/wN9u/RM0msJaK9u8S9+RzsXySP/gOZ
lJeT
             "]]}]}, {}, {}, {}, {}, {}}, {{
          Directive[
           RGBColor[0.368417, 0.506779, 0.709798], 
           AbsoluteThickness[1.6]], 
          LineBox[CompressedData["
1:eJwt0ldXDwAcx+G/vUXlBXSn96BjFS0tO6shpLSolOwoGS0RySyzISOiIgp1
5aobN6EX4vmf4+I553f3u/h+IrIK0wpmBAKBQWYSvFcQSRSp5FBGNQ200UkP
/QwxygQ/meQXU/zmD3+ZJs+Db+Rzi1YO84pN3OQGLfTxkkQ2c4hecrlMHR/4
SA8H6eYAVYyRwHWaaaKLF8STxn6ekcNzsrlEJROcoJEGnvKEOFLJpJMsHpNB
BaPEco6z7OMrG6mlnHGOc4a9fGEDNZTygzJOs4cRYrjIUb5zjJOcYhe72UkJ
n1hPCgM85BH1dJDOZ6K5wDYKGWYtyfRzl3tc4x3vKaKY7dxnBw9Yx3mqOcIg
q0niNbdp4wpveEswvjts4SrtbGWINcF9NJPCKlYSThihLGcZISxlCYtZxEIW
MJ95zGUOs5n1v/V/oq9b8w==
           "]]}, {
          Directive[
           RGBColor[0.880722, 0.611041, 0.142051], 
           AbsoluteThickness[1.6]], 
          LineBox[CompressedData["
1:eJwl0mdbjmEYwPH7qZCECBnRsLKFhq3MhiJ79KYPwOfwUXyACqFkaz497WxZ
JSIz8/ccXvyO/3le1/3qPq6MqvMV50JBEDRwMSYImiwX9BHpsUFwT2foTJKp
tmdoJrOYzS1nNfzkF0/Y5vyxbtffukOXsJTFLGcZO3nqfoXe1yx9oM84ZK7V
ldqkc3Ue87liDzTEG/M+fa37NZYDbCKHjYTd5eqQntZh7WSheRFp1Nsn6EQ+
mot1VEs0nlIOspVe51u0T8u0nGlMJ4kEppLIFB767jOnzDd1nB8McNJZo07W
T1qpY/qdfk7Yb+gkfa9n9YN+o4fj9gaN03d6Rkf0C1/popsO3nLU/TW9S5E5
j3xS2UzE+TG9rjHaqoMcMdfpHXab17GeBexhLy/dvaKNbHu7btCr+oe/POew
s0t6m13mVaxmDgUU8sLdGm3WFF1Li7lCL0ffQ/SfRf+lNztEP3300kM3XXQS
oYMw7bTRSgvNof9v/x+FAGY5
           "]]}}}], {}}, {
      GraphicsComplexBox[CompressedData["
1:eJztWXlU01faDtZ1WhgRWiVUbG0dg9W61KC1ra82uI2WglJbUeIy1VqIS7FV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       "], {{{
          EdgeForm[], 
          Directive[
           Opacity[0]], 
          GraphicsGroupBox[{
            PolygonBox[CompressedData["
1:eJxNl3eUltURxu+1Y4sVFBusAiKiuKhsBFbpICvIIlIEFhcVpDelqoiySIkV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             "]], 
            PolygonBox[CompressedData["
1:eJwtlHdsT2EUhr/T1I69t5qxI/YeqRFSQQhqVW2p1h6pLTbVaqlG7L33DEIE
IRojCDFix0hEJEiI8bw5/ePJ77nN7b3fPec9Jyo+qU9iRAghDiKhj4UwAs7j
K2AOXhvW4buhL94PTuIZ0AsfCmfwJfAHtnL9RfdCPXw+PMafQG88Ds7iS6EO
Phce4A+hOh4NR/A0qIZPg9t4NvyCSVzf4DcaKuItYBt+C8rjTWErfh3q453h
GL4eyuKj4AJ+CargVSEVPwHN8Ez4gH+E1vhe+I6nQA18JtzB70IZvDFswS9C
QzwFXuKv9U7VDlYGv+cfbOf6K78joQI+Ea4F/6aBeCycwjOhPz4A9gTvQRRe
GA4HP/NqPC/sx19AE7wgHAxeg+Z4ITgUvEZF8AA7g9d4LV4ADuDvgj/7L787
gr+jBNe5YV/wMxXFDXbhRyEL3wAxwWvW0bwG+nadaQKeoB7hV6EDvsA8C5+h
OJ5LNQ7eo2J4hHnWlLFO5j1RL/RNrfAe5r2KhW54V/MsbIIxeDm4gl8O3mv1
SL1Rz1vileC4+gDdzTOrrOpvMVz3hHQ8AcbiP1XH4M9UrX7kvEs1G8f1ePN3
6RuHm2dc2T4HJfE8kIWfDv6s0ebZ0zPVm985Z1GPapnXWLXVGUrh+c2ftQxK
4yNhc/BMRaoX8A2fC43wDHgf/J5U8xnVbL6B9uYzp1lTpqbgU+Fm8B7OwKeb
Z1MzN9l8xjRbukdZ1YxptpTZVfggZQB/Dm3MZzY7J0MrzDOtLD+DIeaZVVa1
M/S/g81ro2csU+3Mv+UpLMIXQnLwHaJdURceBd8Zi/Vt5vfqHmV1pflZlNlh
eBq8Db5zlN1kuB88w5rlmnAv+EzPM8+osql3bDTPhLLwCWbh7cxd/5NuvjO1
K1XzePOZ0ixph6o32kHaPerRbPMdqt2pM2i3KNPKsnaMdmdl812jHdrWfOdo
12hmtVsawKvgOyYRTzKfrZawBu9i7rpH357PvPeqgXaPdqx2q3aQZm+5eW80
g/8BNXLIZQ==
             "]]}]}, {
          EdgeForm[], 
          Directive[
           Opacity[0]], 
          GraphicsGroupBox[{
            PolygonBox[CompressedData["
1:eJxNmHuQz2UUxnfPFyFkc61UKJXQPaJEIRSJUKII5dbQdJ+m+1Qz1TQ1Tfea
Zqppyv1+X/frIvfdlqy1drGEtS65Lp2n8/kNfzxz3vf9vu95z+U55/3Nr8HA
UT1GWlpa2vD0tLTEZVefdHGc8Pl8x1jHOEcjXxvjMtMxLz3G+rbcMdUxzbHM
MYX5As6Ndyx1TObbQtYmOBY7JjomOZYgtW8R3/Utm336voJ7pjvWXWBLlmOG
Y6ZjFXKWYzVytmMNco5jM3dI70r06fx69MnvDUj5sREp2zchdf4GCxvWOnKw
Ub7+hVQ8LvE9Z13mOioTV9moGN9jYXsuexXLi33tJLbfaeHzFsdpxyn8OODY
7djj2M79ikcTi3GeI581+bQDKZ9aWvi81VHAmvzbiZR/zSx06I4qPj6Df9V8
XIZ/91rkpgg78rBlL3frzuq+55zLQrCJ2KdZ6N7vKOaczvT09fYWtuxDRwHf
i7nnILKYmOViz6V+zizOpVvoPuQo5R7FbJCvD7TQMYZcK39jybXmF/n3Co7D
Pn7KZXnHER93ctnRcdTH11nUw1wfJxb3aM/RC+4cB2ekdyq5lb0T4I44JH4v
Zp6qAc0nwx/NUzwWZ8ZzTnpVByvQK/6uZC4eZzFP1YPmU7hfelP1IL2NLXj8
J7FYB59SXN9IXNbzLVUb67F7M/tSdbIBH7L5Jr5twbZUbYhL4nYu9ve14E0J
PudwXvWxFV9utagf1cZtFnVyClvFjzzsy2ee4n0+NhUwb2pRJ9vSz/Ne38Tn
wgv8KeSb9Ipv4qvqoAjbbrKok10+vsOiTsqwWbX9N3fLL3F1BPem9ImD4mR7
OCsbelnUwDlicpCzu9krHovjJeitAB/FN/G6lH3ioNbFyX2c2cv5Ur4PtqiH
chY8F7ePwZcTcEa5KiMf4sspOJNDLLKZn+RcQ9dR0XE8PXJ7Fh01LWryEHE9
R2wzLHqB7LrfYq59Tzj68E22dKFPXuGyBnXeD9sVowHcK/sPExf5Wc9lber5
Skcd9tW3qHPtbeCyEjb3cHkZuZRP4tsq/Bbf5Gsti5pXbB9yWRdeKk46l4pN
VQuuKAbqm4pDW4ueqBg85vKqC3zOIA4PWPQy5a2FxfuRSw5rkHvdX5N4Pu6y
HXPZXpV+rb5VCzsr4vtxYlMb3uyAH3vxtS7xlo2Xw0Oda8BZxawe8czAfuVo
qGMIfUQ2Xo2d11hwQmevt+ibc7GnIfl61uUzFv10CDITLl3LnuYW/WIL+6Vn
dHr0FK2rhwy3qG31MdnTmLu0Pozau5F17RmGveofz1nUp+pZ7+ztFjnvw5rq
T3fezL3ScQv6dVa9YBu2NCdfdzGWzR0s8qpYt3J5N+vqja0Ya28LzmbRR9T3
VrJPvra2WFdvEUdUD+KF6lD9azPntE99s43FehHf28A93dmSPfexXsT+tuwR
PzrCH9mvPvUP/aoDvuxkXkANav+/Pu5s8V5Kh/q26qQyHG0H36vA1zPc04mz
ivGDxFk87Yye7i4fsajDn8WfJOr2LfEwCZ5+7ThrUR+Jr5VPQq/4rX6qXqq+
0Rd+znD5vUWtKNfqO7vw5WkLnmqvzqjPpLFPfBBXh8KBxozFp0WOxRa18SRn
S+hhXcmd7kxP4t5yLmsn4fsn4hHcKnGUksvXlesk4vG8o04Sdf62Y5ZF3b7i
WID9k1zWT4IjryrOCXXp8geLOtb93+H7MccXFlz4xjHHorcecrwAT+Sbaka1
8ys5qEQefmF8wnHcopefdJwi36n3Sf22CXrWolNxzORN0NsnTn2sOFlw4SBn
DyAHoUd3qIdmwM2e8FOyF2PlcAC86s267lG8uln0OuXkYfKiOlY9N4Nn4pve
Ab3vPeCsevmjFv1ev0O6s7cb+ZXOEehRT5AuzfXuy/bB2F8OX+Sf+p/62mj6
jMaN7HxPFMc01r4x9K7h5KIp4yZgBPf2g3+6J9fxqcVbJHv6wBPVnd7+I4wV
K/Xb/qxrLNkfnvwIh8QfvQ+9iac4cgCeHFTsLXrMEQset4aPEy04udzxhsW7
sMpllsVbIR6tgEs7HPkW/WOrY4tFftZYnNF+8XqyBbfF2an4+5ujYhJ5/cmx
2uItUQ1Nt6ijMsdpi16W5xhJPLNdbreI605HgUWe35FdSXDiXUehxTvxovIJ
fz53ud+iv78vviTBlY2ODRbvzzrHWov38EvHEov38jXHGQveqr5/t6jxHAub
ZM9cx0cWv2UUj13EZI9jtwVPixyj4ID64kyL3qi++JVFbzRfn2bx+2Cp41uL
30TVkuCKeDLG5WiL3x/V1UuSqJ//15JY/9AxGy6Mc4y1+O1QNQn/5bt62zKL
Gszw8UJ81P36w0E2vOT4wIKTtXytZhK/a/Y5PrN4S+XTe/i1kHhJj+I0n1gt
QJf0ZJIb5aXY5R8WtTvP8bLFeye73sS2/wArOF5T
             "]], 
            PolygonBox[CompressedData["
1:eJwtlGdsj1EUh9//fYsWpbSl1CpCiKAIQsQKUUKpGJHQEF+tD4i9Yu9NzMSI
2CNWisSIEWqH2LNWzaLV1np+uefDk9977z3vvefeM1KGjcoY6YIgGAJRcIvB
TWgADWE/JIVBMEhGMBZ+Qkd4HQmCFegPaG9z1yEb6kF90Oa3ba9s7MuzVzn4
q2/mJkBVxoPRL8zlo7+gE0RgOeyFOOgAK+EgVISusBW2QQykwCwYDbFQBS7B
RSiIeJszcNp8l80WyILqEA0j4Ib5noXNFfQy1IY6sA42QimoBVP0L/4PRfOx
34SGjLuhaTAGPkBreMD6UvQ9tLS5mXAPGsFj1uei4yEekuEOPIdUKGT9MHrB
zs5jvAGdB9XMp3+QyPkD0QSYpDXGmYpbxNtOhEr2TzxrZ9EaMADOW0wUi5ro
IVjl/Ld8+mh30l3aoidhH/RXHCEXljm/1sbuugRa2Z3PWU4FduZ8mAqVLQYH
nH+DZIvxHphjeysHSuLvNDQJesBVu6PuVtditdn5b8UsR7G2t22MHrecSbEz
n8IT53NTMZgM1yzWypkCeyO9jXJyNjx0fq+ciK+VhdDU+ZopE/ocVO5l6FzL
KeWSbBbBDGgOxcwtRt9BOedr4hV6F4oi3ka1tACaOF9TOxVnzuiJpuuN+K4A
udj3Y/zGfNDZLeARbIfutsdR53M0zXIgGttTsuP/Xugze2O9baqdtcP5NZ25
HhxziWgJwXdU6HNfOZ8H36Gd7am7vIRmdqcA2zV8h87vcUSxYG44GmW5FBf6
uyin7sNbu8sf/j9hb5BuObDa+ZxPsD2Lne8R6g1d4Bt8Vcwi3if1snEWO/U0
1cYxu0um+aIaU23JJ9W2eoR6g2q8bOh9ki99ne9da8139bBdEBv6NcX0BZRm
3MfeQN/T0d7Oz32Cz87Xinz8rb3Nd8VUvVU9UL1PPTaB/wvRzs7XeJHzPTNi
czGh7xnqFTpDtbfb+V6rGvwPPj+w/g==
             "]]}]}, {}, {}, {}, {}, {}}, {{
          Directive[
           RGBColor[0.368417, 0.506779, 0.709798], 
           AbsoluteThickness[1.6]], 
          LineBox[CompressedData["
1:eJwt0/dXT2EcB/Dvt6VJU3ZCFMleGWVkJipZkUQIKSIikops2QnZ/Ff+i+zx
uocfXud9z/s8z+ec+9z7ZNc1VZwIh0Khz0QQPGeQRyEllFFFNbXU00AjzbTQ
ShvtdNBJF930M8zgDPZQzS1uEkM6u9nFDa4TTRolzOQ9UWzhBdfopZur9PCc
K3RxmXw6KaaDGVxiOkW8I5Kd7GA7eVwkglSq2EY7FwiTwjTOk0sZmxlgBW1M
5Rw5LOdt+N/hJrOGKbzhj7pUPqOVs5xhMst4zW9rRshsPjCJiWyikE8sZTVZ
vOKXPRvlBJYwyA/dcLmIjyxmPKc4TQs/rdkgn9LPExayiiTms4CxNDOOk3y3
b51cz2xeMqRLkPep4wAVPCCRh8zhIJlU8ojHzGUeo6hnNIf4YuYROYbDHKWB
r/rjspFjNBH82N/0a+VdCrjDLFYSy232UsNW7hFHH/uopZz9xDOSXjNzSSeN
VFJIDr5TcMYkkRi8P/HEERvcA2KIJorI//fuL2TsTkE=
           "]]}, {
          Directive[
           RGBColor[0.880722, 0.611041, 0.142051], 
           AbsoluteThickness[1.6]], 
          LineBox[CompressedData["
1:eJwl0+dXzQEYwPHfLYTsvUL+HC/9Cc7xUpEZMrJHKDsJZbX3RPZI90Yke1Nm
tlDGx/Hic77P89z7uy/uuTd+euK0hFAQBBFmRAVBnmWqHuIeG4jQzFVmcp+N
XKOF6yTwgIfcpI1UEnnOC+5wl9XM5RTzGRMdBMV6ljMMt3fpd77xkx+McB/J
H/MOfrOTwK1So3SPhjRTq5lgXqxNZHOZRq4QZgkXuUQOU7z/oOZygBomui3V
9RxnP3FuKXqBUeZe9CaafW5VjDcv0nXsZZI9S3dznnLi3Sq0nl2MsyfrObZT
xli3BbqWEibbS3Ub3fyih4WcZqjXhzGEInsdayigkHS+kMFX5nGSBo4y2nOz
dCWx5nbt4BYD7Kt0oB7Tl6Rxm0Fug8kz57OZV2zhNbN5y1be0Mk7kvjAR97z
mU/M4QR9feZhjdEjWssy+tgf6WNu0M++XPvrE31KK89YwaZ/35ffd5h66qil
hmqqqKSCcsoopYRiiiikgPzQ///LX6XnflU=
           "]]}}}], {}}, {{}, {}}}, {
    DisplayFunction -> Identity, PlotRangePadding -> {{
        Scaled[0.02], 
        Scaled[0.02]}, {
        Scaled[0.02], 
        Scaled[0.02]}}, PlotRangeClipping -> True, ImagePadding -> All, 
     DisplayFunction -> Identity, AspectRatio -> 1, Axes -> {False, False}, 
     AxesLabel -> {None, None}, AxesOrigin -> {Automatic, Automatic}, 
     DisplayFunction :> Identity, Epilog -> {
       InsetBox[
        FormBox[
         StyleBox[
         "\"\\!\\(\\*StyleBox[\\\"c\\\",FontSlant->\\\"Italic\\\"]\\)\\!\\(\\*\
StyleBox[\\\" \\\",FontSlant->\\\"Italic\\\"]\\)= 1\"", FontSize -> 13, 
          GrayLevel[0], Background -> GrayLevel[1], StripOnInput -> False], 
         TraditionalForm], {0.63, 0.55}, {-1, 0}], 
       InsetBox[
        FormBox[
         StyleBox[
         "\"\\!\\(\\*StyleBox[\\\"c\\\",FontSlant->\\\"Italic\\\"]\\)\\!\\(\\*\
StyleBox[\\\" \\\",FontSlant->\\\"Italic\\\"]\\)= 0.8\"", FontSize -> 13, 
          GrayLevel[0], Background -> GrayLevel[1], StripOnInput -> False], 
         TraditionalForm], {0.63, 0.385}, {-1, 0}], 
       InsetBox[
        FormBox[
         StyleBox[
         "\"\\!\\(\\*StyleBox[\\\"c\\\",FontSlant->\\\"Italic\\\"]\\)\\!\\(\\*\
StyleBox[\\\" \\\",FontSlant->\\\"Italic\\\"]\\)= 0.6\"", FontSize -> 13, 
          GrayLevel[0], Background -> GrayLevel[1], StripOnInput -> False], 
         TraditionalForm], {0.63, 0.165}, {-1, 0}], 
       InsetBox[
        FormBox[
         StyleBox[
         "\"\\!\\(\\*StyleBox[\\\"c\\\",FontSlant->\\\"Italic\\\"]\\)\\!\\(\\*\
StyleBox[\\\" \\\",FontSlant->\\\"Italic\\\"]\\)= 0.5\"", FontSize -> 13, 
          GrayLevel[0], Background -> GrayLevel[1], StripOnInput -> False], 
         TraditionalForm], {0.83, 0.035}, {-1, 0}], 
       GeometricTransformationBox[
        InsetBox[
         FormBox[
          StyleBox["\"drive invades and fixes\"", FontSize -> 15, 
           GrayLevel[0], StripOnInput -> False], TraditionalForm], {-0.18, 
         0.5}, {-1, 0}], {{{0, -1}, {1, 0}}, Center}], 
       InsetBox[
        FormBox[
         StyleBox["\"drive doesn't invade\"", FontSize -> 15, 
          GrayLevel[0], StripOnInput -> False], TraditionalForm], {0.55, 
        0.75}, {-1, 0}], 
       InsetBox[
        FormBox[
         StyleBox["\"polymorphism\"", FontSize -> 15, 
          GrayLevel[0], StripOnInput -> False], TraditionalForm], {0.47, 
        0.07}, {-1, 0}]}, Frame -> {{True, True}, {True, True}}, 
     FrameLabel -> {{None, None}, {None, None}}, 
     FrameTicks -> {{Automatic, Automatic}, {Automatic, Automatic}}, 
     GridLinesStyle -> Directive[
       GrayLevel[0.5, 0.4]], 
     Method -> {
      "DefaultGraphicsInteraction" -> {
        "Version" -> 1.2, "TrackMousePosition" -> {True, False}, 
         "Effects" -> {
          "Highlight" -> {"ratio" -> 2}, "HighlightPoint" -> {"ratio" -> 2}, 
           "Droplines" -> {
            "freeformCursorMode" -> True, 
             "placement" -> {"x" -> "All", "y" -> "None"}}}}, 
       "ScalingFunctions" -> None, "TransparentPolygonMesh" -> True, 
       "AxesInFront" -> True}, PlotRange -> {{0, 1}, {0, 1}}, 
     PlotRangeClipping -> True, PlotRangePadding -> {
       Scaled[0.02], 
       Scaled[0.02]}, Ticks -> {Automatic, Automatic}}],InterpretationBox[
    Cell[
     BoxData[
      FormBox[
      "\"female selection \\!\\(\\*StyleBox[\\\"(\\\",FontSlant->\\\"Italic\\\
\"]\\)\\!\\(\\*SubscriptBox[StyleBox[\\\"s\\\",FontSlant->\\\"Italic\\\"], \\\
\"f\\\"]\\))\"", TextForm]], "InlineText"], 
    Text["female selection \
\!\(\*StyleBox[\"(\",FontSlant->\"Italic\"]\)\!\(\*SubscriptBox[StyleBox[\"s\"\
,FontSlant->\"Italic\"], \"f\"]\))"]],RotationBox[
    InterpretationBox[
     Cell[
      BoxData[
       FormBox[
       "\"dominance \
(\\!\\(\\*StyleBox[\\\"h\\\",FontSlant->\\\"Italic\\\"]\\))\"", TextForm]], 
      "InlineText"], 
     Text["dominance (\!\(\*StyleBox[\"h\",FontSlant->\"Italic\"]\))"]], 
    BoxRotation -> 1.5707963267948966`],TemplateBox[{
     InterpretationBox[
      Cell[
       BoxData[
        FormBox[
         StyleBox["\"B\"", Bold, FontSize -> 20, StripOnInput -> False], 
         TextForm]], "InlineText"], 
      Text[
       Style["B", Bold, FontSize -> 20]]], 
     InterpretationBox[
      Cell[
       BoxData[
        FormBox[
         StyleBox[
         "\"   \\!\\(\\*StyleBox[SubscriptBox[\\\"\[Lambda]\\\", \
RowBox[{\\\"f\\\", \\\"  \
\\\"}]],FontWeight->\\\"Bold\\\",FontSlant->\\\"Italic\\\"]\\)\\!\\(\\*\
StyleBox[\\\"\[Rule]\\\",FontWeight->\\\"Bold\\\",FontSlant->\\\"Italic\\\"]\\\
)\\!\\(\\*StyleBox[\\\" \\\",FontWeight->\\\"Bold\\\",FontSlant->\\\"Italic\\\
\"]\\)\\!\\(\\*StyleBox[\\\"\[Infinity]\\\",FontWeight->\\\"Bold\\\",\
FontSlant->\\\"Italic\\\"]\\)\"", StripOnInput -> False], TextForm]], 
       "InlineText"], 
      Text[
       Style[
       "   \!\(\*StyleBox[SubscriptBox[\"\[Lambda]\", RowBox[{\"f\", \"  \
\"}]],FontWeight->\"Bold\",FontSlant->\"Italic\"]\)\!\(\*StyleBox[\"\[Rule]\",\
FontWeight->\"Bold\",FontSlant->\"Italic\"]\)\!\(\*StyleBox[\" \
\",FontWeight->\"Bold\",FontSlant->\"Italic\"]\)\!\(\*StyleBox[\"\[Infinity]\"\
,FontWeight->\"Bold\",FontSlant->\"Italic\"]\)"]]]}, "RowDefault"]},
  "Labeled",
  DisplayFunction->(GridBox[{{"", 
      ItemBox[#4, Alignment -> {Left, Inherited}, DefaultBaseStyle -> 
       "LabeledLabel"]}, {
      ItemBox[#3, DefaultBaseStyle -> "LabeledLabel"], 
      TagBox[
       ItemBox[
        PaneBox[
         TagBox[#, "SkipImageSizeLevel"], Alignment -> {Center, Baseline}, 
         BaselinePosition -> Baseline, ImageSize -> Large], DefaultBaseStyle -> 
        "Labeled"], "SkipImageSizeLevel"]}, {"", 
      ItemBox[#2, DefaultBaseStyle -> "LabeledLabel"]}}, 
    GridBoxAlignment -> {"Columns" -> {{Center}}, "Rows" -> {{Center}}}, 
    AutoDelete -> False, 
    GridBoxItemSize -> {"Columns" -> {{Automatic}}, "Rows" -> {{Automatic}}}, 
    BaselinePosition -> {2, 2}]& ),
  InterpretationFunction->(RowBox[{"Labeled", "[", 
     RowBox[{#, ",", 
       RowBox[{"{", 
         RowBox[{#2, ",", #3, ",", #4}], "}"}], ",", 
       RowBox[{"{", 
         RowBox[{"Bottom", ",", "Left", ",", 
           RowBox[{"{", 
             RowBox[{"Top", ",", "Left"}], "}"}]}], "}"}], ",", 
       RowBox[{"{", 
         RowBox[{"ImageSize", "\[Rule]", "Large"}], "}"}]}], 
     "]"}]& )]], "Output",
 CellChangeTimes->{{3.814006148431986*^9, 3.814006157968115*^9}, {
   3.8140061879904633`*^9, 3.8140062736784477`*^9}, {3.814006431513706*^9, 
   3.814006461889913*^9}, 3.814007135210734*^9, {3.814007168715795*^9, 
   3.814007181841032*^9}, 3.814011877087171*^9, 3.81401199119109*^9, 
   3.814085984015634*^9, 3.814086040256649*^9, 3.8141965724436617`*^9, 
   3.814197129329672*^9},
 CellLabel->
  "Out[1524]=",ExpressionUUID->"b0473242-a123-4e0d-b37b-a6bae7a6eaa9"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"Style", "[", 
  RowBox[{
   RowBox[{"Grid", "[", 
    RowBox[{"{", 
     RowBox[{
      RowBox[{"{", "boundaries1", "}"}], ",", 
      RowBox[{"{", "boundaries2", "}"}]}], "}"}], "]"}], ",", 
   RowBox[{"ImageSizeMultipliers", "\[Rule]", 
    RowBox[{"{", 
     RowBox[{"1", ",", " ", "1"}], "}"}]}]}], "]"}]], "Input",
 CellChangeTimes->{{3.8096815126894083`*^9, 3.809681536773089*^9}, {
  3.809681724452579*^9, 3.809681732370373*^9}},
 CellLabel->
  "In[1525]:=",ExpressionUUID->"8676f9bc-6683-46c6-b546-1e57bb6e6b69"],

Cell[BoxData[
 StyleBox[
  TagBox[GridBox[{
     {
      TemplateBox[{GraphicsBox[{{
           GraphicsComplexBox[CompressedData["
1:eJzlmnl0zPf6x0eo2kZ/1LVMItq6YmjsEku1T2KI9l5iJJYrdGqp/DD2uhhL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            "], {{{
               EdgeForm[], 
               Directive[
                Opacity[0]], 
               GraphicsGroupBox[{
                 PolygonBox[CompressedData["
1:eJxNmHn8V1Max+8pSkkpVCIpJJKUSosoSaVFq2gPIaTs+26KjApZspMlJmMJ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                  "]], 
                 PolygonBox[CompressedData["
1:eJwt08tLVVEUx/Gzo5l/gWTi1dQyLUNJDC9KoihBoIJIIKFQaJGm5ruHZqHV
7eGjgjTn4SMte6ulKXYdKImYNGnc3+DE72L/Bh8W556z9157rXVDtY3lDQeC
IMjFQdS5IPiNCvzj+SbxFTIRwgCWUYJSvEQEh3EC7XiIVOSgD39QjSPowhzy
EMa01ti3pzGKMRxHIt5iRu+myWeC2IZsZGEQf1GDBHTiCY7ijHLbxUXleANv
kIxDuIRNlGGe/R8Tm3BSe0xiSmeN836EeB/FyMA17OACYrCN28q9ClvO19Bq
V6lve5CkNVabu4hVjSzXR8rFcv6FNUR5Pme/45vOOosh3EIB0lCvM+2s/6xZ
If5U7axnH/FBvc1T7Z7pW6vhEn7orkXEYdxBoXryBZ+xx/t84j0sqhZWU5ud
bvXCZiiKdedzj8cGWlX786rFa5xSTewun5zfu0C9uIxj6onl8lV3t5xe4Dni
kIJanWF7RwI/K/16ZzOzgF7nZzFda1edn2XbY9b5GbLZydVsdCh3m5Gnzs9w
WDk9wHXnZ99m/h3ea631rJl4Rb2x/1ALGnS23fmq8zNTpZrabH13vvY2Y/uX
hmnu
                  "]]}]}, {
               EdgeForm[], 
               Directive[
                Opacity[0]], 
               GraphicsGroupBox[{
                 PolygonBox[CompressedData["
1:eJxNmHmYz1UUxmfOrVCRJWPfyVJkbI09MlqUGUI0USSppFWhon3TnlZt2je0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                  "]], 
                 PolygonBox[CompressedData["
1:eJwtlOmLj1EUx5/fPUPW7Ip4MYksZc0WYwkx1pkh+zIzDGOY+Y2xjOxLIska
IgwisiQ1JSX5G8goa8g2sq9J+Hyd++LT+Z77PM957j3LzSxM55WFJEnyIAPq
W5K8ZGEtrIA8/ClQhS6FReglcAm9Gh7CQPw92OnQGt0IimOMpuhX2HWwCtri
/8bmw3p4BBWs3cBOE+g25s9OQRZ6KOxFF0BLdAGcRy+ExujlcBNdAp3Rv+Je
N2tP+O1gA/o0tEf/xRYGX/sBzcz3dhxamf9DsSvgM9TTP7CH4Sl0MY89L/hZ
SvGvxDP1RPeA7egdsBTdCTahL8N3aIK/ErsxeG6eQGXMUYn5O3qmHGuvZaxd
jXuui14G11UL+Ak5+CewW+ELNDCv3VF4Af3wd2GzYTB6EDxIJck2/GfQ3VzX
pnzvb9D7gp9B706Fk8G/GY6+ozzBfijH/4AtgmuQxk9gProahpjXVLVUDZ9D
H/yd2A7QG10bY2ntXoypWFnQF/0OeyD4Gcaan1FnOwLj8Meb61yYiJ4Ex9Cz
ob95DVW73TGXd+O/ldP3MEo5jWeYYN4T6gXF+AgZ5r1wCEabr0kfhDH42RY1
jETX6Nxxz5PNe0C1r4q9cjvmWj2jWt6PtVFNc/G7wZbgNa2JOa2OMZX7Xua9
pRoE9Cw4gz4LI9AzzHu9HOqg58K54DOj2ayEW8FndBj6dfDZSgev/Z+o1QMd
ef4t+GypZzPxP8Xe00wMMM+xcqs7QHvvaj4bOkORZgkuoBcH/3dD81nXHjS7
j4PfHZph3QVf47u6EzQ7xXAx+AwtUG+Z1/Z/TPM7SXeR3kmhZ8Ka4DXN117M
e0M5aG4+M5oVzVwL8xwqd/pmjmKb11I5fRtzqlyqp/8BHFSjSA==
                  "]]}]}, {}, {}, {}, {}, {}}, {{
               Directive[
                RGBColor[0.368417, 0.506779, 0.709798], 
                AbsoluteThickness[1.6]], 
               LineBox[CompressedData["
1:eJwt0kk3FQAYx+GLikLoXtLSju9Q3wEfwdCcW0gUGYqSNM8TRZSSFBVlSlSX
SGhaWbdta+G551g857yr9/wXv6yCcH5JTCAQ6CWW6J1ODrvIo5gqWmmjjxEi
LLPCP/6zygZPkggSIp0MtpPJDhr4xSme0E09I0QHlXGfOn5yiMd84z0fGKaW
qyxzkHvc5SRLHKCTWd4xxFtquMIi+ynlDtX8YB8dXOIvl7nNCRbYyyNmGOQN
AxznIt/ZzRFa+cMFqujnNa/Yw0M+84VpInzlFpXMU0w7LfzmPDc5TDPnOMYL
XtJHEW1M8omPnGWK69ygiTNU8IxenlPIA8aZYIxGrhFmlNM8pZwuejjKHAXR
jVqopIhcdpJNiCDbSCOVFLaSTBKJbGEzCcSziY3Rvohb73UNscFaxQ==
                "]]}, {
               Directive[
                RGBColor[0.880722, 0.611041, 0.142051], 
                AbsoluteThickness[1.6]], 
               LineBox[CompressedData["
1:eJwl0mdbjmEAx+H7qZQVRZkZFTLTEBkpNIWGTYooIpUR+ja+Ct/ASrR3IYSI
rNPhxXn8/td1PS+eF3dibXNlUygIgkc8DguCToeHOsOp8CDYyz62kcqA+ycs
s5/qTf2pK3Q5p9lPLjmMecvTX9qqv3Wc7XYa6TxzPqBvNEMzGbSfs9J+oU36
R1dpAmc4yyHeuj+o7zTQZtrtfJ3QLN3JsN3BGvu13tJIJu1EjeIes/nkbpdm
s5su52Tt0TsazZS9XhfwgIV8dXddb3CVBq6xgUaOc4JKvvvtHj3JD3ujbqLf
TtGtbGYLFZQz7W2R3qXXHiXGbiOWb85HtYxj9Dmv08tcoZZ66ljKF++3tVuT
dB73mc8SRtzP0bl8tg/rEUrpdF6rNVykmllcIo54wongo98WaQnFvHJerVVc
4DxhLGbI2zkNUcB75x1ayAe7RV9qDmn//jdxRDLt+x1njFFGGGaIQQbop49e
euimK/T/+/8Ln2Nc0Q==
                "]]}}}], {}}, {
           GraphicsComplexBox[CompressedData["
1:eJzlWntUlGUan9BFWR1avGWAWB4vgyFqOKREPOp4a1McCDFFx0uJtShsauZg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            "], {{{
               EdgeForm[], 
               Directive[
                Opacity[0]], 
               GraphicsGroupBox[{
                 PolygonBox[CompressedData["
1:eJxNmHncl2MWxp87a5IS2YrwKpEUqQh5W7xKSaUikSTrEIoYUzS2KcqSspOl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                  "]], 
                 PolygonBox[CompressedData["
1:eJwt0+tPj2Ecx/H7ktMwRklOUVOTkSmNNltbzdaambEhcigzUgqFchg2Rc5n
Hjg9QB44G2ZSjq05rAeRh/4C/4P3Z9fnwWvXfrvv33Vf1/f7+WbVNK5sGJIk
SRmG4m1IkkMYg0k4iJvIxGhcxR+s9jv1eI5ZyEUdjmMCctCIDqQhD+vxGnNw
g+9VsvZhid7j9ybWnyhHL7/3sbZgFLKwET+wFF08v8Q6iFUYiQv4jRUYgXP4
heUYjla8wVzMw1q88Nlb2O8E62csQjEe4hmyEbAB31Gqe/P+R9YPfqY7V+Eb
SnCf52dYB7AMw1CNflTgL893sb7CbNfkPY6iEAuwA1uQgRmoRY17k+na17nW
6sFOvEQqUlyrZtdCNevBVkxBkc+6x2cr8d2aXKtS96bBe6lHJ7EN09zT27jl
3o1nPYXdmImJ2I/tmO4eHvaddJexuOuaqVb6z2M8wWSMw0Xcc+/7k/jtdd5b
Z3iATu/1j+dHXEPVriDELO11NpSpQe+pvZSZL7gTYhYXu9bvMN81b8MnTHVN
n7qGRc5EN45hoXumXirzee7ptRDPVOA7deGR76ZvaJau+OyaqcshZlTZTEcv
voaYlfwQs30+xNlQxpXtsyHOgjKurJ0OMZvKnLKmGdbsVrj3a0LslTKgWTgQ
4t6aCWW/PcRaaAauhzgTxa6BernZZ1NP/wMXt3B7
                  "]]}]}, {
               EdgeForm[], 
               Directive[
                Opacity[0]], 
               GraphicsGroupBox[{
                 PolygonBox[CompressedData["
1:eJxNmHl0lsUVxpM7giKbQZB9SQDZUXaJYZPIHhOwLpgKhmKrooJi61qXtlbF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                  "]], 
                 PolygonBox[CompressedData["
1:eJwtlPmLj1EUxr9zz9hlzQ+2NEhkGSLLkL3sxgzDkGGaMYuJGfuabBOyZ6kp
JaWmlJIkf4WQfc8SWbM1JMrncc4Pn85z7vu+995zz3PfnLK6wtqUyWTyIRu+
kLSwTGYN8RA8gEryRuIMGIDuD4XoDdAevRquosuhE7oWrqFXwD19R36YOAVG
o/NgN3oRPIJq8ovJ5+xtvqbW2gZfYTpjB5PPMQu9DE6j18MC9DPiRjgObchX
wRV0KTRHvyBuTl5TS/JvMfdayCHPhhL0VpiN7mxe2xHtCd0dtqCPJZ/rbnyr
OfvyrA/MRK+D+1BBfj75WAm6H2xPvucPMJl8H/EADEPPMV+rIp69J+5P/s4T
qGHsMnE+3IYy8rPEvFirlc4k+Zra6zw4mnzPBeh8qE4+pl5+j7NpjLk+EZcm
n7Oc/A5xLJyDleQ/o5eXolcdoCx5z6rQTdH7CzCQ/A1xF+yAUeTXiU1ZPjbO
3APq/R54DSPJq4hFyWvPyFNxBh9hqjwVe5QXvoQ35Ykf4Ql5QXuSF9tBQXJP
TkDfIA6FvfAWxjNWQyyGpzDIfK/N4CUMJ6+Ut+A55JLvjDMZgn4Xc2nslnxt
3svc5N5rDXOTe1C9+xW9VQ+1l1LGzsSe5N2b+i483JNnPcx7q7HH0QOdvWoa
jJ4E9eglYOhiOIE+ldzL6qF6J08/jB6pN9rTCPSrqE1n3o28q3nt8shnmEZe
l9xT6t1y8obo4UJ0kfna6tFE8zuou1cf7/6OXjZErX/jrFXzH/nYvPfqgf4V
Hc29rX/GGPM7p7umf8RidC/zf8fJ5LUmc62au5ifgWrfpB5qb+Z3WTVkyWvm
z/57CN3W/K7qTv4Dln6TDg==
                  "]]}]}, {}, {}, {}, {}, {}}, {{
               Directive[
                RGBColor[0.368417, 0.506779, 0.709798], 
                AbsoluteThickness[1.6]], 
               LineBox[CompressedData["
1:eJwt0klTDgAcx/FHlC2SeAFuunfMe5BL2WmhiEhUyhpKiVAoWVLakRSFZEmF
VgrHbm5OTin6PDMdPjP/23/mN991iRmbDi4IBAIdhBC81xLNBuJIIZdiKqmn
nR4GGOMnU/ziN3+Y5i8zzPKP/8R4coY3nKaHU9xgjGQe84hWXvOKJPZxgpec
pIF6yvlBBS00k083iaRRTR6j7OYhH+mni+O8IJfrjLCLVLIZJ4dj1HGV71zj
NkcZZie1XGGSMqrIYogdPKCUCS5TyRG+sJ0aSvjGJW6RyWe2cZ/DFPOJLeyl
iK9cJINODvGcrdyjjwt8oJCb7OE953nGAZroIBjIIJu5w13OMkA8KfTyhDYa
ecs7CjjHfp6STjsJwS21kMxGYlnPGqJYTSSriGAlKwhnOctYyhIWE0Yoi1g4
3/IcpH1ifg==
                "]]}, {
               Directive[
                RGBColor[0.880722, 0.611041, 0.142051], 
                AbsoluteThickness[1.6]], 
               LineBox[CompressedData["
1:eJwl0mlXTlEAhuHzNpCZEqHoTYVEKmOhpEJSCaGMUTLV//MbVBLN84QQRRkr
11o+XOt+zt774wnXt1Q3h4IgeEGv0UNjRBC8JDcyCMY0R8f1sP7SozqhR3RS
f3PXbtVjOqX7NZNBu41t9nb+2JV6mSp2UM0J8jjOSfK5wl/vT+s7PaXvdZF7
drsW6Ac9qFkM26/Zab/RBxpFmGRqieYcM+5L9Yuu0ga67d2ayjrqWMst0rnN
HS7y3dsyndeN2kSfvVf3sZkW4ijnEtlU8NO7JzqgGbqJZmJJ4Ifzx9qve3Q9
z9nAVhacP9JeTdPVPGMNW/jmvFF7NEXnNEYf8tYeZYX9lJV89X1eZ/WCduku
vUkEN4gknhF31zVEMZ98H9ISPtv12qmJmsSSfV9f6RBX7ULOcIBr1LDsLtBp
LdKPelY7NIswcUQx7z+dYpIJxhljlBGGGWKQAfrpC/3/x/8B0k9a0w==
                "]]}}}], {}}, {
           GraphicsComplexBox[CompressedData["
1:eJzlWXtQ01cWThV1rQ1WrBWBCmqtseNqfYTWV48a13V8YKBIW8Ww0soCoaBW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            "], {{{
               EdgeForm[], 
               Directive[
                Opacity[0]], 
               GraphicsGroupBox[{
                 PolygonBox[CompressedData["
1:eJxNl3kcl1MWh+8t+5JsIUuLRKVE2UoUlV0KETNJWSIVIaKaLGlBKpIh1Ahl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                  "]], 
                 PolygonBox[CompressedData["
1:eJwt1FlIVVEUxvG9xIokRUNUKgohBIkQMUSLKJrLBrW0zDTUaE7NoJHoLciK
JhuMSgrMBpuLbCCLioiGl96aLWieLCzCov6LtR5+rG9z7zn3nLXXvsnlVfmV
ESGEHESiWkK4gH54wnoldRU+kl9iFnkN2smvMZG8A53kv1hOXoEP5BeYQN6E
DvIvbCSfQBLesJ5PvY90PGW9kPoAGXjOeib1EGIRh/UYhS989h3N5NH4Guye
hWhAD0RjBlbjG5+3YQ75LPr6Ow7CZrFnK8dQ8h0MwDnWg6m3kIL9rIdQbyMV
jaynUfehO6IwBsPxKdhvDiPfxUC0sh5BvYc0PGJdQj2KBCRiLEbiR7Ae7yRP
QsA/yi7qbs/rME6sp9rLx8H2Igd/gu3JNvJ2/CZX4Dj5mFjW39yKLWLvrt/Z
QK7FZ/IrNJIPIx4PWd+kTsYzcn/qErE9ivNrtPfTEeN7UI88dEMkDmgfvbfa
s0ViPYzymblKveK91mv2YAq6IAKLfY+jfSZvUK/7s/WkHkGT91afuRXXvPcx
PkvFYt+N9b3N91lL8b2dKja7qT4LBWKzrzORiTqxWa/BUrEzo2flHXL9mfVZ
u/qz1/u76ztkifVQe1cbbK8qxK7VPcsm7/Vr61gvo9aInY33wWZhrljWmTiI
IrG90pk7Qz4tNgtvMZ5cJpZ/ooV80e+d7LOoZ1rPss6knsV5/mx6JvUsLhCb
9QyfZT1TbcFmejZOohf64BIui91bZ0Q/K0Vv/8553Xux2dUzWCl2BnT29UxX
+T31XnqmT1HXivVa76H/JfpO+i76n/IfONGb+Q==
                  "]]}]}, {
               EdgeForm[], 
               Directive[
                Opacity[0]], 
               GraphicsGroupBox[{
                 PolygonBox[CompressedData["
1:eJxNmHmUl2MUx6f7pLRLi9I6aDKlpkVNxbSnopSQNUtkX1IoirKUkF1ECJFs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                  "]], 
                 PolygonBox[CompressedData["
1:eJwtlPmLj1EUh79zzyRr1mQQBmNXxFiHGWOZwYxhFmYQWbLGJEtI1owQIqUk
f4GI7MQYjC1LtoQskbUQUmR7Pp37w9P5nPfe995z7znnpk4vL1wYEolEMSTD
E5xMSyS2YbfDPeiDvxFbAVehG/5q7Bo4ChPw92J7whFoh78U2x6OQ3f8tdh0
uAAt8BdgyzUH3QGWobPB0AGmoKfCZXgO42AF1GWsDsxCz4ZG6CYwFz0QXkIZ
/n7sUKiCxzAGlkAlPIJ80MEvQlvmLw4ecy10TZiBngkf4UNcSzE81D6Mb8Zu
hWvQFX9V8DO+h3eQFffsyNileLZCqIZXUa8MfpbqqHWmE1DKt33YftALfQeb
q3uGz/BJJPm37/AN0qAAhjH/adSlMZZs81wqpr9wCkpgIvRn7DR2PGyCX/BT
OYciOKgx5uxR3nQm9Os4VgY98MfCLnQxHIN65rnRGinma+jf+XAYapvfrWI+
BC3NayEFhqN/xNh3wD+4BSNgEvyBJOZMVuzQEN0AnnEf0/DPQjP8UdiM4Lk8
F/dWTpua14xqZR6ch+b4edjMqAeY34W+9UZ/iXe/AU5CDfO9lAPV6hn0nOA1
q9qujLWlGk+Yn0GxK+Y08zPqbKr5NuhUWKS9Yy1WxVpVTbbGb2V+N5qThx5t
XhvrVCPovub/qkcbm8esWNUTb4L3cEXM+QFINu8t3Zl6+XpcSz1dH/82dn08
011IN78r3emNmPPimBPVvt4EvQXqgc74XWA5uht0Mu9h9a6+ZeB/Df42bAn+
dlxR3QZ/Q4bgZ5n/qzcoF51j/nbshJHKrbnWP4PRg8xzlx9r9X7MvWpWvXQT
mxO8p9S7D+Le6mG9FS+Cv116M94G73H1tnqimPES2B3XKEIXmteevhWgfwfv
BfXAf/WUlYo=
                  "]]}]}, {}, {}, {}, {}, {}}, {{
               Directive[
                RGBColor[0.368417, 0.506779, 0.709798], 
                AbsoluteThickness[1.6]], 
               LineBox[CompressedData["
1:eJwt09dWDwAABvB/ysyIeAB33FtRNITMyiq7FA0kmWVkj4ZVlJEd0jBCsjIi
yXEjXoAXqBfo1zkufud8d9/N941Pz0veFhQIBLoYQH8ex0SiSCKDAoopoZIa
ammgmVbaaKeTn/yim9/84S+9xCmZSxFr+E4sx/nIfR5wl1ru0cYH7rCa26xi
K53EcIbTnCKVW0Qzh0JS+MYsjvGO99xkJTdYwRY6mMk+9rKcr0RxlLdcp4Zr
LCOXL0Syh90spZ0ZHCGbH+Swi2Q+M53DvOYNV0mimiVUkcgVIijmJS28oJVX
7GQxn5jGITbTRRY7KCCTTSyikstcYioHeUYzT8ngOdvJJ52NLOAiC6lgCgdo
Yh2NrCePeOppoIR5nOQxT9hAGgmUMp8yJnGWc5RzgfNMZj91POIha+kf82xO
0GNr/0gkkgmMJZwxjCaMUYxkBMMJZRhDGcJgBjGQEIL/f6UPFjlisw==
                "]]}, {
               Directive[
                RGBColor[0.880722, 0.611041, 0.142051], 
                AbsoluteThickness[1.6]], 
               LineBox[CompressedData["
1:eJwl0+dejgEYwOHnLVtGKKOMkF1mRWRTlJVRfOwAcg6OxAHYKySjKLO9h5lI
IlRCydXPh+v3v+/7+fa87xOXdzo7PxQEQRlnw4Kg3HJGK5gSHgSFGqbhjOKc
fapGMpox3HQ7zxe+UsU690pdrz26Qaczg2lEE0US1Z7P0js6U4u0lhPmCzpb
7+o4Hc8ELtt/0Esbu9xadbf26x6NZwmLeei2VN/qKX2npUw0RzCJ6/Zf/OY9
B9za9aAO6iE9zBqe2Vfrc52sayk3DxOYQwyZs/UoRzjOMbrdX5BrLtA/ZJqz
SOSpfZU+1pX6RBP0hg6Qbs5gL/vZx2f3MnLM1/Qnrzlpf6Of6OIlr3jEMs9W
sJyr9j62m3ewjRb7Tv2onTTRTDEL3Ut0kV7R72wx1+lmrdcP3GaTPZU53LOP
1TS20mBv5D4x9gcaqxd1ns5nLnEs4JL7N1LMG0mmxt7BLf6O/L4j75MIhvyf
u2mjlRaaaaKRBuqpo5YaqqmikorQ/+/iHwj+bBQ=
                "]]}}}], {}}, {
           GraphicsComplexBox[CompressedData["
1:eJzlWX1Yk9cVT6lFxYYO0VY+/Nq0jX3EihVsnXrQtGorWuInOo3OiVOCYhmz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            "], {{{
               EdgeForm[], 
               Directive[
                Opacity[0]], 
               GraphicsGroupBox[{
                 PolygonBox[CompressedData["
1:eJxNl3mYkFUVxu8RF3aQRURUhACFAEcEgUGRfZsRl2GRRUKQkG0ADRNhYhmU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                  "]], 
                 PolygonBox[CompressedData["
1:eJwtlHeIz3EYxz8PRyklXUodcvaWvffmHC77zMMpcifRiTp7dLZOVnH2kXFn
RGcre+8dEkLICCG83j3fP159Xt/f7zuf5/154lPSk9IKhRAGQwyMsRBOQxyc
5Lgi62EoDukct2ZdCq/wN9AJbwC38QJoic+HZ/gLaIxXh6v4QxiNr4cfeA5U
wo9ACcjgeCtrbbiO/4GOeGfzez+AsvghKAYJHFdgXQHv8VRIwrfBX/wfDMJ3
Q2EoAiPgFJTWc/XdrPXhFn4BEvBGcBc/DkPwdnAfvwgTzWuiWui3yfgU83Oz
YTg+yfxdM4PfO978Wj2jHL7QvDY9g9euMlwKXsP+eD24iZ+HDHwWPMbzIBev
CzfwEL1rNbgS/J0b4lXhMn4HhqmPUEp95Hgq61x4ih+ANniy+be/gxS8O3wJ
fo+1eGKUBfVwMb4EXuKJwbPQCl4Hz0RbfDm8Dd6TRXiW+bfqmgXmPVTvnsMW
PFOZw38Hz0KHqDfKRAF0g6Mcl4x63w8OBs+Astk++jZldKZ5hpQd1UxZbGH+
LGWyOT4PzgW/RzO8DjzB82ENPg1O4N+C13a1uavG2XgP+Ip/hpV4V/iEf4AN
+Eb4pSzDJjwHcoP/th/PN//vJ8w277F6q3c4hvdRRvHYqJbqkXqjmq7De5tf
+x1G4l3gY/AMlceXme/N5OB7WXtOey0uysLQaK8pE2kwwfzcs8F7MQMeBe+J
3mU83Av+TspeGfNzlUFlRTVRLZSZpngtOBP8HlWiHqp3WcGzMsA828qM3l01
VO30Der1OPPsquc1zGsSG9VwH+ve6F6quXo13Xy2qGe9zGuk2mjGpJrPBM0C
zZQ55nsoRHsgz3wPa+8WZd0B282zrRmy03zPaK8Ya1/YZe7K0CrWUebXKhNj
zWeOZo32sGaNZoJmgWaOZtFAiDGfSTVhs3n2r8EePN18lumcJtE5+k89+A8O
wsbJ
                  "]]}]}, {
               EdgeForm[], 
               Directive[
                Opacity[0]], 
               GraphicsGroupBox[{
                 PolygonBox[CompressedData["
1:eJxNmHds1XUUxfvuAxktlEIBAVuhslQQFFFRQQTFgQqCijhYKioiU0BRWTJM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                  "]], 
                 PolygonBox[CompressedData["
1:eJwt1OdrVUEQxuFz9yj2FkksiUossYGKvaGIXVRMiGKLaFRUrBAISPAPUFRs
2LH3KLELimLvBbGDXVQs2Csa4+9l5sPDnt2z957dmdlNz52ZNSNEUZSNUthK
5yiaow+uIi2Ooqm00/EVT5GK/tiMD2iGWjiGyswfSzsOS1CO/kiNYxd+YiAG
oybv7tF2xkS8wHcMQF2UIMGcTNp0vMcm/c6/+Qbv0NvHzmgdzJ9EOwWX0Zj+
HNpGvvYv6Od7KMQptMZQ7EVZ5g8PtuZn2Ia+6IEHWIYkdMEF7PPndijCebTF
CGzAK/RECj5hntaQsP/ciJXohRwsQmm+P0zrwCEcQRN0xTU05H2B9uVrqU8/
39e0HA3oz6ZNxho897VHeBQsx8ptN+xGKeZnB/vmR2zxXKoGfmEBilnvIH+u
x/xZPqbnExoLNvYN1RjL9Rxe95pQLbTHHpThfRutUznl+aI/5/neVnisCnxv
Bzx3+b63g7QZwfao3F7yvSrHL7EfozzmymWR50I5LU9/KW11n1OMfxiCkoTF
otBrQTH5iz+ak7A5r1GBOaM9Z3FsNarazMJnVGJsjNdYxdhyplzpN4c95zme
06q8rxLbt3VmdNau0E4LdubOIjW2/1bMbgSrcdV2RxzXXng/nralx34tbYdg
OTiJdT5XNX4O69EpWI3qLO4MlludSf12Plr4f9zFnWC1qLHHXkOqne64jVte
W62wEL+DnW3F7C3tXDQNdkZvok5sd0mGx2qV14Zipr0/DFbrioHuhhTmTAh2
R9TgOTm2b2ksKbY7Q3eFYlCb/v1gd81k7MAPj41qVHfDaX+nO0K1oBrM85rY
jtXB7rY07Su2O0p3U6bHbnGws64Y6iw/8VjoTP8HAg2XOA==
                  "]]}]}, {}, {}, {}, {}, {}}, {{
               Directive[
                RGBColor[0.368417, 0.506779, 0.709798], 
                AbsoluteThickness[1.6]], 
               LineBox[CompressedData["
1:eJwt01dXDgAABuDPKCvJuktkpOzILiIjRGRTIiMyys6MkpXxiQ8hFEU0zWz1
G/pBns5x8ZzzXr1X7xubV5h1pEcgEOiiJ915OAmkkEk2+RRRTAnlVBAkRBXV
1FBHA4000Uwnk5TPoYSLvKKWiczmIOv4yQRO8ZGX1FDNC57TxgeesZYnZPKU
PXSwlwucZw1VJDCLc5zlMY+IZyYZPGQ1JzlBK7sJsYoHrCSPH4wniQJW8J04
jtNMC/dJp5Ll7OIb45jBUn6xjCUcI5e/7OQAabQzlqPk8Icd7GcxXxlDEYvI
5guxTGcbv9nOQoKkco/RFNLEFhrZSj6JbGIzKdxhAXcZxWG6h7SeT0QzlbeU
cYViGpjGOzZQzjzO8J6NfGYE15jPVZK5Tgw3qeAGt7nFSA5RzxvqyGIfk3nN
JS5zmlKmMJc220wmnmEMZQiDiWIQkQwkggH0px996UM4YfSm1/8v/QM4P1fV

                "]]}, {
               Directive[
                RGBColor[0.880722, 0.611041, 0.142051], 
                AbsoluteThickness[1.6]], 
               LineBox[CompressedData["
1:eJwl0ulWT1EYwOHzr5RKihIRDaTcjktwAVxHZR5DhUylMoco0TwPKhGNhsgY
JaF4rD486/e+e6/z5aydsWvPzt2hIAiqaDDUsyMsCHKZ5gPvKeEj9/jMeT5x
gS9UkxUeBPf1IiOMMkYeX7nEBJOcJp9+Bqhjne/Xk8yQfYMGhPhrD9cwNtFo
j9JKjdSr2kqG+YjG6E19w1vOcowXnGCb+y5dpbFMmd+RYy7Qc9zhLlW8pINU
9526Wfv0NlvNx3WYNrLt7drLLTLtR/U5N2gh3dkhPUMT2+3Neo2V5usarYf1
GRvN9ZqiDdpDMY94TDkR7ip0hR7UQZ5SwxZne7WQNeZZneMVa+1Fmqi1+oNS
XpPk7KGm6T79SRnzXGGB/fzmD79YYpEDPCHetwms5oG9m5PE2Wf0G+N85xSX
//9nZrzFCfropYduuuikg3baaKWFZppoDC2/539McXGr
                "]]}}}], {}}}, {
         DisplayFunction -> Identity, PlotRangePadding -> {{
             Scaled[0.02], 
             Scaled[0.02]}, {
             Scaled[0.02], 
             Scaled[0.02]}}, PlotRangeClipping -> True, ImagePadding -> All, 
          DisplayFunction -> Identity, AspectRatio -> 1, 
          Axes -> {False, False}, AxesLabel -> {None, None}, 
          AxesOrigin -> {Automatic, Automatic}, DisplayFunction :> Identity, 
          Epilog -> {
            InsetBox[
             FormBox[
              StyleBox[
              "\"\\!\\(\\*StyleBox[\\\"c\\\",FontSlant->\\\"Italic\\\"]\\)\\!\
\\(\\*StyleBox[\\\" \\\",FontSlant->\\\"Italic\\\"]\\)= 1\"", FontSize -> 13, 
               GrayLevel[0], Background -> GrayLevel[1], StripOnInput -> 
               False], TraditionalForm], {0.63, 0.55}, {-1, 0}], 
            InsetBox[
             FormBox[
              StyleBox[
              "\"\\!\\(\\*StyleBox[\\\"c\\\",FontSlant->\\\"Italic\\\"]\\)\\!\
\\(\\*StyleBox[\\\" \\\",FontSlant->\\\"Italic\\\"]\\)= 0.8\"", FontSize -> 
               13, 
               GrayLevel[0], Background -> GrayLevel[1], StripOnInput -> 
               False], TraditionalForm], {0.63, 0.465}, {-1, 0}], 
            InsetBox[
             FormBox[
              StyleBox[
              "\"\\!\\(\\*StyleBox[\\\"c\\\",FontSlant->\\\"Italic\\\"]\\)\\!\
\\(\\*StyleBox[\\\" \\\",FontSlant->\\\"Italic\\\"]\\)= 0.6\"", FontSize -> 
               13, 
               GrayLevel[0], Background -> GrayLevel[1], StripOnInput -> 
               False], TraditionalForm], {0.63, 0.34}, {-1, 0}], 
            InsetBox[
             FormBox[
              StyleBox[
              "\"\\!\\(\\*StyleBox[\\\"c\\\",FontSlant->\\\"Italic\\\"]\\)\\!\
\\(\\*StyleBox[\\\" \\\",FontSlant->\\\"Italic\\\"]\\)= 0.5\"", FontSize -> 
               13, 
               GrayLevel[0], Background -> GrayLevel[1], StripOnInput -> 
               False], TraditionalForm], {0.63, 0.25}, {-1, 0}], 
            GeometricTransformationBox[
             InsetBox[
              FormBox[
               StyleBox["\"drive invades and fixes\"", FontSize -> 15, 
                GrayLevel[0], StripOnInput -> False], 
               TraditionalForm], {-0.15, 0.47}, {-1, 0}], {{{0, -1}, {1, 0}}, 
              Center}], 
            InsetBox[
             FormBox[
              StyleBox["\"drive doesn't invade\"", FontSize -> 15, 
               GrayLevel[0], StripOnInput -> False], TraditionalForm], {0.55, 
             0.75}, {-1, 0}], 
            InsetBox[
             FormBox[
              StyleBox["\"polymorphism\"", FontSize -> 15, 
               GrayLevel[0], StripOnInput -> False], TraditionalForm], {0.47, 
             0.1}, {-1, 0}]}, Frame -> {{True, True}, {True, True}}, 
          FrameLabel -> {{None, None}, {None, None}}, 
          FrameTicks -> {{Automatic, Automatic}, {Automatic, Automatic}}, 
          GridLinesStyle -> Directive[
            GrayLevel[0.5, 0.4]], 
          Method -> {
           "DefaultGraphicsInteraction" -> {
             "Version" -> 1.2, "TrackMousePosition" -> {True, False}, 
              "Effects" -> {
               "Highlight" -> {"ratio" -> 2}, 
                "HighlightPoint" -> {"ratio" -> 2}, 
                "Droplines" -> {
                 "freeformCursorMode" -> True, 
                  "placement" -> {"x" -> "All", "y" -> "None"}}}}, 
            "ScalingFunctions" -> None, "TransparentPolygonMesh" -> True, 
            "AxesInFront" -> True}, PlotRange -> {{0, 1}, {0, 1}}, 
          PlotRangeClipping -> True, PlotRangePadding -> {
            Scaled[0.02], 
            Scaled[0.02]}, Ticks -> {Automatic, Automatic}}],InterpretationBox[
         Cell[
          BoxData[
           FormBox[
           "\"female selection \
\\!\\(\\*StyleBox[\\\"(\\\",FontSlant->\\\"Italic\\\"]\\)\\!\\(\\*StyleBox[\
SubscriptBox[\\\"s\\\", \\\"f\\\"],FontSlant->\\\"Italic\\\"]\\))\"", 
            TextForm]], "InlineText"], 
         Text[
         "female selection \
\!\(\*StyleBox[\"(\",FontSlant->\"Italic\"]\)\!\(\*StyleBox[SubscriptBox[\"s\"\
, \"f\"],FontSlant->\"Italic\"]\))"]],RotationBox[
         InterpretationBox[
          Cell[
           BoxData[
            FormBox[
            "\"dominance (\\!\\(\\*StyleBox[\\\"h\\\",FontSlant->\\\"Italic\\\
\"]\\))\"", TextForm]], "InlineText"], 
          Text["dominance (\!\(\*StyleBox[\"h\",FontSlant->\"Italic\"]\))"]], 
         BoxRotation -> 1.5707963267948966`],TemplateBox[{
          InterpretationBox[
           Cell[
            BoxData[
             FormBox[
              StyleBox["\"A\"", Bold, FontSize -> 20, StripOnInput -> False], 
              TextForm]], "InlineText"], 
           Text[
            Style["A", Bold, FontSize -> 20]]], 
          InterpretationBox[
           Cell[
            BoxData[
             FormBox[
              StyleBox[
              "\"   \\!\\(\\*StyleBox[SubscriptBox[\\\"\[Lambda]\\\", \
RowBox[{\\\"f\\\", \\\"  \
\\\"}]],FontWeight->\\\"Bold\\\",FontSlant->\\\"Italic\\\"]\\)\\!\\(\\*\
StyleBox[\\\"=\\\",FontWeight->\\\"Bold\\\"]\\)\\!\\(\\*StyleBox[\\\" \
\\\",FontWeight->\\\"Bold\\\"]\\)\\!\\(\\*StyleBox[\\\"2\\\",FontWeight->\\\"\
Bold\\\"]\\)\"", StripOnInput -> False], TextForm]], "InlineText"], 
           Text[
            Style[
            "   \!\(\*StyleBox[SubscriptBox[\"\[Lambda]\", RowBox[{\"f\", \"  \
\"}]],FontWeight->\"Bold\",FontSlant->\"Italic\"]\)\!\(\*StyleBox[\"=\",\
FontWeight->\"Bold\"]\)\!\(\*StyleBox[\" \
\",FontWeight->\"Bold\"]\)\!\(\*StyleBox[\"2\",FontWeight->\"Bold\"]\)"]]]}, 
         "RowDefault"]},
       "Labeled",
       DisplayFunction->(GridBox[{{"", 
           
           ItemBox[#4, Alignment -> {Left, Inherited}, DefaultBaseStyle -> 
            "LabeledLabel"]}, {
           ItemBox[#3, DefaultBaseStyle -> "LabeledLabel"], 
           TagBox[
            ItemBox[
             PaneBox[
              TagBox[#, "SkipImageSizeLevel"], 
              Alignment -> {Center, Baseline}, BaselinePosition -> Baseline, 
              ImageSize -> Large], DefaultBaseStyle -> "Labeled"], 
            "SkipImageSizeLevel"]}, {"", 
           ItemBox[#2, DefaultBaseStyle -> "LabeledLabel"]}}, 
         GridBoxAlignment -> {"Columns" -> {{Center}}, "Rows" -> {{Center}}}, 
         AutoDelete -> False, 
         GridBoxItemSize -> {
          "Columns" -> {{Automatic}}, "Rows" -> {{Automatic}}}, 
         BaselinePosition -> {2, 2}]& ),
       InterpretationFunction->(RowBox[{"Labeled", "[", 
          RowBox[{#, ",", 
            RowBox[{"{", 
              RowBox[{#2, ",", #3, ",", #4}], "}"}], ",", 
            RowBox[{"{", 
              RowBox[{"Bottom", ",", "Left", ",", 
                RowBox[{"{", 
                  RowBox[{"Top", ",", "Left"}], "}"}]}], "}"}], ",", 
            RowBox[{"{", 
              RowBox[{"ImageSize", "\[Rule]", "Large"}], "}"}]}], "]"}]& )]},
     {
      TemplateBox[{GraphicsBox[{{
           GraphicsComplexBox[CompressedData["
1:eJzlmnl0zPf6x0eo2kZ/1LVMItq6YmjsEku1T2KI9l5iJJYrdGqp/DD2uhhL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            "], {{{
               EdgeForm[], 
               Directive[
                Opacity[0]], 
               GraphicsGroupBox[{
                 PolygonBox[CompressedData["
1:eJxNmHn8V1Max+8pSkkpVCIpJJKUSosoSaVFq2gPIaTs+26KjApZspMlJmMJ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                  "]], 
                 PolygonBox[CompressedData["
1:eJwt08tLVVEUx/Gzo5l/gWTi1dQyLUNJDC9KoihBoIJIIKFQaJGm5ruHZqHV
7eGjgjTn4SMte6ulKXYdKImYNGnc3+DE72L/Bh8W556z9157rXVDtY3lDQeC
IMjFQdS5IPiNCvzj+SbxFTIRwgCWUYJSvEQEh3EC7XiIVOSgD39QjSPowhzy
EMa01ti3pzGKMRxHIt5iRu+myWeC2IZsZGEQf1GDBHTiCY7ijHLbxUXleANv
kIxDuIRNlGGe/R8Tm3BSe0xiSmeN836EeB/FyMA17OACYrCN28q9ClvO19Bq
V6lve5CkNVabu4hVjSzXR8rFcv6FNUR5Pme/45vOOosh3EIB0lCvM+2s/6xZ
If5U7axnH/FBvc1T7Z7pW6vhEn7orkXEYdxBoXryBZ+xx/t84j0sqhZWU5ud
bvXCZiiKdedzj8cGWlX786rFa5xSTewun5zfu0C9uIxj6onl8lV3t5xe4Dni
kIJanWF7RwI/K/16ZzOzgF7nZzFda1edn2XbY9b5GbLZydVsdCh3m5Gnzs9w
WDk9wHXnZ99m/h3ea631rJl4Rb2x/1ALGnS23fmq8zNTpZrabH13vvY2Y/uX
hmnu
                  "]]}]}, {
               EdgeForm[], 
               Directive[
                Opacity[0]], 
               GraphicsGroupBox[{
                 PolygonBox[CompressedData["
1:eJxNmHmYz1UUxmfOrVCRJWPfyVJkbI09MlqUGUI0USSppFWhon3TnlZt2je0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                  "]], 
                 PolygonBox[CompressedData["
1:eJwtlOmLj1EUx5/fPUPW7Ip4MYksZc0WYwkx1pkh+zIzDGOY+Y2xjOxLIska
IgwisiQ1JSX5G8goa8g2sq9J+Hyd++LT+Z77PM957j3LzSxM55WFJEnyIAPq
W5K8ZGEtrIA8/ClQhS6FReglcAm9Gh7CQPw92OnQGt0IimOMpuhX2HWwCtri
/8bmw3p4BBWs3cBOE+g25s9OQRZ6KOxFF0BLdAGcRy+ExujlcBNdAp3Rv+Je
N2tP+O1gA/o0tEf/xRYGX/sBzcz3dhxamf9DsSvgM9TTP7CH4Sl0MY89L/hZ
SvGvxDP1RPeA7egdsBTdCTahL8N3aIK/ErsxeG6eQGXMUYn5O3qmHGuvZaxd
jXuui14G11UL+Ak5+CewW+ELNDCv3VF4Af3wd2GzYTB6EDxIJck2/GfQ3VzX
pnzvb9D7gp9B706Fk8G/GY6+ozzBfijH/4AtgmuQxk9gProahpjXVLVUDZ9D
H/yd2A7QG10bY2ntXoypWFnQF/0OeyD4Gcaan1FnOwLj8Meb61yYiJ4Ex9Cz
ob95DVW73TGXd+O/ldP3MEo5jWeYYN4T6gXF+AgZ5r1wCEabr0kfhDH42RY1
jETX6Nxxz5PNe0C1r4q9cjvmWj2jWt6PtVFNc/G7wZbgNa2JOa2OMZX7Xua9
pRoE9Cw4gz4LI9AzzHu9HOqg58K54DOj2ayEW8FndBj6dfDZSgev/Z+o1QMd
ef4t+GypZzPxP8Xe00wMMM+xcqs7QHvvaj4bOkORZgkuoBcH/3dD81nXHjS7
j4PfHZph3QVf47u6EzQ7xXAx+AwtUG+Z1/Z/TPM7SXeR3kmhZ8Ka4DXN117M
e0M5aG4+M5oVzVwL8xwqd/pmjmKb11I5fRtzqlyqp/8BHFSjSA==
                  "]]}]}, {}, {}, {}, {}, {}}, {{
               Directive[
                RGBColor[0.368417, 0.506779, 0.709798], 
                AbsoluteThickness[1.6]], 
               LineBox[CompressedData["
1:eJwt0kk3FQAYx+GLikLoXtLSju9Q3wEfwdCcW0gUGYqSNM8TRZSSFBVlSlSX
SGhaWbdta+G551g857yr9/wXv6yCcH5JTCAQ6CWW6J1ODrvIo5gqWmmjjxEi
LLPCP/6zygZPkggSIp0MtpPJDhr4xSme0E09I0QHlXGfOn5yiMd84z0fGKaW
qyxzkHvc5SRLHKCTWd4xxFtquMIi+ynlDtX8YB8dXOIvl7nNCRbYyyNmGOQN
AxznIt/ZzRFa+cMFqujnNa/Yw0M+84VpInzlFpXMU0w7LfzmPDc5TDPnOMYL
XtJHEW1M8omPnGWK69ygiTNU8IxenlPIA8aZYIxGrhFmlNM8pZwuejjKHAXR
jVqopIhcdpJNiCDbSCOVFLaSTBKJbGEzCcSziY3Rvohb73UNscFaxQ==
                "]]}, {
               Directive[
                RGBColor[0.880722, 0.611041, 0.142051], 
                AbsoluteThickness[1.6]], 
               LineBox[CompressedData["
1:eJwl0mdbjmEAx+H7qZQVRZkZFTLTEBkpNIWGTYooIpUR+ja+Ct/ASrR3IYSI
rNPhxXn8/td1PS+eF3dibXNlUygIgkc8DguCToeHOsOp8CDYyz62kcqA+ycs
s5/qTf2pK3Q5p9lPLjmMecvTX9qqv3Wc7XYa6TxzPqBvNEMzGbSfs9J+oU36
R1dpAmc4yyHeuj+o7zTQZtrtfJ3QLN3JsN3BGvu13tJIJu1EjeIes/nkbpdm
s5su52Tt0TsazZS9XhfwgIV8dXddb3CVBq6xgUaOc4JKvvvtHj3JD3ujbqLf
TtGtbGYLFZQz7W2R3qXXHiXGbiOWb85HtYxj9Dmv08tcoZZ66ljKF++3tVuT
dB73mc8SRtzP0bl8tg/rEUrpdF6rNVykmllcIo54wongo98WaQnFvHJerVVc
4DxhLGbI2zkNUcB75x1ayAe7RV9qDmn//jdxRDLt+x1njFFGGGaIQQbop49e
euimK/T/+/8Ln2Nc0Q==
                "]]}}}], {}}, {
           GraphicsComplexBox[CompressedData["
1:eJzlWXlUk1cWj5W6VOOGVg3ua6xYWzQuFb3YOHbR2mBRWyqxUrFiQAq1tGHc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            "], {{{
               EdgeForm[], 
               Directive[
                Opacity[0]], 
               GraphicsGroupBox[{
                 PolygonBox[CompressedData["
1:eJxNmHncl2MWxp87u8iuUEivVirtUiRjGUQIIYpsSSUtIu0iyhIpDSFLSais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                  "]], 
                 PolygonBox[CompressedData["
1:eJwt0/dL1lEUx/F71PayjCapj0IQBEEQIoJto4iGmVY0xTIqkvYOWj9YUdEE
kzYVDWhPgkpo/d5/EEU0KYoKo96H8/nhxbn36/M83ns+55upbahcnpVSqkAO
xlpKTzAYTf6c+hRDcI79DeoS/GHdnroVW/CR/ScsYn0VA5CP86jFN/7WnToJ
w9HKvgN1CkYjC52wEzuwJsVnRrE+jT56VkNdgYTeGIFHGIg6Ho+hPsYg/x32
M6mzLc6WS52MkfjHviP1Nm7hLfu/uMf6Lt6n+B/j0IKhfi8eXaIuxA/WPagb
0WixbuHZMuozDFOP6qiV6KLvLKBORWf1ZDF1GrriF/tdFnfws3ejXsQBfGff
T9nMwecUGfldZ+FDijt7VnPxJUVm3pvpeJeiRystepKrnm/DdmTjK5+5qYw8
mww2YL5Fdv6Z3dijs77GIdaH0RY/2R+lHrOYjd/sT1lk5ln1QhWaLXrtZ/K7
T0Q79eCOMvJsinBfM+KzUYxX6rn3upz6EuuUTZmymqe7eGYvfG4sZreUuh9n
0Vc98yzGI0eZeFYT0EaZ7cNJ9NRMrKY+R4nFzO+lHkeeZmi9zlSmDDbpzOXK
cK3OVKoM/e7XUKgezLDomffKZ77JIpOM/sd1HFFv/Ttn1KNi3cnPtsri3fUz
+iwuxZsUM7nZ4h3NV2Z+18vorzv7u1uPAot3+CEeWGRdoF5dsJhF79kVHFT2
/hsnlGGRetaAaotZ9nfuP9rfgdI=
                  "]]}]}, {
               EdgeForm[], 
               Directive[
                Opacity[0]], 
               GraphicsGroupBox[{
                 PolygonBox[CompressedData["
1:eJxNmHe4z3UUx+/vfCoUl6yLyMgohbqurEQpSiIZUUkRbaU0tGlvRYqSor1F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                  "]], 
                 PolygonBox[CompressedData["
1:eJwtlPlrz3Ecx797v5HcRzS3sSVXbrYcM6YRm9kaiwlznxOJ3HLfpdRKKS35
CzT3lTP3ETJEiTRXrnJ7PHu9fnj0er4+78/xfj9fr9cnpbSsYFFIJBKFUAOa
x0SiG6znYjaMQGdDDnoSdESnwRp0FtyG1762Cfqw1g+2oEfBD/gJ06EAhrD2
hzgbdkAt8pqQgZ4Gp6A3+WZiKlyAYvJDehbGovMgH10Go9Gf/Nn5MJT8EnEG
7IS25G1gJXoVvIJbMAcmQh3WGsFidCZ8haewwPc8nLXLxJmwC+qS14OFfk/9
aO/Qs7rWAp0My9ErYAy6drS97YWAPkPs7nsYRp4VzetiOC6fyPcTk2EgOjfa
s6XQBN0YlqCXymP0FeIs2AMjySPMRe+Gt3AP5sEUaMhag2je6cxHoWm0d1Ul
mVcP0NuCefYRSrh2mJjne7upvgi2xyLy+6oDlEPXaDWQ9+qh1Gg1V63VM/no
lzo37PPaVnlvqMbjyR/5tw7ABPInxHFwUH1G/o6YDhVwEZpxbRmxmv1fI3Ym
L/JvqBbXieuC1aRXtJ5Vr26Ex+6xvFVPtUc/I26H1XBDtYpWG51Z3l/1NdXg
O3xTnSEXurBeqT7wb3aKNjOalbXBai/Pyr0H5O0H4qBgHveI1nPqtQ0wIFqN
Vdut0DdajVVbzdgdP5POopk55x4ecI9+wy/3ppB43j096N9Urathqtf8C0xm
/Yh73h99N9gs655W5D2j7U3/hNbo9Gh7Uw8MRj8P9i3N+HuvWYV7pF497b2o
nj0L7aLNZob36kP3Uj3bgbWUaL2pGr+AzGheqsYnoGW0WUsLNqsng822Zlb/
gmM6c7B/Qg73vpHvwWak0ntyn9fsH/wN9u/RM0msJaK9u8S9+RzsXySP/gOZ
lJeT
                  "]]}]}, {}, {}, {}, {}, {}}, {{
               Directive[
                RGBColor[0.368417, 0.506779, 0.709798], 
                AbsoluteThickness[1.6]], 
               LineBox[CompressedData["
1:eJwt0ldXDwAcx+G/vUXlBXSn96BjFS0tO6shpLSolOwoGS0RySyzISOiIgp1
5aobN6EX4vmf4+I553f3u/h+IrIK0wpmBAKBQWYSvFcQSRSp5FBGNQ200UkP
/QwxygQ/meQXU/zmD3+ZJs+Db+Rzi1YO84pN3OQGLfTxkkQ2c4hecrlMHR/4
SA8H6eYAVYyRwHWaaaKLF8STxn6ekcNzsrlEJROcoJEGnvKEOFLJpJMsHpNB
BaPEco6z7OMrG6mlnHGOc4a9fGEDNZTygzJOs4cRYrjIUb5zjJOcYhe72UkJ
n1hPCgM85BH1dJDOZ6K5wDYKGWYtyfRzl3tc4x3vKaKY7dxnBw9Yx3mqOcIg
q0niNbdp4wpveEswvjts4SrtbGWINcF9NJPCKlYSThihLGcZISxlCYtZxEIW
MJ95zGUOs5n1v/V/oq9b8w==
                "]]}, {
               Directive[
                RGBColor[0.880722, 0.611041, 0.142051], 
                AbsoluteThickness[1.6]], 
               LineBox[CompressedData["
1:eJwl0mdbjmEYwPH7qZCECBnRsLKFhq3MhiJ79KYPwOfwUXyACqFkaz497WxZ
JSIz8/ccXvyO/3le1/3qPq6MqvMV50JBEDRwMSYImiwX9BHpsUFwT2foTJKp
tmdoJrOYzS1nNfzkF0/Y5vyxbtffukOXsJTFLGcZO3nqfoXe1yx9oM84ZK7V
ldqkc3Ue87liDzTEG/M+fa37NZYDbCKHjYTd5eqQntZh7WSheRFp1Nsn6EQ+
mot1VEs0nlIOspVe51u0T8u0nGlMJ4kEppLIFB767jOnzDd1nB8McNJZo07W
T1qpY/qdfk7Yb+gkfa9n9YN+o4fj9gaN03d6Rkf0C1/popsO3nLU/TW9S5E5
j3xS2UzE+TG9rjHaqoMcMdfpHXab17GeBexhLy/dvaKNbHu7btCr+oe/POew
s0t6m13mVaxmDgUU8sLdGm3WFF1Li7lCL0ffQ/SfRf+lNztEP3300kM3XXQS
oYMw7bTRSgvNof9v/x+FAGY5
                "]]}}}], {}}, {
           GraphicsComplexBox[CompressedData["
1:eJztWXlU01faDtZ1WhgRWiVUbG0dg9W61KC1ra82uI2WglJbUeIy1VqIS7FV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            "], {{{
               EdgeForm[], 
               Directive[
                Opacity[0]], 
               GraphicsGroupBox[{
                 PolygonBox[CompressedData["
1:eJxNl3eUltURxu+1Y4sVFBusAiKiuKhsBFbpICvIIlIEFhcVpDelqoiySIkV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                  "]], 
                 PolygonBox[CompressedData["
1:eJwtlHdsT2EUhr/T1I69t5qxI/YeqRFSQQhqVW2p1h6pLTbVaqlG7L33DEIE
IRojCDFix0hEJEiI8bw5/ePJ77nN7b3fPec9Jyo+qU9iRAghDiKhj4UwAs7j
K2AOXhvW4buhL94PTuIZ0AsfCmfwJfAHtnL9RfdCPXw+PMafQG88Ds7iS6EO
Phce4A+hOh4NR/A0qIZPg9t4NvyCSVzf4DcaKuItYBt+C8rjTWErfh3q453h
GL4eyuKj4AJ+CargVSEVPwHN8Ez4gH+E1vhe+I6nQA18JtzB70IZvDFswS9C
QzwFXuKv9U7VDlYGv+cfbOf6K78joQI+Ea4F/6aBeCycwjOhPz4A9gTvQRRe
GA4HP/NqPC/sx19AE7wgHAxeg+Z4ITgUvEZF8AA7g9d4LV4ADuDvgj/7L787
gr+jBNe5YV/wMxXFDXbhRyEL3wAxwWvW0bwG+nadaQKeoB7hV6EDvsA8C5+h
OJ5LNQ7eo2J4hHnWlLFO5j1RL/RNrfAe5r2KhW54V/MsbIIxeDm4gl8O3mv1
SL1Rz1vileC4+gDdzTOrrOpvMVz3hHQ8AcbiP1XH4M9UrX7kvEs1G8f1ePN3
6RuHm2dc2T4HJfE8kIWfDv6s0ebZ0zPVm985Z1GPapnXWLXVGUrh+c2ftQxK
4yNhc/BMRaoX8A2fC43wDHgf/J5U8xnVbL6B9uYzp1lTpqbgU+Fm8B7OwKeb
Z1MzN9l8xjRbukdZ1YxptpTZVfggZQB/Dm3MZzY7J0MrzDOtLD+DIeaZVVa1
M/S/g81ro2csU+3Mv+UpLMIXQnLwHaJdURceBd8Zi/Vt5vfqHmV1pflZlNlh
eBq8Db5zlN1kuB88w5rlmnAv+EzPM8+osql3bDTPhLLwCWbh7cxd/5NuvjO1
K1XzePOZ0ixph6o32kHaPerRbPMdqt2pM2i3KNPKsnaMdmdl812jHdrWfOdo
12hmtVsawKvgOyYRTzKfrZawBu9i7rpH357PvPeqgXaPdqx2q3aQZm+5eW80
g/8BNXLIZQ==
                  "]]}]}, {
               EdgeForm[], 
               Directive[
                Opacity[0]], 
               GraphicsGroupBox[{
                 PolygonBox[CompressedData["
1:eJxNmHuQz2UUxnfPFyFkc61UKJXQPaJEIRSJUKII5dbQdJ+m+1Qz1TQ1Tfea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                  "]], 
                 PolygonBox[CompressedData["
1:eJwtlGdsj1EUh9//fYsWpbSl1CpCiKAIQsQKUUKpGJHQEF+tD4i9Yu9NzMSI
2CNWisSIEWqH2LNWzaLV1np+uefDk9977z3vvefeM1KGjcoY6YIgGAJRcIvB
TWgADWE/JIVBMEhGMBZ+Qkd4HQmCFegPaG9z1yEb6kF90Oa3ba9s7MuzVzn4
q2/mJkBVxoPRL8zlo7+gE0RgOeyFOOgAK+EgVISusBW2QQykwCwYDbFQBS7B
RSiIeJszcNp8l80WyILqEA0j4Ib5noXNFfQy1IY6sA42QimoBVP0L/4PRfOx
34SGjLuhaTAGPkBreMD6UvQ9tLS5mXAPGsFj1uei4yEekuEOPIdUKGT9MHrB
zs5jvAGdB9XMp3+QyPkD0QSYpDXGmYpbxNtOhEr2TzxrZ9EaMADOW0wUi5ro
IVjl/Ld8+mh30l3aoidhH/RXHCEXljm/1sbuugRa2Z3PWU4FduZ8mAqVLQYH
nH+DZIvxHphjeysHSuLvNDQJesBVu6PuVtditdn5b8UsR7G2t22MHrecSbEz
n8IT53NTMZgM1yzWypkCeyO9jXJyNjx0fq+ciK+VhdDU+ZopE/ocVO5l6FzL
KeWSbBbBDGgOxcwtRt9BOedr4hV6F4oi3ka1tACaOF9TOxVnzuiJpuuN+K4A
udj3Y/zGfNDZLeARbIfutsdR53M0zXIgGttTsuP/Xugze2O9baqdtcP5NZ25
HhxziWgJwXdU6HNfOZ8H36Gd7am7vIRmdqcA2zV8h87vcUSxYG44GmW5FBf6
uyin7sNbu8sf/j9hb5BuObDa+ZxPsD2Lne8R6g1d4Bt8Vcwi3if1snEWO/U0
1cYxu0um+aIaU23JJ9W2eoR6g2q8bOh9ki99ne9da8139bBdEBv6NcX0BZRm
3MfeQN/T0d7Oz32Cz87Xinz8rb3Nd8VUvVU9UL1PPTaB/wvRzs7XeJHzPTNi
czGh7xnqFTpDtbfb+V6rGvwPPj+w/g==
                  "]]}]}, {}, {}, {}, {}, {}}, {{
               Directive[
                RGBColor[0.368417, 0.506779, 0.709798], 
                AbsoluteThickness[1.6]], 
               LineBox[CompressedData["
1:eJwt0/dXT2EcB/Dvt6VJU3ZCFMleGWVkJipZkUQIKSIikops2QnZ/Ff+i+zx
uocfXud9z/s8z+ec+9z7ZNc1VZwIh0Khz0QQPGeQRyEllFFFNbXU00AjzbTQ
ShvtdNBJF930M8zgDPZQzS1uEkM6u9nFDa4TTRolzOQ9UWzhBdfopZur9PCc
K3RxmXw6KaaDGVxiOkW8I5Kd7GA7eVwkglSq2EY7FwiTwjTOk0sZmxlgBW1M
5Rw5LOdt+N/hJrOGKbzhj7pUPqOVs5xhMst4zW9rRshsPjCJiWyikE8sZTVZ
vOKXPRvlBJYwyA/dcLmIjyxmPKc4TQs/rdkgn9LPExayiiTms4CxNDOOk3y3
b51cz2xeMqRLkPep4wAVPCCRh8zhIJlU8ojHzGUeo6hnNIf4YuYROYbDHKWB
r/rjspFjNBH82N/0a+VdCrjDLFYSy232UsNW7hFHH/uopZz9xDOSXjNzSSeN
VFJIDr5TcMYkkRi8P/HEERvcA2KIJorI//fuL2TsTkE=
                "]]}, {
               Directive[
                RGBColor[0.880722, 0.611041, 0.142051], 
                AbsoluteThickness[1.6]], 
               LineBox[CompressedData["
1:eJwl0+dXzQEYwPHfLYTsvUL+HC/9Cc7xUpEZMrJHKDsJZbX3RPZI90Yke1Nm
tlDGx/Hic77P89z7uy/uuTd+euK0hFAQBBFmRAVBnmWqHuIeG4jQzFVmcp+N
XKOF6yTwgIfcpI1UEnnOC+5wl9XM5RTzGRMdBMV6ljMMt3fpd77xkx+McB/J
H/MOfrOTwK1So3SPhjRTq5lgXqxNZHOZRq4QZgkXuUQOU7z/oOZygBomui3V
9RxnP3FuKXqBUeZe9CaafW5VjDcv0nXsZZI9S3dznnLi3Sq0nl2MsyfrObZT
xli3BbqWEibbS3Ub3fyih4WcZqjXhzGEInsdayigkHS+kMFX5nGSBo4y2nOz
dCWx5nbt4BYD7Kt0oB7Tl6Rxm0Fug8kz57OZV2zhNbN5y1be0Mk7kvjAR97z
mU/M4QR9feZhjdEjWssy+tgf6WNu0M++XPvrE31KK89YwaZ/35ffd5h66qil
hmqqqKSCcsoopYRiiiikgPzQ///LX6XnflU=
                "]]}}}], {}}, {{}, {}}}, {
         DisplayFunction -> Identity, PlotRangePadding -> {{
             Scaled[0.02], 
             Scaled[0.02]}, {
             Scaled[0.02], 
             Scaled[0.02]}}, PlotRangeClipping -> True, ImagePadding -> All, 
          DisplayFunction -> Identity, AspectRatio -> 1, 
          Axes -> {False, False}, AxesLabel -> {None, None}, 
          AxesOrigin -> {Automatic, Automatic}, DisplayFunction :> Identity, 
          Epilog -> {
            InsetBox[
             FormBox[
              StyleBox[
              "\"\\!\\(\\*StyleBox[\\\"c\\\",FontSlant->\\\"Italic\\\"]\\)\\!\
\\(\\*StyleBox[\\\" \\\",FontSlant->\\\"Italic\\\"]\\)= 1\"", FontSize -> 13, 
               GrayLevel[0], Background -> GrayLevel[1], StripOnInput -> 
               False], TraditionalForm], {0.63, 0.55}, {-1, 0}], 
            InsetBox[
             FormBox[
              StyleBox[
              "\"\\!\\(\\*StyleBox[\\\"c\\\",FontSlant->\\\"Italic\\\"]\\)\\!\
\\(\\*StyleBox[\\\" \\\",FontSlant->\\\"Italic\\\"]\\)= 0.8\"", FontSize -> 
               13, 
               GrayLevel[0], Background -> GrayLevel[1], StripOnInput -> 
               False], TraditionalForm], {0.63, 0.385}, {-1, 0}], 
            InsetBox[
             FormBox[
              StyleBox[
              "\"\\!\\(\\*StyleBox[\\\"c\\\",FontSlant->\\\"Italic\\\"]\\)\\!\
\\(\\*StyleBox[\\\" \\\",FontSlant->\\\"Italic\\\"]\\)= 0.6\"", FontSize -> 
               13, 
               GrayLevel[0], Background -> GrayLevel[1], StripOnInput -> 
               False], TraditionalForm], {0.63, 0.165}, {-1, 0}], 
            InsetBox[
             FormBox[
              StyleBox[
              "\"\\!\\(\\*StyleBox[\\\"c\\\",FontSlant->\\\"Italic\\\"]\\)\\!\
\\(\\*StyleBox[\\\" \\\",FontSlant->\\\"Italic\\\"]\\)= 0.5\"", FontSize -> 
               13, 
               GrayLevel[0], Background -> GrayLevel[1], StripOnInput -> 
               False], TraditionalForm], {0.83, 0.035}, {-1, 0}], 
            GeometricTransformationBox[
             InsetBox[
              FormBox[
               StyleBox["\"drive invades and fixes\"", FontSize -> 15, 
                GrayLevel[0], StripOnInput -> False], 
               TraditionalForm], {-0.18, 0.5}, {-1, 0}], {{{0, -1}, {1, 0}}, 
              Center}], 
            InsetBox[
             FormBox[
              StyleBox["\"drive doesn't invade\"", FontSize -> 15, 
               GrayLevel[0], StripOnInput -> False], TraditionalForm], {0.55, 
             0.75}, {-1, 0}], 
            InsetBox[
             FormBox[
              StyleBox["\"polymorphism\"", FontSize -> 15, 
               GrayLevel[0], StripOnInput -> False], TraditionalForm], {0.47, 
             0.07}, {-1, 0}]}, Frame -> {{True, True}, {True, True}}, 
          FrameLabel -> {{None, None}, {None, None}}, 
          FrameTicks -> {{Automatic, Automatic}, {Automatic, Automatic}}, 
          GridLinesStyle -> Directive[
            GrayLevel[0.5, 0.4]], 
          Method -> {
           "DefaultGraphicsInteraction" -> {
             "Version" -> 1.2, "TrackMousePosition" -> {True, False}, 
              "Effects" -> {
               "Highlight" -> {"ratio" -> 2}, 
                "HighlightPoint" -> {"ratio" -> 2}, 
                "Droplines" -> {
                 "freeformCursorMode" -> True, 
                  "placement" -> {"x" -> "All", "y" -> "None"}}}}, 
            "ScalingFunctions" -> None, "TransparentPolygonMesh" -> True, 
            "AxesInFront" -> True}, PlotRange -> {{0, 1}, {0, 1}}, 
          PlotRangeClipping -> True, PlotRangePadding -> {
            Scaled[0.02], 
            Scaled[0.02]}, Ticks -> {Automatic, Automatic}}],InterpretationBox[
         Cell[
          BoxData[
           FormBox[
           "\"female selection \
\\!\\(\\*StyleBox[\\\"(\\\",FontSlant->\\\"Italic\\\"]\\)\\!\\(\\*\
SubscriptBox[StyleBox[\\\"s\\\",FontSlant->\\\"Italic\\\"], \\\"f\\\"]\\))\"",
             TextForm]], "InlineText"], 
         Text[
         "female selection \
\!\(\*StyleBox[\"(\",FontSlant->\"Italic\"]\)\!\(\*SubscriptBox[StyleBox[\"s\"\
,FontSlant->\"Italic\"], \"f\"]\))"]],RotationBox[
         InterpretationBox[
          Cell[
           BoxData[
            FormBox[
            "\"dominance (\\!\\(\\*StyleBox[\\\"h\\\",FontSlant->\\\"Italic\\\
\"]\\))\"", TextForm]], "InlineText"], 
          Text["dominance (\!\(\*StyleBox[\"h\",FontSlant->\"Italic\"]\))"]], 
         BoxRotation -> 1.5707963267948966`],TemplateBox[{
          InterpretationBox[
           Cell[
            BoxData[
             FormBox[
              StyleBox["\"B\"", Bold, FontSize -> 20, StripOnInput -> False], 
              TextForm]], "InlineText"], 
           Text[
            Style["B", Bold, FontSize -> 20]]], 
          InterpretationBox[
           Cell[
            BoxData[
             FormBox[
              StyleBox[
              "\"   \\!\\(\\*StyleBox[SubscriptBox[\\\"\[Lambda]\\\", \
RowBox[{\\\"f\\\", \\\"  \
\\\"}]],FontWeight->\\\"Bold\\\",FontSlant->\\\"Italic\\\"]\\)\\!\\(\\*\
StyleBox[\\\"\[Rule]\\\",FontWeight->\\\"Bold\\\",FontSlant->\\\"Italic\\\"]\\\
)\\!\\(\\*StyleBox[\\\" \\\",FontWeight->\\\"Bold\\\",FontSlant->\\\"Italic\\\
\"]\\)\\!\\(\\*StyleBox[\\\"\[Infinity]\\\",FontWeight->\\\"Bold\\\",\
FontSlant->\\\"Italic\\\"]\\)\"", StripOnInput -> False], TextForm]], 
            "InlineText"], 
           Text[
            Style[
            "   \!\(\*StyleBox[SubscriptBox[\"\[Lambda]\", RowBox[{\"f\", \"  \
\"}]],FontWeight->\"Bold\",FontSlant->\"Italic\"]\)\!\(\*StyleBox[\"\[Rule]\",\
FontWeight->\"Bold\",FontSlant->\"Italic\"]\)\!\(\*StyleBox[\" \
\",FontWeight->\"Bold\",FontSlant->\"Italic\"]\)\!\(\*StyleBox[\"\[Infinity]\"\
,FontWeight->\"Bold\",FontSlant->\"Italic\"]\)"]]]}, "RowDefault"]},
       "Labeled",
       DisplayFunction->(GridBox[{{"", 
           
           ItemBox[#4, Alignment -> {Left, Inherited}, DefaultBaseStyle -> 
            "LabeledLabel"]}, {
           ItemBox[#3, DefaultBaseStyle -> "LabeledLabel"], 
           TagBox[
            ItemBox[
             PaneBox[
              TagBox[#, "SkipImageSizeLevel"], 
              Alignment -> {Center, Baseline}, BaselinePosition -> Baseline, 
              ImageSize -> Large], DefaultBaseStyle -> "Labeled"], 
            "SkipImageSizeLevel"]}, {"", 
           ItemBox[#2, DefaultBaseStyle -> "LabeledLabel"]}}, 
         GridBoxAlignment -> {"Columns" -> {{Center}}, "Rows" -> {{Center}}}, 
         AutoDelete -> False, 
         GridBoxItemSize -> {
          "Columns" -> {{Automatic}}, "Rows" -> {{Automatic}}}, 
         BaselinePosition -> {2, 2}]& ),
       InterpretationFunction->(RowBox[{"Labeled", "[", 
          RowBox[{#, ",", 
            RowBox[{"{", 
              RowBox[{#2, ",", #3, ",", #4}], "}"}], ",", 
            RowBox[{"{", 
              RowBox[{"Bottom", ",", "Left", ",", 
                RowBox[{"{", 
                  RowBox[{"Top", ",", "Left"}], "}"}]}], "}"}], ",", 
            RowBox[{"{", 
              RowBox[{"ImageSize", "\[Rule]", "Large"}], "}"}]}], "]"}]& )]}
    },
    AutoDelete->False,
    GridBoxItemSize->{"Columns" -> {{Automatic}}, "Rows" -> {{Automatic}}}],
   "Grid"],
  StripOnInput->False,
  ImageSizeMultipliers->{1, 1}]], "Output",
 CellChangeTimes->{{3.8096815222475367`*^9, 3.809681537477656*^9}, 
   3.809681602926483*^9, {3.8096816858881283`*^9, 3.809681732769496*^9}, 
   3.8096830842626753`*^9, 3.814005335346916*^9, 3.814007215264979*^9, 
   3.814086050851211*^9, 3.814196573154961*^9, 3.814197129416876*^9},
 CellLabel->
  "Out[1525]=",ExpressionUUID->"8c0c23ee-7111-4c7b-ad71-11ce17a9c81d"]
}, Open  ]],

Cell[BoxData[""], "Input",
 CellChangeTimes->{{3.7952692709305477`*^9, 3.795269271739058*^9}},
 CellLabel->
  "In[1526]:=",ExpressionUUID->"ba3545b6-dc5e-48ca-b776-6b5eb92fe946"]
}, Closed]],

Cell[CellGroupData[{

Cell["With simulations", "Subsubsection",
 CellChangeTimes->{{3.783864001958191*^9, 3.783864010458066*^9}, {
  3.800015761994898*^9, 
  3.800015766313696*^9}},ExpressionUUID->"7fc3a320-42d1-49cb-83ff-\
688f08f0cd53"],

Cell[BoxData[{
 RowBox[{
  RowBox[{"data", " ", "=", " ", 
   RowBox[{
   "Import", "[", "\"\<data/per_hmap_b2.4_sm0_fm1_mm20_delta1_c1.csv\>\"", 
    "]"}]}], ";"}], "\[IndentingNewLine]", 
 RowBox[{
  RowBox[{
   RowBox[{"b3sm0fm1mm20delta1c1HMap", " ", "=", " ", 
    RowBox[{"MatrixPlot", "[", 
     RowBox[{"data", ",", 
      RowBox[{"DataReversed", "\[Rule]", "True"}], ",", 
      RowBox[{"Frame", "\[Rule]", "True"}], ",", " ", 
      RowBox[{"FrameTicks", "\[Rule]", 
       RowBox[{"{", 
        RowBox[{
         RowBox[{"{", 
          RowBox[{"ticks", ",", "None"}], "}"}], ",", 
         RowBox[{"{", 
          RowBox[{"ticks", ",", "None"}], "}"}]}], "}"}]}], ",", " ", 
      RowBox[{"FrameTicksStyle", "\[Rule]", 
       RowBox[{"Directive", "[", 
        RowBox[{"White", ",", 
         RowBox[{"FontSize", "\[Rule]", "12"}]}], "]"}]}], ",", 
      RowBox[{"FrameStyle", "\[Rule]", 
       RowBox[{"Directive", "[", "White", "]"}]}], ",", 
      RowBox[{"ColorFunctionScaling", "\[Rule]", "False"}], ",", 
      RowBox[{"ColorFunction", "\[Rule]", 
       RowBox[{"(", 
        RowBox[{
         RowBox[{"Lighter", "[", 
          RowBox[{"Orange", ",", 
           RowBox[{"1", "-", "#"}]}], "]"}], "&"}], ")"}]}]}], "]"}]}], ";"}],
   "\[IndentingNewLine]"}], "\[IndentingNewLine]", 
 RowBox[{
  RowBox[{"sm0fm1delta1c1RPlot", " ", "=", " ", 
   RowBox[{"RegionPlot", "[", 
    RowBox[{
     RowBox[{"{", 
      RowBox[{
       RowBox[{"(", 
        RowBox[{"unstabConditions0", "/.", 
         RowBox[{"{", 
          RowBox[{
           RowBox[{"sm", "\[Rule]", "0"}], ",", 
           RowBox[{"\[Delta]", "\[Rule]", "1"}], ",", 
           RowBox[{"c", "\[Rule]", "1"}]}], "}"}]}], ")"}], ",", 
       RowBox[{
        RowBox[{"(", 
         RowBox[{"unstabConditions0", "/.", 
          RowBox[{"{", 
           RowBox[{
            RowBox[{"sm", "\[Rule]", "0"}], ",", 
            RowBox[{"\[Delta]", "\[Rule]", "1"}], ",", 
            RowBox[{"c", "\[Rule]", "1"}]}], "}"}]}], ")"}], "&&", 
        RowBox[{"(", 
         RowBox[{"unstabConditions1", "/.", 
          RowBox[{"{", 
           RowBox[{
            RowBox[{"sm", "\[Rule]", "0"}], ",", 
            RowBox[{"\[Delta]", "\[Rule]", "1"}], ",", 
            RowBox[{"c", "\[Rule]", "1"}]}], "}"}]}], ")"}]}]}], "}"}], ",", 
     RowBox[{"{", 
      RowBox[{"sf", ",", "0", ",", "1"}], "}"}], ",", 
     RowBox[{"{", 
      RowBox[{"h", ",", "0", ",", "1"}], "}"}], ",", 
     RowBox[{"FrameTicksStyle", "\[Rule]", 
      RowBox[{"Directive", "[", 
       RowBox[{"FontSize", "\[Rule]", "12"}], "]"}]}], ",", 
     RowBox[{"PlotStyle", "\[Rule]", 
      RowBox[{"Opacity", "[", "0", "]"}]}], ",", 
     RowBox[{"Epilog", "\[Rule]", 
      RowBox[{"{", 
       RowBox[{
        RowBox[{"Text", "[", 
         RowBox[{
          RowBox[{"Style", "[", 
           RowBox[{"\"\<drive invades\\nand fixes\>\"", ",", 
            RowBox[{"FontSize", "\[Rule]", "15"}], ",", "Black"}], "]"}], ",", 
          RowBox[{"{", 
           RowBox[{"0.06", ",", "0.66"}], "}"}], ",", 
          RowBox[{"{", 
           RowBox[{
            RowBox[{"-", "1"}], ",", "0"}], "}"}]}], "]"}], ",", 
        "\[IndentingNewLine]", 
        RowBox[{"Text", "[", 
         RowBox[{
          RowBox[{"Style", "[", 
           RowBox[{"\"\<drive doesn't invade\>\"", ",", 
            RowBox[{"FontSize", "\[Rule]", "15"}], ",", "Black"}], "]"}], ",", 
          RowBox[{"{", 
           RowBox[{"0.55", ",", "0.7"}], "}"}], ",", 
          RowBox[{"{", 
           RowBox[{
            RowBox[{"-", "1"}], ",", "0"}], "}"}]}], "]"}], ",", 
        RowBox[{"Text", "[", 
         RowBox[{
          RowBox[{"Style", "[", 
           RowBox[{"\"\<polymorphism\>\"", ",", 
            RowBox[{"FontSize", "\[Rule]", "15"}], ",", "Black"}], "]"}], ",", 
          RowBox[{"{", 
           RowBox[{"0.5", ",", "0.25"}], "}"}], ",", 
          RowBox[{"{", 
           RowBox[{
            RowBox[{"-", "1"}], ",", "0"}], "}"}]}], "]"}]}], "}"}]}]}], 
    "]"}]}], ";"}], "\[IndentingNewLine]", 
 RowBox[{
  RowBox[{"plot1", " ", "=", " ", 
   RowBox[{"Labeled", "[", 
    RowBox[{
     RowBox[{"Overlay", "[", 
      RowBox[{"{", 
       RowBox[{"b3sm0fm1mm20delta1c1HMap", ",", "sm0fm1delta1c1RPlot"}], 
       "}"}], "]"}], ",", 
     RowBox[{"{", 
      RowBox[{
       RowBox[{"Text", "[", 
        RowBox[{
        "Style", "[", 
         "\"\<female selection (\!\(\*StyleBox[SubscriptBox[\"s\", \
\"f\"],FontSlant->\"Italic\"]\))\>\"", "]"}], "]"}], ",", 
       RowBox[{"Rotate", "[", 
        RowBox[{
         RowBox[{"Text", "[", 
          RowBox[{
          "Style", "[", 
           "\"\<dominance \
(\!\(\*StyleBox[\"h\",FontSlant->\"Italic\"]\))\>\"", "]"}], "]"}], ",", 
         RowBox[{"90", "Degree"}]}], "]"}], ",", 
       RowBox[{"Row", "[", 
        RowBox[{"{", 
         RowBox[{
          RowBox[{"Text", "[", 
           RowBox[{"Style", "[", 
            RowBox[{"\"\<A\>\"", ",", "Bold", ",", 
             RowBox[{"FontSize", "\[Rule]", "20"}]}], "]"}], "]"}], ",", 
          RowBox[{
          "Text", "[", 
           "\"\<   \!\(\*StyleBox[SubscriptBox[\"\[Lambda]\", \
\"m\"],FontWeight->\"Bold\",FontSlant->\"Italic\"]\)\!\(\*StyleBox[\" \
\",FontWeight->\"Bold\"]\)\!\(\*StyleBox[\"=\",FontWeight->\"Bold\"]\)\!\(\*\
StyleBox[\" \
\",FontWeight->\"Bold\"]\)\!\(\*StyleBox[\"20\",FontWeight->\"Bold\"]\)\>\"", 
           "]"}]}], "}"}], "]"}]}], "}"}], ",", 
     RowBox[{"{", 
      RowBox[{"Bottom", ",", "Left", ",", " ", 
       RowBox[{"{", 
        RowBox[{"Top", ",", "Left"}], "}"}]}], "}"}], ",", 
     RowBox[{"ImageSize", "\[Rule]", "Large"}]}], "]"}]}], ";"}]}], "Input",
 CellChangeTimes->{{3.783864032459453*^9, 3.783864058242403*^9}, 
   3.783864110824005*^9, 3.783864635294197*^9, {3.7838649846843224`*^9, 
   3.7838650491938334`*^9}, {3.783865240177537*^9, 3.783865259716415*^9}, {
   3.783866498488085*^9, 3.783866502884075*^9}, 3.783868455035202*^9, {
   3.7838697607894087`*^9, 3.78386978003507*^9}, {3.7839420309269123`*^9, 
   3.783942069017425*^9}, {3.7855703720964327`*^9, 3.785570403112005*^9}, {
   3.785570466933082*^9, 3.785570520570827*^9}, {3.795178390161525*^9, 
   3.795178392925769*^9}, {3.795267239546633*^9, 3.795267267850564*^9}, {
   3.795267425683845*^9, 3.7952674452123413`*^9}, {3.7952677709432793`*^9, 
   3.795267794254538*^9}, {3.795267903565819*^9, 3.795267961053879*^9}, 
   3.8000155078472023`*^9, {3.800288261591107*^9, 3.800288262140415*^9}, {
   3.800288890616727*^9, 3.800288974077612*^9}, 3.800289028567947*^9, {
   3.809687691995982*^9, 3.809687711923624*^9}, {3.809687762053939*^9, 
   3.809687769359063*^9}, {3.809687825292707*^9, 3.809687908043638*^9}, {
   3.809688406170827*^9, 3.809688443392905*^9}, {3.8141970404678383`*^9, 
   3.814197041154595*^9}},
 CellLabel->
  "In[1527]:=",ExpressionUUID->"75fbc257-4811-44f3-8f78-ca480e9a62ac"],

Cell[BoxData[{
 RowBox[{
  RowBox[{"data", " ", "=", " ", 
   RowBox[{
   "Import", "[", "\"\<data/per_hmap_b2.4_sm0_fm1_mm2_delta1_c1.csv\>\"", 
    "]"}]}], ";"}], "\[IndentingNewLine]", 
 RowBox[{
  RowBox[{"b3sm0fm1mm20delta1c1HMap", " ", "=", " ", 
   RowBox[{"MatrixPlot", "[", 
    RowBox[{"data", ",", 
     RowBox[{"DataReversed", "\[Rule]", "True"}], ",", 
     RowBox[{"Frame", "\[Rule]", "True"}], ",", " ", 
     RowBox[{"FrameTicks", "\[Rule]", 
      RowBox[{"{", 
       RowBox[{
        RowBox[{"{", 
         RowBox[{"ticks", ",", "None"}], "}"}], ",", 
        RowBox[{"{", 
         RowBox[{"ticks", ",", "None"}], "}"}]}], "}"}]}], ",", " ", 
     RowBox[{"FrameTicksStyle", "\[Rule]", 
      RowBox[{"Directive", "[", 
       RowBox[{"White", ",", 
        RowBox[{"FontSize", "\[Rule]", "12"}]}], "]"}]}], ",", 
     RowBox[{"FrameStyle", "\[Rule]", 
      RowBox[{"Directive", "[", "White", "]"}]}], ",", 
     RowBox[{"ColorFunctionScaling", "\[Rule]", "False"}], ",", 
     RowBox[{"ColorFunction", "\[Rule]", 
      RowBox[{"(", 
       RowBox[{
        RowBox[{"Lighter", "[", 
         RowBox[{"Orange", ",", 
          RowBox[{"1", "-", "#"}]}], "]"}], "&"}], ")"}]}]}], "]"}]}], 
  ";"}], "\[IndentingNewLine]", 
 RowBox[{
  RowBox[{"sm0fm1delta1c1RPlot", " ", "=", " ", 
   RowBox[{"RegionPlot", "[", 
    RowBox[{
     RowBox[{"{", 
      RowBox[{
       RowBox[{"(", 
        RowBox[{"unstabConditions0", "/.", 
         RowBox[{"{", 
          RowBox[{
           RowBox[{"sm", "\[Rule]", "0"}], ",", 
           RowBox[{"\[Delta]", "\[Rule]", "1"}], ",", 
           RowBox[{"c", "\[Rule]", "1"}]}], "}"}]}], ")"}], ",", 
       RowBox[{
        RowBox[{"(", 
         RowBox[{"unstabConditions0", "/.", 
          RowBox[{"{", 
           RowBox[{
            RowBox[{"sm", "\[Rule]", "0"}], ",", 
            RowBox[{"\[Delta]", "\[Rule]", "1"}], ",", 
            RowBox[{"c", "\[Rule]", "1"}]}], "}"}]}], ")"}], "&&", 
        RowBox[{"(", 
         RowBox[{"unstabConditions1", "/.", 
          RowBox[{"{", 
           RowBox[{
            RowBox[{"sm", "\[Rule]", "0"}], ",", 
            RowBox[{"\[Delta]", "\[Rule]", "1"}], ",", 
            RowBox[{"c", "\[Rule]", "1"}]}], "}"}]}], ")"}]}]}], "}"}], ",", 
     RowBox[{"{", 
      RowBox[{"sf", ",", "0", ",", "1"}], "}"}], ",", 
     RowBox[{"{", 
      RowBox[{"h", ",", "0", ",", "1"}], "}"}], ",", 
     RowBox[{"FrameTicksStyle", "\[Rule]", 
      RowBox[{"Directive", "[", 
       RowBox[{"FontSize", "\[Rule]", "12"}], "]"}]}], ",", 
     RowBox[{"PlotStyle", "\[Rule]", 
      RowBox[{"Opacity", "[", "0", "]"}]}], ",", 
     RowBox[{"Epilog", "\[Rule]", 
      RowBox[{"{", 
       RowBox[{
        RowBox[{"Text", "[", 
         RowBox[{
          RowBox[{"Style", "[", 
           RowBox[{"\"\<drive invades\\nand fixes\>\"", ",", 
            RowBox[{"FontSize", "\[Rule]", "15"}], ",", "Black"}], "]"}], ",", 
          RowBox[{"{", 
           RowBox[{"0.06", ",", "0.66"}], "}"}], ",", 
          RowBox[{"{", 
           RowBox[{
            RowBox[{"-", "1"}], ",", "0"}], "}"}]}], "]"}], ",", 
        "\[IndentingNewLine]", 
        RowBox[{"Text", "[", 
         RowBox[{
          RowBox[{"Style", "[", 
           RowBox[{"\"\<drive doesn't invade\>\"", ",", 
            RowBox[{"FontSize", "\[Rule]", "15"}], ",", "Black"}], "]"}], ",", 
          RowBox[{"{", 
           RowBox[{"0.55", ",", "0.7"}], "}"}], ",", 
          RowBox[{"{", 
           RowBox[{
            RowBox[{"-", "1"}], ",", "0"}], "}"}]}], "]"}], ",", 
        RowBox[{"Text", "[", 
         RowBox[{
          RowBox[{"Style", "[", 
           RowBox[{"\"\<polymorphism\>\"", ",", 
            RowBox[{"FontSize", "\[Rule]", "15"}], ",", "Black"}], "]"}], ",", 
          RowBox[{"{", 
           RowBox[{"0.5", ",", "0.25"}], "}"}], ",", 
          RowBox[{"{", 
           RowBox[{
            RowBox[{"-", "1"}], ",", "0"}], "}"}]}], "]"}]}], "}"}]}]}], 
    "]"}]}], ";"}], "\[IndentingNewLine]", 
 RowBox[{
  RowBox[{"plot2", " ", "=", " ", 
   RowBox[{"Labeled", "[", 
    RowBox[{
     RowBox[{"Overlay", "[", 
      RowBox[{"{", 
       RowBox[{"b3sm0fm1mm20delta1c1HMap", ",", "sm0fm1delta1c1RPlot"}], 
       "}"}], "]"}], ",", 
     RowBox[{"{", 
      RowBox[{
       RowBox[{"Text", "[", 
        RowBox[{
        "Style", "[", 
         "\"\<female selection (\!\(\*StyleBox[SubscriptBox[\"s\", \
\"f\"],FontSlant->\"Italic\"]\))\>\"", "]"}], "]"}], ",", 
       RowBox[{"Rotate", "[", 
        RowBox[{
         RowBox[{"Text", "[", 
          RowBox[{
          "Style", "[", 
           "\"\<dominance \
(\!\(\*StyleBox[\"h\",FontSlant->\"Italic\"]\)\!\(\*StyleBox[\")\",FontSlant->\
\"Italic\"]\)\>\"", "]"}], "]"}], ",", 
         RowBox[{"90", " ", "Degree"}]}], "]"}], ",", 
       RowBox[{"Row", "[", 
        RowBox[{"{", 
         RowBox[{
          RowBox[{"Text", "[", 
           RowBox[{"Style", "[", 
            RowBox[{"\"\<B\>\"", ",", "Bold", ",", 
             RowBox[{"FontSize", "\[Rule]", "20"}]}], "]"}], "]"}], ",", 
          RowBox[{
          "Text", "[", 
           "\"\<   \!\(\*StyleBox[SubscriptBox[\"\[Lambda]\", \
\"m\"],FontWeight->\"Bold\",FontSlant->\"Italic\"]\)\!\(\*StyleBox[\" \
\",FontWeight->\"Bold\"]\)\!\(\*StyleBox[\"=\",FontWeight->\"Bold\"]\)\!\(\*\
StyleBox[\" \",FontWeight->\"Bold\"]\)\!\(\*StyleBox[\"2\",FontWeight->\"Bold\
\"]\)\>\"", "]"}]}], "}"}], "]"}]}], "}"}], ",", 
     RowBox[{"{", 
      RowBox[{"Bottom", ",", "Left", ",", 
       RowBox[{"{", 
        RowBox[{"Top", ",", "Left"}], "}"}]}], "}"}]}], "]"}]}], 
  ";"}], "\[IndentingNewLine]", 
 RowBox[{
  RowBox[{"legend", " ", "=", " ", 
   RowBox[{"BarLegend", "[", 
    RowBox[{
     RowBox[{"{", 
      RowBox[{
       RowBox[{"(", 
        RowBox[{
         RowBox[{"Lighter", "[", 
          RowBox[{"Orange", ",", 
           RowBox[{"1", "-", "#"}]}], "]"}], "&"}], ")"}], ",", 
       RowBox[{"{", 
        RowBox[{"0", ",", "1"}], "}"}]}], "}"}], ",", 
     RowBox[{"LegendLayout", "\[Rule]", "\"\<Row\>\""}], ",", 
     RowBox[{
     "LegendLabel", "\[Rule]", 
      "\"\<    proportion of simulations resulting in polymorphism\>\""}], 
     ",", 
     RowBox[{"LegendMargins", "\[Rule]", 
      RowBox[{"{", 
       RowBox[{
        RowBox[{"{", 
         RowBox[{"20", ",", "0"}], "}"}], ",", 
        RowBox[{"{", 
         RowBox[{"10", ",", "10"}], "}"}]}], "}"}]}]}], "]"}]}], 
  ";"}], "\[IndentingNewLine]"}], "Input",
 CellChangeTimes->{{3.783868450285521*^9, 3.783868465418487*^9}, {
   3.783869800950856*^9, 3.783869803683094*^9}, {3.785077484107643*^9, 
   3.7850774851564217`*^9}, {3.7851489523912373`*^9, 3.785148964245707*^9}, {
   3.7851490055782547`*^9, 3.7851490643785057`*^9}, {3.795268054852161*^9, 
   3.795268095313209*^9}, {3.796043047710814*^9, 3.796043048258975*^9}, 
   3.800015509363681*^9, {3.800269577306037*^9, 3.8002696175784492`*^9}, {
   3.800269657033944*^9, 3.8002696681290617`*^9}, {3.8002697036244164`*^9, 
   3.800269712374805*^9}, {3.800269743450572*^9, 3.800269807159594*^9}, {
   3.800269881349058*^9, 3.800269901943734*^9}, {3.8002877804629602`*^9, 
   3.800287783553776*^9}, {3.800287892864393*^9, 3.800287999155079*^9}, {
   3.800288120615766*^9, 3.8002882101537647`*^9}, {3.8002882663944197`*^9, 
   3.8002882671922417`*^9}, {3.8002883144573193`*^9, 3.800288320341447*^9}, {
   3.80028870924118*^9, 3.800288803448347*^9}, {3.800288875159225*^9, 
   3.8002888752690277`*^9}, {3.800288988539633*^9, 3.800289013085424*^9}, 
   3.800289060991527*^9, {3.8096877737272263`*^9, 3.809687774048586*^9}, {
   3.809687922169828*^9, 3.8096879645395403`*^9}, {3.809688476006184*^9, 
   3.809688516181436*^9}, {3.809691703077112*^9, 3.809691737669281*^9}, {
   3.81419704588407*^9, 3.814197046539134*^9}},
 CellLabel->
  "In[1531]:=",ExpressionUUID->"79eda932-1e09-4bc1-a529-84819b1e2ba6"],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"Style", "[", 
  RowBox[{
   RowBox[{"Grid", "[", 
    RowBox[{"{", 
     RowBox[{
      RowBox[{"{", "plot1", "}"}], " ", ",", 
      RowBox[{"{", "plot2", "}"}], ",", 
      RowBox[{"{", "}"}], ",", 
      RowBox[{"{", "legend", "}"}]}], "}"}], "]"}], ",", 
   RowBox[{"ImageSizeMultipliers", "\[Rule]", 
    RowBox[{"{", 
     RowBox[{"1", ",", "1"}], "}"}]}]}], "]"}]], "Input",
 CellChangeTimes->{{3.796043055741898*^9, 3.796043056909668*^9}, {
  3.809687986629891*^9, 3.809687992019884*^9}, {3.809688131641711*^9, 
  3.809688186852771*^9}, {3.809688636164486*^9, 3.809688650213469*^9}, {
  3.809688738032606*^9, 3.809688794221554*^9}, {3.809691600441263*^9, 
  3.809691644471856*^9}},
 CellLabel->
  "In[1536]:=",ExpressionUUID->"8e8e219b-ab40-4c41-b0e1-d876ee75b557"],

Cell[BoxData[
 StyleBox[
  TagBox[GridBox[{
     {
      TemplateBox[{OverlayBox[{
          GraphicsBox[
           RasterBox[CompressedData["
1:eJztljFMAkEQRUnsqLHF2GFtbbbEFmoqE23FFmtqbLXFmpqaWgsTDWrUCAgc
3AE2WOvN7g5R8jmPkByns4l5OLe3Ozt/Zue2D47zhxuJRGLT/Pm/9XhXcWHZ
zdXTZVclm9XiXrLP9lq2VUrV+sp/mKu/sD1FD3oqTw/u2U7T0l1Vogm3bP/I
FCqNqaM82qbF9iMaPXVx7o8Htut1HXX55U2z+sp2+rfYVY2pv+Ad27V/PdWm
jR8D56N9fS8zhb557znQfkrnfDPnuolcx7DUek+MnuNAvVA+2Lyh1zJPK48z
Wn/fdyc7UPRapb20/yjfNB3DmV3PcxQtl21FruOqiPRC8Q+rC1ofxXNxnOd1
QfmD/EG667wZKDpes/PLPIkvd2m4rI+1I33PKI5DtWXue2tHeqH7GdVp2LoO
myfI/x06p8c6B8Un7tT9y+O4WntYvdC9iuoRzdf3wsDEuxO4L/pu0f1+xM+/
7zsxfk2Wnh93or6vOV8vSK+w341oPqovVI/XV/7wOF+s3eqk9/15LiEiqjsU
f6Qj6ptovtZ7zHpaO9Lxr9ZjVET3v26jQ3Mvzvqj1tXl+rZ2lCdWvxO6/0ei
45oQ9TukL8qTxX1EuG5EdYfqt2jq1t4HUfsvXC3lO+p/UvqvUCgUCoVCoVAo
FAqFQqFwHfkJOLVeow==
            "], {{0, 0}, {21, 21}}, {0, 1}], Frame -> True, 
           FrameLabel -> {None, None}, FrameStyle -> Directive[
             GrayLevel[1]], FrameTicks -> {{{{0.5, 
                FormBox["0", TraditionalForm]}, {4.5, 
                FormBox["0.2`", TraditionalForm]}, {8.5, 
                FormBox["0.4`", TraditionalForm]}, {12.5, 
                FormBox["0.6`", TraditionalForm]}, {16.5, 
                FormBox["0.8`", TraditionalForm]}, {20.5, 
                FormBox["1", TraditionalForm]}}, None}, {{{0.5, 
                FormBox["0", TraditionalForm]}, {4.5, 
                FormBox["0.2`", TraditionalForm]}, {8.5, 
                FormBox["0.4`", TraditionalForm]}, {12.5, 
                FormBox["0.6`", TraditionalForm]}, {16.5, 
                FormBox["0.8`", TraditionalForm]}, {20.5, 
                FormBox["1", TraditionalForm]}}, None}}, FrameTicksStyle -> 
           Directive[
             GrayLevel[1], FontSize -> 12], GridLinesStyle -> Directive[
             GrayLevel[0.5, 0.4]], 
           Method -> {
            "AxisPadding" -> Scaled[0.02], "DefaultBoundaryStyle" -> 
             Automatic, 
             "DefaultGraphicsInteraction" -> {
              "Version" -> 1.2, "TrackMousePosition" -> {True, False}, 
               "Effects" -> {
                "Highlight" -> {"ratio" -> 2}, 
                 "HighlightPoint" -> {"ratio" -> 2}, 
                 "Droplines" -> {
                  "freeformCursorMode" -> True, 
                   "placement" -> {"x" -> "All", "y" -> "None"}}}}, 
             "DefaultPlotStyle" -> Automatic, "DomainPadding" -> Scaled[0.02],
              "RangePadding" -> Scaled[0.05]}], 
          GraphicsBox[{
            GraphicsComplexBox[CompressedData["
1:eJzlmnl0zPf6x0eo2kZ/1LVMItq6YmjsEku1T2KI9l5iJJYrdGqp/DD2uhhL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             "], {{{
                EdgeForm[], 
                Directive[
                 Opacity[0]], 
                GraphicsGroupBox[{
                  PolygonBox[CompressedData["
1:eJxNmHn8V1Max+8pSkkpVCIpJJKUSosoSaVFq2gPIaTs+26KjApZspMlJmMJ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                   "]], 
                  PolygonBox[CompressedData["
1:eJwt08tLVVEUx/Gzo5l/gWTi1dQyLUNJDC9KoihBoIJIIKFQaJGm5ruHZqHV
7eGjgjTn4SMte6ulKXYdKImYNGnc3+DE72L/Bh8W556z9157rXVDtY3lDQeC
IMjFQdS5IPiNCvzj+SbxFTIRwgCWUYJSvEQEh3EC7XiIVOSgD39QjSPowhzy
EMa01ti3pzGKMRxHIt5iRu+myWeC2IZsZGEQf1GDBHTiCY7ijHLbxUXleANv
kIxDuIRNlGGe/R8Tm3BSe0xiSmeN836EeB/FyMA17OACYrCN28q9ClvO19Bq
V6lve5CkNVabu4hVjSzXR8rFcv6FNUR5Pme/45vOOosh3EIB0lCvM+2s/6xZ
If5U7axnH/FBvc1T7Z7pW6vhEn7orkXEYdxBoXryBZ+xx/t84j0sqhZWU5ud
bvXCZiiKdedzj8cGWlX786rFa5xSTewun5zfu0C9uIxj6onl8lV3t5xe4Dni
kIJanWF7RwI/K/16ZzOzgF7nZzFda1edn2XbY9b5GbLZydVsdCh3m5Gnzs9w
WDk9wHXnZ99m/h3ea631rJl4Rb2x/1ALGnS23fmq8zNTpZrabH13vvY2Y/uX
hmnu
                   "]]}]}, {
                EdgeForm[], 
                Directive[
                 Opacity[0]], 
                GraphicsGroupBox[{
                  PolygonBox[CompressedData["
1:eJxNmHmYz1UUxmfOrVCRJWPfyVJkbI09MlqUGUI0USSppFWhon3TnlZt2je0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                   "]], 
                  PolygonBox[CompressedData["
1:eJwtlOmLj1EUx5/fPUPW7Ip4MYksZc0WYwkx1pkh+zIzDGOY+Y2xjOxLIska
IgwisiQ1JSX5G8goa8g2sq9J+Hyd++LT+Z77PM957j3LzSxM55WFJEnyIAPq
W5K8ZGEtrIA8/ClQhS6FReglcAm9Gh7CQPw92OnQGt0IimOMpuhX2HWwCtri
/8bmw3p4BBWs3cBOE+g25s9OQRZ6KOxFF0BLdAGcRy+ExujlcBNdAp3Rv+Je
N2tP+O1gA/o0tEf/xRYGX/sBzcz3dhxamf9DsSvgM9TTP7CH4Sl0MY89L/hZ
SvGvxDP1RPeA7egdsBTdCTahL8N3aIK/ErsxeG6eQGXMUYn5O3qmHGuvZaxd
jXuui14G11UL+Ak5+CewW+ELNDCv3VF4Af3wd2GzYTB6EDxIJck2/GfQ3VzX
pnzvb9D7gp9B706Fk8G/GY6+ozzBfijH/4AtgmuQxk9gProahpjXVLVUDZ9D
H/yd2A7QG10bY2ntXoypWFnQF/0OeyD4Gcaan1FnOwLj8Meb61yYiJ4Ex9Cz
ob95DVW73TGXd+O/ldP3MEo5jWeYYN4T6gXF+AgZ5r1wCEabr0kfhDH42RY1
jETX6Nxxz5PNe0C1r4q9cjvmWj2jWt6PtVFNc/G7wZbgNa2JOa2OMZX7Xua9
pRoE9Cw4gz4LI9AzzHu9HOqg58K54DOj2ayEW8FndBj6dfDZSgev/Z+o1QMd
ef4t+GypZzPxP8Xe00wMMM+xcqs7QHvvaj4bOkORZgkuoBcH/3dD81nXHjS7
j4PfHZph3QVf47u6EzQ7xXAx+AwtUG+Z1/Z/TPM7SXeR3kmhZ8Ka4DXN117M
e0M5aG4+M5oVzVwL8xwqd/pmjmKb11I5fRtzqlyqp/8BHFSjSA==
                   "]]}]}, {}, {}, {}, {}, {}}, {{
                Directive[
                 RGBColor[0.368417, 0.506779, 0.709798], 
                 AbsoluteThickness[1.6]], 
                LineBox[CompressedData["
1:eJwt0kk3FQAYx+GLikLoXtLSju9Q3wEfwdCcW0gUGYqSNM8TRZSSFBVlSlSX
SGhaWbdta+G551g857yr9/wXv6yCcH5JTCAQ6CWW6J1ODrvIo5gqWmmjjxEi
LLPCP/6zygZPkggSIp0MtpPJDhr4xSme0E09I0QHlXGfOn5yiMd84z0fGKaW
qyxzkHvc5SRLHKCTWd4xxFtquMIi+ynlDtX8YB8dXOIvl7nNCRbYyyNmGOQN
AxznIt/ZzRFa+cMFqujnNa/Yw0M+84VpInzlFpXMU0w7LfzmPDc5TDPnOMYL
XtJHEW1M8omPnGWK69ygiTNU8IxenlPIA8aZYIxGrhFmlNM8pZwuejjKHAXR
jVqopIhcdpJNiCDbSCOVFLaSTBKJbGEzCcSziY3Rvohb73UNscFaxQ==
                 "]]}, {
                Directive[
                 RGBColor[0.880722, 0.611041, 0.142051], 
                 AbsoluteThickness[1.6]], 
                LineBox[CompressedData["
1:eJwl0mdbjmEAx+H7qZQVRZkZFTLTEBkpNIWGTYooIpUR+ja+Ct/ASrR3IYSI
rNPhxXn8/td1PS+eF3dibXNlUygIgkc8DguCToeHOsOp8CDYyz62kcqA+ycs
s5/qTf2pK3Q5p9lPLjmMecvTX9qqv3Wc7XYa6TxzPqBvNEMzGbSfs9J+oU36
R1dpAmc4yyHeuj+o7zTQZtrtfJ3QLN3JsN3BGvu13tJIJu1EjeIes/nkbpdm
s5su52Tt0TsazZS9XhfwgIV8dXddb3CVBq6xgUaOc4JKvvvtHj3JD3ujbqLf
TtGtbGYLFZQz7W2R3qXXHiXGbiOWb85HtYxj9Dmv08tcoZZ66ljKF++3tVuT
dB73mc8SRtzP0bl8tg/rEUrpdF6rNVykmllcIo54wongo98WaQnFvHJerVVc
4DxhLGbI2zkNUcB75x1ayAe7RV9qDmn//jdxRDLt+x1njFFGGGaIQQbop49e
euimK/T/+/8Ln2Nc0Q==
                 "]]}}}], {}}, {
           DisplayFunction -> Identity, PlotRangePadding -> {{
               Scaled[0.02], 
               Scaled[0.02]}, {
               Scaled[0.02], 
               Scaled[0.02]}}, PlotRangeClipping -> True, ImagePadding -> All,
             DisplayFunction -> Identity, AspectRatio -> 1, 
            Axes -> {False, False}, AxesLabel -> {None, None}, 
            AxesOrigin -> {Automatic, Automatic}, DisplayFunction :> Identity,
             Epilog -> {
              InsetBox[
               FormBox[
                StyleBox["\"drive invades\\nand fixes\"", FontSize -> 15, 
                 GrayLevel[0], StripOnInput -> False], TraditionalForm], {
               0.06, 0.66}, {-1, 0}], 
              InsetBox[
               FormBox[
                StyleBox["\"drive doesn't invade\"", FontSize -> 15, 
                 GrayLevel[0], StripOnInput -> False], TraditionalForm], {
               0.55, 0.7}, {-1, 0}], 
              InsetBox[
               FormBox[
                StyleBox["\"polymorphism\"", FontSize -> 15, 
                 GrayLevel[0], StripOnInput -> False], TraditionalForm], {0.5,
                0.25}, {-1, 0}]}, Frame -> {{True, True}, {True, True}}, 
            FrameLabel -> {{None, None}, {None, None}}, 
            FrameTicks -> {{Automatic, Automatic}, {Automatic, Automatic}}, 
            FrameTicksStyle -> Directive[FontSize -> 12], GridLinesStyle -> 
            Directive[
              GrayLevel[0.5, 0.4]], 
            Method -> {
             "DefaultGraphicsInteraction" -> {
               "Version" -> 1.2, "TrackMousePosition" -> {True, False}, 
                "Effects" -> {
                 "Highlight" -> {"ratio" -> 2}, 
                  "HighlightPoint" -> {"ratio" -> 2}, 
                  "Droplines" -> {
                   "freeformCursorMode" -> True, 
                    "placement" -> {"x" -> "All", "y" -> "None"}}}}, 
              "ScalingFunctions" -> None, "TransparentPolygonMesh" -> True, 
              "AxesInFront" -> True}, PlotRange -> {{0, 1}, {0, 1}}, 
            PlotRangeClipping -> True, PlotRangePadding -> {
              Scaled[0.02], 
              Scaled[0.02]}, Ticks -> {Automatic, Automatic}}]}],
        InterpretationBox[
         Cell[
          BoxData[
           FormBox[
            StyleBox[
            "\"female selection (\\!\\(\\*StyleBox[SubscriptBox[\\\"s\\\", \\\
\"f\\\"],FontSlant->\\\"Italic\\\"]\\))\"", StripOnInput -> False], 
            TextForm]], "InlineText"], 
         Text[
          Style[
          "female selection (\!\(\*StyleBox[SubscriptBox[\"s\", \
\"f\"],FontSlant->\"Italic\"]\))"]]],RotationBox[
         InterpretationBox[
          Cell[
           BoxData[
            FormBox[
             StyleBox[
             "\"dominance \
(\\!\\(\\*StyleBox[\\\"h\\\",FontSlant->\\\"Italic\\\"]\\))\"", StripOnInput -> 
              False], TextForm]], "InlineText"], 
          Text[
           Style[
           "dominance (\!\(\*StyleBox[\"h\",FontSlant->\"Italic\"]\))"]]], 
         BoxRotation -> 1.5707963267948966`],TemplateBox[{
          InterpretationBox[
           Cell[
            BoxData[
             FormBox[
              StyleBox["\"A\"", Bold, FontSize -> 20, StripOnInput -> False], 
              TextForm]], "InlineText"], 
           Text[
            Style["A", Bold, FontSize -> 20]]], 
          InterpretationBox[
           Cell[
            BoxData[
             FormBox[
             "\"   \\!\\(\\*StyleBox[SubscriptBox[\\\"\[Lambda]\\\", \
\\\"m\\\"],FontWeight->\\\"Bold\\\",FontSlant->\\\"Italic\\\"]\\)\\!\\(\\*\
StyleBox[\\\" \
\\\",FontWeight->\\\"Bold\\\"]\\)\\!\\(\\*StyleBox[\\\"=\\\",FontWeight->\\\"\
Bold\\\"]\\)\\!\\(\\*StyleBox[\\\" \
\\\",FontWeight->\\\"Bold\\\"]\\)\\!\\(\\*StyleBox[\\\"20\\\",FontWeight->\\\"\
Bold\\\"]\\)\"", TextForm]], "InlineText"], 
           Text[
           "   \!\(\*StyleBox[SubscriptBox[\"\[Lambda]\", \"m\"],FontWeight->\
\"Bold\",FontSlant->\"Italic\"]\)\!\(\*StyleBox[\" \",FontWeight->\"Bold\"]\)\
\!\(\*StyleBox[\"=\",FontWeight->\"Bold\"]\)\!\(\*StyleBox[\" \",FontWeight->\
\"Bold\"]\)\!\(\*StyleBox[\"20\",FontWeight->\"Bold\"]\)"]]}, "RowDefault"]},
       "Labeled",
       DisplayFunction->(GridBox[{{"", 
           
           ItemBox[#4, Alignment -> {Left, Inherited}, DefaultBaseStyle -> 
            "LabeledLabel"]}, {
           ItemBox[#3, DefaultBaseStyle -> "LabeledLabel"], 
           TagBox[
            ItemBox[
             PaneBox[
              TagBox[#, "SkipImageSizeLevel"], 
              Alignment -> {Center, Baseline}, BaselinePosition -> Baseline, 
              ImageSize -> Large], DefaultBaseStyle -> "Labeled"], 
            "SkipImageSizeLevel"]}, {"", 
           ItemBox[#2, DefaultBaseStyle -> "LabeledLabel"]}}, 
         GridBoxAlignment -> {"Columns" -> {{Center}}, "Rows" -> {{Center}}}, 
         AutoDelete -> False, 
         GridBoxItemSize -> {
          "Columns" -> {{Automatic}}, "Rows" -> {{Automatic}}}, 
         BaselinePosition -> {2, 2}]& ),
       InterpretationFunction->(RowBox[{"Labeled", "[", 
          RowBox[{#, ",", 
            RowBox[{"{", 
              RowBox[{#2, ",", #3, ",", #4}], "}"}], ",", 
            RowBox[{"{", 
              RowBox[{"Bottom", ",", "Left", ",", 
                RowBox[{"{", 
                  RowBox[{"Top", ",", "Left"}], "}"}]}], "}"}], ",", 
            RowBox[{"ImageSize", "\[Rule]", "Large"}]}], "]"}]& )]},
     {
      TemplateBox[{OverlayBox[{
          GraphicsBox[
           RasterBox[CompressedData["
1:eJztljtMAkEQhknsrLHF2GFtrVNiCzWVibZSY02NrbZaU1NTa2Gi8RE1gh4c
cDwarfVmbtbE5Hdj44VjNjGf+bOPufl3ZtnYO6zsr+RyubXkL/5fxpyyylap
W8+3ZlRtdt6L1dkPfUDtQmNcbj87vba9entWCyie3ezceOcf8OjT6Uk87p2e
5wV9qvCCO6fLvgHFp3wd5Z2P4uEwCgHVeeG1dx8U/7Jxi8fY+aa65Cukc074
i9ffI070G0XleOGV11/Je596bNiD0+VeDhKfn5wu6wdU5AmP3vkSd0hsc6nr
dIl7mHz3a+r5T4uXF/GIXD5UR3lDeRaGCb/vCbo/yEdUvx88MUzuVde7P4rn
9/iXh1KPE+J0N8Ze31Fdo776V192+bwhsT3NnjeeHb5AIxeX71yJe0jSfpa3
3tG7j+oL+YLyifZB/Rbtj/yV+zalTf6AyNvH0PcahcjHY/ZjROtJP1Yd+SXl
OXLvvOrIR+07+g6ojvxV/+T3wjT1vC06kV/IF/T7ENWdvK8Td47qVo/pUutK
30nVpT4j179Vl/qfuDpW3epxsYh8RPVr/maDqA+bv9mmvbNGo9FoNBqNRqPR
aDQajUaj8b/5CRGr5+4=
            "], {{0, 0}, {21, 21}}, {0, 1}], Frame -> True, 
           FrameLabel -> {None, None}, FrameStyle -> Directive[
             GrayLevel[1]], FrameTicks -> {{{{0.5, 
                FormBox["0", TraditionalForm]}, {4.5, 
                FormBox["0.2`", TraditionalForm]}, {8.5, 
                FormBox["0.4`", TraditionalForm]}, {12.5, 
                FormBox["0.6`", TraditionalForm]}, {16.5, 
                FormBox["0.8`", TraditionalForm]}, {20.5, 
                FormBox["1", TraditionalForm]}}, None}, {{{0.5, 
                FormBox["0", TraditionalForm]}, {4.5, 
                FormBox["0.2`", TraditionalForm]}, {8.5, 
                FormBox["0.4`", TraditionalForm]}, {12.5, 
                FormBox["0.6`", TraditionalForm]}, {16.5, 
                FormBox["0.8`", TraditionalForm]}, {20.5, 
                FormBox["1", TraditionalForm]}}, None}}, FrameTicksStyle -> 
           Directive[
             GrayLevel[1], FontSize -> 12], GridLinesStyle -> Directive[
             GrayLevel[0.5, 0.4]], 
           Method -> {
            "AxisPadding" -> Scaled[0.02], "DefaultBoundaryStyle" -> 
             Automatic, 
             "DefaultGraphicsInteraction" -> {
              "Version" -> 1.2, "TrackMousePosition" -> {True, False}, 
               "Effects" -> {
                "Highlight" -> {"ratio" -> 2}, 
                 "HighlightPoint" -> {"ratio" -> 2}, 
                 "Droplines" -> {
                  "freeformCursorMode" -> True, 
                   "placement" -> {"x" -> "All", "y" -> "None"}}}}, 
             "DefaultPlotStyle" -> Automatic, "DomainPadding" -> Scaled[0.02],
              "RangePadding" -> Scaled[0.05]}], 
          GraphicsBox[{
            GraphicsComplexBox[CompressedData["
1:eJzlmnl0zPf6x0eo2kZ/1LVMItq6YmjsEku1T2KI9l5iJJYrdGqp/DD2uhhL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             "], {{{
                EdgeForm[], 
                Directive[
                 Opacity[0]], 
                GraphicsGroupBox[{
                  PolygonBox[CompressedData["
1:eJxNmHn8V1Max+8pSkkpVCIpJJKUSosoSaVFq2gPIaTs+26KjApZspMlJmMJ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                   "]], 
                  PolygonBox[CompressedData["
1:eJwt08tLVVEUx/Gzo5l/gWTi1dQyLUNJDC9KoihBoIJIIKFQaJGm5ruHZqHV
7eGjgjTn4SMte6ulKXYdKImYNGnc3+DE72L/Bh8W556z9157rXVDtY3lDQeC
IMjFQdS5IPiNCvzj+SbxFTIRwgCWUYJSvEQEh3EC7XiIVOSgD39QjSPowhzy
EMa01ti3pzGKMRxHIt5iRu+myWeC2IZsZGEQf1GDBHTiCY7ijHLbxUXleANv
kIxDuIRNlGGe/R8Tm3BSe0xiSmeN836EeB/FyMA17OACYrCN28q9ClvO19Bq
V6lve5CkNVabu4hVjSzXR8rFcv6FNUR5Pme/45vOOosh3EIB0lCvM+2s/6xZ
If5U7axnH/FBvc1T7Z7pW6vhEn7orkXEYdxBoXryBZ+xx/t84j0sqhZWU5ud
bvXCZiiKdedzj8cGWlX786rFa5xSTewun5zfu0C9uIxj6onl8lV3t5xe4Dni
kIJanWF7RwI/K/16ZzOzgF7nZzFda1edn2XbY9b5GbLZydVsdCh3m5Gnzs9w
WDk9wHXnZ99m/h3ea631rJl4Rb2x/1ALGnS23fmq8zNTpZrabH13vvY2Y/uX
hmnu
                   "]]}]}, {
                EdgeForm[], 
                Directive[
                 Opacity[0]], 
                GraphicsGroupBox[{
                  PolygonBox[CompressedData["
1:eJxNmHmYz1UUxmfOrVCRJWPfyVJkbI09MlqUGUI0USSppFWhon3TnlZt2je0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                   "]], 
                  PolygonBox[CompressedData["
1:eJwtlOmLj1EUx5/fPUPW7Ip4MYksZc0WYwkx1pkh+zIzDGOY+Y2xjOxLIska
IgwisiQ1JSX5G8goa8g2sq9J+Hyd++LT+Z77PM957j3LzSxM55WFJEnyIAPq
W5K8ZGEtrIA8/ClQhS6FReglcAm9Gh7CQPw92OnQGt0IimOMpuhX2HWwCtri
/8bmw3p4BBWs3cBOE+g25s9OQRZ6KOxFF0BLdAGcRy+ExujlcBNdAp3Rv+Je
N2tP+O1gA/o0tEf/xRYGX/sBzcz3dhxamf9DsSvgM9TTP7CH4Sl0MY89L/hZ
SvGvxDP1RPeA7egdsBTdCTahL8N3aIK/ErsxeG6eQGXMUYn5O3qmHGuvZaxd
jXuui14G11UL+Ak5+CewW+ELNDCv3VF4Af3wd2GzYTB6EDxIJck2/GfQ3VzX
pnzvb9D7gp9B706Fk8G/GY6+ozzBfijH/4AtgmuQxk9gProahpjXVLVUDZ9D
H/yd2A7QG10bY2ntXoypWFnQF/0OeyD4Gcaan1FnOwLj8Meb61yYiJ4Ex9Cz
ob95DVW73TGXd+O/ldP3MEo5jWeYYN4T6gXF+AgZ5r1wCEabr0kfhDH42RY1
jETX6Nxxz5PNe0C1r4q9cjvmWj2jWt6PtVFNc/G7wZbgNa2JOa2OMZX7Xua9
pRoE9Cw4gz4LI9AzzHu9HOqg58K54DOj2ayEW8FndBj6dfDZSgev/Z+o1QMd
ef4t+GypZzPxP8Xe00wMMM+xcqs7QHvvaj4bOkORZgkuoBcH/3dD81nXHjS7
j4PfHZph3QVf47u6EzQ7xXAx+AwtUG+Z1/Z/TPM7SXeR3kmhZ8Ka4DXN117M
e0M5aG4+M5oVzVwL8xwqd/pmjmKb11I5fRtzqlyqp/8BHFSjSA==
                   "]]}]}, {}, {}, {}, {}, {}}, {{
                Directive[
                 RGBColor[0.368417, 0.506779, 0.709798], 
                 AbsoluteThickness[1.6]], 
                LineBox[CompressedData["
1:eJwt0kk3FQAYx+GLikLoXtLSju9Q3wEfwdCcW0gUGYqSNM8TRZSSFBVlSlSX
SGhaWbdta+G551g857yr9/wXv6yCcH5JTCAQ6CWW6J1ODrvIo5gqWmmjjxEi
LLPCP/6zygZPkggSIp0MtpPJDhr4xSme0E09I0QHlXGfOn5yiMd84z0fGKaW
qyxzkHvc5SRLHKCTWd4xxFtquMIi+ynlDtX8YB8dXOIvl7nNCRbYyyNmGOQN
AxznIt/ZzRFa+cMFqujnNa/Yw0M+84VpInzlFpXMU0w7LfzmPDc5TDPnOMYL
XtJHEW1M8omPnGWK69ygiTNU8IxenlPIA8aZYIxGrhFmlNM8pZwuejjKHAXR
jVqopIhcdpJNiCDbSCOVFLaSTBKJbGEzCcSziY3Rvohb73UNscFaxQ==
                 "]]}, {
                Directive[
                 RGBColor[0.880722, 0.611041, 0.142051], 
                 AbsoluteThickness[1.6]], 
                LineBox[CompressedData["
1:eJwl0mdbjmEAx+H7qZQVRZkZFTLTEBkpNIWGTYooIpUR+ja+Ct/ASrR3IYSI
rNPhxXn8/td1PS+eF3dibXNlUygIgkc8DguCToeHOsOp8CDYyz62kcqA+ycs
s5/qTf2pK3Q5p9lPLjmMecvTX9qqv3Wc7XYa6TxzPqBvNEMzGbSfs9J+oU36
R1dpAmc4yyHeuj+o7zTQZtrtfJ3QLN3JsN3BGvu13tJIJu1EjeIes/nkbpdm
s5su52Tt0TsazZS9XhfwgIV8dXddb3CVBq6xgUaOc4JKvvvtHj3JD3ujbqLf
TtGtbGYLFZQz7W2R3qXXHiXGbiOWb85HtYxj9Dmv08tcoZZ66ljKF++3tVuT
dB73mc8SRtzP0bl8tg/rEUrpdF6rNVykmllcIo54wongo98WaQnFvHJerVVc
4DxhLGbI2zkNUcB75x1ayAe7RV9qDmn//jdxRDLt+x1njFFGGGaIQQbop49e
euimK/T/+/8Ln2Nc0Q==
                 "]]}}}], {}}, {
           DisplayFunction -> Identity, PlotRangePadding -> {{
               Scaled[0.02], 
               Scaled[0.02]}, {
               Scaled[0.02], 
               Scaled[0.02]}}, PlotRangeClipping -> True, ImagePadding -> All,
             DisplayFunction -> Identity, AspectRatio -> 1, 
            Axes -> {False, False}, AxesLabel -> {None, None}, 
            AxesOrigin -> {Automatic, Automatic}, DisplayFunction :> Identity,
             Epilog -> {
              InsetBox[
               FormBox[
                StyleBox["\"drive invades\\nand fixes\"", FontSize -> 15, 
                 GrayLevel[0], StripOnInput -> False], TraditionalForm], {
               0.06, 0.66}, {-1, 0}], 
              InsetBox[
               FormBox[
                StyleBox["\"drive doesn't invade\"", FontSize -> 15, 
                 GrayLevel[0], StripOnInput -> False], TraditionalForm], {
               0.55, 0.7}, {-1, 0}], 
              InsetBox[
               FormBox[
                StyleBox["\"polymorphism\"", FontSize -> 15, 
                 GrayLevel[0], StripOnInput -> False], TraditionalForm], {0.5,
                0.25}, {-1, 0}]}, Frame -> {{True, True}, {True, True}}, 
            FrameLabel -> {{None, None}, {None, None}}, 
            FrameTicks -> {{Automatic, Automatic}, {Automatic, Automatic}}, 
            FrameTicksStyle -> Directive[FontSize -> 12], GridLinesStyle -> 
            Directive[
              GrayLevel[0.5, 0.4]], 
            Method -> {
             "DefaultGraphicsInteraction" -> {
               "Version" -> 1.2, "TrackMousePosition" -> {True, False}, 
                "Effects" -> {
                 "Highlight" -> {"ratio" -> 2}, 
                  "HighlightPoint" -> {"ratio" -> 2}, 
                  "Droplines" -> {
                   "freeformCursorMode" -> True, 
                    "placement" -> {"x" -> "All", "y" -> "None"}}}}, 
              "ScalingFunctions" -> None, "TransparentPolygonMesh" -> True, 
              "AxesInFront" -> True}, PlotRange -> {{0, 1}, {0, 1}}, 
            PlotRangeClipping -> True, PlotRangePadding -> {
              Scaled[0.02], 
              Scaled[0.02]}, Ticks -> {Automatic, Automatic}}]}],
        InterpretationBox[
         Cell[
          BoxData[
           FormBox[
            StyleBox[
            "\"female selection (\\!\\(\\*StyleBox[SubscriptBox[\\\"s\\\", \\\
\"f\\\"],FontSlant->\\\"Italic\\\"]\\))\"", StripOnInput -> False], 
            TextForm]], "InlineText"], 
         Text[
          Style[
          "female selection (\!\(\*StyleBox[SubscriptBox[\"s\", \
\"f\"],FontSlant->\"Italic\"]\))"]]],RotationBox[
         InterpretationBox[
          Cell[
           BoxData[
            FormBox[
             StyleBox[
             "\"dominance \
(\\!\\(\\*StyleBox[\\\"h\\\",FontSlant->\\\"Italic\\\"]\\)\\!\\(\\*StyleBox[\\\
\")\\\",FontSlant->\\\"Italic\\\"]\\)\"", StripOnInput -> False], TextForm]], 
           "InlineText"], 
          Text[
           Style[
           "dominance \
(\!\(\*StyleBox[\"h\",FontSlant->\"Italic\"]\)\!\(\*StyleBox[\")\",FontSlant->\
\"Italic\"]\)"]]], BoxRotation -> 1.5707963267948966`],TemplateBox[{
          InterpretationBox[
           Cell[
            BoxData[
             FormBox[
              StyleBox["\"B\"", Bold, FontSize -> 20, StripOnInput -> False], 
              TextForm]], "InlineText"], 
           Text[
            Style["B", Bold, FontSize -> 20]]], 
          InterpretationBox[
           Cell[
            BoxData[
             FormBox[
             "\"   \\!\\(\\*StyleBox[SubscriptBox[\\\"\[Lambda]\\\", \
\\\"m\\\"],FontWeight->\\\"Bold\\\",FontSlant->\\\"Italic\\\"]\\)\\!\\(\\*\
StyleBox[\\\" \
\\\",FontWeight->\\\"Bold\\\"]\\)\\!\\(\\*StyleBox[\\\"=\\\",FontWeight->\\\"\
Bold\\\"]\\)\\!\\(\\*StyleBox[\\\" \
\\\",FontWeight->\\\"Bold\\\"]\\)\\!\\(\\*StyleBox[\\\"2\\\",FontWeight->\\\"\
Bold\\\"]\\)\"", TextForm]], "InlineText"], 
           Text[
           "   \!\(\*StyleBox[SubscriptBox[\"\[Lambda]\", \"m\"],FontWeight->\
\"Bold\",FontSlant->\"Italic\"]\)\!\(\*StyleBox[\" \",FontWeight->\"Bold\"]\)\
\!\(\*StyleBox[\"=\",FontWeight->\"Bold\"]\)\!\(\*StyleBox[\" \",FontWeight->\
\"Bold\"]\)\!\(\*StyleBox[\"2\",FontWeight->\"Bold\"]\)"]]}, "RowDefault"]},
       "Labeled",
       DisplayFunction->(GridBox[{{"", 
           
           ItemBox[#4, Alignment -> {Left, Inherited}, DefaultBaseStyle -> 
            "LabeledLabel"]}, {
           ItemBox[#3, DefaultBaseStyle -> "LabeledLabel"], 
           TagBox[
            ItemBox[
             PaneBox[
              TagBox[#, "SkipImageSizeLevel"], 
              Alignment -> {Center, Baseline}, BaselinePosition -> Baseline], 
             DefaultBaseStyle -> "Labeled"], "SkipImageSizeLevel"]}, {"", 
           ItemBox[#2, DefaultBaseStyle -> "LabeledLabel"]}}, 
         GridBoxAlignment -> {"Columns" -> {{Center}}, "Rows" -> {{Center}}}, 
         AutoDelete -> False, 
         GridBoxItemSize -> {
          "Columns" -> {{Automatic}}, "Rows" -> {{Automatic}}}, 
         BaselinePosition -> {2, 2}]& ),
       InterpretationFunction->(RowBox[{"Labeled", "[", 
          RowBox[{#, ",", 
            RowBox[{"{", 
              RowBox[{#2, ",", #3, ",", #4}], "}"}], ",", 
            RowBox[{"{", 
              RowBox[{"Bottom", ",", "Left", ",", 
                RowBox[{"{", 
                  RowBox[{"Top", ",", "Left"}], "}"}]}], "}"}]}], "]"}]& )]},
     {""},
     {
      TemplateBox[{StyleBox[
         StyleBox[
          PaneBox[
           TagBox[
            GridBox[{{
               StyleBox[
                TagBox[
                 FormBox[
                 "\"    proportion of simulations resulting in \
polymorphism\"", TraditionalForm], TraditionalForm, Editable -> True], {
                FontFamily -> "Arial"}, Background -> Automatic, StripOnInput -> 
                False]}, {
               GraphicsBox[{
                 RasterBox[CompressedData["
1:eJx11juMVFUcx/ExNoaaGrHDzqileEoshdrKqK2YWBi0sIBWKCzEwmAgMSZk
g1HjatQsi8jDBfc9s7CP+37fGQqxduY/v7Mn+SbeZLPwv3PP//WZ2Xnh7ffP
vPvsYDB4Zvpzb/oz+/f8euL+73d3evHYhXbihlfPnjwynBzG/f8XTsXnji6E
+PLTE29dXJ64C+3swRD3r5venL4ixL+8PLsm7hW7Qtw/f8QShfgs69WzEzc7
7VQ8Poz7c2fVnl4M8TfshRNnty+GuM/3nl0hbo8fmzhLczLEfR1HrZEQ/3fW
7tOx5tQfxn1983mE+IOV2TVW3yHu6/7AEof4NWt4rD5C/JL6ed7qDfGPrcCx
6uoO477Pef4QP2OJx8oT4q+r//l5If6iDXqs50Lcz2V+v4MjP68Orno5auGq
l5cWrnq5aOGq1/5buOq15xaueu2zhatee2vgqtd+Grjq5aqBq17zDnG/p/lc
G7jqNb8GrjrNqYarTvOo4apT3zVcdeqvhqtOfdRw1aneGq461VXBVaf8FVx1
ylPBVafzKrjq9FwFV53uV3DlPVVw1cpRCVetvJRw1cpFCVet9l/CVas9l3DV
ap8lXLXaWwFXrfZTwFWrPRRw1WreBVy1mmsBV63mV8BVoznlcNVoHjlcNeo7
h6tG/eVw1aiPEPfvk/nnaw5XjerK4Kpxln4lg6tGeTK4anReBleNnsvgqtH9
DK78+zCDq1qOUriq5SWFq1ouUriqtf8UrmrtOYWrWvtM4arW3hK4qrWfBK5q
uUrgqta8E7iqNdcErmrNL4GrSnOK4arSPGK4qtR3DFeV+ovhqlIfMVxVchXD
VaW6Iriq5CqCq0p5IriqdF4EV5Wei+Cq0v0IrvznVARXpVwcwFWpfR7AVak9
HMBVqfntw1WpvvfhqlS9+3BVKs8eXJW6vwdXpea/B1el5ryL71elsy2e24Wr
0tm6TuzCVeG2bR6P4apw562PEJ/nK9zLlv8xXBUusecewVWh90WI+8/f+f4f
wVWhve3AVaF578BVoTntwFWh/kZwVaiuEVwVOm8EV/7zfQRXuVwM4SrXPodw
lWsPQ7jKNb9tuMrV9zZc5ap3G65y5dmCq1z3t+Aql5ctuMq15024yuVqE65y
udqEq0yuNuAqk6sNuMrkagOuMrlah6tMjtbhKtP+1+Eq097W4CrTvNfgKtOc
1uAqU3+rcJWprlW4ynTeKlz5v3+rcJVqb3/DVap5PISrVHkewlWqvT+Aq1Tz
XoGrVHtcgatU+/oLrlIX2V7uw1XqXrM53Yer1H1h+e/BVer+scR34Sp1b1rD
d+Eqcdetvztwlbjn7Lk7cJW4d2zPf8JV4pZsfrfhKnHHra7bcJW4T2zQf8BV
4ka2n1twlbhXre9bcJXovGW4SrT/m3CVaK434SpRvUtw5b83LMFVrL5/h6tY
e/4NrmJ5+BWuYu39F7iKtd+f4SrWHhfhKta+foKrWHv5Ea5i95E1/gNcxe4l
+/09XMWusr6/g6vYXbECbsBV5G7YgwtwFbnjVth1uIrcJev3W7iafv+wPN/A
VeQ+tfqvwVXknlidX8NV5JbsvK/gKnJXLO9luPLnfw5X/pzP4Mq//jxc+fiH
cOW/Vw0G/wH1sC5T
                  "], {{-110, 
                    Rational[-15, 2]}, {110, 
                    Rational[15, 2]}}], {Antialiasing -> False, 
                  AbsoluteThickness[0.1], 
                  Directive[
                   Opacity[0.3], 
                   GrayLevel[0]], 
                  LineBox[
                   NCache[{{110, 
                    Rational[-15, 2]}, {-110, 
                    Rational[-15, 2]}, {-110, 
                    Rational[15, 2]}, {110, 
                    Rational[15, 2]}, {110, 
                    Rational[-15, 2]}}, {{
                    110, -7.5}, {-110, -7.5}, {-110, 7.5}, {110, 7.5}, {
                    110, -7.5}}]]}, {
                  CapForm[None], {}}, 
                 StyleBox[{Antialiasing -> False, 
                   StyleBox[
                    LineBox[{{-110., -7.5}, {110., -7.5}}], 
                    Directive[
                    AbsoluteThickness[0.2], 
                    Opacity[0.3], 
                    GrayLevel[0]], StripOnInput -> False], 
                   StyleBox[
                    StyleBox[{{
                    StyleBox[
                    LineBox[{{{-110., -7.5}, 
                    Offset[{0, -4.}, {-110., -7.5}]}, {{-66., -7.5}, 
                    Offset[{0, -4.}, {-66., -7.5}]}, {{-22., -7.5}, 
                    Offset[{0, -4.}, {-22., -7.5}]}, {{22., -7.5}, 
                    Offset[{0, -4.}, {22., -7.5}]}, {{66., -7.5}, 
                    Offset[{0, -4.}, {66., -7.5}]}, {{110., -7.5}, 
                    Offset[{0, -4.}, {110., -7.5}]}}], 
                    Directive[
                    AbsoluteThickness[0.2], 
                    GrayLevel[0.4]], StripOnInput -> False], 
                    StyleBox[
                    LineBox[{{{-99., -7.5}, 
                    Offset[{0., -2.5}, {-99., -7.5}]}, {{-88., -7.5}, 
                    Offset[{0., -2.5}, {-88., -7.5}]}, {{-77., -7.5}, 
                    Offset[{0., -2.5}, {-77., -7.5}]}, {{-55., -7.5}, 
                    Offset[{0., -2.5}, {-55., -7.5}]}, {{-44., -7.5}, 
                    Offset[{0., -2.5}, {-44., -7.5}]}, {{-33., -7.5}, 
                    Offset[{0., -2.5}, {-33., -7.5}]}, {{-11., -7.5}, 
                    Offset[{0., -2.5}, {-11., -7.5}]}, {{0., -7.5}, 
                    Offset[{0., -2.5}, {0., -7.5}]}, {{
                    11.000000000000014`, -7.5}, 
                    Offset[{0., -2.5}, {11.000000000000014`, -7.5}]}, {{
                    33., -7.5}, 
                    Offset[{0., -2.5}, {33., -7.5}]}, {{44., -7.5}, 
                    Offset[{0., -2.5}, {44., -7.5}]}, {{55., -7.5}, 
                    Offset[{0., -2.5}, {55., -7.5}]}, {{77., -7.5}, 
                    Offset[{0., -2.5}, {77., -7.5}]}, {{88., -7.5}, 
                    Offset[{0., -2.5}, {88., -7.5}]}, {{99., -7.5}, 
                    Offset[{0., -2.5}, {99., -7.5}]}}], 
                    Directive[
                    AbsoluteThickness[0.2], 
                    GrayLevel[0.4], 
                    Opacity[0.3]], StripOnInput -> False]}, 
                    StyleBox[
                    StyleBox[{{
                    StyleBox[{
                    InsetBox[
                    FormBox["0", TraditionalForm], 
                    Offset[{0., -7.}, {-110., -7.5}], {0., 1.}, Automatic, {1,
                     0}], 
                    InsetBox[
                    FormBox[
                    TagBox[
                    InterpretationBox["\"0.2\"", 0.2, AutoDelete -> True], 
                    NumberForm[#, {
                    DirectedInfinity[1], 1}]& ], TraditionalForm], 
                    Offset[{0., -7.}, {-66., -7.5}], {0., 1.}, Automatic, {1, 
                    0}], 
                    InsetBox[
                    FormBox[
                    TagBox[
                    InterpretationBox["\"0.4\"", 0.4, AutoDelete -> True], 
                    NumberForm[#, {
                    DirectedInfinity[1], 1}]& ], TraditionalForm], 
                    Offset[{0., -7.}, {-22., -7.5}], {0., 1.}, Automatic, {1, 
                    0}], 
                    InsetBox[
                    FormBox[
                    TagBox[
                    InterpretationBox["\"0.6\"", 0.6, AutoDelete -> True], 
                    NumberForm[#, {
                    DirectedInfinity[1], 1}]& ], TraditionalForm], 
                    Offset[{0., -7.}, {22., -7.5}], {0., 1.}, Automatic, {1, 
                    0}], 
                    InsetBox[
                    FormBox[
                    TagBox[
                    InterpretationBox["\"0.8\"", 0.8, AutoDelete -> True], 
                    NumberForm[#, {
                    DirectedInfinity[1], 1}]& ], TraditionalForm], 
                    Offset[{0., -7.}, {66., -7.5}], {0., 1.}, Automatic, {1, 
                    0}], 
                    InsetBox[
                    FormBox[
                    TagBox[
                    InterpretationBox["\"1.0\"", 1., AutoDelete -> True], 
                    NumberForm[#, {
                    DirectedInfinity[1], 1}]& ], TraditionalForm], 
                    Offset[{0., -7.}, {110., -7.5}], {0., 1.}, Automatic, {1, 
                    0}]}, 
                    Directive[
                    AbsoluteThickness[0.2], 
                    GrayLevel[0.4]], {
                    Directive[
                    Opacity[1], 
                    GrayLevel[0.3]], 
                    Directive[
                    Opacity[1], 
                    GrayLevel[0.3]]}, StripOnInput -> False], 
                    
                    StyleBox[{{}, {}, {}, {}, {}, {}, {}, {}, {}, {}, {}, {}, \
{}, {}, {}}, 
                    Directive[
                    AbsoluteThickness[0.2], 
                    GrayLevel[0.4], 
                    Opacity[0.3]], {
                    Directive[
                    Opacity[1], 
                    GrayLevel[0.3]], 
                    Directive[
                    Opacity[1], 
                    GrayLevel[0.3]]}, StripOnInput -> False]}, {}}, {
                    Directive[
                    Opacity[1], 
                    GrayLevel[0.3]], 
                    Directive[
                    Opacity[1], 
                    GrayLevel[0.3]]}, StripOnInput -> False], "GraphicsLabel",
                     StripOnInput -> False]}, "GraphicsTicks", StripOnInput -> 
                    False], {
                    Directive[
                    AbsoluteThickness[0.2], 
                    GrayLevel[0.4]], 
                    Directive[
                    AbsoluteThickness[0.2], 
                    GrayLevel[0.4], 
                    Opacity[0.3]]}, StripOnInput -> False]}, "GraphicsAxes", 
                  StripOnInput -> False]}, PlotRangePadding -> Scaled[0.02], 
                PlotRange -> All, Frame -> True, 
                FrameTicks -> {{True, False}, {False, False}}, FrameStyle -> 
                Opacity[0], FrameTicksStyle -> Opacity[0], 
                ImageSize -> {220, Automatic}, BaseStyle -> {}]}}, 
             GridBoxAlignment -> {"Columns" -> {{Center}}}, AutoDelete -> 
             False, GridBoxItemSize -> {
              "Columns" -> {{Automatic}}, "Rows" -> {{Automatic}}}, 
             GridBoxSpacings -> {
              "Columns" -> {{Automatic}}, "Rows" -> {{Automatic}}}], "Grid"], 
           Alignment -> Left, AppearanceElements -> None, 
           ImageMargins -> {{20, 0}, {10, 10}}, ImageSizeAction -> 
           "ResizeToFit"], LineIndent -> 0, StripOnInput -> False], {
         FontFamily -> "Arial"}, Background -> Automatic, StripOnInput -> 
         False]},
       "BarLegend",
       DisplayFunction->(#& ),
       InterpretationFunction:>(RowBox[{"BarLegend", "[", 
          RowBox[{
            RowBox[{"{", 
              RowBox[{
                RowBox[{
                  RowBox[{"Lighter", "[", 
                    RowBox[{"Orange", ",", 
                    RowBox[{"1", "-", "#1"}]}], "]"}], "&"}], ",", 
                RowBox[{"{", 
                  RowBox[{"0", ",", "1"}], "}"}]}], "}"}], ",", 
            RowBox[{"LegendLayout", "\[Rule]", "\"Row\""}], ",", 
            
            RowBox[{"LegendLabel", "\[Rule]", 
              "\"    proportion of simulations resulting in polymorphism\""}],
             ",", 
            RowBox[{"LegendMargins", "\[Rule]", 
              RowBox[{"{", 
                RowBox[{
                  RowBox[{"{", 
                    RowBox[{"20", ",", "0"}], "}"}], ",", 
                  RowBox[{"{", 
                    RowBox[{"10", ",", "10"}], "}"}]}], "}"}]}]}], "]"}]& )]}
    },
    AutoDelete->False,
    GridBoxItemSize->{"Columns" -> {{Automatic}}, "Rows" -> {{Automatic}}}],
   "Grid"],
  StripOnInput->False,
  ImageSizeMultipliers->{1, 1}]], "Output",
 CellChangeTimes->{
  3.79526811748978*^9, 3.796043057804194*^9, 3.800269119294242*^9, 
   3.8002696709371758`*^9, 3.8002697172574577`*^9, 3.80026975220112*^9, {
   3.8002889208676357`*^9, 3.800288977562524*^9}, 3.8002890342206583`*^9, 
   3.809677029249707*^9, {3.809687849706558*^9, 3.8096879113231087`*^9}, 
   3.809687969891954*^9, 3.8096880536618032`*^9, {3.8096881566046257`*^9, 
   3.809688163771237*^9}, {3.809688391081849*^9, 3.809688453778226*^9}, {
   3.809688501710206*^9, 3.8096885221144733`*^9}, 3.809688651295617*^9, 
   3.8096886964154997`*^9, {3.8096887409537907`*^9, 3.809688795293412*^9}, {
   3.80969161086134*^9, 3.809691645547707*^9}, {3.809691730752619*^9, 
   3.809691761712017*^9}, 3.814005339413619*^9, 3.814196574006446*^9, 
   3.8141971302741213`*^9},
 CellLabel->
  "Out[1536]=",ExpressionUUID->"ea2f01aa-b0a2-4b32-bbb9-ecc663a0b60d"]
}, Open  ]],

Cell[BoxData[{
 RowBox[{
  RowBox[{"data", " ", "=", 
   RowBox[{"N", "[", 
    RowBox[{
     RowBox[{
     "Import", "[", "\"\<data/pop_hmap_b2.4_sm0_fm1_mm2_delta1_c1.csv\>\"", 
      "]"}], " ", "-", " ", "161"}], "]"}]}], ";"}], "\[IndentingNewLine]", 
 RowBox[{
  RowBox[{"POPb24sm0fm1mm2delta1c1HMap", " ", "=", " ", 
   RowBox[{"MatrixPlot", "[", 
    RowBox[{"data", ",", 
     RowBox[{"DataReversed", "\[Rule]", "True"}], ",", 
     RowBox[{"Frame", "\[Rule]", "True"}], ",", " ", 
     RowBox[{"FrameTicks", "\[Rule]", 
      RowBox[{"{", 
       RowBox[{
        RowBox[{"{", 
         RowBox[{"ticks", ",", "None"}], "}"}], ",", 
        RowBox[{"{", 
         RowBox[{"ticks", ",", "None"}], "}"}]}], "}"}]}], ",", " ", 
     RowBox[{"FrameTicksStyle", "\[Rule]", 
      RowBox[{"Directive", "[", 
       RowBox[{"White", ",", 
        RowBox[{"FontSize", "\[Rule]", "12"}]}], "]"}]}], ",", 
     RowBox[{"FrameStyle", "\[Rule]", 
      RowBox[{"Directive", "[", "White", "]"}]}], ",", 
     RowBox[{"ColorFunction", "\[Rule]", 
      RowBox[{"(", 
       RowBox[{
        RowBox[{"Blend", "[", 
         RowBox[{
          RowBox[{"{", 
           RowBox[{
            RowBox[{"{", 
             RowBox[{"0", ",", "White"}], "}"}], ",", 
            RowBox[{"{", 
             RowBox[{"250", ",", 
              RowBox[{"Darker", "[", 
               RowBox[{"Green", ",", "0.6"}], "]"}]}], "}"}]}], "}"}], ",", 
          "#"}], "]"}], "&"}], ")"}]}], ",", 
     RowBox[{"ColorFunctionScaling", "\[Rule]", "False"}]}], "]"}]}], 
  ";"}], "\[IndentingNewLine]", 
 RowBox[{
  RowBox[{"sm0fm1delta1c1RPlot", " ", "=", " ", 
   RowBox[{"RegionPlot", "[", 
    RowBox[{
     RowBox[{"{", 
      RowBox[{
       RowBox[{"(", 
        RowBox[{"unstabConditions0", "/.", 
         RowBox[{"{", 
          RowBox[{
           RowBox[{"sm", "\[Rule]", "0"}], ",", 
           RowBox[{"\[Delta]", "\[Rule]", "1"}], ",", 
           RowBox[{"c", "\[Rule]", "1"}]}], "}"}]}], ")"}], ",", 
       RowBox[{
        RowBox[{"(", 
         RowBox[{"unstabConditions0", "/.", 
          RowBox[{"{", 
           RowBox[{
            RowBox[{"sm", "\[Rule]", "0"}], ",", 
            RowBox[{"\[Delta]", "\[Rule]", "1"}], ",", 
            RowBox[{"c", "\[Rule]", "1"}]}], "}"}]}], ")"}], "&&", 
        RowBox[{"(", 
         RowBox[{"unstabConditions1", "/.", 
          RowBox[{"{", 
           RowBox[{
            RowBox[{"sm", "\[Rule]", "0"}], ",", 
            RowBox[{"\[Delta]", "\[Rule]", "1"}], ",", 
            RowBox[{"c", "\[Rule]", "1"}]}], "}"}]}], ")"}]}]}], "}"}], ",", 
     RowBox[{"{", 
      RowBox[{"sf", ",", "0", ",", "1"}], "}"}], ",", 
     RowBox[{"{", 
      RowBox[{"h", ",", "0", ",", "1"}], "}"}], ",", 
     RowBox[{"FrameTicksStyle", "\[Rule]", 
      RowBox[{"Directive", "[", 
       RowBox[{"FontSize", "\[Rule]", "12"}], "]"}]}], ",", 
     RowBox[{"PlotStyle", "\[Rule]", 
      RowBox[{"Opacity", "[", "0", "]"}]}], ",", 
     RowBox[{"Epilog", "\[Rule]", 
      RowBox[{"{", 
       RowBox[{
        RowBox[{"Text", "[", 
         RowBox[{
          RowBox[{"Style", "[", 
           RowBox[{"\"\<drive invades and fixes\>\"", ",", 
            RowBox[{"FontSize", "\[Rule]", "14"}], ",", "Black"}], "]"}], ",", 
          RowBox[{"{", 
           RowBox[{"0.027", ",", "0.66"}], "}"}], ",", 
          RowBox[{"{", 
           RowBox[{
            RowBox[{"-", "1"}], ",", "0"}], "}"}]}], "]"}], ",", 
        "\[IndentingNewLine]", 
        RowBox[{"Text", "[", 
         RowBox[{
          RowBox[{"Style", "[", 
           RowBox[{"\"\<drive doesn't invade\>\"", ",", 
            RowBox[{"FontSize", "\[Rule]", "15"}], ",", "Black"}], "]"}], ",", 
          RowBox[{"{", 
           RowBox[{"0.55", ",", "0.75"}], "}"}], ",", 
          RowBox[{"{", 
           RowBox[{
            RowBox[{"-", "1"}], ",", "0"}], "}"}]}], "]"}], ",", 
        RowBox[{"Text", "[", 
         RowBox[{
          RowBox[{"Style", "[", 
           RowBox[{"\"\<polymorphism\>\"", ",", 
            RowBox[{"FontSize", "\[Rule]", "15"}], ",", "Black"}], "]"}], ",", 
          RowBox[{"{", 
           RowBox[{"0.5", ",", "0.25"}], "}"}], ",", 
          RowBox[{"{", 
           RowBox[{
            RowBox[{"-", "1"}], ",", "0"}], "}"}]}], "]"}]}], "}"}]}]}], 
    "]"}]}], ";"}], "\[IndentingNewLine]", 
 RowBox[{
  RowBox[{"plot4", " ", "=", " ", 
   RowBox[{"Labeled", "[", 
    RowBox[{
     RowBox[{"Overlay", "[", 
      RowBox[{"{", 
       RowBox[{"POPb24sm0fm1mm2delta1c1HMap", ",", "sm0fm1delta1c1RPlot"}], 
       "}"}], "]"}], ",", 
     RowBox[{"{", 
      RowBox[{
       RowBox[{"Text", "[", 
        RowBox[{"Style", "[", 
         RowBox[{"\"\<\!\(\*SubscriptBox[\(s\), \(f\)]\)\>\"", ",", 
          RowBox[{"FontSize", "\[Rule]", "16"}]}], "]"}], "]"}], ",", 
       RowBox[{"Text", "[", 
        RowBox[{"Style", "[", 
         RowBox[{"\"\<h\>\"", ",", 
          RowBox[{"FontSize", "\[Rule]", "16"}]}], "]"}], "]"}], ",", 
       RowBox[{"Text", "[", 
        RowBox[{"Style", "[", 
         RowBox[{"\"\<D\>\"", ",", "Bold", ",", 
          RowBox[{"FontSize", "->", "18"}]}], "]"}], "]"}]}], "}"}], ",", 
     RowBox[{"{", 
      RowBox[{"Bottom", ",", "Left", ",", 
       RowBox[{"{", 
        RowBox[{"Top", ",", "Left"}], "}"}]}], "}"}]}], "]"}]}], 
  ";"}]}], "Input",
 CellChangeTimes->CompressedData["
1:eJwdzs8rg3EAx/FnLHFYm2RYLUR+1KxMrGmmh9lmB5uF4WDzaFlGQ6slP2Ly
4w9g5eB3mB2QtsvyB6xHSwonTmxKQ2kTUTyf7+HT6/I5vEsZt8WRQVGUlBvU
et3yR32SDu0tE6PnplYY7tw2wopRloHzqTYnjPnME1DO+jxwUbH7A3kjQcET
pygSzYehoW4xzFrnS2FPg7AYrvbrZTBlY2qgPbavhF3KOJH/GWiCBX8eDey7
D2qJtE4HnYaNy2wD13F6cQW1mukiO+dZ+qACdkya1LAsc64ZsisCK2S21ERL
SaUNHpceEuvfbofhR5VsHF6fGL3kp7tZhJH0yBEMvM4SRVLxFywfkBDVjs3q
Qc4H/w6xxpyQQ+lCjgJSM6wKFlruiAu1tS1Q5NIQ1WyuAeYxQSM0ClXWlClJ
Zy/xeqF/7CQ+ZU7SvbMFCbhWZ3+BErr9Hf4Gt58pV5Ju/A4T/wF97dwu
  "],
 CellLabel->
  "In[1537]:=",ExpressionUUID->"87999efd-af59-4b0a-b732-41545c56ff4c"],

Cell[BoxData[{
 RowBox[{
  RowBox[{"data", " ", "=", 
   RowBox[{"N", "[", 
    RowBox[{
     RowBox[{
     "Import", "[", "\"\<data/pop_hmap_b2.4_sm0_fm1_mm20_delta1_c1.csv\>\"", 
      "]"}], " ", "-", " ", "161"}], "]"}]}], ";"}], "\[IndentingNewLine]", 
 RowBox[{
  RowBox[{"POPb24sm0fm1mm20delta1c1HMap", " ", "=", " ", 
   RowBox[{"MatrixPlot", "[", 
    RowBox[{"data", ",", 
     RowBox[{"DataReversed", "\[Rule]", "True"}], ",", 
     RowBox[{"Frame", "\[Rule]", "True"}], ",", " ", 
     RowBox[{"FrameTicks", "\[Rule]", 
      RowBox[{"{", 
       RowBox[{
        RowBox[{"{", 
         RowBox[{"ticks", ",", "None"}], "}"}], ",", 
        RowBox[{"{", 
         RowBox[{"ticks", ",", "None"}], "}"}]}], "}"}]}], ",", " ", 
     RowBox[{"FrameTicksStyle", "\[Rule]", 
      RowBox[{"Directive", "[", 
       RowBox[{"White", ",", 
        RowBox[{"FontSize", "\[Rule]", "12"}]}], "]"}]}], ",", 
     RowBox[{"FrameStyle", "\[Rule]", 
      RowBox[{"Directive", "[", "White", "]"}]}], ",", 
     RowBox[{"ColorFunction", "\[Rule]", 
      RowBox[{"(", 
       RowBox[{
        RowBox[{"Blend", "[", 
         RowBox[{
          RowBox[{"{", 
           RowBox[{
            RowBox[{"{", 
             RowBox[{"0", ",", "White"}], "}"}], ",", 
            RowBox[{"{", 
             RowBox[{"250", ",", 
              RowBox[{"Darker", "[", 
               RowBox[{"Green", ",", "0.6"}], "]"}]}], "}"}]}], "}"}], ",", 
          "#"}], "]"}], "&"}], ")"}]}], ",", 
     RowBox[{"ColorFunctionScaling", "\[Rule]", "False"}]}], "]"}]}], 
  ";"}], "\[IndentingNewLine]", 
 RowBox[{
  RowBox[{"sm0fm1delta1c1RPlot", " ", "=", " ", 
   RowBox[{"RegionPlot", "[", 
    RowBox[{
     RowBox[{"{", 
      RowBox[{
       RowBox[{"(", 
        RowBox[{"unstabConditions0", "/.", 
         RowBox[{"{", 
          RowBox[{
           RowBox[{"sm", "\[Rule]", "0"}], ",", 
           RowBox[{"\[Delta]", "\[Rule]", "1"}], ",", 
           RowBox[{"c", "\[Rule]", "1"}]}], "}"}]}], ")"}], ",", 
       RowBox[{
        RowBox[{"(", 
         RowBox[{"unstabConditions0", "/.", 
          RowBox[{"{", 
           RowBox[{
            RowBox[{"sm", "\[Rule]", "0"}], ",", 
            RowBox[{"\[Delta]", "\[Rule]", "1"}], ",", 
            RowBox[{"c", "\[Rule]", "1"}]}], "}"}]}], ")"}], "&&", 
        RowBox[{"(", 
         RowBox[{"unstabConditions1", "/.", 
          RowBox[{"{", 
           RowBox[{
            RowBox[{"sm", "\[Rule]", "0"}], ",", 
            RowBox[{"\[Delta]", "\[Rule]", "1"}], ",", 
            RowBox[{"c", "\[Rule]", "1"}]}], "}"}]}], ")"}]}]}], "}"}], ",", 
     RowBox[{"{", 
      RowBox[{"sf", ",", "0", ",", "1"}], "}"}], ",", 
     RowBox[{"{", 
      RowBox[{"h", ",", "0", ",", "1"}], "}"}], ",", 
     RowBox[{"FrameTicksStyle", "\[Rule]", 
      RowBox[{"Directive", "[", 
       RowBox[{"FontSize", "\[Rule]", "12"}], "]"}]}], ",", 
     RowBox[{"PlotStyle", "\[Rule]", 
      RowBox[{"Opacity", "[", "0", "]"}]}], ",", 
     RowBox[{"Epilog", "\[Rule]", 
      RowBox[{"{", 
       RowBox[{
        RowBox[{"Text", "[", 
         RowBox[{
          RowBox[{"Style", "[", 
           RowBox[{"\"\<drive invades and fixes\>\"", ",", 
            RowBox[{"FontSize", "\[Rule]", "14"}], ",", "Black"}], "]"}], ",", 
          RowBox[{"{", 
           RowBox[{"0.027", ",", "0.66"}], "}"}], ",", 
          RowBox[{"{", 
           RowBox[{
            RowBox[{"-", "1"}], ",", "0"}], "}"}]}], "]"}], ",", 
        "\[IndentingNewLine]", 
        RowBox[{"Text", "[", 
         RowBox[{
          RowBox[{"Style", "[", 
           RowBox[{"\"\<drive doesn't invade\>\"", ",", 
            RowBox[{"FontSize", "\[Rule]", "15"}], ",", "Black"}], "]"}], ",", 
          RowBox[{"{", 
           RowBox[{"0.55", ",", "0.75"}], "}"}], ",", 
          RowBox[{"{", 
           RowBox[{
            RowBox[{"-", "1"}], ",", "0"}], "}"}]}], "]"}], ",", 
        RowBox[{"Text", "[", 
         RowBox[{
          RowBox[{"Style", "[", 
           RowBox[{"\"\<polymorphism\>\"", ",", 
            RowBox[{"FontSize", "\[Rule]", "15"}], ",", "Black"}], "]"}], ",", 
          RowBox[{"{", 
           RowBox[{"0.5", ",", "0.25"}], "}"}], ",", 
          RowBox[{"{", 
           RowBox[{
            RowBox[{"-", "1"}], ",", "0"}], "}"}]}], "]"}]}], "}"}]}]}], 
    "]"}]}], ";"}], "\[IndentingNewLine]", 
 RowBox[{
  RowBox[{"plot3", " ", "=", " ", 
   RowBox[{"Labeled", "[", 
    RowBox[{
     RowBox[{"Overlay", "[", 
      RowBox[{"{", 
       RowBox[{"POPb24sm0fm1mm20delta1c1HMap", ",", "sm0fm1delta1c1RPlot"}], 
       "}"}], "]"}], ",", 
     RowBox[{"{", 
      RowBox[{
       RowBox[{"Text", "[", 
        RowBox[{"Style", "[", 
         RowBox[{"\"\<\!\(\*SubscriptBox[\(s\), \(f\)]\)\>\"", ",", 
          RowBox[{"FontSize", "\[Rule]", "16"}]}], "]"}], "]"}], ",", 
       RowBox[{"Text", "[", 
        RowBox[{"Style", "[", 
         RowBox[{"\"\<h\>\"", ",", 
          RowBox[{"FontSize", "\[Rule]", "16"}]}], "]"}], "]"}], ",", 
       RowBox[{"Text", "[", 
        RowBox[{"Style", "[", 
         RowBox[{"\"\<C\>\"", ",", "Bold", ",", 
          RowBox[{"FontSize", "->", "18"}]}], "]"}], "]"}]}], "}"}], ",", 
     RowBox[{"{", 
      RowBox[{"Bottom", ",", "Left", ",", 
       RowBox[{"{", 
        RowBox[{"Top", ",", "Left"}], "}"}]}], "}"}]}], "]"}]}], 
  ";"}]}], "Input",
 CellChangeTimes->{{3.7960441256332073`*^9, 3.796044147222172*^9}, {
   3.796044186732882*^9, 3.7960442296975603`*^9}, {3.796044298393612*^9, 
   3.7960442985251017`*^9}, {3.796044347875667*^9, 3.796044377730986*^9}, {
   3.796044733677886*^9, 3.796044859194654*^9}, {3.796045031760931*^9, 
   3.796045041218829*^9}, {3.80001555644664*^9, 3.800015562573621*^9}, {
   3.8002890790809383`*^9, 3.800289081790513*^9}, {3.800289133624392*^9, 
   3.80028917270232*^9}, 3.8002892066690083`*^9, {3.80028968911656*^9, 
   3.800289742553235*^9}, {3.814197057378613*^9, 3.814197058005789*^9}},
 CellLabel->
  "In[1541]:=",ExpressionUUID->"4099ea53-91e3-427c-92a1-4137dd3a0fc5"],

Cell[BoxData[
 RowBox[{
  RowBox[{
   RowBox[{"legend2", " ", "=", " ", 
    RowBox[{"BarLegend", "[", 
     RowBox[{
      RowBox[{"{", 
       RowBox[{
        RowBox[{"(", 
         RowBox[{
          RowBox[{"Blend", "[", 
           RowBox[{
            RowBox[{"{", 
             RowBox[{
              RowBox[{"{", 
               RowBox[{"0", ",", "White"}], "}"}], ",", 
              RowBox[{"{", 
               RowBox[{"250", ",", 
                RowBox[{"Darker", "[", 
                 RowBox[{"Green", ",", "0.6"}], "]"}]}], "}"}]}], "}"}], ",", 
            "#"}], "]"}], "&"}], ")"}], ",", 
        RowBox[{"{", 
         RowBox[{"0", ",", "250"}], "}"}]}], "}"}], ",", 
      RowBox[{"LegendLayout", "\[Rule]", "\"\<Row\>\""}], ",", 
      RowBox[{
      "LegendLabel", "\[Rule]", 
       "\"\<    increase in population size relative to WT\>\""}], ",", 
      RowBox[{"LegendMargins", "\[Rule]", 
       RowBox[{"{", 
        RowBox[{
         RowBox[{"{", 
          RowBox[{"25", ",", "0"}], "}"}], ",", 
         RowBox[{"{", 
          RowBox[{"10", ",", "10"}], "}"}]}], "}"}]}]}], "]"}]}], ";"}], 
  "\[IndentingNewLine]"}]], "Input",
 CellChangeTimes->{{3.796043451059404*^9, 3.796043550708375*^9}, {
   3.796043876761118*^9, 3.796043881854583*^9}, {3.796044326742072*^9, 
   3.796044342447345*^9}, 3.8000155655459747`*^9, {3.809762816742606*^9, 
   3.809762868455666*^9}, {3.8097664412673273`*^9, 3.8097664413897343`*^9}, 
   3.809775468956585*^9},
 CellLabel->
  "In[1545]:=",ExpressionUUID->"acc6758e-b3de-48d8-b63f-0eac2af82e07"],

Cell[CellGroupData[{

Cell[BoxData[{
 RowBox[{
  RowBox[{
   RowBox[{"birthPlot2", " ", "=", " ", 
    RowBox[{"Labeled", "[", 
     RowBox[{
      RowBox[{"Overlay", "[", 
       RowBox[{"{", 
        RowBox[{"POPb24sm0fm1mm20delta1c1HMap", ",", "sm0fm1delta1c1RPlot"}], 
        "}"}], "]"}], ",", 
      RowBox[{"{", 
       RowBox[{
        RowBox[{"Text", "[", 
         RowBox[{
         "Style", "[", 
          "\"\<female selection (\!\(\*StyleBox[SubscriptBox[\"s\", \
\"f\"],FontSlant->\"Italic\"]\))\>\"", "]"}], "]"}], ",", 
        RowBox[{"Rotate", "[", 
         RowBox[{
          RowBox[{"Text", "[", 
           RowBox[{
           "Style", "[", 
            "\"\<dominance \
(\!\(\*StyleBox[\"h\",FontSlant->\"Italic\"]\))\>\"", "]"}], "]"}], ",", 
          RowBox[{"90", " ", "Degree"}]}], "]"}], ",", 
        RowBox[{"Text", "[", 
         RowBox[{"Style", "[", 
          RowBox[{"\"\<B\>\"", ",", "Bold", ",", 
           RowBox[{"FontSize", "\[Rule]", "20"}]}], "]"}], "]"}]}], "}"}], 
      ",", 
      RowBox[{"{", 
       RowBox[{"Bottom", ",", "Left", ",", 
        RowBox[{"{", 
         RowBox[{"Top", ",", "Left"}], "}"}]}], "}"}]}], "]"}]}], ";"}], 
  "\[IndentingNewLine]"}], "\[IndentingNewLine]", 
 RowBox[{"Style", "[", 
  RowBox[{
   RowBox[{"Grid", "[", 
    RowBox[{"{", 
     RowBox[{
      RowBox[{"{", "birthPlot", "}"}], ",", 
      RowBox[{"{", "birthPlot2", "}"}], ",", 
      RowBox[{"{", "}"}], ",", 
      RowBox[{"{", "legend2", "}"}]}], "}"}], "]"}], ",", 
   RowBox[{"ImageSizeMultipliers", "\[Rule]", 
    RowBox[{"{", 
     RowBox[{"1", ",", " ", "1"}], "}"}]}]}], "]"}]}], "Input",
 CellChangeTimes->{{3.80976265837462*^9, 3.809762686350124*^9}, {
  3.809762737609069*^9, 3.80976277250113*^9}, {3.8097628897791023`*^9, 
  3.809762941512966*^9}},
 CellLabel->
  "In[1546]:=",ExpressionUUID->"3dca4a35-c88b-497b-ae64-e3bd9f6696a6"],

Cell[BoxData[
 StyleBox[
  TagBox[GridBox[{
     {
      TemplateBox[{TemplateBox[{
          GraphicsBox[{{{{}, {}, 
              TagBox[{
                Directive[
                 Opacity[1.], 
                 RGBColor[0.368417, 0.506779, 0.709798], 
                 AbsoluteThickness[1.6]], 
                LineBox[CompressedData["
1:eJwVjn0wlAkAxhdpfcV633ylkrs92bpw4tTqvA85Kt+VmL0IW9cltSxFd8Lu
a+/U9XWX2+lCajIu5c4sey7nI+Sjayokg0hluvXRZsm2Rbu5vT+eeeaZ+c1v
HuckwfZ9hgwGY6s+/3d6WYDv81daqtKl0VBCXaQq7wYmqGe0lKVrR6Y8eZSK
vBa8if9WSyW4aJnlSjWllmyzf6jVUmuUnLaOCAYu8MPUMNRRQnHERdPVi7HJ
P7K7iqmjqnt2tyzXmkOijS44TeiokbryjOEuEj9/fWWu1E5H5dn0X+qU2OJS
tzK5ermOiuWm6pq5Dqgto8P6XfS8TXXksysrMR4qI5x9dVTGeSuJoJqN0BKL
EjlfRzGGe5rZe9xB+N2WDcl11NApm7oTHlyU5noMcWI+UClvjiXPBgbjYMCy
/fxfFijFI6fNnEXRsL7zRrszk4HqmiLVy73xePteoEq2N4Dbpw9zHL/gY8eb
uH3MTgNgxZ4iRch+FJzpzRQKDeEe3Vc4p01Gp5y3PpIwwljSuvbTnoeRfTla
KL9lhEtdn9Wn16XC1yGJ92PiIoS19NnPsIVIjRltH2cZQ+Z6QuomTUeO9+T9
pY3GkCtCDllNZkB62Si3Jm4xbnoRafKlR8HuOVczYM4EkTNpsSooEw1n3H9K
/YOJ1wvFHOGOLKxTRpv3h5lgo59PeHfYMSy1mjZpmzcB8huORkR8C+NbRVUf
F5viXXtSEG/9d0jxqh80DTEDL/yJI2dlNiq3Jl1u0pjhgWg61nEuG5qXy1gK
qTmkEyZef9cfR+2OYKPCQAtoOnq5FpIc5G2RlMoVFhg5Ef+7cEMuPnqybeCq
ZAlOmvibDTzLha9PfqGDpyVuBvZFjQTnIW5l15xBryVctAqZoTQPmrunR1NF
VhhYu03lOZaHUEN3LodgQem51mHMXQQzFS/KIIkF8gBflpohAt2E68U3WWDR
Tbte1Isg2lkVk2Zqjb1WnpWLtCJw7/cHmSZaY3pX3OLnfmI8tQnMN/nLGm0x
W9wNc8Qw/WSjrT+TwOwPPi87asWYsh2f2BRPIGtsNp39WozYLidCWEUg0Vg1
VbKaRlVMUYHlAoHsg73Z4TwaLa3dOBxNokLg5X3kLI1Y9ZkwZjkJTrD/yW9u
0eh4EOAh+I2EIqXB1beZhleFhui/RiLgkH3FkhYaVvHxA+U3SCw8PBVR3arn
O9fxg2Qk4kodvebbaXgX3cuSNJD4vPhwYv49GkSAWZlRL4ndRUeOiQdp5C1v
+v7gIxJl91MSoh7TmNKkHejtI+FnzlqzaojGPzcG3a4OkPhSmr6vcVjP21TU
BYyQcDaLO655SkM1HtwlmiCx/1e2D19BY/ft97KJSRI3GttCPMZo3C2pKoxS
6v8Opm/W6Xf5djues4rE8Qy7YekEjfgGxYtmNYlHoU41d5Q07kkvdrpqSJzf
Ovqs8BUNblr49XNvSagt/p1JmKJh41IrSJgnUe7W2vhORUPMSN5+5z2JDZXX
stqmaUw/XuHtoSOhdBuzPTej9//ZY3fhAwl2K33hq9d6/1nJ/MICiQJBvs5l
lsZ/5QQZWA==
                 "]]}, Annotation[#, "Charting`Private`Tag$50435#1"]& ], 
              TagBox[{
                Directive[
                 Opacity[1.], 
                 RGBColor[0.880722, 0.611041, 0.142051], 
                 AbsoluteThickness[1.6]], 
                LineBox[CompressedData["
1:eJwV0nk41AkcBvBxjivH/BA9rmm3dFmSSVH7e6kQ1TrCNhsKLaEclS5lxvgV
2Rldpototcoo1i2Jjs0xhS2JNtSqh8Iazz6uGcfs7B/f530+z/M+719fZmis
7z5lGo3mo7j/06DsyKv716Sk8PXcfbPiNLKNHn7xX4XtFiVFCZYWkR9qlX/z
yZWSb01zmDEDw+TYgVvVv9+Wkg69W4W19BmSxoRYT6RwFu1hqLYqrM6ekrRW
SMnWayK5Rbo+9vhNr/cUS8nEpznDZcnmiFO/sq2wXUp2pR/2nd1tBc4DVohG
p5QsynwWckjKRJ7lodTGD1Ky0jqkJsRuCf4e/qfNdUJKfiQclTcKVyI09VOY
s5WMvFoUK7FnOiC8so2/KlFGLq69m93uDvSMV7JsT8pIb7PArf75gL9DTu9q
jozk+mawa+YBt/JoG8dzMrLn9fC0caULrEs1Wl1uykjNHiv3dVabMHTPVSew
UUZaXDBXNxvbgvjbVee4xjOkU8QgfSTRC6cv3Tz7tmqGDPZel+XN8AM165/G
Z8yRvHRBhkdPEHKT7d4vD5wn19YOLH1+NgzRrosiwrLk5PQzbY2E4ggYNE/M
7jxKg/kua9OqN9GYmomVRJkoQV2pUMc7MhZ+E0H76E1KkDQE1IiT45Em6Dia
kKCMPmbltQnmITRVsNd4M1TQ/GJN30n+YSTl+SdUNKjgrcMw2Th3BM6moeyM
vapguNorxZNHERfY//yLvhrE5XKO/ZljOM0aajV8pIbaS56s6bLjEOapJJcH
qaN2SvUWHp7At6/Ol3dr0yEoIvXTWk6iTmB7Ia6Yjk0Xl/n0VifBZsRfu2u7
BnJClXtjsk7BUG9M4w+ZBkYCuuO8jp+GWsONkm+yNSF+YONiuikZMQ4P32l6
aeGipctgvCYH97aG5tVPaiHc0SXJN4qDyeFF+gNCbQjdnH6NfcxBlZ+7yuXN
OgjivC7lmXLB8aByKwZ08GZk9sXnSC4W93p251ML0JOoFhNRzoWzY+plU3td
XE8Yl7XQUhBk0S5V6tDFLx7Se2NbUjAp5vfHcfWgI2dzW/gp2KZs67ScoY+K
uJ5VVm0p0JKwfZRC9eHJypxv1OKBVw9Rdo0+NuRGqit78sDdWRIYr2kAtx+a
eU9SeHBq7XLT3GsAeyfxiZFHPHww2pyqUW2AocAX1uIJHjSXrDd2oTMQUGP3
XH1ZKkaNv3zdEMzANktRaebuVPzYbslIKGFgdGq1YX9GKkoCb6TpyhnI1ho+
/6gmFU+e/omD/gTeh/OrGgcV/XHBdnoBARvn6bssBoXGNle72DsE1Hfd0R5T
2KFwktF1l0A9v9hPRFDQCw7uLigiYGzSc8nCSNFvsglzKyVQx/brVDOhwLrx
8hhVRyBSKNrbbk4p/kTrtkoHgWbPiKYdyylwzOrPRL8hcOXnBR/VVlAYnYzf
39FJgLa/oa9O4Zaid9/ldxOwXx0kWLFK0TcqfODaR6D/s2m1qi0FyRf3du5X
AiYHw9ZUOFDY/Wym9OuQYl/ALtnPoiDOKbnsM0IgwKrSwHIthQLfhWymhECB
qyg93ZFCcN3A58fjBEYHVroHOVF4KbzetGySgOFs56SBMwWn+B2i81MEysp1
+Y0KGy2tit0jI+BxxuWA7UYKKbQo3+YZAgk/uVV/UnjsL3OW3RyBY+6SwSvf
K/YrXy28Ok+gON9F7kUq9jMpmVxOYKDJdU6u8H8+lEzI
                 "]]}, Annotation[#, "Charting`Private`Tag$50435#2"]& ], 
              TagBox[{
                Directive[
                 Opacity[1.], 
                 RGBColor[0.560181, 0.691569, 0.194885], 
                 AbsoluteThickness[1.6]], 
                LineBox[CompressedData["
1:eJwVkXs01AkUx38MeTbG/AYhNKy2x1bY2GLw9SipFOJwZqNQFNZjKtPkOb9S
elBOk80jtBhRZ7EeK2mQWnRS65jyOImsFZOTdmMq0uzsH/fc8z3ncz733nPZ
4fH+h1UJgtijrP+7i6lHsQlBIKnTbGf413xX7vPUlnJl5o5WjMyZPHSV88u8
gmgEroQ0iH5iyVxzTXqkWssInH35Y+Xp7I+u3QdYs3E6BPwCTnp3eC7DIVXH
FDZdyd8cW7FSqANF+QEtKYPArkfDUcMhDDjIqq22GirnaVdJ+q4ZoOy4WzCN
TaB0KsnZa9ECLkZRfzdaEdALW7WL483G8N1LvCOrCdzUDbeLFFmCoRi41Lue
QPa0/eW4tdZIOx/X8bMDgZkl6wtBe9aBW1y47rvdBFwDulaPuNlBv3v+SwCf
gE3y+cWVezkYq20LGBIQ8Ow+PJX9Fwe1+efvhKQo9ymoF/bxneEbY7Y/Ukjg
g8813b4SF+TQt93jXyTQHl0+PpQEaAeKBAUlBKZ2GEovtbtDbdxOPtpFgF3b
8U7g4YWPi/Gz0StUcJkSW3Sr+WLffMhhjS4VVLXV5c9/DUZWTj+fx1NFtdST
mBQdRFcD93tfJg28VOHtRvYhpJQG8hraaDBSe1IS3xIFJ+Nw7sUwNayvWKIK
XkUjIWj80RRDHclu6e5penFIs5f1su6ro3LLDfem9ATkldLS60OWQWy/vYfx
OBHf9F2pH9TRAC9vL0tmegytOZtyE37VgLSJVyHddhwbZgJ1Bnw0sW3RndXN
PQGW3nvNhwuaEMfSfYr5SVBvK6yxKtKC8xpB7ugxPmI33xvS2qUN2VPLB0X8
k7jjHV4qkWsr/1nm8PSoAPK3JozJPB3wa3MNjNxOoWmfF03kqYs3yOH0WCQj
Y0dmScOkLkjfT0tOk8mwHNk5WJa5HNZVy3Oo0hQ4/XBGZGxHB2Ue8eFTWCpC
zJ99VumnI9Q+vnaIngb54+zxBKEeilIjOvkP0rBbdZPjWiYDncbfJvhGpEN7
luunEs6Ak3pkyj8L6TgtQXVRMwM6XapZ1O4MCANqghK19GEc0TKReDUDjr0D
27XC9NHenZIfNZKBUQPPM5q/62Nig6SOYyWElvVWQzcNJmImU5I0w4V4Zzg1
zQllwqKSVVxfJkTwMwsmr4YJ2QmX/OFXQtQEFWbRFUx49FrPzxhS6HjwJ+IC
SZASm0PNfhSC53J8NMQkPns7kJ1nKfzx1N0mvpKE16Zc+p1zFDZXyZkDt0hs
tC2YEmVR0AsNHRTfJmFi3r8l8oKS79oQsb2ORLTtrZ0aORTsC5+czGwlocox
cPS4RoHprl1O6ydxysxN8EsZhYyVkrMxUhLO7Yzgc+UU3skTj/Y/J+HqEW8Y
W0Gh5/bQxrJBEmmpc7b2lUreoOqu+ysSpqFjE4+qKcxOeT0TTpMo9ubgdR2F
/Z2LddMyEsl2tqYPf6Pw+EaNyG+GROwRyxfiegpifyMue5bE3n7WTEwjhdDW
yYn2ORJVY/Lrc80UnuQVdK2RK+9Jvnj8xV0Kjol7qq98JFF60N+muYWCweqm
+IMLJF6+yd2R3EqBIqL9uxeVvn+NKvbfp/B+2MzeZolEoPHb184Spb+xz+j6
VxLXdZkK8zal/3LmgkJBYlRw9YtCmf8D2h4fFw==
                 "]]}, 
               Annotation[#, "Charting`Private`Tag$50435#3"]& ]}}, {}}, {
           DisplayFunction -> Identity, Ticks -> {Automatic, Automatic}, 
            AxesOrigin -> {1., 0}, FrameTicks -> {{Automatic, 
               Charting`ScaledFrameTicks[{Identity, Identity}]}, {Automatic, 
               Charting`ScaledFrameTicks[{Identity, Identity}]}}, 
            GridLines -> {None, None}, DisplayFunction -> Identity, 
            PlotRangePadding -> {{
               Scaled[0.02], 
               Scaled[0.02]}, {
               Scaled[0.05], 
               Scaled[0.05]}}, PlotRangeClipping -> True, ImagePadding -> All,
             DisplayFunction -> Identity, AspectRatio -> 
            NCache[GoldenRatio^(-1), 0.6180339887498948], 
            Axes -> {True, True}, AxesLabel -> {None, None}, 
            AxesOrigin -> {1., 0}, DisplayFunction :> Identity, 
            Frame -> {{False, False}, {False, False}}, 
            FrameLabel -> {{None, None}, {None, None}}, 
            FrameTicks -> {{Automatic, Automatic}, {Automatic, Automatic}}, 
            GridLines -> {None, None}, GridLinesStyle -> Directive[
              GrayLevel[0.5, 0.4]], 
            Method -> {
             "DefaultBoundaryStyle" -> Automatic, 
              "DefaultGraphicsInteraction" -> {
               "Version" -> 1.2, "TrackMousePosition" -> {True, False}, 
                "Effects" -> {
                 "Highlight" -> {"ratio" -> 2}, 
                  "HighlightPoint" -> {"ratio" -> 2}, 
                  "Droplines" -> {
                   "freeformCursorMode" -> True, 
                    "placement" -> {"x" -> "All", "y" -> "None"}}}}, 
              "DefaultMeshStyle" -> AbsolutePointSize[6], "ScalingFunctions" -> 
              None, "CoordinatesToolOptions" -> {"DisplayFunction" -> ({
                  (Identity[#]& )[
                   Part[#, 1]], 
                  (Identity[#]& )[
                   Part[#, 2]]}& ), "CopiedValueFunction" -> ({
                  (Identity[#]& )[
                   Part[#, 1]], 
                  (Identity[#]& )[
                   Part[#, 2]]}& )}}, 
            PlotRange -> {{1, 5}, {0., 679.9999947755102}}, PlotRangeClipping -> 
            True, PlotRangePadding -> {{
               Scaled[0.02], 
               Scaled[0.02]}, {
               Scaled[0.02], 
               Scaled[0.02]}}, Ticks -> {Automatic, Automatic}}], 
          InterpretationBox[
           Cell[
            BoxData[
             FormBox[
             "\"birth rate (\\!\\(\\*StyleBox[\\\"b\\\",FontSlant->\\\"Italic\
\\\"]\\))\"", TextForm]], "InlineText"], 
           Text[
           "birth rate (\!\(\*StyleBox[\"b\",FontSlant->\"Italic\"]\))"]], 
          RotationBox[
           InterpretationBox[
            Cell[
             BoxData[
              FormBox[
              "\"population size \
(\\!\\(\\*StyleBox[\\\"N\\\",FontSlant->\\\"Italic\\\"]\\))\"", TextForm]], 
             "InlineText"], 
            Text[
            "population size \
(\!\(\*StyleBox[\"N\",FontSlant->\"Italic\"]\))"]], BoxRotation -> 
           1.5707963267948966`], 
          InterpretationBox[
           Cell[
            BoxData[
             FormBox[
              StyleBox["\"A\"", Bold, FontSize -> 20, StripOnInput -> False], 
              TextForm]], "InlineText"], 
           Text[
            Style["A", Bold, FontSize -> 20]]]}, "Labeled", 
         DisplayFunction -> (GridBox[{{"", 
             
             ItemBox[#4, Alignment -> {Left, Inherited}, DefaultBaseStyle -> 
              "LabeledLabel"]}, {
             ItemBox[#3, DefaultBaseStyle -> "LabeledLabel"], 
             TagBox[
              ItemBox[
               PaneBox[
                TagBox[#, "SkipImageSizeLevel"], 
                Alignment -> {Center, Baseline}, BaselinePosition -> 
                Baseline], DefaultBaseStyle -> "Labeled"], 
              "SkipImageSizeLevel"]}, {"", 
             ItemBox[#2, DefaultBaseStyle -> "LabeledLabel"]}}, 
           GridBoxAlignment -> {
            "Columns" -> {{Center}}, "Rows" -> {{Center}}}, AutoDelete -> 
           False, GridBoxItemSize -> {
            "Columns" -> {{Automatic}}, "Rows" -> {{Automatic}}}, 
           BaselinePosition -> {2, 2}]& ), 
         InterpretationFunction -> (RowBox[{"Labeled", "[", 
            RowBox[{#, ",", 
              RowBox[{"{", 
                RowBox[{#2, ",", #3, ",", #4}], "}"}], ",", 
              RowBox[{"{", 
                RowBox[{"Bottom", ",", "Left", ",", 
                  RowBox[{"{", 
                    RowBox[{"Top", ",", "Left"}], "}"}]}], "}"}]}], "]"}]& )],
        TemplateBox[{
         "\"strong drive at equilibrium\"", "\"weak drive at fixation\"", 
          "\"wild-type\""}, "LineLegend", DisplayFunction -> (StyleBox[
           StyleBox[
            PaneBox[
             TagBox[
              GridBox[{{
                 TagBox[
                  GridBox[{{
                    GraphicsBox[{{
                    Directive[
                    EdgeForm[
                    Directive[
                    Opacity[0.3], 
                    GrayLevel[0]]], 
                    PointSize[0.5], 
                    AbsoluteThickness[1.6], 
                    RGBColor[0.368417, 0.506779, 0.709798]], {
                    LineBox[{{0, 10}, {20, 10}}]}}, {
                    Directive[
                    EdgeForm[
                    Directive[
                    Opacity[0.3], 
                    GrayLevel[0]]], 
                    PointSize[0.5], 
                    AbsoluteThickness[1.6], 
                    RGBColor[0.368417, 0.506779, 0.709798]], {}}}, 
                    AspectRatio -> Full, ImageSize -> {20, 10}, 
                    PlotRangePadding -> None, ImagePadding -> Automatic, 
                    BaselinePosition -> (Scaled[0.1] -> Baseline)], #}, {
                    GraphicsBox[{{
                    Directive[
                    EdgeForm[
                    Directive[
                    Opacity[0.3], 
                    GrayLevel[0]]], 
                    PointSize[0.5], 
                    AbsoluteThickness[1.6], 
                    RGBColor[0.880722, 0.611041, 0.142051]], {
                    LineBox[{{0, 10}, {20, 10}}]}}, {
                    Directive[
                    EdgeForm[
                    Directive[
                    Opacity[0.3], 
                    GrayLevel[0]]], 
                    PointSize[0.5], 
                    AbsoluteThickness[1.6], 
                    RGBColor[0.880722, 0.611041, 0.142051]], {}}}, 
                    AspectRatio -> Full, ImageSize -> {20, 10}, 
                    PlotRangePadding -> None, ImagePadding -> Automatic, 
                    BaselinePosition -> (Scaled[0.1] -> Baseline)], #2}, {
                    GraphicsBox[{{
                    Directive[
                    EdgeForm[
                    Directive[
                    Opacity[0.3], 
                    GrayLevel[0]]], 
                    PointSize[0.5], 
                    AbsoluteThickness[1.6], 
                    RGBColor[0.560181, 0.691569, 0.194885]], {
                    LineBox[{{0, 10}, {20, 10}}]}}, {
                    Directive[
                    EdgeForm[
                    Directive[
                    Opacity[0.3], 
                    GrayLevel[0]]], 
                    PointSize[0.5], 
                    AbsoluteThickness[1.6], 
                    RGBColor[0.560181, 0.691569, 0.194885]], {}}}, 
                    AspectRatio -> Full, ImageSize -> {20, 10}, 
                    PlotRangePadding -> None, ImagePadding -> Automatic, 
                    BaselinePosition -> (Scaled[0.1] -> Baseline)], #3}}, 
                   GridBoxAlignment -> {
                    "Columns" -> {Center, Left}, "Rows" -> {{Baseline}}}, 
                   AutoDelete -> False, 
                   GridBoxDividers -> {
                    "Columns" -> {{False}}, "Rows" -> {{False}}}, 
                   GridBoxItemSize -> {
                    "Columns" -> {{All}}, "Rows" -> {{All}}}, 
                   GridBoxSpacings -> {
                    "Columns" -> {{0.5}}, "Rows" -> {{0.8}}}], "Grid"]}}, 
               GridBoxAlignment -> {"Columns" -> {{Left}}, "Rows" -> {{Top}}},
                AutoDelete -> False, 
               GridBoxItemSize -> {
                "Columns" -> {{Automatic}}, "Rows" -> {{Automatic}}}, 
               GridBoxSpacings -> {"Columns" -> {{1}}, "Rows" -> {{0}}}], 
              "Grid"], Alignment -> Left, AppearanceElements -> None, 
             ImageMargins -> {{5, 5}, {5, 5}}, ImageSizeAction -> 
             "ResizeToFit"], LineIndent -> 0, StripOnInput -> False], {
           FontFamily -> "Arial"}, Background -> Automatic, StripOnInput -> 
           False]& ), InterpretationFunction :> (RowBox[{"LineLegend", "[", 
            RowBox[{
              RowBox[{"{", 
                RowBox[{
                  InterpretationBox[
                   ButtonBox[
                    TooltipBox[
                    GraphicsBox[{{
                    GrayLevel[0], 
                    RectangleBox[{0, 0}]}, {
                    GrayLevel[0], 
                    RectangleBox[{1, -1}]}, {
                    RGBColor[0.368417, 0.506779, 0.709798], 
                    RectangleBox[{0, -1}, {2, 1}]}}, DefaultBaseStyle -> 
                    "ColorSwatchGraphics", AspectRatio -> 1, Frame -> True, 
                    FrameStyle -> 
                    RGBColor[
                    0.24561133333333335`, 0.3378526666666667, 
                    0.4731986666666667], FrameTicks -> None, PlotRangePadding -> 
                    None, ImageSize -> 
                    Dynamic[{
                    Automatic, 1.35 CurrentValue["FontCapHeight"]/
                    AbsoluteCurrentValue[Magnification]}]], 
                    StyleBox[
                    RowBox[{"RGBColor", "[", 
                    RowBox[{"0.368417`", ",", "0.506779`", ",", "0.709798`"}],
                     "]"}], NumberMarks -> False]], Appearance -> None, 
                    BaseStyle -> {}, BaselinePosition -> Baseline, 
                    DefaultBaseStyle -> {}, ButtonFunction :> 
                    With[{Typeset`box$ = EvaluationBox[]}, 
                    If[
                    Not[
                    AbsoluteCurrentValue["Deployed"]], 
                    SelectionMove[Typeset`box$, All, Expression]; 
                    FrontEnd`Private`$ColorSelectorInitialAlpha = 1; 
                    FrontEnd`Private`$ColorSelectorInitialColor = 
                    RGBColor[0.368417, 0.506779, 0.709798]; 
                    FrontEnd`Private`$ColorSelectorUseMakeBoxes = True; 
                    MathLink`CallFrontEnd[
                    FrontEnd`AttachCell[Typeset`box$, 
                    FrontEndResource["RGBColorValueSelector"], {
                    0, {Left, Bottom}}, {Left, Top}, 
                    "ClosingActions" -> {
                    "SelectionDeparture", "ParentChanged", 
                    "EvaluatorQuit"}]]]], BaseStyle -> Inherited, Evaluator -> 
                    Automatic, Method -> "Preemptive"], 
                   RGBColor[0.368417, 0.506779, 0.709798], Editable -> False, 
                   Selectable -> False], ",", 
                  InterpretationBox[
                   ButtonBox[
                    TooltipBox[
                    GraphicsBox[{{
                    GrayLevel[0], 
                    RectangleBox[{0, 0}]}, {
                    GrayLevel[0], 
                    RectangleBox[{1, -1}]}, {
                    RGBColor[0.880722, 0.611041, 0.142051], 
                    RectangleBox[{0, -1}, {2, 1}]}}, DefaultBaseStyle -> 
                    "ColorSwatchGraphics", AspectRatio -> 1, Frame -> True, 
                    FrameStyle -> 
                    RGBColor[
                    0.587148, 0.40736066666666665`, 0.09470066666666668], 
                    FrameTicks -> None, PlotRangePadding -> None, ImageSize -> 
                    Dynamic[{
                    Automatic, 1.35 CurrentValue["FontCapHeight"]/
                    AbsoluteCurrentValue[Magnification]}]], 
                    StyleBox[
                    RowBox[{"RGBColor", "[", 
                    RowBox[{"0.880722`", ",", "0.611041`", ",", "0.142051`"}],
                     "]"}], NumberMarks -> False]], Appearance -> None, 
                    BaseStyle -> {}, BaselinePosition -> Baseline, 
                    DefaultBaseStyle -> {}, ButtonFunction :> 
                    With[{Typeset`box$ = EvaluationBox[]}, 
                    If[
                    Not[
                    AbsoluteCurrentValue["Deployed"]], 
                    SelectionMove[Typeset`box$, All, Expression]; 
                    FrontEnd`Private`$ColorSelectorInitialAlpha = 1; 
                    FrontEnd`Private`$ColorSelectorInitialColor = 
                    RGBColor[0.880722, 0.611041, 0.142051]; 
                    FrontEnd`Private`$ColorSelectorUseMakeBoxes = True; 
                    MathLink`CallFrontEnd[
                    FrontEnd`AttachCell[Typeset`box$, 
                    FrontEndResource["RGBColorValueSelector"], {
                    0, {Left, Bottom}}, {Left, Top}, 
                    "ClosingActions" -> {
                    "SelectionDeparture", "ParentChanged", 
                    "EvaluatorQuit"}]]]], BaseStyle -> Inherited, Evaluator -> 
                    Automatic, Method -> "Preemptive"], 
                   RGBColor[0.880722, 0.611041, 0.142051], Editable -> False, 
                   Selectable -> False], ",", 
                  InterpretationBox[
                   ButtonBox[
                    TooltipBox[
                    GraphicsBox[{{
                    GrayLevel[0], 
                    RectangleBox[{0, 0}]}, {
                    GrayLevel[0], 
                    RectangleBox[{1, -1}]}, {
                    RGBColor[0.560181, 0.691569, 0.194885], 
                    RectangleBox[{0, -1}, {2, 1}]}}, DefaultBaseStyle -> 
                    "ColorSwatchGraphics", AspectRatio -> 1, Frame -> True, 
                    FrameStyle -> 
                    RGBColor[
                    0.37345400000000006`, 0.461046, 0.12992333333333334`], 
                    FrameTicks -> None, PlotRangePadding -> None, ImageSize -> 
                    Dynamic[{
                    Automatic, 1.35 CurrentValue["FontCapHeight"]/
                    AbsoluteCurrentValue[Magnification]}]], 
                    StyleBox[
                    RowBox[{"RGBColor", "[", 
                    RowBox[{"0.560181`", ",", "0.691569`", ",", "0.194885`"}],
                     "]"}], NumberMarks -> False]], Appearance -> None, 
                    BaseStyle -> {}, BaselinePosition -> Baseline, 
                    DefaultBaseStyle -> {}, ButtonFunction :> 
                    With[{Typeset`box$ = EvaluationBox[]}, 
                    If[
                    Not[
                    AbsoluteCurrentValue["Deployed"]], 
                    SelectionMove[Typeset`box$, All, Expression]; 
                    FrontEnd`Private`$ColorSelectorInitialAlpha = 1; 
                    FrontEnd`Private`$ColorSelectorInitialColor = 
                    RGBColor[0.560181, 0.691569, 0.194885]; 
                    FrontEnd`Private`$ColorSelectorUseMakeBoxes = True; 
                    MathLink`CallFrontEnd[
                    FrontEnd`AttachCell[Typeset`box$, 
                    FrontEndResource["RGBColorValueSelector"], {
                    0, {Left, Bottom}}, {Left, Top}, 
                    "ClosingActions" -> {
                    "SelectionDeparture", "ParentChanged", 
                    "EvaluatorQuit"}]]]], BaseStyle -> Inherited, Evaluator -> 
                    Automatic, Method -> "Preemptive"], 
                   RGBColor[0.560181, 0.691569, 0.194885], Editable -> False, 
                   Selectable -> False]}], "}"}], ",", 
              RowBox[{"{", 
                RowBox[{#, ",", #2, ",", #3}], "}"}]}], "]"}]& ), Editable -> 
         True]},
       "Legended",
       DisplayFunction->(GridBox[{{
           TagBox[
            ItemBox[
             PaneBox[
              TagBox[#, "SkipImageSizeLevel"], 
              Alignment -> {Center, Baseline}, BaselinePosition -> Baseline], 
             DefaultBaseStyle -> "Labeled"], "SkipImageSizeLevel"]}, {
           ItemBox[#2, DefaultBaseStyle -> "LabeledLabel"]}}, 
         GridBoxAlignment -> {"Columns" -> {{Center}}, "Rows" -> {{Center}}}, 
         AutoDelete -> False, GridBoxItemSize -> Automatic, 
         BaselinePosition -> {1, 1}]& ),
       Editable->True,
       InterpretationFunction->(RowBox[{"Legended", "[", 
          RowBox[{#, ",", 
            RowBox[{"Placed", "[", 
              RowBox[{#2, ",", "Below"}], "]"}]}], "]"}]& )]},
     {
      TemplateBox[{OverlayBox[{
          GraphicsBox[
           RasterBox[CompressedData["
1:eJzt13s0VXkbwPEjpbkwyKVSNG5nxGTKpSnJw5QSMnFUHG8q5UxEGjXuTaI0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            "], {{0, 0}, {21, 21}}, {0, 1}], Frame -> True, 
           FrameLabel -> {None, None}, FrameStyle -> Directive[
             GrayLevel[1]], FrameTicks -> {{{{0.5, 
                FormBox["0", TraditionalForm]}, {4.5, 
                FormBox["0.2`", TraditionalForm]}, {8.5, 
                FormBox["0.4`", TraditionalForm]}, {12.5, 
                FormBox["0.6`", TraditionalForm]}, {16.5, 
                FormBox["0.8`", TraditionalForm]}, {20.5, 
                FormBox["1", TraditionalForm]}}, None}, {{{0.5, 
                FormBox["0", TraditionalForm]}, {4.5, 
                FormBox["0.2`", TraditionalForm]}, {8.5, 
                FormBox["0.4`", TraditionalForm]}, {12.5, 
                FormBox["0.6`", TraditionalForm]}, {16.5, 
                FormBox["0.8`", TraditionalForm]}, {20.5, 
                FormBox["1", TraditionalForm]}}, None}}, FrameTicksStyle -> 
           Directive[
             GrayLevel[1], FontSize -> 12], GridLinesStyle -> Directive[
             GrayLevel[0.5, 0.4]], 
           Method -> {
            "AxisPadding" -> Scaled[0.02], "DefaultBoundaryStyle" -> 
             Automatic, 
             "DefaultGraphicsInteraction" -> {
              "Version" -> 1.2, "TrackMousePosition" -> {True, False}, 
               "Effects" -> {
                "Highlight" -> {"ratio" -> 2}, 
                 "HighlightPoint" -> {"ratio" -> 2}, 
                 "Droplines" -> {
                  "freeformCursorMode" -> True, 
                   "placement" -> {"x" -> "All", "y" -> "None"}}}}, 
             "DefaultPlotStyle" -> Automatic, "DomainPadding" -> Scaled[0.02],
              "RangePadding" -> Scaled[0.05]}], 
          GraphicsBox[{
            GraphicsComplexBox[CompressedData["
1:eJzlmnl0zPf6x0eo2kZ/1LVMItq6YmjsEku1T2KI9l5iJJYrdGqp/DD2uhhL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             "], {{{
                EdgeForm[], 
                Directive[
                 Opacity[0]], 
                GraphicsGroupBox[{
                  PolygonBox[CompressedData["
1:eJxNmHn8V1Max+8pSkkpVCIpJJKUSosoSaVFq2gPIaTs+26KjApZspMlJmMJ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                   "]], 
                  PolygonBox[CompressedData["
1:eJwt08tLVVEUx/Gzo5l/gWTi1dQyLUNJDC9KoihBoIJIIKFQaJGm5ruHZqHV
7eGjgjTn4SMte6ulKXYdKImYNGnc3+DE72L/Bh8W556z9157rXVDtY3lDQeC
IMjFQdS5IPiNCvzj+SbxFTIRwgCWUYJSvEQEh3EC7XiIVOSgD39QjSPowhzy
EMa01ti3pzGKMRxHIt5iRu+myWeC2IZsZGEQf1GDBHTiCY7ijHLbxUXleANv
kIxDuIRNlGGe/R8Tm3BSe0xiSmeN836EeB/FyMA17OACYrCN28q9ClvO19Bq
V6lve5CkNVabu4hVjSzXR8rFcv6FNUR5Pme/45vOOosh3EIB0lCvM+2s/6xZ
If5U7axnH/FBvc1T7Z7pW6vhEn7orkXEYdxBoXryBZ+xx/t84j0sqhZWU5ud
bvXCZiiKdedzj8cGWlX786rFa5xSTewun5zfu0C9uIxj6onl8lV3t5xe4Dni
kIJanWF7RwI/K/16ZzOzgF7nZzFda1edn2XbY9b5GbLZydVsdCh3m5Gnzs9w
WDk9wHXnZ99m/h3ea631rJl4Rb2x/1ALGnS23fmq8zNTpZrabH13vvY2Y/uX
hmnu
                   "]]}]}, {
                EdgeForm[], 
                Directive[
                 Opacity[0]], 
                GraphicsGroupBox[{
                  PolygonBox[CompressedData["
1:eJxNmHmYz1UUxmfOrVCRJWPfyVJkbI09MlqUGUI0USSppFWhon3TnlZt2je0
L9oXlYTqqedRUdEmBlNSDY06b+fzfcYf73O+dzv3Peee+97fTIsJZwyfajk5
Oatyc3KS20+8scxxtGMf75jumObfW3z8O8f3jsu9vdrtRscmxxrHt47S3OhX
u5rP2e52h6OMdWsdvzvWOX5wbMX+6PgD+5NjG/Znx5/YXxx/Ydc7NrOnOP2G
b/n9m/Ffc2P/jfCqYI3iyLXgof3/IQaN5Vj40thO5oq7WcwVx2RhxfEjYqxq
EWcpvnazGBf33S2suO9hYcVxvH+Xu63i9gQLnuXwqIDXnvjV994W/eJV3cL+
66htwUl71rDoUxw1LeIU930s+tTey8KffNVhnTheaRFHnqMe3zrD+lhxaYCV
j4ZY8WqEFa9a7ClOjekTryZYcWmKFccWFjGIS3PWq92Svn0dbXbhdQ3frR37
0VefuMuohbb0iW9H9msGarJHe2IQ9w5Y8d0fK74HYLW+Hf4awkk503l2wq+4
dyYece8Kb3E8iP20zwr6uzh64Fdj3Zgr7gXMFZeeWHF5x/G840THwfjQPr0Y
V6y9seLVB3ug40X49XV0Zx/tfbZjkuNQx3P47O941HGnY7BjEFwVx0OOuexz
CL56sL4nXAqZqzjOckyEyxx8H+Eoxq/mDiR++RrCHMV3JHaA4yis5hax9yD6
urL+ZHwPt9Aw+ZKOreNOqzaup2+MY4JFHeU7RjBfe45iv6GOkfSpfSxrNfcJ
xzzmXuqY4hjnKGFc605ln7GMjcHHeY7TmPu0Y7ZjuuMtx2LiO9cxmTxdYaG5
G9DXUnQt082t6GtZbqVOb0FL1hJ32S6a+/suOqv1qg/Vr2o309xt7LMDHV2D
tpWiu5tpZ7r8J7r7N+2NrCtn3j/4Ot/3uIx8bEHPKuAojfoXftIqtTPd1H3b
ihZrbBs6m+mxrMakGS2JRfosfS1nb2mXtExcpF1qV6Ctamfaujc8pFfV4SG9
UjvTdO2t+OSvKpxqMk9cpGHSmasstLWuVepmA+LZF76ZttbHlzQtzyq1tSE8
GtOuTXzN4VGHvkxnG1ulzjaBVzPatVjXjHzVYryM90Q5uNZCZ1vBRbrUBn5t
aWf6uD/c2zGW6Wmmr+0Zy8NHa3h0ZL24qP46EVNn2tLJfOZrne54F/buRlv7
daedaXcH9uvBmPY/iLbGCmhnWplpp2xPq9RNjUlrpYn9LPSzD9zEQ7okncre
yHpwGsCYfEkTB1uljhfASdo1EE6DaItTIe0C1hUS20D8zsWnuD5socXS1uss
tPVwq9RKaZb0uxg/mZ4OYQ/1F+F7KGses9B9vQfST+mo9PRtC22SHkn/pGFj
8DWCPvkeSVu+RtFezPdQ4izi+wYLLR7NmhL8vup4xULLtYd0Qvr5pIXmnmKh
neOYv8ix0OI31MUWOiotnkqf7v45+FM8NzE2nlxOItfz4CDOzzqesXgfzsTm
c/7Kt2rjJcdJ9J2Bvyr4KuC8i8mZYr7EMcNxPHYK3/3Jo/L8puMNi3ejhDgV
4zT6lIf7WX86e04lxtst3nLFNN9xEWM3w0/cHnTMIqZ3sPnEcRaxXGih0VmO
p7DvHRZ5VA7vtnifirEXEOcU8j+W9efjYwFnM4N9ZpLH89hPc1/jbEfB+yLO
bwa5U7yzyYX838v6EuKYxTlNZ95Y1l4ML+FScn4Z3+KmWtQdUG2+53jX4q1d
bvG30PWciazqcYmFPkrPP7S4e9IH1ewczlH9H9BWHUlLpWOfWmhKPn7eZ0y5
0F3uTR50Z46xqNX55EOc74P3xxbvShU46/fBN7lx3+/iXMTxFngqp6rvkdh7
yNs9tI+D1wsWvy3l+2b832hxV0ez71KL92w5/aqHLY7BKeroatbXJT7NzcP3
8+RKeVhpoa0fYHUPHrfQIN1J1cITcJ9Lfg7DPsz3o8wf5niEfuVSdSNN0G8Q
3ZcF1ILePOl7R85Yc/Sbah57KQ+ql6cs6kh9T9KvXN1L3may30S4isdgakD6
JR0+mbbqQW+H9EL3qzPtA+nTt36T34kfxTKHtfLzMvP6wncRnF9iTDlXDanO
WmGf5XsCbdXeXewxDK76LuQM7+AcVeOqeem59F66rzqU9ugOqP4nM1ZEnzRL
tbGY/mM4Z+lLPtzepwayu6T5mQap3nTnVMeq4SuZr/q5gn6dlep5yS7+ljBH
taa61DrVku6eakx1qt9/+j21nb/VpZPXMWcO0Jr+7CEeq5mr9ap/afMyznEZ
+RnOfkvxvxQOmp/VtDjoLr9uUcu6H924I/KxHD/6bbOCM5rk4xNS1Nxi1qv+
tW4zZ72S+foNon/YlJLDFYzlc9bSaOmA7s5tFu9Cb9q9OD/tIf1TTm6Fs+be
zvzbmK+7tskxKMW859irC2f9KX2dfLxjinM73O2hKXwc6faIFLzOdVs7xdtQ
5PaoFHdrJX7Ef6D3bSSuHv7dPQXHXm7LLd6C18mP6m2DY0AKvevqtkuKermB
PKvuRnjf0Sn0oJbbGSnej0K3uSnepb3ctkzh83i3zVPclzFuR6d4P0902zqF
Jox3u3sK3Rrn9oQUeqQznMg5futolaLmS9yOTaFTx7ltmiK3VcUlxfu60zEt
xZv9Hfn6/z12e2qKvxe/0pmwtoPbA1K8a5PdnpJCvxTHGvKzkPqWHhYrphRv
yZeORim0or3bdine2s8d9VPoTAO3DVO8x5851ln0/+w2L8Xb/4ejXoo6/8Yx
hLwpjgqLWFY5frLQtN28P6X4vfK9apB8fuH4wYLPdsfIFGf9o6NOCt3d0221
FL87GrttkuLt/9rRLIV/5bIK+ZRtwbfOYbXFWfzm2COFTotLXfgcm2I/7aXz
VzyKRXXdh9ru6bYgxduY77ZzinevlNpV3apOh8NfufyMvCnfXxCjcnYSedP/
WWvwv1bVZk3qUzXyFXdQuTyY2tCZt+XcZddafI/y7x1uH7Dw1wafymUv+ChP
qzgL1bLOQPlf7+ib4nflNkf1FNrf322/FL9H/gP4VXsD
                   "]], 
                  PolygonBox[CompressedData["
1:eJwtlOmLj1EUx5/fPUPW7Ip4MYksZc0WYwkx1pkh+zIzDGOY+Y2xjOxLIska
IgwisiQ1JSX5G8goa8g2sq9J+Hyd++LT+Z77PM957j3LzSxM55WFJEnyIAPq
W5K8ZGEtrIA8/ClQhS6FReglcAm9Gh7CQPw92OnQGt0IimOMpuhX2HWwCtri
/8bmw3p4BBWs3cBOE+g25s9OQRZ6KOxFF0BLdAGcRy+ExujlcBNdAp3Rv+Je
N2tP+O1gA/o0tEf/xRYGX/sBzcz3dhxamf9DsSvgM9TTP7CH4Sl0MY89L/hZ
SvGvxDP1RPeA7egdsBTdCTahL8N3aIK/ErsxeG6eQGXMUYn5O3qmHGuvZaxd
jXuui14G11UL+Ak5+CewW+ELNDCv3VF4Af3wd2GzYTB6EDxIJck2/GfQ3VzX
pnzvb9D7gp9B706Fk8G/GY6+ozzBfijH/4AtgmuQxk9gProahpjXVLVUDZ9D
H/yd2A7QG10bY2ntXoypWFnQF/0OeyD4Gcaan1FnOwLj8Meb61yYiJ4Ex9Cz
ob95DVW73TGXd+O/ldP3MEo5jWeYYN4T6gXF+AgZ5r1wCEabr0kfhDH42RY1
jETX6Nxxz5PNe0C1r4q9cjvmWj2jWt6PtVFNc/G7wZbgNa2JOa2OMZX7Xua9
pRoE9Cw4gz4LI9AzzHu9HOqg58K54DOj2ayEW8FndBj6dfDZSgev/Z+o1QMd
ef4t+GypZzPxP8Xe00wMMM+xcqs7QHvvaj4bOkORZgkuoBcH/3dD81nXHjS7
j4PfHZph3QVf47u6EzQ7xXAx+AwtUG+Z1/Z/TPM7SXeR3kmhZ8Ka4DXN117M
e0M5aG4+M5oVzVwL8xwqd/pmjmKb11I5fRtzqlyqp/8BHFSjSA==
                   "]]}]}, {}, {}, {}, {}, {}}, {{
                Directive[
                 RGBColor[0.368417, 0.506779, 0.709798], 
                 AbsoluteThickness[1.6]], 
                LineBox[CompressedData["
1:eJwt0kk3FQAYx+GLikLoXtLSju9Q3wEfwdCcW0gUGYqSNM8TRZSSFBVlSlSX
SGhaWbdta+G551g857yr9/wXv6yCcH5JTCAQ6CWW6J1ODrvIo5gqWmmjjxEi
LLPCP/6zygZPkggSIp0MtpPJDhr4xSme0E09I0QHlXGfOn5yiMd84z0fGKaW
qyxzkHvc5SRLHKCTWd4xxFtquMIi+ynlDtX8YB8dXOIvl7nNCRbYyyNmGOQN
AxznIt/ZzRFa+cMFqujnNa/Yw0M+84VpInzlFpXMU0w7LfzmPDc5TDPnOMYL
XtJHEW1M8omPnGWK69ygiTNU8IxenlPIA8aZYIxGrhFmlNM8pZwuejjKHAXR
jVqopIhcdpJNiCDbSCOVFLaSTBKJbGEzCcSziY3Rvohb73UNscFaxQ==
                 "]]}, {
                Directive[
                 RGBColor[0.880722, 0.611041, 0.142051], 
                 AbsoluteThickness[1.6]], 
                LineBox[CompressedData["
1:eJwl0mdbjmEAx+H7qZQVRZkZFTLTEBkpNIWGTYooIpUR+ja+Ct/ASrR3IYSI
rNPhxXn8/td1PS+eF3dibXNlUygIgkc8DguCToeHOsOp8CDYyz62kcqA+ycs
s5/qTf2pK3Q5p9lPLjmMecvTX9qqv3Wc7XYa6TxzPqBvNEMzGbSfs9J+oU36
R1dpAmc4yyHeuj+o7zTQZtrtfJ3QLN3JsN3BGvu13tJIJu1EjeIes/nkbpdm
s5su52Tt0TsazZS9XhfwgIV8dXddb3CVBq6xgUaOc4JKvvvtHj3JD3ujbqLf
TtGtbGYLFZQz7W2R3qXXHiXGbiOWb85HtYxj9Dmv08tcoZZ66ljKF++3tVuT
dB73mc8SRtzP0bl8tg/rEUrpdF6rNVykmllcIo54wongo98WaQnFvHJerVVc
4DxhLGbI2zkNUcB75x1ayAe7RV9qDmn//jdxRDLt+x1njFFGGGaIQQbop49e
euimK/T/+/8Ln2Nc0Q==
                 "]]}}}], {}}, {
           DisplayFunction -> Identity, PlotRangePadding -> {{
               Scaled[0.02], 
               Scaled[0.02]}, {
               Scaled[0.02], 
               Scaled[0.02]}}, PlotRangeClipping -> True, ImagePadding -> All,
             DisplayFunction -> Identity, AspectRatio -> 1, 
            Axes -> {False, False}, AxesLabel -> {None, None}, 
            AxesOrigin -> {Automatic, Automatic}, DisplayFunction :> Identity,
             Epilog -> {
              InsetBox[
               FormBox[
                StyleBox["\"drive invades and fixes\"", FontSize -> 14, 
                 GrayLevel[0], StripOnInput -> False], TraditionalForm], {
               0.027, 0.66}, {-1, 0}], 
              InsetBox[
               FormBox[
                StyleBox["\"drive doesn't invade\"", FontSize -> 15, 
                 GrayLevel[0], StripOnInput -> False], TraditionalForm], {
               0.55, 0.75}, {-1, 0}], 
              InsetBox[
               FormBox[
                StyleBox["\"polymorphism\"", FontSize -> 15, 
                 GrayLevel[0], StripOnInput -> False], TraditionalForm], {0.5,
                0.25}, {-1, 0}]}, Frame -> {{True, True}, {True, True}}, 
            FrameLabel -> {{None, None}, {None, None}}, 
            FrameTicks -> {{Automatic, Automatic}, {Automatic, Automatic}}, 
            FrameTicksStyle -> Directive[FontSize -> 12], GridLinesStyle -> 
            Directive[
              GrayLevel[0.5, 0.4]], 
            Method -> {
             "DefaultGraphicsInteraction" -> {
               "Version" -> 1.2, "TrackMousePosition" -> {True, False}, 
                "Effects" -> {
                 "Highlight" -> {"ratio" -> 2}, 
                  "HighlightPoint" -> {"ratio" -> 2}, 
                  "Droplines" -> {
                   "freeformCursorMode" -> True, 
                    "placement" -> {"x" -> "All", "y" -> "None"}}}}, 
              "ScalingFunctions" -> None, "TransparentPolygonMesh" -> True, 
              "AxesInFront" -> True}, PlotRange -> {{0, 1}, {0, 1}}, 
            PlotRangeClipping -> True, PlotRangePadding -> {
              Scaled[0.02], 
              Scaled[0.02]}, Ticks -> {Automatic, Automatic}}]}],
        InterpretationBox[
         Cell[
          BoxData[
           FormBox[
            StyleBox[
            "\"female selection (\\!\\(\\*StyleBox[SubscriptBox[\\\"s\\\", \\\
\"f\\\"],FontSlant->\\\"Italic\\\"]\\))\"", StripOnInput -> False], 
            TextForm]], "InlineText"], 
         Text[
          Style[
          "female selection (\!\(\*StyleBox[SubscriptBox[\"s\", \
\"f\"],FontSlant->\"Italic\"]\))"]]],RotationBox[
         InterpretationBox[
          Cell[
           BoxData[
            FormBox[
             StyleBox[
             "\"dominance \
(\\!\\(\\*StyleBox[\\\"h\\\",FontSlant->\\\"Italic\\\"]\\))\"", StripOnInput -> 
              False], TextForm]], "InlineText"], 
          Text[
           Style[
           "dominance (\!\(\*StyleBox[\"h\",FontSlant->\"Italic\"]\))"]]], 
         BoxRotation -> 1.5707963267948966`],InterpretationBox[
         Cell[
          BoxData[
           FormBox[
            StyleBox["\"B\"", Bold, FontSize -> 20, StripOnInput -> False], 
            TextForm]], "InlineText"], 
         Text[
          Style["B", Bold, FontSize -> 20]]]},
       "Labeled",
       DisplayFunction->(GridBox[{{"", 
           
           ItemBox[#4, Alignment -> {Left, Inherited}, DefaultBaseStyle -> 
            "LabeledLabel"]}, {
           ItemBox[#3, DefaultBaseStyle -> "LabeledLabel"], 
           TagBox[
            ItemBox[
             PaneBox[
              TagBox[#, "SkipImageSizeLevel"], 
              Alignment -> {Center, Baseline}, BaselinePosition -> Baseline], 
             DefaultBaseStyle -> "Labeled"], "SkipImageSizeLevel"]}, {"", 
           ItemBox[#2, DefaultBaseStyle -> "LabeledLabel"]}}, 
         GridBoxAlignment -> {"Columns" -> {{Center}}, "Rows" -> {{Center}}}, 
         AutoDelete -> False, 
         GridBoxItemSize -> {
          "Columns" -> {{Automatic}}, "Rows" -> {{Automatic}}}, 
         BaselinePosition -> {2, 2}]& ),
       InterpretationFunction->(RowBox[{"Labeled", "[", 
          RowBox[{#, ",", 
            RowBox[{"{", 
              RowBox[{#2, ",", #3, ",", #4}], "}"}], ",", 
            RowBox[{"{", 
              RowBox[{"Bottom", ",", "Left", ",", 
                RowBox[{"{", 
                  RowBox[{"Top", ",", "Left"}], "}"}]}], "}"}]}], "]"}]& )]},
     {""},
     {
      TemplateBox[{StyleBox[
         StyleBox[
          PaneBox[
           TagBox[
            GridBox[{{
               StyleBox[
                TagBox[
                 FormBox[
                 "\"    increase in population size relative to WT\"", 
                  TraditionalForm], TraditionalForm, Editable -> True], {
                FontFamily -> "Arial"}, Background -> Automatic, StripOnInput -> 
                False]}, {
               GraphicsBox[{
                 RasterBox[CompressedData["
1:eJxdmAusXFUVhisa0hBCSEVRUCyotdgYLGh8gb+IiIqVlzGEENPcFNI0CBUr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                  "], {{-110, 
                    Rational[-15, 2]}, {110, 
                    Rational[15, 2]}}], {Antialiasing -> False, 
                  AbsoluteThickness[0.1], 
                  Directive[
                   Opacity[0.3], 
                   GrayLevel[0]], 
                  LineBox[
                   NCache[{{110, 
                    Rational[-15, 2]}, {-110, 
                    Rational[-15, 2]}, {-110, 
                    Rational[15, 2]}, {110, 
                    Rational[15, 2]}, {110, 
                    Rational[-15, 2]}}, {{
                    110, -7.5}, {-110, -7.5}, {-110, 7.5}, {110, 7.5}, {
                    110, -7.5}}]]}, {
                  CapForm[None], {}}, 
                 StyleBox[{Antialiasing -> False, 
                   StyleBox[
                    LineBox[{{-110., -7.5}, {110., -7.5}}], 
                    Directive[
                    AbsoluteThickness[0.2], 
                    Opacity[0.3], 
                    GrayLevel[0]], StripOnInput -> False], 
                   StyleBox[
                    StyleBox[{{
                    StyleBox[
                    LineBox[{{{-110., -7.5}, 
                    Offset[{0, -4.}, {-110., -7.5}]}, {{-66., -7.5}, 
                    Offset[{0, -4.}, {-66., -7.5}]}, {{-22., -7.5}, 
                    Offset[{0, -4.}, {-22., -7.5}]}, {{22., -7.5}, 
                    Offset[{0, -4.}, {22., -7.5}]}, {{66., -7.5}, 
                    Offset[{0, -4.}, {66., -7.5}]}, {{110., -7.5}, 
                    Offset[{0, -4.}, {110., -7.5}]}}], 
                    Directive[
                    AbsoluteThickness[0.2], 
                    GrayLevel[0.4]], StripOnInput -> False], 
                    StyleBox[
                    LineBox[{{{-101.2, -7.5}, 
                    Offset[{0., -2.5}, {-101.2, -7.5}]}, {{-92.4, -7.5}, 
                    
                    Offset[{
                    0., -2.5}, {-92.4, -7.5}]}, {{-83.60000000000001, -7.5}, 
                    
                    Offset[{
                    0., -2.5}, {-83.60000000000001, -7.5}]}, {{-74.8, -7.5}, 
                    Offset[{0., -2.5}, {-74.8, -7.5}]}, {{-57.2, -7.5}, 
                    Offset[{0., -2.5}, {-57.2, -7.5}]}, {{-48.4, -7.5}, 
                    Offset[{0., -2.5}, {-48.4, -7.5}]}, {{-39.6, -7.5}, 
                    Offset[{0., -2.5}, {-39.6, -7.5}]}, {{-30.8, -7.5}, 
                    
                    Offset[{
                    0., -2.5}, {-30.8, -7.5}]}, {{-13.200000000000001`, -7.5}, 
                    
                    Offset[{
                    0., -2.5}, {-13.200000000000001`, -7.5}]}, {{-4.4, -7.5}, 
                    Offset[{0., -2.5}, {-4.4, -7.5}]}, {{4.4, -7.5}, 
                    Offset[{0., -2.5}, {4.4, -7.5}]}, {{
                    13.200000000000001`, -7.5}, 
                    Offset[{0., -2.5}, {13.200000000000001`, -7.5}]}, {{
                    30.8, -7.5}, 
                    Offset[{0., -2.5}, {30.8, -7.5}]}, {{39.6, -7.5}, 
                    Offset[{0., -2.5}, {39.6, -7.5}]}, {{48.4, -7.5}, 
                    Offset[{0., -2.5}, {48.4, -7.5}]}, {{57.2, -7.5}, 
                    Offset[{0., -2.5}, {57.2, -7.5}]}, {{74.8, -7.5}, 
                    Offset[{0., -2.5}, {74.8, -7.5}]}, {{
                    83.60000000000001, -7.5}, 
                    Offset[{0., -2.5}, {83.60000000000001, -7.5}]}, {{
                    92.4, -7.5}, 
                    Offset[{0., -2.5}, {92.4, -7.5}]}, {{101.2, -7.5}, 
                    Offset[{0., -2.5}, {101.2, -7.5}]}}], 
                    Directive[
                    AbsoluteThickness[0.2], 
                    GrayLevel[0.4], 
                    Opacity[0.3]], StripOnInput -> False]}, 
                    StyleBox[
                    StyleBox[{{
                    StyleBox[{
                    InsetBox[
                    FormBox["0", TraditionalForm], 
                    Offset[{0., -7.}, {-110., -7.5}], {0., 1.}, Automatic, {1,
                     0}], 
                    InsetBox[
                    FormBox["50", TraditionalForm], 
                    Offset[{0., -7.}, {-66., -7.5}], {0., 1.}, Automatic, {1, 
                    0}], 
                    InsetBox[
                    FormBox["100", TraditionalForm], 
                    Offset[{0., -7.}, {-22., -7.5}], {0., 1.}, Automatic, {1, 
                    0}], 
                    InsetBox[
                    FormBox["150", TraditionalForm], 
                    Offset[{0., -7.}, {22., -7.5}], {0., 1.}, Automatic, {1, 
                    0}], 
                    InsetBox[
                    FormBox["200", TraditionalForm], 
                    Offset[{0., -7.}, {66., -7.5}], {0., 1.}, Automatic, {1, 
                    0}], 
                    InsetBox[
                    FormBox["250", TraditionalForm], 
                    Offset[{0., -7.}, {110., -7.5}], {0., 1.}, Automatic, {1, 
                    0}]}, 
                    Directive[
                    AbsoluteThickness[0.2], 
                    GrayLevel[0.4]], {
                    Directive[
                    Opacity[1], 
                    GrayLevel[0.3]], 
                    Directive[
                    Opacity[1], 
                    GrayLevel[0.3]]}, StripOnInput -> False], 
                    
                    StyleBox[{{}, {}, {}, {}, {}, {}, {}, {}, {}, {}, {}, {}, \
{}, {}, {}, {}, {}, {}, {}, {}}, 
                    Directive[
                    AbsoluteThickness[0.2], 
                    GrayLevel[0.4], 
                    Opacity[0.3]], {
                    Directive[
                    Opacity[1], 
                    GrayLevel[0.3]], 
                    Directive[
                    Opacity[1], 
                    GrayLevel[0.3]]}, StripOnInput -> False]}, {}}, {
                    Directive[
                    Opacity[1], 
                    GrayLevel[0.3]], 
                    Directive[
                    Opacity[1], 
                    GrayLevel[0.3]]}, StripOnInput -> False], "GraphicsLabel",
                     StripOnInput -> False]}, "GraphicsTicks", StripOnInput -> 
                    False], {
                    Directive[
                    AbsoluteThickness[0.2], 
                    GrayLevel[0.4]], 
                    Directive[
                    AbsoluteThickness[0.2], 
                    GrayLevel[0.4], 
                    Opacity[0.3]]}, StripOnInput -> False]}, "GraphicsAxes", 
                  StripOnInput -> False]}, PlotRangePadding -> Scaled[0.02], 
                PlotRange -> All, Frame -> True, 
                FrameTicks -> {{True, False}, {False, False}}, FrameStyle -> 
                Opacity[0], FrameTicksStyle -> Opacity[0], 
                ImageSize -> {220, Automatic}, BaseStyle -> {}]}}, 
             GridBoxAlignment -> {"Columns" -> {{Center}}}, AutoDelete -> 
             False, GridBoxItemSize -> {
              "Columns" -> {{Automatic}}, "Rows" -> {{Automatic}}}, 
             GridBoxSpacings -> {
              "Columns" -> {{Automatic}}, "Rows" -> {{Automatic}}}], "Grid"], 
           Alignment -> Left, AppearanceElements -> None, 
           ImageMargins -> {{25, 0}, {10, 10}}, ImageSizeAction -> 
           "ResizeToFit"], LineIndent -> 0, StripOnInput -> False], {
         FontFamily -> "Arial"}, Background -> Automatic, StripOnInput -> 
         False]},
       "BarLegend",
       DisplayFunction->(#& ),
       InterpretationFunction:>(RowBox[{"BarLegend", "[", 
          RowBox[{
            RowBox[{"{", 
              RowBox[{
                RowBox[{
                  RowBox[{"Blend", "[", 
                    RowBox[{
                    RowBox[{"{", 
                    RowBox[{
                    RowBox[{"{", 
                    RowBox[{"0", ",", "White"}], "}"}], ",", 
                    RowBox[{"{", 
                    RowBox[{"250", ",", 
                    RowBox[{"Darker", "[", 
                    RowBox[{"Green", ",", "0.6`"}], "]"}]}], "}"}]}], "}"}], 
                    ",", "#1"}], "]"}], "&"}], ",", 
                RowBox[{"{", 
                  RowBox[{"0", ",", "250"}], "}"}]}], "}"}], ",", 
            RowBox[{"LegendLayout", "\[Rule]", "\"Row\""}], ",", 
            
            RowBox[{"LegendLabel", "\[Rule]", 
              "\"    increase in population size relative to WT\""}], ",", 
            RowBox[{"LegendMargins", "\[Rule]", 
              RowBox[{"{", 
                RowBox[{
                  RowBox[{"{", 
                    RowBox[{"25", ",", "0"}], "}"}], ",", 
                  RowBox[{"{", 
                    RowBox[{"10", ",", "10"}], "}"}]}], "}"}]}]}], "]"}]& )]}
    },
    AutoDelete->False,
    GridBoxItemSize->{"Columns" -> {{Automatic}}, "Rows" -> {{Automatic}}}],
   "Grid"],
  StripOnInput->False,
  ImageSizeMultipliers->{1, 1}]], "Output",
 CellChangeTimes->{
  3.809762686689597*^9, 3.809762772918667*^9, {3.809762914784112*^9, 
   3.809762941930785*^9}, 3.809766445575305*^9, 3.814005340795231*^9, 
   3.814196574830812*^9, 3.814197131359394*^9},
 CellLabel->
  "Out[1547]=",ExpressionUUID->"6325db03-fb28-45a9-98c4-8a3455f04c28"]
}, Open  ]]
}, Closed]],

Cell[CellGroupData[{

Cell["Population stacked line graphs", "Subsubsection",
 CellChangeTimes->{{3.795271840235935*^9, 
  3.795271851857108*^9}},ExpressionUUID->"21a6cac6-75f5-4e55-822c-\
5284ed2cd5f9"],

Cell[BoxData[
 RowBox[{
  RowBox[{
  "data", "=", 
   "\"\<[89, 0, 0, 64, 0]\n[92, 0, 0, 100, 0]\n[96, 0, 0, 80, 0]\n[85, 0, 0, \
89, 0]\n[74, 0, 0, 101, 0]\n[74, 0, 0, 75, 0]\n[74, 0, 0, 80, 0]\n[67, 0, 0, \
79, 0]\n[74, 0, 0, 63, 0]\n[73, 0, 0, 75, 0]\n[82, 0, 0, 66, 0]\n[74, 0, 0, \
89, 0]\n[68, 0, 0, 83, 0]\n[76, 0, 0, 61, 0]\n[67, 0, 0, 81, 0]\n[65, 0, 0, \
66, 0]\n[76, 0, 0, 57, 0]\n[88, 0, 0, 61, 0]\n[85, 0, 0, 90, 0]\n[71, 0, 0, \
96, 0]\n[69, 0, 0, 76, 0]\n[64, 0, 0, 63, 0]\n[67, 0, 0, 57, 0]\n[79, 0, 0, \
58, 0]\n[81, 0, 0, 91, 0]\n[77, 0, 0, 73, 0]\n[75, 0, 0, 79, 0]\n[80, 0, 0, \
76, 0]\n[87, 0, 0, 80, 0]\n[102, 0, 0, 69, 0]\n[114, 0, 0, 97, 0]\n[103, 0, \
0, 113, 0]\n[119, 0, 0, 92, 0]\n[117, 0, 0, 120, 0]\n[95, 0, 0, 120, 0]\n[83, \
0, 0, 97, 0]\n[70, 0, 0, 95, 0]\n[83, 0, 0, 58, 0]\n[75, 0, 0, 94, 0]\n[69, \
0, 0, 78, 0]\n[71, 0, 0, 74, 0]\n[62, 0, 0, 83, 0]\n[62, 0, 0, 62, 0]\n[74, \
0, 0, 49, 0]\n[83, 0, 0, 72, 0]\n[86, 0, 0, 88, 0]\n[83, 0, 0, 87, 0]\n[81, \
0, 0, 81, 0]\n[93, 0, 0, 71, 0]\n[84, 0, 0, 101, 0]\n[73, 0, 0, 83, 0]\n[88, \
0, 0, 63, 0]\n[85, 0, 0, 82, 0]\n[83, 0, 0, 83, 0]\n[84, 0, 0, 82, 0]\n[72, \
0, 0, 88, 0]\n[74, 0, 0, 76, 0]\n[72, 0, 0, 76, 0]\n[74, 0, 0, 75, 0]\n[58, \
0, 0, 81, 0]\n[67, 0, 0, 55, 0]\n[79, 0, 0, 58, 0]\n[83, 0, 0, 75, 0]\n[84, \
0, 0, 85, 0]\n[84, 0, 0, 87, 0]\n[80, 0, 0, 97, 0]\n[82, 0, 0, 82, 0]\n[84, \
0, 0, 83, 0]\n[83, 0, 0, 83, 0]\n[93, 0, 0, 79, 0]\n[94, 0, 0, 101, 0]\n[105, \
0, 0, 78, 0]\n[96, 0, 0, 107, 0]\n[107, 0, 0, 79, 0]\n[108, 0, 0, 105, 0]\n\
[96, 0, 0, 109, 0]\n[79, 0, 0, 109, 0]\n[80, 0, 0, 75, 0]\n[85, 0, 0, 75, 0]\n\
[101, 0, 0, 81, 0]\n[95, 0, 0, 108, 0]\n[106, 0, 0, 73, 0]\n[122, 0, 0, 95, \
0]\n[107, 0, 0, 117, 0]\n[96, 0, 0, 113, 0]\n[106, 0, 0, 90, 0]\n[106, 0, 0, \
104, 0]\n[102, 0, 0, 82, 0]\n[101, 0, 0, 95, 0]\n[86, 0, 0, 97, 0]\n[87, 0, \
0, 93, 0]\n[73, 0, 0, 94, 0]\n[82, 0, 0, 79, 0]\n[87, 0, 0, 88, 0]\n[95, 0, \
0, 83, 0]\n[91, 0, 0, 95, 0]\n[77, 0, 0, 99, 0]\n[75, 0, 0, 73, 0]\n[66, 0, \
0, 72, 0]\n[68, 0, 0, 72, 0]\n[64, 0, 0, 74, 0]\n[70, 0, 0, 66, 0]\n[78, 0, \
0, 80, 0]\n[70, 0, 0, 81, 0]\n[73, 0, 0, 70, 0]\n[83, 0, 0, 65, 0]\n[87, 0, \
0, 84, 0]\n[85, 0, 0, 84, 0]\n[86, 0, 0, 87, 0]\n[86, 0, 0, 75, 0]\n[91, 0, \
0, 99, 0]\n[100, 0, 0, 88, 0]\n[97, 0, 0, 101, 0]\n[95, 0, 0, 85, 0]\n[100, \
0, 0, 82, 0]\n[98, 0, 0, 91, 0]\n[95, 0, 0, 101, 0]\n[85, 0, 0, 112, 0]\n[95, \
0, 0, 65, 0]\n[90, 0, 0, 97, 0]\n[92, 0, 0, 80, 0]\n[93, 0, 0, 93, 0]\n[94, \
0, 0, 96, 0]\n[74, 0, 0, 101, 0]\n[80, 0, 0, 65, 0]\n[100, 0, 0, 69, 0]\n[89, \
0, 0, 107, 0]\n[82, 0, 0, 91, 0]\n[94, 0, 0, 68, 0]\n[82, 0, 0, 114, 0]\n[86, \
0, 0, 77, 0]\n[94, 0, 0, 76, 0]\n[82, 0, 0, 106, 0]\n[85, 0, 0, 81, 0]\n[89, \
0, 0, 84, 0]\n[74, 0, 0, 81, 0]\n[89, 0, 0, 62, 0]\n[91, 0, 0, 76, 0]\n[87, \
0, 0, 101, 0]\n[79, 0, 0, 83, 0]\n[81, 0, 0, 94, 0]\n[72, 0, 0, 83, 0]\n[71, \
0, 0, 76, 0]\n[76, 0, 0, 61, 0]\n[86, 0, 0, 76, 0]\n[88, 0, 0, 86, 0]\n[102, \
0, 0, 83, 0]\n[108, 0, 0, 96, 0]\n[104, 0, 0, 111, 0]\n[105, 0, 0, 84, 0]\n\
[96, 0, 0, 103, 0]\n[85, 0, 0, 90, 0]\n[97, 0, 0, 73, 0]\n[100, 0, 0, 82, 0]\n\
[98, 0, 0, 81, 0]\n[92, 0, 0, 100, 0]\n[110, 0, 0, 79, 0]\n[112, 0, 0, 111, \
0]\n[112, 0, 0, 102, 0]\n[121, 0, 0, 97, 0]\n[130, 0, 0, 110, 0]\n[115, 0, 0, \
113, 0]\n[96, 0, 0, 98, 0]\n[78, 0, 0, 103, 0]\n[81, 0, 0, 75, 0]\n[90, 0, 0, \
82, 0]\n[90, 0, 0, 91, 0]\n[95, 0, 0, 84, 0]\n[92, 0, 0, 90, 0]\n[98, 0, 0, \
74, 0]\n[97, 0, 0, 95, 0]\n[93, 0, 0, 97, 0]\n[82, 0, 0, 96, 0]\n[88, 0, 0, \
79, 0]\n[105, 0, 0, 73, 0]\n[101, 0, 0, 101, 0]\n[80, 0, 0, 106, 0]\n[88, 0, \
0, 71, 0]\n[79, 0, 0, 100, 0]\n[84, 0, 0, 84, 0]\n[80, 0, 0, 87, 0]\n[71, 0, \
0, 86, 0]\n[63, 0, 0, 77, 0]\n[73, 0, 0, 60, 0]\n[78, 0, 0, 76, 0]\n[83, 0, \
0, 85, 0]\n[79, 0, 0, 82, 0]\n[84, 0, 0, 82, 0]\n[81, 0, 0, 82, 0]\n[86, 0, \
0, 74, 0]\n[90, 0, 0, 94, 0]\n[98, 0, 0, 87, 0]\n[90, 0, 0, 93, 0]\n[86, 0, \
0, 96, 0]\n[96, 0, 0, 85, 0]\n[85, 0, 0, 107, 0]\n[76, 0, 0, 94, 0]\n[66, 0, \
0, 77, 0]\n[68, 0, 0, 63, 0]\n[71, 0, 0, 66, 0]\>\""}], ";"}]], "Input",
 CellChangeTimes->{{3.795272284341484*^9, 3.795272326377041*^9}, {
   3.795272430311056*^9, 3.795272459172511*^9}, {3.795272530041852*^9, 
   3.795272533166935*^9}, {3.795272652688711*^9, 3.7952726746618843`*^9}, {
   3.795272735321974*^9, 3.7952727381191797`*^9}, {3.795272824377345*^9, 
   3.795272825532275*^9}, 3.7952734874197083`*^9, {3.795273882383634*^9, 
   3.795273895947878*^9}, {3.795276704028225*^9, 3.795276709421897*^9}},
 CellLabel->
  "In[1548]:=",ExpressionUUID->"60c79b22-7316-4011-8f89-cac020b7b7fc"],

Cell["convert from python array to mathematica", "Text",
 CellChangeTimes->{{3.800015610612782*^9, 
  3.800015616968832*^9}},ExpressionUUID->"8fe71c00-62e9-43e1-90ab-\
956b048f3168"],

Cell[BoxData[{
 RowBox[{
  RowBox[{"d", "=", 
   RowBox[{"StringReplace", "[", 
    RowBox[{"data", ",", 
     RowBox[{"{", 
      RowBox[{
       RowBox[{"\"\<[\>\"", "\[Rule]", "\"\<{\>\""}], ",", 
       RowBox[{"\"\<]\>\"", "\[Rule]", "\"\<}\>\""}]}], "}"}]}], "]"}]}], 
  ";"}], "\[IndentingNewLine]", 
 RowBox[{
  RowBox[{"dataFixed", " ", "=", " ", 
   RowBox[{"ToExpression", "[", 
    RowBox[{"StringSplit", "[", 
     RowBox[{
      RowBox[{"StringSplit", "[", 
       RowBox[{"data", ",", "\"\<\\n\>\""}], "]"}], ",", 
      RowBox[{"{", 
       RowBox[{"\"\<,\>\"", ",", "\"\<[\>\"", ",", "\"\<]\>\""}], "}"}]}], 
     "]"}], "]"}]}], ";"}]}], "Input",
 CellChangeTimes->{{3.795272441051826*^9, 3.795272445461129*^9}, {
   3.7952724907309933`*^9, 3.795272519494508*^9}, {3.795272563878426*^9, 
   3.7952726013414*^9}, {3.795272706113306*^9, 3.7952727256411*^9}, {
   3.795272832756119*^9, 3.795272845052567*^9}, {3.7952735643832817`*^9, 
   3.7952735730661182`*^9}, {3.795273612938019*^9, 3.795273620150263*^9}, {
   3.7952736627541733`*^9, 3.795273706176141*^9}, {3.7952738055286007`*^9, 
   3.7952738307147007`*^9}, {3.795273921731044*^9, 3.795273988249963*^9}, {
   3.795274021080566*^9, 3.7952740215279408`*^9}, {3.795274122896686*^9, 
   3.79527416090571*^9}, {3.795274191080389*^9, 3.795274220497113*^9}, 
   3.795274970633462*^9},
 CellLabel->
  "In[1549]:=",ExpressionUUID->"c36681d9-a9a3-4d1a-b71f-c237a134533b"],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{
  RowBox[{"WTplot", " ", "=", 
   RowBox[{"Labeled", "[", 
    RowBox[{
     RowBox[{"Show", "[", 
      RowBox[{
       RowBox[{"ListLinePlot", "[", 
        RowBox[{
         RowBox[{"{", 
          RowBox[{
           RowBox[{"dataFixed", "[", 
            RowBox[{"[", 
             RowBox[{"All", ",", "1"}], "]"}], "]"}], ",", 
           RowBox[{"dataFixed", "[", 
            RowBox[{"[", 
             RowBox[{"All", ",", "2"}], "]"}], "]"}], ",", 
           RowBox[{"dataFixed", "[", 
            RowBox[{"[", 
             RowBox[{"All", ",", "3"}], "]"}], "]"}], ",", 
           RowBox[{"dataFixed", "[", 
            RowBox[{"[", 
             RowBox[{"All", ",", "4"}], "]"}], "]"}], ",", 
           RowBox[{"dataFixed", "[", 
            RowBox[{"[", 
             RowBox[{"All", ",", "5"}], "]"}], "]"}]}], "}"}], ",", 
         RowBox[{"PlotLayout", "\[Rule]", "\"\<Stacked\>\""}], ",", 
         "\[IndentingNewLine]", " ", 
         RowBox[{"PlotStyle", "\[Rule]", "None"}], ",", "\[IndentingNewLine]", 
         RowBox[{"Filling", "\[Rule]", 
          RowBox[{"{", 
           RowBox[{
            RowBox[{"1", "\[Rule]", 
             RowBox[{"{", 
              RowBox[{"Axis", ",", 
               RowBox[{"Lighter", "[", 
                RowBox[{"Red", ",", "0.7"}], "]"}]}], "}"}]}], ",", 
            RowBox[{"2", "\[Rule]", 
             RowBox[{"{", 
              RowBox[{
               RowBox[{"{", "1", "}"}], ",", 
               RowBox[{"Lighter", "[", 
                RowBox[{"Red", ",", "0.5"}], "]"}]}], "}"}]}], ",", 
            RowBox[{"3", "\[Rule]", 
             RowBox[{"{", 
              RowBox[{
               RowBox[{"{", "2", "}"}], ",", 
               RowBox[{"Lighter", "[", 
                RowBox[{"Red", ",", "0.3"}], "]"}]}], "}"}]}], ",", 
            RowBox[{"4", "\[Rule]", 
             RowBox[{"{", 
              RowBox[{
               RowBox[{"{", "3", "}"}], ",", 
               RowBox[{"Lighter", "[", 
                RowBox[{"Blue", ",", "0.7"}], "]"}]}], "}"}]}], ",", 
            RowBox[{"5", "\[Rule]", 
             RowBox[{"{", 
              RowBox[{
               RowBox[{"{", "4", "}"}], ",", 
               RowBox[{"Lighter", "[", 
                RowBox[{"Blue", ",", "0.5"}], "]"}]}], "}"}]}]}], "}"}]}], 
         ",", "\[IndentingNewLine]", 
         RowBox[{"PlotStyle", "\[Rule]", 
          RowBox[{"{", 
           RowBox[{
            RowBox[{"Lighter", "[", 
             RowBox[{"Red", ",", "0.7"}], "]"}], ",", 
            RowBox[{"Lighter", "[", 
             RowBox[{"Red", ",", "0.5"}], "]"}], ",", 
            RowBox[{"Lighter", "[", 
             RowBox[{"Red", ",", "0.3"}], "]"}], ",", 
            RowBox[{"Lighter", "[", 
             RowBox[{"Blue", ",", "0.7"}], "]"}], ",", 
            RowBox[{"Lighter", "[", 
             RowBox[{"Blue", ",", "0.5"}], "]"}]}], "}"}]}], ",", 
         RowBox[{"TicksStyle", "\[Rule]", "Medium"}]}], "]"}], ",", 
       RowBox[{"Plot", "[", 
        RowBox[{
         RowBox[{"{", "162", "}"}], ",", 
         RowBox[{"{", 
          RowBox[{"x", ",", "0", ",", "200"}], "}"}], ",", " ", 
         RowBox[{"PlotStyle", "\[Rule]", 
          RowBox[{"{", 
           RowBox[{"Dotted", ",", "Black"}], "}"}]}]}], "]"}]}], "]"}], ",", 
     RowBox[{"{", 
      RowBox[{
       RowBox[{
       "Text", "[", 
        "\"\<population size \
(\!\(\*StyleBox[\"N\",FontSlant->\"Italic\"]\))\>\"", "]"}], ",", 
       RowBox[{"Text", "[", "\"\<generations\>\"", "]"}]}], "}"}], ",", 
     RowBox[{"{", 
      RowBox[{"Left", ",", "Bottom"}], "}"}], ",", 
     RowBox[{"RotateLabel", "\[Rule]", "True"}]}], "]"}]}], ";"}]], "Input",
 CellChangeTimes->{{3.795274319982539*^9, 3.795274375055476*^9}, {
   3.795274493214356*^9, 3.7952745495383863`*^9}, {3.7952745840543756`*^9, 
   3.795274600641532*^9}, {3.795274668816704*^9, 3.795274700336207*^9}, {
   3.795274770765216*^9, 3.795274843551141*^9}, {3.79527488828366*^9, 
   3.7952749558487787`*^9}, 3.795276715968938*^9, {3.795277903215914*^9, 
   3.795277903719549*^9}, {3.7952782514403057`*^9, 3.79527826767161*^9}, 
   3.809689789004426*^9, 3.8096898526668053`*^9, {3.80968988407157*^9, 
   3.809689888907542*^9}, {3.8096899460287313`*^9, 3.809689969900029*^9}, 
   3.809775487766411*^9},
 CellLabel->
  "In[1551]:=",ExpressionUUID->"cd126c38-24a2-4d35-8221-8463fb431d8d"],

Cell[BoxData[
 TemplateBox[{
  "Filling","invfillentry",
   "\"\\!\\(\\*RowBox[{\\\"6\\\", \\\"\[Rule]\\\", RowBox[{\\\"{\\\", \
RowBox[{\\\"Axis\\\", \\\",\\\", RowBox[{\\\"{\\\", \
RowBox[{InterpretationBox[ButtonBox[TooltipBox[GraphicsBox[List[List[\
GrayLevel[0], RectangleBox[List[0, 0]]], List[GrayLevel[0], \
RectangleBox[List[1, -1]]], List[RGBColor[1.`, 0.7`, 0.7`], \
RectangleBox[List[0, -1], List[2, 1]]]], Rule[DefaultBaseStyle, \
\\\"ColorSwatchGraphics\\\"], Rule[AspectRatio, 1], Rule[Frame, True], \
Rule[FrameStyle, RGBColor[0.6666666666666667`, 0.4666666666666667`, \
0.4666666666666667`]], Rule[FrameTicks, None], Rule[PlotRangePadding, None], \
Rule[ImageSize, Dynamic[List[Automatic, Times[1.35`, \
CurrentValue[\\\"FontCapHeight\\\"], \
Power[AbsoluteCurrentValue[Magnification], -1]]]]]], \
StyleBox[RowBox[List[\\\"RGBColor\\\", \\\"[\\\", RowBox[List[\\\"1.`\\\", \\\
\",\\\", \\\"0.7`\\\", \\\",\\\", \\\"0.7`\\\"]], \\\"]\\\"]], \
Rule[NumberMarks, False]]], Rule[Appearance, None], Rule[BaseStyle, List[]], \
Rule[BaselinePosition, Baseline], Rule[DefaultBaseStyle, List[]], \
RuleDelayed[ButtonFunction, With[List[Set[Typeset`box$, EvaluationBox[]]], \
If[Not[AbsoluteCurrentValue[\\\"Deployed\\\"]], \
CompoundExpression[SelectionMove[Typeset`box$, All, Expression], \
Set[FrontEnd`Private`$ColorSelectorInitialAlpha, 1], \
Set[FrontEnd`Private`$ColorSelectorInitialColor, RGBColor[1.`, 0.7`, 0.7`]], \
Set[FrontEnd`Private`$ColorSelectorUseMakeBoxes, True], \
MathLink`CallFrontEnd[FrontEnd`AttachCell[Typeset`box$, \
FrontEndResource[\\\"RGBColorValueSelector\\\"], List[0, List[Left, Bottom]], \
List[Left, Top], Rule[\\\"ClosingActions\\\", \
List[\\\"SelectionDeparture\\\", \\\"ParentChanged\\\", \
\\\"EvaluatorQuit\\\"]]]]]]]], Rule[BaseStyle, Inherited], Rule[Evaluator, \
Automatic], Rule[Method, \\\"Preemptive\\\"]], RGBColor[1.`, 0.7`, 0.7`], \
Rule[Editable, False], Rule[Selectable, False]], \\\",\\\", \
InterpretationBox[ButtonBox[TooltipBox[GraphicsBox[List[List[GrayLevel[0], \
RectangleBox[List[0, 0]]], List[GrayLevel[0], RectangleBox[List[1, -1]]], \
List[RGBColor[1.`, 0.7`, 0.7`], RectangleBox[List[0, -1], List[2, 1]]]], \
Rule[DefaultBaseStyle, \\\"ColorSwatchGraphics\\\"], Rule[AspectRatio, 1], \
Rule[Frame, True], Rule[FrameStyle, RGBColor[0.6666666666666667`, \
0.4666666666666667`, 0.4666666666666667`]], Rule[FrameTicks, None], \
Rule[PlotRangePadding, None], Rule[ImageSize, Dynamic[List[Automatic, \
Times[1.35`, CurrentValue[\\\"FontCapHeight\\\"], \
Power[AbsoluteCurrentValue[Magnification], -1]]]]]], \
StyleBox[RowBox[List[\\\"RGBColor\\\", \\\"[\\\", RowBox[List[\\\"1.`\\\", \\\
\",\\\", \\\"0.7`\\\", \\\",\\\", \\\"0.7`\\\"]], \\\"]\\\"]], \
Rule[NumberMarks, False]]], Rule[Appearance, None], Rule[BaseStyle, List[]], \
Rule[BaselinePosition, Baseline], Rule[DefaultBaseStyle, List[]], \
RuleDelayed[ButtonFunction, With[List[Set[Typeset`box$, EvaluationBox[]]], \
If[Not[AbsoluteCurrentValue[\\\"Deployed\\\"]], \
CompoundExpression[SelectionMove[Typeset`box$, All, Expression], \
Set[FrontEnd`Private`$ColorSelectorInitialAlpha, 1], \
Set[FrontEnd`Private`$ColorSelectorInitialColor, RGBColor[1.`, 0.7`, 0.7`]], \
Set[FrontEnd`Private`$ColorSelectorUseMakeBoxes, True], \
MathLink`CallFrontEnd[FrontEnd`AttachCell[Typeset`box$, \
FrontEndResource[\\\"RGBColorValueSelector\\\"], List[0, List[Left, Bottom]], \
List[Left, Top], Rule[\\\"ClosingActions\\\", \
List[\\\"SelectionDeparture\\\", \\\"ParentChanged\\\", \
\\\"EvaluatorQuit\\\"]]]]]]]], Rule[BaseStyle, Inherited], Rule[Evaluator, \
Automatic], Rule[Method, \\\"Preemptive\\\"]], RGBColor[1.`, 0.7`, 0.7`], \
Rule[Editable, False], Rule[Selectable, False]]}], \\\"}\\\"}]}], \
\\\"}\\\"}]}]\\) is not a valid Filling specification.\"",2,1551,43,
   32715148735184203895,"Local"},
  "MessageTemplate"]], "Message", "MSG",
 CellChangeTimes->{
  3.809689793445969*^9, 3.809689857733173*^9, {3.809689952524951*^9, 
   3.809689970958087*^9}, 3.814005345147523*^9, 3.814196575419015*^9, 
   3.814197132003066*^9},
 CellLabel->
  "During evaluation of \
In[1551]:=",ExpressionUUID->"8f2818ca-8da8-44d6-b676-a1aa7b4fb700"],

Cell[BoxData[
 TemplateBox[{
  "Filling","invfillentry",
   "\"\\!\\(\\*RowBox[{\\\"7\\\", \\\"\[Rule]\\\", RowBox[{\\\"{\\\", \
RowBox[{RowBox[{\\\"{\\\", \\\"1\\\", \\\"}\\\"}], \\\",\\\", RowBox[{\\\"{\\\
\", RowBox[{InterpretationBox[ButtonBox[TooltipBox[GraphicsBox[List[List[\
GrayLevel[0], RectangleBox[List[0, 0]]], List[GrayLevel[0], \
RectangleBox[List[1, -1]]], List[RGBColor[1.`, 0.5`, 0.5`], \
RectangleBox[List[0, -1], List[2, 1]]]], Rule[DefaultBaseStyle, \
\\\"ColorSwatchGraphics\\\"], Rule[AspectRatio, 1], Rule[Frame, True], \
Rule[FrameStyle, RGBColor[0.6666666666666667`, 0.33333333333333337`, \
0.33333333333333337`]], Rule[FrameTicks, None], Rule[PlotRangePadding, None], \
Rule[ImageSize, Dynamic[List[Automatic, Times[1.35`, \
CurrentValue[\\\"FontCapHeight\\\"], \
Power[AbsoluteCurrentValue[Magnification], -1]]]]]], \
StyleBox[RowBox[List[\\\"RGBColor\\\", \\\"[\\\", RowBox[List[\\\"1.`\\\", \\\
\",\\\", \\\"0.5`\\\", \\\",\\\", \\\"0.5`\\\"]], \\\"]\\\"]], \
Rule[NumberMarks, False]]], Rule[Appearance, None], Rule[BaseStyle, List[]], \
Rule[BaselinePosition, Baseline], Rule[DefaultBaseStyle, List[]], \
RuleDelayed[ButtonFunction, With[List[Set[Typeset`box$, EvaluationBox[]]], \
If[Not[AbsoluteCurrentValue[\\\"Deployed\\\"]], \
CompoundExpression[SelectionMove[Typeset`box$, All, Expression], \
Set[FrontEnd`Private`$ColorSelectorInitialAlpha, 1], \
Set[FrontEnd`Private`$ColorSelectorInitialColor, RGBColor[1.`, 0.5`, 0.5`]], \
Set[FrontEnd`Private`$ColorSelectorUseMakeBoxes, True], \
MathLink`CallFrontEnd[FrontEnd`AttachCell[Typeset`box$, \
FrontEndResource[\\\"RGBColorValueSelector\\\"], List[0, List[Left, Bottom]], \
List[Left, Top], Rule[\\\"ClosingActions\\\", \
List[\\\"SelectionDeparture\\\", \\\"ParentChanged\\\", \
\\\"EvaluatorQuit\\\"]]]]]]]], Rule[BaseStyle, Inherited], Rule[Evaluator, \
Automatic], Rule[Method, \\\"Preemptive\\\"]], RGBColor[1.`, 0.5`, 0.5`], \
Rule[Editable, False], Rule[Selectable, False]], \\\",\\\", \
InterpretationBox[ButtonBox[TooltipBox[GraphicsBox[List[List[GrayLevel[0], \
RectangleBox[List[0, 0]]], List[GrayLevel[0], RectangleBox[List[1, -1]]], \
List[RGBColor[1.`, 0.5`, 0.5`], RectangleBox[List[0, -1], List[2, 1]]]], \
Rule[DefaultBaseStyle, \\\"ColorSwatchGraphics\\\"], Rule[AspectRatio, 1], \
Rule[Frame, True], Rule[FrameStyle, RGBColor[0.6666666666666667`, \
0.33333333333333337`, 0.33333333333333337`]], Rule[FrameTicks, None], \
Rule[PlotRangePadding, None], Rule[ImageSize, Dynamic[List[Automatic, \
Times[1.35`, CurrentValue[\\\"FontCapHeight\\\"], \
Power[AbsoluteCurrentValue[Magnification], -1]]]]]], \
StyleBox[RowBox[List[\\\"RGBColor\\\", \\\"[\\\", RowBox[List[\\\"1.`\\\", \\\
\",\\\", \\\"0.5`\\\", \\\",\\\", \\\"0.5`\\\"]], \\\"]\\\"]], \
Rule[NumberMarks, False]]], Rule[Appearance, None], Rule[BaseStyle, List[]], \
Rule[BaselinePosition, Baseline], Rule[DefaultBaseStyle, List[]], \
RuleDelayed[ButtonFunction, With[List[Set[Typeset`box$, EvaluationBox[]]], \
If[Not[AbsoluteCurrentValue[\\\"Deployed\\\"]], \
CompoundExpression[SelectionMove[Typeset`box$, All, Expression], \
Set[FrontEnd`Private`$ColorSelectorInitialAlpha, 1], \
Set[FrontEnd`Private`$ColorSelectorInitialColor, RGBColor[1.`, 0.5`, 0.5`]], \
Set[FrontEnd`Private`$ColorSelectorUseMakeBoxes, True], \
MathLink`CallFrontEnd[FrontEnd`AttachCell[Typeset`box$, \
FrontEndResource[\\\"RGBColorValueSelector\\\"], List[0, List[Left, Bottom]], \
List[Left, Top], Rule[\\\"ClosingActions\\\", \
List[\\\"SelectionDeparture\\\", \\\"ParentChanged\\\", \
\\\"EvaluatorQuit\\\"]]]]]]]], Rule[BaseStyle, Inherited], Rule[Evaluator, \
Automatic], Rule[Method, \\\"Preemptive\\\"]], RGBColor[1.`, 0.5`, 0.5`], \
Rule[Editable, False], Rule[Selectable, False]]}], \\\"}\\\"}]}], \
\\\"}\\\"}]}]\\) is not a valid Filling specification.\"",2,1551,44,
   32715148735184203895,"Local"},
  "MessageTemplate"]], "Message", "MSG",
 CellChangeTimes->{
  3.809689793445969*^9, 3.809689857733173*^9, {3.809689952524951*^9, 
   3.809689970958087*^9}, 3.814005345147523*^9, 3.814196575419015*^9, 
   3.814197132087689*^9},
 CellLabel->
  "During evaluation of \
In[1551]:=",ExpressionUUID->"33a9fc75-1dac-4d12-9811-08e97d74c169"],

Cell[BoxData[
 TemplateBox[{
  "Filling","invfillentry",
   "\"\\!\\(\\*RowBox[{\\\"8\\\", \\\"\[Rule]\\\", RowBox[{\\\"{\\\", \
RowBox[{RowBox[{\\\"{\\\", \\\"2\\\", \\\"}\\\"}], \\\",\\\", RowBox[{\\\"{\\\
\", RowBox[{InterpretationBox[ButtonBox[TooltipBox[GraphicsBox[List[List[\
GrayLevel[0], RectangleBox[List[0, 0]]], List[GrayLevel[0], \
RectangleBox[List[1, -1]]], List[RGBColor[1.`, 0.3`, 0.3`], \
RectangleBox[List[0, -1], List[2, 1]]]], Rule[DefaultBaseStyle, \
\\\"ColorSwatchGraphics\\\"], Rule[AspectRatio, 1], Rule[Frame, True], \
Rule[FrameStyle, RGBColor[0.6666666666666667`, 0.2`, 0.2`]], Rule[FrameTicks, \
None], Rule[PlotRangePadding, None], Rule[ImageSize, Dynamic[List[Automatic, \
Times[1.35`, CurrentValue[\\\"FontCapHeight\\\"], \
Power[AbsoluteCurrentValue[Magnification], -1]]]]]], \
StyleBox[RowBox[List[\\\"RGBColor\\\", \\\"[\\\", RowBox[List[\\\"1.`\\\", \\\
\",\\\", \\\"0.3`\\\", \\\",\\\", \\\"0.3`\\\"]], \\\"]\\\"]], \
Rule[NumberMarks, False]]], Rule[Appearance, None], Rule[BaseStyle, List[]], \
Rule[BaselinePosition, Baseline], Rule[DefaultBaseStyle, List[]], \
RuleDelayed[ButtonFunction, With[List[Set[Typeset`box$, EvaluationBox[]]], \
If[Not[AbsoluteCurrentValue[\\\"Deployed\\\"]], \
CompoundExpression[SelectionMove[Typeset`box$, All, Expression], \
Set[FrontEnd`Private`$ColorSelectorInitialAlpha, 1], \
Set[FrontEnd`Private`$ColorSelectorInitialColor, RGBColor[1.`, 0.3`, 0.3`]], \
Set[FrontEnd`Private`$ColorSelectorUseMakeBoxes, True], \
MathLink`CallFrontEnd[FrontEnd`AttachCell[Typeset`box$, \
FrontEndResource[\\\"RGBColorValueSelector\\\"], List[0, List[Left, Bottom]], \
List[Left, Top], Rule[\\\"ClosingActions\\\", \
List[\\\"SelectionDeparture\\\", \\\"ParentChanged\\\", \
\\\"EvaluatorQuit\\\"]]]]]]]], Rule[BaseStyle, Inherited], Rule[Evaluator, \
Automatic], Rule[Method, \\\"Preemptive\\\"]], RGBColor[1.`, 0.3`, 0.3`], \
Rule[Editable, False], Rule[Selectable, False]], \\\",\\\", \
InterpretationBox[ButtonBox[TooltipBox[GraphicsBox[List[List[GrayLevel[0], \
RectangleBox[List[0, 0]]], List[GrayLevel[0], RectangleBox[List[1, -1]]], \
List[RGBColor[1.`, 0.3`, 0.3`], RectangleBox[List[0, -1], List[2, 1]]]], \
Rule[DefaultBaseStyle, \\\"ColorSwatchGraphics\\\"], Rule[AspectRatio, 1], \
Rule[Frame, True], Rule[FrameStyle, RGBColor[0.6666666666666667`, 0.2`, \
0.2`]], Rule[FrameTicks, None], Rule[PlotRangePadding, None], Rule[ImageSize, \
Dynamic[List[Automatic, Times[1.35`, CurrentValue[\\\"FontCapHeight\\\"], \
Power[AbsoluteCurrentValue[Magnification], -1]]]]]], \
StyleBox[RowBox[List[\\\"RGBColor\\\", \\\"[\\\", RowBox[List[\\\"1.`\\\", \\\
\",\\\", \\\"0.3`\\\", \\\",\\\", \\\"0.3`\\\"]], \\\"]\\\"]], \
Rule[NumberMarks, False]]], Rule[Appearance, None], Rule[BaseStyle, List[]], \
Rule[BaselinePosition, Baseline], Rule[DefaultBaseStyle, List[]], \
RuleDelayed[ButtonFunction, With[List[Set[Typeset`box$, EvaluationBox[]]], \
If[Not[AbsoluteCurrentValue[\\\"Deployed\\\"]], \
CompoundExpression[SelectionMove[Typeset`box$, All, Expression], \
Set[FrontEnd`Private`$ColorSelectorInitialAlpha, 1], \
Set[FrontEnd`Private`$ColorSelectorInitialColor, RGBColor[1.`, 0.3`, 0.3`]], \
Set[FrontEnd`Private`$ColorSelectorUseMakeBoxes, True], \
MathLink`CallFrontEnd[FrontEnd`AttachCell[Typeset`box$, \
FrontEndResource[\\\"RGBColorValueSelector\\\"], List[0, List[Left, Bottom]], \
List[Left, Top], Rule[\\\"ClosingActions\\\", \
List[\\\"SelectionDeparture\\\", \\\"ParentChanged\\\", \
\\\"EvaluatorQuit\\\"]]]]]]]], Rule[BaseStyle, Inherited], Rule[Evaluator, \
Automatic], Rule[Method, \\\"Preemptive\\\"]], RGBColor[1.`, 0.3`, 0.3`], \
Rule[Editable, False], Rule[Selectable, False]]}], \\\"}\\\"}]}], \
\\\"}\\\"}]}]\\) is not a valid Filling specification.\"",2,1551,45,
   32715148735184203895,"Local"},
  "MessageTemplate"]], "Message", "MSG",
 CellChangeTimes->{
  3.809689793445969*^9, 3.809689857733173*^9, {3.809689952524951*^9, 
   3.809689970958087*^9}, 3.814005345147523*^9, 3.814196575419015*^9, 
   3.814197132190824*^9},
 CellLabel->
  "During evaluation of \
In[1551]:=",ExpressionUUID->"a1d348a0-1634-4fff-8d06-e9029b0ef8f8"],

Cell[BoxData[
 TemplateBox[{
  "General","stop",
   "\"Further output of \\!\\(\\*StyleBox[RowBox[{\\\"Filling\\\", \
\\\"::\\\", \\\"invfillentry\\\"}], \\\"MessageName\\\"]\\) will be \
suppressed during this calculation.\"",2,1551,46,32715148735184203895,"Local"},
  "MessageTemplate"]], "Message", "MSG",
 CellChangeTimes->{
  3.809689793445969*^9, 3.809689857733173*^9, {3.809689952524951*^9, 
   3.809689970958087*^9}, 3.814005345147523*^9, 3.814196575419015*^9, 
   3.814197132266862*^9},
 CellLabel->
  "During evaluation of \
In[1551]:=",ExpressionUUID->"2d016115-946c-46cc-a116-0ac21c068a58"]
}, Open  ]],

Cell[BoxData[{
 RowBox[{
  RowBox[{
  "data", "=", 
   "\"\<[71, 2, 0, 79, 0]\n[69, 1, 0, 77, 1]\n[85, 3, 0, 61, 0]\n[77, 0, 0, \
95, 1]\n[72, 2, 0, 87, 0]\n[69, 2, 0, 68, 1]\n[58, 1, 0, 77, 1]\n[66, 3, 0, \
56, 0]\n[71, 2, 0, 71, 3]\n[63, 2, 0, 67, 1]\n[74, 3, 0, 67, 1]\n[78, 5, 0, \
76, 0]\n[80, 4, 0, 76, 4]\n[78, 11, 0, 79, 1]\n[81, 7, 0, 72, 5]\n[72, 11, 0, \
80, 3]\n[59, 15, 0, 86, 3]\n[65, 13, 0, 56, 7]\n[61, 26, 0, 60, 7]\n[75, 28, \
0, 77, 6]\n[68, 20, 1, 93, 16]\n[66, 23, 2, 69, 9]\n[62, 25, 2, 70, 17]\n[55, \
24, 6, 82, 10]\n[61, 24, 2, 74, 20]\n[48, 37, 5, 57, 8]\n[65, 45, 4, 50, 24]\n\
[67, 63, 6, 59, 11]\n[78, 45, 8, 81, 24]\n[82, 48, 5, 71, 19]\n[89, 64, 9, \
69, 22]\n[73, 75, 21, 73, 27]\n[67, 71, 14, 68, 51]\n[49, 76, 32, 57, 28]\n\
[61, 85, 30, 54, 37]\n[57, 106, 41, 43, 46]\n[41, 121, 67, 45, 42]\n[49, 123, \
72, 51, 57]\n[45, 113, 80, 40, 42]\n[48, 133, 84, 50, 56]\n[51, 138, 95, 33, \
47]\n[33, 162, 120, 30, 61]\n[29, 152, 150, 31, 50]\n[34, 126, 144, 33, 66]\n\
[28, 131, 167, 30, 57]\n[32, 104, 164, 21, 63]\n[18, 108, 189, 17, 41]\n[15, \
135, 182, 17, 75]\n[13, 116, 208, 11, 47]\n[10, 121, 246, 6, 33]\n[6, 141, \
265, 11, 33]\n[21, 112, 250, 8, 63]\n[2, 110, 256, 4, 34]\n[3, 98, 283, 7, \
27]\n[5, 92, 295, 17, 71]\n[5, 92, 273, 6, 47]\n[4, 87, 294, 10, 22]\n[11, \
99, 233, 13, 101]\n[2, 81, 267, 4, 21]\n[6, 86, 249, 8, 101]\n[4, 67, 281, 3, \
14]\n[4, 55, 214, 23, 173]\n[1, 45, 202, 2, 16]\n[3, 43, 237, 8, 74]\n[3, 54, \
283, 1, 25]\n[2, 40, 339, 7, 46]\n[1, 64, 352, 1, 27]\n[0, 46, 321, 5, 88]\n\
[1, 33, 331, 0, 16]\n[0, 22, 277, 8, 209]\n[1, 20, 254, 0, 7]\n[0, 5, 149, \
12, 272]\n[1, 10, 149, 1, 5]\n[0, 17, 112, 3, 165]\n[0, 19, 152, 1, 7]\n[1, \
28, 136, 7, 125]\n[2, 27, 195, 3, 12]\n[0, 39, 194, 10, 116]\n[1, 39, 247, 3, \
11]\n[4, 52, 175, 12, 173]\n[1, 48, 204, 1, 11]\n[0, 27, 184, 18, 146]\n[2, \
34, 196, 0, 17]\n[0, 32, 274, 3, 57]\n[1, 30, 345, 0, 6]\n[0, 5, 128, 13, \
410]\n[0, 5, 111, 0, 1]\n[0, 0, 32, 7, 214]\n[0, 2, 41, 0, 2]\n[0, 1, 46, 0, \
40]\n[0, 2, 83, 0, 0]\n[0, 0, 0, 0, 0]\>\""}], ";"}], "\[IndentingNewLine]", 
 RowBox[{
  RowBox[{"d", "=", 
   RowBox[{"StringReplace", "[", 
    RowBox[{"data", ",", 
     RowBox[{"{", 
      RowBox[{
       RowBox[{"\"\<[\>\"", "\[Rule]", "\"\<{\>\""}], ",", 
       RowBox[{"\"\<]\>\"", "\[Rule]", "\"\<}\>\""}]}], "}"}]}], "]"}]}], 
  ";"}], "\[IndentingNewLine]", 
 RowBox[{
  RowBox[{"dataFixed", " ", "=", " ", 
   RowBox[{"ToExpression", "[", 
    RowBox[{"StringSplit", "[", 
     RowBox[{
      RowBox[{"StringSplit", "[", 
       RowBox[{"data", ",", "\"\<\\n\>\""}], "]"}], ",", 
      RowBox[{"{", 
       RowBox[{"\"\<,\>\"", ",", "\"\<[\>\"", ",", "\"\<]\>\""}], "}"}]}], 
     "]"}], "]"}]}], ";"}], "\[IndentingNewLine]"}], "Input",
 CellChangeTimes->{{3.795275047381508*^9, 3.7952750734131536`*^9}, {
   3.79527570965047*^9, 3.7952757168600616`*^9}, {3.795276006045801*^9, 
   3.795276007123762*^9}, {3.795276053227146*^9, 3.795276056745636*^9}, 
   3.79527630981946*^9, 3.7952763494262238`*^9, 3.795276488893711*^9, 
   3.795276775131549*^9, {3.795276999281911*^9, 3.795277002957556*^9}},
 CellLabel->
  "In[1552]:=",ExpressionUUID->"a74d00b9-b60a-4cd6-ad5d-82aafbb3a80d"],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{
  RowBox[{"extPlot", " ", "=", 
   RowBox[{"Labeled", "[", 
    RowBox[{
     RowBox[{"Show", "[", 
      RowBox[{
       RowBox[{"ListLinePlot", "[", 
        RowBox[{
         RowBox[{"{", 
          RowBox[{
           RowBox[{"dataFixed", "[", 
            RowBox[{"[", 
             RowBox[{"All", ",", "1"}], "]"}], "]"}], ",", 
           RowBox[{"dataFixed", "[", 
            RowBox[{"[", 
             RowBox[{"All", ",", "2"}], "]"}], "]"}], ",", 
           RowBox[{"dataFixed", "[", 
            RowBox[{"[", 
             RowBox[{"All", ",", "3"}], "]"}], "]"}], ",", 
           RowBox[{"dataFixed", "[", 
            RowBox[{"[", 
             RowBox[{"All", ",", "4"}], "]"}], "]"}], ",", 
           RowBox[{"dataFixed", "[", 
            RowBox[{"[", 
             RowBox[{"All", ",", "5"}], "]"}], "]"}]}], "}"}], ",", 
         RowBox[{"PlotLayout", "\[Rule]", "\"\<Stacked\>\""}], ",", " ", 
         RowBox[{"Filling", "\[Rule]", 
          RowBox[{"{", 
           RowBox[{
            RowBox[{"1", "\[Rule]", 
             RowBox[{"{", 
              RowBox[{"Axis", ",", 
               RowBox[{"Lighter", "[", 
                RowBox[{"Red", ",", "0.7"}], "]"}]}], "}"}]}], ",", 
            RowBox[{"2", "\[Rule]", 
             RowBox[{"{", 
              RowBox[{
               RowBox[{"{", "1", "}"}], ",", 
               RowBox[{"Lighter", "[", 
                RowBox[{"Red", ",", "0.5"}], "]"}]}], "}"}]}], ",", 
            RowBox[{"3", "\[Rule]", 
             RowBox[{"{", 
              RowBox[{
               RowBox[{"{", "2", "}"}], ",", 
               RowBox[{"Lighter", "[", 
                RowBox[{"Red", ",", "0.3"}], "]"}]}], "}"}]}], ",", 
            RowBox[{"4", "\[Rule]", 
             RowBox[{"{", 
              RowBox[{
               RowBox[{"{", "3", "}"}], ",", 
               RowBox[{"Lighter", "[", 
                RowBox[{"Blue", ",", "0.7"}], "]"}]}], "}"}]}], ",", 
            RowBox[{"5", "\[Rule]", 
             RowBox[{"{", 
              RowBox[{
               RowBox[{"{", "4", "}"}], ",", 
               RowBox[{"Lighter", "[", 
                RowBox[{"Blue", ",", "0.5"}], "]"}]}], "}"}]}]}], "}"}]}], 
         ",", "\[IndentingNewLine]", 
         RowBox[{"PlotStyle", "\[Rule]", 
          RowBox[{"{", 
           RowBox[{
            RowBox[{"Lighter", "[", 
             RowBox[{"Red", ",", "0.7"}], "]"}], ",", 
            RowBox[{"Lighter", "[", 
             RowBox[{"Red", ",", "0.5"}], "]"}], ",", 
            RowBox[{"Lighter", "[", 
             RowBox[{"Red", ",", "0.3"}], "]"}], ",", 
            RowBox[{"Lighter", "[", 
             RowBox[{"Blue", ",", "0.7"}], "]"}], ",", 
            RowBox[{"Lighter", "[", 
             RowBox[{"Blue", ",", "0.5"}], "]"}]}], "}"}]}], ",", 
         RowBox[{"TicksStyle", "\[Rule]", "Medium"}]}], "]"}], ",", 
       RowBox[{"Plot", "[", 
        RowBox[{
         RowBox[{"{", "162", "}"}], ",", 
         RowBox[{"{", 
          RowBox[{"x", ",", "0", ",", "200"}], "}"}], ",", " ", 
         RowBox[{"PlotStyle", "\[Rule]", 
          RowBox[{"{", 
           RowBox[{"Dotted", ",", "Black"}], "}"}]}]}], "]"}]}], "]"}], ",", 
     RowBox[{"{", 
      RowBox[{
       RowBox[{
       "Text", "[", 
        "\"\<population size \
(\!\(\*StyleBox[\"N\",FontSlant->\"Italic\"]\))\>\"", "]"}], ",", 
       RowBox[{"Text", "[", "\"\<generations\>\"", "]"}]}], "}"}], ",", 
     RowBox[{"{", 
      RowBox[{"Left", ",", "Bottom"}], "}"}], ",", 
     RowBox[{"RotateLabel", "\[Rule]", "True"}]}], "]"}]}], ";"}]], "Input",
 CellChangeTimes->{{3.7952726044363947`*^9, 3.7952726080817213`*^9}, {
   3.795272875877472*^9, 3.795272893357538*^9}, {3.795273179147029*^9, 
   3.7952732005416803`*^9}, {3.79527328690026*^9, 3.795273302501566*^9}, {
   3.795274241785592*^9, 3.7952742707178097`*^9}, 3.795274307936372*^9, {
   3.795275325957712*^9, 3.7952753480706263`*^9}, {3.795275384523851*^9, 
   3.795275531244299*^9}, 3.795276013290897*^9, {3.7952760938746862`*^9, 
   3.795276166704401*^9}, {3.795276198298027*^9, 3.795276232818214*^9}, {
   3.795276500318206*^9, 3.795276504994471*^9}, {3.79527698727664*^9, 
   3.79527700482725*^9}, {3.795277893958102*^9, 3.795277894539732*^9}, 
   3.7952782303810873`*^9, 3.7952782787675667`*^9, 3.80968977365691*^9, 
   3.809775494528908*^9},
 CellLabel->
  "In[1555]:=",ExpressionUUID->"e2a04a6a-a1cb-4268-a08e-d0828a036520"],

Cell[BoxData[
 TemplateBox[{
  "Filling","invfillentry",
   "\"\\!\\(\\*RowBox[{\\\"6\\\", \\\"\[Rule]\\\", RowBox[{\\\"{\\\", \
RowBox[{\\\"Axis\\\", \\\",\\\", RowBox[{\\\"{\\\", \
RowBox[{InterpretationBox[ButtonBox[TooltipBox[GraphicsBox[List[List[\
GrayLevel[0], RectangleBox[List[0, 0]]], List[GrayLevel[0], \
RectangleBox[List[1, -1]]], List[RGBColor[1.`, 0.7`, 0.7`], \
RectangleBox[List[0, -1], List[2, 1]]]], Rule[DefaultBaseStyle, \
\\\"ColorSwatchGraphics\\\"], Rule[AspectRatio, 1], Rule[Frame, True], \
Rule[FrameStyle, RGBColor[0.6666666666666667`, 0.4666666666666667`, \
0.4666666666666667`]], Rule[FrameTicks, None], Rule[PlotRangePadding, None], \
Rule[ImageSize, Dynamic[List[Automatic, Times[1.35`, \
CurrentValue[\\\"FontCapHeight\\\"], \
Power[AbsoluteCurrentValue[Magnification], -1]]]]]], \
StyleBox[RowBox[List[\\\"RGBColor\\\", \\\"[\\\", RowBox[List[\\\"1.`\\\", \\\
\",\\\", \\\"0.7`\\\", \\\",\\\", \\\"0.7`\\\"]], \\\"]\\\"]], \
Rule[NumberMarks, False]]], Rule[Appearance, None], Rule[BaseStyle, List[]], \
Rule[BaselinePosition, Baseline], Rule[DefaultBaseStyle, List[]], \
RuleDelayed[ButtonFunction, With[List[Set[Typeset`box$, EvaluationBox[]]], \
If[Not[AbsoluteCurrentValue[\\\"Deployed\\\"]], \
CompoundExpression[SelectionMove[Typeset`box$, All, Expression], \
Set[FrontEnd`Private`$ColorSelectorInitialAlpha, 1], \
Set[FrontEnd`Private`$ColorSelectorInitialColor, RGBColor[1.`, 0.7`, 0.7`]], \
Set[FrontEnd`Private`$ColorSelectorUseMakeBoxes, True], \
MathLink`CallFrontEnd[FrontEnd`AttachCell[Typeset`box$, \
FrontEndResource[\\\"RGBColorValueSelector\\\"], List[0, List[Left, Bottom]], \
List[Left, Top], Rule[\\\"ClosingActions\\\", \
List[\\\"SelectionDeparture\\\", \\\"ParentChanged\\\", \
\\\"EvaluatorQuit\\\"]]]]]]]], Rule[BaseStyle, Inherited], Rule[Evaluator, \
Automatic], Rule[Method, \\\"Preemptive\\\"]], RGBColor[1.`, 0.7`, 0.7`], \
Rule[Editable, False], Rule[Selectable, False]], \\\",\\\", \
InterpretationBox[ButtonBox[TooltipBox[GraphicsBox[List[List[GrayLevel[0], \
RectangleBox[List[0, 0]]], List[GrayLevel[0], RectangleBox[List[1, -1]]], \
List[RGBColor[1.`, 0.7`, 0.7`], RectangleBox[List[0, -1], List[2, 1]]]], \
Rule[DefaultBaseStyle, \\\"ColorSwatchGraphics\\\"], Rule[AspectRatio, 1], \
Rule[Frame, True], Rule[FrameStyle, RGBColor[0.6666666666666667`, \
0.4666666666666667`, 0.4666666666666667`]], Rule[FrameTicks, None], \
Rule[PlotRangePadding, None], Rule[ImageSize, Dynamic[List[Automatic, \
Times[1.35`, CurrentValue[\\\"FontCapHeight\\\"], \
Power[AbsoluteCurrentValue[Magnification], -1]]]]]], \
StyleBox[RowBox[List[\\\"RGBColor\\\", \\\"[\\\", RowBox[List[\\\"1.`\\\", \\\
\",\\\", \\\"0.7`\\\", \\\",\\\", \\\"0.7`\\\"]], \\\"]\\\"]], \
Rule[NumberMarks, False]]], Rule[Appearance, None], Rule[BaseStyle, List[]], \
Rule[BaselinePosition, Baseline], Rule[DefaultBaseStyle, List[]], \
RuleDelayed[ButtonFunction, With[List[Set[Typeset`box$, EvaluationBox[]]], \
If[Not[AbsoluteCurrentValue[\\\"Deployed\\\"]], \
CompoundExpression[SelectionMove[Typeset`box$, All, Expression], \
Set[FrontEnd`Private`$ColorSelectorInitialAlpha, 1], \
Set[FrontEnd`Private`$ColorSelectorInitialColor, RGBColor[1.`, 0.7`, 0.7`]], \
Set[FrontEnd`Private`$ColorSelectorUseMakeBoxes, True], \
MathLink`CallFrontEnd[FrontEnd`AttachCell[Typeset`box$, \
FrontEndResource[\\\"RGBColorValueSelector\\\"], List[0, List[Left, Bottom]], \
List[Left, Top], Rule[\\\"ClosingActions\\\", \
List[\\\"SelectionDeparture\\\", \\\"ParentChanged\\\", \
\\\"EvaluatorQuit\\\"]]]]]]]], Rule[BaseStyle, Inherited], Rule[Evaluator, \
Automatic], Rule[Method, \\\"Preemptive\\\"]], RGBColor[1.`, 0.7`, 0.7`], \
Rule[Editable, False], Rule[Selectable, False]]}], \\\"}\\\"}]}], \
\\\"}\\\"}]}]\\) is not a valid Filling specification.\"",2,1555,47,
   32715148735184203895,"Local"},
  "MessageTemplate"]], "Message", "MSG",
 CellChangeTimes->{
  3.8002691222071257`*^9, 3.809677030703689*^9, {3.809689775327434*^9, 
   3.809689779697974*^9}, 3.814005347046734*^9, 3.8141965760780153`*^9, 
   3.814197132679696*^9},
 CellLabel->
  "During evaluation of \
In[1555]:=",ExpressionUUID->"82d2c2ea-11ad-4be2-b954-32e487473612"],

Cell[BoxData[
 TemplateBox[{
  "Filling","invfillentry",
   "\"\\!\\(\\*RowBox[{\\\"7\\\", \\\"\[Rule]\\\", RowBox[{\\\"{\\\", \
RowBox[{RowBox[{\\\"{\\\", \\\"1\\\", \\\"}\\\"}], \\\",\\\", RowBox[{\\\"{\\\
\", RowBox[{InterpretationBox[ButtonBox[TooltipBox[GraphicsBox[List[List[\
GrayLevel[0], RectangleBox[List[0, 0]]], List[GrayLevel[0], \
RectangleBox[List[1, -1]]], List[RGBColor[1.`, 0.5`, 0.5`], \
RectangleBox[List[0, -1], List[2, 1]]]], Rule[DefaultBaseStyle, \
\\\"ColorSwatchGraphics\\\"], Rule[AspectRatio, 1], Rule[Frame, True], \
Rule[FrameStyle, RGBColor[0.6666666666666667`, 0.33333333333333337`, \
0.33333333333333337`]], Rule[FrameTicks, None], Rule[PlotRangePadding, None], \
Rule[ImageSize, Dynamic[List[Automatic, Times[1.35`, \
CurrentValue[\\\"FontCapHeight\\\"], \
Power[AbsoluteCurrentValue[Magnification], -1]]]]]], \
StyleBox[RowBox[List[\\\"RGBColor\\\", \\\"[\\\", RowBox[List[\\\"1.`\\\", \\\
\",\\\", \\\"0.5`\\\", \\\",\\\", \\\"0.5`\\\"]], \\\"]\\\"]], \
Rule[NumberMarks, False]]], Rule[Appearance, None], Rule[BaseStyle, List[]], \
Rule[BaselinePosition, Baseline], Rule[DefaultBaseStyle, List[]], \
RuleDelayed[ButtonFunction, With[List[Set[Typeset`box$, EvaluationBox[]]], \
If[Not[AbsoluteCurrentValue[\\\"Deployed\\\"]], \
CompoundExpression[SelectionMove[Typeset`box$, All, Expression], \
Set[FrontEnd`Private`$ColorSelectorInitialAlpha, 1], \
Set[FrontEnd`Private`$ColorSelectorInitialColor, RGBColor[1.`, 0.5`, 0.5`]], \
Set[FrontEnd`Private`$ColorSelectorUseMakeBoxes, True], \
MathLink`CallFrontEnd[FrontEnd`AttachCell[Typeset`box$, \
FrontEndResource[\\\"RGBColorValueSelector\\\"], List[0, List[Left, Bottom]], \
List[Left, Top], Rule[\\\"ClosingActions\\\", \
List[\\\"SelectionDeparture\\\", \\\"ParentChanged\\\", \
\\\"EvaluatorQuit\\\"]]]]]]]], Rule[BaseStyle, Inherited], Rule[Evaluator, \
Automatic], Rule[Method, \\\"Preemptive\\\"]], RGBColor[1.`, 0.5`, 0.5`], \
Rule[Editable, False], Rule[Selectable, False]], \\\",\\\", \
InterpretationBox[ButtonBox[TooltipBox[GraphicsBox[List[List[GrayLevel[0], \
RectangleBox[List[0, 0]]], List[GrayLevel[0], RectangleBox[List[1, -1]]], \
List[RGBColor[1.`, 0.5`, 0.5`], RectangleBox[List[0, -1], List[2, 1]]]], \
Rule[DefaultBaseStyle, \\\"ColorSwatchGraphics\\\"], Rule[AspectRatio, 1], \
Rule[Frame, True], Rule[FrameStyle, RGBColor[0.6666666666666667`, \
0.33333333333333337`, 0.33333333333333337`]], Rule[FrameTicks, None], \
Rule[PlotRangePadding, None], Rule[ImageSize, Dynamic[List[Automatic, \
Times[1.35`, CurrentValue[\\\"FontCapHeight\\\"], \
Power[AbsoluteCurrentValue[Magnification], -1]]]]]], \
StyleBox[RowBox[List[\\\"RGBColor\\\", \\\"[\\\", RowBox[List[\\\"1.`\\\", \\\
\",\\\", \\\"0.5`\\\", \\\",\\\", \\\"0.5`\\\"]], \\\"]\\\"]], \
Rule[NumberMarks, False]]], Rule[Appearance, None], Rule[BaseStyle, List[]], \
Rule[BaselinePosition, Baseline], Rule[DefaultBaseStyle, List[]], \
RuleDelayed[ButtonFunction, With[List[Set[Typeset`box$, EvaluationBox[]]], \
If[Not[AbsoluteCurrentValue[\\\"Deployed\\\"]], \
CompoundExpression[SelectionMove[Typeset`box$, All, Expression], \
Set[FrontEnd`Private`$ColorSelectorInitialAlpha, 1], \
Set[FrontEnd`Private`$ColorSelectorInitialColor, RGBColor[1.`, 0.5`, 0.5`]], \
Set[FrontEnd`Private`$ColorSelectorUseMakeBoxes, True], \
MathLink`CallFrontEnd[FrontEnd`AttachCell[Typeset`box$, \
FrontEndResource[\\\"RGBColorValueSelector\\\"], List[0, List[Left, Bottom]], \
List[Left, Top], Rule[\\\"ClosingActions\\\", \
List[\\\"SelectionDeparture\\\", \\\"ParentChanged\\\", \
\\\"EvaluatorQuit\\\"]]]]]]]], Rule[BaseStyle, Inherited], Rule[Evaluator, \
Automatic], Rule[Method, \\\"Preemptive\\\"]], RGBColor[1.`, 0.5`, 0.5`], \
Rule[Editable, False], Rule[Selectable, False]]}], \\\"}\\\"}]}], \
\\\"}\\\"}]}]\\) is not a valid Filling specification.\"",2,1555,48,
   32715148735184203895,"Local"},
  "MessageTemplate"]], "Message", "MSG",
 CellChangeTimes->{
  3.8002691222071257`*^9, 3.809677030703689*^9, {3.809689775327434*^9, 
   3.809689779697974*^9}, 3.814005347046734*^9, 3.8141965760780153`*^9, 
   3.8141971327302732`*^9},
 CellLabel->
  "During evaluation of \
In[1555]:=",ExpressionUUID->"9ce418c3-8f1e-43e6-a65d-2853efa2d6bd"],

Cell[BoxData[
 TemplateBox[{
  "Filling","invfillentry",
   "\"\\!\\(\\*RowBox[{\\\"8\\\", \\\"\[Rule]\\\", RowBox[{\\\"{\\\", \
RowBox[{RowBox[{\\\"{\\\", \\\"2\\\", \\\"}\\\"}], \\\",\\\", RowBox[{\\\"{\\\
\", RowBox[{InterpretationBox[ButtonBox[TooltipBox[GraphicsBox[List[List[\
GrayLevel[0], RectangleBox[List[0, 0]]], List[GrayLevel[0], \
RectangleBox[List[1, -1]]], List[RGBColor[1.`, 0.3`, 0.3`], \
RectangleBox[List[0, -1], List[2, 1]]]], Rule[DefaultBaseStyle, \
\\\"ColorSwatchGraphics\\\"], Rule[AspectRatio, 1], Rule[Frame, True], \
Rule[FrameStyle, RGBColor[0.6666666666666667`, 0.2`, 0.2`]], Rule[FrameTicks, \
None], Rule[PlotRangePadding, None], Rule[ImageSize, Dynamic[List[Automatic, \
Times[1.35`, CurrentValue[\\\"FontCapHeight\\\"], \
Power[AbsoluteCurrentValue[Magnification], -1]]]]]], \
StyleBox[RowBox[List[\\\"RGBColor\\\", \\\"[\\\", RowBox[List[\\\"1.`\\\", \\\
\",\\\", \\\"0.3`\\\", \\\",\\\", \\\"0.3`\\\"]], \\\"]\\\"]], \
Rule[NumberMarks, False]]], Rule[Appearance, None], Rule[BaseStyle, List[]], \
Rule[BaselinePosition, Baseline], Rule[DefaultBaseStyle, List[]], \
RuleDelayed[ButtonFunction, With[List[Set[Typeset`box$, EvaluationBox[]]], \
If[Not[AbsoluteCurrentValue[\\\"Deployed\\\"]], \
CompoundExpression[SelectionMove[Typeset`box$, All, Expression], \
Set[FrontEnd`Private`$ColorSelectorInitialAlpha, 1], \
Set[FrontEnd`Private`$ColorSelectorInitialColor, RGBColor[1.`, 0.3`, 0.3`]], \
Set[FrontEnd`Private`$ColorSelectorUseMakeBoxes, True], \
MathLink`CallFrontEnd[FrontEnd`AttachCell[Typeset`box$, \
FrontEndResource[\\\"RGBColorValueSelector\\\"], List[0, List[Left, Bottom]], \
List[Left, Top], Rule[\\\"ClosingActions\\\", \
List[\\\"SelectionDeparture\\\", \\\"ParentChanged\\\", \
\\\"EvaluatorQuit\\\"]]]]]]]], Rule[BaseStyle, Inherited], Rule[Evaluator, \
Automatic], Rule[Method, \\\"Preemptive\\\"]], RGBColor[1.`, 0.3`, 0.3`], \
Rule[Editable, False], Rule[Selectable, False]], \\\",\\\", \
InterpretationBox[ButtonBox[TooltipBox[GraphicsBox[List[List[GrayLevel[0], \
RectangleBox[List[0, 0]]], List[GrayLevel[0], RectangleBox[List[1, -1]]], \
List[RGBColor[1.`, 0.3`, 0.3`], RectangleBox[List[0, -1], List[2, 1]]]], \
Rule[DefaultBaseStyle, \\\"ColorSwatchGraphics\\\"], Rule[AspectRatio, 1], \
Rule[Frame, True], Rule[FrameStyle, RGBColor[0.6666666666666667`, 0.2`, \
0.2`]], Rule[FrameTicks, None], Rule[PlotRangePadding, None], Rule[ImageSize, \
Dynamic[List[Automatic, Times[1.35`, CurrentValue[\\\"FontCapHeight\\\"], \
Power[AbsoluteCurrentValue[Magnification], -1]]]]]], \
StyleBox[RowBox[List[\\\"RGBColor\\\", \\\"[\\\", RowBox[List[\\\"1.`\\\", \\\
\",\\\", \\\"0.3`\\\", \\\",\\\", \\\"0.3`\\\"]], \\\"]\\\"]], \
Rule[NumberMarks, False]]], Rule[Appearance, None], Rule[BaseStyle, List[]], \
Rule[BaselinePosition, Baseline], Rule[DefaultBaseStyle, List[]], \
RuleDelayed[ButtonFunction, With[List[Set[Typeset`box$, EvaluationBox[]]], \
If[Not[AbsoluteCurrentValue[\\\"Deployed\\\"]], \
CompoundExpression[SelectionMove[Typeset`box$, All, Expression], \
Set[FrontEnd`Private`$ColorSelectorInitialAlpha, 1], \
Set[FrontEnd`Private`$ColorSelectorInitialColor, RGBColor[1.`, 0.3`, 0.3`]], \
Set[FrontEnd`Private`$ColorSelectorUseMakeBoxes, True], \
MathLink`CallFrontEnd[FrontEnd`AttachCell[Typeset`box$, \
FrontEndResource[\\\"RGBColorValueSelector\\\"], List[0, List[Left, Bottom]], \
List[Left, Top], Rule[\\\"ClosingActions\\\", \
List[\\\"SelectionDeparture\\\", \\\"ParentChanged\\\", \
\\\"EvaluatorQuit\\\"]]]]]]]], Rule[BaseStyle, Inherited], Rule[Evaluator, \
Automatic], Rule[Method, \\\"Preemptive\\\"]], RGBColor[1.`, 0.3`, 0.3`], \
Rule[Editable, False], Rule[Selectable, False]]}], \\\"}\\\"}]}], \
\\\"}\\\"}]}]\\) is not a valid Filling specification.\"",2,1555,49,
   32715148735184203895,"Local"},
  "MessageTemplate"]], "Message", "MSG",
 CellChangeTimes->{
  3.8002691222071257`*^9, 3.809677030703689*^9, {3.809689775327434*^9, 
   3.809689779697974*^9}, 3.814005347046734*^9, 3.8141965760780153`*^9, 
   3.814197132795782*^9},
 CellLabel->
  "During evaluation of \
In[1555]:=",ExpressionUUID->"c02ce28f-8a7b-40bf-9e74-fb267e325ba5"],

Cell[BoxData[
 TemplateBox[{
  "General","stop",
   "\"Further output of \\!\\(\\*StyleBox[RowBox[{\\\"Filling\\\", \
\\\"::\\\", \\\"invfillentry\\\"}], \\\"MessageName\\\"]\\) will be \
suppressed during this calculation.\"",2,1555,50,32715148735184203895,"Local"},
  "MessageTemplate"]], "Message", "MSG",
 CellChangeTimes->{
  3.8002691222071257`*^9, 3.809677030703689*^9, {3.809689775327434*^9, 
   3.809689779697974*^9}, 3.814005347046734*^9, 3.8141965760780153`*^9, 
   3.8141971328625517`*^9},
 CellLabel->
  "During evaluation of \
In[1555]:=",ExpressionUUID->"7ffb3887-d6c1-4a3b-9fa9-269d87d5e4c3"]
}, Open  ]],

Cell[BoxData[
 RowBox[{
  RowBox[{
  "data", " ", "=", " ", 
   "\"\<[68, 0, 0, 75, 1]\n[56, 2, 0, 83, 0]\n[60, 2, 0, 57, 1]\n[66, 0, 0, \
57, 2]\n[59, 3, 0, 72, 0]\n[60, 1, 0, 63, 1]\n[59, 4, 0, 62, 0]\n[67, 4, 0, \
57, 2]\n[58, 6, 0, 75, 1]\n[72, 5, 1, 59, 2]\n[69, 4, 1, 81, 6]\n[67, 15, 1, \
64, 2]\n[75, 6, 0, 79, 18]\n[70, 22, 1, 67, 1]\n[85, 19, 0, 75, 16]\n[74, 33, \
0, 81, 4]\n[88, 21, 1, 82, 16]\n[78, 43, 6, 78, 7]\n[87, 42, 0, 79, 28]\n[70, \
58, 3, 70, 12]\n[83, 35, 1, 91, 29]\n[72, 44, 2, 75, 14]\n[68, 35, 4, 73, 22]\
\n[70, 44, 6, 60, 13]\n[66, 48, 5, 77, 19]\n[70, 43, 3, 71, 22]\n[69, 41, 5, \
68, 19]\n[74, 38, 4, 86, 25]\n[62, 57, 4, 71, 19]\n[71, 44, 7, 67, 24]\n[60, \
58, 9, 60, 21]\n[58, 66, 4, 72, 33]\n[57, 61, 10, 59, 25]\n[65, 66, 10, 63, \
27]\n[65, 74, 5, 56, 26]\n[80, 68, 8, 67, 27]\n[83, 74, 7, 75, 22]\n[76, 71, \
6, 71, 34]\n[73, 64, 8, 64, 34]\n[69, 68, 17, 72, 30]\n[71, 69, 12, 61, 39]\n\
[61, 80, 8, 59, 27]\n[70, 68, 7, 70, 32]\n[66, 72, 12, 67, 29]\n[70, 67, 3, \
72, 38]\n[65, 70, 11, 64, 21]\n[75, 59, 14, 80, 23]\n[71, 57, 3, 80, 37]\n\
[70, 56, 9, 70, 18]\n[75, 58, 5, 76, 29]\n[76, 71, 12, 69, 15]\n[89, 56, 6, \
89, 43]\n[58, 74, 11, 72, 20]\n[64, 56, 9, 78, 27]\n[59, 64, 12, 61, 25]\n\
[49, 65, 10, 67, 35]\n[43, 65, 10, 65, 33]\n[48, 64, 8, 49, 38]\n[38, 81, 14, \
45, 31]\n[40, 73, 20, 45, 38]\n[41, 72, 16, 49, 33]\n[46, 80, 14, 43, 25]\n\
[54, 98, 14, 52, 22]\n[64, 78, 17, 71, 39]\n[71, 83, 17, 41, 31]\n[53, 113, \
17, 65, 32]\n[72, 86, 24, 68, 50]\n[54, 111, 15, 58, 40]\n[55, 108, 16, 54, \
38]\n[66, 106, 22, 67, 21]\n[81, 75, 10, 95, 49]\n[70, 74, 13, 82, 28]\n[67, \
67, 10, 73, 32]\n[65, 82, 2, 76, 16]\n[86, 51, 2, 84, 40]\n[62, 70, 6, 67, \
17]\n[59, 41, 6, 80, 38]\n[45, 53, 11, 58, 17]\n[59, 48, 1, 52, 32]\n[50, 68, \
6, 49, 20]\n[51, 53, 11, 72, 23]\n[50, 54, 7, 77, 30]\n[40, 51, 8, 67, 26]\n\
[40, 46, 8, 55, 26]\n[38, 55, 9, 41, 29]\n[39, 57, 11, 40, 22]\n[53, 61, 16, \
40, 27]\n[54, 73, 13, 61, 26]\n[67, 67, 19, 57, 25]\n[88, 58, 10, 62, 35]\n\
[67, 85, 12, 52, 29]\n[62, 112, 17, 57, 41]\n[60, 104, 26, 61, 34]\n[72, 99, \
19, 51, 64]\n[49, 113, 34, 41, 25]\n[58, 108, 25, 69, 52]\n[53, 97, 28, 43, \
41]\n[43, 105, 29, 41, 37]\n[63, 93, 31, 43, 39]\n[43, 111, 27, 52, 39]\n[51, \
97, 33, 61, 41]\n[49, 104, 24, 49, 38]\n[44, 97, 23, 71, 43]\n[58, 86, 25, \
54, 33]\n[68, 94, 19, 54, 34]\n[62, 86, 19, 66, 42]\n[65, 100, 19, 55, 33]\n\
[64, 101, 16, 67, 42]\n[72, 99, 12, 75, 36]\n[75, 77, 9, 77, 35]\n[76, 68, \
11, 72, 28]\n[71, 63, 11, 83, 32]\n[76, 73, 3, 62, 33]\n[71, 69, 6, 67, 22]\n\
[79, 72, 11, 85, 24]\n[81, 68, 10, 64, 34]\n[68, 84, 9, 64, 36]\n[66, 83, 13, \
72, 26]\n[84, 71, 7, 60, 27]\n[78, 87, 16, 63, 31]\n[78, 73, 10, 76, 45]\n\
[56, 88, 20, 65, 27]\n[60, 83, 18, 67, 40]\n[59, 81, 12, 49, 31]\n[55, 84, \
10, 70, 32]\n[66, 76, 15, 60, 37]\n[62, 86, 16, 54, 27]\n[71, 79, 12, 74, 35]\
\n[62, 64, 17, 75, 34]\n[60, 72, 9, 64, 40]\n[67, 76, 9, 62, 32]\n[79, 79, 9, \
58, 29]\n[60, 93, 13, 75, 25]\n[71, 71, 13, 98, 41]\n[77, 77, 9, 61, 31]\n\
[80, 79, 12, 86, 30]\n[61, 62, 7, 81, 44]\n[38, 68, 6, 56, 30]\n[53, 60, 14, \
44, 27]\n[56, 63, 12, 54, 18]\n[58, 57, 4, 70, 33]\n[64, 62, 6, 62, 21]\n[68, \
66, 2, 85, 21]\n[82, 37, 6, 86, 25]\n[86, 50, 4, 82, 11]\n[100, 48, 2, 72, \
26]\n[84, 67, 7, 86, 16]\n[86, 51, 5, 82, 34]\n[81, 59, 5, 84, 16]\n[86, 56, \
4, 82, 23]\n[93, 41, 6, 89, 27]\n[86, 52, 5, 83, 18]\n[84, 45, 5, 83, 25]\n\
[77, 51, 5, 90, 15]\n[84, 43, 3, 87, 27]\n[65, 43, 5, 81, 26]\n[72, 51, 1, \
63, 14]\n[74, 46, 2, 81, 20]\n[78, 45, 3, 73, 21]\n[90, 50, 3, 60, 19]\n[79, \
62, 5, 77, 28]\n[66, 65, 4, 77, 20]\n[61, 57, 7, 83, 34]\n[62, 58, 5, 65, 24]\
\n[53, 58, 6, 71, 29]\n[43, 61, 10, 65, 30]\n[42, 64, 8, 47, 32]\n[45, 58, \
22, 45, 23]\n[43, 69, 15, 61, 25]\n[38, 70, 11, 71, 27]\n[59, 59, 11, 47, 31]\
\n[45, 70, 21, 47, 28]\n[57, 71, 20, 55, 38]\n[49, 96, 19, 57, 38]\n[60, 94, \
20, 52, 41]\n[49, 109, 25, 58, 28]\n[70, 87, 19, 63, 43]\n[70, 80, 10, 74, \
36]\n[65, 66, 12, 71, 37]\n[53, 67, 9, 66, 35]\n[58, 76, 9, 58, 23]\n[67, 71, \
13, 77, 32]\n[65, 63, 15, 67, 33]\n[73, 74, 13, 64, 28]\n[76, 80, 13, 94, 27]\
\n[67, 60, 6, 89, 35]\n[67, 59, 9, 58, 20]\n[62, 64, 4, 75, 30]\n[58, 71, 8, \
62, 17]\n[72, 49, 4, 68, 38]\n[59, 61, 4, 62, 21]\n[67, 64, 6, 59, 29]\n[68, \
68, 11, 63, 26]\n[75, 66, 14, 90, 27]\n[77, 59, 9, 71, 44]\n[67, 88, 11, 62, \
20]\n[77, 80, 6, 73, 43]\n[68, 89, 12, 70, 18]\n[73, 88, 8, 71, 38]\n[85, 93, \
18, 83, 28]\>\""}], ";"}]], "Input",
 CellChangeTimes->{{3.795275571284932*^9, 3.795275574163827*^9}, {
   3.795275616678177*^9, 3.795275628122179*^9}, {3.795275688850383*^9, 
   3.795275689702538*^9}, 3.795276909545538*^9},
 CellLabel->
  "In[1556]:=",ExpressionUUID->"2cad1aeb-c8b1-403a-8e2f-5a21b3436931"],

Cell[CellGroupData[{

Cell[BoxData[{
 RowBox[{
  RowBox[{"d", "=", 
   RowBox[{"StringReplace", "[", 
    RowBox[{"data", ",", 
     RowBox[{"{", 
      RowBox[{
       RowBox[{"\"\<[\>\"", "\[Rule]", "\"\<{\>\""}], ",", 
       RowBox[{"\"\<]\>\"", "\[Rule]", "\"\<}\>\""}]}], "}"}]}], "]"}]}], 
  ";"}], "\[IndentingNewLine]", 
 RowBox[{
  RowBox[{"dataFixed", " ", "=", " ", 
   RowBox[{"ToExpression", "[", 
    RowBox[{"StringSplit", "[", 
     RowBox[{
      RowBox[{"StringSplit", "[", 
       RowBox[{"data", ",", "\"\<\\n\>\""}], "]"}], ",", 
      RowBox[{"{", 
       RowBox[{"\"\<,\>\"", ",", "\"\<[\>\"", ",", "\"\<]\>\""}], "}"}]}], 
     "]"}], "]"}]}], ";"}], "\[IndentingNewLine]", 
 RowBox[{
  RowBox[{"polyPlot", " ", "=", " ", 
   RowBox[{"Labeled", "[", 
    RowBox[{
     RowBox[{"Show", "[", 
      RowBox[{
       RowBox[{"ListLinePlot", "[", 
        RowBox[{
         RowBox[{"{", 
          RowBox[{
           RowBox[{"dataFixed", "[", 
            RowBox[{"[", 
             RowBox[{"All", ",", "1"}], "]"}], "]"}], ",", 
           RowBox[{"dataFixed", "[", 
            RowBox[{"[", 
             RowBox[{"All", ",", "2"}], "]"}], "]"}], ",", 
           RowBox[{"dataFixed", "[", 
            RowBox[{"[", 
             RowBox[{"All", ",", "3"}], "]"}], "]"}], ",", 
           RowBox[{"dataFixed", "[", 
            RowBox[{"[", 
             RowBox[{"All", ",", "4"}], "]"}], "]"}], ",", 
           RowBox[{"dataFixed", "[", 
            RowBox[{"[", 
             RowBox[{"All", ",", "5"}], "]"}], "]"}]}], "}"}], ",", 
         RowBox[{"PlotLayout", "\[Rule]", "\"\<Stacked\>\""}], ",", " ", 
         RowBox[{"Filling", "\[Rule]", 
          RowBox[{"{", 
           RowBox[{
            RowBox[{"1", "\[Rule]", 
             RowBox[{"{", 
              RowBox[{"Axis", ",", 
               RowBox[{"Lighter", "[", 
                RowBox[{"Red", ",", "0.7"}], "]"}]}], "}"}]}], ",", 
            RowBox[{"2", "\[Rule]", 
             RowBox[{"{", 
              RowBox[{
               RowBox[{"{", "1", "}"}], ",", 
               RowBox[{"Lighter", "[", 
                RowBox[{"Red", ",", "0.5"}], "]"}]}], "}"}]}], ",", 
            RowBox[{"3", "\[Rule]", 
             RowBox[{"{", 
              RowBox[{
               RowBox[{"{", "2", "}"}], ",", 
               RowBox[{"Lighter", "[", 
                RowBox[{"Red", ",", "0.3"}], "]"}]}], "}"}]}], ",", 
            RowBox[{"4", "\[Rule]", 
             RowBox[{"{", 
              RowBox[{
               RowBox[{"{", "3", "}"}], ",", 
               RowBox[{"Lighter", "[", 
                RowBox[{"Blue", ",", "0.7"}], "]"}]}], "}"}]}], ",", 
            RowBox[{"5", "\[Rule]", 
             RowBox[{"{", 
              RowBox[{
               RowBox[{"{", "4", "}"}], ",", 
               RowBox[{"Lighter", "[", 
                RowBox[{"Blue", ",", "0.5"}], "]"}]}], "}"}]}]}], "}"}]}], 
         ",", "\[IndentingNewLine]", 
         RowBox[{"PlotStyle", "\[Rule]", 
          RowBox[{"{", 
           RowBox[{
            RowBox[{"Lighter", "[", 
             RowBox[{"Red", ",", "0.7"}], "]"}], ",", 
            RowBox[{"Lighter", "[", 
             RowBox[{"Red", ",", "0.5"}], "]"}], ",", 
            RowBox[{"Lighter", "[", 
             RowBox[{"Red", ",", "0.3"}], "]"}], ",", 
            RowBox[{"Lighter", "[", 
             RowBox[{"Blue", ",", "0.7"}], "]"}], ",", 
            RowBox[{"Lighter", "[", 
             RowBox[{"Blue", ",", "0.5"}], "]"}]}], "}"}]}], ",", 
         RowBox[{"TicksStyle", "\[Rule]", "Medium"}]}], "]"}], ",", 
       RowBox[{"Plot", "[", 
        RowBox[{
         RowBox[{"{", "162", "}"}], ",", 
         RowBox[{"{", 
          RowBox[{"x", ",", "0", ",", "200"}], "}"}], ",", " ", 
         RowBox[{"PlotStyle", "\[Rule]", 
          RowBox[{"{", 
           RowBox[{"Dotted", ",", "Black"}], "}"}]}]}], "]"}]}], "]"}], ",", 
     RowBox[{"{", 
      RowBox[{
       RowBox[{
       "Text", "[", 
        "\"\<population size \
(\!\(\*StyleBox[\"N\",FontSlant->\"Italic\"]\))\>\"", "]"}], ",", 
       RowBox[{"Text", "[", "\"\<generations\>\"", "]"}]}], "}"}], ",", 
     RowBox[{"{", 
      RowBox[{"Left", ",", "Bottom"}], "}"}], ",", 
     RowBox[{"RotateLabel", "\[Rule]", "True"}]}], "]"}]}], 
  ";"}], "\[IndentingNewLine]"}], "Input",
 CellChangeTimes->{{3.795275598205471*^9, 3.795275606915362*^9}, {
   3.795277740878084*^9, 3.7952777814093103`*^9}, {3.795278077783441*^9, 
   3.7952781192492743`*^9}, {3.795278149953835*^9, 3.795278211196074*^9}, {
   3.795278522218637*^9, 3.795278660322172*^9}, {3.8096891230398684`*^9, 
   3.8096891262335033`*^9}, {3.809689156902836*^9, 3.809689162857648*^9}, {
   3.809689335540588*^9, 3.8096894301045103`*^9}, {3.80968951616259*^9, 
   3.8096895727455606`*^9}, {3.809689611923848*^9, 3.809689713871299*^9}, 
   3.8097755000766687`*^9},
 CellLabel->
  "In[1557]:=",ExpressionUUID->"7f436d6f-fe8c-4383-9ffd-7a415fb2a786"],

Cell[BoxData[
 TemplateBox[{
  "Filling","invfillentry",
   "\"\\!\\(\\*RowBox[{\\\"6\\\", \\\"\[Rule]\\\", RowBox[{\\\"{\\\", \
RowBox[{\\\"Axis\\\", \\\",\\\", RowBox[{\\\"{\\\", \
RowBox[{InterpretationBox[ButtonBox[TooltipBox[GraphicsBox[List[List[\
GrayLevel[0], RectangleBox[List[0, 0]]], List[GrayLevel[0], \
RectangleBox[List[1, -1]]], List[RGBColor[1.`, 0.7`, 0.7`], \
RectangleBox[List[0, -1], List[2, 1]]]], Rule[DefaultBaseStyle, \
\\\"ColorSwatchGraphics\\\"], Rule[AspectRatio, 1], Rule[Frame, True], \
Rule[FrameStyle, RGBColor[0.6666666666666667`, 0.4666666666666667`, \
0.4666666666666667`]], Rule[FrameTicks, None], Rule[PlotRangePadding, None], \
Rule[ImageSize, Dynamic[List[Automatic, Times[1.35`, \
CurrentValue[\\\"FontCapHeight\\\"], \
Power[AbsoluteCurrentValue[Magnification], -1]]]]]], \
StyleBox[RowBox[List[\\\"RGBColor\\\", \\\"[\\\", RowBox[List[\\\"1.`\\\", \\\
\",\\\", \\\"0.7`\\\", \\\",\\\", \\\"0.7`\\\"]], \\\"]\\\"]], \
Rule[NumberMarks, False]]], Rule[Appearance, None], Rule[BaseStyle, List[]], \
Rule[BaselinePosition, Baseline], Rule[DefaultBaseStyle, List[]], \
RuleDelayed[ButtonFunction, With[List[Set[Typeset`box$, EvaluationBox[]]], \
If[Not[AbsoluteCurrentValue[\\\"Deployed\\\"]], \
CompoundExpression[SelectionMove[Typeset`box$, All, Expression], \
Set[FrontEnd`Private`$ColorSelectorInitialAlpha, 1], \
Set[FrontEnd`Private`$ColorSelectorInitialColor, RGBColor[1.`, 0.7`, 0.7`]], \
Set[FrontEnd`Private`$ColorSelectorUseMakeBoxes, True], \
MathLink`CallFrontEnd[FrontEnd`AttachCell[Typeset`box$, \
FrontEndResource[\\\"RGBColorValueSelector\\\"], List[0, List[Left, Bottom]], \
List[Left, Top], Rule[\\\"ClosingActions\\\", \
List[\\\"SelectionDeparture\\\", \\\"ParentChanged\\\", \
\\\"EvaluatorQuit\\\"]]]]]]]], Rule[BaseStyle, Inherited], Rule[Evaluator, \
Automatic], Rule[Method, \\\"Preemptive\\\"]], RGBColor[1.`, 0.7`, 0.7`], \
Rule[Editable, False], Rule[Selectable, False]], \\\",\\\", \
InterpretationBox[ButtonBox[TooltipBox[GraphicsBox[List[List[GrayLevel[0], \
RectangleBox[List[0, 0]]], List[GrayLevel[0], RectangleBox[List[1, -1]]], \
List[RGBColor[1.`, 0.7`, 0.7`], RectangleBox[List[0, -1], List[2, 1]]]], \
Rule[DefaultBaseStyle, \\\"ColorSwatchGraphics\\\"], Rule[AspectRatio, 1], \
Rule[Frame, True], Rule[FrameStyle, RGBColor[0.6666666666666667`, \
0.4666666666666667`, 0.4666666666666667`]], Rule[FrameTicks, None], \
Rule[PlotRangePadding, None], Rule[ImageSize, Dynamic[List[Automatic, \
Times[1.35`, CurrentValue[\\\"FontCapHeight\\\"], \
Power[AbsoluteCurrentValue[Magnification], -1]]]]]], \
StyleBox[RowBox[List[\\\"RGBColor\\\", \\\"[\\\", RowBox[List[\\\"1.`\\\", \\\
\",\\\", \\\"0.7`\\\", \\\",\\\", \\\"0.7`\\\"]], \\\"]\\\"]], \
Rule[NumberMarks, False]]], Rule[Appearance, None], Rule[BaseStyle, List[]], \
Rule[BaselinePosition, Baseline], Rule[DefaultBaseStyle, List[]], \
RuleDelayed[ButtonFunction, With[List[Set[Typeset`box$, EvaluationBox[]]], \
If[Not[AbsoluteCurrentValue[\\\"Deployed\\\"]], \
CompoundExpression[SelectionMove[Typeset`box$, All, Expression], \
Set[FrontEnd`Private`$ColorSelectorInitialAlpha, 1], \
Set[FrontEnd`Private`$ColorSelectorInitialColor, RGBColor[1.`, 0.7`, 0.7`]], \
Set[FrontEnd`Private`$ColorSelectorUseMakeBoxes, True], \
MathLink`CallFrontEnd[FrontEnd`AttachCell[Typeset`box$, \
FrontEndResource[\\\"RGBColorValueSelector\\\"], List[0, List[Left, Bottom]], \
List[Left, Top], Rule[\\\"ClosingActions\\\", \
List[\\\"SelectionDeparture\\\", \\\"ParentChanged\\\", \
\\\"EvaluatorQuit\\\"]]]]]]]], Rule[BaseStyle, Inherited], Rule[Evaluator, \
Automatic], Rule[Method, \\\"Preemptive\\\"]], RGBColor[1.`, 0.7`, 0.7`], \
Rule[Editable, False], Rule[Selectable, False]]}], \\\"}\\\"}]}], \
\\\"}\\\"}]}]\\) is not a valid Filling specification.\"",2,1559,51,
   32715148735184203895,"Local"},
  "MessageTemplate"]], "Message", "MSG",
 CellChangeTimes->{{3.8096896852133512`*^9, 3.8096897147192917`*^9}, 
   3.814005348507642*^9, 3.8141965766992683`*^9, 3.814197133328616*^9},
 CellLabel->
  "During evaluation of \
In[1557]:=",ExpressionUUID->"0d9cdaff-73ae-4ddd-a287-45fe2a537dad"],

Cell[BoxData[
 TemplateBox[{
  "Filling","invfillentry",
   "\"\\!\\(\\*RowBox[{\\\"7\\\", \\\"\[Rule]\\\", RowBox[{\\\"{\\\", \
RowBox[{RowBox[{\\\"{\\\", \\\"1\\\", \\\"}\\\"}], \\\",\\\", RowBox[{\\\"{\\\
\", RowBox[{InterpretationBox[ButtonBox[TooltipBox[GraphicsBox[List[List[\
GrayLevel[0], RectangleBox[List[0, 0]]], List[GrayLevel[0], \
RectangleBox[List[1, -1]]], List[RGBColor[1.`, 0.5`, 0.5`], \
RectangleBox[List[0, -1], List[2, 1]]]], Rule[DefaultBaseStyle, \
\\\"ColorSwatchGraphics\\\"], Rule[AspectRatio, 1], Rule[Frame, True], \
Rule[FrameStyle, RGBColor[0.6666666666666667`, 0.33333333333333337`, \
0.33333333333333337`]], Rule[FrameTicks, None], Rule[PlotRangePadding, None], \
Rule[ImageSize, Dynamic[List[Automatic, Times[1.35`, \
CurrentValue[\\\"FontCapHeight\\\"], \
Power[AbsoluteCurrentValue[Magnification], -1]]]]]], \
StyleBox[RowBox[List[\\\"RGBColor\\\", \\\"[\\\", RowBox[List[\\\"1.`\\\", \\\
\",\\\", \\\"0.5`\\\", \\\",\\\", \\\"0.5`\\\"]], \\\"]\\\"]], \
Rule[NumberMarks, False]]], Rule[Appearance, None], Rule[BaseStyle, List[]], \
Rule[BaselinePosition, Baseline], Rule[DefaultBaseStyle, List[]], \
RuleDelayed[ButtonFunction, With[List[Set[Typeset`box$, EvaluationBox[]]], \
If[Not[AbsoluteCurrentValue[\\\"Deployed\\\"]], \
CompoundExpression[SelectionMove[Typeset`box$, All, Expression], \
Set[FrontEnd`Private`$ColorSelectorInitialAlpha, 1], \
Set[FrontEnd`Private`$ColorSelectorInitialColor, RGBColor[1.`, 0.5`, 0.5`]], \
Set[FrontEnd`Private`$ColorSelectorUseMakeBoxes, True], \
MathLink`CallFrontEnd[FrontEnd`AttachCell[Typeset`box$, \
FrontEndResource[\\\"RGBColorValueSelector\\\"], List[0, List[Left, Bottom]], \
List[Left, Top], Rule[\\\"ClosingActions\\\", \
List[\\\"SelectionDeparture\\\", \\\"ParentChanged\\\", \
\\\"EvaluatorQuit\\\"]]]]]]]], Rule[BaseStyle, Inherited], Rule[Evaluator, \
Automatic], Rule[Method, \\\"Preemptive\\\"]], RGBColor[1.`, 0.5`, 0.5`], \
Rule[Editable, False], Rule[Selectable, False]], \\\",\\\", \
InterpretationBox[ButtonBox[TooltipBox[GraphicsBox[List[List[GrayLevel[0], \
RectangleBox[List[0, 0]]], List[GrayLevel[0], RectangleBox[List[1, -1]]], \
List[RGBColor[1.`, 0.5`, 0.5`], RectangleBox[List[0, -1], List[2, 1]]]], \
Rule[DefaultBaseStyle, \\\"ColorSwatchGraphics\\\"], Rule[AspectRatio, 1], \
Rule[Frame, True], Rule[FrameStyle, RGBColor[0.6666666666666667`, \
0.33333333333333337`, 0.33333333333333337`]], Rule[FrameTicks, None], \
Rule[PlotRangePadding, None], Rule[ImageSize, Dynamic[List[Automatic, \
Times[1.35`, CurrentValue[\\\"FontCapHeight\\\"], \
Power[AbsoluteCurrentValue[Magnification], -1]]]]]], \
StyleBox[RowBox[List[\\\"RGBColor\\\", \\\"[\\\", RowBox[List[\\\"1.`\\\", \\\
\",\\\", \\\"0.5`\\\", \\\",\\\", \\\"0.5`\\\"]], \\\"]\\\"]], \
Rule[NumberMarks, False]]], Rule[Appearance, None], Rule[BaseStyle, List[]], \
Rule[BaselinePosition, Baseline], Rule[DefaultBaseStyle, List[]], \
RuleDelayed[ButtonFunction, With[List[Set[Typeset`box$, EvaluationBox[]]], \
If[Not[AbsoluteCurrentValue[\\\"Deployed\\\"]], \
CompoundExpression[SelectionMove[Typeset`box$, All, Expression], \
Set[FrontEnd`Private`$ColorSelectorInitialAlpha, 1], \
Set[FrontEnd`Private`$ColorSelectorInitialColor, RGBColor[1.`, 0.5`, 0.5`]], \
Set[FrontEnd`Private`$ColorSelectorUseMakeBoxes, True], \
MathLink`CallFrontEnd[FrontEnd`AttachCell[Typeset`box$, \
FrontEndResource[\\\"RGBColorValueSelector\\\"], List[0, List[Left, Bottom]], \
List[Left, Top], Rule[\\\"ClosingActions\\\", \
List[\\\"SelectionDeparture\\\", \\\"ParentChanged\\\", \
\\\"EvaluatorQuit\\\"]]]]]]]], Rule[BaseStyle, Inherited], Rule[Evaluator, \
Automatic], Rule[Method, \\\"Preemptive\\\"]], RGBColor[1.`, 0.5`, 0.5`], \
Rule[Editable, False], Rule[Selectable, False]]}], \\\"}\\\"}]}], \
\\\"}\\\"}]}]\\) is not a valid Filling specification.\"",2,1559,52,
   32715148735184203895,"Local"},
  "MessageTemplate"]], "Message", "MSG",
 CellChangeTimes->{{3.8096896852133512`*^9, 3.8096897147192917`*^9}, 
   3.814005348507642*^9, 3.8141965766992683`*^9, 3.814197133392158*^9},
 CellLabel->
  "During evaluation of \
In[1557]:=",ExpressionUUID->"d3803404-8a16-41f0-b819-b3fb923b1689"],

Cell[BoxData[
 TemplateBox[{
  "Filling","invfillentry",
   "\"\\!\\(\\*RowBox[{\\\"8\\\", \\\"\[Rule]\\\", RowBox[{\\\"{\\\", \
RowBox[{RowBox[{\\\"{\\\", \\\"2\\\", \\\"}\\\"}], \\\",\\\", RowBox[{\\\"{\\\
\", RowBox[{InterpretationBox[ButtonBox[TooltipBox[GraphicsBox[List[List[\
GrayLevel[0], RectangleBox[List[0, 0]]], List[GrayLevel[0], \
RectangleBox[List[1, -1]]], List[RGBColor[1.`, 0.3`, 0.3`], \
RectangleBox[List[0, -1], List[2, 1]]]], Rule[DefaultBaseStyle, \
\\\"ColorSwatchGraphics\\\"], Rule[AspectRatio, 1], Rule[Frame, True], \
Rule[FrameStyle, RGBColor[0.6666666666666667`, 0.2`, 0.2`]], Rule[FrameTicks, \
None], Rule[PlotRangePadding, None], Rule[ImageSize, Dynamic[List[Automatic, \
Times[1.35`, CurrentValue[\\\"FontCapHeight\\\"], \
Power[AbsoluteCurrentValue[Magnification], -1]]]]]], \
StyleBox[RowBox[List[\\\"RGBColor\\\", \\\"[\\\", RowBox[List[\\\"1.`\\\", \\\
\",\\\", \\\"0.3`\\\", \\\",\\\", \\\"0.3`\\\"]], \\\"]\\\"]], \
Rule[NumberMarks, False]]], Rule[Appearance, None], Rule[BaseStyle, List[]], \
Rule[BaselinePosition, Baseline], Rule[DefaultBaseStyle, List[]], \
RuleDelayed[ButtonFunction, With[List[Set[Typeset`box$, EvaluationBox[]]], \
If[Not[AbsoluteCurrentValue[\\\"Deployed\\\"]], \
CompoundExpression[SelectionMove[Typeset`box$, All, Expression], \
Set[FrontEnd`Private`$ColorSelectorInitialAlpha, 1], \
Set[FrontEnd`Private`$ColorSelectorInitialColor, RGBColor[1.`, 0.3`, 0.3`]], \
Set[FrontEnd`Private`$ColorSelectorUseMakeBoxes, True], \
MathLink`CallFrontEnd[FrontEnd`AttachCell[Typeset`box$, \
FrontEndResource[\\\"RGBColorValueSelector\\\"], List[0, List[Left, Bottom]], \
List[Left, Top], Rule[\\\"ClosingActions\\\", \
List[\\\"SelectionDeparture\\\", \\\"ParentChanged\\\", \
\\\"EvaluatorQuit\\\"]]]]]]]], Rule[BaseStyle, Inherited], Rule[Evaluator, \
Automatic], Rule[Method, \\\"Preemptive\\\"]], RGBColor[1.`, 0.3`, 0.3`], \
Rule[Editable, False], Rule[Selectable, False]], \\\",\\\", \
InterpretationBox[ButtonBox[TooltipBox[GraphicsBox[List[List[GrayLevel[0], \
RectangleBox[List[0, 0]]], List[GrayLevel[0], RectangleBox[List[1, -1]]], \
List[RGBColor[1.`, 0.3`, 0.3`], RectangleBox[List[0, -1], List[2, 1]]]], \
Rule[DefaultBaseStyle, \\\"ColorSwatchGraphics\\\"], Rule[AspectRatio, 1], \
Rule[Frame, True], Rule[FrameStyle, RGBColor[0.6666666666666667`, 0.2`, \
0.2`]], Rule[FrameTicks, None], Rule[PlotRangePadding, None], Rule[ImageSize, \
Dynamic[List[Automatic, Times[1.35`, CurrentValue[\\\"FontCapHeight\\\"], \
Power[AbsoluteCurrentValue[Magnification], -1]]]]]], \
StyleBox[RowBox[List[\\\"RGBColor\\\", \\\"[\\\", RowBox[List[\\\"1.`\\\", \\\
\",\\\", \\\"0.3`\\\", \\\",\\\", \\\"0.3`\\\"]], \\\"]\\\"]], \
Rule[NumberMarks, False]]], Rule[Appearance, None], Rule[BaseStyle, List[]], \
Rule[BaselinePosition, Baseline], Rule[DefaultBaseStyle, List[]], \
RuleDelayed[ButtonFunction, With[List[Set[Typeset`box$, EvaluationBox[]]], \
If[Not[AbsoluteCurrentValue[\\\"Deployed\\\"]], \
CompoundExpression[SelectionMove[Typeset`box$, All, Expression], \
Set[FrontEnd`Private`$ColorSelectorInitialAlpha, 1], \
Set[FrontEnd`Private`$ColorSelectorInitialColor, RGBColor[1.`, 0.3`, 0.3`]], \
Set[FrontEnd`Private`$ColorSelectorUseMakeBoxes, True], \
MathLink`CallFrontEnd[FrontEnd`AttachCell[Typeset`box$, \
FrontEndResource[\\\"RGBColorValueSelector\\\"], List[0, List[Left, Bottom]], \
List[Left, Top], Rule[\\\"ClosingActions\\\", \
List[\\\"SelectionDeparture\\\", \\\"ParentChanged\\\", \
\\\"EvaluatorQuit\\\"]]]]]]]], Rule[BaseStyle, Inherited], Rule[Evaluator, \
Automatic], Rule[Method, \\\"Preemptive\\\"]], RGBColor[1.`, 0.3`, 0.3`], \
Rule[Editable, False], Rule[Selectable, False]]}], \\\"}\\\"}]}], \
\\\"}\\\"}]}]\\) is not a valid Filling specification.\"",2,1559,53,
   32715148735184203895,"Local"},
  "MessageTemplate"]], "Message", "MSG",
 CellChangeTimes->{{3.8096896852133512`*^9, 3.8096897147192917`*^9}, 
   3.814005348507642*^9, 3.8141965766992683`*^9, 3.814197133465164*^9},
 CellLabel->
  "During evaluation of \
In[1557]:=",ExpressionUUID->"e53813d8-9ef9-4148-9473-fe5cba18a1af"],

Cell[BoxData[
 TemplateBox[{
  "General","stop",
   "\"Further output of \\!\\(\\*StyleBox[RowBox[{\\\"Filling\\\", \
\\\"::\\\", \\\"invfillentry\\\"}], \\\"MessageName\\\"]\\) will be \
suppressed during this calculation.\"",2,1559,54,32715148735184203895,"Local"},
  "MessageTemplate"]], "Message", "MSG",
 CellChangeTimes->{{3.8096896852133512`*^9, 3.8096897147192917`*^9}, 
   3.814005348507642*^9, 3.8141965766992683`*^9, 3.814197133536414*^9},
 CellLabel->
  "During evaluation of \
In[1557]:=",ExpressionUUID->"5e34575e-b1bc-4c6b-af71-4b25d0d82dd8"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"Style", "[", "\[IndentingNewLine]", 
  RowBox[{
   RowBox[{"Grid", "[", 
    RowBox[{
     RowBox[{"{", "\[IndentingNewLine]", 
      RowBox[{
       RowBox[{"{", 
        RowBox[{"Labeled", "[", 
         RowBox[{"WTplot", ",", 
          RowBox[{"{", 
           RowBox[{"Text", "[", 
            RowBox[{"Style", "[", 
             RowBox[{"\"\<A\>\"", ",", "Bold", ",", 
              RowBox[{"FontSize", "\[Rule]", "20"}]}], "]"}], "]"}], "}"}], 
          ",", 
          RowBox[{"{", 
           RowBox[{"{", 
            RowBox[{"Top", ",", "Left"}], "}"}], "}"}]}], "]"}], "}"}], ",", 
       RowBox[{"{", 
        RowBox[{"Labeled", "[", 
         RowBox[{"extPlot", ",", 
          RowBox[{"{", 
           RowBox[{"Text", "[", 
            RowBox[{"Style", "[", 
             RowBox[{"\"\<B\>\"", ",", "Bold", ",", 
              RowBox[{"FontSize", "\[Rule]", "20"}]}], "]"}], "]"}], "}"}], 
          ",", 
          RowBox[{"{", 
           RowBox[{"{", 
            RowBox[{"Top", ",", "Left"}], "}"}], "}"}]}], "]"}], "}"}], ",", 
       RowBox[{"{", 
        RowBox[{"Labeled", "[", 
         RowBox[{"polyPlot", ",", 
          RowBox[{"{", 
           RowBox[{"Text", "[", 
            RowBox[{"Style", "[", 
             RowBox[{"\"\<C\>\"", ",", "Bold", ",", 
              RowBox[{"FontSize", "\[Rule]", "20"}]}], "]"}], "]"}], "}"}], 
          ",", 
          RowBox[{"{", 
           RowBox[{"{", 
            RowBox[{"Top", ",", "Left"}], "}"}], "}"}]}], "]"}], "}"}], ",", 
       RowBox[{"{", 
        RowBox[{"SwatchLegend", "[", 
         RowBox[{
          RowBox[{"{", 
           RowBox[{
            RowBox[{"Lighter", "[", 
             RowBox[{"Red", ",", "0.7"}], "]"}], ",", 
            RowBox[{"Lighter", "[", 
             RowBox[{"Red", ",", "0.5"}], "]"}], ",", 
            RowBox[{"Lighter", "[", 
             RowBox[{"Red", ",", "0.3"}], "]"}], ",", 
            RowBox[{"Lighter", "[", 
             RowBox[{"Blue", ",", "0.7"}], "]"}], ",", 
            RowBox[{"Lighter", "[", 
             RowBox[{"Blue", ",", "0.5"}], "]"}]}], "}"}], ",", 
          RowBox[{"{", 
           RowBox[{
            RowBox[{"Text", "[", "\"\<XX\>\"", "]"}], ",", 
            RowBox[{
            "Text", "[", "\"\<\!\(\*SubscriptBox[\(X\), \(d\)]\)X\>\"", "]"}],
             ",", 
            RowBox[{
            "Text", "[", 
             "\"\<\!\(\*SubscriptBox[\(X\), \(d\)]\)\!\(\*SubscriptBox[\(X\), \
\(d\)]\)\>\"", "]"}], ",", 
            RowBox[{"Text", "[", "\"\<XY\>\"", "]"}], ",", 
            RowBox[{
            "Text", "[", "\"\<\!\(\*SubscriptBox[\(X\), \(d\)]\)Y\>\"", 
             "]"}]}], "}"}], ",", 
          RowBox[{"LegendMarkerSize", "\[Rule]", "24"}], ",", 
          RowBox[{"LabelStyle", "\[Rule]", 
           RowBox[{"{", 
            RowBox[{"FontSize", "\[Rule]", "14"}], "}"}]}], ",", 
          RowBox[{"LegendLayout", "\[Rule]", 
           RowBox[{"{", 
            RowBox[{"\"\<Row\>\"", ",", "1"}], "}"}]}], ",", 
          RowBox[{"LegendMargins", "\[Rule]", 
           RowBox[{"{", 
            RowBox[{
             RowBox[{"{", 
              RowBox[{"48", ",", "0"}], "}"}], ",", 
             RowBox[{"{", 
              RowBox[{"10", ",", "10"}], "}"}]}], "}"}]}]}], "]"}], "}"}]}], 
      "}"}], ",", "\[IndentingNewLine]", 
     RowBox[{"Alignment", "\[Rule]", 
      RowBox[{"{", 
       RowBox[{
        RowBox[{"{", "Center", "}"}], ",", 
        RowBox[{"{", "Center", "}"}], ",", 
        RowBox[{"{", "Center", "}"}], ",", 
        RowBox[{"{", "Right", "}"}]}], "}"}]}]}], "\[IndentingNewLine]", 
    "]"}], ",", 
   RowBox[{"ImageSizeMultipliers", "\[Rule]", 
    RowBox[{"{", 
     RowBox[{"1", ",", " ", "1"}], "}"}]}]}], "]"}]], "Input",
 CellChangeTimes->CompressedData["
1:eJwdyEsowwEAx/E1j6i12kHWIqHWVrZWduGAeTW1EPJYWkzbWBNjklFykUIm
p9kOHjssm7VamdrYP1sO+yvyWmZlTrItTSGXxf/3P3z71LdSPdWjYTIYjDIq
uJnd/nlvz8hYDg8tQVoZKUpDo7IA6gVnHNgb13Ihz3Zng/fNb7TWgPAQxl9Z
Hrg/J/LB6gnyEhJrNSQ8D+7EIF+0+ww3DB8J2ClVfcHPwVFar7cjB30visI0
pWRvqYT2QVwOi7LCEXN3RmazxDSQmVelhze81kmoTCpmofN2eB5G/BGXTp2R
JX+dbiioj3mhZZHjg7Zo7gQq2rr80G5eCEJuUeQCRlMD15A/Jn6ER8UO6Thl
fihdB73T6w1QK2XLoGMl1A9XjacqSBjYOpiQRwzwgFAa4Uy4xQSvwsdm2CSR
L0Mmx70Fh0qDLvpX2AOw7++bgGG5hISmUO0T/AeTy+q4
  "],
 CellLabel->
  "In[1560]:=",ExpressionUUID->"2168ced3-f738-483f-bf38-1fe97cb7500d"],

Cell[BoxData[
 StyleBox[
  TagBox[GridBox[{
     {
      TemplateBox[{TemplateBox[{
          GraphicsBox[{{{{}, 
              GraphicsComplexBox[CompressedData["
1:eJztnUuIpFcVx8v4IDSCiaQlgYhtCEGKEPKavJO5ec1kep4909Xvx1fVXY/p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               "], {{{}, {
                  EdgeForm[], 
                  RGBColor[1., 0.7, 0.7], 
                  GraphicsGroupBox[{
                    PolygonBox[CompressedData["
1:eJwl1WV0FgQUBuBv9KgRG12jWdDd3bmxDTYYMLo2umN0ty0YYLekAQomKA1K
mKCkgoJBx7PDj+e85/6/977hKWkxqZkCgUAQa4Ie5f7sgcABVhlWsoLlLGMp
S1jMIhaygPnMYy5zSGc2s5jJDKYzjalMYTKTmMgExjOOsYwhjVRGM4qRjGA4
wxjKEAYziIGkMID+9COZvvQhiUR604sE4omjJ7HE0IPudKMrXehMJzrSgfa0
oy1taE0rWtKC5jSjKU1oTCMa0oD61KMudahNLWpSg+pUoyrRRBFJBFWoTCUq
UoHylKMs4ZShNKUoSQmKU4yiFKEwhQgjlIIUID/5CCEvechNLnISTA6yk42s
ZCEzmQgKerRQD8R97nGXO9zmFje5wf/8x7/8w3Wu8Td/cZUr/MkfXOYSF7nA
ec7xO79xljP8yi/8zE/8yA+c5hQnOcH3fMdxjnGUIxzmEAc5kHETfMs37GMv
X/MVX/IFn/MZe9jNp3zCLnbyMR/xIR+wg+1sYytb2Mz7vMe7vMPbvMWbvMHr
vMarvMLLvMQmNvIiL/A8z7GB9TzLMzzNUzzJEzzOY6xjbcYfYDWrWMkKlrOM
pSxhMYtYyALmM4+5zCGd2cxiJjOYzjSmMoXJTGIiExjPOMYyhjRSGc0oRjKC
4QxjKEMYzCAGksIA+tOPZPrShyQS6U0vEognjp7EEkMPutONrnShM53oSAfa
0462tKE1rWhJC5rTjKY0oTGNaEgD6lOPutShNrWoSQ2qU42qRBNFJBFUoTKV
qEgFylOOsoRThtKUoiQlKE4xilKEwhQijFAKUoD85COEvOQhN7nISTA5UBGB
bGQlC5nJ6JKHcvmsNw==
                    "]]}]}, {}, {}, {}, {}, {}, {}, {}, {}, {}, {}, {}, {
                  EdgeForm[], 
                  RGBColor[1., 0.5, 0.5], 
                  GraphicsGroupBox[{
                    PolygonBox[CompressedData["
1:eJwl1kPUGFgMBtDatm3b9tS2bdv2oO3Utm3btm3bmvufWdyTl1VWeV8SN+5Q
uX2QQIECBSZB0P9rQO8ZKBjBCUFIQhGaMIQlHOGJQEQiEZkoRCUa0YlBTGIR
mzjEJR7xA+aSkEQkJglJSUZyUpCSVKQmDWlJR3oykJFMZCYLWclGdnKQk1zk
Jg95yUd+ClCQQhSmCEUpRnFKUJJS/EFpylCWcpSnAhWpRGWqUJVqVKcGNalF
bepQl3rUpwENaURjmtCUZjSnBS1pRWva0JZ2tKcDHelEZ7rQlW50pwc96UVv
+tCXfvRnAAMZxGCGMJRhDGcEIxnFaMYwlj/5i7/5h3GMZwL/MpFJTGYKU5nG
dGYwk1nMZg5zmcd8FrCQRSxmCUtZxnJWsJJVrGYNa1nHejawkU1sZgtb2cZ2
drCTXexmD3vZx34OcJBDHOYIRznGcU5wklOc5gxnOcd5LnCRS1zmCle5xnVu
cJNb3OYOd7nHfR7wkEc85glPecZzXvCSV7zmDW95x3s+8JFPfOYLX/nGd37w
k1/8JmD5AxOEoAQjOCEISShCE4awhCM8EYhIJCIThahEIzoxiEksYhOHuMQj
PglISCISk4SkJCM5KUhJKlKThrSkIz0ZyEgmMpOFrGQjOznISS5yk4e85CM/
BShIIQpThKIUozglKEkp/qA0ZShLOcpTgYpUojJVqEo1qlODmtSiNnWoSz3q
04CGNKIxTWhKM5rTgpa0ojVtaEs72tOBjnSiM13oSje604Oe9KI3fehLP/oz
gIEMYjBDGMowhjOCkYxiNGMYy5/8xd+B/8+P+MQjLnGITSxiEoPoRCMqUYhM
JCISgfCEIyxhCE0oQhKC4AQjKEEC8iogqPgttH7xkx985xtf+cJnPvGRD7zn
HW95w2te8ZIXPOcZT3nCYx7xkAfc5x53ucNtbnGTG1znGle5wmUucZELnOcc
ZznDaU5xkhMc5xhHOcJhDnGQA+xnH3vZw252sZMdbGcbW9nCZjaxkQ2sZx1r
WcNqVrGSFSxnGUtZwmIWsZAFzGcec5nDbGYxkxlMZxpTmcJkJjGRf5nAeMbx
D3/zF38yljGMZhQjGcFwhjGUIQxmEAMZQH/60Zc+9KYXPelBd7rRlS50phMd
6UDAIdSOtrShNa1oSQua04ymNKExjWhIA+pTj7rUoTa1qEkNqlONqlShMpWo
SAXKU46ylKE0f1CKkpSgOMUoShEKU4iCFCA/+chLHnKTi5zkIDvZyEoWMpOJ
jGQgPelISxpSk4qUpCA5yUhKEhKTiIQkID7xiEscYhOLmMQgOtGIShQiE4mI
RCA84QhLGEITipCEIDjBCEoQAhNwsP72//ziJz/4zje+8oXPfOIjH3jPO97y
hte84iUveM4znvKExzziIQ+4zz3ucofb3OImN7jONa5yhctc4iIXOM85znKG
05ziJCc4zjGOcoTDHOIgB9jPPvayh93sYic72M42trKFzWxiIxtYzzrWsobV
rGIlK1jOMpayhMUsYiELmM885jKH2cxiJjOYzjSmMoXJTGIi/zKB8YzjH/4D
/6uFlA==
                    "]]}]}, {}, {}, {}, {}, {}, {}, {}, {}, {}, {
                  EdgeForm[], 
                  RGBColor[1., 0.3, 0.3], 
                  GraphicsGroupBox[{
                    PolygonBox[CompressedData["
1:eJwl1VOwGAgSBdAX27Y1sZ1MbCcT27Y9sW3btm3btm3tmdqPU9X/t293ogZt
K7YJHBAQEIj4QQICxhrGMZ4JTGQSk5nCVKYxnRnMZBazmcNc5jGfBSxkEYtZ
wlKWsZwVrGQVq1nDWtaxng1sZBOb2cJWtrGdHexkF7vZw172sZ8DHOQQhznC
UY5xnBOc5BSnOcNZznGeC1zkEpe5wlWucZ0b3OQWt7nDXe5xnwc85BGPecJT
nvGcF7zkFa95w1ve8Z4PfOQTn/nCV77xnR/85Be/+fNfiMIMRGCCEJRgBCcE
IQlFaMIQlnCEJwIRiURkohCVaEQnBjGJRWziEJd4xCcBCUlEYpKQlGQkJwUp
ScVfpCYNaUlHejKQkUxkJgtZyUZ2cpCTXOQmD3nJR37+pgAFKURhilCUYhSn
BCUpRWnKUJZylKcCFalEZarwD1WpRnVqUJNa1KYOdalHfRrQkEY0pglNaUZz
WtCSVrTmvxK2pR3t6UBHOtGZLnSlG93pQU960Zs+9KUf/fmXAQxkEIMZwlCG
MZwRjGQUoxnDWMYxnglMZBKTmcJUpjGdGcxkFrOZw1zmMZ8FLGQRi1nCUpax
nBWsZBWrWcNa1rGeDWxkE5vZwla2sZ0d7GQXu9nDXvaxnwMc5BCHOcJRjnGc
E5zkFKc5w1nOcZ4LXOQSl7nCVa5xnRvc5Ba3ucNd7nGfBzzkEY95wlOe8ZwX
vOQVr3nDW97xng985BOf+cJXvvGdH/zkF7/5Q4AbHojABCEowQhOCEISitCE
ISzhCE8EIhKJyEQhKtGITgxiEovYxCEu8YL8/3esCipPVrCcZSxlCYtZxEIW
MJ95zGUOs5nFTGYwnWlMZQqTmcREJjCecYxlDKMZxUhGMJxhDGUIgxnEQAbw
L/3pR1/60Jte9KQH3elGV7rQmU50pAPtaUdb2tCaVrSkBc1pRlOa0JhGNKQB
9alHXepQm1rUpAbVqUZV/qEKlalERSpQnnKUpQylKUVJSlCcYhSlCIUpREEK
8Df5yUde8pCbXOQkB9nJRlaykJlMZCQD6UlHWtKQmr9IRUpSkJxkJCUJiUlE
QhIQn3jEJQ6xiUVMYhCdaEQlCpGJREQiEJ5whCUMoQlFSEIQnGAEJQiBCUQA
f+zub37xkx985xtf+cJnPvGRD7znHW95w2te8ZIXPOcZT3nCYx7xkAfc5x53
ucNtbnGTG1znGle5wmUucZELnOccZznDaU5xkhMc5xhHOcJhDnGQA+xnH3vZ
w252sZMdbGcbW9nCZjaxkQ2sZx1rWcPq//rPSlawnGUsZQmLWcRCFjCfecxl
DrOZxUxmMJ1pTGUKk5nERCYwnnGMZQyjGcVIRjCcYQxlCIMZxEAG8C/96Udf
+tCbXvSkB93pRle60JlOdKQD7WlHW9rQmla0pAXNaUZTmtCYRjSkAfWpR13q
UJta1KQG1alGVf6hCpWpREUqUJ5ylKUMpSlFSUpQnGIUpQiFKURBCvA3+clH
XvKQm1zkJAfZyUZWspCZTGQkA+lJR1rSkJq/SEVKUpCcZCQlCYlJREIS8D+K
ipsh
                    "]]}]}, {}, {}, {}, {}, {}, {}, {}, {
                  EdgeForm[], 
                  RGBColor[0.7, 0.7, 1.], 
                  GraphicsGroupBox[{
                    PolygonBox[CompressedData["
1:eJwV1MO2IAgCRMFXtm3btm3btm12l23btm3btm11TcwivuCezESNO1RuHzgg
ICAQ8YMEBCQgIYlITBKSkozkpCAlqUhNGtKSjvRkICOZyEwWspKN7OQgJ7nI
TR7yko/8FKAghShMEYpSjOKUoCSlKE0ZylKO8lSgIpWoTBWqUo3q1KAmtahN
HepSj/o0oCGNaEwTmtKM5rSgJa1oTRva0o72dKAjnehMF7rSje70oCe96E0f
+tKP/gxgIIMYzBCG8g//MozhjGAkoxjNGMYyjvFMYCKTmMwUpjKN6cxgJrOY
zRzmMo/5LGAhi1jMEpayjOWsYCWrWM0a1rKO9WxgI5vYzBa2so3t7GAnu9jN
Hvayj/0c4CCHOMwRjnKM45zgJKc4zRnOco7zXOAil7jMFa5yjevc4Ca3uM0d
7nKP+zzgIY94zBOe8oznvOAlr3jNG97yjvd84COf+MwXvvKN7/zgJ7/4zR/+
4y8BQe2RwAQhKMEITghCEorQhCEs4QhPBCISichEISrRiE4MYhKL2MQhLvGI
TwISkojEJCEpyUhOClKSitSkIS3pSE8GMpKJzGQhK9nITg5ykovc5CEv+chP
AQpSiMIUoSjFKE4JSlKK0pShLOUoTwUqUonKVKEq1ahODWpSi9rUoS71qE8D
GtKIxjShKc1oTgta0orWtKEt7WhPBzrSic50oSvd6E4PetKL3vShL/3ozwAG
MojBDGEo//AvwxjOCEYyitGMYSzjGM8EJjKJyUxhKtOYzgxmMovZzGEu85jP
AhayiMUsYSnLWM4KVrKK1az5f8fgepKbXOQkB9nJRlaykJlMZCQD6UlHWtKQ
mlSkJAXJSUZSkpCYRCQkAfGJR1ziEJtYxCQG0YlGVKIQmUhEJALhCUdYwhCa
UIQkBMEJRlCCEJhABPA3mN3yh9/84ic/+M43vvKFz3ziIx94zzve8obXvOIl
L3jOM57yhMc84iEPuM897nKH29ziJje4zjWucoXLXOIiFzjPOc5yhtOc4iQn
OM4xjnKEwxziIAfYzz72sofd7GInO9jONrayhc1sYiMbWM861rKG1axiJStY
zjKWsoTFLGIhC5jPPOYyh9nMYiYzmM40pjKFyUxiIhMYzzjGMobRjGIkIxjO
MP7lH4YyhMEMYiAD6E8/+tKH3vSiJz3oTje60oXOdKIjHWhPO9rShta0oiUt
aE4zmtKExjSiIQ2oTz3qUofa1KImNahONapShcpUoiIVKE85ylKG0pSiJCUo
TjGKUoTCFKIgBchPPvKSh9zkIic5yE42spKFzGQiIxlITzrSkobUpCIlKUhO
MpKShMQkIiEJiE884hKH2MQiJjGITjSiEoXIRCIiEQhPOMIShtCEIiQhCE4w
ghKEwAQigL9+7z/+8Jtf/OQH3/nGV77wmU985APvecdb3vCaV7zkBc95xlOe
8JhHPOQB97nHXe5wm1vc5AbXucZVrnCZS1zkAuc5x1nOcJpTnOQExznGUY5w
mEMc5AD72cde9rCbXexkB9vZxla2sJlNbGQD61nHWv4HwvSSzA==
                    "]]}]}, {}, {}, {}, {}, {}, {
                  EdgeForm[], 
                  RGBColor[0.5, 0.5, 1.], 
                  GraphicsGroupBox[{
                    PolygonBox[CompressedData["
1:eJwV08O2IAgCRMHXZdu2baPLtm3bdpdt27Zt27ZtT8wifiDz3IQN21VsGygg
IOAf4gUOCFgTJCBgLetYzwY2sonNbGEr29jODnayi93sYS/72M8BDnKIwxzh
KMc4zglOcorTnOEs5zjPBS5yictc4SrXuM4NbnKL29zhLve4zwMe8ojHPOEp
z3jOC17yite84S3veM8HPvKJz3zhK9/4zg9+8ovf/OEvAUFtSiACE4SgBCM4
IQhJKEIThrCEIzwRiEgkIhOFqEQjOjGISSxiE4e4xCM+CUhIIhKThKQkIzkp
SEkqUpOGtKQjPRnISCYyk4WsZCM7OchJLnKTh7zkIz8FKMi/FKIwRShKMYpT
gpKUojRlKEs5ylOBilSiMlWoSjWqU4Oa1KI2dahLPerTgIY0ojFNaEozmtOC
lrSiNW1oSzva04GOdKIzXehKN7rTg570ojd96Es/+jOAgQxiMEP4j6EMYzgj
GMkoRjOGsYxjPBOYyCQmM4WpTGM6M5jJLGYzh7nMYz4LWMgiFrOEpSxjOStY
ySpWs4a1rGM9G9jIJjazha1sYzs72MkudrOHvexjPwc4yCEOc4SjHOM4JzjJ
KU5zhrOc4zwXuMglLnOFq1zjOje4yS1uc4e73OM+D3jIIx7zhKc84zkveMkr
XvOGt7zjPR/4yCc+84WvfOM7P/jJL37zh78EBNM/gQhMEIISjOCEICShCE0Y
whKO8EQgIpGITBSiEo3oxCAmsYhNHOISj/gkICGJSEwSkpKM5KQgJalITRrS
ko70ZCAjmchMFrKSjezkICe5yE0eTgb3J8c5xlGOcJhDHOQA+9nHXvawm13s
ZAfb2cZWtrCZTWxkA+tZx1rWsJpVrGQFy1nGUpawmEUsZAHzmcdc5jCbWcxk
BtOZxlSmMJlJTGQC4xnHWMYwmlGMZATDGcZQ/mMIgxnEQAbQn370pQ+96UVP
etCdbnSlC53pREc60J52tKUNrWlFS1rQnGY0pQmNaURDGlCfetSlDrWpRU1q
UJ1qVKUKlalERSpQnnKUpQylKUVJSlCcYhSlCIUpxL8UpAD5yUde8pCbXOQk
B9nJRlaykJlMZCQD6UlHWtKQmlSkJAXJSUZSkpCYRCQkAfGJR1ziEJtYxCQG
0YlGVKIQmUhEJALhCUdYwhCaUIQkBMEJRlCCEJhA/EMAfzXzh9/84ic/+M43
vvKFz3ziIx94zzve8obXvOIlL3jOM57yhMc84iEPuM897nKH29ziJje4zjWu
coXLXOIiFzjPOc5yhtOc+n//nOA4xzjKEQ5ziIMcYD/72MsedrOLnexgO9vY
yhY2s4mNbGA961jLGlazipWsYDnLWMoSFrOIhSxgPvOYyxxmM4uZzGA605jK
FCYziYlMYDzjGMsYRjOKkYxgOMMYyn8MYTCDGMgA+tOPvvShN73oSQ+6042u
dKEznehIB9rTjra0oTWtaEkLmtOMpjShMY1oSAPqU4+61KE2tahJDapTjapU
oTKVqEgFylOOspShNKUoSQmKU4yiFKEwhfiXghQgP/nIy/8Ax2aoWQ==
                    
                    "]]}]}, {}, {}}, {{}, {}, {}, {}, {}, {}, {}, {}}, {{}, \
{}, {}, {}, {}}, {{}, {}, {}, {}, {}, {}, {}, {}, {}, {}, {}, {}, {}, {}, {}, \
{}, {}, {}, {}, {}, {}, {}, {}, {}, {}, {}}, {{}, {}, {}, {}, {}, {}, {}, {}, \
{}, {}, {}, {}, {}, {}, {}, {}, {}, {}, {}, {}, {}, {}, {}, {}, {}, {}}}]}, \
{{}, {{{}, {}, {
                 Hue[0.67, 0.6, 0.6], 
                 Directive[
                  Opacity[0]], 
                 LineBox[CompressedData["
1:eJxdmMlvG2UYxqcFpCriUCQigVSkUFUVGlWopYXuzdu9dbc4jZ3E2WzHTowv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                  "]]}, {
                 Hue[0.9060679774997897, 0.6, 0.6], 
                 Directive[
                  Opacity[0]], 
                 LineBox[CompressedData["
1:eJxdmMlvG2UYxqcFpCriUCQigVSkUFUVGlWopYXuzdu9dbc4jZ3E2WzHTowv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                  "]]}, {
                 Hue[0.1421359549995791, 0.6, 0.6], 
                 Directive[
                  Opacity[0]], 
                 LineBox[CompressedData["
1:eJxdmMlvG2UYxqcFpCriUCQigVSkUFUVGlWopYXuzdu9dbc4jZ3E2WzHTowv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                  "]]}, {
                 Hue[0.37820393249936934`, 0.6, 0.6], 
                 Directive[
                  Opacity[0]], 
                 LineBox[CompressedData["
1:eJxdmEmIJGUWx9MVKTzoYIqCYo4MMiQibuXay+tueyu7q2vL2rsqKvfO7p45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                  "]]}, {
                 Hue[0.6142719099991583, 0.6, 0.6], 
                 Directive[
                  Opacity[0]], 
                 LineBox[CompressedData["
1:eJxdmEmIJGUWx9MVKTzoYIqCYo4MMiQibuXay+tueyu7q2vL2rsqKvfO7p45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                  "]]}}, {{
                 Directive[
                  Opacity[0]]}, {
                 Directive[
                  Opacity[0]]}, {
                 Directive[
                  Opacity[0]]}, {
                 Directive[
                  Opacity[0]]}, {
                 Directive[
                  Opacity[0]]}}, {{
                 Directive[
                  Opacity[0]]}, {
                 Directive[
                  Opacity[0]]}, {
                 Directive[
                  Opacity[0]]}, {
                 Directive[
                  Opacity[0]]}, {
                 Directive[
                  Opacity[
                  0]]}, {}, {}, {}, {}, {}, {}, {}, {}, {}, {}, {}, {}, {}, \
{}, {}, {}, {}, {}, {}, {}, {}}, {{
                 Directive[
                  Opacity[0]]}, {
                 Directive[
                  Opacity[0]]}, {
                 Directive[
                  Opacity[0]]}, {
                 Directive[
                  Opacity[0]]}, {
                 Directive[
                  Opacity[
                  0]]}, {}, {}, {}, {}, {}, {}, {}, {}, {}, {}, {}, {}, {}, \
{}, {}, {}, {}, {}, {}, {}, {}}}}, {{}, {{}, {}}}}, {{{{}, {}, 
               TagBox[{
                 Directive[
                  Opacity[1.], 
                  AbsoluteThickness[1.6], 
                  Dashing[{0, Small}], 
                  GrayLevel[0]], 
                 LineBox[CompressedData["
1:eJxTTMoPSmViYGAwAWIQrdT/d/E8uYt2DCDgkOJwauLtwhfp/+1hfK+97H7P
0vkdYPyzvGacLLEKcL5Tv+uB5f4acH6q/i9eQ14DON9XM/WjmLsJnL/pSm3p
SQsLON9aSGs9S70NnJ/+e9LcpREOcH7+sefG632d4PyuKDbVvzkucH7y8eo0
www3OP8gu2RgdqQHnG8o2DDtZYkXnJ/w2VbHL8cHzl8476jM914/OH9zn9Xt
S0EBcH7JLtuAuT6BcP7+U0s/dUYHwfmZ7pZ7MyOC4fw5Kwy3ZmeFwPkTP+9t
LCgPhfOfrSotyygKg/OPuE2a0dQUDucz+t961N8QAeczOG5jXFAZCed3iH/3
vdQeBecXRCUf+t4cDec/bDRJsZ8aA+eLTJ55q3NhLJy/TECr8vjsOER4Fcm7
qq+Oh/PnKSyQsQ1JgPP31vxiyLiA4Ct6s/9RDk+E87ec1+y+eQXB7zn2z0Ah
NgnOn1z8MivhDoJfIyrHrxSZDOev1GQ9VfIAwWf7wih5MC4Fzl9usLFu3nME
XzNpld/f3FQ4/7bnzk2hrxH8S4urdO2L0+D8H5sDbZQ+Ifj/dxRtbctLh/O9
Ddxca74i+Mdj9TOzizPgfCv2gA1bdyL46zYUfP//H8EHAM8H5s8=
                  "]]}, 
                Annotation[#, "Charting`Private`Tag$144166#1"]& ]}}, {}}}, {
           DisplayFunction -> Identity, DisplayFunction -> Identity, 
            DisplayFunction -> Identity, AspectRatio -> 
            NCache[GoldenRatio^(-1), 0.6180339887498948], 
            Axes -> {True, True}, AxesLabel -> {None, None}, 
            AxesOrigin -> {0., 0}, DisplayFunction :> Identity, 
            Frame -> {{False, False}, {False, False}}, 
            FrameLabel -> {{None, None}, {None, None}}, 
            FrameTicks -> {{Automatic, Automatic}, {Automatic, Automatic}}, 
            GridLines -> {None, None}, GridLinesStyle -> Directive[
              GrayLevel[0.5, 0.4]], 
            Method -> {
             "OptimizePlotMarkers" -> True, 
              "CoordinatesToolOptions" -> {"DisplayFunction" -> ({
                  (Identity[#]& )[
                   Part[#, 1]], 
                  (Identity[#]& )[
                   Part[#, 2]]}& ), "CopiedValueFunction" -> ({
                  (Identity[#]& )[
                   Part[#, 1]], 
                  (Identity[#]& )[
                   Part[#, 2]]}& )}}, PlotRange -> {{0., 200.}, {0, 240.}}, 
            PlotRangeClipping -> True, PlotRangePadding -> {{
               Scaled[0.02], 
               Scaled[0.02]}, {
               Scaled[0.02], 
               Scaled[0.05]}}, Ticks -> {Automatic, Automatic}, TicksStyle -> 
            Medium}], 
          InterpretationBox[
           Cell[
            BoxData[
             FormBox[
             "\"population size \
(\\!\\(\\*StyleBox[\\\"N\\\",FontSlant->\\\"Italic\\\"]\\))\"", TextForm]], 
            "InlineText"], 
           Text[
           "population size \
(\!\(\*StyleBox[\"N\",FontSlant->\"Italic\"]\))"]], 
          InterpretationBox[
           Cell[
            BoxData[
             FormBox["\"generations\"", TextForm]], "InlineText"], 
           Text["generations"]]}, "Labeled", DisplayFunction -> (GridBox[{{
             ItemBox[
              RotationBox[#2, BoxRotation -> 1.5707963267948966`], 
              DefaultBaseStyle -> "LabeledLabel"], 
             TagBox[
              ItemBox[
               PaneBox[
                TagBox[#, "SkipImageSizeLevel"], 
                Alignment -> {Center, Baseline}, BaselinePosition -> 
                Baseline], DefaultBaseStyle -> "Labeled"], 
              "SkipImageSizeLevel"]}, {"", 
             ItemBox[#3, DefaultBaseStyle -> "LabeledLabel"]}}, 
           GridBoxAlignment -> {
            "Columns" -> {{Center}}, "Rows" -> {{Center}}}, AutoDelete -> 
           False, GridBoxItemSize -> {
            "Columns" -> {{Automatic}}, "Rows" -> {{Automatic}}}, 
           BaselinePosition -> {1, 2}]& ), 
         InterpretationFunction -> (RowBox[{"Labeled", "[", 
            RowBox[{#, ",", 
              RowBox[{"{", 
                RowBox[{#2, ",", #3}], "}"}], ",", 
              RowBox[{"{", 
                RowBox[{"Left", ",", "Bottom"}], "}"}], ",", 
              RowBox[{"RotateLabel", "\[Rule]", "True"}]}], "]"}]& )],
        InterpretationBox[
         Cell[
          BoxData[
           FormBox[
            StyleBox["\"A\"", Bold, FontSize -> 20, StripOnInput -> False], 
            TextForm]], "InlineText"], 
         Text[
          Style["A", Bold, FontSize -> 20]]]},
       "Labeled",
       DisplayFunction->(GridBox[{{
           
           ItemBox[#2, Alignment -> {Left, Inherited}, DefaultBaseStyle -> 
            "LabeledLabel"]}, {
           TagBox[
            ItemBox[
             PaneBox[
              TagBox[#, "SkipImageSizeLevel"], 
              Alignment -> {Center, Baseline}, BaselinePosition -> Baseline], 
             DefaultBaseStyle -> "Labeled"], "SkipImageSizeLevel"]}}, 
         GridBoxAlignment -> {"Columns" -> {{Center}}, "Rows" -> {{Center}}}, 
         AutoDelete -> False, 
         GridBoxItemSize -> {
          "Columns" -> {{Automatic}}, "Rows" -> {{Automatic}}}, 
         BaselinePosition -> {2, 1}]& ),
       InterpretationFunction->(RowBox[{"Labeled", "[", 
          RowBox[{#, ",", 
            RowBox[{"{", #2, "}"}], ",", 
            RowBox[{"{", 
              RowBox[{"{", 
                RowBox[{"Top", ",", "Left"}], "}"}], "}"}]}], "]"}]& )]},
     {
      TemplateBox[{TemplateBox[{
          GraphicsBox[{{{{}, 
              GraphicsComplexBox[CompressedData["
1:eJztmjtsJVcZxycRRZTKSLtSkII0WJEVrZB1d0P2RYhPHHazhM3uXb8A8cis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               "], {{{}, {
                  EdgeForm[], 
                  RGBColor[1., 0.7, 0.7], 
                  GraphicsGroupBox[{
                    
                    PolygonBox[{{87, 89, 88}, {104, 106, 105}, {89, 91, 90}, {
                    102, 104, 103}, {91, 93, 92}}], 
                    PolygonBox[{{95, 98, 97, 96}}], 
                    PolygonBox[{{98, 102, 101, 100, 99}}], 
                    PolygonBox[CompressedData["
1:eJwt0FkzggEAheEPZc0SEom0WSu7shYlOy2IWz+A/3/nMePimXfmXJ7010/n
uz8Igj4+/vtp6Ok7b7zSpUObFi8888QjD9xzxy03NLmmQZ0rLqlR5YJzzjjl
hGMqlDnikAP22WOXHbYpUaTAFptssM4aq+TJkSVDmhVSLLNEkkUSLDBPnDli
zDLDNFGmmGSCcSKMMcoIwwwxSJgQA/z9/AvQ9xFd
                    "]]}]}, {}, {}, {}, {}, {}, {}, {}, {}, {}, {}, {}, {
                  EdgeForm[], 
                  RGBColor[1., 0.5, 0.5], 
                  GraphicsGroupBox[{
                    
                    PolygonBox[{{1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 
                    14, 15, 16, 17, 18, 19, 20, 21, 22, 23, 24, 25, 26, 27, 
                    28, 29, 30, 31, 32, 33, 34, 35, 36, 37, 38, 39, 40, 41, 
                    42, 43, 44, 45, 46, 47, 48, 49, 50, 51, 52, 53, 54, 55, 
                    56, 57, 58, 59, 60, 61, 62, 63, 64, 65, 66, 67, 68, 69, 
                    70, 71, 72, 73, 74, 75, 76, 77, 78, 79, 80, 81, 82, 83, 
                    84, 85, 86, 87, 88, 89, 90, 91, 92, 93, 94, 95, 96, 97, 
                    98, 99, 100, 101, 102, 103, 104, 105, 106, 107, 108, 109, 
                    110, 111, 112, 113, 114, 115, 116, 117, 118, 119, 120, 
                    240, 239, 238, 237, 236, 235, 234, 233, 232, 231, 230, 
                    229, 228, 227, 226, 225, 224, 223, 222, 221, 220, 219, 
                    218, 217, 216, 215, 214, 213, 212, 211, 210, 209, 208, 
                    207, 206, 205, 204, 203, 202, 201, 200, 199, 198, 197, 
                    196, 195, 194, 193, 192, 191, 190, 189, 188, 187, 186, 
                    185, 184, 183, 182, 181, 180, 179, 178, 177, 176, 175, 
                    174, 173, 172, 171, 170, 169, 168, 167, 166, 165, 164, 
                    163, 162, 161, 160, 159, 158, 157, 156, 155, 154, 153, 
                    152, 151, 150, 149, 148, 147, 146, 145, 144, 143, 142, 
                    141, 140, 139, 138, 137, 136, 135, 134, 133, 132, 131, 
                    130, 129, 128, 127, 126, 125, 124, 123, 122, 
                    121}}]}]}, {}, {}, {}, {}, {}, {}, {}, {}, {}, {
                  EdgeForm[], 
                  RGBColor[1., 0.3, 0.3], 
                  GraphicsGroupBox[{
                    PolygonBox[CompressedData["
1:eJwt01OXFQAAhdF7J9vmZNu2Odk18ames91k27Zt27Zte69VD3udH/CtExre
LaxrSCAQCPKW7vSgJ73oTR/60o/+DGAggxjMEIYyjOGMYCSjiGA0YxjLOMYz
gYlMYjJTmMo0pjODmcxiNnOYyzzms4CFLGIxS1jKMpazgpWsYjVrWMs61rOB
jWxiM1vYyja2s4Od7GI3e9jLPvZzgIMc4jBHOMoxjnOCk5ziNGc4yznOc4GL
XOIyV7jKNa5zg5vc4jZ3uMs97vOAhzziMU94yjOe84KXvOI1b/737SJ0ZzrR
kQ60J5x2tKUNrWlFS1rQnGY0pQmNaURDGhBGfepRlzrUphY1qUF1qlGVKlSm
EhWpQHnKUZYylKYUJSlBcYpRlCIUphAFKUB+8pGXPOQmFznJQXaykZUsZCYT
GclAKOlJR1rSkJpUpCQFyUlGUpKQmEQkJAHxiUdc4hCbWMQkBtGJRlSiEJlI
hBAM/jviH/ObX/zkB9/5xle+8JlPfOQD73nHX8I+e4c=
                    "]]}]}, {}, {}, {}, {}, {}, {}, {}, {
                  EdgeForm[], 
                  RGBColor[0.7, 0.7, 1.], 
                  GraphicsGroupBox[{
                    PolygonBox[CompressedData["
1:eJwlzFN3FgAAANBvy7a5bNu17JZdy9qya7mWbdu2bdv1E7J1z+nhvt6Q8Miw
iOBAIBDEO97zgY984jNf+Mo3vvODn/ziN3/4G/ifBBFMDGISi9jEIS7xiE8C
EpKIxCQhKclITgpSkorUpCEt6UhPBjKSicxkISshZCM7OchJLnKTh7zkIz8F
KEghClOEohSjOCUoSSlKU4aylKM8FahIJSpThapUI5Tq1KAmtahNHepSj/o0
oCGNaEwTwmhKM5rTgpa0ojVtaEs72tOBjnSiM+F0oSvd6E4PevKWN7zmFS95
wXOe8ZQnPOYRD3nAfe5xlzvc5hY3ucF1rnGVK1zmEhe5wHnOcZYznOYUJznB
cY5xlCMc5hAHOcB+9rGXPexmFzvZwXa2sZUtbGYTG9nAetaxljWsZhUrWcFy
lrGUJSxmEQtZwHzmMZc5zGYWM5nBdKKZxlSmMJlJTGQC4xlHFGMZw2hGMZIR
DGcYQxnCYAYxkAH0J5II+tGXPvSmF/8A+DN0fw==
                    "]]}]}, {}, {}, {}, {}, {}, {
                  EdgeForm[], 
                  RGBColor[0.5, 0.5, 1.], 
                  GraphicsGroupBox[{
                    PolygonBox[CompressedData["
1:eJwV01N0FgAABeB/y7a5bNu17Bafe0tbts3lWrZt27Zt17Lx9fCdc5/vPTek
dURYeHAgEAgimjZCW9rRng6EE0FHOtGZLnSlG93pQU960Zs+9KUf/RnAQAYx
mCEMZRjDGcFIRjGaMYxlHJGMZwITmcRkpjCVaUwnihnMZBazmcNc5jGfBSxk
EYtZwlKWsZwVrGQVq1nDWtaxng1sZBOb2cJWtrGdHexkF7vZw172sZ8DHOQQ
hznCUY5xnBOc5BSnOcNZznGeC1zkEpe5wlWucZ0b3OQWt7nDXe5xn1bGbkkL
mtOMpoTRhMY0oiENqE896lKH2tSiJjWoTijVqEoVKlOJilSgPOUoSxlKU4qS
lKA4xShKEQpTiIIUID/5yEsecpOLnOQgO9kIIStZyEwmMpKB9KQjLWlITSpS
koLkJCMpSUhMIhKSgPjEIy5xiE0sYhKDYIL+H4m/ev7Db37xkx985xtf+cJn
PvGRD7wnmne85Q2vecVLXvCcZzzlCY95xEMe8A8gMn1n
                    
                    "]]}]}, {}, {}}, {{}, {}, {}, {}, {}, {}, {}, {}}, {{}, \
{}, {}, {}, {}}, {{}, {}, {}, {}, {}, {}, {}, {}, {}, {}, {}, {}, {}, {}, {}, \
{}, {}, {}, {}, {}, {}, {}, {}, {}, {}, {}}, {{}, {}, {}, {}, {}, {}, {}, {}, \
{}, {}, {}, {}, {}, {}, {}, {}, {}, {}, {}, {}, {}, {}, {}, {}, {}, {}}}]}, \
{{}, {{{}, {}, {
                 Hue[0.67, 0.6, 0.6], 
                 Directive[
                  PointSize[0.006944444444444445], 
                  AbsoluteThickness[1.6], 
                  RGBColor[1., 0.7, 0.7]], 
                 LineBox[CompressedData["
1:eJxdlElKQ0EURUtH4khBQUFBQxARCUns25R97NL8JM6cCE6cuAT/JrIJcRHZ
hJtw5MSxJv+dW1K/4HO5deu+9/6rZvnpNXked849Dr8R2viuZTBo+142fmoZ
9NueBYa+7d+zMW7r+gn6BHoPfdI/ZHoXfQq9Q/xpw0/5Z8AEfdarHpufC/lN
n7f4vRb6AniPvgg2mF8yTFvUt2T4JX8h5DO9YPFnlb+Iv4leBOVfQU+YX8n9
3yp6B30V7KCvgV3m1yz/p/T1nB8+kF6iP/KX/BjrTS/n+lv2H9H/V4iv/lX8
72i8qX9V4rfQq9TXRN/A30DfsDpeFH+TetWPLVD93cav9Tug4u+i38L3Qj7j
+2AdPAj1Gj8kv/gRfvXnGK7zWsP/APdw6k99Lt4J8+RPxVXfKfwOFD9HP8v5
xS/Rta4e+9Jr5i/iutKL+D+d4qgucfrirkD5xatxXp0z1RnOrerQPRHn3rob
kH1PxXlXnPaVdyTss96Fu5irj3pXhvfeUPdQXPG1D9QT9kX+ZsxTcd5Fp33W
O5fj4dySL5wL+cM7yqp2rKc5fZDTXRLrOofB/6//AZhW5y4=
                  "]], 
                 
                 LineBox[{{75., 0.}, {76., 1.}, {77., 2.}, {78., 0.}, {79., 
                  1.}, {80., 4.}, {81., 1.}, {82., 0.}, {83., 2.}, {84., 
                  0.}, {85., 1.}, {86., 0.}}]}, {
                 Hue[0.9060679774997897, 0.6, 0.6], 
                 Directive[
                  PointSize[0.006944444444444445], 
                  AbsoluteThickness[1.6], 
                  RGBColor[1., 0.5, 0.5]], 
                 LineBox[CompressedData["
1:eJxdlT1vE0EQhpdUiAqkRAIJJBNFKLKiKE4gfHNDIHyHnB3fAQJBCAQQDS3l
/gl6avgT9yfSp6GhSkNDA+eb513bd5I1fvbdnZnd25k7v/dlsD8TQvha/0bW
n6OsMTawonn+OIeBMcFt7NuP5plhfAd7HFuin7By5GauYPykW9vB/6nW+ln8
K/6cKR/XT7ut+uhn4Bz9LOu30M95Ht1t9A7+lH/H533ro8+neK7Pe/5JX0DP
0Rfs3+jJpF9I+3MdrrS/RXiIvujxD4boXWyB3nX7Wee3hH+d75L9bKaV6Mut
818mf61fSfH9fFbsWDOg+L3W+ffIT+e/iq73t8r+tX4N/wX6mg2b1/gM/SL6
C/hSytftetqf8+Xkz/lK6zyvMl/6tRZfJ9+X8A30Xfgm/l7Dt1p65rbzDjb8
7WFhe4t+G36DLmZ+2HC7+x4dNjjcYXwfC1eKf9ft9w/o4o/om/j/hL45PT/c
wyo+fCj/97HsJ8KR/YQH5MN+ItzReT0kPhzFzA+PyEc6HF5hH7s17kcUP0d/
Ms0RrnRfthjnfUe40v18iuU+RrhS/W5Pr4vqGzqXfJotn84jSufc1Zc0LyiO
1ou5J1G83spLdaI6pE+mvkifieIe9T+wnDp1f9Kp09R3qQNrcYTTPZTOvUt9
Q+8F1r2I0qnD9B42sKpT1ZGYuoviWfYzZvc3Zu9ThR3y3fFx3Ys6UtOHam7+
/OI7l3T6VWlVs/5v5n23tDj5HYyaf5T5d6M0m9Srlr+D8frG3+9aLyfi6x7W
z3+Zt0Dt
                  "]]}, {
                 Hue[0.1421359549995791, 0.6, 0.6], 
                 Directive[
                  PointSize[0.006944444444444445], 
                  AbsoluteThickness[1.6], 
                  RGBColor[1., 0.3, 0.3]], 
                 LineBox[CompressedData["
1:eJxdlctrU0EUxseuxFUECwoVrqVICaWkrY+q1R5frY+2JmlyRVCJjwoGtfVd
Xd2/QNy4NkvRTVau75/QTfdZuwqCa23m/L4L5kL5+t3fzJkzc+ecHHu4Wd8Y
CSFs7/0N1J/+YhSrWzM+f9yHujHANavZ9/iM8H4d3Y+m8AOWDsKMNnlfcrV1
4h8cmn+I+Fp/1JSP88OueQ1+xPZFX4WPMX8VftTzKN+CJ8RT/omP+1KDjxfr
OR/3/As+Aa/CJ+zv4FkUP17szzk+1/4m8Q34pK+/24CX0Sa8bD8G0tb5TRFf
5zvlvJzCp4fOf5r8Nb9SrO/nU/HzC1p/Zuj8Z8hP5z8L1/ebZf+aP0f8JnzO
GvEz3oafgN/Bnyz24/4UXP70EJ/Haz9nivXcnyU/jT+Hv4tfQB+g511bLfwF
1578Ius9xptrskFcfHiEXnTt4DN8eIJe4v1TFN95Br/M+xcoPn8Ov+L6eQuO
z1/Cr7p2X8Pl38CXUHF8S3yZeK/gy//HC9fgjM/wpbfw6+T3Ho7PP8BvuPbF
8QEfbqLbcHyGDyuuiebjS+Krrlvkk+F7ym/NtSqOD+9Q+kSJ88zUNzZR+kwX
bnhTPPwO+akvlT4ynz7RZz3D7+AzfJfzz/EVfV/q0LhfRV/kfmb4pEX91+33
oD6/Mj6Ht7i/6rsd5pv6FPcxwyfcP/WxvvKh7nuMN3yH88rwrTbz1Sd0n6nb
hPulus70ffEd7adpP2M/vE881f09+lTTxqJX/UvN+Whqa3Ecdam+slcn3q9S
+xb76ZKv106tGsdTJ+ore3XhvxupLUTOPVPfMvrfbmqf4j8rHu9XavOxIasP
qS+G8A8aflQr
                  "]]}, {
                 Hue[0.37820393249936934`, 0.6, 0.6], 
                 Directive[
                  PointSize[0.006944444444444445], 
                  AbsoluteThickness[1.6], 
                  RGBColor[0.7, 0.7, 1.]], 
                 LineBox[CompressedData["
1:eJxdlbtPFFEUh2+ojBUmkGiiyWgIIYSQBUQR0T0+eIio6+qMlWHYhxF8LD5Q
tJrawtDYb2lWC/4Dp7aiobe2orFVd+/5zk12Jtn89rvffc/MnbO1VrU54Jzb
7f56qddRWbMusb/+lDu962dNqKCZ1sSXdwYk6VXbMX9Ms23+uPbzpY4f1Mya
9H9CvvlqTfyQZtTAD2v/s9b+JONv4E8pywb+NGn+jM7jU0p5RP115hdpbpk/
R9r84cz6H4FT/EgYT3O00H407IeWj2n+quPHNH/b+sbJBn5c1x83KJ8gm/gJ
3b+O+clC+8mwn1peYv9sf2ExPxXaq5+S735+Nv9pzTzFT+s4q7b+Gfqv4Wfk
77/u9cPWd579ewLP0p/xhTC+5sWwX8pzsPlLhfXNh/GVLzOe+QX8U/gKfgu+
ynw24TLjPYdFU17SDs6e4a/Rnv4yY2t/nf6sPRy9wt/Q3H6Nh1PzNzX33+Dh
A9gtau7t4OH2O/yS5qB5WMwv49/i4Rx2K4xn7eEcdrc0j97j4e2P+FV4Fw8L
9d3tgof3PuDXKLf+4X3quzv968ngyNZ3tzB/uGT+HuW2/7C08BXq4wWuWH+w
MJ8cjmz99/vXK/CBzRcucT9zY7v/Vcp5ngW25z+Dw/lUlZXea3jI+5Obt/P2
QT8LnPK8ZsY8rzkcsX73sN8LnLNfGdymvxx2L8hYs8J6BQ73FxY7T2MZ8uXr
9Id3j/W8OYyl5b83CeVk9z3Sz0kiX/2xynsj5hc5r8wv8V1IpOL75T3JrP4y
37VEFrznOcuTsA6dTyKf/Z81zvlE5vyBzjnkHpHO/QerdF2I
                  "]]}, {
                 Hue[0.6142719099991583, 0.6, 0.6], 
                 Directive[
                  PointSize[0.006944444444444445], 
                  AbsoluteThickness[1.6], 
                  RGBColor[0.5, 0.5, 1.]], 
                 LineBox[CompressedData["
1:eJw9lDlsE0EUhkepEJWRiAQSSEsURZEVRU7CEcLh4YaQgDHxUiE2PrghJlzm
XloK5AK3yDVQpKHfmspCSu+CiipNarD3ff+uFP359p93zFvPHKiulRsjzrnW
4G+o9mwVTWu+kj7bxVQ2q54FpnHVf0ufEdNP8neYduXvhGv4ObhB/l3GQQN/
N1zHHzX2it9D/VX8vcbRKv4+4uXvp16EH5DvFv0F9C9/jHzqf8z/GMov5R/P
8pk/ntUzfwJW/IRfGbbxU/OZNO3X8Cetz9E6fp791fHzPkzj5U9l8zJ/yvxt
+dP0p/hp/z1V+QXTRPMtZOvNn2H/8mcs/3PNf5b1mu+s1clrf3PMV7+fOdt/
R/kP4t+BD9HPXfhwtj/jI+htdB5ffJR+xAv0r36Pmcaqd5x69+AT+I/gk+R/
CBfJvwZ7016TOHgLdqfY32N8eEP+adPSOj6cPMU/g77AFz9Dz5oWXuLDCevd
OdPoFT6cwO488fLhnvwLaAsfLlHPXeT9axTeYL27hMqH47e8X6Q/OBar/mXT
QPGwe48umTblwznYLZu2lR9uqv8r5JcP5+RfNS29w4e7+n4l0x7z9nBf3wsO
3pgmWg+7a7ynvod7+p5wELMO7j4hvmwakc/DCecrhtvrnP+y/z2UxQesU7zO
y/VsveWD+5yHGG6yvwTusd6toMzfw22dD1jzT+BA56lCPc1TrHnA0Uf2U/Ff
0/uI/Am+v2n3zWbFf0n3q/smNB2cX7tnQ99K70v27/EH58Duq9D/+Td4Pixb
vftwkbnFYVY3Xd4J7X7s0EcSZvuwfkL/Of1nyfL9Df18GsA95G6gzv0Hxthh
Sg==
                  "]]}}, {{
                 Directive[
                  PointSize[0.006944444444444445], 
                  AbsoluteThickness[1.6], 
                  RGBColor[1., 0.7, 0.7]]}, {
                 Directive[
                  PointSize[0.006944444444444445], 
                  AbsoluteThickness[1.6], 
                  RGBColor[1., 0.5, 0.5]]}, {
                 Directive[
                  PointSize[0.006944444444444445], 
                  AbsoluteThickness[1.6], 
                  RGBColor[1., 0.3, 0.3]]}, {
                 Directive[
                  PointSize[0.006944444444444445], 
                  AbsoluteThickness[1.6], 
                  RGBColor[0.7, 0.7, 1.]]}, {
                 Directive[
                  PointSize[0.006944444444444445], 
                  AbsoluteThickness[1.6], 
                  RGBColor[0.5, 0.5, 1.]]}}, {{
                 Directive[
                  PointSize[0.006944444444444445], 
                  AbsoluteThickness[1.6], 
                  RGBColor[1., 0.7, 0.7]]}, {
                 Directive[
                  PointSize[0.006944444444444445], 
                  AbsoluteThickness[1.6], 
                  RGBColor[1., 0.5, 0.5]]}, {
                 Directive[
                  PointSize[0.006944444444444445], 
                  AbsoluteThickness[1.6], 
                  RGBColor[1., 0.3, 0.3]]}, {
                 Directive[
                  PointSize[0.006944444444444445], 
                  AbsoluteThickness[1.6], 
                  RGBColor[0.7, 0.7, 1.]]}, {
                 Directive[
                  PointSize[0.006944444444444445], 
                  AbsoluteThickness[1.6], 
                  RGBColor[
                  0.5, 0.5, 
                   1.]]}, {}, {}, {}, {}, {}, {}, {}, {}, {}, {}, {}, {}, {}, \
{}, {}, {}, {}, {}, {}, {}, {}}, {{
                 Directive[
                  PointSize[0.006944444444444445], 
                  AbsoluteThickness[1.6], 
                  RGBColor[1., 0.7, 0.7]]}, {
                 Directive[
                  PointSize[0.006944444444444445], 
                  AbsoluteThickness[1.6], 
                  RGBColor[1., 0.5, 0.5]]}, {
                 Directive[
                  PointSize[0.006944444444444445], 
                  AbsoluteThickness[1.6], 
                  RGBColor[1., 0.3, 0.3]]}, {
                 Directive[
                  PointSize[0.006944444444444445], 
                  AbsoluteThickness[1.6], 
                  RGBColor[0.7, 0.7, 1.]]}, {
                 Directive[
                  PointSize[0.006944444444444445], 
                  AbsoluteThickness[1.6], 
                  RGBColor[
                  0.5, 0.5, 
                   1.]]}, {}, {}, {}, {}, {}, {}, {}, {}, {}, {}, {}, {}, {}, \
{}, {}, {}, {}, {}, {}, {}, {}}}}, {{}, {{}, {}}}}, {{{{}, {}, 
               TagBox[{
                 Directive[
                  Opacity[1.], 
                  AbsoluteThickness[1.6], 
                  Dashing[{0, Small}], 
                  GrayLevel[0]], 
                 LineBox[CompressedData["
1:eJxTTMoPSmViYGAwAWIQrdT/d/E8uYt2DCDgkOJwauLtwhfp/+1hfK+97H7P
0vkdYPyzvGacLLEKcL5Tv+uB5f4acH6q/i9eQ14DON9XM/WjmLsJnL/pSm3p
SQsLON9aSGs9S70NnJ/+e9LcpREOcH7+sefG632d4PyuKDbVvzkucH7y8eo0
www3OP8gu2RgdqQHnG8o2DDtZYkXnJ/w2VbHL8cHzl8476jM914/OH9zn9Xt
S0EBcH7JLtuAuT6BcP7+U0s/dUYHwfmZ7pZ7MyOC4fw5Kwy3ZmeFwPkTP+9t
LCgPhfOfrSotyygKg/OPuE2a0dQUDucz+t961N8QAeczOG5jXFAZCed3iH/3
vdQeBecXRCUf+t4cDec/bDRJsZ8aA+eLTJ55q3NhLJy/TECr8vjsOER4Fcm7
qq+Oh/PnKSyQsQ1JgPP31vxiyLiA4Ct6s/9RDk+E87ec1+y+eQXB7zn2z0Ah
NgnOn1z8MivhDoJfIyrHrxSZDOev1GQ9VfIAwWf7wih5MC4Fzl9usLFu3nME
XzNpld/f3FQ4/7bnzk2hrxH8S4urdO2L0+D8H5sDbZQ+Ifj/dxRtbctLh/O9
Ddxca74i+Mdj9TOzizPgfCv2gA1bdyL46zYUfP//H8EHAM8H5s8=
                  "]]}, 
                Annotation[#, "Charting`Private`Tag$150742#1"]& ]}}, {}}}, {
           DisplayFunction -> Identity, DisplayFunction -> Identity, 
            DisplayFunction -> Identity, AspectRatio -> 
            NCache[GoldenRatio^(-1), 0.6180339887498948], 
            Axes -> {True, True}, AxesLabel -> {None, None}, 
            AxesOrigin -> {0., 0}, DisplayFunction :> Identity, 
            Frame -> {{False, False}, {False, False}}, 
            FrameLabel -> {{None, None}, {None, None}}, 
            FrameTicks -> {{Automatic, Automatic}, {Automatic, Automatic}}, 
            GridLines -> {None, None}, GridLinesStyle -> Directive[
              GrayLevel[0.5, 0.4]], 
            Method -> {
             "OptimizePlotMarkers" -> True, 
              "CoordinatesToolOptions" -> {"DisplayFunction" -> ({
                  (Identity[#]& )[
                   Part[#, 1]], 
                  (Identity[#]& )[
                   Part[#, 2]]}& ), "CopiedValueFunction" -> ({
                  (Identity[#]& )[
                   Part[#, 1]], 
                  (Identity[#]& )[
                   Part[#, 2]]}& )}}, PlotRange -> {{0., 92.}, {0, 556.}}, 
            PlotRangeClipping -> True, PlotRangePadding -> {{
               Scaled[0.02], 
               Scaled[0.02]}, {
               Scaled[0.02], 
               Scaled[0.05]}}, Ticks -> {Automatic, Automatic}, TicksStyle -> 
            Medium}], 
          InterpretationBox[
           Cell[
            BoxData[
             FormBox[
             "\"population size \
(\\!\\(\\*StyleBox[\\\"N\\\",FontSlant->\\\"Italic\\\"]\\))\"", TextForm]], 
            "InlineText"], 
           Text[
           "population size \
(\!\(\*StyleBox[\"N\",FontSlant->\"Italic\"]\))"]], 
          InterpretationBox[
           Cell[
            BoxData[
             FormBox["\"generations\"", TextForm]], "InlineText"], 
           Text["generations"]]}, "Labeled", DisplayFunction -> (GridBox[{{
             ItemBox[
              RotationBox[#2, BoxRotation -> 1.5707963267948966`], 
              DefaultBaseStyle -> "LabeledLabel"], 
             TagBox[
              ItemBox[
               PaneBox[
                TagBox[#, "SkipImageSizeLevel"], 
                Alignment -> {Center, Baseline}, BaselinePosition -> 
                Baseline], DefaultBaseStyle -> "Labeled"], 
              "SkipImageSizeLevel"]}, {"", 
             ItemBox[#3, DefaultBaseStyle -> "LabeledLabel"]}}, 
           GridBoxAlignment -> {
            "Columns" -> {{Center}}, "Rows" -> {{Center}}}, AutoDelete -> 
           False, GridBoxItemSize -> {
            "Columns" -> {{Automatic}}, "Rows" -> {{Automatic}}}, 
           BaselinePosition -> {1, 2}]& ), 
         InterpretationFunction -> (RowBox[{"Labeled", "[", 
            RowBox[{#, ",", 
              RowBox[{"{", 
                RowBox[{#2, ",", #3}], "}"}], ",", 
              RowBox[{"{", 
                RowBox[{"Left", ",", "Bottom"}], "}"}], ",", 
              RowBox[{"RotateLabel", "\[Rule]", "True"}]}], "]"}]& )],
        InterpretationBox[
         Cell[
          BoxData[
           FormBox[
            StyleBox["\"B\"", Bold, FontSize -> 20, StripOnInput -> False], 
            TextForm]], "InlineText"], 
         Text[
          Style["B", Bold, FontSize -> 20]]]},
       "Labeled",
       DisplayFunction->(GridBox[{{
           
           ItemBox[#2, Alignment -> {Left, Inherited}, DefaultBaseStyle -> 
            "LabeledLabel"]}, {
           TagBox[
            ItemBox[
             PaneBox[
              TagBox[#, "SkipImageSizeLevel"], 
              Alignment -> {Center, Baseline}, BaselinePosition -> Baseline], 
             DefaultBaseStyle -> "Labeled"], "SkipImageSizeLevel"]}}, 
         GridBoxAlignment -> {"Columns" -> {{Center}}, "Rows" -> {{Center}}}, 
         AutoDelete -> False, 
         GridBoxItemSize -> {
          "Columns" -> {{Automatic}}, "Rows" -> {{Automatic}}}, 
         BaselinePosition -> {2, 1}]& ),
       InterpretationFunction->(RowBox[{"Labeled", "[", 
          RowBox[{#, ",", 
            RowBox[{"{", #2, "}"}], ",", 
            RowBox[{"{", 
              RowBox[{"{", 
                RowBox[{"Top", ",", "Left"}], "}"}], "}"}]}], "]"}]& )]},
     {
      TemplateBox[{TemplateBox[{
          GraphicsBox[{{{{}, 
              GraphicsComplexBox[CompressedData["
1:eJzt3EuvZFUZxvEjI+KoTSDRRJOSEGMI0fYOCvTmKgg0DdrNvSkaWtpWLmqc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               "], {{{}, {
                  EdgeForm[], 
                  RGBColor[1., 0.7, 0.7], 
                  GraphicsGroupBox[{
                    PolygonBox[CompressedData["
1:eJwt09VSUAEARdFrJyIGGCiigIWdYCIWKKgY2IEdgImBYrdiYHd3F3ZhB+rw
Qy5nfFhzzg/s6KzcjJzyQRCUo+z/hlcMggj++L/5RSk/+cF3vvGVL3zmEx/5
QAnvecdb3vCaV7zkBc95xlOKecJjHvGQB9znHne5w21ucZMbXOcaV7nCZS5x
kQuc5xxnOcNpTnGSExznGEc5wmEOUcRBDrCffRSylz3sZhc72cF2trGVLWxm
ExvZwHoKWMda8lnDalaxkjxWsJxlLGUJi8klh2wWsZAFzGcec5nDbGYxkyxm
MJ1pTGUKk5nERCYwnkzGMZYxjCaDUYxkBOmkMZxhpJLCUIYwmEEMJJkBJNGf
fvSlD73pRSIJ9KQH3elGV7rQmU50pAPtaUc8bWlDa1rRkjhiiaEFzYmmGVE0
pQmRNKYRDWlABOHUpx51qUMYtQmlFiHUpAbVqUZVqlCZSsgmqMC/pv4C76tY
dw==
                    "]]}]}, {}, {}, {}, {}, {}, {}, {}, {}, {}, {}, {}, {
                  EdgeForm[], 
                  RGBColor[1., 0.5, 0.5], 
                  GraphicsGroupBox[{
                    PolygonBox[CompressedData["
1:eJwl1GV3EAQABdCN7u7c6O7OMbphdMPo2ujO0d3dKKE0tqB0d6fSKKFggEFe
Dh/ueef9gPeCwyPDIqIFBAQEsjbwY37o0YlBTGIRmzjEJR7xSUBCEpGYJCQl
GclJQUpSkZo0pCUd6clARjKRmSCCyUJWspGdHOQkF7nJQ17ykZ8CFKQQhSlC
UYpRnBKUpBSlKUNZylGeClQkhEqEUpkqVKUa1alBTWpRmzrUpR71aUBDwmhE
Y5rQlGY0pwUtaUVr2tCWdrSnA+F0pBOd6UJXutGdHvSkF72JIJI+9KUf/RnA
QAYxmCEMZRjDGcFIRjGaMYwlinGMZwITmcRkpjCVaUxnBjOZxWzmMJd5zGcB
C1nEYpawlGUsZwUrWcVq1rCWT/iUdaxnAxv5jM/ZxGa2sJVtbGcHO/mCL/mK
r/mGb/mO79nFbn7gR/awl33s5wAHOcRhjnCUYxznBCc5xWnOcJZzAR93tIbV
rGIlK1jOMpayhMUsYiELmM885jKH2cxiJjOYzjSmMoXJTGIiExjPOKIYyxhG
M4qRjGA4wxjKEAYziIEMoD/96EsfIomgN73oSQ+6042udKEznehIOB1oTzva
0obWtKIlLWhOM5rShMY0IoyGNKA+9ahLHWpTi5rUoDrVqEoVKhNKJUKoSAXK
U46ylKE0pShJCYpTjKIUoTCFKEgB8pOPvOQhN7nISQ6yk42sZCGYIDKTiYxk
ID3pSEsaUpOKlKQgOclIShISk4iEJCA+8YhLHGITi5jEIDrRCOTDmb8Tb3nD
a17xP//xL//wkhf8zV/8yR885xm/8xtPecJjHvErv/CQB9znHne5w21u8TM/
cZMbXOcaV7nCZS5xkQuc5z38srYP
                    "]]}]}, {}, {}, {}, {}, {}, {}, {}, {}, {}, {
                  EdgeForm[], 
                  RGBColor[1., 0.3, 0.3], 
                  GraphicsGroupBox[{
                    PolygonBox[CompressedData["
1:eJwl1FN0GAYUBuAktW2mtpXaZmq3qZliq92t2Mqt2MqtGMuh7lbbtpnaxtfT
h+/e5//c89/gsPDQnkEBAQGBLDCO2Ec5xnFOcJJTnOYMZznHeS5wkUtc5gpX
ucZ1bnCTW9wmgjvc5R73ecBDHvGYJzzlGc95wUte8Zo3vOUd7/kQ8DlAIEFE
IjJRiEo0ohODmMQiNnGISzzik4CEJCIxSUhKMpKTgpSkIjVpSEs60hNMBjKS
icxkISvZyE4OcpKL3OQhL/nITwEKUojCFKEoxQihOCUoSSlKU4aylKM8FahI
JSpThapUozo1qEktalOHutQjlPo0oCGNaEwTmtKM5rSgJa1oTRvaEkY72tOB
jnSiM13oSje604OehNOL3vShL1/wJf3ozwAGMojBDGEowxjOCEYyitF8xdeM
YSzjGM83fMsEJjKJyUxhKt/xPdOYzgxm8gM/MovZzGEu85jPT/wc+LlDfRWq
D73pRTifStaD7nSjK13oTCc60oH2tCOMtrShNa1oSQua04ymNKExjWhIA+oT
Sj3qUofa1KImNahONapShcpUoiIVKE85ylKG0pSiJCUoTgjFKEoRClOIghQg
P/nISx5yk4uc5CA72chKFjKTiYxkIJj0pCMtaUhNKlKSguQkIylJSEwiEpKA
+MQjLnGITSxiEoPoRCMqUYhMJIII/PQY+eCm73nHW97wmle85AXPecZTnvCY
RzzkAfe5x13uEMFtbnGTG1znGle5wmUucZELnOccZznDaU5xkhMc5xhHOcJh
DnGQA+xnH3vZw252sZMdbGcbW9nCZjbxP/+xkQ2sZx1rWcNqVvEv//A3f7GS
FSxnGUtZwp/8we/8xq/8wmIWsZCPf1DHlg==
                    "]]}]}, {}, {}, {}, {}, {}, {}, {}, {
                  EdgeForm[], 
                  RGBColor[0.7, 0.7, 1.], 
                  GraphicsGroupBox[{
                    PolygonBox[CompressedData["
1:eJwV01V4EAQUBtAV3TG6RsNGje5QRBEQUQkFpZRQWsGibVExUSmJkaM2GDC6
R43u7o4xaqRnD+c+3/+/3w3r2q9N36CAgIBAphrTmE4UM5jJLGYzh7lEM4/5
LGAhi4ghlsUsIY6lLGM58axgJatYzRrWso71bGAjm9jMFhLYyja2s4OdJLKL
3exhL/vYzwEOcojDHOEoxzjOCU5yitOc4SznOM8FLnKJy1zhKte4zg1ucovb
JHGHZO5yj/s84CEpPOIxT3jKM56nFu8AgQQRTAhpSEs60pOBjGQiM1nISjay
k4Oc5CI3oeQhL/nITwEKUojCFKEoxQijOCUoSSlKU4aylKM84URQgYpUojJV
iKQq1ahODWpSi9rUoS71qE8DGtKIxjThBV6kKS/RjJd5hea8Sgta0orXaM3r
tOEN3uQt2tKO9nTgbd6hI514l/foTBe60o3uvM8H9KAnvejNh3xEH1Ifpx/9
GcBABqXuHywHTWhMIxrSgPrUoy51qE0talKD6lSjKpFUoTKVqEgFIginPOUo
SxlKU4qSlKA4YRSjKEUoTCEKUoD85CMveQglN7nISQ6yk42sZCEzmchIBtKT
jrSkIYRggggkgOd6ecZTnvCYR6TwkAfc5x53SeYOSdzmFje5wXWucZUrXOYS
F7nAec5xljOc5hQnOcFxjnGUIxzmEAc5wH72sZc97GYXiexkB9vZxlYS2MJm
NrGRDaxnHWtZw2pWsZIVxLOcZSwljiUsJpYYFrGQBcxnHtHMZQ6zmcVMZhDF
dKYxlSn8x2QmMZEJjOdf/uFvxvEXf/IHv/MbvzKWX/iZnxjDj/zA93zHt3zD
13zFaEYxkhEMZxhD+ZIv+JzP+JQhDOYTPuZ/h73aHA==
                    "]]}]}, {}, {}, {}, {}, {}, {
                  EdgeForm[], 
                  RGBColor[0.5, 0.5, 1.], 
                  GraphicsGroupBox[{
                    PolygonBox[CompressedData["
1:eJwV0mV0FgQUBuCPbXSHCoqESKNIKApIoyLdYYLSSEkzuru7R3d3w+ga3Q2j
uzb64cfznnP/3HPfc276Os0qNQ0KBAIxCBOtDK1pQ1va0Z4OdCSUTnSmC13p
Rnd60JNe9KYPfelHfwYwkEEMZghDGcZwRjCSUYxmDGMZx3gmMJFJTGYKU5lG
GNOZwUxmMZs5zGUe81nAQhaxmCUsZRnLWcFKVrGaNaxlHevZwEY2sZktbGUb
2wlnBzvZxW72sJd97OcABzlEBIc5wlGOcZwTnOQUpznDWc5xngtc5BKXucJV
rnGdSG5wk1vc5g53ucd9HvCQRzzmCU95xnNeEEU0L3nFa97wlncEgv0OQQQT
QkxiEZs4xCUe8UlAQhKRmCQkJRnJScFHfMwnpCQVn/IZqfmcNKQlHen5ggx8
SUYykZksZCUb2cnBV3xNTr4hF7nJQ16+5Tvy8T0/kJ8CFORHClGYIhSlGMUp
8eH+ED1ITjKSkoTEJCIhCYhPPOISh9jEIiYhBBNEDAK8s/stb3jNK14STRQv
eM4znvKExzziIQ+4zz3ucofb3OImN4jkOte4yhUuc4mLXOA85zjLGU5zipOc
4DjHOMoRDhPBIQ5ygP3sYy972M0udrKDcLazja1sYTOb2MgG1rOOtaxhNatY
yQqWs4ylLGExi1jIAuYzj7nMYTazmMkMphPGNKYyhclMYiITGM84xjKG0Yxi
JCMYzjCGMoTBDGIgA+hPP/rSh970oic96E43utKFznQilI50oD3taEsbWtOK
/2lJC5rTjKb8RxMa04iGNKA+9ajLv/xDHWrzN3/xJ3/wO79Ri5rUoDrVqEoV
KlOJilSgPOUoSxlK8yul+IWf+YmSvAeQKuyi
                    
                    "]]}]}, {}, {}}, {{}, {}, {}, {}, {}, {}, {}, {}}, {{}, \
{}, {}, {}, {}}, {{}, {}, {}, {}, {}, {}, {}, {}, {}, {}, {}, {}, {}, {}, {}, \
{}, {}, {}, {}, {}, {}, {}, {}, {}, {}, {}}, {{}, {}, {}, {}, {}, {}, {}, {}, \
{}, {}, {}, {}, {}, {}, {}, {}, {}, {}, {}, {}, {}, {}, {}, {}, {}, {}}}]}, \
{{}, {{{}, {}, {
                 Hue[0.67, 0.6, 0.6], 
                 Directive[
                  PointSize[
                   NCache[
                    Rational[1, 180], 0.005555555555555556]], 
                  AbsoluteThickness[1.6], 
                  RGBColor[1., 0.7, 0.7]], 
                 LineBox[CompressedData["
1:eJxdlslqVUEQho+uxJWCgoLCUURExHkeWzOYaBKT6I1REz0Ocdr4Cv0od60v
cV7BRfb3Edy41+R8X0PXhVB896+urqqu7psTH36sbu1tmub3/78dO3z+PBjs
Shrtfv7CywkH7EL6ufvZi13g+33YJb7fDz/DHhhsXib+wbS2Y7bVD6Evoh9O
v3bjLKIfKfEH/ehgt9WPhfXH0ZfQ28H2y+TXpj1VfSexxjuFXcWeHmxagc+U
eAOfhfU/Rz76n2f9iPwuDPV9f45+EX/XX8KuYS+j63+F/Ubw1bDfNfQX8PVQ
zw2s/jfD+luw9d0O+h3qke+G/O9hPd/75KP/A6znk7D0P8PZ+XrIfuSTAzeP
sNSb4VL/FJb+ZTiZ73RdT54O+c7wPflmWX0WNr/ZoD8O+cNlXubq/fJcWD9f
++f5up7mCfGZlwyXeX6KdX84qy+EfeWZEGcurLOvniN9K32CyxxQd5ZDXlr7
VPph3tRV7p39s0/EK+donSGOcZPzsVr7ORfu41yXvPT3/nkO1JWcM+dZZi76
oDu3nrN5ec456GWu7e+o5iRP1Xk6974bZU58Z+wfXOpfq/36oDcvsdYvOwey
fYWdt2YdSz/Sep1flpnHXnaOXtV6ku2f7D2DvbfN67C/7NzI7NfD5f15g3V+
ZecEdm56dedmg++dv41az0HvZddvso65SJshf7jMv7pz8hadeClwhn0netl7
/w4/+yd7/2T3h8u70NX1tV2tJ5n9Ov2dT5l8x8G/D/EnsPe0eY+lfy3sO5bU
uT+duvOpTn/GMv3oZfKdyL5bH4jj/jLnk2DnvJO9x8F/DDsPfdAn6t7Lj7Xe
fqzjJ9h+dUHP6tQ3VnfejQ9PZPv/CeZdaWHPM6kTv5N9X2DnZazu74XM+zDR
39+3LXTeg3ar9k9wOX913/ewfix7X2Ten4nxvD+fa72VfQ8Cd3D5f0Kdesdw
mT/Y+zIJ65svMOfTwr4/6Uu9vguc9bf/svvLnj9cfv++1vFa2PuR1FnfweV3
NqwfB70P8Sbq/v/wDR6lf/6CmNQ=
                  "]]}, {
                 Hue[0.9060679774997897, 0.6, 0.6], 
                 Directive[
                  PointSize[
                   NCache[
                    Rational[1, 180], 0.005555555555555556]], 
                  AbsoluteThickness[1.6], 
                  RGBColor[1., 0.5, 0.5]], 
                 LineBox[CompressedData["
1:eJxdlstqFEEUQJusxJWCgoJCKyISxPf7NeUrGjVPjTGjGTtjJlE3/kJ9yqzj
T/QvuHDfn+DGvTF9TkFVQ7icunXrVt3X5Mz4x+pkpqqqX/t//2X//Rn0ciWs
HXx/Bz/39r/Z5cAG5ELYO/hmkAusH0IusX4YucT6kV7GZc4/GvLzjyX7Xn+8
l78XWT+B/SL6k4X+FHrtT/f+v3t+3ct2hXvV+F9Bfxbp/nO9DG/h8/AqfAF/
7+BZ9PJF5DryEnLI/S5jP2T9CveTr3LeR/gaPIKv97Ju4BtI+Sbnb8K3kPJt
/G3Ad9j/Cb5b8D38y/fRe78HSPUPOV9/j2D1g+J+AfmZ9VDoH/ey4X1R3kL/
hPUxEp5+Qf+U/erhWvtnOUc4qn/OeerhoH4u5zhXnP+il533h1vz9RKpHo7E
o5rHv/r5/D3VK/z5Pjh6/msk9RPh1np6wzr5iXBlfu1z9XDKL31ofcSCq6X8
PdG5YH7ou5Z4BTigj+q3kbL29PEUfYAbOMJxgp2svX09xh42PxE23i2c8suc
MN5pbtBfUf0H7OHU78wN6z3NEfSx4BZu37O+xj5kUK7n65X+3W9+OKclP0H2
/XAYF36NH37aSeHX+MON8YfDdnEv8yfjz7jZ362sf+JsvYSNfH+Ea/TOPfvZ
uKb8D/P7OaeTf9h6s09S/cG+N81J/cucZ18Zb/tm6v3tK+7rHO2038z3B/vS
/in09m1j/EbF/Uf5+dE5ob1642df27+w89m+t7+tM+Pp3DPf/q5pF2SkvwP6
d9154JxM+VbPfuew9VK5DtvXvjPN9WH+rlDmxTgXeWqLOHjvNP/Hub1z3HvZ
d8bLOdQV69471Y1+PH8rt0v+7Rv0xq0u42T8jat5VG8fqCd+bVFnXaGv7DP9
w8bRPjTvaa4TL+dMWcf2mXPFPu+Kvq6cS84J2L4MctnX+tfevnW/dQeb1845
aVx22GfcYeMY1GPfwOn/ITi93/36h81v537zv8t5nF/D5jXA5rtRb/4LnsLW
TwunubKb37f6mr+nhlP9w9ZDU+ij9va/+62zgjvYOqu+4Wcn/AP2d7/T
                  "]]}, {
                 Hue[0.1421359549995791, 0.6, 0.6], 
                 Directive[
                  PointSize[
                   NCache[
                    Rational[1, 180], 0.005555555555555556]], 
                  AbsoluteThickness[1.6], 
                  RGBColor[1., 0.3, 0.3]], 
                 LineBox[CompressedData["
1:eJxdlctOFEEUQCusjCtNNNFEk9YYY4jx/X5RPvCFDAwKIgi2yAi68RfqU3qN
fkT/ggv38wlu3AvT51TSNcnk5vStW3XrvurM1o+l0VQI4ff+/0B2v78znRzG
5cnv38zPvf3f9GJkAXI+7k1+U8h5vh9CLvD9cJzY7y3w/Ugn0yL7H439/Y9l
+05/PK4ciD8Dvp/AfoD+ZCez/hR67U93fnx3/6qT7RD/qvhrcv4Q/Vmk68+x
31v4PLwEX2C/d/B0J6N8EX4PX0Ku4d9l9lvj+xX2k6+iX4evof8EX+9kVcM3
OlnLN1m/Ad9iP/l2cd4d5CbybmF/DynfR/8RflDoH6J3v0ecL88U94lI/E8F
h8edbD6jh8MW8gnfvyDlbfRPO1mpl7V/1smoHg5wmOV81qfZwv4591EPJ/wN
L9hfPZy0f9lJ85fgsfd/hb33h8f69xp2f9nz5/DX/eeK/d8gyUeCc73Y5+Qv
wY32gz6nQf+8wBwwHgluzc9i/z4RrkesV/8VO7iFA32c4AjXcILjLnbD4nz6
uvF8+3wLe/VwC9fGlzlhfiJsfztHIv3cOlfsb+cG8XeO2F9J1l5exW6Zdcyb
KH9AwvZ7C0fzs8I6/Ydzv6iHW9n4ce54B3u43en7JbdwIh9htbCHo/mDK/MP
16wP3LMx/wUbh2xvXPSfuFTUW1wr7OEx6/Pc9Pz14nzYenWOV9rDtf1Lnp0n
0blq/GHnh3UR9Z8+zf2z0T/PPnZ9W7B9nf3fLOzte+0LvXWrvxHO81K2fpwz
2td9f5wbzi/r1Pmf3znj43rsnUP2dwsbb+de7fnuS/6r4lz72jzZ9/kehV/6
aZxb515xL891zhg36yD3I+z7UsveX2a9cXcf7dwnvxOeb1xg+964+S42RRyN
j3PBujXP1kX2k7qxTq2LsF2cv93317oo8xaLOsp1Jxt/6xj/xgUH+974w+Y3
yuyX68T7w+az8d3C/1b2HqP+90DfOS+cU353Dhon3zXj5bunP02h95007+NC
n+eY/QfneQjn/MPmIzm3PV97z1dPPJ3zeW7s9tdXsHXuu23d1a5H7ztinhvX
2/+u5z7jgsO3vM9/L1nFBA==
                  "]]}, {
                 Hue[0.37820393249936934`, 0.6, 0.6], 
                 Directive[
                  PointSize[
                   NCache[
                    Rational[1, 180], 0.005555555555555556]], 
                  AbsoluteThickness[1.6], 
                  RGBColor[0.7, 0.7, 1.]], 
                 LineBox[CompressedData["
1:eJxdlklrFUEURgtX4krBgIJCKyIi4jxPqTjPxmgcoomdZ0yconHhvn5Kr+Pa
fa3d+Q/6J7hxb/LqnIbuB4/L6a/urVt1a9o3+jbzfksI4c/Gf9OW39/JsWkX
4+z492/y5/rG7/dipEGxzWJcH/+2xOebzX7V6FuLzfPo24o+sYC+HX2B+Dvi
OP76W/Sd6DX6RPGfNf4u7Fv03cVO6L+n2Ki+t3BSr7A1+VWl3Q/j7y+2drwH
iq3kg+jv4EPwCD6MXcIeGfBR2i+T3zHir6AfJ1/5BPYD9mSxzUf4FPYT9jTx
1M8UG9XPohvvHP0tw+cHfAG2/UXyN79Lxba2v0x/9n8F1v8q+es/OYgXsfin
ONCn+K4+1W8frpHPZ3R5Ff069gs6XDk/N8hXHW7Ub2LhBLfqt4hn/3BSv018
/eFEf+FOsbU63NX7brFZ/e7A/x72K9/vDfK7Dzt/cFffB+jMd5Kt70PG9x4d
juqP+pwGHB734yU4Ml9hmvExngg7fwkO32gHRzg84Tv1jnAFJ3h6jXYDDjPk
pz/c1VMmv2x7839arPWKcHQ+4ZbzIMPd+fAMf+Y3wq7/BHs+ZDjrP4v/CH+4
5fxKsOdVhiP9heeMz/zhbr3KjDfDwfG/wFKPCNdwgivWZ4Yb9+dL/KhHhBvn
X93+YddLeEV86wdH94Ns/eDa/ueKtZ5xrq+nuX78bHv9X2MdP1wxngS7PjNc
m/8b2rn+4W6/w55HGW4ZX5gf5A+7XhMc7R/O1o97OhEvwtYnyc6f7c2fe7Zb
P3Bt/dStH9ztf+7hRH9V3dcj7P6q4W5+6v54G9n1pj/1ab33zZ99sp16ee97
HkffAewX3wm15yHsfd7A3ncZ9j50X7q+w6jfXwVbnzhg97H1SqP+eDwHunrZ
Hm4HeuBd4/6oYOsVYddXPdCT/oyvgT2/skz9WuN5/yz19Wqp31+EvQ9qzz3r
77nq/Nve/o3P/eM5bD/W1fOwu+dW+t+Nb92D9xns/eY57ngybL1b+3P+8XO9
VubpeQa7fx1Ht/7h2vk3nv2bp/U3nu8D31nWH17zPHYerT/se9J5cn+67q1v
HrT3nWFc37m+R2xnft26+tofp/ePfo6vkT3v7He1n5fvJd+Ra9/jf489wME=

                  "]]}, {
                 Hue[0.6142719099991583, 0.6, 0.6], 
                 Directive[
                  PointSize[
                   NCache[
                    Rational[1, 180], 0.005555555555555556]], 
                  AbsoluteThickness[1.6], 
                  RGBColor[0.5, 0.5, 1.]], 
                 LineBox[CompressedData["
1:eJxdlkluFTEQQFtZIVZBAgkkkBqEEEIIwjyDmcIMYYaEBCcQCHOAC/govYYN
R/AR4AZ9hGzYM/i9ltxf+iq9rsHlsl329qVP95Ynmqb59ff/T5bf2rkil8LD
/7/fhfvFgEGR3WL49v83UThH9OuQC+jXh+//xNsFvk8WGZ4TfwPf9d84xCv6
TYWj+s0j/y0DF/3WkX4beahv0Ue+t0jj7ygyOd+djC/vKrJ9Ce+mHkvwHuK/
gveO7PcRb4X89qNfQT+FlA8g3yIPkp98CP4AHyYf9UeK7NUfhd/Ax8hXPj7K
5wT2xjuJXj418j/N/N7DZ0b5nB3xudH8AvHJN8HDfM+TD/ETHBzvAvwJPTyz
iv4i8eEEx4/oLyHVw4l4zWXszQ/u9Z8uMsNpus6nuTLKH47qryLNH+7f8f1a
nW8acXO9yB+f0V+v821ukI/5w53xb6I3P7hRf4t8WO8ED/vhdm2f4Kj+DhJO
svnPFPmT+oeZen4JTswvw/kL/nexRx/gNdcT7vTX/iv+94r84fhw63rI5JPh
VffXfb4z/wA31lN+jZ3seXtAfDjAwXrJxM8PRv4Pi+yXsYMTnOAWznCn/6Mi
Pb9Bdr/AHfXIsuv3uMhJ1iPAP92P6ql3hqP1e4Id+vCk1ie4d3zY/dI8Rbr+
8HAe4IR9ht0vzSzS9Z+t4yd4yv2jvf5zRbbOH45wgif1n6vn0zzDz/Fh+0uC
XY8Mu7+beeLrD9v/0nwdP8Pu72ahnl+APS/Je1x/ODu+9yz5eS/b3xKcXT94
2D+xnk8L239DrO19Fwz9Vsa+0958Ydejh4f6c8+vMd92sdYHOHNe4oh9N0T6
Y6fe82o8zrPvquD8l4hD/VrY+zDA3g8Rdj2T9rDvEuuVYdejN57r94I42Lcv
6vwCnMknjuyTzHw7uKUeGfY+60f2De8k6+W7yXyD7yjHf1nnl/Snvh3cWn/Z
+sv2v+Wa7ZO+V4J64kV4uA9h77tOe+sPT1Fv+7T3TcO70X7Twp4/35U9+ggP
5wvO9O/OdyicYfdLr7/7j3vJ89bCQz+D7S8Rtj4Jtj91sO+BDPv+6WXfj9Yd
vevgeQiuE/WI3pPuf/Wef3jVfrtS5997ju3f5DH1NfwBeCe9Mg==
                  "]]}}, {{
                 Directive[
                  PointSize[
                   NCache[
                    Rational[1, 180], 0.005555555555555556]], 
                  AbsoluteThickness[1.6], 
                  RGBColor[1., 0.7, 0.7]]}, {
                 Directive[
                  PointSize[
                   NCache[
                    Rational[1, 180], 0.005555555555555556]], 
                  AbsoluteThickness[1.6], 
                  RGBColor[1., 0.5, 0.5]]}, {
                 Directive[
                  PointSize[
                   NCache[
                    Rational[1, 180], 0.005555555555555556]], 
                  AbsoluteThickness[1.6], 
                  RGBColor[1., 0.3, 0.3]]}, {
                 Directive[
                  PointSize[
                   NCache[
                    Rational[1, 180], 0.005555555555555556]], 
                  AbsoluteThickness[1.6], 
                  RGBColor[0.7, 0.7, 1.]]}, {
                 Directive[
                  PointSize[
                   NCache[
                    Rational[1, 180], 0.005555555555555556]], 
                  AbsoluteThickness[1.6], 
                  RGBColor[0.5, 0.5, 1.]]}}, {{
                 Directive[
                  PointSize[
                   NCache[
                    Rational[1, 180], 0.005555555555555556]], 
                  AbsoluteThickness[1.6], 
                  RGBColor[1., 0.7, 0.7]]}, {
                 Directive[
                  PointSize[
                   NCache[
                    Rational[1, 180], 0.005555555555555556]], 
                  AbsoluteThickness[1.6], 
                  RGBColor[1., 0.5, 0.5]]}, {
                 Directive[
                  PointSize[
                   NCache[
                    Rational[1, 180], 0.005555555555555556]], 
                  AbsoluteThickness[1.6], 
                  RGBColor[1., 0.3, 0.3]]}, {
                 Directive[
                  PointSize[
                   NCache[
                    Rational[1, 180], 0.005555555555555556]], 
                  AbsoluteThickness[1.6], 
                  RGBColor[0.7, 0.7, 1.]]}, {
                 Directive[
                  PointSize[
                   NCache[
                    Rational[1, 180], 0.005555555555555556]], 
                  AbsoluteThickness[1.6], 
                  RGBColor[
                  0.5, 0.5, 
                   1.]]}, {}, {}, {}, {}, {}, {}, {}, {}, {}, {}, {}, {}, {}, \
{}, {}, {}, {}, {}, {}, {}, {}}, {{
                 Directive[
                  PointSize[
                   NCache[
                    Rational[1, 180], 0.005555555555555556]], 
                  AbsoluteThickness[1.6], 
                  RGBColor[1., 0.7, 0.7]]}, {
                 Directive[
                  PointSize[
                   NCache[
                    Rational[1, 180], 0.005555555555555556]], 
                  AbsoluteThickness[1.6], 
                  RGBColor[1., 0.5, 0.5]]}, {
                 Directive[
                  PointSize[
                   NCache[
                    Rational[1, 180], 0.005555555555555556]], 
                  AbsoluteThickness[1.6], 
                  RGBColor[1., 0.3, 0.3]]}, {
                 Directive[
                  PointSize[
                   NCache[
                    Rational[1, 180], 0.005555555555555556]], 
                  AbsoluteThickness[1.6], 
                  RGBColor[0.7, 0.7, 1.]]}, {
                 Directive[
                  PointSize[
                   NCache[
                    Rational[1, 180], 0.005555555555555556]], 
                  AbsoluteThickness[1.6], 
                  RGBColor[
                  0.5, 0.5, 
                   1.]]}, {}, {}, {}, {}, {}, {}, {}, {}, {}, {}, {}, {}, {}, \
{}, {}, {}, {}, {}, {}, {}, {}}}}, {{}, {{}, {}}}}, {{{{}, {}, 
               TagBox[{
                 Directive[
                  Opacity[1.], 
                  AbsoluteThickness[1.6], 
                  Dashing[{0, Small}], 
                  GrayLevel[0]], 
                 LineBox[CompressedData["
1:eJxTTMoPSmViYGAwAWIQrdT/d/E8uYt2DCDgkOJwauLtwhfp/+1hfK+97H7P
0vkdYPyzvGacLLEKcL5Tv+uB5f4acH6q/i9eQ14DON9XM/WjmLsJnL/pSm3p
SQsLON9aSGs9S70NnJ/+e9LcpREOcH7+sefG632d4PyuKDbVvzkucH7y8eo0
www3OP8gu2RgdqQHnG8o2DDtZYkXnJ/w2VbHL8cHzl8476jM914/OH9zn9Xt
S0EBcH7JLtuAuT6BcP7+U0s/dUYHwfmZ7pZ7MyOC4fw5Kwy3ZmeFwPkTP+9t
LCgPhfOfrSotyygKg/OPuE2a0dQUDucz+t961N8QAeczOG5jXFAZCed3iH/3
vdQeBecXRCUf+t4cDec/bDRJsZ8aA+eLTJ55q3NhLJy/TECr8vjsOER4Fcm7
qq+Oh/PnKSyQsQ1JgPP31vxiyLiA4Ct6s/9RDk+E87ec1+y+eQXB7zn2z0Ah
NgnOn1z8MivhDoJfIyrHrxSZDOev1GQ9VfIAwWf7wih5MC4Fzl9usLFu3nME
XzNpld/f3FQ4/7bnzk2hrxH8S4urdO2L0+D8H5sDbZQ+Ifj/dxRtbctLh/O9
Ddxca74i+Mdj9TOzizPgfCv2gA1bdyL46zYUfP//H8EHAM8H5s8=
                  "]]}, 
                Annotation[#, "Charting`Private`Tag$159597#1"]& ]}}, {}}}, {
           DisplayFunction -> Identity, DisplayFunction -> Identity, 
            DisplayFunction -> Identity, AspectRatio -> 
            NCache[GoldenRatio^(-1), 0.6180339887498948], 
            Axes -> {True, True}, AxesLabel -> {None, None}, 
            AxesOrigin -> {0., 0}, DisplayFunction :> Identity, 
            Frame -> {{False, False}, {False, False}}, 
            FrameLabel -> {{None, None}, {None, None}}, 
            FrameTicks -> {{Automatic, Automatic}, {Automatic, Automatic}}, 
            GridLines -> {None, None}, GridLinesStyle -> Directive[
              GrayLevel[0.5, 0.4]], 
            Method -> {
             "OptimizePlotMarkers" -> True, 
              "CoordinatesToolOptions" -> {"DisplayFunction" -> ({
                  (Identity[#]& )[
                   Part[#, 1]], 
                  (Identity[#]& )[
                   Part[#, 2]]}& ), "CopiedValueFunction" -> ({
                  (Identity[#]& )[
                   Part[#, 1]], 
                  (Identity[#]& )[
                   Part[#, 2]]}& )}}, PlotRange -> {{0., 200.}, {0, 312.}}, 
            PlotRangeClipping -> True, PlotRangePadding -> {{
               Scaled[0.02], 
               Scaled[0.02]}, {
               Scaled[0.02], 
               Scaled[0.05]}}, Ticks -> {Automatic, Automatic}, TicksStyle -> 
            Medium}], 
          InterpretationBox[
           Cell[
            BoxData[
             FormBox[
             "\"population size \
(\\!\\(\\*StyleBox[\\\"N\\\",FontSlant->\\\"Italic\\\"]\\))\"", TextForm]], 
            "InlineText"], 
           Text[
           "population size \
(\!\(\*StyleBox[\"N\",FontSlant->\"Italic\"]\))"]], 
          InterpretationBox[
           Cell[
            BoxData[
             FormBox["\"generations\"", TextForm]], "InlineText"], 
           Text["generations"]]}, "Labeled", DisplayFunction -> (GridBox[{{
             ItemBox[
              RotationBox[#2, BoxRotation -> 1.5707963267948966`], 
              DefaultBaseStyle -> "LabeledLabel"], 
             TagBox[
              ItemBox[
               PaneBox[
                TagBox[#, "SkipImageSizeLevel"], 
                Alignment -> {Center, Baseline}, BaselinePosition -> 
                Baseline], DefaultBaseStyle -> "Labeled"], 
              "SkipImageSizeLevel"]}, {"", 
             ItemBox[#3, DefaultBaseStyle -> "LabeledLabel"]}}, 
           GridBoxAlignment -> {
            "Columns" -> {{Center}}, "Rows" -> {{Center}}}, AutoDelete -> 
           False, GridBoxItemSize -> {
            "Columns" -> {{Automatic}}, "Rows" -> {{Automatic}}}, 
           BaselinePosition -> {1, 2}]& ), 
         InterpretationFunction -> (RowBox[{"Labeled", "[", 
            RowBox[{#, ",", 
              RowBox[{"{", 
                RowBox[{#2, ",", #3}], "}"}], ",", 
              RowBox[{"{", 
                RowBox[{"Left", ",", "Bottom"}], "}"}], ",", 
              RowBox[{"RotateLabel", "\[Rule]", "True"}]}], "]"}]& )],
        InterpretationBox[
         Cell[
          BoxData[
           FormBox[
            StyleBox["\"C\"", Bold, FontSize -> 20, StripOnInput -> False], 
            TextForm]], "InlineText"], 
         Text[
          Style["C", Bold, FontSize -> 20]]]},
       "Labeled",
       DisplayFunction->(GridBox[{{
           
           ItemBox[#2, Alignment -> {Left, Inherited}, DefaultBaseStyle -> 
            "LabeledLabel"]}, {
           TagBox[
            ItemBox[
             PaneBox[
              TagBox[#, "SkipImageSizeLevel"], 
              Alignment -> {Center, Baseline}, BaselinePosition -> Baseline], 
             DefaultBaseStyle -> "Labeled"], "SkipImageSizeLevel"]}}, 
         GridBoxAlignment -> {"Columns" -> {{Center}}, "Rows" -> {{Center}}}, 
         AutoDelete -> False, 
         GridBoxItemSize -> {
          "Columns" -> {{Automatic}}, "Rows" -> {{Automatic}}}, 
         BaselinePosition -> {2, 1}]& ),
       InterpretationFunction->(RowBox[{"Labeled", "[", 
          RowBox[{#, ",", 
            RowBox[{"{", #2, "}"}], ",", 
            RowBox[{"{", 
              RowBox[{"{", 
                RowBox[{"Top", ",", "Left"}], "}"}], "}"}]}], "]"}]& )]},
     {
      TemplateBox[{InterpretationBox[
         Cell[
          BoxData[
           FormBox["\"XX\"", TextForm]], "InlineText"], 
         Text["XX"]],InterpretationBox[
         Cell[
          BoxData[
           FormBox[
           "\"\\!\\(\\*SubscriptBox[\\(X\\), \\(d\\)]\\)X\"", TextForm]], 
          "InlineText"], 
         Text["\!\(\*SubscriptBox[\(X\), \(d\)]\)X"]],InterpretationBox[
         Cell[
          BoxData[
           FormBox[
           "\"\\!\\(\\*SubscriptBox[\\(X\\), \
\\(d\\)]\\)\\!\\(\\*SubscriptBox[\\(X\\), \\(d\\)]\\)\"", TextForm]], 
          "InlineText"], 
         Text[
         "\!\(\*SubscriptBox[\(X\), \(d\)]\)\!\(\*SubscriptBox[\(X\), \
\(d\)]\)"]],InterpretationBox[
         Cell[
          BoxData[
           FormBox["\"XY\"", TextForm]], "InlineText"], 
         Text["XY"]],InterpretationBox[
         Cell[
          BoxData[
           FormBox[
           "\"\\!\\(\\*SubscriptBox[\\(X\\), \\(d\\)]\\)Y\"", TextForm]], 
          "InlineText"], 
         Text["\!\(\*SubscriptBox[\(X\), \(d\)]\)Y"]]},
       "SwatchLegend",
       DisplayFunction->(StyleBox[
         StyleBox[
          PaneBox[
           TagBox[
            GridBox[{{
               TagBox[
                GridBox[{{
                   GraphicsBox[{
                    Directive[
                    EdgeForm[
                    Directive[
                    Opacity[0.3], 
                    GrayLevel[0]]], 
                    PointSize[0.5], 
                    AbsoluteThickness[1.6], 
                    RGBColor[1., 0.7, 0.7]], 
                    RectangleBox[{0, 0}, {24, 24}, "RoundingRadius" -> 0]}, 
                    AspectRatio -> Full, ImageSize -> {24, 24}, 
                    PlotRangePadding -> None, ImagePadding -> Automatic, 
                    BaselinePosition -> (Scaled[0.2666666666666667] -> 
                    Baseline)], #, 
                   GraphicsBox[{
                    Directive[
                    EdgeForm[
                    Directive[
                    Opacity[0.3], 
                    GrayLevel[0]]], 
                    PointSize[0.5], 
                    AbsoluteThickness[1.6], 
                    RGBColor[1., 0.5, 0.5]], 
                    RectangleBox[{0, 0}, {24, 24}, "RoundingRadius" -> 0]}, 
                    AspectRatio -> Full, ImageSize -> {24, 24}, 
                    PlotRangePadding -> None, ImagePadding -> Automatic, 
                    BaselinePosition -> (Scaled[0.2666666666666667] -> 
                    Baseline)], #2, 
                   GraphicsBox[{
                    Directive[
                    EdgeForm[
                    Directive[
                    Opacity[0.3], 
                    GrayLevel[0]]], 
                    PointSize[0.5], 
                    AbsoluteThickness[1.6], 
                    RGBColor[1., 0.3, 0.3]], 
                    RectangleBox[{0, 0}, {24, 24}, "RoundingRadius" -> 0]}, 
                    AspectRatio -> Full, ImageSize -> {24, 24}, 
                    PlotRangePadding -> None, ImagePadding -> Automatic, 
                    BaselinePosition -> (Scaled[0.2666666666666667] -> 
                    Baseline)], #3, 
                   GraphicsBox[{
                    Directive[
                    EdgeForm[
                    Directive[
                    Opacity[0.3], 
                    GrayLevel[0]]], 
                    PointSize[0.5], 
                    AbsoluteThickness[1.6], 
                    RGBColor[0.7, 0.7, 1.]], 
                    RectangleBox[{0, 0}, {24, 24}, "RoundingRadius" -> 0]}, 
                    AspectRatio -> Full, ImageSize -> {24, 24}, 
                    PlotRangePadding -> None, ImagePadding -> Automatic, 
                    BaselinePosition -> (Scaled[0.2666666666666667] -> 
                    Baseline)], #4, 
                   GraphicsBox[{
                    Directive[
                    EdgeForm[
                    Directive[
                    Opacity[0.3], 
                    GrayLevel[0]]], 
                    PointSize[0.5], 
                    AbsoluteThickness[1.6], 
                    RGBColor[0.5, 0.5, 1.]], 
                    RectangleBox[{0, 0}, {24, 24}, "RoundingRadius" -> 0]}, 
                    AspectRatio -> Full, ImageSize -> {24, 24}, 
                    PlotRangePadding -> None, ImagePadding -> Automatic, 
                    BaselinePosition -> (Scaled[0.2666666666666667] -> 
                    Baseline)], #5}}, 
                 GridBoxAlignment -> {
                  "Columns" -> {Center, Left}, "Rows" -> {{Baseline}}}, 
                 AutoDelete -> False, 
                 GridBoxDividers -> {
                  "Columns" -> {{False}}, "Rows" -> {{False}}}, 
                 GridBoxItemSize -> {"Columns" -> {{All}}, "Rows" -> {{All}}},
                  GridBoxSpacings -> {"Columns" -> {{0.5, 0.5}}}], "Grid"]}}, 
             GridBoxAlignment -> {"Columns" -> {{Left}}, "Rows" -> {{Top}}}, 
             AutoDelete -> False, 
             GridBoxDividers -> {"Columns" -> {{None}}, "Rows" -> {{None}}}, 
             GridBoxItemSize -> {"Columns" -> {{All}}, "Rows" -> {{All}}}, 
             GridBoxSpacings -> {"Columns" -> {{0}}, "Rows" -> {{1}}}], 
            "Grid"], Alignment -> Left, AppearanceElements -> None, 
           ImageMargins -> {{48, 0}, {10, 10}}, ImageSizeAction -> 
           "ResizeToFit"], LineIndent -> 0, StripOnInput -> False], {
         FontSize -> 14, FontFamily -> "Arial"}, Background -> Automatic, 
         StripOnInput -> False]& ),
       Editable->True,
       InterpretationFunction:>(RowBox[{"SwatchLegend", "[", 
          RowBox[{
            RowBox[{"{", 
              RowBox[{
                InterpretationBox[
                 ButtonBox[
                  TooltipBox[
                   GraphicsBox[{{
                    GrayLevel[0], 
                    RectangleBox[{0, 0}]}, {
                    GrayLevel[0], 
                    RectangleBox[{1, -1}]}, {
                    RGBColor[1., 0.7, 0.7], 
                    RectangleBox[{0, -1}, {2, 1}]}}, DefaultBaseStyle -> 
                    "ColorSwatchGraphics", AspectRatio -> 1, Frame -> True, 
                    FrameStyle -> 
                    RGBColor[
                    0.6666666666666667, 0.4666666666666667, 
                    0.4666666666666667], FrameTicks -> None, PlotRangePadding -> 
                    None, ImageSize -> 
                    Dynamic[{
                    Automatic, 1.35 CurrentValue["FontCapHeight"]/
                    AbsoluteCurrentValue[Magnification]}]], 
                   StyleBox[
                    RowBox[{"RGBColor", "[", 
                    RowBox[{"1.`", ",", "0.7`", ",", "0.7`"}], "]"}], 
                    NumberMarks -> False]], Appearance -> None, 
                  BaseStyle -> {}, BaselinePosition -> Baseline, 
                  DefaultBaseStyle -> {}, ButtonFunction :> 
                  With[{Typeset`box$ = EvaluationBox[]}, 
                    If[
                    Not[
                    AbsoluteCurrentValue["Deployed"]], 
                    SelectionMove[Typeset`box$, All, Expression]; 
                    FrontEnd`Private`$ColorSelectorInitialAlpha = 1; 
                    FrontEnd`Private`$ColorSelectorInitialColor = 
                    RGBColor[1., 0.7, 0.7]; 
                    FrontEnd`Private`$ColorSelectorUseMakeBoxes = True; 
                    MathLink`CallFrontEnd[
                    FrontEnd`AttachCell[Typeset`box$, 
                    FrontEndResource["RGBColorValueSelector"], {
                    0, {Left, Bottom}}, {Left, Top}, 
                    "ClosingActions" -> {
                    "SelectionDeparture", "ParentChanged", 
                    "EvaluatorQuit"}]]]], BaseStyle -> Inherited, Evaluator -> 
                  Automatic, Method -> "Preemptive"], 
                 RGBColor[1., 0.7, 0.7], Editable -> False, Selectable -> 
                 False], ",", 
                InterpretationBox[
                 ButtonBox[
                  TooltipBox[
                   GraphicsBox[{{
                    GrayLevel[0], 
                    RectangleBox[{0, 0}]}, {
                    GrayLevel[0], 
                    RectangleBox[{1, -1}]}, {
                    RGBColor[1., 0.5, 0.5], 
                    RectangleBox[{0, -1}, {2, 1}]}}, DefaultBaseStyle -> 
                    "ColorSwatchGraphics", AspectRatio -> 1, Frame -> True, 
                    FrameStyle -> 
                    RGBColor[
                    0.6666666666666667, 0.33333333333333337`, 
                    0.33333333333333337`], FrameTicks -> None, 
                    PlotRangePadding -> None, ImageSize -> 
                    Dynamic[{
                    Automatic, 1.35 CurrentValue["FontCapHeight"]/
                    AbsoluteCurrentValue[Magnification]}]], 
                   StyleBox[
                    RowBox[{"RGBColor", "[", 
                    RowBox[{"1.`", ",", "0.5`", ",", "0.5`"}], "]"}], 
                    NumberMarks -> False]], Appearance -> None, 
                  BaseStyle -> {}, BaselinePosition -> Baseline, 
                  DefaultBaseStyle -> {}, ButtonFunction :> 
                  With[{Typeset`box$ = EvaluationBox[]}, 
                    If[
                    Not[
                    AbsoluteCurrentValue["Deployed"]], 
                    SelectionMove[Typeset`box$, All, Expression]; 
                    FrontEnd`Private`$ColorSelectorInitialAlpha = 1; 
                    FrontEnd`Private`$ColorSelectorInitialColor = 
                    RGBColor[1., 0.5, 0.5]; 
                    FrontEnd`Private`$ColorSelectorUseMakeBoxes = True; 
                    MathLink`CallFrontEnd[
                    FrontEnd`AttachCell[Typeset`box$, 
                    FrontEndResource["RGBColorValueSelector"], {
                    0, {Left, Bottom}}, {Left, Top}, 
                    "ClosingActions" -> {
                    "SelectionDeparture", "ParentChanged", 
                    "EvaluatorQuit"}]]]], BaseStyle -> Inherited, Evaluator -> 
                  Automatic, Method -> "Preemptive"], 
                 RGBColor[1., 0.5, 0.5], Editable -> False, Selectable -> 
                 False], ",", 
                InterpretationBox[
                 ButtonBox[
                  TooltipBox[
                   GraphicsBox[{{
                    GrayLevel[0], 
                    RectangleBox[{0, 0}]}, {
                    GrayLevel[0], 
                    RectangleBox[{1, -1}]}, {
                    RGBColor[1., 0.3, 0.3], 
                    RectangleBox[{0, -1}, {2, 1}]}}, DefaultBaseStyle -> 
                    "ColorSwatchGraphics", AspectRatio -> 1, Frame -> True, 
                    FrameStyle -> RGBColor[0.6666666666666667, 0.2, 0.2], 
                    FrameTicks -> None, PlotRangePadding -> None, ImageSize -> 
                    Dynamic[{
                    Automatic, 1.35 CurrentValue["FontCapHeight"]/
                    AbsoluteCurrentValue[Magnification]}]], 
                   StyleBox[
                    RowBox[{"RGBColor", "[", 
                    RowBox[{"1.`", ",", "0.3`", ",", "0.3`"}], "]"}], 
                    NumberMarks -> False]], Appearance -> None, 
                  BaseStyle -> {}, BaselinePosition -> Baseline, 
                  DefaultBaseStyle -> {}, ButtonFunction :> 
                  With[{Typeset`box$ = EvaluationBox[]}, 
                    If[
                    Not[
                    AbsoluteCurrentValue["Deployed"]], 
                    SelectionMove[Typeset`box$, All, Expression]; 
                    FrontEnd`Private`$ColorSelectorInitialAlpha = 1; 
                    FrontEnd`Private`$ColorSelectorInitialColor = 
                    RGBColor[1., 0.3, 0.3]; 
                    FrontEnd`Private`$ColorSelectorUseMakeBoxes = True; 
                    MathLink`CallFrontEnd[
                    FrontEnd`AttachCell[Typeset`box$, 
                    FrontEndResource["RGBColorValueSelector"], {
                    0, {Left, Bottom}}, {Left, Top}, 
                    "ClosingActions" -> {
                    "SelectionDeparture", "ParentChanged", 
                    "EvaluatorQuit"}]]]], BaseStyle -> Inherited, Evaluator -> 
                  Automatic, Method -> "Preemptive"], 
                 RGBColor[1., 0.3, 0.3], Editable -> False, Selectable -> 
                 False], ",", 
                InterpretationBox[
                 ButtonBox[
                  TooltipBox[
                   GraphicsBox[{{
                    GrayLevel[0], 
                    RectangleBox[{0, 0}]}, {
                    GrayLevel[0], 
                    RectangleBox[{1, -1}]}, {
                    RGBColor[0.7, 0.7, 1.], 
                    RectangleBox[{0, -1}, {2, 1}]}}, DefaultBaseStyle -> 
                    "ColorSwatchGraphics", AspectRatio -> 1, Frame -> True, 
                    FrameStyle -> 
                    RGBColor[
                    0.4666666666666667, 0.4666666666666667, 
                    0.6666666666666667], FrameTicks -> None, PlotRangePadding -> 
                    None, ImageSize -> 
                    Dynamic[{
                    Automatic, 1.35 CurrentValue["FontCapHeight"]/
                    AbsoluteCurrentValue[Magnification]}]], 
                   StyleBox[
                    RowBox[{"RGBColor", "[", 
                    RowBox[{"0.7`", ",", "0.7`", ",", "1.`"}], "]"}], 
                    NumberMarks -> False]], Appearance -> None, 
                  BaseStyle -> {}, BaselinePosition -> Baseline, 
                  DefaultBaseStyle -> {}, ButtonFunction :> 
                  With[{Typeset`box$ = EvaluationBox[]}, 
                    If[
                    Not[
                    AbsoluteCurrentValue["Deployed"]], 
                    SelectionMove[Typeset`box$, All, Expression]; 
                    FrontEnd`Private`$ColorSelectorInitialAlpha = 1; 
                    FrontEnd`Private`$ColorSelectorInitialColor = 
                    RGBColor[0.7, 0.7, 1.]; 
                    FrontEnd`Private`$ColorSelectorUseMakeBoxes = True; 
                    MathLink`CallFrontEnd[
                    FrontEnd`AttachCell[Typeset`box$, 
                    FrontEndResource["RGBColorValueSelector"], {
                    0, {Left, Bottom}}, {Left, Top}, 
                    "ClosingActions" -> {
                    "SelectionDeparture", "ParentChanged", 
                    "EvaluatorQuit"}]]]], BaseStyle -> Inherited, Evaluator -> 
                  Automatic, Method -> "Preemptive"], 
                 RGBColor[0.7, 0.7, 1.], Editable -> False, Selectable -> 
                 False], ",", 
                InterpretationBox[
                 ButtonBox[
                  TooltipBox[
                   GraphicsBox[{{
                    GrayLevel[0], 
                    RectangleBox[{0, 0}]}, {
                    GrayLevel[0], 
                    RectangleBox[{1, -1}]}, {
                    RGBColor[0.5, 0.5, 1.], 
                    RectangleBox[{0, -1}, {2, 1}]}}, DefaultBaseStyle -> 
                    "ColorSwatchGraphics", AspectRatio -> 1, Frame -> True, 
                    FrameStyle -> 
                    RGBColor[
                    0.33333333333333337`, 0.33333333333333337`, 
                    0.6666666666666667], FrameTicks -> None, PlotRangePadding -> 
                    None, ImageSize -> 
                    Dynamic[{
                    Automatic, 1.35 CurrentValue["FontCapHeight"]/
                    AbsoluteCurrentValue[Magnification]}]], 
                   StyleBox[
                    RowBox[{"RGBColor", "[", 
                    RowBox[{"0.5`", ",", "0.5`", ",", "1.`"}], "]"}], 
                    NumberMarks -> False]], Appearance -> None, 
                  BaseStyle -> {}, BaselinePosition -> Baseline, 
                  DefaultBaseStyle -> {}, ButtonFunction :> 
                  With[{Typeset`box$ = EvaluationBox[]}, 
                    If[
                    Not[
                    AbsoluteCurrentValue["Deployed"]], 
                    SelectionMove[Typeset`box$, All, Expression]; 
                    FrontEnd`Private`$ColorSelectorInitialAlpha = 1; 
                    FrontEnd`Private`$ColorSelectorInitialColor = 
                    RGBColor[0.5, 0.5, 1.]; 
                    FrontEnd`Private`$ColorSelectorUseMakeBoxes = True; 
                    MathLink`CallFrontEnd[
                    FrontEnd`AttachCell[Typeset`box$, 
                    FrontEndResource["RGBColorValueSelector"], {
                    0, {Left, Bottom}}, {Left, Top}, 
                    "ClosingActions" -> {
                    "SelectionDeparture", "ParentChanged", 
                    "EvaluatorQuit"}]]]], BaseStyle -> Inherited, Evaluator -> 
                  Automatic, Method -> "Preemptive"], 
                 RGBColor[0.5, 0.5, 1.], Editable -> False, Selectable -> 
                 False]}], "}"}], ",", 
            RowBox[{"{", 
              RowBox[{#, ",", #2, ",", #3, ",", #4, ",", #5}], "}"}], ",", 
            RowBox[{"LegendMarkerSize", "\[Rule]", "24"}], ",", 
            RowBox[{"LabelStyle", "\[Rule]", 
              RowBox[{"{", 
                RowBox[{"FontSize", "\[Rule]", "14"}], "}"}]}], ",", 
            RowBox[{"LegendLayout", "\[Rule]", 
              RowBox[{"{", 
                RowBox[{"\"Row\"", ",", "1"}], "}"}]}], ",", 
            RowBox[{"LegendMargins", "\[Rule]", 
              RowBox[{"{", 
                RowBox[{
                  RowBox[{"{", 
                    RowBox[{"48", ",", "0"}], "}"}], ",", 
                  RowBox[{"{", 
                    RowBox[{"10", ",", "10"}], "}"}]}], "}"}]}]}], "]"}]& )]}
    },
    AutoDelete->False,
    GridBoxAlignment->{"Columns" -> {Center}, "Rows" -> {Center}},
    GridBoxItemSize->{"Columns" -> {{Automatic}}, "Rows" -> {{Automatic}}}],
   "Grid"],
  StripOnInput->False,
  ImageSizeMultipliers->{1, 1}]], "Output",
 CellChangeTimes->{{3.8096909106656733`*^9, 3.809690974450738*^9}, {
   3.809691111915864*^9, 3.809691144174444*^9}, 3.809691232106159*^9, {
   3.809691273833362*^9, 3.809691279058724*^9}, {3.809691316985715*^9, 
   3.8096913380071783`*^9}, 3.814005349456761*^9, 3.8141965770337*^9, 
   3.814197133752973*^9},
 CellLabel->
  "Out[1560]=",ExpressionUUID->"1d00c814-0062-4dc4-a8ed-9925d08da0fa"]
}, Open  ]]
}, Closed]]
}, Closed]]
}, Open  ]]
},
WindowSize->{Full, Full},
WindowMargins->{{0, Automatic}, {Automatic, -180}},
Magnification->1.25,
FrontEndVersion->"12.0 for Mac OS X x86 (64-bit) (April 8, 2019)",
StyleDefinitions->"Default.nb"
]
(* End of Notebook Content *)

(* Internal cache information *)
(*CellTagsOutline
CellTagsIndex->{}
*)
(*CellTagsIndex
CellTagsIndex->{}
*)
(*NotebookFileOutline
Notebook[{
Cell[CellGroupData[{
Cell[580, 22, 215, 5, 200, "Title",ExpressionUUID->"f963561f-d28a-4a18-ba87-157215e0ae56"],
Cell[798, 29, 205, 3, 67, "Subtitle",ExpressionUUID->"f0b9a4d6-bcd4-48f7-9312-c0514f50288c"],
Cell[CellGroupData[{
Cell[1028, 36, 155, 3, 84, "Section",ExpressionUUID->"4310ba37-7551-4d2a-a43b-1de62e5d9dda"],
Cell[1186, 41, 2043, 42, 303, "Text",ExpressionUUID->"49957b2d-c3fb-449d-9a2a-f23d490ae09c"]
}, Closed]],
Cell[CellGroupData[{
Cell[3266, 88, 246, 4, 65, "Section",ExpressionUUID->"ecb2965d-0ae6-4644-846b-a4438bf10635"],
Cell[3515, 94, 3998, 117, 361, "Text",ExpressionUUID->"bbf45c70-2c7b-4d53-8e72-65085059f88a"],
Cell[7516, 213, 1103, 22, 78, "Text",ExpressionUUID->"159b7ed4-0e7f-402a-a842-a76148633e12"],
Cell[8622, 237, 672, 11, 73, "Text",ExpressionUUID->"eecf2aaf-f731-49f3-bfda-0e917861cdd7"],
Cell[CellGroupData[{
Cell[9319, 252, 2026, 56, 361, "Input",ExpressionUUID->"933bbd05-fc28-431a-9273-1f317681a387"],
Cell[11348, 310, 1348, 39, 70, "Output",ExpressionUUID->"16d45062-7510-4809-b453-50a35732a81e"]
}, Open  ]],
Cell[12711, 352, 913, 14, 102, "Text",ExpressionUUID->"6495c331-3979-4c05-ad07-e0a3df45f3ae"],
Cell[13627, 368, 280, 7, 37, "Input",ExpressionUUID->"b8cdac10-be99-4143-9f4a-cbcfca14e837"],
Cell[13910, 377, 1742, 51, 231, "Text",ExpressionUUID->"440e961e-4cff-444f-9479-667039b4eae6"]
}, Closed]],
Cell[CellGroupData[{
Cell[15689, 433, 586, 15, 66, "Section",ExpressionUUID->"64106deb-c0a7-4c27-9d28-350dff73efe7"],
Cell[CellGroupData[{
Cell[16300, 452, 169, 3, 67, "Subsection",ExpressionUUID->"39c0b246-04bf-4ac8-9a23-0d4f89147337"],
Cell[16472, 457, 1464, 26, 108, "Text",ExpressionUUID->"952fafe7-9bdc-4222-85bb-f6cec541a6e0"],
Cell[17939, 485, 5914, 180, 685, "Input",ExpressionUUID->"c7291398-a127-4761-862e-ec5f4fc0601c"],
Cell[23856, 667, 289, 7, 44, "Text",ExpressionUUID->"3b8c5c61-7339-4028-8e8d-7a59c34f43d9"],
Cell[24148, 676, 977, 28, 142, "Input",ExpressionUUID->"3c17a47e-acfe-4409-983e-0716addff0f1"]
}, Closed]],
Cell[CellGroupData[{
Cell[25162, 709, 196, 3, 46, "Subsection",ExpressionUUID->"6cec54af-8f67-495d-ae4b-902b41bf33b2"],
Cell[25361, 714, 508, 9, 44, "Text",ExpressionUUID->"73ee4ad7-1b79-49e0-94b4-98acb39dd9bd"],
Cell[CellGroupData[{
Cell[25894, 727, 2625, 72, 94, "Input",ExpressionUUID->"de30a10f-65b2-4f17-9a3d-9934a3ef9d10"],
Cell[28522, 801, 5576, 146, 70, "Output",ExpressionUUID->"12fdd420-b47c-4fea-b67f-734a62e0777f"]
}, Open  ]],
Cell[34113, 950, 222, 4, 44, "Text",ExpressionUUID->"eb65e4ed-19e0-45dc-8f5c-864c8f1c51d6"],
Cell[34338, 956, 623, 16, 37, "Input",ExpressionUUID->"c3fd749e-7bec-4cdf-8c2a-e20d7b5d47c5"],
Cell[34964, 974, 387, 7, 44, "Text",ExpressionUUID->"5d4314a8-b012-4d04-b525-0aad925df173"],
Cell[CellGroupData[{
Cell[35376, 985, 2208, 54, 116, "Input",ExpressionUUID->"b95805b1-e618-48a4-86f1-166d4a8e9edf"],
Cell[37587, 1041, 1581, 31, 70, "Output",ExpressionUUID->"7eea0df3-e312-4ff9-9003-fab15e21f3e1"],
Cell[39171, 1074, 2727, 65, 70, "Output",ExpressionUUID->"d14df597-d291-4b5b-940a-27a3b5d1a11d"],
Cell[41901, 1141, 1328, 23, 70, "Output",ExpressionUUID->"f3698d69-2c83-45f0-823e-0e01b1544ba1"],
Cell[43232, 1166, 1480, 28, 70, "Output",ExpressionUUID->"1b762892-0378-4e0c-b8d4-7449844f0bb0"]
}, Open  ]],
Cell[44727, 1197, 217, 5, 44, "Text",ExpressionUUID->"a2ebe6e0-9a3a-4805-9370-fa17194a141a"],
Cell[CellGroupData[{
Cell[44969, 1206, 1242, 32, 90, "Input",ExpressionUUID->"bafee36e-ee18-4c2c-add2-b705d82ee33d"],
Cell[46214, 1240, 34435, 814, 70, "Output",ExpressionUUID->"9eae98eb-9b60-4173-be0b-3b853ab230b0"],
Cell[80652, 2056, 4742, 131, 70, "Output",ExpressionUUID->"5e2aab2e-0478-4fbb-89c8-bda8cb3001c5"]
}, Open  ]],
Cell[85409, 2190, 241, 6, 44, "Text",ExpressionUUID->"bb02e358-4fa4-428e-afcb-5f3933e40e86"],
Cell[CellGroupData[{
Cell[85675, 2200, 480, 11, 37, "Input",ExpressionUUID->"5fd12814-a562-43f7-bff3-28f08afc2467"],
Cell[86158, 2213, 945, 17, 70, "Output",ExpressionUUID->"72e30847-6871-4d15-9421-a242db0bbb9d"]
}, Open  ]],
Cell[87118, 2233, 317, 7, 44, "Text",ExpressionUUID->"31a4b22f-1b38-4dbb-b93e-621a9cd9e860"],
Cell[CellGroupData[{
Cell[87460, 2244, 2527, 72, 168, "Input",ExpressionUUID->"76e79208-a384-4c72-9517-ab46e19f04b3"],
Cell[89990, 2318, 614, 10, 70, "Output",ExpressionUUID->"ca5b5032-efd0-4446-b419-be905d55a573"],
Cell[90607, 2330, 612, 10, 70, "Output",ExpressionUUID->"6488e1bd-1c6f-4547-bf26-b7b41c73425f"]
}, Open  ]],
Cell[91234, 2343, 319, 7, 44, "Text",ExpressionUUID->"5ff8ff81-da59-4f3f-b306-d25ac6466a2d"],
Cell[CellGroupData[{
Cell[91578, 2354, 2089, 65, 193, "Input",ExpressionUUID->"27c4ce68-c42f-4132-bce1-f9f2f51adf48"],
Cell[93670, 2421, 3118, 92, 70, "Output",ExpressionUUID->"dac1b305-824a-4147-b6cf-b4bb658a7b9e"],
Cell[96791, 2515, 255, 5, 70, "Output",ExpressionUUID->"797da42e-b36e-4881-9932-422c3de6e91d"],
Cell[97049, 2522, 255, 5, 70, "Output",ExpressionUUID->"dcf1802f-61aa-4082-8057-dc2085b1b51e"]
}, Open  ]],
Cell[97319, 2530, 747, 17, 50, "Text",ExpressionUUID->"61bb9a29-74e9-418c-bb86-9b818c625f55"],
Cell[CellGroupData[{
Cell[98091, 2551, 1267, 30, 64, "Input",ExpressionUUID->"b212b1df-35b8-4272-996f-97cfee5f7bd3"],
Cell[99361, 2583, 539, 10, 70, "Output",ExpressionUUID->"5829f78e-11c2-4920-a939-57a2e567a899"]
}, Open  ]],
Cell[99915, 2596, 321, 7, 44, "Text",ExpressionUUID->"819ae32d-3371-438e-83b2-ce5a625c27cb"],
Cell[CellGroupData[{
Cell[100261, 2607, 671, 19, 86, "Input",ExpressionUUID->"2ef5f460-c768-47ac-90cc-5ed41d836f6f"],
Cell[100935, 2628, 199, 4, 70, "Output",ExpressionUUID->"62c3fbaf-d7de-4955-bff1-28e455f6fe54"]
}, Open  ]]
}, Closed]],
Cell[CellGroupData[{
Cell[101183, 2638, 159, 3, 46, "Subsection",ExpressionUUID->"c7603756-d652-48a5-8d25-7c22c8140a38"],
Cell[101345, 2643, 531, 11, 102, "Text",ExpressionUUID->"069fa669-e6f5-414b-8bb6-0768dbae7863"],
Cell[101879, 2656, 1008, 27, 64, "Input",ExpressionUUID->"0311a470-b3b5-4128-9aee-344bd1c02d18"],
Cell[102890, 2685, 241, 6, 44, "Text",ExpressionUUID->"849fe3c7-f7a5-4759-a193-b19237b95a9c"],
Cell[CellGroupData[{
Cell[103156, 2695, 210, 3, 56, "Subsubsection",ExpressionUUID->"79b91767-43e9-45e7-8f2d-900ada457e51"],
Cell[103369, 2700, 538, 9, 73, "Text",ExpressionUUID->"bb9328f9-fea8-4cba-ada5-68fd7946ad96"],
Cell[CellGroupData[{
Cell[103932, 2713, 850, 23, 64, "Input",ExpressionUUID->"7f47ae55-1386-43af-8177-ce1b2387ab73"],
Cell[104785, 2738, 1509, 33, 70, "Output",ExpressionUUID->"1548c55b-67b8-42b5-b844-9a327c3f3c8c"]
}, Open  ]],
Cell[106309, 2774, 2531, 50, 164, "Text",ExpressionUUID->"c6b7c149-1d9d-486e-8d79-e0b69ffdc7ac"],
Cell[CellGroupData[{
Cell[108865, 2828, 1822, 49, 189, "Input",ExpressionUUID->"a062c177-c1b5-4aea-89af-f51112f7b4e6"],
Cell[110690, 2879, 1399, 35, 70, "Output",ExpressionUUID->"ddb9ee80-7091-4e9d-a6fd-550af5402d74"]
}, Open  ]],
Cell[112104, 2917, 218, 5, 44, "Text",ExpressionUUID->"3f84c290-93d1-4128-a2e3-b4d2a996e2ab"],
Cell[CellGroupData[{
Cell[112347, 2926, 366, 9, 37, "Input",ExpressionUUID->"bc064fba-63b8-47ae-aa2c-a70d69dcf620"],
Cell[112716, 2937, 1197, 28, 70, "Output",ExpressionUUID->"d26856b4-bb3f-4050-a6da-2ac9f7950954"]
}, Open  ]],
Cell[113928, 2968, 1718, 41, 217, "Text",ExpressionUUID->"1cf4ddfc-43a4-471a-9b4e-dfe5df17d9c2"],
Cell[CellGroupData[{
Cell[115671, 3013, 640, 16, 37, "Input",ExpressionUUID->"ba5dece9-5a47-4e41-9ea8-ddec4b0874db"],
Cell[116314, 3031, 701, 11, 70, "Output",ExpressionUUID->"cb214e2d-51de-436c-bc6b-e11e1e1135d9"]
}, Open  ]],
Cell[CellGroupData[{
Cell[117052, 3047, 524, 13, 37, "Input",ExpressionUUID->"7fde27ef-5868-4523-8056-b480154601d9"],
Cell[117579, 3062, 713, 11, 70, "Output",ExpressionUUID->"81f42865-d0a0-433e-9086-66a6cba4e3da"]
}, Open  ]],
Cell[118307, 3076, 253, 6, 44, "Text",ExpressionUUID->"ce2389da-627b-495a-8e96-9e6424cc0855"],
Cell[CellGroupData[{
Cell[118585, 3086, 1009, 22, 37, "Input",ExpressionUUID->"5530f374-81a1-4254-82c7-fd3e53be49c9"],
Cell[119597, 3110, 1334, 26, 70, "Output",ExpressionUUID->"669ee6e1-5491-45d7-bae6-5675aaeb1f83"]
}, Open  ]],
Cell[120946, 3139, 194, 3, 44, "Text",ExpressionUUID->"a7fe70a1-d54b-44b0-8921-4c0b2d57c4b4"],
Cell[CellGroupData[{
Cell[121165, 3146, 695, 16, 37, "Input",ExpressionUUID->"b37fc450-b899-412b-9f90-3bbcfa772767"],
Cell[121863, 3164, 784, 15, 70, "Output",ExpressionUUID->"edce6e00-2e91-4b7c-906f-fb16ad73cca3"]
}, Open  ]]
}, Closed]],
Cell[CellGroupData[{
Cell[122696, 3185, 207, 3, 45, "Subsubsection",ExpressionUUID->"532130bb-df22-432d-be64-3fff631f92c5"],
Cell[122906, 3190, 226, 5, 44, "Text",ExpressionUUID->"27fb58c4-df16-4b7d-8d6e-ff8caf4127a2"],
Cell[CellGroupData[{
Cell[123157, 3199, 872, 22, 64, "Input",ExpressionUUID->"e6b57997-0174-4c54-ae73-1a8b8e77e142"],
Cell[124032, 3223, 2194, 52, 70, "Output",ExpressionUUID->"14349bd1-4d87-4bac-9dec-516f8ee4fa33"]
}, Open  ]],
Cell[126241, 3278, 3071, 66, 200, "Text",ExpressionUUID->"8cc660ae-b589-4086-aacc-522dd5b2a3cf"],
Cell[CellGroupData[{
Cell[129337, 3348, 989, 27, 94, "Input",ExpressionUUID->"e5f02f23-dcee-4bc6-91ee-8d5917ca99af"],
Cell[130329, 3377, 1061, 33, 70, "Output",ExpressionUUID->"e26f52ca-46cb-4df1-b9cf-e718b0ae0acc"],
Cell[131393, 3412, 896, 27, 70, "Output",ExpressionUUID->"5fe05b59-72a1-4ede-a803-68b8ba53b75b"]
}, Open  ]],
Cell[CellGroupData[{
Cell[132326, 3444, 2063, 57, 224, "Input",ExpressionUUID->"4f2268e3-8b21-42ca-9c21-5f26150d9928"],
Cell[134392, 3503, 952, 29, 70, "Output",ExpressionUUID->"f7de4613-1dc1-4969-b75c-fda2256b921f"]
}, Open  ]],
Cell[135359, 3535, 269, 6, 44, "Text",ExpressionUUID->"f9cb2b4e-b36a-47ac-9f6c-6aabd0033d26"],
Cell[CellGroupData[{
Cell[135653, 3545, 199, 5, 37, "Input",ExpressionUUID->"47a59791-140b-49c7-ad27-d57f3a116d92"],
Cell[135855, 3552, 905, 30, 70, "Output",ExpressionUUID->"fdfd170c-057e-4e00-84ac-17c8422a404d"]
}, Open  ]],
Cell[136775, 3585, 1490, 41, 246, "Text",ExpressionUUID->"ea849c1f-d447-44c8-a831-dda615a758e0"],
Cell[CellGroupData[{
Cell[138290, 3630, 2232, 59, 116, "Input",ExpressionUUID->"6851f3e7-d9e9-4279-9edf-bd03416640c2"],
Cell[140525, 3691, 1469, 31, 70, "Output",ExpressionUUID->"a39878f8-78a0-48a9-a1a2-8548eff096cc"],
Cell[141997, 3724, 1491, 32, 70, "Output",ExpressionUUID->"0d326dd4-b466-4acf-a0dc-aa8076598e89"],
Cell[143491, 3758, 1048, 18, 70, "Output",ExpressionUUID->"9d98d3a1-1f76-4752-88d8-b1b992cab3f0"],
Cell[144542, 3778, 961, 14, 70, "Output",ExpressionUUID->"ccb76f38-d119-4c6c-abc9-5087a899c08a"]
}, Open  ]],
Cell[145518, 3795, 653, 16, 102, "Text",ExpressionUUID->"9af9eb1e-7c80-466b-af4a-62179fc49259"],
Cell[CellGroupData[{
Cell[146196, 3815, 633, 13, 37, "Input",ExpressionUUID->"491fb21d-1290-499e-87eb-f450adfd4779"],
Cell[146832, 3830, 925, 14, 70, "Output",ExpressionUUID->"e62d4d91-a610-4f42-8f94-af359b527e44"]
}, Open  ]],
Cell[147772, 3847, 181, 3, 44, "Text",ExpressionUUID->"82cebce6-a58c-41d6-9fe7-68f3c37b6f42"],
Cell[CellGroupData[{
Cell[147978, 3854, 228, 4, 37, "Input",ExpressionUUID->"eba6001c-7ecc-487c-bc04-abc003cfbda8"],
Cell[148209, 3860, 1222, 28, 70, "Output",ExpressionUUID->"463bdf59-0013-4c54-bd28-cf8aa8c64cf9"]
}, Open  ]],
Cell[149446, 3891, 216, 3, 44, "Text",ExpressionUUID->"65db19cc-62c1-4302-9fc4-09cc92a8709c"],
Cell[CellGroupData[{
Cell[149687, 3898, 576, 14, 37, "Input",ExpressionUUID->"de8f62a2-64e2-404d-aae8-663b919d7c2d"],
Cell[150266, 3914, 739, 16, 70, "Output",ExpressionUUID->"d70e4ebf-e872-4ddc-8d19-edee238d7459"]
}, Open  ]]
}, Closed]]
}, Closed]],
Cell[CellGroupData[{
Cell[151066, 3937, 167, 3, 46, "Subsection",ExpressionUUID->"84474009-2768-4dd8-97c9-3a4b7276fabc"],
Cell[151236, 3942, 422, 10, 102, "Text",ExpressionUUID->"a6911daa-5507-424e-90e6-cb28c25dcca6"],
Cell[CellGroupData[{
Cell[151683, 3956, 1039, 30, 90, "Input",ExpressionUUID->"a84301ba-cf0e-4539-a1f4-955b7eb24eff"],
Cell[152725, 3988, 323, 6, 70, "Output",ExpressionUUID->"a9fcffa3-1ba9-4eb2-8060-c38425e13d68"],
Cell[153051, 3996, 4004, 94, 70, "Output",ExpressionUUID->"096912d8-e96d-4233-b417-bf0eaf30a9b1"],
Cell[157058, 4092, 515, 14, 70, "Output",ExpressionUUID->"f5964664-00d2-4470-b2f2-61f5cc56846d"]
}, Open  ]],
Cell[157588, 4109, 431, 8, 73, "Text",ExpressionUUID->"a1a39302-2e9c-467b-853f-9039b002746c"],
Cell[CellGroupData[{
Cell[158044, 4121, 1202, 27, 64, "Input",ExpressionUUID->"8dfa1109-9f28-44f2-ab69-f549df3c952a"],
Cell[159249, 4150, 453, 7, 70, "Output",ExpressionUUID->"3e8a93d7-6ba5-4512-a7b9-d5cc32dd4360"],
Cell[159705, 4159, 453, 7, 70, "Output",ExpressionUUID->"b886918b-da0f-4148-a97d-80a922096c2c"]
}, Open  ]],
Cell[160173, 4169, 380, 7, 44, "Text",ExpressionUUID->"92f7d12e-62d8-49ec-bd54-9a11abb6bf3e"],
Cell[160556, 4178, 67807, 1696, 1132, "Input",ExpressionUUID->"e2a5396b-cee0-4f52-8545-db86db08f2f0"],
Cell[228366, 5876, 156, 3, 44, "Text",ExpressionUUID->"d1f15f6d-3212-4137-81ef-697918aa6d83"],
Cell[CellGroupData[{
Cell[228547, 5883, 2836, 73, 138, "Input",ExpressionUUID->"a22582de-5b0f-4ce0-a46a-bc9a52795dd9"],
Cell[231386, 5958, 22107, 436, 70, "Output",ExpressionUUID->"7e6aa397-65c5-41f0-ac8c-0258435f0c59"]
}, Open  ]]
}, Closed]]
}, Closed]],
Cell[CellGroupData[{
Cell[253554, 6401, 536, 14, 66, "Section",ExpressionUUID->"a3d9aa4f-4097-4134-9b5c-9a4fe1d5ffb9"],
Cell[CellGroupData[{
Cell[254115, 6419, 167, 3, 67, "Subsection",ExpressionUUID->"d6a593b5-fc9a-4ff0-b568-77f6abb26a87"],
Cell[254285, 6424, 506, 10, 102, "Text",ExpressionUUID->"69f9f8a0-3224-4ce1-95a7-20e790020c64"],
Cell[254794, 6436, 14067, 417, 807, "Input",ExpressionUUID->"49a453be-289a-46d8-a5bb-2cff568a2eac"],
Cell[268864, 6855, 214, 6, 44, "Text",ExpressionUUID->"84e39e22-f4cc-4793-a45a-e8a087e71a1e"],
Cell[269081, 6863, 862, 25, 142, "Input",ExpressionUUID->"c9182c0f-7901-46d8-8bbb-a2f0e2c978ec"]
}, Closed]],
Cell[CellGroupData[{
Cell[269980, 6893, 198, 3, 46, "Subsection",ExpressionUUID->"8c3a6cc1-cc9d-4c72-b273-ef5ed5c3ee86"],
Cell[270181, 6898, 583, 10, 73, "Text",ExpressionUUID->"3fa6a303-336e-49b9-ba10-d0f9b7c8282d"],
Cell[270767, 6910, 703, 18, 37, "Input",ExpressionUUID->"797883cf-eaed-4f58-8766-21be4be1bc4a"],
Cell[271473, 6930, 404, 8, 44, "Text",ExpressionUUID->"1f0fa615-a1a8-4a49-8f28-362abf728df4"],
Cell[CellGroupData[{
Cell[271902, 6942, 1571, 37, 64, "Input",ExpressionUUID->"2fd847cc-8c70-4b60-852d-932e9004fbd7"],
Cell[273476, 6981, 1034, 27, 70, "Output",ExpressionUUID->"112ee71e-507d-4a44-92c1-285475b460fc"],
Cell[274513, 7010, 2061, 53, 70, "Output",ExpressionUUID->"1ff45ad7-bfc0-40a6-8796-741664783276"]
}, Open  ]],
Cell[276589, 7066, 395, 8, 44, "Text",ExpressionUUID->"3ee33cf4-9caa-438a-91a8-e6c61ff9bfce"],
Cell[276987, 7076, 2643, 69, 150, "Input",ExpressionUUID->"1c43e52d-6fae-4d7c-8290-b58af3761bb2"],
Cell[279633, 7147, 301, 6, 44, "Text",ExpressionUUID->"ac86ebce-ccb5-4e46-895b-f4f370499b0b"],
Cell[CellGroupData[{
Cell[279959, 7157, 751, 20, 37, "Input",ExpressionUUID->"24e38814-b48b-4912-8a57-eba6bbb961af"],
Cell[280713, 7179, 113954, 2839, 70, "Output",ExpressionUUID->"3ba05cf0-15cb-4431-b1a8-b31ed054edce"]
}, Open  ]]
}, Closed]],
Cell[CellGroupData[{
Cell[394716, 10024, 157, 3, 46, "Subsection",ExpressionUUID->"bd99752f-e3a9-4507-a41e-16f72d13af99"],
Cell[394876, 10029, 334, 8, 102, "Text",ExpressionUUID->"2e906936-97d9-44e1-a935-2be5e7061234"],
Cell[395213, 10039, 1116, 29, 64, "Input",ExpressionUUID->"f15f9af6-18ce-4a44-acc5-bbe08526fe99"],
Cell[396332, 10070, 239, 6, 44, "Text",ExpressionUUID->"fe4e26a8-f4e0-41b2-b4dc-01f75780b3bf"],
Cell[CellGroupData[{
Cell[396596, 10080, 208, 3, 56, "Subsubsection",ExpressionUUID->"ddc39688-4044-459a-8251-86bb830d06bd"],
Cell[396807, 10085, 468, 8, 73, "Text",ExpressionUUID->"299b84b6-b272-4a8b-b5a7-5add72f98c62"],
Cell[CellGroupData[{
Cell[397300, 10097, 1009, 24, 64, "Input",ExpressionUUID->"ce804ada-232a-456f-84c5-36048946ac34"],
Cell[398312, 10123, 1598, 37, 70, "Output",ExpressionUUID->"b8275a37-31a0-4c71-ae64-965dc064d9ae"]
}, Open  ]],
Cell[399925, 10163, 2512, 52, 171, "Text",ExpressionUUID->"3314a751-f981-4828-bc7d-4866dc84ed1b"],
Cell[CellGroupData[{
Cell[402462, 10219, 1769, 49, 198, "Input",ExpressionUUID->"f2985155-9fb4-430b-a258-99e2fb6a1626"],
Cell[404234, 10270, 1086, 29, 70, "Output",ExpressionUUID->"883882bc-7fea-4519-8d44-997b117b938f"]
}, Open  ]],
Cell[405335, 10302, 170, 3, 44, "Text",ExpressionUUID->"57f608a2-425f-4a96-acf0-62ca320654b2"],
Cell[CellGroupData[{
Cell[405530, 10309, 247, 7, 37, "Input",ExpressionUUID->"cd15c18a-b09c-410c-b1f2-430be7bad0f9"],
Cell[405780, 10318, 936, 26, 70, "Output",ExpressionUUID->"f51e1940-4940-4cd9-91cd-22503cacfd8b"]
}, Open  ]],
Cell[406731, 10347, 1420, 37, 217, "Text",ExpressionUUID->"88ee105b-b6e1-4def-a77f-522399abc6bf"],
Cell[CellGroupData[{
Cell[408176, 10388, 1643, 43, 90, "Input",ExpressionUUID->"52b9dcb3-9b7e-4ec5-b7ae-59ca876cc397"],
Cell[409822, 10433, 1115, 27, 70, "Output",ExpressionUUID->"4c98ecf0-81bb-42c5-a010-6b750005e07f"],
Cell[410940, 10462, 1115, 27, 70, "Output",ExpressionUUID->"38f36a22-7bfa-409e-a9ad-068d6af4ee23"],
Cell[412058, 10491, 597, 9, 70, "Output",ExpressionUUID->"dcbc15c1-1d15-4e82-a93b-21a6302d6939"]
}, Open  ]],
Cell[412670, 10503, 378, 7, 44, "Text",ExpressionUUID->"6b47a709-ea8f-4dcd-9dcf-4cc6acfed59b"],
Cell[CellGroupData[{
Cell[413073, 10514, 700, 18, 37, "Input",ExpressionUUID->"ca6a9cb5-1fa5-45f3-b3f0-9b6ffdce575c"],
Cell[413776, 10534, 1011, 26, 70, "Output",ExpressionUUID->"44829f00-2da3-476e-8be2-82f3c37d989e"]
}, Open  ]],
Cell[414802, 10563, 170, 3, 44, "Text",ExpressionUUID->"30fbf15a-a200-4476-8ddb-0150815c1608"],
Cell[CellGroupData[{
Cell[414997, 10570, 610, 14, 37, "Input",ExpressionUUID->"a57737c8-21c9-4cbd-bdf3-ef59defff834"],
Cell[415610, 10586, 819, 21, 70, "Output",ExpressionUUID->"f55c05fd-d07f-4418-8673-45d9c298035c"]
}, Open  ]]
}, Closed]],
Cell[CellGroupData[{
Cell[416478, 10613, 211, 3, 45, "Subsubsection",ExpressionUUID->"0cf784ab-e1fd-40fd-aabe-51311338b372"],
Cell[416692, 10618, 200, 5, 44, "Text",ExpressionUUID->"5dffe054-dd46-4f7a-b6d9-eb3b1c467ab4"],
Cell[CellGroupData[{
Cell[416917, 10627, 786, 21, 64, "Input",ExpressionUUID->"56682ae3-a806-4c2e-8a48-cc078ee8ed44"],
Cell[417706, 10650, 3520, 80, 70, "Output",ExpressionUUID->"c5a617cd-1425-41dd-91b9-09bdde328bbc"]
}, Open  ]],
Cell[421241, 10733, 3054, 68, 200, "Text",ExpressionUUID->"dce823c8-8f9a-4560-8b4f-352fa58efdac"],
Cell[CellGroupData[{
Cell[424320, 10805, 2174, 57, 198, "Input",ExpressionUUID->"dfa473d8-2b51-429d-b510-5c1e5c37f9aa"],
Cell[426497, 10864, 1256, 35, 70, "Output",ExpressionUUID->"99711265-4b94-4f5e-870c-d29f8ac76404"]
}, Open  ]],
Cell[427768, 10902, 239, 5, 44, "Text",ExpressionUUID->"64fd5d32-57a1-45d1-8bc8-3b887885e303"],
Cell[CellGroupData[{
Cell[428032, 10911, 269, 6, 37, "Input",ExpressionUUID->"8e4a5cd7-1809-43ec-a6c7-a8fb59e6e802"],
Cell[428304, 10919, 1141, 35, 70, "Output",ExpressionUUID->"ac09b42e-a463-40b9-8809-0c1dfe5510da"]
}, Open  ]],
Cell[429460, 10957, 1561, 42, 246, "Text",ExpressionUUID->"0010841b-1eca-4e62-9dfb-ed7eff11b6fe"],
Cell[CellGroupData[{
Cell[431046, 11003, 2025, 56, 116, "Input",ExpressionUUID->"866dbab8-0523-4787-88c5-77d23c472993"],
Cell[433074, 11061, 1128, 32, 70, "Output",ExpressionUUID->"2b73118d-1bcb-49b9-9740-0d3679e45fad"],
Cell[434205, 11095, 1076, 30, 70, "Output",ExpressionUUID->"5ac98a83-df50-4c17-bcb7-d7ab6108c8dc"],
Cell[435284, 11127, 582, 12, 70, "Output",ExpressionUUID->"43a08ec8-b4df-4dfb-9af8-8c150f512416"],
Cell[435869, 11141, 490, 8, 70, "Output",ExpressionUUID->"07cbc167-45e6-47ce-9e30-dd5acfcd19e7"]
}, Open  ]],
Cell[436374, 11152, 424, 8, 44, "Text",ExpressionUUID->"21c5f077-202a-4e67-be8c-e0c1e0eb41bc"],
Cell[CellGroupData[{
Cell[436823, 11164, 818, 18, 37, "Input",ExpressionUUID->"966610af-88a3-49f2-8723-ecdc07880cce"],
Cell[437644, 11184, 638, 10, 70, "Output",ExpressionUUID->"813b299a-4a5b-48a8-8003-d91dc644c0ba"]
}, Open  ]],
Cell[438297, 11197, 822, 19, 73, "Text",ExpressionUUID->"f34cac6b-8b66-4371-8621-80a37628061d"],
Cell[CellGroupData[{
Cell[439144, 11220, 302, 6, 37, "Input",ExpressionUUID->"5e4271bd-a10a-4cce-9caf-6fc1a0aee066"],
Cell[439449, 11228, 1221, 36, 70, "Output",ExpressionUUID->"6369344b-a73e-4cb8-b820-bff32b89eecd"]
}, Open  ]]
}, Closed]]
}, Closed]]
}, Closed]],
Cell[CellGroupData[{
Cell[440743, 11272, 558, 15, 66, "Section",ExpressionUUID->"ae63bde1-64a0-45d8-a5df-efb083e37f75"],
Cell[CellGroupData[{
Cell[441326, 11291, 163, 2, 67, "Subsection",ExpressionUUID->"857e1452-59f6-4db3-9370-0c89dc3e3a7b"],
Cell[441492, 11295, 724, 15, 217, "Text",ExpressionUUID->"f5cc5e9e-2d93-4dde-82cb-9b3a43b74208"],
Cell[442219, 11312, 2804, 82, 391, "Input",ExpressionUUID->"212cac65-fd16-437c-b577-9747db50f982"],
Cell[445026, 11396, 214, 6, 44, "Text",ExpressionUUID->"2fc5c8bb-b413-4c0a-80f7-0ac214eb88a3"],
Cell[445243, 11404, 857, 25, 142, "Input",ExpressionUUID->"fbf2e3c1-0190-4b27-9424-bf3a77c868ba"]
}, Closed]],
Cell[CellGroupData[{
Cell[446137, 11434, 223, 3, 46, "Subsection",ExpressionUUID->"b849157a-262e-4cbf-9118-8d1e8c4b3334"],
Cell[446363, 11439, 339, 7, 44, "Text",ExpressionUUID->"9a26f5fd-a2fa-42e2-abf3-46cb5b5baeb8"],
Cell[CellGroupData[{
Cell[446727, 11450, 2570, 71, 130, "Input",ExpressionUUID->"915927a4-d9f8-450f-832c-1069351e9a1e"],
Cell[449300, 11523, 4898, 150, 70, "Output",ExpressionUUID->"b58b7436-e2d4-496c-81b0-c709638835e3"]
}, Open  ]],
Cell[454213, 11676, 177, 3, 44, "Text",ExpressionUUID->"6057c0d4-b727-46fe-9189-f0a38bbd200e"],
Cell[454393, 11681, 724, 18, 37, "Input",ExpressionUUID->"af9006e4-f765-4961-b4f8-c03ebc39bb15"],
Cell[455120, 11701, 264, 6, 44, "Text",ExpressionUUID->"fb77491e-bd41-415d-bd14-88d89493f794"],
Cell[CellGroupData[{
Cell[455409, 11711, 1893, 51, 116, "Input",ExpressionUUID->"ff3a0256-827a-4e99-9958-7f4408ab7370"],
Cell[457305, 11764, 582, 12, 70, "Output",ExpressionUUID->"2fe10b64-a2b9-402d-add5-ed7158ffe041"],
Cell[457890, 11778, 734, 17, 70, "Output",ExpressionUUID->"d9334b4c-9932-48e6-b6be-7e1617afdf26"]
}, Open  ]],
Cell[458639, 11798, 238, 5, 44, "Text",ExpressionUUID->"7dbef16d-c4ba-4902-8cf2-180cdecaccb4"],
Cell[458880, 11805, 871, 23, 37, "Input",ExpressionUUID->"5d83063f-2a14-4474-9a7d-a9aab5e80959"],
Cell[CellGroupData[{
Cell[459776, 11832, 451, 12, 37, "Input",ExpressionUUID->"7eb4b35b-93f8-4eee-ae84-40debbf22b38"],
Cell[460230, 11846, 3278, 94, 70, "Output",ExpressionUUID->"ce6fc66f-12f8-4453-b78e-a878a126492c"]
}, Open  ]]
}, Closed]],
Cell[CellGroupData[{
Cell[463557, 11946, 157, 3, 46, "Subsection",ExpressionUUID->"dd05fb51-e055-4cb7-bb15-81c0cc3c5f9d"],
Cell[463717, 11951, 334, 8, 102, "Text",ExpressionUUID->"47972617-7c6d-4b93-9067-80c6647915f7"],
Cell[464054, 11961, 1164, 29, 64, "Input",ExpressionUUID->"5a43fb10-1189-4adb-b947-0359d89fbcc5"],
Cell[465221, 11992, 263, 6, 44, "Text",ExpressionUUID->"c475972b-93ec-4c87-87a7-4ed38c09ad07"],
Cell[CellGroupData[{
Cell[465509, 12002, 234, 4, 56, "Subsubsection",ExpressionUUID->"52a97988-5a75-4509-936e-06464b1a879c"],
Cell[465746, 12008, 162, 3, 44, "Text",ExpressionUUID->"6e06c72a-772e-4df2-adc5-6c3ec47dee3f"],
Cell[CellGroupData[{
Cell[465933, 12015, 1058, 25, 64, "Input",ExpressionUUID->"05134b4a-647c-45fc-a865-ba3d57c68168"],
Cell[466994, 12042, 1221, 29, 70, "Output",ExpressionUUID->"64713bff-66ef-408e-ac32-075cf301a5a9"]
}, Open  ]],
Cell[468230, 12074, 2505, 51, 168, "Text",ExpressionUUID->"cb3ec2b4-cd18-4d16-8684-5a33b61da948"],
Cell[CellGroupData[{
Cell[470760, 12129, 1776, 47, 189, "Input",ExpressionUUID->"60c71e76-7525-4a3a-97d0-63b6944d2b63"],
Cell[472539, 12178, 1111, 29, 70, "Output",ExpressionUUID->"06bdfeba-862b-4c2b-8944-a32b7c8a8b12"]
}, Open  ]],
Cell[473665, 12210, 170, 3, 44, "Text",ExpressionUUID->"990c23e7-a271-4e5a-b33a-88dea689bc15"],
Cell[CellGroupData[{
Cell[473860, 12217, 159, 3, 37, "Input",ExpressionUUID->"92bc3fda-d49d-41c0-a32b-6a3de94c7e02"],
Cell[474022, 12222, 533, 16, 70, "Output",ExpressionUUID->"15978260-aead-4eae-9ca7-cf3c57d812c7"]
}, Open  ]],
Cell[CellGroupData[{
Cell[474592, 12243, 313, 8, 37, "Input",ExpressionUUID->"43d153fb-eb05-4992-b2bc-6bc01f67ef36"],
Cell[474908, 12253, 928, 24, 70, "Output",ExpressionUUID->"8d54243d-b6e6-4f83-80f3-a786c04bac3b"]
}, Open  ]],
Cell[475851, 12280, 1420, 37, 217, "Text",ExpressionUUID->"d6b47106-1413-48c6-8b15-e893e39ce732"],
Cell[CellGroupData[{
Cell[477296, 12321, 1693, 44, 90, "Input",ExpressionUUID->"f81c340e-6f9a-495e-a4fb-f8522f1ec28b"],
Cell[478992, 12367, 953, 22, 70, "Output",ExpressionUUID->"a935214a-6eb2-4f38-a8f2-c4f423a3595f"],
Cell[479948, 12391, 963, 22, 70, "Output",ExpressionUUID->"b31606ee-5624-4ecc-b4e4-3f5af30febb2"],
Cell[480914, 12415, 619, 9, 70, "Output",ExpressionUUID->"6af46331-b7a5-43eb-afc5-15a21d117d38"]
}, Open  ]],
Cell[481548, 12427, 378, 7, 44, "Text",ExpressionUUID->"86fd9338-dd80-4568-b298-512a1d31063e"],
Cell[CellGroupData[{
Cell[481951, 12438, 746, 18, 37, "Input",ExpressionUUID->"8c892f84-4ed8-4152-959e-bd6460391a65"],
Cell[482700, 12458, 849, 21, 70, "Output",ExpressionUUID->"ecea66e5-5689-4cc3-b8ac-c9532b465ef3"]
}, Open  ]],
Cell[483564, 12482, 170, 3, 44, "Text",ExpressionUUID->"d36a44fc-c0bc-4ed2-a7e9-1177e72d41c7"],
Cell[CellGroupData[{
Cell[483759, 12489, 658, 15, 37, "Input",ExpressionUUID->"d61046b6-a41d-406a-926f-df53053cec8e"],
Cell[484420, 12506, 578, 12, 70, "Output",ExpressionUUID->"8747b626-48d0-46f3-97e6-e1eaa14bff1e"]
}, Open  ]]
}, Closed]],
Cell[CellGroupData[{
Cell[485047, 12524, 237, 4, 45, "Subsubsection",ExpressionUUID->"9448e96a-9b8d-4993-a76d-d262a1ec1243"],
Cell[485287, 12530, 200, 5, 44, "Text",ExpressionUUID->"5e5dbf94-c343-4c9f-a04b-a9a03660486d"],
Cell[CellGroupData[{
Cell[485512, 12539, 837, 22, 64, "Input",ExpressionUUID->"4d52337a-32db-4790-b7e5-808754491976"],
Cell[486352, 12563, 2293, 54, 70, "Output",ExpressionUUID->"9845ca6f-247f-4c2f-a282-e60f91764288"]
}, Open  ]],
Cell[488660, 12620, 3060, 67, 200, "Text",ExpressionUUID->"75774f4d-26ff-42a6-b382-e34a5c2c7e24"],
Cell[CellGroupData[{
Cell[491745, 12691, 2169, 56, 198, "Input",ExpressionUUID->"7e78625f-aa15-4880-88d4-b4d8b1e2b50b"],
Cell[493917, 12749, 1241, 33, 70, "Output",ExpressionUUID->"1b7b201f-282d-4ee2-a0b4-1da70df94894"]
}, Open  ]],
Cell[495173, 12785, 239, 5, 44, "Text",ExpressionUUID->"3e1c4ca1-1e86-4c4e-9d39-b51c02a3b3aa"],
Cell[CellGroupData[{
Cell[495437, 12794, 320, 7, 37, "Input",ExpressionUUID->"70f2be14-e7b7-480a-ab97-979e3cd0984b"],
Cell[495760, 12803, 1110, 33, 70, "Output",ExpressionUUID->"a06bc303-9f62-469f-b0e2-1f194523403c"]
}, Open  ]],
Cell[496885, 12839, 1561, 42, 246, "Text",ExpressionUUID->"b1e1aa0a-d623-4803-93b4-15419b2e8379"],
Cell[CellGroupData[{
Cell[498471, 12885, 2074, 57, 116, "Input",ExpressionUUID->"7f52a622-8937-410a-88ec-95958fc17686"],
Cell[500548, 12944, 1011, 27, 70, "Output",ExpressionUUID->"954154d8-a97c-42ba-8afa-41f7a9da4d51"],
Cell[501562, 12973, 1027, 28, 70, "Output",ExpressionUUID->"e5aa4eb8-c218-476b-ba22-904b8b4646b6"],
Cell[502592, 13003, 610, 13, 70, "Output",ExpressionUUID->"786a77f7-de8f-4b06-9fa4-31fda5552368"],
Cell[503205, 13018, 518, 9, 70, "Output",ExpressionUUID->"c68377aa-5875-43ae-ac14-f3d44a58a099"]
}, Open  ]],
Cell[503738, 13030, 424, 8, 44, "Text",ExpressionUUID->"2e9457fc-050d-4c87-833d-f09c218d7e7b"],
Cell[CellGroupData[{
Cell[504187, 13042, 818, 18, 37, "Input",ExpressionUUID->"ffcc99d6-a5f0-4ae8-a087-25cf2910c2ba"],
Cell[505008, 13062, 667, 11, 70, "Output",ExpressionUUID->"a2fc5805-70d9-4a63-84aa-b77ff4115eba"]
}, Open  ]],
Cell[505690, 13076, 822, 19, 73, "Text",ExpressionUUID->"fe1c83ee-be24-4e16-943d-a90f1fee73e1"],
Cell[CellGroupData[{
Cell[506537, 13099, 350, 7, 37, "Input",ExpressionUUID->"8d8b0218-772f-4eeb-a34b-589827e37067"],
Cell[506890, 13108, 1115, 30, 70, "Output",ExpressionUUID->"47d051a0-b998-4322-b2d0-2bd0cfada6c4"]
}, Open  ]]
}, Closed]]
}, Closed]]
}, Closed]],
Cell[CellGroupData[{
Cell[508078, 13146, 184, 3, 56, "Chapter",ExpressionUUID->"9ba9e339-5f1d-42fd-8e6e-d3494987d8b8"],
Cell[508265, 13151, 1312, 31, 90, "Input",ExpressionUUID->"548e99b5-8fd9-4ae2-a7c2-64982b0bc649"],
Cell[CellGroupData[{
Cell[509602, 13186, 208, 4, 56, "Subsubsection",ExpressionUUID->"7d1cc74b-1f3d-4fbf-b304-bd7a55bca76f"],
Cell[509813, 13192, 183, 3, 44, "Text",ExpressionUUID->"fa3eebbc-479f-4bc3-bfb9-9231b76136a9"],
Cell[CellGroupData[{
Cell[510021, 13199, 3016, 73, 246, "Input",ExpressionUUID->"66571b6d-e08d-4aef-9b29-1c627f42dff1"],
Cell[513040, 13274, 441, 11, 70, "Message",ExpressionUUID->"7a8bd29f-ad80-41f5-9ef3-b16cbab556a8"],
Cell[513484, 13287, 405, 10, 70, "Message",ExpressionUUID->"079f19e2-f100-4fb6-9bd5-c8e5b9dc5480"],
Cell[513892, 13299, 16995, 300, 70, "Output",ExpressionUUID->"a5c51404-aaa5-435b-8aa5-dbddc3ab8e58"]
}, Open  ]],
Cell[CellGroupData[{
Cell[530924, 13604, 6974, 178, 584, "Input",ExpressionUUID->"4c9ca0f4-6cb8-42e6-bb4c-6dcdc4cfdea2"],
Cell[537901, 13784, 55834, 1021, 70, "Output",ExpressionUUID->"a32d642c-a66e-427c-bf8b-7eb833dd75f9"]
}, Open  ]],
Cell[CellGroupData[{
Cell[593772, 14810, 22245, 490, 1182, "Input",ExpressionUUID->"8c2b3f1e-e249-476d-86d8-5c0252a58cd0"],
Cell[616020, 15302, 68515, 1230, 70, "Output",ExpressionUUID->"5d51ac5a-5023-4e5d-aa69-59f1a9a324be"],
Cell[684538, 16534, 53029, 960, 70, "Output",ExpressionUUID->"b0473242-a123-4e0d-b37b-a6bae7a6eaa9"]
}, Open  ]],
Cell[CellGroupData[{
Cell[737604, 17499, 544, 14, 37, "Input",ExpressionUUID->"8676f9bc-6683-46c6-b546-1e57bb6e6b69"],
Cell[738151, 17515, 123665, 2209, 70, "Output",ExpressionUUID->"8c0c23ee-7111-4c7b-ad71-11ce17a9c81d"]
}, Open  ]],
Cell[861831, 19727, 178, 3, 37, "Input",ExpressionUUID->"ba3545b6-dc5e-48ca-b776-6b5eb92fe946"]
}, Closed]],
Cell[CellGroupData[{
Cell[862046, 19735, 216, 4, 45, "Subsubsection",ExpressionUUID->"7fc3a320-42d1-49cb-83ff-688f08f0cd53"],
Cell[862265, 19741, 6897, 166, 402, "Input",ExpressionUUID->"75fbc257-4811-44f3-8f78-ca480e9a62ac"],
Cell[869165, 19909, 7858, 193, 454, "Input",ExpressionUUID->"79eda932-1e09-4bc1-a529-84819b1e2ba6"],
Cell[CellGroupData[{
Cell[877048, 20106, 797, 19, 37, "Input",ExpressionUUID->"8e8e219b-ab40-4c41-b0e1-d876ee75b557"],
Cell[877848, 20127, 60075, 1142, 1169, "Output",ExpressionUUID->"ea2f01aa-b0a2-4b32-bbb9-ecc663a0b60d"]
}, Open  ]],
Cell[937938, 21272, 5935, 153, 376, "Input",ExpressionUUID->"87999efd-af59-4b0a-b732-41545c56ff4c"],
Cell[943876, 21427, 5994, 151, 376, "Input",ExpressionUUID->"4099ea53-91e3-427c-92a1-4137dd3a0fc5"],
Cell[949873, 21580, 1560, 41, 90, "Input",ExpressionUUID->"acc6758e-b3de-48d8-b63f-0eac2af82e07"],
Cell[CellGroupData[{
Cell[951458, 21625, 1864, 51, 142, "Input",ExpressionUUID->"3dca4a35-c88b-497b-ae64-e3bd9f6696a6"],
Cell[953325, 21678, 59931, 1164, 1093, "Output",ExpressionUUID->"6325db03-fb28-45a9-98c4-8a3455f04c28"]
}, Open  ]]
}, Closed]],
Cell[CellGroupData[{
Cell[1013305, 22848, 181, 3, 45, "Subsubsection",ExpressionUUID->"21a6cac6-75f5-4e55-822c-5284ed2cd5f9"],
Cell[1013489, 22853, 4569, 62, 5212, "Input",ExpressionUUID->"60c79b22-7316-4011-8f89-cac020b7b7fc"],
Cell[1018061, 22917, 182, 3, 44, "Text",ExpressionUUID->"8fe71c00-62e9-43e1-90ab-956b048f3168"],
Cell[1018246, 22922, 1433, 31, 64, "Input",ExpressionUUID->"c36681d9-a9a3-4d1a-b71f-c237a134533b"],
Cell[CellGroupData[{
Cell[1019704, 22957, 4387, 104, 246, "Input",ExpressionUUID->"cd126c38-24a2-4d35-8221-8463fb431d8d"],
Cell[1024094, 23063, 4152, 67, 70, "Message",ExpressionUUID->"8f2818ca-8da8-44d6-b676-a1aa7b4fb700"],
Cell[1028249, 23132, 4185, 67, 70, "Message",ExpressionUUID->"33a9fc75-1dac-4d12-9811-08e97d74c169"],
Cell[1032437, 23201, 4117, 65, 70, "Message",ExpressionUUID->"a1d348a0-1634-4fff-8d06-e9029b0ef8f8"],
Cell[1036557, 23268, 602, 13, 70, "Message",ExpressionUUID->"2d016115-946c-46cc-a116-0ac21c068a58"]
}, Open  ]],
Cell[1037174, 23284, 3205, 55, 2482, "Input",ExpressionUUID->"a74d00b9-b60a-4cd6-ad5d-82aafbb3a80d"],
Cell[CellGroupData[{
Cell[1040404, 23343, 4422, 104, 194, "Input",ExpressionUUID->"e2a04a6a-a1cb-4268-a08e-d0828a036520"],
Cell[1044829, 23449, 4156, 67, 70, "Message",ExpressionUUID->"82d2c2ea-11ad-4be2-b954-32e487473612"],
Cell[1048988, 23518, 4191, 67, 70, "Message",ExpressionUUID->"9ce418c3-8f1e-43e6-a65d-2853efa2d6bd"],
Cell[1053182, 23587, 4121, 65, 70, "Message",ExpressionUUID->"c02ce28f-8a7b-40bf-9e74-fb267e325ba5"],
Cell[1057306, 23654, 608, 13, 70, "Message",ExpressionUUID->"7ffb3887-d6c1-4a3b-9fa9-269d87d5e4c3"]
}, Open  ]],
Cell[1057929, 23670, 4759, 65, 5212, "Input",ExpressionUUID->"2cad1aeb-c8b1-403a-8e2f-5a21b3436931"],
Cell[CellGroupData[{
Cell[1062713, 23739, 4906, 121, 272, "Input",ExpressionUUID->"7f436d6f-fe8c-4383-9ffd-7a415fb2a786"],
Cell[1067622, 23862, 4107, 65, 70, "Message",ExpressionUUID->"0d9cdaff-73ae-4ddd-a287-45fe2a537dad"],
Cell[1071732, 23929, 4140, 65, 70, "Message",ExpressionUUID->"d3803404-8a16-41f0-b819-b3fb923b1689"],
Cell[1075875, 23996, 4072, 63, 70, "Message",ExpressionUUID->"e53813d8-9ef9-4148-9473-fe5cba18a1af"],
Cell[1079950, 24061, 557, 11, 70, "Message",ExpressionUUID->"5e34575e-b1bc-4c6b-af71-4b25d0d82dd8"]
}, Open  ]],
Cell[CellGroupData[{
Cell[1080544, 24077, 4359, 109, 272, "Input",ExpressionUUID->"2168ced3-f738-483f-bf38-1fe97cb7500d"],
Cell[1084906, 24188, 103521, 2011, 1139, "Output",ExpressionUUID->"1d00c814-0062-4dc4-a8ed-9925d08da0fa"]
}, Open  ]]
}, Closed]]
}, Closed]]
}, Open  ]]
}
]
*)

