(* Content-type: application/vnd.wolfram.mathematica *)

(*** Wolfram Notebook File ***)
(* http://www.wolfram.com/nb *)

(* CreatedBy='Mathematica 12.0' *)

(*CacheID: 234*)
(* Internal cache information:
NotebookFileLineBreakTest
NotebookFileLineBreakTest
NotebookDataPosition[       158,          7]
NotebookDataLength[    190347,       4775]
NotebookOptionsPosition[    177565,       4567]
NotebookOutlinePosition[    177984,       4584]
CellTagsIndexPosition[    177941,       4581]
WindowFrame->Normal*)

(* Beginning of Notebook Content *)
Notebook[{

Cell[CellGroupData[{
Cell["\<\
X-linked meiotic drive can boost population size and persistence\
\>", "Title",
 CellChangeTimes->{{3.813651645117669*^9, 
  3.813651659263967*^9}},ExpressionUUID->"58a35404-c46a-4052-9e2f-\
f80fb751f245"],

Cell["Carl Mackintosh, Andrew Pomiankowski, and Michael F Scott", "Subtitle",
 CellChangeTimes->{{3.813651663235971*^9, 
  3.8136516773311377`*^9}},ExpressionUUID->"f5221977-f0e4-4d02-b1d3-\
4c538645795d"],

Cell[CellGroupData[{

Cell["Overview", "Section",
 CellChangeTimes->{{3.814018860757616*^9, 
  3.8140188621077747`*^9}},ExpressionUUID->"bb378928-8785-464b-8dab-\
ff0c8fc86d79"],

Cell["\<\
In this file, we explore an alternative form of density-dependence in which \
competition for resources can cause the population size to be depressed as \
population birth rate increases. Specifically, we assume that juvenile \
survival declines with the number of other juveniles. This means that very \
high birth rates can lead to extreme competition and reduce the number of \
juveniles that survival to adulthood compared to lower birth rates. That is, \
the strength of competition is an increasing function of population birth \
rates. \
\>", "Text",
 CellChangeTimes->{{3.814092458241453*^9, 3.814092462602716*^9}, {
  3.8140925373123817`*^9, 3.814092618614622*^9}, {3.814092851373975*^9, 
  3.81409299416639*^9}},ExpressionUUID->"250706bb-65ca-4f28-9a37-\
c5dbe94cb31c"]
}, Open  ]],

Cell[CellGroupData[{

Cell["Common Notation and Substitutions [ENTER FIRST] ", "Section",
 CellChangeTimes->{{3.814018923386217*^9, 3.8140189329701023`*^9}, {
  3.814019466209839*^9, 
  3.8140194721274967`*^9}},ExpressionUUID->"aeb6c763-514d-4534-b55f-\
747741b817bf"],

Cell[TextData[{
 "We track adult (diploid) genotypes at an X linked locus that experiences \
meiotic drive in males. There are therefore 5 genotypes: 3 female genotypes \
and 2 male genotypes. In this notebook we use upper case D for the wildtype \
allele (which could also be called ",
 Cell[BoxData[
  FormBox[
   SubscriptBox["S", "t"], TraditionalForm]],
  FormatType->"TraditionalForm",ExpressionUUID->
  "212874ac-e87e-48f6-88a9-09b4cb5b6722"],
 " or ",
 Cell[BoxData[
  FormBox["X", TraditionalForm]],
  FormatType->"TraditionalForm",ExpressionUUID->
  "51c197be-1066-4787-84ff-dd3f7080bb65"],
 ") and lower case d to give the meiotic drive allele (which could also be \
called ",
 Cell[BoxData[
  FormBox[
   SubscriptBox["S", "r"], TraditionalForm]],
  FormatType->"TraditionalForm",ExpressionUUID->
  "9e81384f-d35e-43c6-8e55-afa5f278f6e0"],
 " or ",
 Cell[BoxData[
  FormBox[
   SubscriptBox["X", "d"], TraditionalForm]],
  FormatType->"TraditionalForm",ExpressionUUID->
  "124164dd-8e29-419f-a9c9-93043636ca76"],
 "). The 5 adult genotypes are therefore\nfemales:\n",
 Cell[BoxData[
  FormBox[
   SubscriptBox["S", "t"], TraditionalForm]],
  FormatType->"TraditionalForm",ExpressionUUID->
  "c13f0fb7-097a-45ec-b87e-6f283967f81e"],
 Cell[BoxData[
  FormBox[
   SubscriptBox["S", "t"], TraditionalForm]],
  FormatType->"TraditionalForm",ExpressionUUID->
  "99596f32-f58e-4f8f-a641-27aebaa1c59a"],
 " = ",
 Cell[BoxData[
  FormBox["X", TraditionalForm]],ExpressionUUID->
  "ef2eb348-ee9f-4a8b-b061-bc33901ae632"],
 Cell[BoxData[
  FormBox["X", TraditionalForm]],ExpressionUUID->
  "a0364fca-962c-4a60-8023-adba80a06182"],
 "= DDf\n",
 Cell[BoxData[
  FormBox[
   SubscriptBox["S", "r"], TraditionalForm]],
  FormatType->"TraditionalForm",ExpressionUUID->
  "5bda6407-7042-4dc1-a7af-bd1282c2f5fd"],
 Cell[BoxData[
  FormBox[
   SubscriptBox["S", "t"], TraditionalForm]],
  FormatType->"TraditionalForm",ExpressionUUID->
  "2c43eb35-b7f5-4f32-b050-3b0dce6841f9"],
 " = ",
 Cell[BoxData[
  FormBox[
   SubscriptBox["X", "d"], TraditionalForm]],ExpressionUUID->
  "600a2552-e243-4b51-a2c6-66218136d558"],
 Cell[BoxData[
  FormBox["X", TraditionalForm]],ExpressionUUID->
  "9a2607dc-2320-474a-a0fc-d82bcf268b0a"],
 " = Ddf\n",
 Cell[BoxData[
  FormBox[
   SubscriptBox["S", "r"], TraditionalForm]],
  FormatType->"TraditionalForm",ExpressionUUID->
  "ccd28cc6-1582-4793-a01d-01025b442748"],
 Cell[BoxData[
  FormBox[
   SubscriptBox["S", "r"], TraditionalForm]],
  FormatType->"TraditionalForm",ExpressionUUID->
  "267f684d-a1b0-4e87-bada-1fc3e4feeaa3"],
 " = ",
 Cell[BoxData[
  FormBox[
   SubscriptBox["X", "d"], TraditionalForm]],ExpressionUUID->
  "c5491461-ea4c-4179-9940-3b6db6c44c88"],
 Cell[BoxData[
  FormBox[
   SubscriptBox["X", "d"], TraditionalForm]],ExpressionUUID->
  "075c8eb8-e3eb-4c5f-92dc-0544299656be"],
 " = ddf\nmales:\n",
 Cell[BoxData[
  FormBox[
   SubscriptBox["S", "t"], TraditionalForm]],ExpressionUUID->
  "33a0b17b-1f7b-4d87-a12e-79d4e737b074"],
 "Y = ",
 Cell[BoxData[
  FormBox["X", TraditionalForm]],ExpressionUUID->
  "c520cf04-0cf4-4092-853e-f389231a1769"],
 Cell[BoxData[
  FormBox["Y", TraditionalForm]],ExpressionUUID->
  "f537ceda-a7e7-4ffb-aa1c-b6909a243517"],
 " = Dm\n",
 Cell[BoxData[
  FormBox[
   SubscriptBox["S", "r"], TraditionalForm]],ExpressionUUID->
  "6263411e-cd1e-4fcb-b5ef-d1120f86e378"],
 "Y = ",
 Cell[BoxData[
  FormBox[
   SubscriptBox["X", "d"], TraditionalForm]],ExpressionUUID->
  "a17586b3-5873-4d9e-8952-60b0d9899359"],
 Cell[BoxData[
  FormBox["Y", TraditionalForm]],ExpressionUUID->
  "1c292321-29b3-43b1-b0cf-c08a6e1a1504"],
 " = dm\nsuch that\nijf is the density of females with genotype ij \nim is \
the density of males with genotype i"
}], "Text",
 CellChangeTimes->{{3.8140190605131702`*^9, 3.814019392288312*^9}, {
  3.8140757952060537`*^9, 3.8140758541839437`*^9}, {3.814077180589332*^9, 
  3.81407720443826*^9}, {3.81407730059503*^9, 
  3.814077315912006*^9}},ExpressionUUID->"828fce9c-f220-461b-a1c4-\
86c2ce6a2967"],

Cell[TextData[{
 "We assume that the X-linked meiotic drive allele biases transmission in \
males, such that a ",
 Cell[BoxData[
  RowBox[{"(", 
   FractionBox[
    RowBox[{"(", 
     RowBox[{"1", "+", "\[Delta]"}], ")"}], "2"], ")"}]],
  CellChangeTimes->{{3.7814428720012417`*^9, 3.781442894240843*^9}, {
   3.814008329923299*^9, 3.8140083738240767`*^9}, {3.814019442825118*^9, 
   3.814019443216568*^9}, {3.81401966596344*^9, 3.814019666468874*^9}, {
   3.81407591820485*^9, 3.814075935309269*^9}, {3.814075978592771*^9, 
   3.814075985638269*^9}, {3.814079326819665*^9, 3.8140793341532917`*^9}, {
   3.8140794393433743`*^9, 3.814079615488195*^9}, {3.8140796473810863`*^9, 
   3.8140798375136547`*^9}},ExpressionUUID->
  "2640c78b-a689-4f7c-b23f-2893f515ef09"],
 "fraction of gametes produced by a meiotic drive male (dm) carry the drive \
allele (d). Furthermore, we assume that the ejaculate size of meiotic drive \
males, relative to wildtype males, is c. "
}], "Text",
 CellChangeTimes->{{3.814080029254443*^9, 
  3.8140801640639563`*^9}},ExpressionUUID->"e76d593a-4d23-4f5f-ad8e-\
d99b12ad4c5c"],

Cell["\<\
We  use the following substitutions to convert between these genotype \
densities and the total density of males (maleNum) and females (femaleNum), \
and the frequency of the drive allele among males (pm) and females (pf), and \
the inbreeding coefficient (F). These quantities are defined by \
substitutionEquations, subeq can be used to convert from the genotype \
densities into male/female densities and allele frequencies. \
\>", "Text",
 CellChangeTimes->{{3.814077206472991*^9, 3.814077212502678*^9}, {
  3.814077289440152*^9, 3.814077399069392*^9}, {3.81407752412192*^9, 
  3.814077623571116*^9}},ExpressionUUID->"5338ade1-35b0-4064-b584-\
adb8676ffbc3"],

Cell[CellGroupData[{

Cell[BoxData[{
 RowBox[{
  RowBox[{"substitutionEquations", "=", 
   RowBox[{"{", "\[IndentingNewLine]", 
    RowBox[{
     RowBox[{"femaleNum", "==", 
      RowBox[{"DDf", "+", "Ddf", "+", "ddf"}]}], ",", "\[IndentingNewLine]", 
     RowBox[{"maleNum", "\[Equal]", 
      RowBox[{"(", 
       RowBox[{"dm", "+", "Dm"}], ")"}]}], ",", "\[IndentingNewLine]", 
     RowBox[{"pm", "==", 
      FractionBox["dm", 
       RowBox[{"dm", "+", "Dm"}]]}], ",", "\[IndentingNewLine]", 
     RowBox[{
      RowBox[{"(", 
       RowBox[{
        RowBox[{
         RowBox[{
          RowBox[{"(", 
           RowBox[{"1", "-", "pf"}], ")"}], "^", "2"}], 
         RowBox[{"(", 
          RowBox[{"1", "-", "F"}], ")"}]}], "+", 
        RowBox[{"F", 
         RowBox[{"(", 
          RowBox[{"1", "-", "pf"}], ")"}]}]}], ")"}], "==", 
      FractionBox["DDf", 
       RowBox[{"DDf", "+", "Ddf", "+", "ddf"}]]}], ",", "\[IndentingNewLine]", 
     RowBox[{
      RowBox[{"2", "pf", 
       RowBox[{"(", 
        RowBox[{"1", "-", "pf"}], ")"}], 
       RowBox[{"(", 
        RowBox[{"1", "-", "F"}], ")"}]}], "==", 
      FractionBox["Ddf", 
       RowBox[{"DDf", "+", "Ddf", "+", "ddf"}]]}], ",", "\[IndentingNewLine]", 
     RowBox[{
      RowBox[{
       RowBox[{
        RowBox[{
         RowBox[{"(", "pf", ")"}], "^", "2"}], 
        RowBox[{"(", 
         RowBox[{"1", "-", "F"}], ")"}]}], "+", 
       RowBox[{"F", " ", "pf"}]}], "==", 
      FractionBox["ddf", 
       RowBox[{"DDf", "+", "Ddf", "+", "ddf"}]]}]}], "\[IndentingNewLine]", 
    "}"}]}], ";"}], "\[IndentingNewLine]", 
 RowBox[{"subeq", "=", 
  RowBox[{"Simplify", "[", 
   RowBox[{"Solve", "[", 
    RowBox[{"substitutionEquations", ",", 
     RowBox[{"{", 
      RowBox[{"dm", ",", "Dm", ",", "DDf", ",", "Ddf", ",", "ddf"}], "}"}]}], 
    "]"}], "]"}]}]}], "Input",
 CellChangeTimes->{{3.814077223260071*^9, 3.8140772805761766`*^9}, {
  3.814077501177114*^9, 3.814077520478444*^9}},
 CellLabel->"In[1]:=",ExpressionUUID->"9abe40f9-a542-40db-9ab7-232aecfce29b"],

Cell[BoxData[
 RowBox[{"{", 
  RowBox[{"{", 
   RowBox[{
    RowBox[{"dm", "\[Rule]", 
     RowBox[{"maleNum", " ", "pm"}]}], ",", 
    RowBox[{"Dm", "\[Rule]", 
     RowBox[{"maleNum", "-", 
      RowBox[{"maleNum", " ", "pm"}]}]}], ",", 
    RowBox[{"DDf", "\[Rule]", 
     RowBox[{
      RowBox[{"-", "femaleNum"}], " ", 
      RowBox[{"(", 
       RowBox[{
        RowBox[{"-", "1"}], "+", "pf"}], ")"}], " ", 
      RowBox[{"(", 
       RowBox[{"1", "+", 
        RowBox[{
         RowBox[{"(", 
          RowBox[{
           RowBox[{"-", "1"}], "+", "F"}], ")"}], " ", "pf"}]}], ")"}]}]}], 
    ",", 
    RowBox[{"Ddf", "\[Rule]", 
     RowBox[{"2", " ", 
      RowBox[{"(", 
       RowBox[{
        RowBox[{"-", "1"}], "+", "F"}], ")"}], " ", "femaleNum", " ", 
      RowBox[{"(", 
       RowBox[{
        RowBox[{"-", "1"}], "+", "pf"}], ")"}], " ", "pf"}]}], ",", 
    RowBox[{"ddf", "\[Rule]", 
     RowBox[{"femaleNum", " ", "pf", " ", 
      RowBox[{"(", 
       RowBox[{"F", "+", "pf", "-", 
        RowBox[{"F", " ", "pf"}]}], ")"}]}]}]}], "}"}], "}"}]], "Output",
 CellChangeTimes->{{3.814077219052555*^9, 3.814077280927372*^9}, 
   3.814077589788096*^9, 3.814080197254603*^9, 3.814098265528617*^9},
 CellLabel->"Out[2]=",ExpressionUUID->"c65020f7-9f22-4315-bbe1-9e9069f32f3a"]
}, Open  ]],

Cell["\<\
Each female will survive to produce b surviving juveniles. 
Density dependence acts to reduce the survival of juveniles and declines with \
the number of other juveniles. Specifically, juvenile survival declines by a \
factor of (1-\[Alpha] N) where N is the total density of male and female \
juveniles. That is \[Alpha] is the per juvenile competitive effect on \
juvenile survival.\
\>", "Text",
 CellChangeTimes->{{3.814078082733198*^9, 3.814078132242939*^9}, {
  3.814078167961803*^9, 3.814078323095675*^9}, {3.814078358488497*^9, 
  3.814078360367148*^9}, {3.814078400246647*^9, 3.81407846691296*^9}, {
  3.814078506703103*^9, 3.814078506939747*^9}, {3.814078603846031*^9, 
  3.8140786211249657`*^9}, {3.814092748910684*^9, 
  3.8140928257117863`*^9}},ExpressionUUID->"61e643bf-7a0e-480f-b76d-\
3f5721dbcd4e"],

Cell[TextData[{
 "Before juveniles reach adulthood, they also experience viability selection \
according to their genotype. The relative fitness of different genotypes is \
given by\n",
 Cell[BoxData[
  FormBox[
   RowBox[{
    SubscriptBox["W", "XX"], "=", "1"}], TraditionalForm]],ExpressionUUID->
  "531197fe-f329-4d91-b0eb-a3900e1b0b35"],
 "\n",
 Cell[BoxData[
  FormBox[
   RowBox[{
    SubscriptBox["W", 
     RowBox[{
      SubscriptBox["X", "d"], "X"}]], "=", 
    RowBox[{"1", "-", 
     RowBox[{"h", " ", "sf"}]}]}], TraditionalForm]],ExpressionUUID->
  "bdee0304-7603-4576-8083-9dfa457a5eab"],
 "\n",
 Cell[BoxData[
  FormBox[
   RowBox[{
    SubscriptBox["W", 
     RowBox[{
      SubscriptBox["X", "d"], 
      SubscriptBox["X", "d"]}]], "=", 
    RowBox[{"1", "-", "sf"}]}], TraditionalForm]],ExpressionUUID->
  "fa5ae154-32cf-4d16-930b-a9125b080b96"],
 "\n",
 Cell[BoxData[
  FormBox[
   RowBox[{
    SubscriptBox["W", "XY"], "=", "1"}], TraditionalForm]],ExpressionUUID->
  "83a8d333-9b2a-4aaf-892b-d7ee1656d528"],
 "\n",
 Cell[BoxData[
  FormBox[
   RowBox[{
    SubscriptBox["W", 
     RowBox[{
      SubscriptBox["X", "d"], "Y"}]], "=", 
    RowBox[{"1", "-", "sm"}]}], TraditionalForm]],ExpressionUUID->
  "d22704f8-577e-4554-bcc5-5d41c9999178"],
 "\nWe use ijfprime and imprime to specify the densities of each genotype \
after one generation (recursion/difference equations for change in genotype \
densities). For example, DDfprime is the density of females with genotype DD \
after one generation. "
}], "Text",
 CellChangeTimes->{{3.814078662208776*^9, 3.8140787408128223`*^9}, {
  3.81407914095951*^9, 3.814079300391842*^9}, {3.814080008982128*^9, 
  3.8140800183201637`*^9}, {3.814092832822933*^9, 
  3.814092836860931*^9}},ExpressionUUID->"dfd6dffa-4c34-45ab-9f7f-\
84084ef3187e"]
}, Closed]],

Cell[CellGroupData[{

Cell[TextData[{
 "Allele Frequency and Population Size Dynamics - females mate with one male \
(",
 Cell[BoxData[
  FormBox[
   RowBox[{
    SubscriptBox["\[Lambda]", "f"], "=", "1"}], TraditionalForm]],
  FormatType->"TraditionalForm",ExpressionUUID->
  "25078cc5-50ec-4322-b61d-66322bbb02dc"],
 ")"
}], "Section",
 CellChangeTimes->{{3.769852030466571*^9, 3.769852037504027*^9}, {
   3.810993843290845*^9, 3.8109938547752533`*^9}, 
   3.814092638937703*^9},ExpressionUUID->"81f0ebfe-5142-4ce4-b8ce-\
04c3cfc325bd"],

Cell[CellGroupData[{

Cell["Recursion Equations", "Subsection",
 CellChangeTimes->{{3.769855665745242*^9, 3.7698556686753263`*^9}, {
  3.769861921705832*^9, 3.769861926034112*^9}, {3.8140999862888927`*^9, 
  3.8140999867493553`*^9}},ExpressionUUID->"f54b6010-1fa4-4a81-89ed-\
8ae2fa491764"],

Cell[TextData[{
 "We consider each pairwise combination of male and female genotypes and the \
probability that each mating type produces each offspring genotype. We assume \
random mating so that each mating combination happens in proportion to the \
frequency of males with each genotype (e.g., ",
 Cell[BoxData[
  FractionBox["dm", 
   RowBox[{"(", 
    RowBox[{"Dm", "+", "dm"}], ")"}]]],
  CellChangeTimes->{{3.7814428720012417`*^9, 3.781442894240843*^9}, {
   3.814008329923299*^9, 3.8140083738240767`*^9}, {3.814019442825118*^9, 
   3.814019443216568*^9}, {3.81401966596344*^9, 3.814019666468874*^9}, {
   3.81407591820485*^9, 3.814075935309269*^9}, {3.814075978592771*^9, 
   3.814075985638269*^9}, {3.814079326819665*^9, 3.8140793341532917`*^9}, {
   3.8140794393433743`*^9, 3.814079615488195*^9}, {3.8140796473810863`*^9, 
   3.8140798375136547`*^9}},ExpressionUUID->
  "e96985d1-306b-4346-90a8-c28265c4007c"],
 "is the frequency of drive males among males). These equations track all \
mating combinations that produce each juvenile genotype. The rate of birth is \
not directly dependent on genotype. "
}], "Text",
 CellChangeTimes->{{3.769852649608509*^9, 3.7698528245115623`*^9}, {
  3.7698528744863863`*^9, 3.7698528770112886`*^9}, {3.769855237904292*^9, 
  3.769855279351633*^9}, {3.769861951370813*^9, 3.76986197473839*^9}, {
  3.814093092507104*^9, 
  3.8140931116481752`*^9}},ExpressionUUID->"07f8cc85-4729-4c30-963f-\
5b1fc2a2df3d"],

Cell[BoxData[{
 RowBox[{
  RowBox[{"DDfjuveniles", "=", 
   RowBox[{"b", "*", 
    RowBox[{"(", 
     RowBox[{
      RowBox[{"DDf", "*", 
       FractionBox["Dm", 
        RowBox[{"(", 
         RowBox[{"Dm", "+", "dm"}], ")"}]], "*", 
       FractionBox["1", "2"]}], " ", "+", "\[IndentingNewLine]", 
      RowBox[{
       FractionBox["Ddf", "2"], "*", 
       FractionBox["Dm", 
        RowBox[{"(", 
         RowBox[{"Dm", "+", "dm"}], ")"}]], "*", 
       FractionBox["1", "2"]}]}], ")"}]}]}], ";"}], "\[IndentingNewLine]", 
 RowBox[{
  RowBox[{"Ddfjuveniles", "=", 
   RowBox[{"b", "*", 
    RowBox[{"(", 
     RowBox[{
      RowBox[{"DDf", "*", 
       FractionBox["dm", 
        RowBox[{"(", 
         RowBox[{"Dm", "+", "dm"}], ")"}]], "*", 
       RowBox[{"(", 
        FractionBox[
         RowBox[{"(", 
          RowBox[{"1", "+", "\[Delta]"}], ")"}], "2"], ")"}]}], "+", 
      "\[IndentingNewLine]", 
      RowBox[{
       FractionBox["Ddf", "2"], 
       RowBox[{"(", 
        RowBox[{
         RowBox[{
          FractionBox["Dm", 
           RowBox[{"(", 
            RowBox[{"Dm", "+", "dm"}], ")"}]], "*", 
          FractionBox["1", "2"]}], "+", 
         RowBox[{
          FractionBox["dm", 
           RowBox[{"(", 
            RowBox[{"Dm", "+", "dm"}], ")"}]], "*", 
          RowBox[{"(", 
           FractionBox[
            RowBox[{"(", 
             RowBox[{"1", "+", "\[Delta]"}], ")"}], "2"], ")"}]}]}], ")"}]}], 
      "+", "\[IndentingNewLine]", 
      RowBox[{"ddf", "*", 
       FractionBox["Dm", 
        RowBox[{"(", 
         RowBox[{"Dm", "+", "dm"}], ")"}]], "*", 
       FractionBox["1", "2"]}]}], ")"}]}]}], ";"}], "\[IndentingNewLine]", 
 RowBox[{
  RowBox[{"ddfjuveniles", "=", 
   RowBox[{"b", "*", 
    RowBox[{"(", 
     RowBox[{
      RowBox[{
       FractionBox["Ddf", "2"], "*", 
       FractionBox["dm", 
        RowBox[{"(", 
         RowBox[{"Dm", "+", "dm"}], ")"}]], "*", 
       RowBox[{"(", 
        FractionBox[
         RowBox[{"(", 
          RowBox[{"1", "+", "\[Delta]"}], ")"}], "2"], ")"}]}], "+", 
      "\[IndentingNewLine]", 
      RowBox[{"ddf", "*", 
       FractionBox["dm", 
        RowBox[{"(", 
         RowBox[{"Dm", "+", "dm"}], ")"}]], "*", 
       RowBox[{"(", 
        FractionBox[
         RowBox[{"(", 
          RowBox[{"1", "+", "\[Delta]"}], ")"}], "2"], ")"}]}]}], ")"}]}]}], 
  ";"}], "\[IndentingNewLine]", 
 RowBox[{
  RowBox[{"Dmjuveniles", "=", 
   RowBox[{"b", "*", 
    RowBox[{"(", 
     RowBox[{
      RowBox[{"DDf", "*", 
       FractionBox["Dm", 
        RowBox[{"(", 
         RowBox[{"Dm", "+", "dm"}], ")"}]], "*", 
       FractionBox["1", "2"]}], "+", "\[IndentingNewLine]", 
      RowBox[{
       FractionBox["Ddf", "2"], "*", 
       RowBox[{"(", 
        RowBox[{
         RowBox[{
          FractionBox["Dm", 
           RowBox[{"(", 
            RowBox[{"Dm", "+", "dm"}], ")"}]], "*", 
          FractionBox["1", "2"]}], "+", 
         RowBox[{
          FractionBox["dm", 
           RowBox[{"(", 
            RowBox[{"Dm", "+", "dm"}], ")"}]], "*", 
          RowBox[{"(", 
           FractionBox[
            RowBox[{"(", 
             RowBox[{"1", "-", "\[Delta]"}], ")"}], "2"], ")"}]}]}], ")"}]}], 
      "+", "\[IndentingNewLine]", 
      RowBox[{"DDf", "*", 
       FractionBox["dm", 
        RowBox[{"(", 
         RowBox[{"Dm", "+", "dm"}], ")"}]], "*", 
       RowBox[{"(", 
        FractionBox[
         RowBox[{"(", 
          RowBox[{"1", "-", "\[Delta]"}], ")"}], "2"], ")"}]}]}], ")"}]}]}], 
  ";"}], "\[IndentingNewLine]", 
 RowBox[{
  RowBox[{"dmjuveniles", "=", 
   RowBox[{"b", "*", 
    RowBox[{"(", 
     RowBox[{
      RowBox[{
       FractionBox["Ddf", "2"], "*", 
       RowBox[{"(", 
        RowBox[{
         RowBox[{
          FractionBox["dm", 
           RowBox[{"(", 
            RowBox[{"Dm", "+", "dm"}], ")"}]], "*", 
          RowBox[{"(", 
           FractionBox[
            RowBox[{"(", 
             RowBox[{"1", "-", "\[Delta]"}], ")"}], "2"], ")"}]}], "+", 
         RowBox[{
          FractionBox["Dm", 
           RowBox[{"(", 
            RowBox[{"Dm", "+", "dm"}], ")"}]], "*", 
          FractionBox["1", "2"]}]}], ")"}]}], "+", "\[IndentingNewLine]", 
      RowBox[{"ddf", "*", 
       RowBox[{"(", 
        RowBox[{
         RowBox[{
          FractionBox["Dm", 
           RowBox[{"(", 
            RowBox[{"Dm", "+", "dm"}], ")"}]], 
          FractionBox["1", "2"]}], "+", 
         RowBox[{
          FractionBox["dm", 
           RowBox[{"(", 
            RowBox[{"Dm", "+", "dm"}], ")"}]], "*", 
          RowBox[{"(", 
           FractionBox[
            RowBox[{"(", 
             RowBox[{"1", "-", "\[Delta]"}], ")"}], "2"], ")"}]}]}], 
        ")"}]}]}], ")"}]}]}], ";"}]}], "Input",
 CellChangeTimes->{{3.769852829076851*^9, 3.7698528517796507`*^9}, {
  3.769852882136455*^9, 3.769852903427639*^9}, {3.769853083166275*^9, 
  3.769853094771209*^9}, {3.7698535488533363`*^9, 3.769853643995084*^9}, {
  3.7698537026503696`*^9, 3.769853703890132*^9}, {3.769853761936763*^9, 
  3.7698540306666117`*^9}, {3.7698541172779818`*^9, 3.769854120625928*^9}, {
  3.769861984777997*^9, 3.769861988603487*^9}, {3.7698620232984533`*^9, 
  3.769862064098259*^9}, {3.769862099032528*^9, 3.7698621100790167`*^9}, {
  3.7698686749830723`*^9, 3.769868735624775*^9}, {3.7699432542587833`*^9, 
  3.769943307197878*^9}, {3.770469703535376*^9, 3.7704697037099543`*^9}, {
  3.814093054683201*^9, 3.814093054685388*^9}, {3.814093146834547*^9, 
  3.814093161540985*^9}, {3.814098451050337*^9, 3.814098461501289*^9}},
 CellLabel->"In[3]:=",ExpressionUUID->"13fe85f1-f84b-4e25-8931-9e84a64c2555"],

Cell["\<\
now survival, where density dependence is a function of juvenile number now \
(and so a function of birth rate)\
\>", "Text",
 CellChangeTimes->{{3.769852860810169*^9, 3.76985286325079*^9}, {
  3.769854402168894*^9, 3.76985440435671*^9}, {3.7698621623671827`*^9, 
  3.7698621667967663`*^9}, {3.769880768035132*^9, 3.76988076880515*^9}, {
  3.810993938994034*^9, 
  3.810993971111459*^9}},ExpressionUUID->"8793f039-88ed-449c-9e97-\
718fc109fc89"],

Cell[BoxData[{
 RowBox[{
  RowBox[{"totJuveniles", "=", 
   RowBox[{
   "DDfjuveniles", "+", "Ddfjuveniles", "+", "ddfjuveniles", "+", 
    "Dmjuveniles", "+", "dmjuveniles"}]}], ";"}], "\[IndentingNewLine]", 
 RowBox[{
  RowBox[{"DDfprime", "=", 
   RowBox[{"DDfjuveniles", 
    RowBox[{"(", 
     RowBox[{"1", "-", 
      RowBox[{"\[Alpha]", " ", "totJuveniles"}]}], ")"}], "1"}]}], 
  ";"}], "\[IndentingNewLine]", 
 RowBox[{
  RowBox[{"Ddfprime", "=", 
   RowBox[{"Ddfjuveniles", 
    RowBox[{"(", 
     RowBox[{"1", "-", 
      RowBox[{"\[Alpha]", " ", "totJuveniles"}]}], ")"}], 
    RowBox[{"(", 
     RowBox[{"1", " ", "-", " ", 
      RowBox[{"h", "*", "sf"}]}], ")"}]}]}], ";"}], "\[IndentingNewLine]", 
 RowBox[{
  RowBox[{"ddfprime", "=", 
   RowBox[{"ddfjuveniles", 
    RowBox[{"(", 
     RowBox[{"1", "-", 
      RowBox[{"\[Alpha]", " ", "totJuveniles"}]}], ")"}], 
    RowBox[{"(", 
     RowBox[{"1", " ", "-", " ", "sf"}], ")"}]}]}], 
  ";"}], "\[IndentingNewLine]", 
 RowBox[{
  RowBox[{"Dmprime", "=", 
   RowBox[{"Dmjuveniles", 
    RowBox[{"(", 
     RowBox[{"1", "-", 
      RowBox[{"\[Alpha]", " ", "totJuveniles"}]}], ")"}], "1"}]}], 
  ";"}], "\[IndentingNewLine]", 
 RowBox[{
  RowBox[{"dmprime", "=", 
   RowBox[{"dmjuveniles", 
    RowBox[{"(", 
     RowBox[{"1", "-", 
      RowBox[{"\[Alpha]", " ", "totJuveniles"}]}], ")"}], 
    RowBox[{"(", 
     RowBox[{"1", " ", "-", " ", "sm"}], ")"}]}]}], ";"}]}], "Input",
 CellChangeTimes->{{3.769862168165381*^9, 3.769862288989339*^9}, 
   3.77235979764887*^9, {3.772359947256694*^9, 3.772360000583913*^9}, 
   3.772360427028974*^9},
 CellLabel->"In[8]:=",ExpressionUUID->"a558d4a4-e400-42df-8228-835e16bf1b63"]
}, Closed]],

Cell[CellGroupData[{

Cell["Equilibrium allele frequency and population size", "Subsection",
 CellChangeTimes->{{3.7699381302033567`*^9, 3.7699381325000343`*^9}, {
   3.769941882879545*^9, 3.769941892061565*^9}, 
   3.8140999922123957`*^9},ExpressionUUID->"bcd72549-4783-409a-91a8-\
53925885f135"],

Cell["\<\
We first look at only the difference in allele frequency and inbreeding \
coefficient to calculate the polymorphic allele frequency equilibrium (when \
changes in allele frequency and inbreeding coefficient are 0)\
\>", "Text",
 CellChangeTimes->{{3.769938591610197*^9, 3.769938597319046*^9}, 
   3.8140986623692017`*^9},ExpressionUUID->"1f0ac6dc-e0b2-4ba8-9330-\
cf03c3724fd2"],

Cell[CellGroupData[{

Cell[BoxData[{
 RowBox[{
  RowBox[{"solveme", "=", 
   RowBox[{"Simplify", "[", 
    RowBox[{
     RowBox[{"{", 
      RowBox[{
       RowBox[{
        RowBox[{"(", 
         FractionBox["dmprime", 
          RowBox[{"Dmprime", "+", "dmprime"}]], ")"}], "-", 
        RowBox[{"(", 
         FractionBox["dm", 
          RowBox[{"dm", "+", "Dm"}]], ")"}]}], ",", 
       RowBox[{
        RowBox[{"(", 
         FractionBox[
          RowBox[{
           RowBox[{"2", " ", "ddfprime"}], "+", "Ddfprime"}], 
          RowBox[{"2", " ", 
           RowBox[{"(", 
            RowBox[{"ddfprime", "+", "Ddfprime", "+", "DDfprime"}], ")"}]}]], 
         ")"}], "-", 
        RowBox[{"(", 
         FractionBox[
          RowBox[{
           RowBox[{"2", " ", "ddf"}], "+", "Ddf"}], 
          RowBox[{"2", " ", 
           RowBox[{"(", 
            RowBox[{"ddf", "+", "Ddf", "+", "DDf"}], ")"}]}]], ")"}]}], ",", 
       RowBox[{
        RowBox[{"(", 
         FractionBox[
          RowBox[{
           RowBox[{"4", " ", "ddfprime", " ", "DDfprime"}], " ", "-", 
           SuperscriptBox["Ddfprime", "2"]}], 
          RowBox[{
           RowBox[{"(", 
            RowBox[{
             RowBox[{"2", " ", "ddfprime"}], "+", "Ddfprime"}], ")"}], " ", 
           RowBox[{"(", 
            RowBox[{"Ddfprime", "+", 
             RowBox[{"2", " ", "DDfprime"}]}], ")"}]}]], ")"}], "-", 
        RowBox[{"(", 
         FractionBox[
          RowBox[{
           RowBox[{"4", " ", "ddf", " ", "DDf"}], "-", 
           SuperscriptBox["Ddf", "2"]}], 
          RowBox[{
           RowBox[{"(", 
            RowBox[{
             RowBox[{"2", " ", "ddf"}], "+", "Ddf"}], ")"}], " ", 
           RowBox[{"(", 
            RowBox[{"Ddf", "+", 
             RowBox[{"2", " ", "DDf"}]}], ")"}]}]], ")"}]}]}], "}"}], "/.", 
     "subeq"}], "]"}]}], ";"}], "\[IndentingNewLine]", 
 RowBox[{"equilibrium", "=", 
  RowBox[{
   RowBox[{"Solve", "[", 
    RowBox[{
     RowBox[{
      RowBox[{"solveme", "[", 
       RowBox[{"[", "1", "]"}], "]"}], "\[Equal]", 
      RowBox[{"{", 
       RowBox[{"0", ",", "0", ",", "0"}], "}"}]}], ",", 
     RowBox[{"{", 
      RowBox[{"pm", ",", "pf", ",", "F"}], "}"}]}], "]"}], "//", 
   "Simplify"}]}]}], "Input",
 CellChangeTimes->{{3.769886175554981*^9, 3.769886204822073*^9}, {
  3.7698862445919847`*^9, 3.769886289012404*^9}, {3.769886438108795*^9, 
  3.769886438482112*^9}, {3.7698864884062853`*^9, 3.7698864893141127`*^9}, {
  3.7698865912608433`*^9, 3.7698865926293077`*^9}, {3.7698868650517*^9, 
  3.76988686519938*^9}, {3.769938705473711*^9, 3.769938752318252*^9}, {
  3.8140985238648443`*^9, 3.814098560197666*^9}},
 CellLabel->"In[14]:=",ExpressionUUID->"910658e9-a5fb-493a-958f-d6083fe6bed0"],

Cell[BoxData[
 RowBox[{"{", 
  RowBox[{"{", 
   RowBox[{
    RowBox[{"pm", "\[Rule]", 
     FractionBox[
      RowBox[{
       RowBox[{"(", 
        RowBox[{
         RowBox[{"-", "1"}], "+", "sm"}], ")"}], " ", 
       RowBox[{"(", 
        RowBox[{"\[Delta]", "-", 
         RowBox[{"sm", " ", 
          RowBox[{"(", 
           RowBox[{"1", "+", "\[Delta]"}], ")"}]}], "+", 
         RowBox[{"h", " ", "sf", " ", 
          RowBox[{"(", 
           RowBox[{
            RowBox[{"-", "2"}], "+", "sm", "-", "\[Delta]", "+", 
            RowBox[{"sm", " ", "\[Delta]"}]}], ")"}]}]}], ")"}]}], 
      RowBox[{
       RowBox[{"2", " ", "sf", " ", 
        RowBox[{"(", 
         RowBox[{
          RowBox[{"-", "1"}], "+", "sm"}], ")"}], " ", 
        RowBox[{"(", 
         RowBox[{"1", "+", "\[Delta]"}], ")"}]}], "+", 
       RowBox[{"h", " ", "sf", " ", 
        RowBox[{"(", 
         RowBox[{
          RowBox[{"-", "2"}], "+", "sm"}], ")"}], " ", 
        RowBox[{"(", 
         RowBox[{
          RowBox[{"-", "2"}], "+", "sm", "-", "\[Delta]", "+", 
          RowBox[{"sm", " ", "\[Delta]"}]}], ")"}]}], "-", 
       RowBox[{"sm", " ", 
        RowBox[{"(", 
         RowBox[{"sm", "-", "\[Delta]", "+", 
          RowBox[{"sm", " ", "\[Delta]"}]}], ")"}]}]}]]}], ",", 
    RowBox[{"pf", "\[Rule]", 
     FractionBox[
      RowBox[{"\[Delta]", "-", 
       RowBox[{"sm", " ", 
        RowBox[{"(", 
         RowBox[{"1", "+", "\[Delta]"}], ")"}]}], "+", 
       RowBox[{"h", " ", "sf", " ", 
        RowBox[{"(", 
         RowBox[{
          RowBox[{"-", "2"}], "+", "sm", "-", "\[Delta]", "+", 
          RowBox[{"sm", " ", "\[Delta]"}]}], ")"}]}]}], 
      RowBox[{"2", " ", "sf", " ", 
       RowBox[{"(", 
        RowBox[{
         RowBox[{
          RowBox[{"-", 
           RowBox[{"(", 
            RowBox[{
             RowBox[{"-", "1"}], "+", "sm"}], ")"}]}], " ", 
          RowBox[{"(", 
           RowBox[{"1", "+", "\[Delta]"}], ")"}]}], "+", 
         RowBox[{"h", " ", 
          RowBox[{"(", 
           RowBox[{
            RowBox[{"-", "2"}], "+", "sm", "-", "\[Delta]", "+", 
            RowBox[{"sm", " ", "\[Delta]"}]}], ")"}]}]}], ")"}]}]]}], ",", 
    RowBox[{"F", "\[Rule]", 
     RowBox[{"-", 
      FractionBox[
       RowBox[{
        RowBox[{
         RowBox[{"-", "2"}], " ", "h", " ", 
         SuperscriptBox["sf", "2"], " ", 
         RowBox[{"(", 
          RowBox[{
           RowBox[{"-", "1"}], "+", "sm"}], ")"}], " ", 
         RowBox[{"(", 
          RowBox[{"1", "+", "\[Delta]"}], ")"}], " ", 
         RowBox[{"(", 
          RowBox[{
           RowBox[{"-", "2"}], "+", "sm", "-", "\[Delta]", "+", 
           RowBox[{"sm", " ", "\[Delta]"}]}], ")"}]}], "+", 
        RowBox[{
         SuperscriptBox["h", "2"], " ", 
         SuperscriptBox["sf", "2"], " ", 
         SuperscriptBox[
          RowBox[{"(", 
           RowBox[{
            RowBox[{"-", "2"}], "+", "sm", "-", "\[Delta]", "+", 
            RowBox[{"sm", " ", "\[Delta]"}]}], ")"}], "2"]}], "+", 
        RowBox[{
         RowBox[{"(", 
          RowBox[{"sm", "-", "\[Delta]", "+", 
           RowBox[{"sm", " ", "\[Delta]"}]}], ")"}], " ", 
         RowBox[{"(", 
          RowBox[{"\[Delta]", "+", 
           RowBox[{"2", " ", "sf", " ", 
            RowBox[{"(", 
             RowBox[{
              RowBox[{"-", "1"}], "+", "sm"}], ")"}], " ", 
            RowBox[{"(", 
             RowBox[{"1", "+", "\[Delta]"}], ")"}]}], "-", 
           RowBox[{"sm", " ", 
            RowBox[{"(", 
             RowBox[{"1", "+", "\[Delta]"}], ")"}]}]}], ")"}]}]}], 
       RowBox[{
        RowBox[{
         SuperscriptBox["h", "2"], " ", 
         SuperscriptBox["sf", "2"], " ", 
         SuperscriptBox[
          RowBox[{"(", 
           RowBox[{
            RowBox[{"-", "2"}], "+", "sm", "-", "\[Delta]", "+", 
            RowBox[{"sm", " ", "\[Delta]"}]}], ")"}], "2"]}], "+", 
        SuperscriptBox[
         RowBox[{"(", 
          RowBox[{"sm", "-", "\[Delta]", "+", 
           RowBox[{"sm", " ", "\[Delta]"}]}], ")"}], "2"], "-", 
        RowBox[{"2", " ", "sf", " ", 
         RowBox[{"(", 
          RowBox[{
           RowBox[{"2", " ", 
            RowBox[{"(", 
             RowBox[{
              RowBox[{"-", "1"}], "+", "sm"}], ")"}], " ", 
            RowBox[{"(", 
             RowBox[{"1", "+", "\[Delta]"}], ")"}]}], "+", 
           RowBox[{"h", " ", 
            SuperscriptBox[
             RowBox[{"(", 
              RowBox[{
               RowBox[{"-", "2"}], "+", "sm", "-", "\[Delta]", "+", 
               RowBox[{"sm", " ", "\[Delta]"}]}], ")"}], "2"]}]}], 
          ")"}]}]}]]}]}]}], "}"}], "}"}]], "Output",
 CellChangeTimes->{3.814098562927105*^9},
 CellLabel->"Out[15]=",ExpressionUUID->"e5c7bb73-f0b3-4b94-b5ca-81777d2d2601"]
}, Open  ]],

Cell["The full system is described by", "Text",
 CellChangeTimes->{{3.7699390900464983`*^9, 3.769939094324548*^9}, 
   3.814098666779276*^9},ExpressionUUID->"54fcd6cb-f965-4ab9-9229-\
21cd96d2548d"],

Cell[BoxData[
 RowBox[{
  RowBox[{"solvemeFull", "=", 
   RowBox[{
    RowBox[{
     RowBox[{"{", 
      RowBox[{
       RowBox[{"DDfprime", "-", "DDf"}], ",", 
       RowBox[{"Ddfprime", "-", "Ddf"}], ",", 
       RowBox[{"ddfprime", "-", "ddf"}], ",", " ", 
       RowBox[{"Dmprime", "-", "Dm"}], ",", " ", 
       RowBox[{"dmprime", "-", "dm"}]}], "}"}], "/.", 
     RowBox[{"subeq", "[", 
      RowBox[{"[", "1", "]"}], "]"}]}], "//", "Simplify"}]}], ";"}]], "Input",\

 CellChangeTimes->{
  3.769938276861104*^9, {3.769938343771866*^9, 3.76993835938778*^9}, {
   3.769939169862314*^9, 3.769939172196684*^9}, {3.76994170126457*^9, 
   3.769941701679681*^9}},
 CellLabel->"In[23]:=",ExpressionUUID->"c12bf7b4-d788-41f0-ac61-5834ea6f32ee"],

Cell["\<\
We can then solve for the population sizes. First we find population sizes \
when drive alleles are absent (pm=pf=0) or fixed (pm=pf=1):\
\>", "Text",
 CellChangeTimes->{{3.7699427703452044`*^9, 3.769942781021439*^9}, 
   3.814098670997642*^9},ExpressionUUID->"e4b6a7b9-c4d0-4c2f-87f2-\
5885b6ef6388"],

Cell[CellGroupData[{

Cell[BoxData[{
 RowBox[{"noDrivePopSizeSol", "=", 
  RowBox[{
   RowBox[{"Solve", "[", 
    RowBox[{
     RowBox[{
      RowBox[{"(", 
       RowBox[{
        RowBox[{"solvemeFull", "/.", 
         RowBox[{"pm", "\[Rule]", "0"}]}], "/.", 
        RowBox[{"pf", "\[Rule]", "0"}]}], ")"}], "\[Equal]", 
      RowBox[{"{", 
       RowBox[{"0", ",", "0", ",", "0", ",", "0", ",", "0"}], "}"}]}], ",", 
     RowBox[{"{", 
      RowBox[{"maleNum", ",", "femaleNum"}], "}"}]}], "]"}], "[", 
   RowBox[{"[", "2", "]"}], "]"}]}], "\[IndentingNewLine]", 
 RowBox[{"fixedDrivePopSizeSol", " ", "=", " ", 
  RowBox[{
   RowBox[{
    RowBox[{"Solve", "[", 
     RowBox[{
      RowBox[{
       RowBox[{"(", 
        RowBox[{
         RowBox[{"solvemeFull", "/.", 
          RowBox[{"pm", "\[Rule]", "1"}]}], "/.", 
         RowBox[{"pf", "\[Rule]", "1"}]}], ")"}], "\[Equal]", 
       RowBox[{"{", 
        RowBox[{"0", ",", "0", ",", "0", ",", "0", ",", "0"}], "}"}]}], ",", 
      RowBox[{"{", 
       RowBox[{"maleNum", ",", "femaleNum"}], "}"}]}], "]"}], "[", 
    RowBox[{"[", "2", "]"}], "]"}], "//", 
   "Simplify"}]}], "\[IndentingNewLine]", 
 RowBox[{"NnoDrive", "=", 
  RowBox[{
   RowBox[{
    RowBox[{"(", 
     RowBox[{"maleNum", " ", "+", " ", "femaleNum"}], ")"}], "/.", 
    "noDrivePopSizeSol"}], "//", "FullSimplify"}]}], "\[IndentingNewLine]", 
 RowBox[{"Nfixed", " ", "=", " ", 
  RowBox[{
   RowBox[{
    RowBox[{"(", 
     RowBox[{"maleNum", " ", "+", " ", "femaleNum"}], ")"}], "/.", 
    "fixedDrivePopSizeSol"}], "//", "FullSimplify"}]}]}], "Input",
 CellChangeTimes->{{3.7699383671627693`*^9, 3.769938423332643*^9}, {
   3.769939175654799*^9, 3.769939177087852*^9}, {3.769939672956071*^9, 
   3.76993969075327*^9}, {3.769940082733008*^9, 3.769940088265189*^9}, {
   3.7699403164181747`*^9, 3.769940322485866*^9}, {3.769942642908152*^9, 
   3.769942689023869*^9}, {3.769942723065901*^9, 3.769942728384009*^9}, {
   3.769942767873897*^9, 3.7699427680733128`*^9}, {3.7713269995636683`*^9, 
   3.7713270250192947`*^9}, {3.7713271122204103`*^9, 
   3.7713271285541897`*^9}, {3.797760549776165*^9, 3.797760654046899*^9}, {
   3.8000165175480967`*^9, 3.800016519503839*^9}, {3.8000167431373367`*^9, 
   3.800016757907406*^9}, 3.814098682598933*^9},
 CellLabel->"In[24]:=",ExpressionUUID->"dda5a460-53b5-4e26-96cd-e5f727801662"],

Cell[BoxData[
 RowBox[{"{", 
  RowBox[{
   RowBox[{"maleNum", "\[Rule]", 
    FractionBox[
     RowBox[{
      RowBox[{"-", "2"}], "+", "b"}], 
     RowBox[{
      SuperscriptBox["b", "2"], " ", "\[Alpha]"}]]}], ",", 
   RowBox[{"femaleNum", "\[Rule]", 
    FractionBox[
     RowBox[{
      RowBox[{"-", "2"}], "+", "b"}], 
     RowBox[{
      SuperscriptBox["b", "2"], " ", "\[Alpha]"}]]}]}], "}"}]], "Output",
 CellChangeTimes->{{3.814098683924067*^9, 3.8140986900191813`*^9}},
 CellLabel->"Out[24]=",ExpressionUUID->"8e645cbd-7402-4923-abd8-7efdff2fe8be"],

Cell[BoxData[
 RowBox[{"{", 
  RowBox[{
   RowBox[{"maleNum", "\[Rule]", 
    RowBox[{"-", 
     FractionBox[
      RowBox[{
       RowBox[{"(", 
        RowBox[{
         RowBox[{"-", "1"}], "+", "sm"}], ")"}], " ", 
       RowBox[{"(", 
        RowBox[{
         RowBox[{"-", "1"}], "+", "\[Delta]"}], ")"}], " ", 
       RowBox[{"(", 
        RowBox[{"2", "+", 
         RowBox[{"b", " ", 
          RowBox[{"(", 
           RowBox[{
            RowBox[{"-", "1"}], "+", "sf"}], ")"}], " ", 
          RowBox[{"(", 
           RowBox[{"1", "+", "\[Delta]"}], ")"}]}]}], ")"}]}], 
      RowBox[{
       SuperscriptBox["b", "2"], " ", 
       SuperscriptBox[
        RowBox[{"(", 
         RowBox[{
          RowBox[{"-", "1"}], "+", "sf"}], ")"}], "2"], " ", "\[Alpha]", " ", 
       
       SuperscriptBox[
        RowBox[{"(", 
         RowBox[{"1", "+", "\[Delta]"}], ")"}], "2"]}]]}]}], ",", 
   RowBox[{"femaleNum", "\[Rule]", 
    FractionBox[
     RowBox[{"2", "+", 
      RowBox[{"b", " ", 
       RowBox[{"(", 
        RowBox[{
         RowBox[{"-", "1"}], "+", "sf"}], ")"}], " ", 
       RowBox[{"(", 
        RowBox[{"1", "+", "\[Delta]"}], ")"}]}]}], 
     RowBox[{
      SuperscriptBox["b", "2"], " ", 
      RowBox[{"(", 
       RowBox[{
        RowBox[{"-", "1"}], "+", "sf"}], ")"}], " ", "\[Alpha]", " ", 
      RowBox[{"(", 
       RowBox[{"1", "+", "\[Delta]"}], ")"}]}]]}]}], "}"}]], "Output",
 CellChangeTimes->{{3.814098683924067*^9, 3.814098690057407*^9}},
 CellLabel->"Out[25]=",ExpressionUUID->"a2a2bb4b-622c-45fc-8f3b-750a56000383"],

Cell[BoxData[
 FractionBox[
  RowBox[{"2", " ", 
   RowBox[{"(", 
    RowBox[{
     RowBox[{"-", "2"}], "+", "b"}], ")"}]}], 
  RowBox[{
   SuperscriptBox["b", "2"], " ", "\[Alpha]"}]]], "Output",
 CellChangeTimes->{{3.814098683924067*^9, 3.814098690093773*^9}},
 CellLabel->"Out[26]=",ExpressionUUID->"52408ef0-1587-4b11-826a-e44fd2e43342"],

Cell[BoxData[
 FractionBox[
  RowBox[{
   RowBox[{"(", 
    RowBox[{
     RowBox[{"-", "2"}], "+", "sf", "+", "sm", "+", 
     RowBox[{"sf", " ", "\[Delta]"}], "-", 
     RowBox[{"sm", " ", "\[Delta]"}]}], ")"}], " ", 
   RowBox[{"(", 
    RowBox[{"2", "+", 
     RowBox[{"b", " ", 
      RowBox[{"(", 
       RowBox[{
        RowBox[{"-", "1"}], "+", "sf"}], ")"}], " ", 
      RowBox[{"(", 
       RowBox[{"1", "+", "\[Delta]"}], ")"}]}]}], ")"}]}], 
  RowBox[{
   SuperscriptBox["b", "2"], " ", 
   SuperscriptBox[
    RowBox[{"(", 
     RowBox[{
      RowBox[{"-", "1"}], "+", "sf"}], ")"}], "2"], " ", "\[Alpha]", " ", 
   SuperscriptBox[
    RowBox[{"(", 
     RowBox[{"1", "+", "\[Delta]"}], ")"}], "2"]}]]], "Output",
 CellChangeTimes->{{3.814098683924067*^9, 3.814098690139229*^9}},
 CellLabel->"Out[27]=",ExpressionUUID->"2737f089-1f46-4091-82c9-e6c0ef9d1724"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{
  RowBox[{
   RowBox[{"(", 
    RowBox[{"maleNum", " ", "+", " ", "femaleNum"}], ")"}], "/.", 
   "noDrivePopSizeSol"}], "//", "Simplify"}]], "Input",
 CellChangeTimes->{{3.77157514672134*^9, 3.7715751660546513`*^9}, {
   3.771575338494068*^9, 3.771575374825197*^9}, {3.7716608350619164`*^9, 
   3.771660843523736*^9}, 3.7723599629256897`*^9, {3.772361250123844*^9, 
   3.7723612505942497`*^9}},
 CellLabel->
  "In[2310]:=",ExpressionUUID->"5cda42b8-bccf-43dc-879a-f819080ecb0f"],

Cell[BoxData[
 RowBox[{"{", 
  RowBox[{"0", ",", 
   FractionBox[
    RowBox[{"2", " ", 
     RowBox[{"(", 
      RowBox[{
       RowBox[{"-", "2"}], "+", "b"}], ")"}]}], 
    RowBox[{
     SuperscriptBox["b", "2"], " ", "\[Alpha]"}]]}], "}"}]], "Output",
 CellChangeTimes->{3.771660843870935*^9, 3.772361253000827*^9, 
  3.772805565029376*^9, 3.7751383478986464`*^9, 3.7754164932491417`*^9, 
  3.797760016924254*^9, 3.797827033007077*^9, 3.8000161685410433`*^9, 
  3.800016685439253*^9, 3.809766822043935*^9},
 CellLabel->
  "Out[2310]=",ExpressionUUID->"b4d65503-c533-4476-8416-1c091cd89955"]
}, Open  ]],

Cell["WT population size is maximised for b=4", "Text",
 CellChangeTimes->{{3.800016776998363*^9, 
  3.800016800102717*^9}},ExpressionUUID->"67e73bba-e5e1-4071-be83-\
336d710090d5"],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{
  RowBox[{"Simplify", "[", 
   RowBox[{
    RowBox[{"Maximize", "[", 
     RowBox[{"NnoDrive", ",", "b"}], "]"}], ",", " ", 
    RowBox[{"\[Alpha]", ">", "0"}]}], "]"}], "[", 
  RowBox[{"[", "2", "]"}], "]"}]], "Input",
 CellChangeTimes->{{3.771660859269701*^9, 3.7716608642335997`*^9}, {
  3.814098704377606*^9, 3.814098752689563*^9}, {3.81409883615376*^9, 
  3.8140988385480623`*^9}, {3.814098914575449*^9, 3.814098915091632*^9}},
 CellLabel->"In[44]:=",ExpressionUUID->"148bc5eb-e7a0-4634-829e-7049146077ff"],

Cell[BoxData[
 RowBox[{"{", 
  RowBox[{"b", "\[Rule]", "4"}], "}"}]], "Output",
 CellChangeTimes->{{3.8140989027005167`*^9, 3.8140989153972816`*^9}},
 CellLabel->"Out[44]=",ExpressionUUID->"82c8f930-e172-4f1f-8ac5-f79ffcafe6a9"]
}, Open  ]],

Cell["\<\
When drive is fixed, the female sex ratio bias increases population \
productivity and means that the maximum population size is reached at a lower \
intrinsic birth rate\
\>", "Text",
 CellChangeTimes->{{3.814098924664686*^9, 
  3.8140989654117317`*^9}},ExpressionUUID->"2d5d1f78-b49c-41db-9774-\
0e1883730af2"],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{
  RowBox[{"Simplify", "[", 
   RowBox[{
    RowBox[{"Maximize", "[", 
     RowBox[{"Nfixed", ",", "b"}], "]"}], ",", " ", 
    RowBox[{"{", 
     RowBox[{
      RowBox[{"\[Alpha]", ">", "0"}], ",", 
      RowBox[{"0", "<", "sf", "<", "1"}], ",", 
      RowBox[{"0", "<", "sm", "<", "1"}], ",", 
      RowBox[{"0", "<", "h"}], ",", " ", 
      RowBox[{
       RowBox[{"h", " ", "sf"}], "<", "1"}], ",", 
      RowBox[{"0", "<", "b"}], ",", " ", 
      RowBox[{"0", "<", "\[Delta]", "<", "1"}], ",", 
      RowBox[{"0", "<", "c", "<", "1"}]}], "}"}]}], "]"}], "[", 
  RowBox[{"[", "2", "]"}], "]"}]], "Input",
 CellChangeTimes->{{3.814098849607452*^9, 3.8140988753576727`*^9}, {
  3.814098917579432*^9, 3.81409891826409*^9}, {3.814098973356359*^9, 
  3.8140989885652323`*^9}},
 CellLabel->"In[47]:=",ExpressionUUID->"5c29a56a-0fdb-46b6-8c89-fd0040e7b572"],

Cell[BoxData[
 RowBox[{"{", 
  RowBox[{"b", "\[Rule]", 
   RowBox[{"-", 
    FractionBox["4", 
     RowBox[{
      RowBox[{"(", 
       RowBox[{
        RowBox[{"-", "1"}], "+", "sf"}], ")"}], " ", 
      RowBox[{"(", 
       RowBox[{"1", "+", "\[Delta]"}], ")"}]}]]}]}], "}"}]], "Output",
 CellChangeTimes->{{3.8140988664537067`*^9, 3.814098876139146*^9}, 
   3.8140989188964577`*^9, {3.814098974386888*^9, 3.814098990010138*^9}},
 CellLabel->"Out[47]=",ExpressionUUID->"5edc3b3f-1d19-45d0-90ad-a3bde678a056"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"equilPopSizeSol", "=", 
  RowBox[{"Simplify", "[", 
   RowBox[{"Solve", "[", 
    RowBox[{
     RowBox[{
      RowBox[{"Flatten", "[", 
       RowBox[{"Simplify", "[", 
        RowBox[{"(", 
         RowBox[{"solvemeFull", "/.", "equilibrium"}], ")"}], "]"}], "]"}], 
      "\[Equal]", 
      RowBox[{"{", 
       RowBox[{"0", ",", "0", ",", "0", ",", "0", ",", "0"}], "}"}]}], ",", 
     " ", 
     RowBox[{"{", 
      RowBox[{"femaleNum", ",", "maleNum"}], "}"}]}], "]"}], "]"}]}]], "Input",\

 CellChangeTimes->{{3.7699384766191463`*^9, 3.7699384860993357`*^9}, {
   3.769939104815619*^9, 3.769939132710162*^9}, {3.769939231123618*^9, 
   3.76993923335522*^9}, {3.769939279772685*^9, 3.7699392839898157`*^9}, {
   3.769939317037973*^9, 3.7699393210350637`*^9}, {3.76993938466969*^9, 
   3.769939390637548*^9}, {3.769939624375873*^9, 3.769939631758957*^9}, 
   3.7699399837680607`*^9, {3.7699400297846003`*^9, 3.7699400629782133`*^9}, 
   3.769941562655718*^9, 3.769941817542982*^9, {3.7699427884805527`*^9, 
   3.769942788978636*^9}, {3.770717009677896*^9, 3.7707170195711117`*^9}, 
   3.771327222334549*^9, 3.800016834431821*^9, {3.814098997609971*^9, 
   3.814099008194984*^9}, {3.814099072906405*^9, 3.8140990890473433`*^9}},
 CellLabel->"In[53]:=",ExpressionUUID->"924ad8cd-c072-4ea4-a85f-7233e1ca5ec3"],

Cell[BoxData[
 RowBox[{"{", 
  RowBox[{
   RowBox[{"{", 
    RowBox[{
     RowBox[{"femaleNum", "\[Rule]", "0"}], ",", 
     RowBox[{"maleNum", "\[Rule]", "0"}]}], "}"}], ",", 
   RowBox[{"{", 
    RowBox[{
     RowBox[{"femaleNum", "\[Rule]", 
      FractionBox[
       RowBox[{
        RowBox[{"b", " ", 
         SuperscriptBox["h", "2"], " ", 
         SuperscriptBox["sf", "2"], " ", 
         SuperscriptBox[
          RowBox[{"(", 
           RowBox[{
            RowBox[{"-", "2"}], "+", "sm", "-", "\[Delta]", "+", 
            RowBox[{"sm", " ", "\[Delta]"}]}], ")"}], "2"]}], "+", 
        RowBox[{
         RowBox[{"(", 
          RowBox[{"sm", "-", "\[Delta]", "+", 
           RowBox[{"sm", " ", "\[Delta]"}]}], ")"}], " ", 
         RowBox[{"(", 
          RowBox[{
           RowBox[{
            RowBox[{"-", "b"}], " ", "\[Delta]"}], "+", 
           RowBox[{"sm", " ", 
            RowBox[{"(", 
             RowBox[{
              RowBox[{"-", "4"}], "+", "b", "+", 
              RowBox[{"b", " ", "\[Delta]"}]}], ")"}]}]}], ")"}]}], "-", 
        RowBox[{"2", " ", "sf", " ", 
         RowBox[{"(", 
          RowBox[{
           RowBox[{
            RowBox[{"-", "4"}], " ", 
            RowBox[{"(", 
             RowBox[{
              RowBox[{"-", "1"}], "+", "sm"}], ")"}], " ", 
            RowBox[{"(", 
             RowBox[{"1", "+", "\[Delta]"}], ")"}]}], "+", 
           RowBox[{"2", " ", "b", " ", 
            RowBox[{"(", 
             RowBox[{
              RowBox[{"-", "1"}], "+", "sm"}], ")"}], " ", 
            RowBox[{"(", 
             RowBox[{"1", "+", "\[Delta]"}], ")"}]}], "-", 
           RowBox[{"2", " ", "h", " ", 
            RowBox[{"(", 
             RowBox[{
              RowBox[{"-", "2"}], "+", "sm"}], ")"}], " ", 
            RowBox[{"(", 
             RowBox[{
              RowBox[{"-", "2"}], "+", "sm", "-", "\[Delta]", "+", 
              RowBox[{"sm", " ", "\[Delta]"}]}], ")"}]}], "+", 
           RowBox[{"b", " ", "h", " ", 
            SuperscriptBox[
             RowBox[{"(", 
              RowBox[{
               RowBox[{"-", "2"}], "+", "sm", "-", "\[Delta]", "+", 
               RowBox[{"sm", " ", "\[Delta]"}]}], ")"}], "2"]}]}], ")"}]}]}], 
       
       RowBox[{
        SuperscriptBox["b", "2"], " ", "\[Alpha]", " ", 
        RowBox[{"(", 
         RowBox[{
          RowBox[{
           SuperscriptBox["h", "2"], " ", 
           SuperscriptBox["sf", "2"], " ", 
           SuperscriptBox[
            RowBox[{"(", 
             RowBox[{
              RowBox[{"-", "2"}], "+", "sm", "-", "\[Delta]", "+", 
              RowBox[{"sm", " ", "\[Delta]"}]}], ")"}], "2"]}], "+", 
          SuperscriptBox[
           RowBox[{"(", 
            RowBox[{"sm", "-", "\[Delta]", "+", 
             RowBox[{"sm", " ", "\[Delta]"}]}], ")"}], "2"], "-", 
          RowBox[{"2", " ", "sf", " ", 
           RowBox[{"(", 
            RowBox[{
             RowBox[{"2", " ", 
              RowBox[{"(", 
               RowBox[{
                RowBox[{"-", "1"}], "+", "sm"}], ")"}], " ", 
              RowBox[{"(", 
               RowBox[{"1", "+", "\[Delta]"}], ")"}]}], "+", 
             RowBox[{"h", " ", 
              SuperscriptBox[
               RowBox[{"(", 
                RowBox[{
                 RowBox[{"-", "2"}], "+", "sm", "-", "\[Delta]", "+", 
                 RowBox[{"sm", " ", "\[Delta]"}]}], ")"}], "2"]}]}], 
            ")"}]}]}], ")"}]}]]}], ",", 
     RowBox[{"maleNum", "\[Rule]", 
      RowBox[{"-", 
       RowBox[{"(", 
        RowBox[{
         RowBox[{"(", 
          RowBox[{
           RowBox[{"(", 
            RowBox[{
             RowBox[{"b", " ", 
              SuperscriptBox["h", "2"], " ", 
              SuperscriptBox["sf", "2"], " ", 
              SuperscriptBox[
               RowBox[{"(", 
                RowBox[{
                 RowBox[{"-", "2"}], "+", "sm", "-", "\[Delta]", "+", 
                 RowBox[{"sm", " ", "\[Delta]"}]}], ")"}], "2"]}], "+", 
             RowBox[{
              RowBox[{"(", 
               RowBox[{"sm", "-", "\[Delta]", "+", 
                RowBox[{"sm", " ", "\[Delta]"}]}], ")"}], " ", 
              RowBox[{"(", 
               RowBox[{
                RowBox[{
                 RowBox[{"-", "b"}], " ", "\[Delta]"}], "+", 
                RowBox[{"sm", " ", 
                 RowBox[{"(", 
                  RowBox[{
                   RowBox[{"-", "4"}], "+", "b", "+", 
                   RowBox[{"b", " ", "\[Delta]"}]}], ")"}]}]}], ")"}]}], "-", 
             
             RowBox[{"2", " ", "sf", " ", 
              RowBox[{"(", 
               RowBox[{
                RowBox[{
                 RowBox[{"-", "4"}], " ", 
                 RowBox[{"(", 
                  RowBox[{
                   RowBox[{"-", "1"}], "+", "sm"}], ")"}], " ", 
                 RowBox[{"(", 
                  RowBox[{"1", "+", "\[Delta]"}], ")"}]}], "+", 
                RowBox[{"2", " ", "b", " ", 
                 RowBox[{"(", 
                  RowBox[{
                   RowBox[{"-", "1"}], "+", "sm"}], ")"}], " ", 
                 RowBox[{"(", 
                  RowBox[{"1", "+", "\[Delta]"}], ")"}]}], "-", 
                RowBox[{"2", " ", "h", " ", 
                 RowBox[{"(", 
                  RowBox[{
                   RowBox[{"-", "2"}], "+", "sm"}], ")"}], " ", 
                 RowBox[{"(", 
                  RowBox[{
                   RowBox[{"-", "2"}], "+", "sm", "-", "\[Delta]", "+", 
                   RowBox[{"sm", " ", "\[Delta]"}]}], ")"}]}], "+", 
                RowBox[{"b", " ", "h", " ", 
                 SuperscriptBox[
                  RowBox[{"(", 
                   RowBox[{
                    RowBox[{"-", "2"}], "+", "sm", "-", "\[Delta]", "+", 
                    RowBox[{"sm", " ", "\[Delta]"}]}], ")"}], "2"]}]}], 
               ")"}]}]}], ")"}], " ", 
           RowBox[{"(", 
            RowBox[{
             RowBox[{"sm", " ", 
              RowBox[{"(", 
               RowBox[{
                RowBox[{"sm", " ", 
                 RowBox[{"(", 
                  RowBox[{
                   RowBox[{"-", "1"}], "+", "\[Delta]"}], ")"}]}], "-", 
                "\[Delta]"}], ")"}], " ", 
              SuperscriptBox[
               RowBox[{"(", 
                RowBox[{"sm", "-", "\[Delta]", "+", 
                 RowBox[{"sm", " ", "\[Delta]"}]}], ")"}], "2"]}], "+", 
             RowBox[{
              SuperscriptBox["sf", "2"], " ", 
              RowBox[{"(", 
               RowBox[{
                RowBox[{
                 RowBox[{"-", "4"}], " ", 
                 SuperscriptBox[
                  RowBox[{"(", 
                   RowBox[{
                    RowBox[{"-", "1"}], "+", "sm"}], ")"}], "2"], " ", 
                 SuperscriptBox[
                  RowBox[{"(", 
                   RowBox[{"1", "+", "\[Delta]"}], ")"}], "2"]}], "+", 
                RowBox[{
                 SuperscriptBox["h", "2"], " ", 
                 RowBox[{"(", 
                  RowBox[{
                   RowBox[{"-", "2"}], "+", "sm"}], ")"}], " ", 
                 RowBox[{"(", 
                  RowBox[{"2", "+", 
                   RowBox[{"sm", " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{"-", "1"}], "+", "\[Delta]"}], ")"}]}], "-", 
                   "\[Delta]"}], ")"}], " ", 
                 SuperscriptBox[
                  RowBox[{"(", 
                   RowBox[{
                    RowBox[{"-", "2"}], "+", "sm", "-", "\[Delta]", "+", 
                    RowBox[{"sm", " ", "\[Delta]"}]}], ")"}], "2"]}], "+", 
                RowBox[{"2", " ", "h", " ", 
                 RowBox[{"(", 
                  RowBox[{
                   RowBox[{"-", "1"}], "+", "sm"}], ")"}], " ", 
                 RowBox[{"(", 
                  RowBox[{"1", "+", "\[Delta]"}], ")"}], " ", 
                 RowBox[{"(", 
                  RowBox[{
                   RowBox[{"-", "8"}], "-", 
                   RowBox[{"2", " ", "\[Delta]"}], "+", 
                   SuperscriptBox["\[Delta]", "2"], "+", 
                   RowBox[{"sm", " ", 
                    RowBox[{"(", 
                    RowBox[{"8", "+", 
                    RowBox[{"3", " ", "\[Delta]"}], "-", 
                    RowBox[{"2", " ", 
                    SuperscriptBox["\[Delta]", "2"]}]}], ")"}]}], "+", 
                   RowBox[{
                    SuperscriptBox["sm", "2"], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{"-", "2"}], "-", "\[Delta]", "+", 
                    SuperscriptBox["\[Delta]", "2"]}], ")"}]}]}], ")"}]}]}], 
               ")"}]}], "-", 
             RowBox[{"2", " ", "sf", " ", 
              RowBox[{"(", 
               RowBox[{"sm", "-", "\[Delta]", "+", 
                RowBox[{"sm", " ", "\[Delta]"}]}], ")"}], " ", 
              RowBox[{"(", 
               RowBox[{
                RowBox[{"h", " ", 
                 SuperscriptBox["sm", "3"], " ", 
                 RowBox[{"(", 
                  RowBox[{
                   RowBox[{"-", "1"}], "+", 
                   SuperscriptBox["\[Delta]", "2"]}], ")"}]}], "+", 
                RowBox[{"\[Delta]", " ", 
                 RowBox[{"(", 
                  RowBox[{"1", "+", "\[Delta]", "-", 
                   RowBox[{"h", " ", 
                    RowBox[{"(", 
                    RowBox[{"2", "+", "\[Delta]"}], ")"}]}]}], ")"}]}], "-", 
                RowBox[{
                 SuperscriptBox["sm", "2"], " ", 
                 RowBox[{"(", 
                  RowBox[{"2", "+", "\[Delta]", "-", 
                   SuperscriptBox["\[Delta]", "2"], "+", 
                   RowBox[{"h", " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{"-", "4"}], "+", "\[Delta]", "+", 
                    RowBox[{"3", " ", 
                    SuperscriptBox["\[Delta]", "2"]}]}], ")"}]}]}], ")"}]}], 
                "+", 
                RowBox[{"sm", " ", 
                 RowBox[{"(", 
                  RowBox[{"2", "-", 
                   RowBox[{"2", " ", 
                    SuperscriptBox["\[Delta]", "2"]}], "+", 
                   RowBox[{"h", " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{"-", "4"}], "+", 
                    RowBox[{"3", " ", "\[Delta]"}], "+", 
                    RowBox[{"3", " ", 
                    SuperscriptBox["\[Delta]", "2"]}]}], ")"}]}]}], ")"}]}]}],
                ")"}]}]}], ")"}]}], ")"}], "/", 
         RowBox[{"(", 
          RowBox[{
           SuperscriptBox["b", "2"], " ", "sf", " ", "\[Alpha]", " ", 
           RowBox[{"(", 
            RowBox[{
             RowBox[{
              RowBox[{"-", 
               RowBox[{"(", 
                RowBox[{
                 RowBox[{"-", "1"}], "+", "sm"}], ")"}]}], " ", 
              RowBox[{"(", 
               RowBox[{"1", "+", "\[Delta]"}], ")"}]}], "+", 
             RowBox[{"h", " ", 
              RowBox[{"(", 
               RowBox[{
                RowBox[{"-", "2"}], "+", "sm", "-", "\[Delta]", "+", 
                RowBox[{"sm", " ", "\[Delta]"}]}], ")"}]}]}], ")"}], " ", 
           SuperscriptBox[
            RowBox[{"(", 
             RowBox[{
              RowBox[{
               SuperscriptBox["h", "2"], " ", 
               SuperscriptBox["sf", "2"], " ", 
               SuperscriptBox[
                RowBox[{"(", 
                 RowBox[{
                  RowBox[{"-", "2"}], "+", "sm", "-", "\[Delta]", "+", 
                  RowBox[{"sm", " ", "\[Delta]"}]}], ")"}], "2"]}], "+", 
              SuperscriptBox[
               RowBox[{"(", 
                RowBox[{"sm", "-", "\[Delta]", "+", 
                 RowBox[{"sm", " ", "\[Delta]"}]}], ")"}], "2"], "-", 
              RowBox[{"2", " ", "sf", " ", 
               RowBox[{"(", 
                RowBox[{
                 RowBox[{"2", " ", 
                  RowBox[{"(", 
                   RowBox[{
                    RowBox[{"-", "1"}], "+", "sm"}], ")"}], " ", 
                  RowBox[{"(", 
                   RowBox[{"1", "+", "\[Delta]"}], ")"}]}], "+", 
                 RowBox[{"h", " ", 
                  SuperscriptBox[
                   RowBox[{"(", 
                    RowBox[{
                    RowBox[{"-", "2"}], "+", "sm", "-", "\[Delta]", "+", 
                    RowBox[{"sm", " ", "\[Delta]"}]}], ")"}], "2"]}]}], 
                ")"}]}]}], ")"}], "2"]}], ")"}]}], ")"}]}]}]}], "}"}]}], 
  "}"}]], "Output",
 CellChangeTimes->{3.7723612807794313`*^9, 3.772805567939139*^9, 
  3.775138354199901*^9, 3.775416496358555*^9, 3.797760019906515*^9, 
  3.7978270361387053`*^9, 3.800016171274366*^9, 3.800016697767928*^9, 
  3.800016852287209*^9, 3.809766838331295*^9, 3.81409901346994*^9, 
  3.814099076865303*^9, 3.8140991147876062`*^9},
 CellLabel->"Out[53]=",ExpressionUUID->"82857bcf-4ac0-4a58-8783-eba9797880eb"]
}, Open  ]],

Cell["\<\
Check that this is an equilibrium solution by checking that the difference \
equations are 0\
\>", "Text",
 CellChangeTimes->{{3.769942894623459*^9, 3.769942902620317*^9}, {
  3.7699429667125473`*^9, 
  3.769942979884016*^9}},ExpressionUUID->"51be3e69-ddfc-4d22-ab0f-\
406020908356"],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"Simplify", "[", 
  RowBox[{
   RowBox[{
    RowBox[{"solvemeFull", "/.", "subeq"}], "/.", "equilibrium"}], "/.", 
   RowBox[{"equilPopSizeSol", "[", 
    RowBox[{"[", "2", "]"}], "]"}]}], "]"}]], "Input",
 CellChangeTimes->{{3.769941357591445*^9, 3.769941391398161*^9}, {
   3.769942887850418*^9, 3.7699428885472918`*^9}, {3.769942995121667*^9, 
   3.769943000854732*^9}, 3.8000168851297493`*^9, {3.814099117510462*^9, 
   3.814099117925303*^9}},
 CellLabel->"In[56]:=",ExpressionUUID->"6ee650c9-96cf-45e6-8be5-906729a1e138"],

Cell[BoxData[
 RowBox[{"{", 
  RowBox[{"{", 
   RowBox[{"{", 
    RowBox[{"0", ",", "0", ",", "0", ",", "0", ",", "0"}], "}"}], "}"}], 
  "}"}]], "Output",
 CellChangeTimes->{{3.769941361055786*^9, 3.7699413918202257`*^9}, 
   3.769941841320743*^9, 3.769943001514942*^9, 3.769953316893544*^9, 
   3.770112919222467*^9, 3.7703732470228987`*^9, 3.771325809200185*^9, 
   3.771327236999894*^9, 3.77236131342944*^9, 3.772805568352688*^9, 
   3.77513835979795*^9, 3.775416496801271*^9, 3.797760020331903*^9, 
   3.797827036577125*^9, 3.800016171656394*^9, 3.800016700219249*^9, 
   3.800016887295944*^9, 3.809766840079547*^9, {3.81409911490854*^9, 
   3.814099118176263*^9}},
 CellLabel->"Out[56]=",ExpressionUUID->"e0521479-690a-4ccd-a3bb-5acfdb2e8456"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"Neq", " ", "=", " ", 
  RowBox[{"Simplify", "[", 
   RowBox[{"(", 
    RowBox[{
     RowBox[{"(", 
      RowBox[{"femaleNum", "+", "maleNum"}], ")"}], "/.", 
     RowBox[{"equilPopSizeSol", "[", 
      RowBox[{"[", "2", "]"}], "]"}]}], ")"}], "]"}]}]], "Input",
 CellChangeTimes->{{3.770036691536613*^9, 3.770036692957157*^9}, {
  3.77157592348055*^9, 3.7715759596789417`*^9}, {3.771576204371656*^9, 
  3.77157624037074*^9}, {3.814099094484907*^9, 3.814099107167481*^9}},
 CellLabel->"In[55]:=",ExpressionUUID->"25536a4d-a783-4d71-869f-b50b867d83e9"],

Cell[BoxData[
 RowBox[{
  RowBox[{"(", 
   RowBox[{
    RowBox[{"(", 
     RowBox[{
      RowBox[{"b", " ", 
       SuperscriptBox["h", "2"], " ", 
       SuperscriptBox["sf", "2"], " ", 
       SuperscriptBox[
        RowBox[{"(", 
         RowBox[{
          RowBox[{"-", "2"}], "+", "sm", "-", "\[Delta]", "+", 
          RowBox[{"sm", " ", "\[Delta]"}]}], ")"}], "2"]}], "+", 
      RowBox[{
       RowBox[{"(", 
        RowBox[{"sm", "-", "\[Delta]", "+", 
         RowBox[{"sm", " ", "\[Delta]"}]}], ")"}], " ", 
       RowBox[{"(", 
        RowBox[{
         RowBox[{
          RowBox[{"-", "b"}], " ", "\[Delta]"}], "+", 
         RowBox[{"sm", " ", 
          RowBox[{"(", 
           RowBox[{
            RowBox[{"-", "4"}], "+", "b", "+", 
            RowBox[{"b", " ", "\[Delta]"}]}], ")"}]}]}], ")"}]}], "-", 
      RowBox[{"2", " ", "sf", " ", 
       RowBox[{"(", 
        RowBox[{
         RowBox[{
          RowBox[{"-", "4"}], " ", 
          RowBox[{"(", 
           RowBox[{
            RowBox[{"-", "1"}], "+", "sm"}], ")"}], " ", 
          RowBox[{"(", 
           RowBox[{"1", "+", "\[Delta]"}], ")"}]}], "+", 
         RowBox[{"2", " ", "b", " ", 
          RowBox[{"(", 
           RowBox[{
            RowBox[{"-", "1"}], "+", "sm"}], ")"}], " ", 
          RowBox[{"(", 
           RowBox[{"1", "+", "\[Delta]"}], ")"}]}], "-", 
         RowBox[{"2", " ", "h", " ", 
          RowBox[{"(", 
           RowBox[{
            RowBox[{"-", "2"}], "+", "sm"}], ")"}], " ", 
          RowBox[{"(", 
           RowBox[{
            RowBox[{"-", "2"}], "+", "sm", "-", "\[Delta]", "+", 
            RowBox[{"sm", " ", "\[Delta]"}]}], ")"}]}], "+", 
         RowBox[{"b", " ", "h", " ", 
          SuperscriptBox[
           RowBox[{"(", 
            RowBox[{
             RowBox[{"-", "2"}], "+", "sm", "-", "\[Delta]", "+", 
             RowBox[{"sm", " ", "\[Delta]"}]}], ")"}], "2"]}]}], ")"}]}]}], 
     ")"}], " ", 
    RowBox[{"(", 
     RowBox[{
      RowBox[{
       SuperscriptBox["h", "2"], " ", 
       SuperscriptBox["sf", "2"], " ", 
       SuperscriptBox[
        RowBox[{"(", 
         RowBox[{
          RowBox[{"-", "2"}], "+", "sm", "-", "\[Delta]", "+", 
          RowBox[{"sm", " ", "\[Delta]"}]}], ")"}], "2"]}], "+", 
      SuperscriptBox[
       RowBox[{"(", 
        RowBox[{"sm", "-", "\[Delta]", "+", 
         RowBox[{"sm", " ", "\[Delta]"}]}], ")"}], "2"], "-", 
      RowBox[{"2", " ", "sf", " ", 
       RowBox[{"(", 
        RowBox[{
         RowBox[{"2", " ", 
          RowBox[{"(", 
           RowBox[{
            RowBox[{"-", "1"}], "+", "sm"}], ")"}], " ", 
          RowBox[{"(", 
           RowBox[{"1", "+", "\[Delta]"}], ")"}]}], "+", 
         RowBox[{"h", " ", 
          SuperscriptBox[
           RowBox[{"(", 
            RowBox[{
             RowBox[{"-", "2"}], "+", "sm", "-", "\[Delta]", "+", 
             RowBox[{"sm", " ", "\[Delta]"}]}], ")"}], "2"]}]}], ")"}]}], "-", 
      RowBox[{
       FractionBox["1", 
        RowBox[{"sf", " ", 
         RowBox[{"(", 
          RowBox[{
           RowBox[{
            RowBox[{"-", 
             RowBox[{"(", 
              RowBox[{
               RowBox[{"-", "1"}], "+", "sm"}], ")"}]}], " ", 
            RowBox[{"(", 
             RowBox[{"1", "+", "\[Delta]"}], ")"}]}], "+", 
           RowBox[{"h", " ", 
            RowBox[{"(", 
             RowBox[{
              RowBox[{"-", "2"}], "+", "sm", "-", "\[Delta]", "+", 
              RowBox[{"sm", " ", "\[Delta]"}]}], ")"}]}]}], ")"}]}]], 
       RowBox[{"(", 
        RowBox[{
         RowBox[{"sm", " ", 
          RowBox[{"(", 
           RowBox[{
            RowBox[{"sm", " ", 
             RowBox[{"(", 
              RowBox[{
               RowBox[{"-", "1"}], "+", "\[Delta]"}], ")"}]}], "-", 
            "\[Delta]"}], ")"}], " ", 
          SuperscriptBox[
           RowBox[{"(", 
            RowBox[{"sm", "-", "\[Delta]", "+", 
             RowBox[{"sm", " ", "\[Delta]"}]}], ")"}], "2"]}], "+", 
         RowBox[{
          SuperscriptBox["sf", "2"], " ", 
          RowBox[{"(", 
           RowBox[{
            RowBox[{
             RowBox[{"-", "4"}], " ", 
             SuperscriptBox[
              RowBox[{"(", 
               RowBox[{
                RowBox[{"-", "1"}], "+", "sm"}], ")"}], "2"], " ", 
             SuperscriptBox[
              RowBox[{"(", 
               RowBox[{"1", "+", "\[Delta]"}], ")"}], "2"]}], "+", 
            RowBox[{
             SuperscriptBox["h", "2"], " ", 
             RowBox[{"(", 
              RowBox[{
               RowBox[{"-", "2"}], "+", "sm"}], ")"}], " ", 
             RowBox[{"(", 
              RowBox[{"2", "+", 
               RowBox[{"sm", " ", 
                RowBox[{"(", 
                 RowBox[{
                  RowBox[{"-", "1"}], "+", "\[Delta]"}], ")"}]}], "-", 
               "\[Delta]"}], ")"}], " ", 
             SuperscriptBox[
              RowBox[{"(", 
               RowBox[{
                RowBox[{"-", "2"}], "+", "sm", "-", "\[Delta]", "+", 
                RowBox[{"sm", " ", "\[Delta]"}]}], ")"}], "2"]}], "+", 
            RowBox[{"2", " ", "h", " ", 
             RowBox[{"(", 
              RowBox[{
               RowBox[{"-", "1"}], "+", "sm"}], ")"}], " ", 
             RowBox[{"(", 
              RowBox[{"1", "+", "\[Delta]"}], ")"}], " ", 
             RowBox[{"(", 
              RowBox[{
               RowBox[{"-", "8"}], "-", 
               RowBox[{"2", " ", "\[Delta]"}], "+", 
               SuperscriptBox["\[Delta]", "2"], "+", 
               RowBox[{"sm", " ", 
                RowBox[{"(", 
                 RowBox[{"8", "+", 
                  RowBox[{"3", " ", "\[Delta]"}], "-", 
                  RowBox[{"2", " ", 
                   SuperscriptBox["\[Delta]", "2"]}]}], ")"}]}], "+", 
               RowBox[{
                SuperscriptBox["sm", "2"], " ", 
                RowBox[{"(", 
                 RowBox[{
                  RowBox[{"-", "2"}], "-", "\[Delta]", "+", 
                  SuperscriptBox["\[Delta]", "2"]}], ")"}]}]}], ")"}]}]}], 
           ")"}]}], "-", 
         RowBox[{"2", " ", "sf", " ", 
          RowBox[{"(", 
           RowBox[{"sm", "-", "\[Delta]", "+", 
            RowBox[{"sm", " ", "\[Delta]"}]}], ")"}], " ", 
          RowBox[{"(", 
           RowBox[{
            RowBox[{"h", " ", 
             SuperscriptBox["sm", "3"], " ", 
             RowBox[{"(", 
              RowBox[{
               RowBox[{"-", "1"}], "+", 
               SuperscriptBox["\[Delta]", "2"]}], ")"}]}], "+", 
            RowBox[{"\[Delta]", " ", 
             RowBox[{"(", 
              RowBox[{"1", "+", "\[Delta]", "-", 
               RowBox[{"h", " ", 
                RowBox[{"(", 
                 RowBox[{"2", "+", "\[Delta]"}], ")"}]}]}], ")"}]}], "-", 
            RowBox[{
             SuperscriptBox["sm", "2"], " ", 
             RowBox[{"(", 
              RowBox[{"2", "+", "\[Delta]", "-", 
               SuperscriptBox["\[Delta]", "2"], "+", 
               RowBox[{"h", " ", 
                RowBox[{"(", 
                 RowBox[{
                  RowBox[{"-", "4"}], "+", "\[Delta]", "+", 
                  RowBox[{"3", " ", 
                   SuperscriptBox["\[Delta]", "2"]}]}], ")"}]}]}], ")"}]}], 
            "+", 
            RowBox[{"sm", " ", 
             RowBox[{"(", 
              RowBox[{"2", "-", 
               RowBox[{"2", " ", 
                SuperscriptBox["\[Delta]", "2"]}], "+", 
               RowBox[{"h", " ", 
                RowBox[{"(", 
                 RowBox[{
                  RowBox[{"-", "4"}], "+", 
                  RowBox[{"3", " ", "\[Delta]"}], "+", 
                  RowBox[{"3", " ", 
                   SuperscriptBox["\[Delta]", "2"]}]}], ")"}]}]}], ")"}]}]}], 
           ")"}]}]}], ")"}]}]}], ")"}]}], ")"}], "/", 
  RowBox[{"(", 
   RowBox[{
    SuperscriptBox["b", "2"], " ", "\[Alpha]", " ", 
    SuperscriptBox[
     RowBox[{"(", 
      RowBox[{
       RowBox[{
        SuperscriptBox["h", "2"], " ", 
        SuperscriptBox["sf", "2"], " ", 
        SuperscriptBox[
         RowBox[{"(", 
          RowBox[{
           RowBox[{"-", "2"}], "+", "sm", "-", "\[Delta]", "+", 
           RowBox[{"sm", " ", "\[Delta]"}]}], ")"}], "2"]}], "+", 
       SuperscriptBox[
        RowBox[{"(", 
         RowBox[{"sm", "-", "\[Delta]", "+", 
          RowBox[{"sm", " ", "\[Delta]"}]}], ")"}], "2"], "-", 
       RowBox[{"2", " ", "sf", " ", 
        RowBox[{"(", 
         RowBox[{
          RowBox[{"2", " ", 
           RowBox[{"(", 
            RowBox[{
             RowBox[{"-", "1"}], "+", "sm"}], ")"}], " ", 
           RowBox[{"(", 
            RowBox[{"1", "+", "\[Delta]"}], ")"}]}], "+", 
          RowBox[{"h", " ", 
           SuperscriptBox[
            RowBox[{"(", 
             RowBox[{
              RowBox[{"-", "2"}], "+", "sm", "-", "\[Delta]", "+", 
              RowBox[{"sm", " ", "\[Delta]"}]}], ")"}], "2"]}]}], ")"}]}]}], 
      ")"}], "2"]}], ")"}]}]], "Output",
 CellChangeTimes->{
  3.772361554150361*^9, 3.7728055701958227`*^9, 3.775138376640626*^9, 
   3.775416498795054*^9, 3.797760022160544*^9, 3.797827038581608*^9, 
   3.80001617344917*^9, 3.800016700323613*^9, 3.8000169002435102`*^9, 
   3.80976684144516*^9, {3.814099107459569*^9, 3.814099114979775*^9}},
 CellLabel->"Out[55]=",ExpressionUUID->"9f7e632d-da7a-408d-8078-19cf8682134f"]
}, Open  ]]
}, Closed]],

Cell[CellGroupData[{

Cell["Stability", "Subsection",
 CellChangeTimes->{{3.7812702047944727`*^9, 
  3.781270206443592*^9}},ExpressionUUID->"45736a70-7990-499e-a832-\
0b9071784a70"],

Cell["\<\
There are at least two immediately apparent equilibria: when drive is absent \
from the population and when drive is fixed in the population. 

We first find the Jacobian for the system:\
\>", "Text",
 CellChangeTimes->{{3.781272256024481*^9, 3.781272395504654*^9}, {
  3.781273708553084*^9, 3.7812737092006273`*^9}, {3.781273897437092*^9, 
  3.781273908742497*^9}, {3.800007455744604*^9, 3.800007456204402*^9}, {
  3.814092010155294*^9, 
  3.814092011347086*^9}},ExpressionUUID->"ab51b2eb-4d40-4a0a-886b-\
7a2b5e6928c2"],

Cell[BoxData[{
 RowBox[{
  RowBox[{"recursions", "=", 
   RowBox[{"{", 
    RowBox[{
    "Dmprime", ",", " ", "dmprime", ",", " ", "DDfprime", ",", " ", 
     "Ddfprime", ",", " ", "ddfprime"}], "}"}]}], 
  ";"}], "\[IndentingNewLine]", 
 RowBox[{
  RowBox[{"jacob", "=", 
   RowBox[{"Simplify", "[", 
    RowBox[{"Transpose", "[", 
     RowBox[{"{", 
      RowBox[{
       RowBox[{"D", "[", 
        RowBox[{"recursions", ",", "Dm"}], "]"}], ",", " ", 
       RowBox[{"D", "[", 
        RowBox[{"recursions", ",", "dm"}], "]"}], ",", " ", 
       RowBox[{"D", "[", 
        RowBox[{"recursions", ",", " ", "DDf"}], "]"}], ",", " ", 
       RowBox[{"D", "[", 
        RowBox[{"recursions", ",", " ", "Ddf"}], "]"}], ",", " ", 
       RowBox[{"D", "[", 
        RowBox[{"recursions", ",", " ", "ddf"}], "]"}]}], "}"}], "]"}], 
    "]"}]}], ";"}]}], "Input",
 CellChangeTimes->{{3.7812738954655247`*^9, 3.781273895657804*^9}},
 CellLabel->"In[73]:=",ExpressionUUID->"4c6c6aea-6ce2-4e48-bf24-4d7f10b04578"],

Cell["\<\
And then we assess the conditions under which the drive-absent and \
drive-fixed equilibria are stable/instable. \
\>", "Text",
 CellChangeTimes->{
  3.8140187547665863`*^9},ExpressionUUID->"afc02abd-9832-41ed-ac68-\
5c1d687df12d"],

Cell[CellGroupData[{

Cell["drive absent (pm=pf=0) equilibrium", "Subsubsection",
 CellChangeTimes->{{3.781273712670188*^9, 3.7812737328189983`*^9}, 
   3.814018761654243*^9},ExpressionUUID->"65608a1c-7b86-45e4-8704-\
d4cfa68e8141"],

Cell["\<\
To evaluate the stability of the equilibrium where drive is absent we \
consider the eigenvalues of the Jacobian at this point (substituting pf->0, \
pm->0, and the equilibrium population size without drive). The equilibrium is \
unstable whenever it has an eigenvalue with absolute value > 1.\
\>", "Text",
 CellChangeTimes->{{3.781273918397637*^9, 3.781273991244995*^9}, {
  3.781275775595746*^9, 3.781275807826384*^9}, {3.814003868404478*^9, 
  3.814003889059251*^9}},ExpressionUUID->"ee1ba8bb-788e-478c-bee9-\
d535f8518781"],

Cell[CellGroupData[{

Cell[BoxData[{
 RowBox[{
  RowBox[{
   RowBox[{"Simplify", "[", 
    RowBox[{
     RowBox[{
      RowBox[{
       RowBox[{"jacob", "/.", "subeq"}], "/.", 
       RowBox[{"pf", "\[Rule]", "0"}]}], "/.", 
      RowBox[{"pm", "\[Rule]", "0"}]}], "/.", "noDrivePopSizeSol"}], "]"}], 
   "[", 
   RowBox[{"[", "1", "]"}], "]"}], ";"}], "\[IndentingNewLine]", 
 RowBox[{"charpolynomial0", "=", 
  RowBox[{"Factor", "[", 
   RowBox[{"Det", "[", 
    RowBox[{"%", "-", 
     RowBox[{"\[Lambda]", "*", 
      RowBox[{"IdentityMatrix", "[", "5", "]"}]}]}], "]"}], "]"}]}]}], "Input",\

 CellChangeTimes->{
  3.781274002989327*^9, {3.813930917830587*^9, 3.813931059531245*^9}, {
   3.814003803361516*^9, 3.814003804854281*^9}, {3.8140085533840113`*^9, 
   3.814008553741354*^9}},
 CellLabel->"In[75]:=",ExpressionUUID->"0296eb6b-c900-4934-aa04-032620399872"],

Cell[BoxData[
 RowBox[{
  RowBox[{"-", 
   FractionBox["1", "4"]}], " ", 
  SuperscriptBox["\[Lambda]", "2"], " ", 
  RowBox[{"(", 
   RowBox[{
    RowBox[{"-", "4"}], "+", "b", "+", 
    RowBox[{"2", " ", "\[Lambda]"}]}], ")"}], " ", 
  RowBox[{"(", 
   RowBox[{
    RowBox[{"-", "1"}], "+", 
    RowBox[{"h", " ", "sf"}], "+", "sm", "-", 
    RowBox[{"h", " ", "sf", " ", "sm"}], "-", "\[Delta]", "+", 
    RowBox[{"h", " ", "sf", " ", "\[Delta]"}], "+", 
    RowBox[{"sm", " ", "\[Delta]"}], "-", 
    RowBox[{"h", " ", "sf", " ", "sm", " ", "\[Delta]"}], "-", "\[Lambda]", 
    "+", 
    RowBox[{"h", " ", "sf", " ", "\[Lambda]"}], "+", 
    RowBox[{"2", " ", 
     SuperscriptBox["\[Lambda]", "2"]}]}], ")"}]}]], "Output",
 CellChangeTimes->{
  3.7812740033403482`*^9, 3.7814573218417273`*^9, 3.7818848609506083`*^9, 
   3.781952821819541*^9, 3.78195289751086*^9, {3.7819538958219957`*^9, 
   3.781953915448163*^9}, 3.7821306544665728`*^9, 3.782747157952016*^9, 
   3.7830788934301643`*^9, 3.785162683931014*^9, 3.785164973948011*^9, 
   3.789732648885466*^9, 3.795174657109274*^9, 3.799320669052409*^9, 
   3.799340142599029*^9, 3.799756425128694*^9, 3.800007655671279*^9, 
   3.8002681846876783`*^9, 3.813930901799224*^9, {3.8139309343336077`*^9, 
   3.8139310598724957`*^9}, 3.8140037587039423`*^9, 3.814003805149556*^9, 
   3.814081435651245*^9, 3.8140996368748817`*^9},
 CellLabel->"Out[76]=",ExpressionUUID->"637a97a0-6d35-4609-bb59-af5cbfa75158"]
}, Open  ]],

Cell[TextData[{
 "The factor ",
 Cell[BoxData[
  RowBox[{
   RowBox[{"-", 
    FractionBox["1", "4"]}], " ", 
   SuperscriptBox["\[Lambda]", "2"], " ", 
   RowBox[{"(", 
    RowBox[{
     RowBox[{"-", "4"}], "+", "b", "+", 
     RowBox[{"2", " ", "\[Lambda]"}]}], ")"}]}]],
  CellChangeTimes->{
   3.7812740033403482`*^9, 3.7814573218417273`*^9, 3.7818848609506083`*^9, 
    3.781952821819541*^9, 3.78195289751086*^9, {3.7819538958219957`*^9, 
    3.781953915448163*^9}, 3.7821306544665728`*^9, 3.782747157952016*^9, 
    3.7830788934301643`*^9, 3.785162683931014*^9, 3.785164973948011*^9, 
    3.789732648885466*^9, 3.795174657109274*^9, 3.799320669052409*^9, 
    3.799340142599029*^9, 3.799756425128694*^9, 3.800007655671279*^9, 
    3.8002681846876783`*^9, 3.813930901799224*^9, {3.8139309343336077`*^9, 
    3.8139310598724957`*^9}, 3.8140037587039423`*^9, 3.814003805149556*^9},
  ExpressionUUID->"442ae86d-947b-47e1-83d4-69e2de14e677"],
 " gives two 0 eigenvalues and and eigenvalue of ",
 Cell[BoxData[
  FractionBox[
   RowBox[{"4", "-", "b"}], "2"]],
  CellChangeTimes->{3.814003986297804*^9},ExpressionUUID->
  "c0806308-6bd2-4390-977e-a2719fe182c4"],
 ", corresponding to the change in overall population size. Thus, we can see \
that the demographic dynamics are orthogonal to the evolutionary dynamics. \n\
\nHere,  we divide through by this factor to simplify the equation and \
evaluate the evolutionary dynamics. We then write the characteristic \
polynomial in the form\n",
 Cell[BoxData[
  RowBox[{
   SuperscriptBox["a\[Lambda]", "2"], "+", "b\[Lambda]", "+", "c"}]],
  CellChangeTimes->{
   3.7812740321927767`*^9, 3.781274073867482*^9, 3.7814285214279013`*^9, 
    3.781457321850852*^9, 3.78188486097983*^9, 3.781952821828101*^9, 
    3.7819528975187283`*^9, {3.781953895876915*^9, 3.7819539154956713`*^9}, 
    3.7821306544745502`*^9, 3.782747157961987*^9, 3.7830788934387407`*^9, 
    3.785162684364028*^9, 3.785164974025779*^9, 3.789732649060177*^9, 
    3.79517465722147*^9, 3.799320669259963*^9, 3.799340142805358*^9, 
    3.799756425312364*^9, 3.800007655751677*^9, 3.80026818474804*^9, 
    3.8139309588718987`*^9, 3.813931061212427*^9, 3.814004119936627*^9},
  ExpressionUUID->"aa07dfbe-082f-4b7a-a382-1ac7d323fa4f"],
 " \nfinding the a, b, and c coefficients. "
}], "Text",
 CellChangeTimes->{{3.8140039558834753`*^9, 3.8140040343097057`*^9}, {
  3.8140040782287703`*^9, 3.814004106285927*^9}, {3.814004210438005*^9, 
  3.814004280376747*^9}},ExpressionUUID->"0335147d-eb3f-47af-bc72-\
ab9e7a3e1f20"],

Cell[CellGroupData[{

Cell[BoxData[{
 RowBox[{
  RowBox[{"Collect", "[", 
   RowBox[{
    RowBox[{"charpolynomial0", "/", 
     RowBox[{"(", 
      RowBox[{
       RowBox[{"-", 
        FractionBox["1", "4"]}], " ", 
       SuperscriptBox["\[Lambda]", "2"], " ", 
       RowBox[{"(", 
        RowBox[{
         RowBox[{"-", "4"}], "+", "b", "+", 
         RowBox[{"2", " ", "\[Lambda]"}]}], ")"}]}], ")"}]}], ",", 
    "\[Lambda]", ",", "Simplify"}], "]"}], ";"}], "\[IndentingNewLine]", 
 RowBox[{
  RowBox[{"Coefficient", "[", 
   RowBox[{"%", ",", 
    RowBox[{"\[Lambda]", "^", "2"}]}], "]"}], ";"}], "\[IndentingNewLine]", 
 RowBox[{
  RowBox[{"charpolySimpleACoef0", "=", 
   RowBox[{"%%", "/", "%"}]}], " ", 
  RowBox[{"(*", 
   RowBox[{
   "simplify", " ", "so", " ", "that", " ", "the", " ", "a", " ", 
    "coefficient", " ", "is", " ", "1"}], "*)"}]}], "\[IndentingNewLine]", 
 RowBox[{
  RowBox[{"acoef0", "=", 
   RowBox[{"Coefficient", "[", 
    RowBox[{"charpolySimpleACoef0", ",", 
     RowBox[{"\[Lambda]", "^", "2"}]}], "]"}]}], ";"}], "\[IndentingNewLine]", 
 RowBox[{
  RowBox[{"bcoef0", "=", 
   RowBox[{"Coefficient", "[", 
    RowBox[{"charpolySimpleACoef0", ",", "\[Lambda]"}], "]"}]}], 
  ";"}], "\[IndentingNewLine]", 
 RowBox[{
  RowBox[{"ccoef0", "=", 
   RowBox[{
    RowBox[{"charpolySimpleACoef0", " ", "-", " ", 
     RowBox[{"bcoef0", "*", "\[Lambda]"}], "-", 
     RowBox[{"acoef0", "*", 
      RowBox[{"\[Lambda]", "^", "2"}]}]}], " ", "//", " ", "Simplify"}]}], 
  ";"}]}], "Input",
 CellChangeTimes->{{3.7812740703585978`*^9, 3.7812740731784153`*^9}, {
   3.781428516566597*^9, 3.781428520312667*^9}, {3.814003772626279*^9, 
   3.814003773254695*^9}, 3.814004117518854*^9, {3.814004306769434*^9, 
   3.814004317051339*^9}},
 CellLabel->"In[77]:=",ExpressionUUID->"87f0de7d-0098-4f70-9f16-10470b21eded"],

Cell[BoxData[
 RowBox[{
  FractionBox["1", "2"], " ", 
  RowBox[{"(", 
   RowBox[{
    RowBox[{
     RowBox[{"-", 
      RowBox[{"(", 
       RowBox[{
        RowBox[{"-", "1"}], "+", 
        RowBox[{"h", " ", "sf"}]}], ")"}]}], " ", 
     RowBox[{"(", 
      RowBox[{
       RowBox[{"-", "1"}], "+", "sm"}], ")"}], " ", 
     RowBox[{"(", 
      RowBox[{"1", "+", "\[Delta]"}], ")"}]}], "+", 
    RowBox[{
     RowBox[{"(", 
      RowBox[{
       RowBox[{"-", "1"}], "+", 
       RowBox[{"h", " ", "sf"}]}], ")"}], " ", "\[Lambda]"}], "+", 
    RowBox[{"2", " ", 
     SuperscriptBox["\[Lambda]", "2"]}]}], ")"}]}]], "Output",
 CellChangeTimes->{
  3.7812740321927767`*^9, 3.781274073867482*^9, 3.7814285214279013`*^9, 
   3.781457321850852*^9, 3.78188486097983*^9, 3.781952821828101*^9, 
   3.7819528975187283`*^9, {3.781953895876915*^9, 3.7819539154956713`*^9}, 
   3.7821306544745502`*^9, 3.782747157961987*^9, 3.7830788934387407`*^9, 
   3.785162684364028*^9, 3.785164974025779*^9, 3.789732649060177*^9, 
   3.79517465722147*^9, 3.799320669259963*^9, 3.799340142805358*^9, 
   3.799756425312364*^9, 3.800007655751677*^9, 3.80026818474804*^9, 
   3.8139309588718987`*^9, 3.813931061212427*^9, 3.814004119936627*^9, 
   3.814081440403721*^9, 3.814099640905692*^9},
 CellLabel->"Out[79]=",ExpressionUUID->"bf0bcb4e-ee9c-4221-9892-b8b157a60955"]
}, Open  ]],

Cell["\<\
The discriminant must be non-negative for the eigenvalues to be real\
\>", "Text",
 CellChangeTimes->{{3.781428281916833*^9, 
  3.781428302970292*^9}},ExpressionUUID->"c94e92f7-10e6-4e8f-8282-\
1077be6164b8"],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{
  RowBox[{
   RowBox[{"bcoef0", "^", "2"}], "-", 
   RowBox[{"4", "acoef0", " ", "ccoef0"}]}], " ", "//", " ", 
  "FullSimplify"}]], "Input",
 CellChangeTimes->{{3.781428313265222*^9, 3.781428316168023*^9}, {
  3.781428426467711*^9, 3.7814285260999413`*^9}},
 CellLabel->"In[83]:=",ExpressionUUID->"8eb1f776-89ec-4765-9c17-11aa44331c0e"],

Cell[BoxData[
 RowBox[{
  FractionBox["1", "4"], " ", 
  RowBox[{"(", 
   RowBox[{
    RowBox[{"-", "1"}], "+", 
    RowBox[{"h", " ", "sf"}]}], ")"}], " ", 
  RowBox[{"(", 
   RowBox[{
    RowBox[{"-", "1"}], "+", 
    RowBox[{"h", " ", "sf"}], "+", 
    RowBox[{"8", " ", 
     RowBox[{"(", 
      RowBox[{
       RowBox[{"-", "1"}], "+", "sm"}], ")"}], " ", 
     RowBox[{"(", 
      RowBox[{"1", "+", "\[Delta]"}], ")"}]}]}], ")"}]}]], "Output",
 CellChangeTimes->{
  3.781428316942605*^9, {3.781428429373097*^9, 3.781428527133175*^9}, 
   3.7814573219234743`*^9, 3.781884861028933*^9, 3.7819528218734293`*^9, 
   3.781952897577923*^9, {3.781953895926058*^9, 3.781953915547658*^9}, 
   3.782130654518096*^9, 3.7827471580481977`*^9, 3.7830788935329447`*^9, 
   3.785162684709345*^9, 3.785164974150434*^9, 3.789732649230698*^9, 
   3.795174657383437*^9, 3.799320669605383*^9, 3.7993401431414948`*^9, 
   3.799756425497038*^9, 3.80000765591433*^9, 3.80026818485813*^9, 
   3.8139310639623957`*^9, 3.8140041234837303`*^9, 3.814081441766411*^9, 
   3.814099642754355*^9},
 CellLabel->"Out[83]=",ExpressionUUID->"32f8574f-7b26-4923-906e-d17d51203240"]
}, Open  ]],

Cell[TextData[{
 "We can see that this is always true as the two terms in the product are \
always \[LessEqual] 0.\n\nThe conditions for the above characteristic \
polynomial ",
 Cell[BoxData[
  FormBox["p", TraditionalForm]],ExpressionUUID->
  "755e8bb5-211e-48b0-b289-ec3cc6ca7764"],
 "  to have roots between -1 and 1 are:\n1. ",
 Cell[BoxData[
  FormBox[
   RowBox[{
    RowBox[{"p", "(", "1", ")"}], " ", ">", " ", "0"}], TraditionalForm]],
  ExpressionUUID->"e7e02f9b-b684-4b6b-8a57-f80af9b5671d"],
 "\n2. ",
 Cell[BoxData[
  FormBox[
   RowBox[{
    RowBox[{"p", "(", 
     RowBox[{"-", "1"}], ")"}], " ", ">", " ", "0"}], TraditionalForm]],
  ExpressionUUID->"1fb0edc3-6cef-4ca9-bb33-ce2a1ddff83b"],
 "\n3.  ",
 Cell[BoxData[
  FormBox[
   RowBox[{
    RowBox[{"-", "2"}], " ", "<", " ", "b", " ", "<", " ", "2"}], 
   TraditionalForm]],ExpressionUUID->"f6453ca4-b24e-4563-bba1-cd1f0c3a454b"],
 "\nWhen any of these are false, the largest eigenvalue is ",
 Cell[BoxData[
  FormBox[
   RowBox[{">", 
    RowBox[{"1", "\:201a", " "}]}], TraditionalForm]],ExpressionUUID->
  "249e9dbd-e6da-4731-9e45-9e037d99d0db"],
 "and the equilibrium is unstable. Conditions 2 and 3 are always true:"
}], "Text",
 CellChangeTimes->{{3.781428661248453*^9, 3.781428712666779*^9}, {
  3.781428780656464*^9, 3.781428782114583*^9}, {3.781429946730907*^9, 
  3.7814299482572327`*^9}, {3.7814360173472347`*^9, 3.781436101352002*^9}, {
  3.781436149893372*^9, 3.7814362551794662`*^9}, {3.781436330451419*^9, 
  3.781436341934223*^9}, {3.781436393282946*^9, 3.781436444944702*^9}, {
  3.781436680869691*^9, 3.78143669857165*^9}, {3.814004641338129*^9, 
  3.814004642068164*^9}},ExpressionUUID->"f7b0820d-78d7-4eca-aee7-\
bc99661078e5"],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"FullSimplify", "[", 
  RowBox[{
   RowBox[{
    RowBox[{"(", 
     RowBox[{"charpolySimpleACoef0", "/.", 
      RowBox[{"\[Lambda]", "\[Rule]", 
       RowBox[{"-", "1"}]}]}], " ", ")"}], ">", " ", "0"}], ",", "  ", 
   RowBox[{"{", 
    RowBox[{
     RowBox[{"1", ">", " ", "\[Delta]", ">", "0"}], ",", 
     RowBox[{"1", ">", "h", ">", "0"}], ",", 
     RowBox[{"1", ">", "sm", ">", "0"}], ",", 
     RowBox[{"0", "<", "sf", "<", "1"}]}], "}"}]}], "]"}]], "Input",
 CellChangeTimes->{{3.781436736869988*^9, 3.781436739183576*^9}},
 CellLabel->"In[84]:=",ExpressionUUID->"1fd3ad56-38b9-4951-ad80-c20f5f77b24c"],

Cell[BoxData["True"], "Output",
 CellChangeTimes->{
  3.781436739697763*^9, 3.78145732216691*^9, 3.781884861062943*^9, 
   3.7819528219341993`*^9, 3.781952897819153*^9, {3.781953895978655*^9, 
   3.781953915596232*^9}, 3.782130654549685*^9, 3.782747158425377*^9, 
   3.7830788938290977`*^9, 3.785162685017118*^9, 3.785164974273959*^9, 
   3.789732649397847*^9, 3.795174657691847*^9, 3.799320669976574*^9, 
   3.799340143764106*^9, 3.799756425685603*^9, 3.8000076560762587`*^9, 
   3.800268185146523*^9, 3.814004596139834*^9, 3.814081444299515*^9, 
   3.8140996449189463`*^9},
 CellLabel->"Out[84]=",ExpressionUUID->"b590d123-c81d-438a-bdc1-5566fea3d97c"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"FullSimplify", "[", 
  RowBox[{
   RowBox[{
    RowBox[{"-", "2"}], "<", "bcoef0", "<", "2"}], ",", 
   RowBox[{"{", 
    RowBox[{
     RowBox[{"1", ">", " ", "\[Delta]", ">", "0"}], ",", 
     RowBox[{"1", ">", "h", ">", "0"}], ",", 
     RowBox[{"1", ">", "sm", ">", "0"}], ",", 
     RowBox[{"0", "<", "sf", "<", "1"}]}], "}"}]}], "]"}]], "Input",
 CellChangeTimes->{{3.781436745619834*^9, 3.781436768260647*^9}},
 CellLabel->"In[85]:=",ExpressionUUID->"83a4d73d-8803-4057-8aba-5ec72c36236e"],

Cell[BoxData["True"], "Output",
 CellChangeTimes->{
  3.7814367687176857`*^9, 3.781457322173564*^9, 3.781884861071455*^9, 
   3.7819528219793453`*^9, 3.7819528978269367`*^9, {3.7819538959919233`*^9, 
   3.781953915604405*^9}, 3.782130654581544*^9, 3.782747158434576*^9, 
   3.78307889402479*^9, 3.785162685768771*^9, 3.785164974470911*^9, 
   3.789732649484159*^9, 3.795174657881214*^9, 3.7993206700828447`*^9, 
   3.799340144112084*^9, 3.799756425778799*^9, 3.8000076561625147`*^9, 
   3.800268185279215*^9, 3.814004597392589*^9, 3.8140814447417097`*^9, 
   3.814099645667891*^9},
 CellLabel->"Out[85]=",ExpressionUUID->"8ed0a354-8746-43d1-b6dd-ad6a4d707b99"]
}, Open  ]],

Cell["\<\
But condition 1 fails when the following inequality is true, giving us the \
instability condition.\
\>", "Text",
 CellChangeTimes->{{3.7814367744575577`*^9, 
  3.7814368080369053`*^9}},ExpressionUUID->"79fa2b1e-ea4f-4428-931c-\
a4bf0561909a"],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"unstabConditions0", " ", "=", " ", 
  RowBox[{"FullSimplify", "[", 
   RowBox[{
    RowBox[{
     RowBox[{"(", 
      RowBox[{"charpolySimpleACoef0", "/.", 
       RowBox[{"\[Lambda]", "\[Rule]", "1"}]}], " ", ")"}], "<", " ", "0"}], 
    ",", "  ", 
    RowBox[{"{", 
     RowBox[{
      RowBox[{"1", ">", " ", "\[Delta]", ">", "0"}], ",", 
      RowBox[{"1", ">", "h", ">", "0"}], ",", 
      RowBox[{"1", ">", "sm", ">", "0"}], ",", 
      RowBox[{"0", "<", "sf", "<", "1"}]}], "}"}]}], "]"}]}]], "Input",
 CellChangeTimes->{{3.781429951183529*^9, 3.7814301179474297`*^9}, {
   3.7814302541642017`*^9, 3.781430298190481*^9}, {3.7814303383056297`*^9, 
   3.781430361497563*^9}, {3.7814312736199713`*^9, 3.78143127584548*^9}, {
   3.781436466857931*^9, 3.781436560917033*^9}, {3.781436631602688*^9, 
   3.781436644697918*^9}, 3.8140046082541723`*^9, {3.814004682873187*^9, 
   3.814004713133438*^9}},
 CellLabel->"In[86]:=",ExpressionUUID->"3a0f72c0-e0a6-48fa-808f-a744cf23697c"],

Cell[BoxData[
 RowBox[{
  RowBox[{"sm", "+", 
   RowBox[{"sm", " ", "\[Delta]"}], "+", 
   RowBox[{"h", " ", "sf", " ", 
    RowBox[{"(", 
     RowBox[{"2", "+", "\[Delta]", "-", 
      RowBox[{"sm", " ", 
       RowBox[{"(", 
        RowBox[{"1", "+", "\[Delta]"}], ")"}]}]}], ")"}]}]}], "<", 
  "\[Delta]"}]], "Output",
 CellChangeTimes->{
  3.7814300190036163`*^9, {3.781430050891747*^9, 3.781430119181625*^9}, {
   3.781430258842821*^9, 3.7814302987538567`*^9}, {3.781430357589197*^9, 
   3.7814303619226294`*^9}, 3.781431276347818*^9, {3.781436471134864*^9, 
   3.7814365067294188`*^9}, 3.781436668610888*^9, 3.7814573222715*^9, 
   3.781884861204401*^9, 3.781952822089045*^9, 3.781952897921483*^9, {
   3.781953896105276*^9, 3.781953915747797*^9}, 3.782130654684663*^9, 
   3.782747158540249*^9, 3.783078894199912*^9, 3.7851626859809427`*^9, 
   3.785164974645007*^9, 3.789732649767761*^9, 3.795174658046834*^9, 
   3.799320670294146*^9, 3.7993401443157*^9, 3.79975642605756*^9, 
   3.800007656424817*^9, 3.800268185436696*^9, 3.8139311136920567`*^9, {
   3.814004601006237*^9, 3.8140046096271353`*^9}, {3.814004683951476*^9, 
   3.814004713521501*^9}, 3.814081445877571*^9, 3.814099647466834*^9},
 CellLabel->"Out[86]=",ExpressionUUID->"bde67e1d-76de-48fb-9d3a-2d59a62f2ed6"]
}, Open  ]],

Cell["which can be re-written using the w fitness notation", "Text",
 CellChangeTimes->{{3.814004650674572*^9, 
  3.814004710715838*^9}},ExpressionUUID->"6ff0c034-af49-4d51-8ea1-\
6d121d8be3fa"],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"Simplify", "[", 
  RowBox[{"unstabConditions0", ",", " ", 
   RowBox[{"{", 
    RowBox[{
     RowBox[{"wXdY", " ", "\[Equal]", " ", 
      RowBox[{"1", " ", "-", " ", "sm"}]}], ",", 
     RowBox[{"wXdX", " ", "\[Equal]", " ", 
      RowBox[{"1", " ", "-", " ", 
       RowBox[{"h", " ", "sf"}]}]}], ",", 
     RowBox[{"wXdXd", " ", "\[Equal]", " ", 
      RowBox[{"1", " ", "-", " ", "sf"}]}]}], "}"}]}], "]"}]], "Input",
 CellChangeTimes->{{3.782747173005354*^9, 3.782747212585889*^9}, {
  3.782748801098444*^9, 3.7827488122111607`*^9}, {3.783078875883896*^9, 
  3.783078876006201*^9}},
 CellLabel->"In[87]:=",ExpressionUUID->"40f9b588-14ff-4c1a-9ccc-73f918eb34ad"],

Cell[BoxData[
 RowBox[{
  RowBox[{"wXdX", "+", 
   RowBox[{"wXdX", " ", "wXdY", " ", 
    RowBox[{"(", 
     RowBox[{"1", "+", "\[Delta]"}], ")"}]}]}], ">", "2"}]], "Output",
 CellChangeTimes->{{3.7827471930133467`*^9, 3.782747213122994*^9}, {
   3.782748802357192*^9, 3.782748813081629*^9}, {3.7830788773278503`*^9, 
   3.783078894293872*^9}, 3.78516268619658*^9, 3.7851649747693663`*^9, 
   3.789732650076954*^9, 3.795174658156262*^9, 3.799320670556473*^9, 
   3.799340144519253*^9, 3.799756426506791*^9, 3.8000076569522038`*^9, 
   3.800268185528966*^9, 3.814004612665675*^9, 3.8140047147797937`*^9, 
   3.81408144704889*^9, 3.814099648786633*^9},
 CellLabel->"Out[87]=",ExpressionUUID->"68c12345-b81e-4431-86b9-455d5c72272b"]
}, Open  ]],

Cell["This is equivalent to \[Phi], as written in the text. ", "Text",
 CellChangeTimes->{{3.8140997893825817`*^9, 
  3.8140998253881407`*^9}},ExpressionUUID->"4c9cab46-2e08-4808-a069-\
962cb6675d59"],

Cell[CellGroupData[{

Cell[BoxData[{
 RowBox[{
  RowBox[{"\[Phi]", " ", "=", " ", 
   RowBox[{"\[Delta]", " ", "-", " ", 
    RowBox[{"(", 
     RowBox[{"sm", "+", 
      RowBox[{"sm", " ", "\[Delta]"}], "+", 
      RowBox[{"h", " ", "sf", " ", 
       RowBox[{"(", 
        RowBox[{"2", "+", "\[Delta]", "-", 
         RowBox[{"sm", " ", 
          RowBox[{"(", 
           RowBox[{"1", "+", "\[Delta]"}], ")"}]}]}], ")"}]}]}], ")"}]}]}], 
  ";"}], "\[IndentingNewLine]", 
 RowBox[{
  RowBox[{"Simplify", "[", 
   RowBox[{"Reduce", "[", 
    RowBox[{"{", 
     RowBox[{"unstabConditions0", ",", 
      RowBox[{"1", ">", " ", "\[Delta]", ">", "0"}], ",", 
      RowBox[{"1", ">", "h", ">", "0"}], ",", 
      RowBox[{"1", ">", "sm", ">", "0"}], ",", 
      RowBox[{"0", "<", "sf", "<", "1"}]}], "}"}], "]"}], "]"}], 
  ";"}], "\[IndentingNewLine]", 
 RowBox[{
  RowBox[{"Simplify", "[", 
   RowBox[{"Reduce", "[", 
    RowBox[{"{", 
     RowBox[{
      RowBox[{"\[Phi]", ">", "0"}], ",", 
      RowBox[{"1", ">", " ", "\[Delta]", ">", "0"}], ",", 
      RowBox[{"1", ">", "h", ">", "0"}], ",", 
      RowBox[{"1", ">", "sm", ">", "0"}], ",", 
      RowBox[{"0", "<", "sf", "<", "1"}]}], "}"}], "]"}], "]"}], 
  ";"}], "\[IndentingNewLine]", 
 RowBox[{"%", "\[Equal]", "%%"}]}], "Input",
 CellChangeTimes->{{3.814099721522447*^9, 3.814099787257207*^9}, {
  3.814099826366563*^9, 3.814099846242858*^9}, {3.8140999152713537`*^9, 
  3.8140999167560987`*^9}},
 CellLabel->
  "In[149]:=",ExpressionUUID->"0f574c14-d9e6-44cc-904d-2c74f53dd191"],

Cell[BoxData["True"], "Output",
 CellChangeTimes->{{3.814099725820879*^9, 3.814099772444579*^9}, {
   3.814099830616767*^9, 3.814099847097879*^9}, 3.814099917193047*^9},
 CellLabel->
  "Out[152]=",ExpressionUUID->"e4688023-1827-4f6c-9abf-43a61c43edb6"]
}, Open  ]]
}, Closed]],

Cell[CellGroupData[{

Cell["drive fixed (pm=pf=1) equilibrium", "Subsubsection",
 CellChangeTimes->{{3.781273735316922*^9, 3.781273740747805*^9}, 
   3.814018765219983*^9},ExpressionUUID->"5b8db823-0ef9-4c73-9c07-\
07e593c1b7a6"],

Cell["\<\
We proceed similarly to before, but now considering the other equilibrium:\
\>", "Text",
 CellChangeTimes->{{3.7814369207715807`*^9, 
  3.781436938486001*^9}},ExpressionUUID->"c52c3617-685e-46eb-9b2f-\
877de8aa1afe"],

Cell[CellGroupData[{

Cell[BoxData[{
 RowBox[{
  RowBox[{
   RowBox[{"Simplify", "[", 
    RowBox[{
     RowBox[{
      RowBox[{
       RowBox[{"jacob", "/.", "subeq"}], "/.", 
       RowBox[{"pf", "\[Rule]", "1"}]}], "/.", 
      RowBox[{"pm", "\[Rule]", "1"}]}], "/.", "fixedDrivePopSizeSol"}], "]"}],
    "[", 
   RowBox[{"[", "1", "]"}], "]"}], ";"}], "\[IndentingNewLine]", 
 RowBox[{"charpolynomial1", "=", 
  RowBox[{"Factor", "[", 
   RowBox[{"Det", "[", 
    RowBox[{"%", "-", 
     RowBox[{"\[Lambda]", "*", 
      RowBox[{"IdentityMatrix", "[", "5", "]"}]}]}], "]"}], "]"}]}]}], "Input",\

 CellChangeTimes->{{3.814004729064748*^9, 3.814004743901525*^9}, {
  3.8140047795551653`*^9, 3.814004791431684*^9}, {3.814005039784062*^9, 
  3.814005043586812*^9}, {3.814008492634818*^9, 3.814008493127555*^9}},
 CellLabel->"In[88]:=",ExpressionUUID->"558514c0-c8c7-4b08-85da-37c3c8b66b22"],

Cell[BoxData[
 FractionBox[
  RowBox[{
   RowBox[{"(", 
    RowBox[{"4", "-", "b", "+", 
     RowBox[{"b", " ", "sf"}], "-", 
     RowBox[{"b", " ", "\[Delta]"}], "+", 
     RowBox[{"b", " ", "sf", " ", "\[Delta]"}], "-", 
     RowBox[{"2", " ", "\[Lambda]"}]}], ")"}], " ", 
   SuperscriptBox["\[Lambda]", "2"], " ", 
   RowBox[{"(", 
    RowBox[{
     RowBox[{"-", "1"}], "+", 
     RowBox[{"h", " ", "sf"}], "-", "\[Lambda]", "+", 
     RowBox[{"h", " ", "sf", " ", "\[Lambda]"}], "+", 
     RowBox[{"sm", " ", "\[Lambda]"}], "-", 
     RowBox[{"h", " ", "sf", " ", "sm", " ", "\[Lambda]"}], "-", 
     RowBox[{"\[Delta]", " ", "\[Lambda]"}], "+", 
     RowBox[{"h", " ", "sf", " ", "\[Delta]", " ", "\[Lambda]"}], "+", 
     RowBox[{"sm", " ", "\[Delta]", " ", "\[Lambda]"}], "-", 
     RowBox[{"h", " ", "sf", " ", "sm", " ", "\[Delta]", " ", "\[Lambda]"}], 
     "+", 
     RowBox[{"2", " ", 
      SuperscriptBox["\[Lambda]", "2"]}], "-", 
     RowBox[{"2", " ", "sf", " ", 
      SuperscriptBox["\[Lambda]", "2"]}], "-", 
     RowBox[{"2", " ", "sm", " ", 
      SuperscriptBox["\[Lambda]", "2"]}], "+", 
     RowBox[{"2", " ", "sf", " ", "sm", " ", 
      SuperscriptBox["\[Lambda]", "2"]}], "+", 
     RowBox[{"2", " ", "\[Delta]", " ", 
      SuperscriptBox["\[Lambda]", "2"]}], "-", 
     RowBox[{"2", " ", "sf", " ", "\[Delta]", " ", 
      SuperscriptBox["\[Lambda]", "2"]}], "-", 
     RowBox[{"2", " ", "sm", " ", "\[Delta]", " ", 
      SuperscriptBox["\[Lambda]", "2"]}], "+", 
     RowBox[{"2", " ", "sf", " ", "sm", " ", "\[Delta]", " ", 
      SuperscriptBox["\[Lambda]", "2"]}]}], ")"}]}], 
  RowBox[{"4", " ", 
   RowBox[{"(", 
    RowBox[{
     RowBox[{"-", "1"}], "+", "sf"}], ")"}], " ", 
   RowBox[{"(", 
    RowBox[{
     RowBox[{"-", "1"}], "+", "sm"}], ")"}], " ", 
   RowBox[{"(", 
    RowBox[{"1", "+", "\[Delta]"}], ")"}]}]]], "Output",
 CellChangeTimes->{{3.814004786393588*^9, 3.814004791710847*^9}, 
   3.8140050447489643`*^9, 3.814005210911131*^9, 3.8140814605227623`*^9, 
   3.814099660811801*^9},
 CellLabel->"Out[89]=",ExpressionUUID->"8d8881fb-8291-487b-a5a6-379638330ddb"]
}, Open  ]],

Cell[TextData[{
 "The factor ",
 Cell[BoxData[
  FractionBox[
   RowBox[{
    RowBox[{"(", 
     RowBox[{"4", "-", "b", "+", 
      RowBox[{"b", " ", "sf"}], "-", 
      RowBox[{"b", " ", "\[Delta]"}], "+", 
      RowBox[{"b", " ", "sf", " ", "\[Delta]"}], "-", 
      RowBox[{"2", " ", "\[Lambda]"}]}], ")"}], " ", 
    SuperscriptBox["\[Lambda]", "2"], " "}], 
   RowBox[{"4", " ", 
    RowBox[{"(", 
     RowBox[{
      RowBox[{"-", "1"}], "+", "sf"}], ")"}], " ", 
    RowBox[{"(", 
     RowBox[{
      RowBox[{"-", "1"}], "+", "sm"}], ")"}], " ", 
    RowBox[{"(", 
     RowBox[{"1", "+", "\[Delta]"}], ")"}]}]]],
  CellChangeTimes->{
   3.7812740033403482`*^9, 3.7814573218417273`*^9, 3.7818848609506083`*^9, 
    3.781952821819541*^9, 3.78195289751086*^9, {3.7819538958219957`*^9, 
    3.781953915448163*^9}, 3.7821306544665728`*^9, 3.782747157952016*^9, 
    3.7830788934301643`*^9, 3.785162683931014*^9, 3.785164973948011*^9, 
    3.789732648885466*^9, 3.795174657109274*^9, 3.799320669052409*^9, 
    3.799340142599029*^9, 3.799756425128694*^9, 3.800007655671279*^9, 
    3.8002681846876783`*^9, 3.813930901799224*^9, {3.8139309343336077`*^9, 
    3.8139310598724957`*^9}, 3.8140037587039423`*^9, 3.814003805149556*^9},
  ExpressionUUID->"fd392500-ab7a-41ab-9dc3-756aa1e78102"],
 " gives two 0 eigenvalues and and eigenvalue of ",
 Cell[BoxData[
  RowBox[{"2", "+", 
   RowBox[{
    FractionBox["1", "2"], " ", "b", " ", 
    RowBox[{"(", 
     RowBox[{
      RowBox[{"-", "1"}], "+", "sf"}], ")"}], " ", 
    RowBox[{"(", 
     RowBox[{"1", "+", "\[Delta]"}], ")"}]}]}]],
  CellChangeTimes->{3.814003986297804*^9},ExpressionUUID->
  "dfc8d803-41ad-4813-8e86-8b57ffd38a7b"],
 ", corresponding to the stability of the ecological equilibrium population \
size. Thus, we can see that the demographic dynamics are orthogonal to the \
evolutionary dynamics. \n\nHere,  we divide through by this factor to \
simplify the equation and evaluate the evolutionary dynamics. We then write \
the characteristic polynomial in the form\n",
 Cell[BoxData[
  RowBox[{
   SuperscriptBox["a\[Lambda]", "2"], "+", "b\[Lambda]", "+", "c"}]],
  CellChangeTimes->{
   3.7812740321927767`*^9, 3.781274073867482*^9, 3.7814285214279013`*^9, 
    3.781457321850852*^9, 3.78188486097983*^9, 3.781952821828101*^9, 
    3.7819528975187283`*^9, {3.781953895876915*^9, 3.7819539154956713`*^9}, 
    3.7821306544745502`*^9, 3.782747157961987*^9, 3.7830788934387407`*^9, 
    3.785162684364028*^9, 3.785164974025779*^9, 3.789732649060177*^9, 
    3.79517465722147*^9, 3.799320669259963*^9, 3.799340142805358*^9, 
    3.799756425312364*^9, 3.800007655751677*^9, 3.80026818474804*^9, 
    3.8139309588718987`*^9, 3.813931061212427*^9, 3.814004119936627*^9},
  ExpressionUUID->"bfefafb3-6f12-4e9d-b61d-f33ae1949a7e"],
 " \nfinding the a, b, and c coefficients. "
}], "Text",
 CellChangeTimes->{{3.814004985097933*^9, 3.814004991560975*^9}, {
  3.814005062086405*^9, 3.814005085289956*^9}, {3.814005142323441*^9, 
  3.8140051616191607`*^9}},ExpressionUUID->"f502fdb9-4698-4f13-8d48-\
f3fa156d4057"],

Cell[CellGroupData[{

Cell[BoxData[{
 RowBox[{"Simplify", "[", 
  RowBox[{"charpolynomial1", "/", 
   RowBox[{"(", 
    FractionBox[
     RowBox[{
      RowBox[{"(", 
       RowBox[{"4", "-", "b", "+", 
        RowBox[{"b", " ", "sf"}], "-", 
        RowBox[{"b", " ", "\[Delta]"}], "+", 
        RowBox[{"b", " ", "sf", " ", "\[Delta]"}], "-", 
        RowBox[{"2", " ", "\[Lambda]"}]}], ")"}], " ", 
      SuperscriptBox["\[Lambda]", "2"], " "}], 
     RowBox[{"4", " ", 
      RowBox[{"(", 
       RowBox[{
        RowBox[{"-", "1"}], "+", "sf"}], ")"}], " ", 
      RowBox[{"(", 
       RowBox[{
        RowBox[{"-", "1"}], "+", "sm"}], ")"}], " ", 
      RowBox[{"(", 
       RowBox[{"1", "+", "\[Delta]"}], ")"}]}]], ")"}]}], 
  "]"}], "\[IndentingNewLine]", 
 RowBox[{"Collect", "[", 
  RowBox[{"%", ",", "\[Lambda]", ",", "Simplify"}], "]"}]}], "Input",
 CellChangeTimes->{{3.814005280882255*^9, 3.814005284996513*^9}},
 CellLabel->"In[90]:=",ExpressionUUID->"287620a3-1950-4474-82ee-e473fcc4471f"],

Cell[BoxData[
 RowBox[{
  RowBox[{"-", "1"}], "+", 
  RowBox[{
   RowBox[{"(", 
    RowBox[{
     RowBox[{"-", "1"}], "+", "sm"}], ")"}], " ", 
   RowBox[{"(", 
    RowBox[{"1", "+", "\[Delta]"}], ")"}], " ", "\[Lambda]"}], "+", 
  RowBox[{"2", " ", 
   RowBox[{"(", 
    RowBox[{
     RowBox[{"-", "1"}], "+", "sf"}], ")"}], " ", 
   RowBox[{"(", 
    RowBox[{
     RowBox[{"-", "1"}], "+", "sm"}], ")"}], " ", 
   RowBox[{"(", 
    RowBox[{"1", "+", "\[Delta]"}], ")"}], " ", 
   SuperscriptBox["\[Lambda]", "2"]}], "+", 
  RowBox[{"h", " ", 
   RowBox[{"(", 
    RowBox[{"sf", "-", 
     RowBox[{"sf", " ", 
      RowBox[{"(", 
       RowBox[{
        RowBox[{"-", "1"}], "+", "sm"}], ")"}], " ", 
      RowBox[{"(", 
       RowBox[{"1", "+", "\[Delta]"}], ")"}], " ", "\[Lambda]"}]}], 
    ")"}]}]}]], "Output",
 CellChangeTimes->{{3.814005268764485*^9, 3.814005285448605*^9}, 
   3.814081462523357*^9, 3.8140996630638657`*^9},
 CellLabel->"Out[90]=",ExpressionUUID->"d10fe94e-1e64-45dd-870e-d886b0d227e2"],

Cell[BoxData[
 RowBox[{
  RowBox[{"-", "1"}], "+", 
  RowBox[{"h", " ", "sf"}], "+", 
  RowBox[{
   RowBox[{"(", 
    RowBox[{"1", "-", 
     RowBox[{"h", " ", "sf"}]}], ")"}], " ", 
   RowBox[{"(", 
    RowBox[{
     RowBox[{"-", "1"}], "+", "sm"}], ")"}], " ", 
   RowBox[{"(", 
    RowBox[{"1", "+", "\[Delta]"}], ")"}], " ", "\[Lambda]"}], "+", 
  RowBox[{"2", " ", 
   RowBox[{"(", 
    RowBox[{
     RowBox[{"-", "1"}], "+", "sf"}], ")"}], " ", 
   RowBox[{"(", 
    RowBox[{
     RowBox[{"-", "1"}], "+", "sm"}], ")"}], " ", 
   RowBox[{"(", 
    RowBox[{"1", "+", "\[Delta]"}], ")"}], " ", 
   SuperscriptBox["\[Lambda]", "2"]}]}]], "Output",
 CellChangeTimes->{{3.814005268764485*^9, 3.814005285448605*^9}, 
   3.814081462523357*^9, 3.814099663069263*^9},
 CellLabel->"Out[91]=",ExpressionUUID->"aca2aabb-05fa-4b80-a5e9-35b3b4acdacf"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[{
 RowBox[{
  RowBox[{"Simplify", "[", 
   RowBox[{"charpolynomial1", "/", 
    RowBox[{"(", 
     FractionBox[
      RowBox[{
       RowBox[{"(", 
        RowBox[{"4", "-", "b", "+", 
         RowBox[{"b", " ", "sf"}], "-", 
         RowBox[{"b", " ", "\[Delta]"}], "+", 
         RowBox[{"b", " ", "sf", " ", "\[Delta]"}], "-", 
         RowBox[{"2", " ", "\[Lambda]"}]}], ")"}], " ", 
       SuperscriptBox["\[Lambda]", "2"], " "}], 
      RowBox[{"4", " ", 
       RowBox[{"(", 
        RowBox[{
         RowBox[{"-", "1"}], "+", "sf"}], ")"}], " ", 
       RowBox[{"(", 
        RowBox[{
         RowBox[{"-", "1"}], "+", "sm"}], ")"}], " ", 
       RowBox[{"(", 
        RowBox[{"1", "+", "\[Delta]"}], ")"}]}]], ")"}]}], "]"}], 
  ";"}], "\[IndentingNewLine]", 
 RowBox[{
  RowBox[{"Collect", "[", 
   RowBox[{"%", ",", "\[Lambda]", ",", "Simplify"}], "]"}], 
  ";"}], "\[IndentingNewLine]", 
 RowBox[{
  RowBox[{"Coefficient", "[", 
   RowBox[{"%", ",", 
    RowBox[{"\[Lambda]", "^", "2"}]}], "]"}], ";"}], "\[IndentingNewLine]", 
 RowBox[{"charpolySimpleACoef1", "=", 
  RowBox[{"Collect", "[", 
   RowBox[{
    RowBox[{"%%", "/", "%"}], ",", "\[Lambda]", ",", "Simplify"}], 
   "]"}]}], "\[IndentingNewLine]", 
 RowBox[{
  RowBox[{"acoef1", "=", 
   RowBox[{"Coefficient", "[", 
    RowBox[{"charpolySimpleACoef1", ",", 
     RowBox[{"\[Lambda]", "^", "2"}]}], "]"}]}], ";"}], "\[IndentingNewLine]", 
 RowBox[{
  RowBox[{"bcoef1", "=", 
   RowBox[{"Coefficient", "[", 
    RowBox[{"charpolySimpleACoef1", ",", "\[Lambda]"}], "]"}]}], 
  ";"}], "\[IndentingNewLine]", 
 RowBox[{
  RowBox[{"ccoef1", "=", 
   RowBox[{"charpolySimpleACoef1", "-", " ", 
    RowBox[{"bcoef1", "*", "\[Lambda]"}], "-", 
    RowBox[{"acoef1", "*", 
     RowBox[{"\[Lambda]", "^", "2"}]}]}]}], ";"}]}], "Input",
 CellChangeTimes->{{3.814005188468329*^9, 3.814005191684287*^9}, {
  3.814005253330809*^9, 3.814005259347432*^9}, {3.8140052975628242`*^9, 
  3.8140053332696457`*^9}},
 CellLabel->"In[92]:=",ExpressionUUID->"64b86bc6-8509-4160-829f-f613f2f9eaf1"],

Cell[BoxData[
 RowBox[{
  FractionBox[
   RowBox[{
    RowBox[{"-", "1"}], "+", 
    RowBox[{"h", " ", "sf"}]}], 
   RowBox[{"2", " ", 
    RowBox[{"(", 
     RowBox[{
      RowBox[{"-", "1"}], "+", "sf"}], ")"}], " ", 
    RowBox[{"(", 
     RowBox[{
      RowBox[{"-", "1"}], "+", "sm"}], ")"}], " ", 
    RowBox[{"(", 
     RowBox[{"1", "+", "\[Delta]"}], ")"}]}]], "+", 
  FractionBox[
   RowBox[{
    RowBox[{"(", 
     RowBox[{"1", "-", 
      RowBox[{"h", " ", "sf"}]}], ")"}], " ", "\[Lambda]"}], 
   RowBox[{"2", " ", 
    RowBox[{"(", 
     RowBox[{
      RowBox[{"-", "1"}], "+", "sf"}], ")"}]}]], "+", 
  SuperscriptBox["\[Lambda]", "2"]}]], "Output",
 CellChangeTimes->{3.81400519422432*^9, 3.814005259847539*^9, 
  3.8140053004626207`*^9, 3.814005333624386*^9, 3.814081463662334*^9, 
  3.814099664043454*^9},
 CellLabel->"Out[95]=",ExpressionUUID->"162333d1-2c10-4dfd-9fbf-01c7a38de571"]
}, Open  ]],

Cell["\<\
The discriminant is nonnegative and so the eigenvalues are real:\
\>", "Text",
 CellChangeTimes->{{3.7814370117088013`*^9, 3.7814370178709927`*^9}, {
  3.7814370976392593`*^9, 
  3.781437109782298*^9}},ExpressionUUID->"e8c43d39-f3c1-4f65-a6e5-\
65b7a6edff55"],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{
  RowBox[{"bcoef1", "^", "2"}], "-", 
  RowBox[{"4", "acoef1", " ", "ccoef1"}]}]], "Input",
 CellLabel->"In[99]:=",ExpressionUUID->"142bdcdf-af7e-43fa-ba2c-0e92873ac6b0"],

Cell[BoxData[
 RowBox[{
  FractionBox[
   SuperscriptBox[
    RowBox[{"(", 
     RowBox[{"1", "-", 
      RowBox[{"h", " ", "sf"}]}], ")"}], "2"], 
   RowBox[{"4", " ", 
    SuperscriptBox[
     RowBox[{"(", 
      RowBox[{
       RowBox[{"-", "1"}], "+", "sf"}], ")"}], "2"]}]], "-", 
  FractionBox[
   RowBox[{"2", " ", 
    RowBox[{"(", 
     RowBox[{
      RowBox[{"-", "1"}], "+", 
      RowBox[{"h", " ", "sf"}]}], ")"}]}], 
   RowBox[{
    RowBox[{"(", 
     RowBox[{
      RowBox[{"-", "1"}], "+", "sf"}], ")"}], " ", 
    RowBox[{"(", 
     RowBox[{
      RowBox[{"-", "1"}], "+", "sm"}], ")"}], " ", 
    RowBox[{"(", 
     RowBox[{"1", "+", "\[Delta]"}], ")"}]}]]}]], "Output",
 CellChangeTimes->{3.81400537457533*^9, 3.81408146557857*^9, 
  3.814099665478208*^9},
 CellLabel->"Out[99]=",ExpressionUUID->"134c175d-0959-4cc6-bdeb-bf4444ea07d7"]
}, Open  ]],

Cell[TextData[{
 "The conditions for the above characteristic polynomial ",
 Cell[BoxData[
  FormBox["p", TraditionalForm]],ExpressionUUID->
  "46aa4bdb-06dc-4c1d-af93-d31b6b5295e0"],
 "  to have roots between -1 and 1 are:\n1. ",
 Cell[BoxData[
  FormBox[
   RowBox[{
    RowBox[{"p", "(", "1", ")"}], " ", ">", " ", "0"}], TraditionalForm]],
  ExpressionUUID->"495437b1-f1aa-4390-bb7c-a020e451ee9a"],
 "\n2. ",
 Cell[BoxData[
  FormBox[
   RowBox[{
    RowBox[{"p", "(", 
     RowBox[{"-", "1"}], ")"}], " ", ">", " ", "0"}], TraditionalForm]],
  ExpressionUUID->"6f478c9d-fd30-4c05-803b-399f3373494d"],
 "\n3.  ",
 Cell[BoxData[
  FormBox[
   RowBox[{
    RowBox[{"-", "2"}], " ", "<", " ", "b", " "}], TraditionalForm]],
  ExpressionUUID->"a4bb8530-891c-4a7b-900f-ab54556da2d8"],
 "\n4.  ",
 Cell[BoxData[
  FormBox[
   RowBox[{"b", " ", "<", " ", "2", " "}], TraditionalForm]],ExpressionUUID->
  "3ea6fbe1-f3ce-45c2-ac8b-3f6709ae2c51"],
 "\nWhen any of these are false, the largest eigenvalue is ",
 Cell[BoxData[
  FormBox[
   RowBox[{">", 
    RowBox[{"1", "\:201a", " "}]}], TraditionalForm]],ExpressionUUID->
  "bf544769-4d4c-4509-938b-daae51d2c318"],
 "and the equilibrium is unstable. \n\nNow checking these conditions for \
instability by reversing the inequalities required for stability"
}], "Text",
 CellChangeTimes->{{3.7814371420798693`*^9, 3.781437152698497*^9}, {
  3.814005459930065*^9, 
  3.814005579369294*^9}},ExpressionUUID->"06a59374-4759-4b18-8760-\
3a177915ed7b"],

Cell[CellGroupData[{

Cell[BoxData[{
 RowBox[{"r1", " ", "=", " ", 
  RowBox[{"FullSimplify", "[", 
   RowBox[{
    RowBox[{
     RowBox[{"(", 
      RowBox[{"charpolySimpleACoef1", "/.", 
       RowBox[{"\[Lambda]", "\[Rule]", "1"}]}], " ", ")"}], "<", " ", "0"}], 
    ",", "  ", 
    RowBox[{"{", 
     RowBox[{
      RowBox[{"1", ">", " ", "\[Delta]", ">", "0"}], ",", 
      RowBox[{"1", ">", "h", ">", "0"}], ",", 
      RowBox[{"1", ">", "sm", ">", "0"}], ",", 
      RowBox[{"0", "<", "sf", "<", "1"}]}], "}"}]}], 
   "]"}]}], "\[IndentingNewLine]", 
 RowBox[{"r2", " ", "=", 
  RowBox[{"FullSimplify", "[", 
   RowBox[{
    RowBox[{
     RowBox[{"(", 
      RowBox[{"charpolySimpleACoef1", "/.", 
       RowBox[{"\[Lambda]", "\[Rule]", 
        RowBox[{"-", "1"}]}]}], " ", ")"}], "<", " ", "0"}], ",", "  ", 
    RowBox[{"{", 
     RowBox[{
      RowBox[{"1", ">", " ", "\[Delta]", ">", "0"}], ",", 
      RowBox[{"1", ">", "h", ">", "0"}], ",", 
      RowBox[{"1", ">", "sm", ">", "0"}], ",", 
      RowBox[{"0", "<", "sf", "<", "1"}]}], "}"}]}], 
   "]"}]}], "\[IndentingNewLine]", 
 RowBox[{"r3", "=", 
  RowBox[{"FullSimplify", "[", 
   RowBox[{
    RowBox[{"bcoef1", "<", 
     RowBox[{"-", "2"}]}], ",", 
    RowBox[{"{", 
     RowBox[{
      RowBox[{"1", ">", " ", "\[Delta]", ">", "0"}], ",", 
      RowBox[{"1", ">", "h", ">", "0"}], ",", 
      RowBox[{"1", ">", "sm", ">", "0"}], ",", 
      RowBox[{"0", "<", "sf", "<", "1"}]}], "}"}]}], 
   "]"}]}], "\[IndentingNewLine]", 
 RowBox[{"r4", "=", 
  RowBox[{"FullSimplify", "[", 
   RowBox[{
    RowBox[{"bcoef1", ">", "2"}], ",", 
    RowBox[{"{", 
     RowBox[{
      RowBox[{"1", ">", " ", "\[Delta]", ">", "0"}], ",", 
      RowBox[{"1", ">", "h", ">", "0"}], ",", 
      RowBox[{"1", ">", "sm", ">", "0"}], ",", 
      RowBox[{"0", "<", "sf", "<", "1"}]}], "}"}]}], "]"}]}]}], "Input",
 CellChangeTimes->{{3.781437173950416*^9, 3.78143720186456*^9}, {
  3.781437293015765*^9, 3.781437392444491*^9}, {3.781437548946312*^9, 
  3.781437594361197*^9}, {3.781439033702999*^9, 3.781439042554227*^9}, {
  3.7814390879120607`*^9, 3.781439133547997*^9}, {3.7814394461550627`*^9, 
  3.78143945760787*^9}},
 CellLabel->
  "In[100]:=",ExpressionUUID->"41130e31-cbdf-4c46-ac55-f8d101b3d310"],

Cell[BoxData[
 RowBox[{
  RowBox[{"\[Delta]", "+", 
   RowBox[{"2", " ", "sf", " ", 
    RowBox[{"(", 
     RowBox[{
      RowBox[{"-", "1"}], "+", "sm"}], ")"}], " ", 
    RowBox[{"(", 
     RowBox[{"1", "+", "\[Delta]"}], ")"}]}], "+", 
   RowBox[{"h", " ", "sf", " ", 
    RowBox[{"(", 
     RowBox[{"2", "+", "\[Delta]", "-", 
      RowBox[{"sm", " ", 
       RowBox[{"(", 
        RowBox[{"1", "+", "\[Delta]"}], ")"}]}]}], ")"}]}]}], "<", 
  RowBox[{"sm", " ", 
   RowBox[{"(", 
    RowBox[{"1", "+", "\[Delta]"}], ")"}]}]}]], "Output",
 CellChangeTimes->{{3.781437175151822*^9, 3.781437202530854*^9}, {
   3.781437295782474*^9, 3.781437352236356*^9}, 3.7814374035892963`*^9, 
   3.781437613171399*^9, 3.781439043535357*^9, {3.781439104876899*^9, 
   3.7814391344053497`*^9}, {3.7814394535688467`*^9, 3.7814394581207047`*^9}, 
   3.781457322620882*^9, 3.78188486163474*^9, 3.781952822497775*^9, 
   3.781952898261024*^9, {3.781953896487258*^9, 3.781953916168281*^9}, 
   3.782130655037745*^9, 3.7827471589942493`*^9, 3.783078894796163*^9, 
   3.785162687556292*^9, 3.785164975528594*^9, 3.7897326510301743`*^9, 
   3.795174658845468*^9, 3.799320671848111*^9, 3.7993401455539513`*^9, 
   3.7997564277832108`*^9, 3.800007657980575*^9, 3.800268186111622*^9, 
   3.8140055881642313`*^9, 3.814081468356298*^9, 3.814099668224862*^9},
 CellLabel->
  "Out[100]=",ExpressionUUID->"6c75f31b-dfe0-4fa3-9ac8-9e688e8c8f7a"],

Cell[BoxData[
 RowBox[{
  RowBox[{"2", "+", 
   RowBox[{"3", " ", "\[Delta]"}], "+", 
   RowBox[{"sf", " ", 
    RowBox[{"(", 
     RowBox[{
      RowBox[{"-", "2"}], "-", 
      RowBox[{
       RowBox[{"(", 
        RowBox[{"2", "+", "h"}], ")"}], " ", "\[Delta]"}], "+", 
      RowBox[{
       RowBox[{"(", 
        RowBox[{"2", "+", "h"}], ")"}], " ", "sm", " ", 
       RowBox[{"(", 
        RowBox[{"1", "+", "\[Delta]"}], ")"}]}]}], ")"}]}]}], "<", 
  RowBox[{"3", " ", "sm", " ", 
   RowBox[{"(", 
    RowBox[{"1", "+", "\[Delta]"}], ")"}]}]}]], "Output",
 CellChangeTimes->{{3.781437175151822*^9, 3.781437202530854*^9}, {
   3.781437295782474*^9, 3.781437352236356*^9}, 3.7814374035892963`*^9, 
   3.781437613171399*^9, 3.781439043535357*^9, {3.781439104876899*^9, 
   3.7814391344053497`*^9}, {3.7814394535688467`*^9, 3.7814394581207047`*^9}, 
   3.781457322620882*^9, 3.78188486163474*^9, 3.781952822497775*^9, 
   3.781952898261024*^9, {3.781953896487258*^9, 3.781953916168281*^9}, 
   3.782130655037745*^9, 3.7827471589942493`*^9, 3.783078894796163*^9, 
   3.785162687556292*^9, 3.785164975528594*^9, 3.7897326510301743`*^9, 
   3.795174658845468*^9, 3.799320671848111*^9, 3.7993401455539513`*^9, 
   3.7997564277832108`*^9, 3.800007657980575*^9, 3.800268186111622*^9, 
   3.8140055881642313`*^9, 3.814081468356298*^9, 3.814099668489266*^9},
 CellLabel->
  "Out[101]=",ExpressionUUID->"08f13fe8-fc94-4949-8b91-07b1c68142e7"],

Cell[BoxData[
 RowBox[{
  RowBox[{"3", "+", 
   RowBox[{"h", " ", "sf"}]}], "<", 
  RowBox[{"4", " ", "sf"}]}]], "Output",
 CellChangeTimes->{{3.781437175151822*^9, 3.781437202530854*^9}, {
   3.781437295782474*^9, 3.781437352236356*^9}, 3.7814374035892963`*^9, 
   3.781437613171399*^9, 3.781439043535357*^9, {3.781439104876899*^9, 
   3.7814391344053497`*^9}, {3.7814394535688467`*^9, 3.7814394581207047`*^9}, 
   3.781457322620882*^9, 3.78188486163474*^9, 3.781952822497775*^9, 
   3.781952898261024*^9, {3.781953896487258*^9, 3.781953916168281*^9}, 
   3.782130655037745*^9, 3.7827471589942493`*^9, 3.783078894796163*^9, 
   3.785162687556292*^9, 3.785164975528594*^9, 3.7897326510301743`*^9, 
   3.795174658845468*^9, 3.799320671848111*^9, 3.7993401455539513`*^9, 
   3.7997564277832108`*^9, 3.800007657980575*^9, 3.800268186111622*^9, 
   3.8140055881642313`*^9, 3.814081468356298*^9, 3.814099668517885*^9},
 CellLabel->
  "Out[102]=",ExpressionUUID->"0491f61e-d6aa-47e7-8695-a8c12ecb7d73"],

Cell[BoxData["False"], "Output",
 CellChangeTimes->{{3.781437175151822*^9, 3.781437202530854*^9}, {
   3.781437295782474*^9, 3.781437352236356*^9}, 3.7814374035892963`*^9, 
   3.781437613171399*^9, 3.781439043535357*^9, {3.781439104876899*^9, 
   3.7814391344053497`*^9}, {3.7814394535688467`*^9, 3.7814394581207047`*^9}, 
   3.781457322620882*^9, 3.78188486163474*^9, 3.781952822497775*^9, 
   3.781952898261024*^9, {3.781953896487258*^9, 3.781953916168281*^9}, 
   3.782130655037745*^9, 3.7827471589942493`*^9, 3.783078894796163*^9, 
   3.785162687556292*^9, 3.785164975528594*^9, 3.7897326510301743`*^9, 
   3.795174658845468*^9, 3.799320671848111*^9, 3.7993401455539513`*^9, 
   3.7997564277832108`*^9, 3.800007657980575*^9, 3.800268186111622*^9, 
   3.8140055881642313`*^9, 3.814081468356298*^9, 3.814099668526672*^9},
 CellLabel->
  "Out[103]=",ExpressionUUID->"17d418be-521a-458c-83b1-1307031034d4"]
}, Open  ]],

Cell[TextData[{
 "Region r3 is independent of drive and relates only to female heterozygote \
and drive homozygote fitness (it can be rewritten as ",
 Cell[BoxData[
  FormBox[
   RowBox[{
    RowBox[{"1", "-", "s_f"}], " ", "<", " ", 
    RowBox[{
     RowBox[{"(", 
      RowBox[{"1", "-", "hs_f"}], ")"}], "/", "4"}]}], TraditionalForm]],
  ExpressionUUID->"efde28cd-c318-4a6f-96cc-f4832f137378"],
 ").\n\nThe region defined by r1 is included in all the other regions, so we \
need only satisfy this inequality"
}], "Text",
 CellChangeTimes->{{3.814005601784459*^9, 
  3.8140056112211018`*^9}},ExpressionUUID->"063735c5-7e61-4dca-a210-\
f8d99284cccd"],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"Reduce", "[", 
  RowBox[{"{", 
   RowBox[{
    RowBox[{"!", "r1"}], ",", "r2", ",", "r3", ",", 
    RowBox[{"1", ">", " ", "\[Delta]", ">", "0"}], ",", 
    RowBox[{"1", ">", "h", ">", "0"}], ",", 
    RowBox[{"1", ">", "sm", ">", "0"}], ",", 
    RowBox[{"0", "<", "sf", "<", "1"}]}], "}"}], "]"}]], "Input",
 CellChangeTimes->{{3.781438491275654*^9, 3.7814385705794687`*^9}, {
  3.781438609675919*^9, 3.781438609925259*^9}, {3.781439361820053*^9, 
  3.781439383480029*^9}, {3.7818742029198303`*^9, 3.7818742196134233`*^9}},
 CellLabel->
  "In[104]:=",ExpressionUUID->"380ac5ed-6f95-416e-8851-3243faf4b72f"],

Cell[BoxData["False"], "Output",
 CellChangeTimes->{
  3.7814385244255323`*^9, {3.78143855490099*^9, 3.781438571329862*^9}, 
   3.781438610236524*^9, {3.781439344792201*^9, 3.781439389623969*^9}, 
   3.7814394716890707`*^9, 3.781440562046509*^9, 3.781457323045916*^9, {
   3.781874206107297*^9, 3.781874221613144*^9}, 3.781884862070531*^9, 
   3.7819528229217377`*^9, 3.781952898586425*^9, {3.7819538968695173`*^9, 
   3.781953916535781*^9}, 3.782130655424851*^9, 3.7827471595772257`*^9, 
   3.783078895331085*^9, 3.7851626882942743`*^9, 3.785164976226273*^9, 
   3.7897326518393*^9, 3.79517465934138*^9, 3.799320672547058*^9, 
   3.799340146185979*^9, 3.799756428376923*^9, 3.8000076587001953`*^9, 
   3.800268186632812*^9, 3.814005589787862*^9, 3.8140814700973053`*^9, 
   3.814099678882833*^9},
 CellLabel->
  "Out[104]=",ExpressionUUID->"e2302f22-85eb-49d9-9337-97f7397c4158"]
}, Open  ]],

Cell["Thus, the instability conditions are:", "Text",
 CellChangeTimes->{{3.814005633840632*^9, 
  3.8140056425480337`*^9}},ExpressionUUID->"f7e9882a-0ffb-4a86-bfbf-\
307fc7c9d3f8"],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"unstabConditions1", " ", "=", " ", "r1"}]], "Input",
 CellChangeTimes->{{3.781440566940112*^9, 3.7814405727307663`*^9}},
 CellLabel->
  "In[105]:=",ExpressionUUID->"8f1858c4-46be-405e-83ee-273c6200485e"],

Cell[BoxData[
 RowBox[{
  RowBox[{"\[Delta]", "+", 
   RowBox[{"2", " ", "sf", " ", 
    RowBox[{"(", 
     RowBox[{
      RowBox[{"-", "1"}], "+", "sm"}], ")"}], " ", 
    RowBox[{"(", 
     RowBox[{"1", "+", "\[Delta]"}], ")"}]}], "+", 
   RowBox[{"h", " ", "sf", " ", 
    RowBox[{"(", 
     RowBox[{"2", "+", "\[Delta]", "-", 
      RowBox[{"sm", " ", 
       RowBox[{"(", 
        RowBox[{"1", "+", "\[Delta]"}], ")"}]}]}], ")"}]}]}], "<", 
  RowBox[{"sm", " ", 
   RowBox[{"(", 
    RowBox[{"1", "+", "\[Delta]"}], ")"}]}]}]], "Output",
 CellChangeTimes->{
  3.781440573583838*^9, 3.7814573230570507`*^9, 3.781884862079309*^9, 
   3.781952822929488*^9, 3.7819528985964413`*^9, {3.7819538968833323`*^9, 
   3.78195391654261*^9}, 3.7821306554345818`*^9, 3.782747159587644*^9, 
   3.78307889534017*^9, 3.785162688501986*^9, 3.78516497629982*^9, 
   3.7897326519250937`*^9, 3.7951746594067907`*^9, 3.799320672638276*^9, 
   3.799340146500635*^9, 3.799756428598815*^9, 3.8000076589437943`*^9, 
   3.800268186690694*^9, 3.814005590612234*^9, 3.8140814712053947`*^9, 
   3.814099680192692*^9},
 CellLabel->
  "Out[105]=",ExpressionUUID->"f94689dc-dfbb-4116-83da-c274977a8eba"]
}, Open  ]],

Cell["which can be re-written using the w fitness notation", "Text",
 CellChangeTimes->{{3.78187228235487*^9, 3.781872375904385*^9}, 
   3.814005627241395*^9},ExpressionUUID->"4762f435-6f50-45e5-8df7-\
3111b78f1d1c"],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"FullSimplify", "[", 
  RowBox[{"unstabConditions1", ",", 
   RowBox[{"{", 
    RowBox[{
     RowBox[{"wXdY", "\[Equal]", 
      RowBox[{"1", "-", "sm"}]}], ",", 
     RowBox[{"wXdX", " ", "\[Equal]", " ", 
      RowBox[{"1", " ", "-", " ", 
       RowBox[{"h", " ", "sf"}]}]}], ",", 
     RowBox[{"wXdXd", " ", "\[Equal]", " ", 
      RowBox[{"1", " ", "-", " ", "sf"}]}]}], "}"}]}], "]"}]], "Input",
 CellChangeTimes->{{3.783082107345252*^9, 3.7830821633262053`*^9}},
 CellLabel->
  "In[106]:=",ExpressionUUID->"10d509a4-52c8-40b6-ab2f-6a3b902947f5"],

Cell[BoxData[
 RowBox[{
  RowBox[{"2", " ", "wXdXd", " ", "wXdY", " ", 
   RowBox[{"(", 
    RowBox[{"1", "+", "\[Delta]"}], ")"}]}], "<", 
  RowBox[{"wXdX", " ", 
   RowBox[{"(", 
    RowBox[{"1", "+", "wXdY", "+", 
     RowBox[{"wXdY", " ", "\[Delta]"}]}], ")"}]}]}]], "Output",
 CellChangeTimes->{{3.783082123089921*^9, 3.7830821639965973`*^9}, 
   3.785162688703109*^9, 3.785164976483889*^9, 3.789732652155534*^9, 
   3.795174659581051*^9, 3.799320672863186*^9, 3.799340146798864*^9, 
   3.79975642881137*^9, 3.800007659149609*^9, 3.800268186858501*^9, 
   3.814081472383049*^9, 3.814099681805407*^9},
 CellLabel->
  "Out[106]=",ExpressionUUID->"35c31739-2f2c-4711-a1c5-c2fb49afbc1a"]
}, Open  ]],

Cell["This is equivalent to \[CurlyPhi], as written in the text. ", "Text",
 CellChangeTimes->{{3.8140997893825817`*^9, 
  3.814099802297583*^9}},ExpressionUUID->"d4e79e97-27aa-485d-b0b1-\
e42c374b7871"],

Cell[CellGroupData[{

Cell[BoxData[{
 RowBox[{
  RowBox[{"\[CurlyPhi]", "=", " ", 
   RowBox[{
    RowBox[{"sm", " ", 
     RowBox[{"(", 
      RowBox[{"1", "+", "\[Delta]"}], ")"}]}], "-", 
    RowBox[{"(", 
     RowBox[{"\[Delta]", "+", 
      RowBox[{"2", " ", "sf", " ", 
       RowBox[{"(", 
        RowBox[{
         RowBox[{"-", "1"}], "+", "sm"}], ")"}], " ", 
       RowBox[{"(", 
        RowBox[{"1", "+", "\[Delta]"}], ")"}]}], "+", 
      RowBox[{"h", " ", "sf", " ", 
       RowBox[{"(", 
        RowBox[{"2", "+", "\[Delta]", "-", 
         RowBox[{"sm", " ", 
          RowBox[{"(", 
           RowBox[{"1", "+", "\[Delta]"}], ")"}]}]}], ")"}]}]}], ")"}]}]}], 
  ";"}], "\[IndentingNewLine]", 
 RowBox[{
  RowBox[{"Simplify", "[", 
   RowBox[{"Reduce", "[", 
    RowBox[{"{", 
     RowBox[{"unstabConditions1", ",", 
      RowBox[{"1", ">", " ", "\[Delta]", ">", "0"}], ",", 
      RowBox[{"1", ">", "h", ">", "0"}], ",", 
      RowBox[{"1", ">", "sm", ">", "0"}], ",", 
      RowBox[{"0", "<", "sf", "<", "1"}]}], "}"}], "]"}], "]"}], 
  ";"}], "\[IndentingNewLine]", 
 RowBox[{
  RowBox[{"Simplify", "[", 
   RowBox[{"Reduce", "[", 
    RowBox[{"{", 
     RowBox[{
      RowBox[{"\[CurlyPhi]", ">", "0"}], ",", 
      RowBox[{"1", ">", " ", "\[Delta]", ">", "0"}], ",", 
      RowBox[{"1", ">", "h", ">", "0"}], ",", 
      RowBox[{"1", ">", "sm", ">", "0"}], ",", 
      RowBox[{"0", "<", "sf", "<", "1"}]}], "}"}], "]"}], "]"}], 
  ";"}], "\[IndentingNewLine]", 
 RowBox[{"%", "\[Equal]", "%%"}]}], "Input",
 CellChangeTimes->{{3.814099721522447*^9, 3.814099787257207*^9}, {
  3.814099863442215*^9, 3.814099902960973*^9}},
 CellLabel->
  "In[145]:=",ExpressionUUID->"8f270242-0732-4f2d-a989-7c5240aa97d7"],

Cell[BoxData["True"], "Output",
 CellChangeTimes->{{3.814099725820879*^9, 3.814099772444579*^9}, 
   3.814099813443368*^9, {3.814099865527564*^9, 3.814099903366467*^9}},
 CellLabel->
  "Out[148]=",ExpressionUUID->"089eaa0c-dab0-4661-a53e-1bb4d9ec831b"]
}, Open  ]]
}, Closed]]
}, Closed]],

Cell[CellGroupData[{

Cell["Critical birth rate", "Subsection",
 CellChangeTimes->{{3.7707171378648453`*^9, 
  3.7707171424036207`*^9}},ExpressionUUID->"9a6be050-7bae-4fba-872a-\
c3b23864a3d4"],

Cell["\<\
The critical birth rate is the lowest birth rate for which a population still \
exists (equilibrium population size, N, bigger than 0). 

We first find the critical birth rate when drive is absent, polymorphic, or \
fixed. \
\>", "Text",
 CellChangeTimes->{{3.813930468388834*^9, 3.8139304720109*^9}, {
  3.8139306445371513`*^9, 
  3.813930713094428*^9}},ExpressionUUID->"898e4dc1-f2fd-4899-993f-\
9f68889ab0c3"],

Cell[CellGroupData[{

Cell[BoxData[{
 RowBox[{"bcritNoDrive", "=", 
  RowBox[{"b", "/.", 
   RowBox[{
    RowBox[{"Solve", "[", 
     RowBox[{
      RowBox[{
       RowBox[{"Simplify", "[", 
        RowBox[{
         RowBox[{"(", 
          RowBox[{"maleNum", " ", "+", " ", "femaleNum"}], ")"}], "/.", 
         "noDrivePopSizeSol"}], "]"}], "\[Equal]", "0"}], ",", "b"}], "]"}], 
    "[", 
    RowBox[{"[", "1", "]"}], "]"}]}]}], "\[IndentingNewLine]", 
 RowBox[{"bcritEquil", "=", 
  RowBox[{"Simplify", "[", 
   RowBox[{"b", "/.", 
    RowBox[{
     RowBox[{"Solve", "[", 
      RowBox[{
       RowBox[{"Neq", "\[Equal]", "0"}], ",", "b"}], "]"}], "[", 
     RowBox[{"[", "1", "]"}], "]"}]}], "]"}]}], "\[IndentingNewLine]", 
 RowBox[{"bcritFixed", "=", 
  RowBox[{"b", "/.", 
   RowBox[{
    RowBox[{"Solve", "[", 
     RowBox[{
      RowBox[{"Nfixed", "\[Equal]", "0"}], ",", "b"}], "]"}], "[", 
    RowBox[{"[", "1", "]"}], "]"}]}]}]}], "Input",
 CellChangeTimes->{{3.813928398062742*^9, 3.813928561860159*^9}, {
  3.814099574852695*^9, 3.814099577381032*^9}},
 CellLabel->"In[68]:=",ExpressionUUID->"d19ef520-56e2-44f9-b45d-912a2d6aad5f"],

Cell[BoxData["2"], "Output",
 CellChangeTimes->{
  3.813928401523972*^9, {3.813928434003956*^9, 3.813928529425839*^9}, 
   3.813928562308132*^9, 3.814081785262821*^9, {3.8140995676921873`*^9, 
   3.8140995777549677`*^9}},
 CellLabel->"Out[68]=",ExpressionUUID->"eda60f85-e96a-4628-be8d-56276c90e672"],

Cell[BoxData[
 FractionBox[
  RowBox[{
   RowBox[{"4", " ", "sm", " ", 
    RowBox[{"(", 
     RowBox[{"sm", "-", "\[Delta]", "+", 
      RowBox[{"sm", " ", "\[Delta]"}]}], ")"}]}], "-", 
   RowBox[{"4", " ", "sf", " ", 
    RowBox[{"(", 
     RowBox[{
      RowBox[{"2", " ", 
       RowBox[{"(", 
        RowBox[{
         RowBox[{"-", "1"}], "+", "sm"}], ")"}], " ", 
       RowBox[{"(", 
        RowBox[{"1", "+", "\[Delta]"}], ")"}]}], "+", 
      RowBox[{"h", " ", 
       RowBox[{"(", 
        RowBox[{
         RowBox[{"-", "2"}], "+", "sm"}], ")"}], " ", 
       RowBox[{"(", 
        RowBox[{
         RowBox[{"-", "2"}], "+", "sm", "-", "\[Delta]", "+", 
         RowBox[{"sm", " ", "\[Delta]"}]}], ")"}]}]}], ")"}]}]}], 
  RowBox[{
   RowBox[{
    SuperscriptBox["h", "2"], " ", 
    SuperscriptBox["sf", "2"], " ", 
    SuperscriptBox[
     RowBox[{"(", 
      RowBox[{
       RowBox[{"-", "2"}], "+", "sm", "-", "\[Delta]", "+", 
       RowBox[{"sm", " ", "\[Delta]"}]}], ")"}], "2"]}], "+", 
   SuperscriptBox[
    RowBox[{"(", 
     RowBox[{"sm", "-", "\[Delta]", "+", 
      RowBox[{"sm", " ", "\[Delta]"}]}], ")"}], "2"], "-", 
   RowBox[{"2", " ", "sf", " ", 
    RowBox[{"(", 
     RowBox[{
      RowBox[{"2", " ", 
       RowBox[{"(", 
        RowBox[{
         RowBox[{"-", "1"}], "+", "sm"}], ")"}], " ", 
       RowBox[{"(", 
        RowBox[{"1", "+", "\[Delta]"}], ")"}]}], "+", 
      RowBox[{"h", " ", 
       SuperscriptBox[
        RowBox[{"(", 
         RowBox[{
          RowBox[{"-", "2"}], "+", "sm", "-", "\[Delta]", "+", 
          RowBox[{"sm", " ", "\[Delta]"}]}], ")"}], "2"]}]}], 
     ")"}]}]}]]], "Output",
 CellChangeTimes->{
  3.813928401523972*^9, {3.813928434003956*^9, 3.813928529425839*^9}, 
   3.813928562308132*^9, 3.814081785262821*^9, {3.8140995676921873`*^9, 
   3.814099578038756*^9}},
 CellLabel->"Out[69]=",ExpressionUUID->"01e09ee8-7e1b-43e6-bc8b-47511622bb21"],

Cell[BoxData[
 RowBox[{"-", 
  FractionBox["2", 
   RowBox[{
    RowBox[{"(", 
     RowBox[{
      RowBox[{"-", "1"}], "+", "sf"}], ")"}], " ", 
    RowBox[{"(", 
     RowBox[{"1", "+", "\[Delta]"}], ")"}]}]]}]], "Output",
 CellChangeTimes->{
  3.813928401523972*^9, {3.813928434003956*^9, 3.813928529425839*^9}, 
   3.813928562308132*^9, 3.814081785262821*^9, {3.8140995676921873`*^9, 
   3.814099578069371*^9}},
 CellLabel->"Out[70]=",ExpressionUUID->"10b4c6cc-e760-40f5-ac20-e8c908e82d5d"]
}, Open  ]],

Cell["\<\
Given that drive is polymorphic (\[Phi]>0,\[CurlyPhi]>0), the critical birth \
rate is always smaller than it would have been without drive
Given that drive reaches fixation (\[Phi]>0,\[CurlyPhi]<0), the critical \
birth rate is always smaller than it would have been without drive\
\>", "Text",
 CellChangeTimes->{{3.813930720878367*^9, 
  3.813930768858816*^9}},ExpressionUUID->"9d14e7a2-85f4-43a8-8581-\
76615094fac3"],

Cell[CellGroupData[{

Cell[BoxData[{
 RowBox[{"Reduce", "[", 
  RowBox[{"{", 
   RowBox[{
    RowBox[{"bcritNoDrive", "<", "bcritEquil"}], ",", " ", 
    RowBox[{"0", "<", "sf", "<", "1"}], ",", 
    RowBox[{"0", "<", "h", "<", "1"}], ",", 
    RowBox[{"0", "<", "sm", "<", "1"}], ",", 
    RowBox[{"0", "<", "b"}], ",", " ", 
    RowBox[{"0", "<", "\[Delta]", "<", "1"}], ",", " ", 
    RowBox[{"\[Phi]", ">", "0"}], ",", 
    RowBox[{"\[CurlyPhi]", ">", "0"}]}], "}"}], 
  "]"}], "\[IndentingNewLine]", 
 RowBox[{"Reduce", "[", 
  RowBox[{"{", 
   RowBox[{
    RowBox[{"bcritNoDrive", "<", "bcritFixed"}], ",", " ", 
    RowBox[{"0", "<", "sf", "<", "1"}], ",", 
    RowBox[{"0", "<", "h", "<", "1"}], ",", 
    RowBox[{"0", "<", "sm", "<", "1"}], ",", 
    RowBox[{"0", "<", "b"}], ",", " ", 
    RowBox[{"0", "<", "\[Delta]", "<", "1"}], ",", " ", 
    RowBox[{"\[Phi]", ">", "0"}], ",", 
    RowBox[{"\[CurlyPhi]", "<", "0"}]}], "}"}], "]"}]}], "Input",
 CellChangeTimes->{{3.81392856865447*^9, 3.81392863122628*^9}, 
   3.813928679185483*^9, {3.81393002322504*^9, 3.813930051570941*^9}, {
   3.813930082291155*^9, 3.813930177188037*^9}},
 CellLabel->
  "In[153]:=",ExpressionUUID->"23732140-3772-4d4d-91fd-476a4d7c84c1"],

Cell[BoxData["False"], "Output",
 CellChangeTimes->{{3.8139286049658337`*^9, 3.813928627951355*^9}, 
   3.813928684723576*^9, {3.8139300258073263`*^9, 3.8139300537471743`*^9}, {
   3.81393010421644*^9, 3.813930128247485*^9}, {3.813930158987651*^9, 
   3.8139301784794807`*^9}, 3.814081786639469*^9, 3.814099586418022*^9, 
   3.8140999241267147`*^9},
 CellLabel->
  "Out[153]=",ExpressionUUID->"024359bc-dfdb-470c-84ae-17b8c0f8d554"],

Cell[BoxData["False"], "Output",
 CellChangeTimes->{{3.8139286049658337`*^9, 3.813928627951355*^9}, 
   3.813928684723576*^9, {3.8139300258073263`*^9, 3.8139300537471743`*^9}, {
   3.81393010421644*^9, 3.813930128247485*^9}, {3.813930158987651*^9, 
   3.8139301784794807`*^9}, 3.814081786639469*^9, 3.814099586418022*^9, 
   3.814099924154397*^9},
 CellLabel->
  "Out[154]=",ExpressionUUID->"ce2db405-f0a4-4801-90f9-7ae419084011"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[{
 RowBox[{
  RowBox[{"funcEq", "[", 
   RowBox[{
   "h_", ",", "sf_", ",", "\[Alpha]_", ",", "\[Delta]_", ",", "c_", ",", 
    "sm_"}], "]"}], ":=", 
  RowBox[{"If", "[", 
   RowBox[{
    RowBox[{
     RowBox[{
      RowBox[{"(", 
       RowBox[{
        RowBox[{"(", 
         RowBox[{
          RowBox[{"b", " ", 
           SuperscriptBox["h", "2"], " ", 
           SuperscriptBox["sf", "2"], " ", 
           SuperscriptBox[
            RowBox[{"(", 
             RowBox[{
              RowBox[{"-", "2"}], "+", "sm", "-", "\[Delta]", "+", 
              RowBox[{"sm", " ", "\[Delta]"}]}], ")"}], "2"]}], "+", 
          RowBox[{
           RowBox[{"(", 
            RowBox[{"sm", "-", "\[Delta]", "+", 
             RowBox[{"sm", " ", "\[Delta]"}]}], ")"}], " ", 
           RowBox[{"(", 
            RowBox[{
             RowBox[{
              RowBox[{"-", "b"}], " ", "\[Delta]"}], "+", 
             RowBox[{"sm", " ", 
              RowBox[{"(", 
               RowBox[{
                RowBox[{"-", "4"}], "+", "b", "+", 
                RowBox[{"b", " ", "\[Delta]"}]}], ")"}]}]}], ")"}]}], "-", 
          RowBox[{"2", " ", "sf", " ", 
           RowBox[{"(", 
            RowBox[{
             RowBox[{
              RowBox[{"-", "4"}], " ", 
              RowBox[{"(", 
               RowBox[{
                RowBox[{"-", "1"}], "+", "sm"}], ")"}], " ", 
              RowBox[{"(", 
               RowBox[{"1", "+", "\[Delta]"}], ")"}]}], "+", 
             RowBox[{"2", " ", "b", " ", 
              RowBox[{"(", 
               RowBox[{
                RowBox[{"-", "1"}], "+", "sm"}], ")"}], " ", 
              RowBox[{"(", 
               RowBox[{"1", "+", "\[Delta]"}], ")"}]}], "-", 
             RowBox[{"2", " ", "h", " ", 
              RowBox[{"(", 
               RowBox[{
                RowBox[{"-", "2"}], "+", "sm"}], ")"}], " ", 
              RowBox[{"(", 
               RowBox[{
                RowBox[{"-", "2"}], "+", "sm", "-", "\[Delta]", "+", 
                RowBox[{"sm", " ", "\[Delta]"}]}], ")"}]}], "+", 
             RowBox[{"b", " ", "h", " ", 
              SuperscriptBox[
               RowBox[{"(", 
                RowBox[{
                 RowBox[{"-", "2"}], "+", "sm", "-", "\[Delta]", "+", 
                 RowBox[{"sm", " ", "\[Delta]"}]}], ")"}], "2"]}]}], 
            ")"}]}]}], ")"}], " ", 
        RowBox[{"(", 
         RowBox[{
          RowBox[{
           SuperscriptBox["h", "2"], " ", 
           SuperscriptBox["sf", "2"], " ", 
           SuperscriptBox[
            RowBox[{"(", 
             RowBox[{
              RowBox[{"-", "2"}], "+", "sm", "-", "\[Delta]", "+", 
              RowBox[{"sm", " ", "\[Delta]"}]}], ")"}], "2"]}], "+", 
          SuperscriptBox[
           RowBox[{"(", 
            RowBox[{"sm", "-", "\[Delta]", "+", 
             RowBox[{"sm", " ", "\[Delta]"}]}], ")"}], "2"], "-", 
          RowBox[{"2", " ", "sf", " ", 
           RowBox[{"(", 
            RowBox[{
             RowBox[{"2", " ", 
              RowBox[{"(", 
               RowBox[{
                RowBox[{"-", "1"}], "+", "sm"}], ")"}], " ", 
              RowBox[{"(", 
               RowBox[{"1", "+", "\[Delta]"}], ")"}]}], "+", 
             RowBox[{"h", " ", 
              SuperscriptBox[
               RowBox[{"(", 
                RowBox[{
                 RowBox[{"-", "2"}], "+", "sm", "-", "\[Delta]", "+", 
                 RowBox[{"sm", " ", "\[Delta]"}]}], ")"}], "2"]}]}], ")"}]}], 
          "-", 
          RowBox[{
           FractionBox["1", 
            RowBox[{"sf", " ", 
             RowBox[{"(", 
              RowBox[{
               RowBox[{
                RowBox[{"-", 
                 RowBox[{"(", 
                  RowBox[{
                   RowBox[{"-", "1"}], "+", "sm"}], ")"}]}], " ", 
                RowBox[{"(", 
                 RowBox[{"1", "+", "\[Delta]"}], ")"}]}], "+", 
               RowBox[{"h", " ", 
                RowBox[{"(", 
                 RowBox[{
                  RowBox[{"-", "2"}], "+", "sm", "-", "\[Delta]", "+", 
                  RowBox[{"sm", " ", "\[Delta]"}]}], ")"}]}]}], ")"}]}]], 
           RowBox[{"(", 
            RowBox[{
             RowBox[{"sm", " ", 
              RowBox[{"(", 
               RowBox[{
                RowBox[{"sm", " ", 
                 RowBox[{"(", 
                  RowBox[{
                   RowBox[{"-", "1"}], "+", "\[Delta]"}], ")"}]}], "-", 
                "\[Delta]"}], ")"}], " ", 
              SuperscriptBox[
               RowBox[{"(", 
                RowBox[{"sm", "-", "\[Delta]", "+", 
                 RowBox[{"sm", " ", "\[Delta]"}]}], ")"}], "2"]}], "+", 
             RowBox[{
              SuperscriptBox["sf", "2"], " ", 
              RowBox[{"(", 
               RowBox[{
                RowBox[{
                 RowBox[{"-", "4"}], " ", 
                 SuperscriptBox[
                  RowBox[{"(", 
                   RowBox[{
                    RowBox[{"-", "1"}], "+", "sm"}], ")"}], "2"], " ", 
                 SuperscriptBox[
                  RowBox[{"(", 
                   RowBox[{"1", "+", "\[Delta]"}], ")"}], "2"]}], "+", 
                RowBox[{
                 SuperscriptBox["h", "2"], " ", 
                 RowBox[{"(", 
                  RowBox[{
                   RowBox[{"-", "2"}], "+", "sm"}], ")"}], " ", 
                 RowBox[{"(", 
                  RowBox[{"2", "+", 
                   RowBox[{"sm", " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{"-", "1"}], "+", "\[Delta]"}], ")"}]}], "-", 
                   "\[Delta]"}], ")"}], " ", 
                 SuperscriptBox[
                  RowBox[{"(", 
                   RowBox[{
                    RowBox[{"-", "2"}], "+", "sm", "-", "\[Delta]", "+", 
                    RowBox[{"sm", " ", "\[Delta]"}]}], ")"}], "2"]}], "+", 
                RowBox[{"2", " ", "h", " ", 
                 RowBox[{"(", 
                  RowBox[{
                   RowBox[{"-", "1"}], "+", "sm"}], ")"}], " ", 
                 RowBox[{"(", 
                  RowBox[{"1", "+", "\[Delta]"}], ")"}], " ", 
                 RowBox[{"(", 
                  RowBox[{
                   RowBox[{"-", "8"}], "-", 
                   RowBox[{"2", " ", "\[Delta]"}], "+", 
                   SuperscriptBox["\[Delta]", "2"], "+", 
                   RowBox[{"sm", " ", 
                    RowBox[{"(", 
                    RowBox[{"8", "+", 
                    RowBox[{"3", " ", "\[Delta]"}], "-", 
                    RowBox[{"2", " ", 
                    SuperscriptBox["\[Delta]", "2"]}]}], ")"}]}], "+", 
                   RowBox[{
                    SuperscriptBox["sm", "2"], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{"-", "2"}], "-", "\[Delta]", "+", 
                    SuperscriptBox["\[Delta]", "2"]}], ")"}]}]}], ")"}]}]}], 
               ")"}]}], "-", 
             RowBox[{"2", " ", "sf", " ", 
              RowBox[{"(", 
               RowBox[{"sm", "-", "\[Delta]", "+", 
                RowBox[{"sm", " ", "\[Delta]"}]}], ")"}], " ", 
              RowBox[{"(", 
               RowBox[{
                RowBox[{"h", " ", 
                 SuperscriptBox["sm", "3"], " ", 
                 RowBox[{"(", 
                  RowBox[{
                   RowBox[{"-", "1"}], "+", 
                   SuperscriptBox["\[Delta]", "2"]}], ")"}]}], "+", 
                RowBox[{"\[Delta]", " ", 
                 RowBox[{"(", 
                  RowBox[{"1", "+", "\[Delta]", "-", 
                   RowBox[{"h", " ", 
                    RowBox[{"(", 
                    RowBox[{"2", "+", "\[Delta]"}], ")"}]}]}], ")"}]}], "-", 
                RowBox[{
                 SuperscriptBox["sm", "2"], " ", 
                 RowBox[{"(", 
                  RowBox[{"2", "+", "\[Delta]", "-", 
                   SuperscriptBox["\[Delta]", "2"], "+", 
                   RowBox[{"h", " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{"-", "4"}], "+", "\[Delta]", "+", 
                    RowBox[{"3", " ", 
                    SuperscriptBox["\[Delta]", "2"]}]}], ")"}]}]}], ")"}]}], 
                "+", 
                RowBox[{"sm", " ", 
                 RowBox[{"(", 
                  RowBox[{"2", "-", 
                   RowBox[{"2", " ", 
                    SuperscriptBox["\[Delta]", "2"]}], "+", 
                   RowBox[{"h", " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{"-", "4"}], "+", 
                    RowBox[{"3", " ", "\[Delta]"}], "+", 
                    RowBox[{"3", " ", 
                    SuperscriptBox["\[Delta]", "2"]}]}], ")"}]}]}], ")"}]}]}],
                ")"}]}]}], ")"}]}]}], ")"}]}], ")"}], "/", 
      RowBox[{"(", 
       RowBox[{
        SuperscriptBox["b", "2"], " ", "\[Alpha]", " ", 
        SuperscriptBox[
         RowBox[{"(", 
          RowBox[{
           RowBox[{
            SuperscriptBox["h", "2"], " ", 
            SuperscriptBox["sf", "2"], " ", 
            SuperscriptBox[
             RowBox[{"(", 
              RowBox[{
               RowBox[{"-", "2"}], "+", "sm", "-", "\[Delta]", "+", 
               RowBox[{"sm", " ", "\[Delta]"}]}], ")"}], "2"]}], "+", 
           SuperscriptBox[
            RowBox[{"(", 
             RowBox[{"sm", "-", "\[Delta]", "+", 
              RowBox[{"sm", " ", "\[Delta]"}]}], ")"}], "2"], "-", 
           RowBox[{"2", " ", "sf", " ", 
            RowBox[{"(", 
             RowBox[{
              RowBox[{"2", " ", 
               RowBox[{"(", 
                RowBox[{
                 RowBox[{"-", "1"}], "+", "sm"}], ")"}], " ", 
               RowBox[{"(", 
                RowBox[{"1", "+", "\[Delta]"}], ")"}]}], "+", 
              RowBox[{"h", " ", 
               SuperscriptBox[
                RowBox[{"(", 
                 RowBox[{
                  RowBox[{"-", "2"}], "+", "sm", "-", "\[Delta]", "+", 
                  RowBox[{"sm", " ", "\[Delta]"}]}], ")"}], "2"]}]}], 
             ")"}]}]}], ")"}], "2"]}], ")"}]}], ">", "0"}], ",", 
    RowBox[{
     RowBox[{"(", 
      RowBox[{
       RowBox[{"(", 
        RowBox[{
         RowBox[{"b", " ", 
          SuperscriptBox["h", "2"], " ", 
          SuperscriptBox["sf", "2"], " ", 
          SuperscriptBox[
           RowBox[{"(", 
            RowBox[{
             RowBox[{"-", "2"}], "+", "sm", "-", "\[Delta]", "+", 
             RowBox[{"sm", " ", "\[Delta]"}]}], ")"}], "2"]}], "+", 
         RowBox[{
          RowBox[{"(", 
           RowBox[{"sm", "-", "\[Delta]", "+", 
            RowBox[{"sm", " ", "\[Delta]"}]}], ")"}], " ", 
          RowBox[{"(", 
           RowBox[{
            RowBox[{
             RowBox[{"-", "b"}], " ", "\[Delta]"}], "+", 
            RowBox[{"sm", " ", 
             RowBox[{"(", 
              RowBox[{
               RowBox[{"-", "4"}], "+", "b", "+", 
               RowBox[{"b", " ", "\[Delta]"}]}], ")"}]}]}], ")"}]}], "-", 
         RowBox[{"2", " ", "sf", " ", 
          RowBox[{"(", 
           RowBox[{
            RowBox[{
             RowBox[{"-", "4"}], " ", 
             RowBox[{"(", 
              RowBox[{
               RowBox[{"-", "1"}], "+", "sm"}], ")"}], " ", 
             RowBox[{"(", 
              RowBox[{"1", "+", "\[Delta]"}], ")"}]}], "+", 
            RowBox[{"2", " ", "b", " ", 
             RowBox[{"(", 
              RowBox[{
               RowBox[{"-", "1"}], "+", "sm"}], ")"}], " ", 
             RowBox[{"(", 
              RowBox[{"1", "+", "\[Delta]"}], ")"}]}], "-", 
            RowBox[{"2", " ", "h", " ", 
             RowBox[{"(", 
              RowBox[{
               RowBox[{"-", "2"}], "+", "sm"}], ")"}], " ", 
             RowBox[{"(", 
              RowBox[{
               RowBox[{"-", "2"}], "+", "sm", "-", "\[Delta]", "+", 
               RowBox[{"sm", " ", "\[Delta]"}]}], ")"}]}], "+", 
            RowBox[{"b", " ", "h", " ", 
             SuperscriptBox[
              RowBox[{"(", 
               RowBox[{
                RowBox[{"-", "2"}], "+", "sm", "-", "\[Delta]", "+", 
                RowBox[{"sm", " ", "\[Delta]"}]}], ")"}], "2"]}]}], ")"}]}]}],
         ")"}], " ", 
       RowBox[{"(", 
        RowBox[{
         RowBox[{
          SuperscriptBox["h", "2"], " ", 
          SuperscriptBox["sf", "2"], " ", 
          SuperscriptBox[
           RowBox[{"(", 
            RowBox[{
             RowBox[{"-", "2"}], "+", "sm", "-", "\[Delta]", "+", 
             RowBox[{"sm", " ", "\[Delta]"}]}], ")"}], "2"]}], "+", 
         SuperscriptBox[
          RowBox[{"(", 
           RowBox[{"sm", "-", "\[Delta]", "+", 
            RowBox[{"sm", " ", "\[Delta]"}]}], ")"}], "2"], "-", 
         RowBox[{"2", " ", "sf", " ", 
          RowBox[{"(", 
           RowBox[{
            RowBox[{"2", " ", 
             RowBox[{"(", 
              RowBox[{
               RowBox[{"-", "1"}], "+", "sm"}], ")"}], " ", 
             RowBox[{"(", 
              RowBox[{"1", "+", "\[Delta]"}], ")"}]}], "+", 
            RowBox[{"h", " ", 
             SuperscriptBox[
              RowBox[{"(", 
               RowBox[{
                RowBox[{"-", "2"}], "+", "sm", "-", "\[Delta]", "+", 
                RowBox[{"sm", " ", "\[Delta]"}]}], ")"}], "2"]}]}], ")"}]}], 
         "-", 
         RowBox[{
          FractionBox["1", 
           RowBox[{"sf", " ", 
            RowBox[{"(", 
             RowBox[{
              RowBox[{
               RowBox[{"-", 
                RowBox[{"(", 
                 RowBox[{
                  RowBox[{"-", "1"}], "+", "sm"}], ")"}]}], " ", 
               RowBox[{"(", 
                RowBox[{"1", "+", "\[Delta]"}], ")"}]}], "+", 
              RowBox[{"h", " ", 
               RowBox[{"(", 
                RowBox[{
                 RowBox[{"-", "2"}], "+", "sm", "-", "\[Delta]", "+", 
                 RowBox[{"sm", " ", "\[Delta]"}]}], ")"}]}]}], ")"}]}]], 
          RowBox[{"(", 
           RowBox[{
            RowBox[{"sm", " ", 
             RowBox[{"(", 
              RowBox[{
               RowBox[{"sm", " ", 
                RowBox[{"(", 
                 RowBox[{
                  RowBox[{"-", "1"}], "+", "\[Delta]"}], ")"}]}], "-", 
               "\[Delta]"}], ")"}], " ", 
             SuperscriptBox[
              RowBox[{"(", 
               RowBox[{"sm", "-", "\[Delta]", "+", 
                RowBox[{"sm", " ", "\[Delta]"}]}], ")"}], "2"]}], "+", 
            RowBox[{
             SuperscriptBox["sf", "2"], " ", 
             RowBox[{"(", 
              RowBox[{
               RowBox[{
                RowBox[{"-", "4"}], " ", 
                SuperscriptBox[
                 RowBox[{"(", 
                  RowBox[{
                   RowBox[{"-", "1"}], "+", "sm"}], ")"}], "2"], " ", 
                SuperscriptBox[
                 RowBox[{"(", 
                  RowBox[{"1", "+", "\[Delta]"}], ")"}], "2"]}], "+", 
               RowBox[{
                SuperscriptBox["h", "2"], " ", 
                RowBox[{"(", 
                 RowBox[{
                  RowBox[{"-", "2"}], "+", "sm"}], ")"}], " ", 
                RowBox[{"(", 
                 RowBox[{"2", "+", 
                  RowBox[{"sm", " ", 
                   RowBox[{"(", 
                    RowBox[{
                    RowBox[{"-", "1"}], "+", "\[Delta]"}], ")"}]}], "-", 
                  "\[Delta]"}], ")"}], " ", 
                SuperscriptBox[
                 RowBox[{"(", 
                  RowBox[{
                   RowBox[{"-", "2"}], "+", "sm", "-", "\[Delta]", "+", 
                   RowBox[{"sm", " ", "\[Delta]"}]}], ")"}], "2"]}], "+", 
               RowBox[{"2", " ", "h", " ", 
                RowBox[{"(", 
                 RowBox[{
                  RowBox[{"-", "1"}], "+", "sm"}], ")"}], " ", 
                RowBox[{"(", 
                 RowBox[{"1", "+", "\[Delta]"}], ")"}], " ", 
                RowBox[{"(", 
                 RowBox[{
                  RowBox[{"-", "8"}], "-", 
                  RowBox[{"2", " ", "\[Delta]"}], "+", 
                  SuperscriptBox["\[Delta]", "2"], "+", 
                  RowBox[{"sm", " ", 
                   RowBox[{"(", 
                    RowBox[{"8", "+", 
                    RowBox[{"3", " ", "\[Delta]"}], "-", 
                    RowBox[{"2", " ", 
                    SuperscriptBox["\[Delta]", "2"]}]}], ")"}]}], "+", 
                  RowBox[{
                   SuperscriptBox["sm", "2"], " ", 
                   RowBox[{"(", 
                    RowBox[{
                    RowBox[{"-", "2"}], "-", "\[Delta]", "+", 
                    SuperscriptBox["\[Delta]", "2"]}], ")"}]}]}], ")"}]}]}], 
              ")"}]}], "-", 
            RowBox[{"2", " ", "sf", " ", 
             RowBox[{"(", 
              RowBox[{"sm", "-", "\[Delta]", "+", 
               RowBox[{"sm", " ", "\[Delta]"}]}], ")"}], " ", 
             RowBox[{"(", 
              RowBox[{
               RowBox[{"h", " ", 
                SuperscriptBox["sm", "3"], " ", 
                RowBox[{"(", 
                 RowBox[{
                  RowBox[{"-", "1"}], "+", 
                  SuperscriptBox["\[Delta]", "2"]}], ")"}]}], "+", 
               RowBox[{"\[Delta]", " ", 
                RowBox[{"(", 
                 RowBox[{"1", "+", "\[Delta]", "-", 
                  RowBox[{"h", " ", 
                   RowBox[{"(", 
                    RowBox[{"2", "+", "\[Delta]"}], ")"}]}]}], ")"}]}], "-", 
               RowBox[{
                SuperscriptBox["sm", "2"], " ", 
                RowBox[{"(", 
                 RowBox[{"2", "+", "\[Delta]", "-", 
                  SuperscriptBox["\[Delta]", "2"], "+", 
                  RowBox[{"h", " ", 
                   RowBox[{"(", 
                    RowBox[{
                    RowBox[{"-", "4"}], "+", "\[Delta]", "+", 
                    RowBox[{"3", " ", 
                    SuperscriptBox["\[Delta]", "2"]}]}], ")"}]}]}], ")"}]}], 
               "+", 
               RowBox[{"sm", " ", 
                RowBox[{"(", 
                 RowBox[{"2", "-", 
                  RowBox[{"2", " ", 
                   SuperscriptBox["\[Delta]", "2"]}], "+", 
                  RowBox[{"h", " ", 
                   RowBox[{"(", 
                    RowBox[{
                    RowBox[{"-", "4"}], "+", 
                    RowBox[{"3", " ", "\[Delta]"}], "+", 
                    RowBox[{"3", " ", 
                    SuperscriptBox["\[Delta]", "2"]}]}], ")"}]}]}], ")"}]}]}],
               ")"}]}]}], ")"}]}]}], ")"}]}], ")"}], "/", 
     RowBox[{"(", 
      RowBox[{
       SuperscriptBox["b", "2"], " ", "\[Alpha]", " ", 
       SuperscriptBox[
        RowBox[{"(", 
         RowBox[{
          RowBox[{
           SuperscriptBox["h", "2"], " ", 
           SuperscriptBox["sf", "2"], " ", 
           SuperscriptBox[
            RowBox[{"(", 
             RowBox[{
              RowBox[{"-", "2"}], "+", "sm", "-", "\[Delta]", "+", 
              RowBox[{"sm", " ", "\[Delta]"}]}], ")"}], "2"]}], "+", 
          SuperscriptBox[
           RowBox[{"(", 
            RowBox[{"sm", "-", "\[Delta]", "+", 
             RowBox[{"sm", " ", "\[Delta]"}]}], ")"}], "2"], "-", 
          RowBox[{"2", " ", "sf", " ", 
           RowBox[{"(", 
            RowBox[{
             RowBox[{"2", " ", 
              RowBox[{"(", 
               RowBox[{
                RowBox[{"-", "1"}], "+", "sm"}], ")"}], " ", 
              RowBox[{"(", 
               RowBox[{"1", "+", "\[Delta]"}], ")"}]}], "+", 
             RowBox[{"h", " ", 
              SuperscriptBox[
               RowBox[{"(", 
                RowBox[{
                 RowBox[{"-", "2"}], "+", "sm", "-", "\[Delta]", "+", 
                 RowBox[{"sm", " ", "\[Delta]"}]}], ")"}], "2"]}]}], 
            ")"}]}]}], ")"}], "2"]}], ")"}]}], ",", "Null"}], 
   "]"}]}], "\[IndentingNewLine]", 
 RowBox[{
  RowBox[{
   RowBox[{"funcNoDrive", "[", 
    RowBox[{
    "h_", ",", "sf_", ",", "\[Alpha]_", ",", "\[Delta]_", ",", "c_", ",", 
     "sm_"}], "]"}], ":=", 
   RowBox[{"If", "[", 
    RowBox[{
     RowBox[{
      FractionBox[
       RowBox[{"2", " ", 
        RowBox[{"(", 
         RowBox[{
          RowBox[{"-", "2"}], "+", "b"}], ")"}]}], 
       RowBox[{
        SuperscriptBox["b", "2"], " ", "\[Alpha]"}]], ">", "0"}], ",", 
     FractionBox[
      RowBox[{"2", " ", 
       RowBox[{"(", 
        RowBox[{
         RowBox[{"-", "2"}], "+", "b"}], ")"}]}], 
      RowBox[{
       SuperscriptBox["b", "2"], " ", "\[Alpha]"}]], ",", "Null"}], "]"}]}], 
  "\[IndentingNewLine]"}], "\[IndentingNewLine]", 
 RowBox[{
  RowBox[{"testParameters4", "=", 
   RowBox[{"{", 
    RowBox[{
     RowBox[{"h", "\[Rule]", "0.1"}], ",", " ", 
     RowBox[{"sf", "\[Rule]", "0.8"}], ",", 
     RowBox[{"\[Alpha]", "\[Rule]", 
      RowBox[{"10", "^", 
       RowBox[{"(", 
        RowBox[{"-", "3"}], ")"}]}]}], ",", 
     RowBox[{"\[Delta]", "\[Rule]", "1"}], ",", 
     RowBox[{"c", "\[Rule]", "1"}], ",", 
     RowBox[{"sm", "\[Rule]", "0"}]}], "}"}]}], ";"}], "\n", 
 RowBox[{"Legended", "[", 
  RowBox[{
   RowBox[{"Labeled", "[", 
    RowBox[{
     RowBox[{"Plot", "[", 
      RowBox[{
       RowBox[{"{", 
        RowBox[{
         RowBox[{
          RowBox[{"funcEq", "[", 
           RowBox[{
           "h", ",", "sf", ",", "\[Alpha]", ",", "\[Delta]", ",", "c", ",", 
            "sm"}], "]"}], "/.", "testParameters4"}], ",", 
         RowBox[{
          RowBox[{"funcNoDrive", "[", 
           RowBox[{
           "h", ",", "sf", ",", "\[Alpha]", ",", "\[Delta]", ",", "c", ",", 
            "sm"}], "]"}], "/.", "testParameters4"}]}], "}"}], ",", 
       RowBox[{"{", 
        RowBox[{"b", ",", "1", ",", "5"}], "}"}], ",", 
       RowBox[{"ImageSize", "\[Rule]", "Medium"}]}], "]"}], ",", 
     RowBox[{"{", 
      RowBox[{
       RowBox[{
       "Text", "[", 
        "\"\<birth rate (\!\(\*StyleBox[\"b\",FontSlant->\"Italic\"]\))\>\"", 
        "]"}], ",", 
       RowBox[{"Rotate", "[", 
        RowBox[{
         RowBox[{
         "Text", "[", 
          "\"\<population size \
(\!\(\*StyleBox[\"N\",FontSlant->\"Italic\"]\))\>\"", "]"}], ",", 
         RowBox[{"90", " ", "Degree"}]}], "]"}]}], "}"}], ",", 
     RowBox[{"{", 
      RowBox[{"Bottom", ",", "Left"}], "}"}]}], "]"}], ",", 
   RowBox[{"Placed", "[", 
    RowBox[{
     RowBox[{"LineLegend", "[", 
      RowBox[{
       RowBox[{"{", 
        RowBox[{
         RowBox[{
          RowBox[{"ColorData", "[", "97", "]"}], "[", "1", "]"}], ",", 
         RowBox[{
          RowBox[{"ColorData", "[", "97", "]"}], "[", "2", "]"}]}], "}"}], 
       ",", 
       RowBox[{"{", 
        RowBox[{
        "\"\<strong drive at equilibrium\>\"", ",", "\"\<wild-type\>\""}], 
        "}"}]}], "]"}], ",", "Below"}], "]"}]}], 
  "]"}], "\[IndentingNewLine]"}], "Input",
 CellChangeTimes->CompressedData["
1:eJwdzk8og3EAxvFlQkS25U/msBFpLe0wWim8ajWEJkOhNDayJE62GjGylvlX
5k/Jm5XeHR28iUhjjVoya6VQ7CDZz7BalMSeHZ4+x+cr1o606pI4HI4wPjjx
ZNsL5RFKwlOzcOGl7xpWR6r88MZqVkfiei1WDdzPMDjlQkI1M/kM1BgKO6ZE
hNoU0V3wlWszwnreZULFEuuAz4KrDcgvkdLeRkJZ72UJybHvHJpU/R7I6nLe
y5oIRasuPmARu5IcayEUT8mkwtk3XxZUrNSKYW6ELoVGQbEUSqJ1lbCtPNAA
Q9zJdvgrOuuB2zPLergQuzPAmqhrFJ46fGYo1/dNw2/NjgUebFnmE3+frB0e
ZWpdcDwjzMLg3IkbmsYGPTDlWx+A6Z3DCf+OpYd2LaHWCx7dcDXo9EJZN/8S
Zn+t3UKRZPEBkgomBHfToj+wdyCs9A/Fezy9KvgPJ77ntg==
  "],
 CellLabel->
  "In[155]:=",ExpressionUUID->"3285256e-8e37-4891-a7f9-3dde3bd5e3ba"],

Cell[BoxData[
 TemplateBox[{
   TemplateBox[{
     GraphicsBox[{{{{}, {}, 
         TagBox[{
           Directive[
            Opacity[1.], 
            RGBColor[0.368417, 0.506779, 0.709798], 
            AbsoluteThickness[1.6]], 
           LineBox[CompressedData["
1:eJwV13lcTc8bB/D2u51zW27aRamQKMmu+kgqUaIiUdq03HtKi1S2SoqKrKGF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            "]]}, Annotation[#, "Charting`Private`Tag$33786#1"]& ], 
         TagBox[{
           Directive[
            Opacity[1.], 
            RGBColor[0.880722, 0.611041, 0.142051], 
            AbsoluteThickness[1.6]], 
           LineBox[CompressedData["
1:eJwV13k8VVsUB/BrdsdzuCJTyJAppDSZfqlUKhUVSuqhKO65UoqXZlRCKpVE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            "]]}, 
          Annotation[#, "Charting`Private`Tag$33786#2"]& ]}}, {}}, {
      DisplayFunction -> Identity, Ticks -> {Automatic, Automatic}, 
       AxesOrigin -> {1., 0}, 
       FrameTicks -> {{Automatic, Automatic}, {Automatic, Automatic}}, 
       GridLines -> {None, None}, DisplayFunction -> Identity, 
       PlotRangePadding -> {{
          Scaled[0.02], 
          Scaled[0.02]}, {
          Scaled[0.05], 
          Scaled[0.05]}}, PlotRangeClipping -> True, ImagePadding -> All, 
       DisplayFunction -> Identity, AspectRatio -> 
       NCache[GoldenRatio^(-1), 0.6180339887498948], Axes -> {True, True}, 
       AxesLabel -> {None, None}, AxesOrigin -> {1., 0}, DisplayFunction :> 
       Identity, Frame -> {{False, False}, {False, False}}, 
       FrameLabel -> {{None, None}, {None, None}}, 
       FrameTicks -> {{Automatic, Automatic}, {Automatic, Automatic}}, 
       GridLines -> {None, None}, GridLinesStyle -> Directive[
         GrayLevel[0.5, 0.4]], ImageSize -> Medium, 
       Method -> {
        "DefaultBoundaryStyle" -> Automatic, 
         "DefaultGraphicsInteraction" -> {
          "Version" -> 1.2, "TrackMousePosition" -> {True, False}, 
           "Effects" -> {
            "Highlight" -> {"ratio" -> 2}, "HighlightPoint" -> {"ratio" -> 2},
              "Droplines" -> {
              "freeformCursorMode" -> True, 
               "placement" -> {"x" -> "All", "y" -> "None"}}}}, 
         "DefaultMeshStyle" -> AbsolutePointSize[6], "ScalingFunctions" -> 
         None, "CoordinatesToolOptions" -> {"DisplayFunction" -> ({
             (Identity[#]& )[
              Part[#, 1]], 
             (Identity[#]& )[
              Part[#, 2]]}& ), "CopiedValueFunction" -> ({
             (Identity[#]& )[
              Part[#, 1]], 
             (Identity[#]& )[
              Part[#, 2]]}& )}}, 
       PlotRange -> {{1, 5}, {0., 249.99999821721505`}}, PlotRangeClipping -> 
       True, PlotRangePadding -> {{
          Scaled[0.02], 
          Scaled[0.02]}, {
          Scaled[0.02], 
          Scaled[0.02]}}, Ticks -> {Automatic, Automatic}}], 
     InterpretationBox[
      Cell[
       BoxData[
        FormBox[
        "\"birth rate (\\!\\(\\*StyleBox[\\\"b\\\",FontSlant->\\\"Italic\\\"]\
\\))\"", TextForm]], "InlineText"], 
      Text["birth rate (\!\(\*StyleBox[\"b\",FontSlant->\"Italic\"]\))"]], 
     RotationBox[
      InterpretationBox[
       Cell[
        BoxData[
         FormBox[
         "\"population size \
(\\!\\(\\*StyleBox[\\\"N\\\",FontSlant->\\\"Italic\\\"]\\))\"", TextForm]], 
        "InlineText"], 
       Text[
       "population size (\!\(\*StyleBox[\"N\",FontSlant->\"Italic\"]\))"]], 
      BoxRotation -> 1.5707963267948966`]}, "Labeled", 
    DisplayFunction -> (GridBox[{{
        ItemBox[#3, DefaultBaseStyle -> "LabeledLabel"], 
        TagBox[
         ItemBox[
          PaneBox[
           TagBox[#, "SkipImageSizeLevel"], Alignment -> {Center, Baseline}, 
           BaselinePosition -> Baseline], DefaultBaseStyle -> "Labeled"], 
         "SkipImageSizeLevel"]}, {"", 
        ItemBox[#2, DefaultBaseStyle -> "LabeledLabel"]}}, 
      GridBoxAlignment -> {"Columns" -> {{Center}}, "Rows" -> {{Center}}}, 
      AutoDelete -> False, 
      GridBoxItemSize -> {
       "Columns" -> {{Automatic}}, "Rows" -> {{Automatic}}}, 
      BaselinePosition -> {1, 2}]& ), 
    InterpretationFunction -> (RowBox[{"Labeled", "[", 
       RowBox[{#, ",", 
         RowBox[{"{", 
           RowBox[{#2, ",", #3}], "}"}], ",", 
         RowBox[{"{", 
           RowBox[{"Bottom", ",", "Left"}], "}"}]}], "]"}]& )], 
   TemplateBox[{"\"strong drive at equilibrium\"", "\"wild-type\""}, 
    "LineLegend", DisplayFunction -> (StyleBox[
      StyleBox[
       PaneBox[
        TagBox[
         GridBox[{{
            TagBox[
             GridBox[{{
                GraphicsBox[{{
                   Directive[
                    EdgeForm[
                    Directive[
                    Opacity[0.3], 
                    GrayLevel[0]]], 
                    PointSize[0.5], 
                    AbsoluteThickness[1.6], 
                    RGBColor[0.368417, 0.506779, 0.709798]], {
                    LineBox[{{0, 10}, {20, 10}}]}}, {
                   Directive[
                    EdgeForm[
                    Directive[
                    Opacity[0.3], 
                    GrayLevel[0]]], 
                    PointSize[0.5], 
                    AbsoluteThickness[1.6], 
                    RGBColor[0.368417, 0.506779, 0.709798]], {}}}, 
                 AspectRatio -> Full, ImageSize -> {20, 10}, PlotRangePadding -> 
                 None, ImagePadding -> Automatic, 
                 BaselinePosition -> (Scaled[0.1] -> Baseline)], #}, {
                GraphicsBox[{{
                   Directive[
                    EdgeForm[
                    Directive[
                    Opacity[0.3], 
                    GrayLevel[0]]], 
                    PointSize[0.5], 
                    AbsoluteThickness[1.6], 
                    RGBColor[0.880722, 0.611041, 0.142051]], {
                    LineBox[{{0, 10}, {20, 10}}]}}, {
                   Directive[
                    EdgeForm[
                    Directive[
                    Opacity[0.3], 
                    GrayLevel[0]]], 
                    PointSize[0.5], 
                    AbsoluteThickness[1.6], 
                    RGBColor[0.880722, 0.611041, 0.142051]], {}}}, 
                 AspectRatio -> Full, ImageSize -> {20, 10}, PlotRangePadding -> 
                 None, ImagePadding -> Automatic, 
                 BaselinePosition -> (Scaled[0.1] -> Baseline)], #2}}, 
              GridBoxAlignment -> {
               "Columns" -> {Center, Left}, "Rows" -> {{Baseline}}}, 
              AutoDelete -> False, 
              GridBoxDividers -> {
               "Columns" -> {{False}}, "Rows" -> {{False}}}, 
              GridBoxItemSize -> {"Columns" -> {{All}}, "Rows" -> {{All}}}, 
              GridBoxSpacings -> {"Columns" -> {{0.5}}, "Rows" -> {{0.8}}}], 
             "Grid"]}}, 
          GridBoxAlignment -> {"Columns" -> {{Left}}, "Rows" -> {{Top}}}, 
          AutoDelete -> False, 
          GridBoxItemSize -> {
           "Columns" -> {{Automatic}}, "Rows" -> {{Automatic}}}, 
          GridBoxSpacings -> {"Columns" -> {{1}}, "Rows" -> {{0}}}], "Grid"], 
        Alignment -> Left, AppearanceElements -> None, 
        ImageMargins -> {{5, 5}, {5, 5}}, ImageSizeAction -> "ResizeToFit"], 
       LineIndent -> 0, StripOnInput -> False], {FontFamily -> "Arial"}, 
      Background -> Automatic, StripOnInput -> False]& ), 
    InterpretationFunction :> (RowBox[{"LineLegend", "[", 
       RowBox[{
         RowBox[{"{", 
           RowBox[{
             InterpretationBox[
              ButtonBox[
               TooltipBox[
                GraphicsBox[{{
                   GrayLevel[0], 
                   RectangleBox[{0, 0}]}, {
                   GrayLevel[0], 
                   RectangleBox[{1, -1}]}, {
                   RGBColor[0.368417, 0.506779, 0.709798], 
                   RectangleBox[{0, -1}, {2, 1}]}}, DefaultBaseStyle -> 
                 "ColorSwatchGraphics", AspectRatio -> 1, Frame -> True, 
                 FrameStyle -> 
                 RGBColor[
                  0.24561133333333335`, 0.3378526666666667, 
                   0.4731986666666667], FrameTicks -> None, PlotRangePadding -> 
                 None, ImageSize -> 
                 Dynamic[{
                   Automatic, 1.35 CurrentValue["FontCapHeight"]/
                    AbsoluteCurrentValue[Magnification]}]], 
                StyleBox[
                 RowBox[{"RGBColor", "[", 
                   RowBox[{"0.368417`", ",", "0.506779`", ",", "0.709798`"}], 
                   "]"}], NumberMarks -> False]], Appearance -> None, 
               BaseStyle -> {}, BaselinePosition -> Baseline, 
               DefaultBaseStyle -> {}, ButtonFunction :> 
               With[{Typeset`box$ = EvaluationBox[]}, 
                 If[
                  Not[
                   AbsoluteCurrentValue["Deployed"]], 
                  SelectionMove[Typeset`box$, All, Expression]; 
                  FrontEnd`Private`$ColorSelectorInitialAlpha = 1; 
                  FrontEnd`Private`$ColorSelectorInitialColor = 
                   RGBColor[0.368417, 0.506779, 0.709798]; 
                  FrontEnd`Private`$ColorSelectorUseMakeBoxes = True; 
                  MathLink`CallFrontEnd[
                    FrontEnd`AttachCell[Typeset`box$, 
                    FrontEndResource["RGBColorValueSelector"], {
                    0, {Left, Bottom}}, {Left, Top}, 
                    "ClosingActions" -> {
                    "SelectionDeparture", "ParentChanged", 
                    "EvaluatorQuit"}]]]], BaseStyle -> Inherited, Evaluator -> 
               Automatic, Method -> "Preemptive"], 
              RGBColor[0.368417, 0.506779, 0.709798], Editable -> False, 
              Selectable -> False], ",", 
             InterpretationBox[
              ButtonBox[
               TooltipBox[
                GraphicsBox[{{
                   GrayLevel[0], 
                   RectangleBox[{0, 0}]}, {
                   GrayLevel[0], 
                   RectangleBox[{1, -1}]}, {
                   RGBColor[0.880722, 0.611041, 0.142051], 
                   RectangleBox[{0, -1}, {2, 1}]}}, DefaultBaseStyle -> 
                 "ColorSwatchGraphics", AspectRatio -> 1, Frame -> True, 
                 FrameStyle -> 
                 RGBColor[
                  0.587148, 0.40736066666666665`, 0.09470066666666668], 
                 FrameTicks -> None, PlotRangePadding -> None, ImageSize -> 
                 Dynamic[{
                   Automatic, 1.35 CurrentValue["FontCapHeight"]/
                    AbsoluteCurrentValue[Magnification]}]], 
                StyleBox[
                 RowBox[{"RGBColor", "[", 
                   RowBox[{"0.880722`", ",", "0.611041`", ",", "0.142051`"}], 
                   "]"}], NumberMarks -> False]], Appearance -> None, 
               BaseStyle -> {}, BaselinePosition -> Baseline, 
               DefaultBaseStyle -> {}, ButtonFunction :> 
               With[{Typeset`box$ = EvaluationBox[]}, 
                 If[
                  Not[
                   AbsoluteCurrentValue["Deployed"]], 
                  SelectionMove[Typeset`box$, All, Expression]; 
                  FrontEnd`Private`$ColorSelectorInitialAlpha = 1; 
                  FrontEnd`Private`$ColorSelectorInitialColor = 
                   RGBColor[0.880722, 0.611041, 0.142051]; 
                  FrontEnd`Private`$ColorSelectorUseMakeBoxes = True; 
                  MathLink`CallFrontEnd[
                    FrontEnd`AttachCell[Typeset`box$, 
                    FrontEndResource["RGBColorValueSelector"], {
                    0, {Left, Bottom}}, {Left, Top}, 
                    "ClosingActions" -> {
                    "SelectionDeparture", "ParentChanged", 
                    "EvaluatorQuit"}]]]], BaseStyle -> Inherited, Evaluator -> 
               Automatic, Method -> "Preemptive"], 
              RGBColor[0.880722, 0.611041, 0.142051], Editable -> False, 
              Selectable -> False]}], "}"}], ",", 
         RowBox[{"{", 
           RowBox[{#, ",", #2}], "}"}]}], "]"}]& ), Editable -> True]},
  "Legended",
  DisplayFunction->(GridBox[{{
      TagBox[
       ItemBox[
        PaneBox[
         TagBox[#, "SkipImageSizeLevel"], Alignment -> {Center, Baseline}, 
         BaselinePosition -> Baseline], DefaultBaseStyle -> "Labeled"], 
       "SkipImageSizeLevel"]}, {
      ItemBox[#2, DefaultBaseStyle -> "LabeledLabel"]}}, 
    GridBoxAlignment -> {"Columns" -> {{Center}}, "Rows" -> {{Center}}}, 
    AutoDelete -> False, GridBoxItemSize -> Automatic, 
    BaselinePosition -> {1, 1}]& ),
  Editable->True,
  InterpretationFunction->(RowBox[{"Legended", "[", 
     RowBox[{#, ",", 
       RowBox[{"Placed", "[", 
         RowBox[{#2, ",", "Below"}], "]"}]}], "]"}]& )]], "Output",
 CellChangeTimes->{{3.809767245982184*^9, 3.8097672803363743`*^9}, {
   3.809767314408186*^9, 3.8097673925111732`*^9}, {3.8140995128066273`*^9, 
   3.814099523984096*^9}, 3.814099936979662*^9},
 CellLabel->
  "Out[158]=",ExpressionUUID->"e78e311d-0b67-4e46-9fb6-dd5776971cfb"]
}, Open  ]]
}, Closed]]
}, Open  ]]
}, Open  ]]
},
WindowSize->{2048, 1035},
WindowMargins->{{Automatic, 1440}, {128, Automatic}},
Magnification->1.25,
FrontEndVersion->"12.1 for Mac OS X x86 (64-bit) (June 19, 2020)",
StyleDefinitions->"Default.nb",
ExpressionUUID->"2fee21e7-bf84-4d39-ba9e-ba71cf581606"
]
(* End of Notebook Content *)

(* Internal cache information *)
(*CellTagsOutline
CellTagsIndex->{}
*)
(*CellTagsIndex
CellTagsIndex->{}
*)
(*NotebookFileOutline
Notebook[{
Cell[CellGroupData[{
Cell[580, 22, 215, 5, 123, "Title",ExpressionUUID->"58a35404-c46a-4052-9e2f-f80fb751f245"],
Cell[798, 29, 205, 3, 67, "Subtitle",ExpressionUUID->"f5221977-f0e4-4d02-b1d3-4c538645795d"],
Cell[CellGroupData[{
Cell[1028, 36, 155, 3, 84, "Section",ExpressionUUID->"bb378928-8785-464b-8dab-ff0c8fc86d79"],
Cell[1186, 41, 789, 13, 102, "Text",ExpressionUUID->"250706bb-65ca-4f28-9a37-c5dbe94cb31c"]
}, Open  ]],
Cell[CellGroupData[{
Cell[2012, 59, 246, 4, 84, "Section",ExpressionUUID->"aeb6c763-514d-4534-b55f-747741b817bf"],
Cell[2261, 65, 3998, 117, 361, "Text",ExpressionUUID->"828fce9c-f220-461b-a1c4-86c2ce6a2967"],
Cell[6262, 184, 1103, 22, 78, "Text",ExpressionUUID->"e76d593a-4d23-4f5f-ad8e-d99b12ad4c5c"],
Cell[7368, 208, 672, 11, 73, "Text",ExpressionUUID->"5338ade1-35b0-4064-b584-adb8676ffbc3"],
Cell[CellGroupData[{
Cell[8065, 223, 2020, 55, 361, "Input",ExpressionUUID->"9abe40f9-a542-40db-9ab7-232aecfce29b"],
Cell[10088, 280, 1292, 37, 42, "Output",ExpressionUUID->"c65020f7-9f22-4315-bbe1-9e9069f32f3a"]
}, Open  ]],
Cell[11395, 320, 825, 14, 102, "Text",ExpressionUUID->"61e643bf-7a0e-480f-b76d-3f5721dbcd4e"],
Cell[12223, 336, 1808, 53, 231, "Text",ExpressionUUID->"dfd6dffa-4c34-45ab-9f7f-84084ef3187e"]
}, Closed]],
Cell[CellGroupData[{
Cell[14068, 394, 516, 14, 66, "Section",ExpressionUUID->"81f0ebfe-5142-4ce4-b8ce-04c3cfc325bd"],
Cell[CellGroupData[{
Cell[14609, 412, 268, 4, 67, "Subsection",ExpressionUUID->"f54b6010-1fa4-4a81-89ed-8ae2fa491764"],
Cell[14880, 418, 1452, 26, 79, "Text",ExpressionUUID->"07f8cc85-4729-4c30-963f-5b1fc2a2df3d"],
Cell[16335, 446, 5630, 165, 635, "Input",ExpressionUUID->"13fe85f1-f84b-4e25-8931-9e84a64c2555"],
Cell[21968, 613, 455, 9, 44, "Text",ExpressionUUID->"8793f039-88ed-449c-9e97-718fc109fc89"],
Cell[22426, 624, 1685, 49, 168, "Input",ExpressionUUID->"a558d4a4-e400-42df-8228-835e16bf1b63"]
}, Closed]],
Cell[CellGroupData[{
Cell[24148, 678, 275, 4, 46, "Subsection",ExpressionUUID->"bcd72549-4783-409a-91a8-53925885f135"],
Cell[24426, 684, 388, 7, 44, "Text",ExpressionUUID->"1f0ac6dc-e0b2-4ba8-9330-cf03c3724fd2"],
Cell[CellGroupData[{
Cell[24839, 695, 2723, 74, 94, "Input",ExpressionUUID->"910658e9-a5fb-493a-958f-d6083fe6bed0"],
Cell[27565, 771, 4781, 134, 122, "Output",ExpressionUUID->"e5c7bb73-f0b3-4b94-b5ca-81777d2d2601"]
}, Open  ]],
Cell[32361, 908, 198, 3, 44, "Text",ExpressionUUID->"54fcd6cb-f965-4ab9-9229-21cd96d2548d"],
Cell[32562, 913, 741, 19, 37, "Input",ExpressionUUID->"c12bf7b4-d788-41f0-ac61-5834ea6f32ee"],
Cell[33306, 934, 311, 6, 44, "Text",ExpressionUUID->"e4b6a7b9-c4d0-4c2f-87f2-5885b6ef6388"],
Cell[CellGroupData[{
Cell[33642, 944, 2332, 55, 116, "Input",ExpressionUUID->"dda5a460-53b5-4e26-96cd-e5f727801662"],
Cell[35977, 1001, 558, 16, 63, "Output",ExpressionUUID->"8e645cbd-7402-4923-abd8-7efdff2fe8be"],
Cell[36538, 1019, 1561, 48, 66, "Output",ExpressionUUID->"a2a2bb4b-622c-45fc-8f3b-750a56000383"],
Cell[38102, 1069, 341, 9, 63, "Output",ExpressionUUID->"52408ef0-1587-4b11-826a-e44fd2e43342"],
Cell[38446, 1080, 870, 26, 66, "Output",ExpressionUUID->"2737f089-1f46-4091-82c9-e6c0ef9d1724"]
}, Open  ]],
Cell[CellGroupData[{
Cell[39353, 1111, 503, 11, 37, "Input",ExpressionUUID->"5cda42b8-bccf-43dc-879a-f819080ecb0f"],
Cell[39859, 1124, 594, 15, 63, "Output",ExpressionUUID->"b4d65503-c533-4476-8416-1c091cd89955"]
}, Open  ]],
Cell[40468, 1142, 181, 3, 44, "Text",ExpressionUUID->"67e73bba-e5e1-4071-be83-336d710090d5"],
Cell[CellGroupData[{
Cell[40674, 1149, 535, 11, 37, "Input",ExpressionUUID->"148bc5eb-e7a0-4634-829e-7049146077ff"],
Cell[41212, 1162, 228, 4, 42, "Output",ExpressionUUID->"82c8f930-e172-4f1f-8ac5-f79ffcafe6a9"]
}, Open  ]],
Cell[41455, 1169, 322, 7, 44, "Text",ExpressionUUID->"2d5d1f78-b49c-41db-9774-0e1883730af2"],
Cell[CellGroupData[{
Cell[41802, 1180, 877, 21, 37, "Input",ExpressionUUID->"5c29a56a-0fdb-46b6-8c89-fd0040e7b572"],
Cell[42682, 1203, 510, 13, 62, "Output",ExpressionUUID->"5edc3b3f-1d19-45d0-90ad-a3bde678a056"]
}, Open  ]],
Cell[CellGroupData[{
Cell[43229, 1221, 1334, 27, 37, "Input",ExpressionUUID->"924ad8cd-c072-4ea4-a85f-7233e1ca5ec3"],
Cell[44566, 1250, 12948, 322, 236, "Output",ExpressionUUID->"82857bcf-4ac0-4a58-8783-eba9797880eb"]
}, Open  ]],
Cell[57529, 1575, 293, 7, 44, "Text",ExpressionUUID->"51be3e69-ddfc-4d22-ab0f-406020908356"],
Cell[CellGroupData[{
Cell[57847, 1586, 549, 11, 37, "Input",ExpressionUUID->"6ee650c9-96cf-45e6-8be5-906729a1e138"],
Cell[58399, 1599, 749, 14, 42, "Output",ExpressionUUID->"e0521479-690a-4ccd-a3bb-5acfdb2e8456"]
}, Open  ]],
Cell[CellGroupData[{
Cell[59185, 1618, 574, 12, 37, "Input",ExpressionUUID->"25536a4d-a783-4d71-869f-b50b867d83e9"],
Cell[59762, 1632, 9362, 254, 210, "Output",ExpressionUUID->"9f7e632d-da7a-408d-8078-19cf8682134f"]
}, Open  ]]
}, Closed]],
Cell[CellGroupData[{
Cell[69173, 1892, 159, 3, 46, "Subsection",ExpressionUUID->"45736a70-7990-499e-a832-0b9071784a70"],
Cell[69335, 1897, 531, 11, 102, "Text",ExpressionUUID->"ab51b2eb-4d40-4a0a-886b-7a2b5e6928c2"],
Cell[69869, 1910, 1003, 26, 64, "Input",ExpressionUUID->"4c6c6aea-6ce2-4e48-bf24-4d7f10b04578"],
Cell[70875, 1938, 241, 6, 44, "Text",ExpressionUUID->"afc02abd-9832-41ed-ac68-5c1d687df12d"],
Cell[CellGroupData[{
Cell[71141, 1948, 210, 3, 56, "Subsubsection",ExpressionUUID->"65608a1c-7b86-45e4-8704-d4cfa68e8141"],
Cell[71354, 1953, 538, 9, 73, "Text",ExpressionUUID->"ee1ba8bb-788e-478c-bee9-d535f8518781"],
Cell[CellGroupData[{
Cell[71917, 1966, 847, 23, 64, "Input",ExpressionUUID->"0296eb6b-c900-4934-aa04-032620399872"],
Cell[72767, 1991, 1458, 31, 59, "Output",ExpressionUUID->"637a97a0-6d35-4609-bb59-af5cbfa75158"]
}, Open  ]],
Cell[74240, 2025, 2531, 50, 164, "Text",ExpressionUUID->"0335147d-eb3f-47af-bc72-ab9e7a3e1f20"],
Cell[CellGroupData[{
Cell[76796, 2079, 1817, 48, 189, "Input",ExpressionUUID->"87f0de7d-0098-4f70-9f16-10470b21eded"],
Cell[78616, 2129, 1347, 33, 59, "Output",ExpressionUUID->"bf0bcb4e-ee9c-4221-9892-b8b157a60955"]
}, Open  ]],
Cell[79978, 2165, 218, 5, 44, "Text",ExpressionUUID->"c94e92f7-10e6-4e8f-8282-1077be6164b8"],
Cell[CellGroupData[{
Cell[80221, 2174, 361, 8, 37, "Input",ExpressionUUID->"8eb1f776-89ec-4765-9c17-11aa44331c0e"],
Cell[80585, 2184, 1149, 27, 59, "Output",ExpressionUUID->"32f8574f-7b26-4923-906e-d17d51203240"]
}, Open  ]],
Cell[81749, 2214, 1718, 41, 217, "Text",ExpressionUUID->"f7b0820d-78d7-4eca-aee7-bc99661078e5"],
Cell[CellGroupData[{
Cell[83492, 2259, 635, 15, 37, "Input",ExpressionUUID->"1fd3ad56-38b9-4951-ad80-c20f5f77b24c"],
Cell[84130, 2276, 654, 10, 42, "Output",ExpressionUUID->"b590d123-c81d-438a-bdc1-5566fea3d97c"]
}, Open  ]],
Cell[CellGroupData[{
Cell[84821, 2291, 519, 12, 37, "Input",ExpressionUUID->"83a4d73d-8803-4057-8aba-5ec72c36236e"],
Cell[85343, 2305, 660, 10, 42, "Output",ExpressionUUID->"8ed0a354-8746-43d1-b6dd-ad6a4d707b99"]
}, Open  ]],
Cell[86018, 2318, 253, 6, 44, "Text",ExpressionUUID->"79fa2b1e-ea4f-4428-931c-a4bf0561909a"],
Cell[CellGroupData[{
Cell[86296, 2328, 1004, 21, 37, "Input",ExpressionUUID->"3a0f72c0-e0a6-48fa-808f-a744cf23697c"],
Cell[87303, 2351, 1282, 24, 42, "Output",ExpressionUUID->"bde67e1d-76de-48fb-9d3a-2d59a62f2ed6"]
}, Open  ]],
Cell[88600, 2378, 194, 3, 44, "Text",ExpressionUUID->"6ff0c034-af49-4d51-8ea1-6d121d8be3fa"],
Cell[CellGroupData[{
Cell[88819, 2385, 690, 15, 37, "Input",ExpressionUUID->"40f9b588-14ff-4c1a-9ccc-73f918eb34ad"],
Cell[89512, 2402, 729, 13, 42, "Output",ExpressionUUID->"68c12345-b81e-4431-86b9-455d5c72272b"]
}, Open  ]],
Cell[90256, 2418, 200, 3, 44, "Text",ExpressionUUID->"4c9cab46-2e08-4808-a069-962cb6675d59"],
Cell[CellGroupData[{
Cell[90481, 2425, 1515, 40, 116, "Input",ExpressionUUID->"0f574c14-d9e6-44cc-904d-2c74f53dd191"],
Cell[91999, 2467, 252, 4, 62, "Output",ExpressionUUID->"e4688023-1827-4f6c-9abf-43a61c43edb6"]
}, Open  ]]
}, Closed]],
Cell[CellGroupData[{
Cell[92300, 2477, 207, 3, 45, "Subsubsection",ExpressionUUID->"5b8db823-0ef9-4c73-9c07-07e593c1b7a6"],
Cell[92510, 2482, 226, 5, 44, "Text",ExpressionUUID->"c52c3617-685e-46eb-9b2f-877de8aa1afe"],
Cell[CellGroupData[{
Cell[92761, 2491, 869, 22, 64, "Input",ExpressionUUID->"558514c0-c8c7-4b08-85da-37c3c8b66b22"],
Cell[93633, 2515, 2114, 50, 68, "Output",ExpressionUUID->"8d8881fb-8291-487b-a5a6-379638330ddb"]
}, Open  ]],
Cell[95762, 2568, 3071, 66, 200, "Text",ExpressionUUID->"f502fdb9-4698-4f13-8d48-f3fa156d4057"],
Cell[CellGroupData[{
Cell[98858, 2638, 984, 26, 94, "Input",ExpressionUUID->"287620a3-1950-4474-82ee-e473fcc4471f"],
Cell[99845, 2666, 1010, 31, 43, "Output",ExpressionUUID->"d10fe94e-1e64-45dd-870e-d886b0d227e2"],
Cell[100858, 2699, 843, 25, 43, "Output",ExpressionUUID->"aca2aabb-05fa-4b80-a5e9-35b3b4acdacf"]
}, Open  ]],
Cell[CellGroupData[{
Cell[101738, 2729, 2058, 56, 224, "Input",ExpressionUUID->"64b86bc6-8509-4160-829f-f613f2f9eaf1"],
Cell[103799, 2787, 901, 28, 63, "Output",ExpressionUUID->"162333d1-2c10-4dfd-9fbf-01c7a38de571"]
}, Open  ]],
Cell[104715, 2818, 269, 6, 44, "Text",ExpressionUUID->"e8c43d39-f3c1-4f65-a6e5-65b7a6edff55"],
Cell[CellGroupData[{
Cell[105009, 2828, 194, 4, 37, "Input",ExpressionUUID->"142bdcdf-af7e-43fa-ba2c-0e92873ac6b0"],
Cell[105206, 2834, 854, 29, 69, "Output",ExpressionUUID->"134c175d-0959-4cc6-bdeb-bf4444ea07d7"]
}, Open  ]],
Cell[106075, 2866, 1490, 41, 246, "Text",ExpressionUUID->"06a59374-4759-4b18-8760-3a177915ed7b"],
Cell[CellGroupData[{
Cell[107590, 2911, 2231, 59, 116, "Input",ExpressionUUID->"41130e31-cbdf-4c46-ac55-f8d101b3d310"],
Cell[109824, 2972, 1416, 30, 42, "Output",ExpressionUUID->"6c75f31b-dfe0-4fa3-9ac8-9e688e8c8f7a"],
Cell[111243, 3004, 1436, 31, 42, "Output",ExpressionUUID->"08f13fe8-fc94-4949-8b91-07b1c68142e7"],
Cell[112682, 3037, 996, 17, 42, "Output",ExpressionUUID->"0491f61e-d6aa-47e7-8695-a8c12ecb7d73"],
Cell[113681, 3056, 906, 13, 42, "Output",ExpressionUUID->"17d418be-521a-458c-83b1-1307031034d4"]
}, Open  ]],
Cell[114602, 3072, 653, 16, 102, "Text",ExpressionUUID->"063735c5-7e61-4dca-a210-f8d99284cccd"],
Cell[CellGroupData[{
Cell[115280, 3092, 632, 13, 37, "Input",ExpressionUUID->"380ac5ed-6f95-416e-8851-3243faf4b72f"],
Cell[115915, 3107, 880, 14, 42, "Output",ExpressionUUID->"e2302f22-85eb-49d9-9337-97f7397c4158"]
}, Open  ]],
Cell[116810, 3124, 181, 3, 44, "Text",ExpressionUUID->"f7e9882a-0ffb-4a86-bfbf-307fc7c9d3f8"],
Cell[CellGroupData[{
Cell[117016, 3131, 227, 4, 37, "Input",ExpressionUUID->"8f1858c4-46be-405e-83ee-273c6200485e"],
Cell[117246, 3137, 1175, 28, 42, "Output",ExpressionUUID->"f94689dc-dfbb-4116-83da-c274977a8eba"]
}, Open  ]],
Cell[118436, 3168, 216, 3, 44, "Text",ExpressionUUID->"4762f435-6f50-45e5-8df7-3111b78f1d1c"],
Cell[CellGroupData[{
Cell[118677, 3175, 575, 14, 37, "Input",ExpressionUUID->"10d509a4-52c8-40b6-ab2f-6a3b902947f5"],
Cell[119255, 3191, 688, 15, 42, "Output",ExpressionUUID->"35c31739-2f2c-4711-a1c5-c2fb49afbc1a"]
}, Open  ]],
Cell[119958, 3209, 203, 3, 44, "Text",ExpressionUUID->"d4e79e97-27aa-485d-b0b1-e42c374b7871"],
Cell[CellGroupData[{
Cell[120186, 3216, 1702, 47, 116, "Input",ExpressionUUID->"8f270242-0732-4f2d-a989-7c5240aa97d7"],
Cell[121891, 3265, 252, 4, 42, "Output",ExpressionUUID->"089eaa0c-dab0-4661-a53e-1bb4d9ec831b"]
}, Open  ]]
}, Closed]]
}, Closed]],
Cell[CellGroupData[{
Cell[122204, 3276, 171, 3, 46, "Subsection",ExpressionUUID->"9a6be050-7bae-4fba-872a-c3b23864a3d4"],
Cell[122378, 3281, 422, 10, 102, "Text",ExpressionUUID->"898e4dc1-f2fd-4899-993f-9f68889ab0c3"],
Cell[CellGroupData[{
Cell[122825, 3295, 1124, 31, 90, "Input",ExpressionUUID->"d19ef520-56e2-44f9-b45d-912a2d6aad5f"],
Cell[123952, 3328, 300, 5, 42, "Output",ExpressionUUID->"eda60f85-e96a-4628-be8d-56276c90e672"],
Cell[124255, 3335, 1917, 57, 69, "Output",ExpressionUUID->"01e09ee8-7e1b-43e6-bc8b-47511622bb21"],
Cell[126175, 3394, 492, 13, 62, "Output",ExpressionUUID->"10b4c6cc-e760-40f5-ac20-e8c908e82d5d"]
}, Open  ]],
Cell[126682, 3410, 431, 8, 73, "Text",ExpressionUUID->"9d14e7a2-85f4-43a8-8581-76615094fac3"],
Cell[CellGroupData[{
Cell[127138, 3422, 1204, 28, 64, "Input",ExpressionUUID->"23732140-3772-4d4d-91fd-476a4d7c84c1"],
Cell[128345, 3452, 432, 7, 42, "Output",ExpressionUUID->"024359bc-dfdb-470c-84ae-17b8c0f8d554"],
Cell[128780, 3461, 430, 7, 42, "Output",ExpressionUUID->"ce2db405-f0a4-4801-90f9-7ae419084011"]
}, Open  ]],
Cell[CellGroupData[{
Cell[129247, 3473, 23448, 608, 776, "Input",ExpressionUUID->"3285256e-8e37-4891-a7f9-3dde3bd5e3ba"],
Cell[152698, 4083, 24815, 478, 409, "Output",ExpressionUUID->"e78e311d-0b67-4e46-9fb6-dd5776971cfb"]
}, Open  ]]
}, Closed]]
}, Open  ]]
}, Open  ]]
}
]
*)

